/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_ARCH_D_
#define HCL_ARCH_D_

#include "Std_includes.h"

/** {\bf HCL\_FPArch\_d} is an interface class to the LAPACK DLAMCH,
    which determines the floating point characteristics of machine
    on which the code is running.  For example, the relative machine
    precision can be determined as follows:
    \begin{verbatim}
       HCL_FPArch_d Arch;  // Create a "floating point architecture" object
       double eps = Arch.eps;  // eps is now machine precision
    \end{verbatim}
    See the list of class members for other available parameters.

    For more information on these parameters, see

       LAPACK User's Guide (2nd edition) by Anderson, Bai, et al., 1995
       (SIAM)
*/

class HCL_FPArch_d
{
   public:

   /**@name Machine parameters */
      
   //@{
   /// {\bf eps} is the relative machine precision.
      double eps;
   /// {\bf sfmin} is the safe minimum (1/sfmin does not overflow).
      double sfmin;
   /// {\bf base} is the base of the floating point arithmetic.
      int base;
   /// {\bf prec} = eps*base.
      double prec;
   /// {\bf t} is the number of digits in the mantissa.
      int t;
   /// {\bf rnd} is 1 if rounding occurs in addition, 0 otherwise.
      int rnd;
   /// {\bf emin} is the minimum exponent before (gradual) underflow.
      int emin;
   /** {\bf rmin} is the underflow threshold: base raised to the power
        (emin-1). */
      double rmin;
   /// {\bf emax} is the largest exponent before overflow.
      int emax;
   /// {\bf rmax} is the overflow threshold: (base to the power emax)*(1-eps).
      double rmax;
   //@}

   /**@name Additional parameters
      The following constants provide reasonable choices for the values of 
      UNDERFLOW and OVERFLOW while also compensating for poor arithmetic in 
      the upper half of the exponent range on certain machines
      (such as Crays.) (REFERENCE: LAPCK routine DLABAD.) */
   //@{
   /** {\bf small} equals sfmin, unless LOG10(1/sfmin) is large, in
       which case {\bf small} is the square root of sfmin.*/
      double small;
   /** {\bf large} equals 1/sfmin, unless LOG10(1/sfmin) is large, in
       which case {\bf large} is the square root of 1/sfmin.*/
      double large;
   //@}

    HCL_FPArch_d();
   /// {\bf Write} displays the values of all parameters on the given stream.
    ostream & Write( ostream & str ) const;
};

#endif
