/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_LINEARSOLVER_D_H_
#define HCL_LINEARSOLVER_D_H_

#include "HCL_LinearOp_d.h"
#include "Table.h"

/** {\bf HCL\_LinearSolver\_d} is the base class for linear solvers, that is,
for objects representing algorithms for solving linear operator equations.
This class is intended to represent "abstract" algorithms, that is,
algorithms such as conjugate gradients which do not use the implementation
of the operator.  */
class HCL_LinearSolver_d: public HCL_Base
{
   public:

  // Default constructor.
      HCL_LinearSolver_d(){}
  // Destructor.
      virtual ~HCL_LinearSolver_d(){}

  /// Solve the equation Ax=b for x. Returns non-zero on error.
      virtual int Solve( const HCL_LinearOp_d & A,const HCL_Vector_d & b,
                                HCL_Vector_d & x ) const = 0;
   /** Access to the parameter table. Algorithmic parameters can
       be accessed via {\bf Parameters().GetValue( "NAME",val )}
       and changed via {\bf Parameters().PutValue( "NAME",val )} */
      virtual Table & Parameters() const = 0;
};

#endif
