/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#ifndef HCL_MATRIXOPS_D_H_
#define HCL_MATRIXOPS_D_H_

#include "HCL_LinearOp_d.h"
#include "mat_inc_d.H"

// Look-ahead declaration.
class HCL_UpperTriangularMatOp_d;

/** {\bf HCL\_SymMatOp\_d} makes a symmetric matrix into a linear operator
in the HCL sense.  Note that the matrix is an instance of the class
{\bf GenSymMatrix_d}, and it must be provided to the constructor.

The methods of this class are the methods of the base class
{\bf HCL\_LinearOp\_d}, and the documentation for the base class can be
consulted for more information. */

class HCL_SymMatOp_d: public HCL_LinearOp_d
{
   private:
      HCL_VectorSpace_d * dom;
      HCL_VectorSpace_d * range;
      int DeleteMat;
   public:

   /** {\bf Matrix} is a pointer to the underlying {\bf GenSymMatrix_d},
       which is the matrix representing the linear operator.  The matrix
       is public so that its methods are available.  For instance,
       a {\bf GenSymMatrix_d} knows how to compute its eigenvalues and
       eigenvectors through the method {\bf eig}. */
      GenSymMatrix_d * Matrix;

   // Default constructor---disabled.
      HCL_SymMatOp_d();
   // Copy constructor---disabled.
      HCL_SymMatOp_d( const HCL_SymMatOp_d & );
   /** Usual constructor; requires the matrix and the domain. (The range
       is assumed to equal the domain.)  Also, a flag must be passed to
       indicate whether the matrix is to be deleted upon destruction of
       the operator (nonzero = delete, zero = no delete). */
      HCL_SymMatOp_d( GenSymMatrix_d * AA,int flag,HCL_VectorSpace_d * d );
   /** Usual constructor; requires the matrix, the domain, and the range.
       Note that the dimensions of the domain and range must agree.
       Also, a flag must be passed to indicate whether the matrix is to be
       deleted upon destruction of the operator (nonzero = delete,
       zero = no delete). */
      HCL_SymMatOp_d( GenSymMatrix_d * AA,int flag,HCL_VectorSpace_d * d,
                      HCL_VectorSpace_d * r );
   // Destructor.
      ~HCL_SymMatOp_d();

  /** Domain space access. Returns a reference to the domain of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
  /** Range space access. Returns a reference to the range of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const { return *range; }
  /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;

   /// {\bf InvImage} computes the action of the inverse on y, giving x.
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /** {\bf Cholesky} creates the linear operator which is 
       represented by the (upper triangular) Cholesky factor of the
       underlying matrix.  The return value is NULL if the matrix
       is not positive definite and therefore does not have
       a Cholesky factorization. */
      HCL_UpperTriangularMatOp_d * Cholesky() const;
   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_GeneralMatOp\_d} makes a matrix into a linear operator
in the HCL sense.  Note that the matrix is an instance of the class
{\bf GeneralMatrix_d}, and it must be provided to the constructor.

The methods of this class are the methods of the base class
{\bf HCL\_LinearOp\_d}, and the documentation for the base class can be
consulted for more information. */

class HCL_GeneralMatOp_d: public HCL_LinearOp_d
{
   private:
      HCL_VectorSpace_d * dom;
      HCL_VectorSpace_d * range;
      int DeleteMat;
   public:


   /** {\bf Matrix} is a pointer to the underlying {\bf GeneralMatrix_d},
       which is the matrix representing the linear operator.  The matrix
       is public so that its methods are available. */
      GeneralMatrix_d * Matrix;

   // Default constructor---disabled.
      HCL_GeneralMatOp_d();
   // Copy constructor---disabled.
      HCL_GeneralMatOp_d( const HCL_GeneralMatOp_d & );
   /** Usual constructor; requires the matrix, the domain, and the range.
       Also, a flag must be passed to indicate whether the matrix is to be
       deleted upon destruction of the operator (nonzero = delete,
       zero = no delete). */
      HCL_GeneralMatOp_d( GeneralMatrix_d * AA,int flag,HCL_VectorSpace_d * d,
                          HCL_VectorSpace_d * r );
   // Destructor.
      ~HCL_GeneralMatOp_d();

  /** Domain space access. Returns a reference to the domain of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
  /** Range space access. Returns a reference to the range of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const { return *range; }
  /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;

   /** {\bf InvImage} computes the action of the inverse on y, giving x.
       This method will return an error if the matrix is not square. */
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y.  This method will return an error if the matrix is not
       square. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;
   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_SymSparseRCSMatOp\_d} makes a sparse symmetric matrix into a
linear operator in the HCL sense.  Note that the matrix is an instance of
the class {\bf SymSparseRCSMatrix_d}, which implements a sparse matrix
in Row Compressed Storage, and it must be provided to the constructor.

The methods of this class are the methods of the base class
{\bf HCL\_LinearOp\_d}, and the documentation for the base class can be
consulted for more information. */

class HCL_SymSparseRCSMatOp_d: public HCL_LinearOp_d
{
   private:
      HCL_VectorSpace_d * dom;
      HCL_VectorSpace_d * range;
      int DeleteMat;
   public:

   /** {\bf Matrix} is a pointer to the underlying {\bf SymSparseRCSMatrix_d},
       which is the matrix representing the linear operator.  The matrix
       is public so that its methods are available. */
      SymSparseRCSMatrix_d * Matrix;

   // Default constructor---disabled.
      HCL_SymSparseRCSMatOp_d();
   // Copy constructor---disabled.
      HCL_SymSparseRCSMatOp_d( const HCL_SymSparseRCSMatOp_d & );
   /** Usual constructor; requires the matrix and the domain. (The range
       is assumed to equal the domain.)  Also, a flag must be passed to
       indicate whether the matrix is to be deleted upon destruction of
       the operator (nonzero = delete, zero = no delete). */
      HCL_SymSparseRCSMatOp_d( SymSparseRCSMatrix_d * AA,int flag,
                               HCL_VectorSpace_d * d );
   /** Usual constructor; requires the matrix, the domain, and the range.
       Note that the dimensions of the domain and range must agree.
       Also, a flag must be passed to indicate whether the matrix is to be
       deleted upon destruction of the operator (nonzero = delete,
       zero = no delete). */
      HCL_SymSparseRCSMatOp_d( SymSparseRCSMatrix_d * AA,int flag,
                               HCL_VectorSpace_d * d,
                               HCL_VectorSpace_d * r );
   // Destructor.
      ~HCL_SymSparseRCSMatOp_d();

  /** Domain space access. Returns a reference to the domain of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
  /** Range space access. Returns a reference to the range of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const { return *range; }
  /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;

   /// {\bf InvImage} computes the action of the inverse on y, giving x.
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;
   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_SparseRCSMatOp\_d} makes a sparse matrix into a linear operator
in the HCL sense.  Note that the matrix is an instance of the class
{\bf SparseRCSMatrix_d}, and it must be provided to the constructor.

The methods of this class are the methods of the base class
{\bf HCL\_LinearOp\_d}, and the documentation for the base class can be
consulted for more information. */

class HCL_SparseRCSMatOp_d: public HCL_LinearOp_d
{
   private:
      HCL_VectorSpace_d * dom;
      HCL_VectorSpace_d * range;
      int DeleteMat;
   public:

   /** {\bf Matrix} is a pointer to the underlying {\bf SparseRCSMatrix_d},
       which is the matrix representing the linear operator.  The matrix
       is public so that its methods are available. */
      SparseRCSMatrix_d * Matrix;

   // Default constructor---disabled.
      HCL_SparseRCSMatOp_d();
   // Copy constructor---disabled.
      HCL_SparseRCSMatOp_d( const HCL_SparseRCSMatOp_d & );
   /** Usual constructor; requires the matrix, the domain, and the range.
       Also, a flag must be passed to indicate whether the matrix is to be
       deleted upon destruction of the operator (nonzero = delete,
       zero = no delete). */
      HCL_SparseRCSMatOp_d( SparseRCSMatrix_d * AA,int flag,
                               HCL_VectorSpace_d * d,
                               HCL_VectorSpace_d * r );
   // Destructor.
      ~HCL_SparseRCSMatOp_d();

   /** Domain space access. Returns a reference to the domain of
       the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
   /** Range space access. Returns a reference to the range of
       the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const { return *range; }
   /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;

   /** {\bf InvImage} computes the action of the inverse on y, giving x.
       This method will return an error if the matrix is not square. */
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y.  This method will return an error if the matrix is not
       square. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;
   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_UpperTriangularMatOp\_d} makes an upper triangular matrix into a
linear operator in the HCL sense.  Note that the matrix is an instance of the
class {\bf UpperTriangle_d}, and it must be provided to the constructor.

The methods of this class are the methods of the base class
{\bf HCL\_LinearOp\_d}, and the documentation for the base class can be
consulted for more information. */

class HCL_UpperTriangularMatOp_d: public HCL_LinearOp_d
{
   private:
      HCL_VectorSpace_d * dom;
      HCL_VectorSpace_d * range;
      int DeleteMat;
   public:


   /** {\bf Matrix} is a pointer to the underlying {\bf UpperTriangle_d},
       which is the matrix representing the linear operator.  The matrix
       is public so that its methods are available. */
      UpperTriangle_d * Matrix;

   // Default constructor---disabled.
      HCL_UpperTriangularMatOp_d();
   // Copy constructor---disabled.
      HCL_UpperTriangularMatOp_d( const HCL_UpperTriangularMatOp_d & );
   /** Usual constructor; requires the matrix, the domain, and the range.
       Also, a flag must be passed to indicate whether the matrix is to be
       deleted upon destruction of the operator (nonzero = delete,
       zero = no delete). */
      HCL_UpperTriangularMatOp_d( UpperTriangle_d * AA,int flag,
                         HCL_VectorSpace_d * d,HCL_VectorSpace_d * r );
   // Destructor.
      ~HCL_UpperTriangularMatOp_d();

  /** Domain space access. Returns a reference to the domain of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
  /** Range space access. Returns a reference to the range of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const { return *range; }
  /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;

   /** {\bf InvImage} computes the action of the inverse on y, giving x.
       This method will return an error if the matrix is not square. */
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y.  This method will return an error if the matrix is not
       square. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;
   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_LowerTriangularMatOp\_d} makes an lower triangular matrix into a
linear operator in the HCL sense.  Note that the matrix is an instance of the
class {\bf LowerTriangle_d}, and it must be provided to the constructor.

The methods of this class are the methods of the base class
{\bf HCL\_LinearOp\_d}, and the documentation for the base class can be
consulted for more information. */

class HCL_LowerTriangularMatOp_d: public HCL_LinearOp_d
{
   private:
      HCL_VectorSpace_d * dom;
      HCL_VectorSpace_d * range;
      int DeleteMat;
   public:


   /** {\bf Matrix} is a pointer to the underlying {\bf LowerTriangle_d},
       which is the matrix representing the linear operator.  The matrix
       is public so that its methods are available. */
      LowerTriangle_d * Matrix;

   // Default constructor---disabled.
      HCL_LowerTriangularMatOp_d();
   // Copy constructor---disabled.
      HCL_LowerTriangularMatOp_d( const HCL_LowerTriangularMatOp_d & );
   /** Usual constructor; requires the matrix, the domain, and the range.
       Also, a flag must be passed to indicate whether the matrix is to be
       deleted upon destruction of the operator (nonzero = delete,
       zero = no delete). */
      HCL_LowerTriangularMatOp_d( LowerTriangle_d * AA,int flag,
                         HCL_VectorSpace_d * d,HCL_VectorSpace_d * r );
   // Destructor.
      ~HCL_LowerTriangularMatOp_d();

  /** Domain space access. Returns a reference to the domain of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
  /** Range space access. Returns a reference to the range of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const { return *range; }
  /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;

   /** {\bf InvImage} computes the action of the inverse on y, giving x.
       This method will return an error if the matrix is not square. */
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y.  This method will return an error if the matrix is not
       square. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;
   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};

#endif
