/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#ifndef HCL_RNRECORDER_D_H_
#define HCL_RNRECORDER_D_H_

#include "HCL_Rn_d.h"
#include "mat_inc_d.H"
#include "HCL_Recorder_d.H"

/** HCL\_RnRecorder Object. Implements the HCL\_Recorder interface for
    the Rn vector class. */

class HCL_RnRecorder_d: public HCL_Recorder_d
{
   private:
      int maxnv;
      int nv;
      int ptr;
      int n;
      HCL_RnSpace_d * sp;
      GeneralMatrix_d * V;
      char * filename;
   public:

   // Default constructor---disabled.
      HCL_RnRecorder_d();
   // Copy constructor---disabled.
      HCL_RnRecorder_d( const HCL_RnRecorder_d & );
   /** Usual constructor; needs the dimension of the vector space and the
       maximum number of vectors to be recorded.  Optionally, the name of
       a file in which to store the vectors upon destruction can be
       included. */
      HCL_RnRecorder_d( int dim,int m,char * fname=NULL );
   // Destructor.
      ~HCL_RnRecorder_d();

   /// Record---store input vector.
      virtual void Record( const HCL_Vector_d & Input );
   /// Rewind---go to first vector recorded.
      virtual void Rewind() const;
   /// Playback---copy current vector onto the input parameter.
      virtual void PlayBack( HCL_Vector_d & Output ) const;
   /// Next---advance to next recorded vector---returns 0 when at end.
      virtual int Next() const;
   /// Number---how many vectors currently stored.
      virtual int Number() const;
   /// Erase---deletes currently stored vectors.
      virtual void Erase();
   /// Save to a file.
      void Save( char * fname ) const;
    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif

