/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include <iostream.h>
#include <iomanip.h>
#include <stdio.h>
#include <float.h>
#include "maxmin_d.h"
#include <math.h>
#include <stdlib.h>
#include "Table.h"

#include "lapack.h"

#include "HCL_Vector_d.h"
#include "HCL_LinearOp_d.h"
#include "HCL_ItEigSolvers_d.h"

// Implicitly Restarted Arnoldi method


// Based on the Fortran code in ARPACK written by:
//    Danny Sorensen               Phuong Vu
//    Richard Lehoucq              CRPC / Rice University
//    Dept. of Computational &     Houston, Texas
//    Applied Mathematics
//    Rice University
//    Houston, Texas


//
// Prototypes for the ARPACK functions we need
//

#ifdef SINGLE_PRECISION

extern "C" void F77NAME(sneigh)(float*,int*,float*,int*,float*,float*,
			float*,float*,int*,float*,int*);
extern "C" void F77NAME(sngets)(int*,char*,int*,int*,float*,float*,float*,
			float*,float*);
extern "C" void F77NAME(snconv)(int*,float*,float*,float*,float*,int*);
extern "C" void F77NAME(snapps2)(int*,int*,float*,float*,float*,int*,
			float*,int*,float*);
extern "C" void F77NAME(ssortc)(char*,int*,int*,float*,float*,float*);

extern "C" void F77NAME(initarpack)();
extern "C" void F77NAME(sneupd1)(int*,float*,float*,float*,
			char*,int*,float*,int*,int*,
			int*,int*,int*,float*,int*,float*,float*,float*,
			float*,int*,int*);
#else

extern "C" void F77NAME(dneigh)(double*,int*,double*,int*,double*,double*,
			double*,double*,int*,double*,int*);
extern "C" void F77NAME(dngets)(int*,char*,int*,int*,double*,double*,double*,
			double*,double*);
extern "C" void F77NAME(dnconv)(int*,double*,double*,double*,double*,int*);
extern "C" void F77NAME(dnapps2)(int*,int*,double*,double*,double*,int*,
			double*,int*,double*);
extern "C" void F77NAME(dsortc)(char*,int*,int*,double*,double*,double*);

extern "C" void F77NAME(initarpack)();
extern "C" void F77NAME(dneupd1)(int*,double*,double*,double*,
			char*,int*,double*,int*,int*,
			int*,int*,int*,double*,int*,double*,double*,double*,
			double*,int*,int*);
#endif

//
// constructor
//

HCL_IRArnoldi_d::HCL_IRArnoldi_d(char *fname) 
{
   ParamTable = new Table( fname, "ItEigSolver IRArnoldi", 20 );
   vsize = 0;
}

//
// destructor
//

// There should be some more stuff to delete here: eigenvectors, etc.?
HCL_IRArnoldi_d::~HCL_IRArnoldi_d()
{
   HCL_delete( ParamTable );

   Clean();
}

//
// Clean()
//
// Delete local variables in free-store
//
void
HCL_IRArnoldi_d::Clean()
{
   if (vsize > 0) {
      for (int i = 0; i < vsize; ++i)
         HCL_delete( V[i] );
      delete[] V;

      delete[] dr;
      delete[] di;

      vsize = 0;
   }
}

//
//  These define the external interfaces to the _Solve() routine.
//  See the header file for details.
//

int 
HCL_IRArnoldi_d::Solve( int &Nev, HCL_LinearOp_d &OP, 
	HCL_Vector_d &resid) 
{
   if( ParamTable->GetValue( "DispFlag", DispFlag))
      DispFlag = 0;
   if (DispFlag > 1) {
      cout << endl << "===========HCL_IRArnoldi_d::Solve===========" << endl;
      cout << "Solving A*x = lambda*x" << endl;
   }
   OP_str = "A*x";
   B_str = "I*x";
   return _Solve(Nev, resid, &OP, NULL, 0, Regular);
}

int 
HCL_IRArnoldi_d::Solve( int &Nev, HCL_LinearOp_d &OP, double sigmar, 
	HCL_Vector_d &resid)
{
   if( ParamTable->GetValue( "DispFlag", DispFlag))
      DispFlag = 0;
   if (DispFlag > 1) {
      cout << endl << "===========HCL_IRArnoldi_d::Solve===========" << endl;
      cout << "Solving A*x = lambda*x with a real shift of " << sigmar << endl;
   }
   OP_str = "Inv[A-sigma*I]*x";
   B_str = "I*x";
   return _Solve(Nev, resid, &OP, NULL, sigmar, ShiftI);
}

int 
HCL_IRArnoldi_d::Solve( int &Nev, HCL_LinearOp_d &OP, HCL_LinearOp_d &M,
	    HCL_Vector_d &resid) 
{
   if( ParamTable->GetValue( "DispFlag", DispFlag))
      DispFlag = 0;
   if (DispFlag > 1) {
      cout << endl << "===========HCL_IRArnoldi_d::Solve===========" << endl;
      cout << "Solving A*x = lambda*M*x" << endl;
   }
   OP_str = "Inv[M]*A*x";
   B_str = "M*x";
   return _Solve(Nev, resid, &OP, &M, 0, Regular);
}

int 
HCL_IRArnoldi_d::Solve( int &Nev, HCL_LinearOp_d &OP, HCL_LinearOp_d &M, 
	    double sigmar, HCL_Vector_d &resid) 
{
   if (DispFlag > 1) {
      cout << endl << "===========HCL_IRArnoldi_d::Solve===========" << endl;
      cout << "Solving A*x = lambda*M*x using a real shift of " << sigmar<<endl;
   }
   OP_str = "Inv[A-sigma*I]*x";
   B_str = "M*x";
   return _Solve(Nev, resid, &OP, &M, sigmar, ShiftI);
}




//
// _Solve()
//
// solve the eigenvalue problem.
//

int 
HCL_IRArnoldi_d::_Solve(int &pnev, HCL_Vector_d &resid, 
      HCL_LinearOp_d *A, HCL_LinearOp_d *M, double sigmar,
      int ptype) 
{
   
   /*------------------------------------%
   | Get the machine dependent constant. |
   %------------------------------------*/
#ifdef SINGLE_PRECISION
   float eps23 = F77NAME(slamch)("Epsilon-Machine");
#else
   double eps23 = F77NAME(dlamch)("Epsilon-Machine");
#endif
   eps23 = pow(eps23, 2.0 / 3.0);

   Clean();
   Initialize();

   type = ptype;

   ishift = 1;

   nev = pnev;
   dom = &A->Domain();

   //
   // Check the arguments
   //

   if (*dom != resid.Space() || *dom != A->Range() || (M && 
		     (*dom != M->Domain() || *dom != M->Range()))) {
      cout << "HCL_IRArnoldi_d::Solve: Error: Incompatible vector spaces "
	      "in parameters." << endl;
      exit(1);
   }

   ncv = (ncv > 0)? ncv: 2*nev+2;  // use ncv=2*nev+2 as a reasonable default
   if ((ncv-nev < 2)) {
      ncv = nev+2;
      cout << "HCL_IRArnoldi_d::Solve: Increasing ncv to minimum allowable "
	      "value (" << ncv << ".)" << endl;
   }

   
   if (DispFlag > 1) {
      cout << "Number of wanted eigenvalues: " << nev << endl;;
      cout << "Dimension of Krylov supspace: " << ncv << endl;
      cout << "Eigenvalue selection criteria: " << Which << endl;
      cout << "Desired error bound: " << Tol << endl;
      cout << "Maximum number of iterations: " << MaxIter << endl;
      if (EigVectors)
         cout << "Will compute eigenvectors" << endl;
      else if (EigVectors  || SchurVectors)
         cout << "Will compute a Schur basis" << endl;
      else
	 cout << "No basis of the eigenspace will be computed" << endl;
      cout << "============================================" << endl;
   }
      
   //
   // Allocate the needed objects
   //

   int i;

   vsize = ncv;
   V = new HCL_Vector_d *[ncv];
   for (i = 0; i < ncv; ++i)
      V[i] = (HCL_Vector_d *)dom->Member();
  
   tmpvec = dom->Member();
   tmpvec2 = dom->Member();

   Q = new GeneralMatrix_d(ncv, ncv);
   R = new GeneralMatrix_d(ncv, ncv);
   H = new GeneralMatrix_d(ncv, ncv);
   
   // zero-out H
   H->zero();

   // allocate all the double precision arrays in one call

   double *dbl_store = new double[15*ncv + 3*ncv*ncv];
   ritzr = dbl_store;
   ritzi = ritzr + ncv;
   bounds = ritzi + ncv;
   workl = bounds + ncv;
   workev = workl + 3*ncv*ncv + 6*ncv;
   heigr = workev + 3*ncv;
   heigi = heigr + ncv;
   hbds = heigi + ncv;
//   double *dbl_store_end = hbds + ncv;

   dr = new double[nev+4];
   di = new double[nev+4];

   select = new int[ncv];


   // zero out the workl vector
   for (i = 0; i < 3*ncv*ncv+6*ncv; ++i)  workl[i] = 0.0;

   // 
   // Set the ARPACK message variables
   //

   F77NAME(initarpack)();


   //
   // Carry out the Implicitly restarted Arnoldi Iteration. 
   //

   int res = IRArnoldi(A, M, &resid);

   if (nconv > 0) {
      
      //
      // Compute the eigenvalues and eigenvectors
      //

      res = ComputeEigPairs(sigmar, &resid, Q);
   }

   //
   // print timing information
   //
   
   if (DispFlag == 1) {
      cout << "HCL_Arnoldi_d: " << nconv << " of " << nev 
	   << " eigenvalues found (" << OP_count << " calls to the operator)" << endl;
   }
   if (DispFlag > 1) {
//      int saved_prec = cout.precision(4);
      char buff[64];
      cout << endl << "=======HCL_IRArnoldi_d::Solve Summary=======" << endl;
      cout << "Number of wanted 'converged' Ritz values: " 
	   << nconv << endl;
      cout << "Number of update iterations taken: " 
	   << iter << endl;
      sprintf(buff, "Number of user %s calls: ", OP_str);
      cout << buff 
	   << OP_count << endl;
      if (M != NULL) {
	 sprintf(buff, "Number of user %s calls: ", B_str);
	 cout << buff 
	      << B_count << endl;
      }
      
      cout << "Number of reorthogonalization steps: " 
	   << Reorthog_count << endl;

      sprintf(buff, "Total time in user %s calls: ", OP_str);
      cout << buff
	   << OP_time << endl;
      if (M != NULL) {
	 sprintf(buff, "Total time in user %s calls: ", B_str);
	 cout << buff
	      << B_time << endl;
      }
      cout << "Total time in Iterative Arnoldi algorithm: " 
	   << Arnoldi_time << endl;
      cout << "Total time in Arnoldi factorization: " 
	   << Factor_time << endl;
      cout << "Total time in reorthogonalization: " 
	   << Reorthog_time << endl;
      cout << "Total time in Hessenberg eig. problem: " 
	   << HessEig_time << endl;
      cout << "Total time in applying shifted QR: " 
	   << QR_time << endl;
      cout << "Total time in computing final Ritz pairs: " 
	   << EigPair_time << endl;
//      cout.precision(saved_prec);


      cout << "============================================" << endl << endl;
      
   }



   //
   // Cleanup and return
   //

   delete[] dbl_store;
   delete[] select;
   delete H;
   delete Q;
   delete R;
   HCL_delete( tmpvec );
   HCL_delete( tmpvec2 );
   cout.precision( OldDispPrecision );
   pnev = nconv; // return the number of eigenvalues actually computed in pnev

   TermCode = res;
   ParamTable->PutValue("TermCode", TermCode);

   return res;

} /* Solve */


int HCL_IRArnoldi_d::Eigenvalue( int i, double &eigr, double &eigi )
{
   if (vsize == 0 || i > nconv)
      return -1;
   
   eigr = dr[i-1];
   eigi = di[i-1];
   return Success;
}

int HCL_IRArnoldi_d::Eigenvalue( int i, double &eigr )
{
   if (vsize == 0 || i > nconv)
      return -1;
   
   eigr = dr[i-1];
   return Success;
}

//
// EigPair()
//
// Return the ith eigenvalue/vector.  Returns nonzero on error;
//
int 
HCL_IRArnoldi_d::EigPair(int i, double &eigr, double &eigi, HCL_Vector_d *vr,
	HCL_Vector_d *vi)
{


   if (vsize == 0 || i > nconv)
      return -1;
   
   eigr = dr[i-1];
   eigi = di[i-1];

   if (EigVectors || SchurVectors && vr) {
      // Find out if the ith eigenvector is part of a conjugate pair
      int iconj = 0;
      int k = 1;
      while (k < i) {
         if (iconj) 
	    iconj = 0;
         else {
	    if (di[k-1] != 0)
	       iconj = 1;
         }
         k++;
      }
   
      if (di[i-1] != 0) {
         // complex eigenvalue
	 if (!vi)
	    return -1;
         if (!iconj) {
	    // First vector of a conjugate pair
	    vr->Copy(*V[i-1]);
	    vi->Copy(*V[i]);
         }
         else {
	    // Second vector of a conjugate pair
	    vr->Copy(*V[i-2]);
	    vi->Copy(*V[i-1]);
	    vi->Mul(-1);
         }
      }
      else {
         // real eigenvalue
         vr->Copy(*V[i-1]);
         if (vi) 
	    vi->Zero();
      }
   }
   return Success;
}
	    
//
// EigPair()
//
// Return the ith eigenvalue/vector.  Returns nonzero on error;
//
int 
HCL_IRArnoldi_d::EigPair(int i, double &eigr, HCL_Vector_d *vr )
{

   if (vsize == 0 || i > nconv)
      return -1;
   
   eigr = dr[i-1];

   if (EigVectors || SchurVectors && vr) {
      vr->Copy(*V[i-1]);
   return Success;
   }
   else
      return ComputeEigVectsFailed;
}
	    
//
// Initialize()
//
// Read values from the parameter table
//

void
HCL_IRArnoldi_d::Initialize()
{
   if( ParamTable->GetValue( "DispPrecision",DispPrecision ) )
      DispPrecision = 6;
   OldDispPrecision = cout.precision( DispPrecision );
   if( ParamTable->GetValue( "SubspaceDim", ncv))
      ncv = 0;
   if( ParamTable->GetValue( "Which", Which))
      strcpy(Which, "LM");
   if( ParamTable->GetValue( "Tol", Tol))
      Tol = 0;
#ifdef SINGLE_PRECISION
   Tol = max(2.0*F77NAME(slamch)("Epsilon-Machine"), Tol);
#else
   Tol = max(2.0*F77NAME(dlamch)("Epsilon-Machine"), Tol);
#endif
   if( ParamTable->GetValue( "MaxItn", MaxIter))
      MaxIter = 100;
   if( ParamTable->GetValue( "EigVectors", EigVectors))
      EigVectors = 1;
   if( ParamTable->GetValue( "SchurVectors", SchurVectors))
      SchurVectors = 1;
   if( ParamTable->GetValue( "DispFlag", DispFlag))
      DispFlag = 0;
   
   TermCode = Success;

   // clear timing variables

   OP_time = 0;
   B_time = 0;
   Arnoldi_time = 0;
   Factor_time = 0;
   Reorthog_time = 0;
   HessEig_time = 0;
   QR_time = 0;
   EigPair_time = 0;

   OP_count = 0;
   B_count = 0;
   Reorthog_count = 0;
}




//
// IRArnoldi()
//
// Implicitly restarted Arnoldi algorithm
//

int
HCL_IRArnoldi_d::IRArnoldi(HCL_LinearOp_d * OP, HCL_LinearOp_d * B,
		  HCL_Vector_d *resid) 
{

   timer_d tme(&Arnoldi_time);

   int i, j, k;

   /*------------------------------------%
   | Get the machine dependent constant. |
   %------------------------------------*/

#ifdef SINGLE_PRECISION
   float eps23 = F77NAME(slamch)("Epsilon-Machine");
#else
   double eps23 = F77NAME(dlamch)("Epsilon-Machine");
#endif
   eps23 = pow(eps23, 2.0 / 3.0);

   int np = max( ncv - nev,0 );
   int nev0 = nev;
   int np0 = np;

   /*------------------------------------%
   | kplusp is the bound on the largest  |
   |        Lanczos factorization built. |
   | nconv is the current number of      |
   |        "converged" eigenvlues.      |
   | iter is the counter on the current  |
   |      iteration step.                |
   %------------------------------------*/

   kplusp = nev + np;
   nconv  = 0;
   iter   = 0;


   
   if (DispFlag>3) {
      cout << endl << "===========HCL_IRArnoldi===========" << endl;
   }

   //
   // For the general problem, force resid into the range of OP
   //

   if (B != NULL) {
      if (type == Regular) {
	 { 
	   timer_d tme(&OP_time);
           OP->Image(*resid, *tmpvec); 
	 }
	 OP_count++;

         resid->Copy(*tmpvec);
      }
      else {
	 {
	   timer_d tme(&B_time);
           B->Image(*resid, *tmpvec);
	 }
	 B_count++;
	 {
	   timer_d tme(&OP_time);
           OP->Image(*tmpvec, *resid);
	 }
	 OP_count++;
      }
   }



   /*-----------------------------%
   | Compute the B-norm of resid. |
   %-----------------------------*/

   if (B != NULL) {
      {
	timer_d tme(&B_time);
        B->Image(*resid, *tmpvec);
      }
      B_count++;
      rnorm = sqrt(abs(resid->Inner(*tmpvec)));
   }
   else {
      tmpvec->Copy(*resid);
      rnorm = resid->Norm();
   }

   if (rnorm == 0) {

      /*----------------------------------------%
      | The initial vector is zero. Error exit. |
      %----------------------------------------*/
 
      return ZeroInitialVector;
   }


   /*---------------------------------------------------------%
   | Compute the first NEV steps of the Arnoldi factorization |
   %---------------------------------------------------------*/

   int Info = 0;
   int res = Arnoldi(*OP, B, *H, V, 0, nev, Info, *resid, rnorm, tmpvec2, type);
   if (res != Success) {
      cout << "HCL_IRArnoldi_d::Solve: Initial Arnoldi call failed" << endl;
      return res;
   }


   /*-------------------------------------------------------------%
   |                                                              |
   |           M A I N  ARNOLDI  I T E R A T I O N  L O O P       |
   |           Each iteration implicitly restarts the Arnoldi     |
   |           factorization in place.                            |
   |                                                              |
   %-------------------------------------------------------------*/

   int ldh = H->numCols();
   int ldq = Q->numCols();
   int ierr;
   int numcnv;

   while(iter++ < MaxIter ) {
      if (DispFlag>2) {
         cout << "HCL_IRArnoldi_d::Solve: **** Start of major iteration number "
	      << iter << " ****" << endl;
      }
      
      /*----------------------------------------------------------%
      | Compute NP additional steps of the Arnoldi factorization. |
      | Adjust NP since NEV might have been updated by last call  |
      | to the shift application routine dnapps.                  |
      %----------------------------------------------------------*/

      np  = kplusp - nev;

      if (DispFlag > 3) {
         cout << "HCL_IRArnoldi_d::Solve: current Arnoldi factorization length: "
	      << nev << endl;
         cout << "HCL_IRArnoldi_d::Solve: extending Arnoldi factorization by: "
	      << np << endl;
      }
      

      /*---------------------------------------------------------%
      | Compute NP additional steps of the Arnoldi factorization |
      %---------------------------------------------------------*/
   
      int Info = 0;
      res = Arnoldi(*OP, B, *H, V, nev, np, Info, *resid, rnorm, tmpvec2, type);
      if (res != Success) {
         cout << "HCL_IRArnoldi_d::Solve: Arnoldi call failed" << endl;
         return res;
      }

      if (DispFlag > 3) {
	 cout << "HCL_IRArnoldi_d::Solve: Corresponding B-norm of the residual: "
	      << rnorm << endl;
      }

      /*-------------------------------------------------------%
      | Compute the eigenvalues and corresponding error bounds |
      | of the current upper Hessenberg matrix.                |
      %-------------------------------------------------------*/
      {
	 timer_d tme(&HessEig_time);

#ifdef SINGLE_PRECISION
         F77NAME(sneigh)(&rnorm, &kplusp, H->data, &ldh, ritzr, ritzi,
		      bounds, Q->data, &ldq, workl, &ierr);
#else
         F77NAME(dneigh)(&rnorm, &kplusp, H->data, &ldh, ritzr, ritzi,
		      bounds, Q->data, &ldq, workl, &ierr);
#endif
      }

      
      if (ierr != 0) {
	 cout << "HCL_IRArnoldi_d::Solve: dneigh call failed" << endl;
	 return LapackEigFailed;
      }

      /*---------------------------------------------------%
      | Make a copy of eigenvalues and corresponding error |
      | bounds obtained from dneigh.                       |
      %---------------------------------------------------*/

      for (k = 0; k < kplusp; ++k) {
	  workl[kplusp*kplusp+k] = ritzr[k];
	  workl[kplusp*kplusp+kplusp+k] = ritzi[k];
	  workl[kplusp*kplusp+2*kplusp+k] = bounds[k];
      }


      /*--------------------------------------------------%
      | Select the wanted Ritz values and their bounds    |
      | to be used in the convergence test.               |
      | The wanted part of the spectrum and corresponding |
      | error bounds are in the last NEV loc. of RITZR,   |
      | RITZI and BOUNDS respectively. The variables NEV  |
      | and NP may be updated if the NEV-th wanted Ritz   |
      | value has a non zero imaginary part. In this case |
      | NEV is increased by one and NP decreased by one.  |
      | NOTE: The last two arguments of dngets are no     |
      | longer used as of version 2.1.                    |
      %--------------------------------------------------*/

      nev = nev0;
      np = np0;
      numcnv = nev;

#ifdef SINGLE_PRECISION
      F77NAME(sngets)(&ishift, Which, &nev, &np, ritzr, ritzi,
		      bounds, NULL, NULL);
#else
      F77NAME(dngets)(&ishift, Which, &nev, &np, ritzr, ritzi,
		      bounds, NULL, NULL);
#endif
      
      if (DispFlag > 2) {
	 HCL_RnVector_d t3(nev, bounds+np);
	 for (k = 0; k < nev; ++k) {
#ifdef SINGLE_PRECISION
	    float temp = max(eps23, F77NAME(slapy2)( &ritzr[k+np], 
						&ritzi[k+np] ));
#else
	    double temp = max(eps23, F77NAME(dlapy2)( &ritzr[k+np], 
						&ritzi[k+np] ));
#endif
	    t3(k+1) /= temp;
	 }
         cout << "HCL_IRArnoldi_d::Solve: Error estimates:" << endl;
         t3.Write(cout);
      }
      
      if (nev == nev0+1) 
	 numcnv = nev0+1;
      
      /*------------------%
      | Convergence test. |
      %------------------*/

      for (k = 0; k < nev; ++k) {
	 workl[2*np+k] = bounds[np+k];
      }

#ifdef SINGLE_PRECISION
      F77NAME(snconv)(&nev, ritzr+np, ritzi+np, workl+2*np, &Tol, &nconv);
#else
      F77NAME(dnconv)(&nev, ritzr+np, ritzi+np, workl+2*np, &Tol, &nconv);
#endif
      
      if (DispFlag > 3) {

         cout << "HCL_IRArnoldi_d::Solve: nev = " << nev << " np = " << np
	      << " numcnv = " << numcnv << " nconv = " << nconv << endl;
         HCL_RnVector_d t1(kplusp, ritzr);
         cout << "HCL_IRArnoldi_d::Solve: real part of eigenvalues of H" << endl;
         t1.Write(cout);
      
         HCL_RnVector_d t2(kplusp, ritzi);
         cout << "HCL_IRArnoldi_d::Solve: imag part of eigenvalues of H" << endl; 
         t2.Write(cout);
      
         HCL_RnVector_d t3(kplusp, bounds);
         cout << "HCL_IRArnoldi_d::Solve: Ritz estimates of the current "
		 "NCV values" << endl;
         t3.Write(cout);
	 
      }
      

      /*--------------------------------------------------------%
      | Count the number of unwanted Ritz values that have zero |
      | Ritz estimates. If any Ritz estimates are equal to zero |
      | then a leading block of H of order equal to at least    |
      | the number of Ritz values with zero Ritz estimates has  |
      | split off. None of these Ritz values may be removed by  |
      | shifting. Decrease NP the number of shifts to apply. If |
      | no shifts may be applied, then prepare to exit          |
      %--------------------------------------------------------*/

      int nptemp = np;
      for (j = 0; j < nptemp; ++j) {
	 if (bounds[j] == 0) {
	    --np; ++nev;
         }
      }

      if (nconv >= numcnv || np == 0)
	 break;  // done!  exit main while loop

      
      if (nconv < numcnv && ishift == 1) {

         /*------------------------------------------------%
         | Do not have all the requested eigenvalues yet.  |
         | To prevent possible stagnation, adjust the size |
         | of NEV.                                         |
         %------------------------------------------------*/

	 int nevbef = nev;
	 nev += ((nconv<(int)(np/2))?nconv:(int)(np/2));

	 if (nev == 1 && kplusp >= 6)
	    nev = kplusp / 2;
	 else if (nev == 1 && kplusp > 3)
	    nev = 2;
	 
	 np = kplusp - nev;

         /*--------------------------------------%
         | If the size of NEV was just increased |
         | resort the eigenvalues.               |
         %--------------------------------------*/

         if (nevbef < nev) {
#ifdef SINGLE_PRECISION
	    F77NAME(sngets)(&ishift, Which, &nev, &np, ritzr, ritzi,
		   bounds, NULL, NULL);
#else
	    F77NAME(dngets)(&ishift, Which, &nev, &np, ritzr, ritzi,
		   bounds, NULL, NULL);
#endif
	 }
	 
	
      }
      if (DispFlag>2) {
	 cout << "HCL_IRArnoldi_d::Solve: no. of converged Ritz values at "
		 "this iter: " << nconv << endl;

	 if (DispFlag>3) {
	    cout << "NEV = " << nev << " NP = " << np << endl;

	    HCL_RnVector_d t1(nev, ritzr+np);
	    cout << "HCL_IRArnoldi_d::Solve: wanted Ritz values -- real part" 
		 << endl;
	    t1.Write(cout);
      
	    HCL_RnVector_d t2(nev, ritzi+np);
	    cout << "HCL_IRArnoldi_d::Solve: wanted Ritz values -- imag part" 
		 << endl; 
	    t2.Write(cout);
   
	    HCL_RnVector_d t3(nev, bounds+np);
	    cout << "HCL_IRArnoldi_d::Solve: Ritz estimates of the wanted "
		    "values" << endl;
	    t3.Write(cout);
	 }
      }

	 

      if (DispFlag>3) {
	 cout << "HCL_IRArnoldi_d::Solve: The number of shifts to apply: " 
	      << np << endl;
         
	 HCL_RnVector_d t1(np, ritzr);
         cout << "HCL_IRArnoldi_d::Solve: real part of the shifts: " << endl;
         t1.Write(cout);
      
         HCL_RnVector_d t2(np, ritzi);
         cout << "HCL_IRArnoldi_d::Solve: imag part of the shifts: " << endl; 
         t2.Write(cout);
      
         HCL_RnVector_d t3(np, bounds);
         cout << "HCL_IRArnoldi_d::Solve: Ritz estimates of the shifts: " << endl;
         t3.Write(cout);
      }


      /*--------------------------------------------------------%
      | Apply the NP implicit shifts by QR bulge chasing.       |
      | Each shift is applied to the whole upper Hessenberg     |
      | matrix H.                                               |
      | The first 2*N locations of WORKD are used as workspace. |
      %--------------------------------------------------------*/

      {
	 
	 timer_d tme(&QR_time);


#ifdef SINGLE_PRECISION
	 F77NAME(snapps2)(&nev, &np, ritzr, ritzi, H->data, &ldh,
			 Q->data, &ldq, workl);
#else
	 F77NAME(dnapps2)(&nev, &np, ritzr, ritzi, H->data, &ldh,
			 Q->data, &ldq, workl);
#endif
	 
	 /*------------------------------------------------%
	 | Compute the (nev+1)-st column of (V*Q) and      |
	 | temporarily store the result in WORKD(N+1:2*N). |
	 | This is needed in the residual update since we  |
	 | cannot GUARANTEE that the corresponding entry   |
	 | of H would be zero as in exact arithmetic.      |
	 %------------------------------------------------*/

	 if ((*H)(nev+1,nev) > 0) {
	    tmpvec->Mul((*Q)(1, nev+1), *V[0]);
	    for (i = 1; i < kplusp; ++i) {
	       tmpvec->AddScale((*Q)(i+1, nev+1), *V[i]);
	    }
	 }

	 /*---------------------------------------------------------%
	 | Compute column 1 to nev of (V*Q) in backward order       |
	 | taking advantage of the upper Hessenberg structure of Q. |
	 %---------------------------------------------------------*/

	 for(j = 1; j <=nev; ++j) {
	    tmpvec2->Mul((*Q)(1, nev-j+1), *V[0]);
	    for (i = 2; i <= kplusp-j+1; ++i) {
	       tmpvec2->AddScale((*Q)(i, nev-j+1), *V[i-1]);
	    }
	    V[kplusp-j]->Copy(*tmpvec2);
	 }


	 /*------------------------------------------------%
	 |  Move v(:,kplusp-nev+1:kplusp) into v(:,1:nev). |
	 %------------------------------------------------*/
    
	 for (i = 0; i < nev; ++i) {
	    V[i]->Copy(*V[kplusp-nev+i]);
	 }
    
	 /*-------------------------------------------------------------%
	 | Copy the (nev+1)-st column of (V*Q) in the appropriate place |
	 %-------------------------------------------------------------*/
    
	 if ((*H)(nev+1,nev) > 0) {
	    V[nev]->Copy(*tmpvec);
	 }
    
	 /*------------------------------------%
	 | Update the residual vector:         |
	 |    r <- sigmak*r + betak*v(:,nev+1) |
	 | where                               |
	 |    sigmak = (e_{kplusp}'*Q)*e_{nev} |
	 |    betak = e_{nev+1}'*H*e_{nev}     |
	 %------------------------------------*/
    
	 resid->Mul((*Q)(kplusp,nev));
	 if ((*H)(nev+1,nev) > 0) {
	    resid->AddScale((*H)(nev+1,nev), *V[nev]);
	 }

   //      cout << "AFTER SHIFTS:" << endl;
   //      for (i = 0; i < nev; ++i) {
   //	   cout << "V["<<i+1<<"] = " << endl;
   //	   V[i]->Write(cout);
   //      }
   //      cout << "resid = " << endl;
   //      resid->Write(cout);


	 if (DispFlag>4) {
	    cout << "HCL_IRArnoldi_d::Solve: updated Hessenberg matrix:" << endl;
	    cout << *H << endl;
	 }
      }
 

      /*--------------------------------------------%
      | Compute the B-norm of the updated residual. |
      | Keep B*RESID in TMPVEC to be used in        |
      | the first step of the next call to dnaitr.  |
      %--------------------------------------------*/

      if (B != NULL) {
	 {
	   timer_d tme(&B_time);
	   B->Image(*resid, *tmpvec);
	 }
	 B_count++;
	 rnorm = sqrt(abs(resid->Inner(*tmpvec)));
      }
      else {
	 rnorm = resid->Norm();
      }


      if (DispFlag>3) {
	 cout << "HCL_IRArnoldi_d::Solve: B-norm of residual for compressed "
		 "factorization: " << rnorm << endl;
	 
	 cout << "HCL_IRArnoldi_d::Solve: Compressed upper Hessenberg matrix H: "
	      << endl;
	 
	 for (i = 1; i <= nev; ++i) {
	    for (j = 1; j <= nev; ++j) {
	       cout << (*H)(i,j) << "," << endl;
	    }
	    cout << endl;
	 }
//	 H->Write(cout);
      }

   }
   /*--------------------------------------------------------------%
   |                                                               |
   |  E N D     O F     M A I N     I T E R A T I O N     L O O P  |
   |                                                               |
   %--------------------------------------------------------------*/

   if (DispFlag>5) {
      HCL_RnVector_d t1(kplusp, workl+kplusp*kplusp);
      cout << "HCL_IRArnoldi_d::Solve: Real part of the eig computed by _neigh: " 
	   << endl;
      t1.Write(cout);

      HCL_RnVector_d t2(kplusp, workl+kplusp*kplusp+kplusp);
      cout << "HCL_IRArnoldi_d::Solve: Imag part of the eig computed by _neigh: "
	   << endl;
      t2.Write(cout);

      HCL_RnVector_d t3(kplusp, workl+kplusp*kplusp+2*kplusp);
      cout << "HCL_IRArnoldi_d::Solve: Ritz estimates computed by _neigh:" 
	   << endl;
      t3.Write(cout);
   }
  
   /*-----------------------------------------------%
   | Prepare to exit. Put the converged Ritz values |
   | and corresponding bounds in RITZ(1:NCONV) and  |
   | BOUNDS(1:NCONV) respectively. Then sort. Be    |
   | careful when NCONV > NP                        |
   %-----------------------------------------------*/

   /*---------------------------------------------%
   | To be consistent with dngets, we first do a  |
   | pre-processing sort in order to keep complex |
   | conjugate pairs together.  This is similar   |
   | to the pre-processing sort used in dngets    |
   | except that the sort is done in the opposite |
   | order.                                       |
   %---------------------------------------------*/

   if (DispFlag>3) {
      HCL_RnVector_d t1(kplusp, ritzr);
      cout << "---------------------------------------------------" << endl;
      cout << "HCL_IRArnoldi_d::Solve: Before sorting" << endl;
      cout << "HCL_IRArnoldi_d::Solve: real part of the ritz values: " 
	   << endl;
      t1.Write(cout);

      HCL_RnVector_d t2(kplusp, ritzi);
      cout << "HCL_IRArnoldi_d::Solve: imaginary part of the ritz values: "
	   << endl;
      t2.Write(cout);

      HCL_RnVector_d t3(kplusp, bounds);
      cout << "HCL_IRArnoldi_d::Solve: ritz estimates:" << endl;
      t3.Write(cout);
      cout << "---------------------------------------------------" << endl;
   }

   char *wprime;

   if (!strncmp(Which,"LM",2)) wprime = "SR";
   if (!strncmp(Which,"SM",2)) wprime = "LR";
   if (!strncmp(Which,"LR",2)) wprime = "SM";
   if (!strncmp(Which,"SR",2)) wprime = "LM";
   if (!strncmp(Which,"LI",2)) wprime = "SM";
   if (!strncmp(Which,"SI",2)) wprime = "LM";

   int tmp = 1;
#ifdef SINGLE_PRECISION
   F77NAME(ssortc)(wprime, &tmp, &kplusp, ritzr, ritzi, bounds);
#else
   F77NAME(dsortc)(wprime, &tmp, &kplusp, ritzr, ritzi, bounds);
#endif
   
   if (DispFlag>3) {
      HCL_RnVector_d t1(kplusp, ritzr);
      cout << "---------------------------------------------------" << endl;
      cout << "HCL_IRArnoldi_d::Solve: Pre-processing step" << endl;
      cout << "HCL_IRArnoldi_d::Solve: Sorted real part of the ritz values: " 
	   << endl;
      t1.Write(cout);

      HCL_RnVector_d t2(kplusp, ritzi);
      cout << "HCL_IRArnoldi_d::Solve: Sorted imaginary part of the ritz values: "
	   << endl;
      t2.Write(cout);

      HCL_RnVector_d t3(kplusp, bounds);
      cout << "HCL_IRArnoldi_d::Solve: Sorted ritz estimates:" << endl;
      t3.Write(cout);
      cout << "---------------------------------------------------" << endl;
   }
   
   /*---------------------------------------------%
   | Now sort Ritz values so that converged Ritz  |
   | values appear within the first NEV locations |
   | of ritzr, ritzi and bounds, and the most     |
   | desired one appears at the front.            |
   %---------------------------------------------*/
   
   if (!strncmp(Which,"LM",2)) wprime = "SM";
   if (!strncmp(Which,"SM",2)) wprime = "LM";
   if (!strncmp(Which,"LR",2)) wprime = "SR";
   if (!strncmp(Which,"SR",2)) wprime = "LR";
   if (!strncmp(Which,"LI",2)) wprime = "SI";
   if (!strncmp(Which,"SI",2)) wprime = "LI";

   tmp = 1;
#ifdef SINGLE_PRECISION
   F77NAME(ssortc)(wprime, &tmp, &kplusp, ritzr, ritzi, bounds);
#else
   F77NAME(dsortc)(wprime, &tmp, &kplusp, ritzr, ritzi, bounds);
#endif

   if (DispFlag>3) {
      cout << "---------------------------------------------------" << endl;
      cout << "HCL_IRArnoldi_d::Solve: Moved converged Ritz value to "
              "the front" << endl;
      HCL_RnVector_d t1(kplusp, ritzr);
      cout << "HCL_IRArnoldi_d::Solve: Sorted real part of the ritz values: " 
	   << endl;
      t1.Write(cout);

      HCL_RnVector_d t2(kplusp, ritzi);
      cout << "HCL_IRArnoldi_d::Solve: Sorted imaginary part of the ritz values: "
	   << endl;
      t2.Write(cout);

      HCL_RnVector_d t3(kplusp, bounds);
      cout << "HCL_IRArnoldi_d::Solve: Sorted ritz estimates:" << endl;
      t3.Write(cout);
      cout << "---------------------------------------------------" << endl;
   }
   
   /*-------------------------------------------------%
   | Scale the Ritz estimate of each Ritz value       |
   | by 1 / max(eps23,magnitude of the Ritz value).   |
   %-------------------------------------------------*/

   for (j = 0; j < nev0; ++j) {
#ifdef SINGLE_PRECISION
      float temp = max(eps23,F77NAME(slapy2)(&ritzr[j],&ritzi[j]));
#else
      double temp = max(eps23,F77NAME(dlapy2)(&ritzr[j],&ritzi[j]));
#endif
      bounds[j] /= temp;
   }

   /*---------------------------------------------------%
   | Sort the Ritz values according to the scaled Ritz  |
   | estimates.  This will push all the converged ones  |
   | towards the front of ritzr, ritzi, bounds          |
   | (in the case when NCONV < NEV.)                    |
   %---------------------------------------------------*/

   wprime = "LR";
   tmp = 1;
#ifdef SINGLE_PRECISION
   F77NAME(ssortc)(wprime, &tmp, &nev0, bounds, ritzr, ritzi);
#else
   F77NAME(dsortc)(wprime, &tmp, &nev0, bounds, ritzr, ritzi);
#endif

   if (DispFlag>3) {
      cout << "---------------------------------------------------" << endl;
      HCL_RnVector_d t1(kplusp, ritzr);
      cout << "HCL_IRArnoldi_d::Solve: sorted by scale Ritz values" << endl;
      cout << "HCL_IRArnoldi_d::Solve: Sorted real part of the ritz values: " 
	   << endl;
      t1.Write(cout);

      HCL_RnVector_d t2(kplusp, ritzi);
      cout << "HCL_IRArnoldi_d::Solve: Sorted imaginary part of the ritz values: "
	   << endl;
      t2.Write(cout);

      HCL_RnVector_d t3(kplusp, bounds);
      cout << "HCL_IRArnoldi_d::Solve: Sorted ritz estimates:" << endl;
      t3.Write(cout);
      cout << "---------------------------------------------------" << endl;
   }
   
   /*---------------------------------------------%
   | Scale the Ritz estimate back to its original |
   | value.                                       |
   %---------------------------------------------*/

   for (j = 0; j < nev0; ++j) {
#ifdef SINGLE_PRECISION
      double temp = max(eps23,F77NAME(slapy2)(&ritzr[j],&ritzi[j]));
#else
      double temp = max(eps23,F77NAME(dlapy2)(&ritzr[j],&ritzi[j]));
#endif
      bounds[j] *= temp;
   }

   /*-----------------------------------------------%
   | Sort the converged Ritz values again so that   |
   | the "threshold" value appears at the front of  |
   | ritzr, ritzi and bound.                        |
   %-----------------------------------------------*/

   tmp = 1;
#ifdef SINGLE_PRECISION
   F77NAME(ssortc)(Which, &tmp, &nconv, ritzr, ritzi, bounds);
#else
   F77NAME(dsortc)(Which, &tmp, &nconv, ritzr, ritzi, bounds);
#endif

   if (DispFlag>3) {
      cout << "---------------------------------------------------" << endl;
      cout << "HCL_IRArnoldi_d::Solve: Moved threshold value to front" << endl;
      HCL_RnVector_d t1(kplusp, ritzr);
      cout << "HCL_IRArnoldi_d::Solve: Sorted real part of the ritz values: " 
	   << endl;
      t1.Write(cout);

      HCL_RnVector_d t2(kplusp, ritzi);
      cout << "HCL_IRArnoldi_d::Solve: Sorted imaginary part of the ritz values: "
	   << endl;
      t2.Write(cout);

      HCL_RnVector_d t3(kplusp, bounds);
      cout << "HCL_IRArnoldi_d::Solve: Sorted ritz estimates:" << endl;
      t3.Write(cout);
      cout << "---------------------------------------------------" << endl;
   }
   
   /*-----------------------------------%
   | Max iterations have been exceeded. |
   %-----------------------------------*/

   if (iter > MaxIter && nconv < numcnv) {
      return TooManyIterations;
   }

   /*--------------------%
   | No shifts to apply. |
   %--------------------*/

   else if (np == 0 && nconv < numcnv) {
      return NoShiftsAvailable;
   }
   else
      return Success;

}
      
   
//
//  Compute the Eigenvalues/vectors
//

int
HCL_IRArnoldi_d::ComputeEigPairs(double sigmar,  
			  HCL_Vector_d *resid, GeneralMatrix_d *QQ)
{

   //
   // Allocate storage
   //

      double * tritzr = workl+kplusp*kplusp;
      double * tritzi = tritzr + kplusp;
      double * tbounds = tritzi + kplusp;
     
   int one = 1;

   int i, j, k;

   timer_d tme(&EigPair_time);

   /*------------------------------------%
   | Get the machine dependent constant. |
   %------------------------------------*/

#ifdef SINGLE_PRECISION
   double eps23 = F77NAME(slamch)("Epsilon-Machine");
#else
   double eps23 = F77NAME(dlamch)("Epsilon-Machine");
#endif
   eps23 = pow(eps23, 2.0 / 3.0);

   if (EigVectors || SchurVectors) {

      double thres;   // Threshold eigenvalue used for re-ordering

      /*------------------------------------------%
      | Get converged Ritz value on the boundary. |
      | Note: converged Ritz values have been     |
      | placed in the first NCONV locations in    |
      | workl(ritzr) and workl(ritzi).  They have |
      | been sorted (in _naup2) according to the  |
      | WHICH selection criterion.                |
      %------------------------------------------*/

      if (!strncmp(Which, "LM", 2) || !strncmp(Which, "SM",2)) {
#ifdef SINGLE_PRECISION
	 thres = F77NAME(slapy2)(ritzr, ritzi);
#else
	 thres = F77NAME(dlapy2)(ritzr, ritzi);
#endif
      }
      else if (!strncmp(Which, "LR", 2) || !strncmp(Which, "SR",2))
	 thres = ritzr[0];
      else
//	 thres = ritzi[0];  Changed by MSG
	 thres = abs( ritzi[0] );
      
      if (DispFlag > 3) {
	 cout << "ComputeEigPairs: Threshold eigenvalue used for re-ordering: "
	      << thres << endl;
      }

      /*---------------------------------------------------------%
      | Check to see if all converged Ritz values appear at the  |
      | top of the upper quasi-triangular matrix computed by     |
      | _neigh in _naup2.  This is done in the following way:    |
      |                                                          |
      | 1) For each Ritz value obtained from _neigh, compare it  |
      |    with the threshold Ritz value computed above to       |
      |    determine whether it is a wanted one.                 |
      |                                                          |
      | 2) If it is wanted, then check the corresponding Ritz    |
      |    estimate to see if it has converged.  If it has, set  |
      |    correponding entry in the logical array SELECT to     |
      |    .TRUE..                                               |
      |                                                          |
      | If SELECT(j) = .TRUE. and j > NCONV, then there is a     |
      | converged Ritz value that does not appear at the top of  |
      | the upper quasi-triangular matrix computed by _neigh in  |
      | _naup2.  Reordering is needed.                           |
      %---------------------------------------------------------*/

      int reord = 0;
      int ktrord = 0;
      double temp1;

// There seems to be a bug in the following for-loop.  I'm changing
// all references to ritzr and ritzi to tritzr and tritzi (MSG 22 May 1998)
// Ditto bounds to tbounds


      for (j = 0; j < ncv; ++j) {
	 select[j] = 0;
 
	 if (!strncmp(Which, "LM", 2)) {
#ifdef SINGLE_PRECISION
//	    if (F77NAME(slapy2)(ritzr+j,ritzi+j) >= thres) {
	    if (F77NAME(slapy2)(tritzr+j,tritzi+j) >= thres) {
//	       temp1 = max(eps23, F77NAME(slapy2)(ritzr+j,ritzi+j));
	       temp1 = max(eps23, F77NAME(slapy2)(tritzr+j,tritzi+j));
#else
//	    if (F77NAME(dlapy2)(ritzr+j,ritzi+j) >= thres) {
	    if (F77NAME(dlapy2)(tritzr+j,tritzi+j) >= thres) {
//	       temp1 = max(eps23, F77NAME(dlapy2)(ritzr+j,ritzi+j));
	       temp1 = max(eps23, F77NAME(dlapy2)(tritzr+j,tritzi+j));
#endif
//	       if (bounds[j] <= Tol*temp1)
	       if (tbounds[j] <= Tol*temp1)
		  select[j] = 1;
	    }
	 } else if (!strncmp(Which, "SM", 2)) {
#ifdef SINGLE_PRECISION
//	    if (F77NAME(slapy2)(ritzr+j, ritzi+j) <= thres) {
	    if (F77NAME(slapy2)(tritzr+j, tritzi+j) <= thres) {
//	       temp1 = max(eps23, F77NAME(slapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(slapy2)(tritzr+j, tritzi+j ));
#else
//	    if (F77NAME(dlapy2)(ritzr+j, ritzi+j) <= thres) {
	    if (F77NAME(dlapy2)(tritzr+j, tritzi+j) <= thres) {
//	       temp1 = max(eps23, F77NAME(dlapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(dlapy2)(tritzr+j, tritzi+j ));
#endif
//	       if (bounds[j] <= Tol*temp1)
	       if (tbounds[j] <= Tol*temp1)
		  select[j] = 1;
	    }
	 } else if (!strncmp(Which, "LR", 2)) {
//	    if (ritzr[j] >= thres) {
	    if (tritzr[j] >= thres) {
#ifdef SINGLE_PRECISION
//	       temp1 = max(eps23, F77NAME(slapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(slapy2)(tritzr+j, tritzi+j ));
#else
//	       temp1 = max(eps23, F77NAME(dlapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(dlapy2)(tritzr+j, tritzi+j ));
#endif
//	       if (bounds[j] <= Tol*temp1)
	       if (tbounds[j] <= Tol*temp1)
		  select[j] = 1;
	    }
	 } else if (!strncmp(Which, "SR", 2)) {
//	    if (ritzr[j] <= thres) {
	    if (tritzr[j] <= thres) {
#ifdef SINGLE_PRECISION
//	       temp1 = max(eps23, F77NAME(slapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(slapy2)(tritzr+j, tritzi+j ));
#else
//	       temp1 = max(eps23, F77NAME(dlapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(dlapy2)(tritzr+j, tritzi+j ));
#endif
//	       if (bounds[j] <= Tol*temp1)
	       if (tbounds[j] <= Tol*temp1)
		  select[j] = 1;
	    }
	 } else if (!strncmp(Which, "LI", 2)) {
//	    if (ritzi[j] >= thres) {
	    if (tritzi[j] >= thres) {
#ifdef SINGLE_PRECISION
//	       temp1 = max(eps23, F77NAME(slapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(slapy2)(tritzr+j, tritzi+j ));
#else
//	       temp1 = max(eps23, F77NAME(dlapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(dlapy2)(tritzr+j, tritzi+j ));
#endif
//	       if (bounds[j] <= Tol*temp1)
	       if (tbounds[j] <= Tol*temp1)
		  select[j] = 1;
	    }
	 } else if (!strncmp(Which, "SI", 2)) {
//	    if (ritzi[j] <= thres) {
	    if (tritzi[j] <= thres) {
#ifdef SINGLE_PRECISION
//	       temp1 = max(eps23, F77NAME(slapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(slapy2)(tritzr+j, tritzi+j ));
#else
//	       temp1 = max(eps23, F77NAME(dlapy2)(ritzr+j, ritzi+j ));
	       temp1 = max(eps23, F77NAME(dlapy2)(tritzr+j, tritzi+j ));
#endif
//	       if (bounds[j] <= Tol*temp1)
	       if (tbounds[j] <= Tol*temp1)
		  select[j] = 1;
	    }
         }
         
	 if (j+1 > nconv)
	    reord = (select[j] || reord);
	 if (select[j])
            ++ktrord;
      }
 
      if (DispFlag >  3) {
         cout << "HCL_IRArnoldi_d::Solve: Number of specified eigenvalues: " << ktrord << endl;
         cout << "HCL_IRArnoldi_d::Solve: Number of converged eigenvalues: " << nconv << endl;
      }
  
         /*----------------------------------------------------------%
         | Call LAPACK routine dlahqr to compute the real Schur form |
         | of the upper Hessenberg matrix returned by DNAUPD.        |
         | Make a copy of the upper Hessenberg matrix.               |
         | Initialize the Schur vector matrix QQ to the identity.     |
         %----------------------------------------------------------*/


      int s = ncv*ncv;
      int ierr = 0;
      double alpha = 1.0, beta = 0.0;
#ifdef SINGLE_PRECISION
      F77NAME(scopy)(&s, H->data, &one, R->data, &one);
      F77NAME(slaset)("All", &ncv, &ncv, &beta, &alpha, QQ->data, &ncv);
      
      F77NAME(slahqr)(&one, &one, &ncv, &one, &ncv, R->data, &ncv, 
		      heigr, heigi, &one, &ncv, QQ->data, &ncv, &ierr);
      F77NAME(scopy)(&ncv, QQ->data+ncv-1, &ncv, bounds, &one);
#else
      F77NAME(dcopy)(&s, H->data, &one, R->data, &one);
      F77NAME(dlaset)("All", &ncv, &ncv, &beta, &alpha, QQ->data, &ncv);
      
      F77NAME(dlahqr)(&one, &one, &ncv, &one, &ncv, R->data, &ncv, 
		      heigr, heigi, &one, &ncv, QQ->data, &ncv, &ierr);
      F77NAME(dcopy)(&ncv, QQ->data+ncv-1, &ncv, bounds, &one);
#endif
    
      if (ierr) {
	 if (DispFlag)
	    cout << "HCL_IRArnoldi_d::Solve: dlahqr failed: " << ierr << endl;
         return ComputeSchurVectsFailed;
      }
       
      if (DispFlag > 3) {
          
         cout << "HCL_IRArnoldi_d::Solve: Real part of the eigenvalues of H:" 
	      << endl;
         HCL_RnVector_d t1(ncv, heigr); 
         t1.Write(cout);
   
         cout << "HCL_IRArnoldi_d::Solve: Imaginary part of the eigenvalues of H:"
	      << endl;
         HCL_RnVector_d t2(ncv, heigi); 
         t2.Write(cout);
   
         cout << "HCL_IRArnoldi_d::Solve: The upper quasi-triangular matrix:" 
	      << endl;
         cout << *R << endl;
      }
   
      if (reord) {

         /*----------------------------------------------------%
         | Reorder the computed upper quasi-triangular matrix. |
         %----------------------------------------------------*/

#ifdef SINGLE_PRECISION
         int dummy;
	 F77NAME(strsen)("None", "V", select, &ncv, R->data, &ncv,
			 QQ->data, &ncv, heigr, heigi, &nconv, NULL, NULL,
			 bounds, &ncv, &dummy, &one, &ierr);
#else
         int dummy;
	 F77NAME(dtrsen)("None", "V", select, &ncv, R->data, &ncv,
			 QQ->data, &ncv, heigr, heigi, &nconv, NULL, NULL,
			 bounds, &ncv, &dummy, &one, &ierr);
#endif
	 
	 if (ierr == 1) {
	    if (DispFlag)
	       cout << "HCL_IRArnoldi_d::Solve: dtrsen failed: " << ierr << endl;
            return 1;
	 }
 
         if (DispFlag > 3) {
	    
	    cout << "HCL_IRArnoldi_d::Solve: After reordering: "<< endl;
             
            cout << "HCL_IRArnoldi_d::Solve: Real part of the eigenvalues of H:" 
		 << endl;
            HCL_RnVector_d t1(ncv, heigr); 
            t1.Write(cout);
      
            cout << "HCL_IRArnoldi_d::Solve: Imaginary part of the eigenvalues "
		    "of H:" << endl;
            HCL_RnVector_d t2(ncv, heigi); 
            t2.Write(cout);
         
	    cout << "HCL_IRArnoldi_d::Solve: The upper quasi-triangular matrix:" 
		 << endl;
            cout << *R << endl;
         }
      }

      /*--------------------------------------%
      | Copy the last row of the Schur vector |
      | into workl(ihbds).  This will be used |
      | to compute the Ritz estimates of      |
      | converged Ritz values.                |
      %--------------------------------------*/

#ifdef SINGLE_PRECISION
      F77NAME(scopy)(&ncv, QQ->data+ncv-1, &ncv, hbds, &one);
#else
      F77NAME(dcopy)(&ncv, QQ->data+ncv-1, &ncv, hbds, &one);
#endif


      /*---------------------------------------------------%
      | Place the computed eigenvalues of H into DR and DI |
      | if a spectral transformation was not used.         |
      %---------------------------------------------------*/

      if (type == Regular) {
#ifdef SINGLE_PRECISION
	 F77NAME(scopy)(&nconv, heigr, &one, dr, &one);
	 F77NAME(scopy)(&nconv, heigi, &one, di, &one);
#else
	 F77NAME(dcopy)(&nconv, heigr, &one, dr, &one);
	 F77NAME(dcopy)(&nconv, heigi, &one, di, &one);
#endif
      }


      /*---------------------------------------------------------%
      | Compute the QR factorization of the matrix representing  |
      | the wanted invariant subspace located in the first NCONV |
      | columns of workl(invsub,ldq).                            |
      %---------------------------------------------------------*/

#ifdef SINGLE_PRECISION
      F77NAME(sgeqr2)(&ncv, &nconv, QQ->data, &ncv, workev, workev+ncv,
		     &ierr);
#else
      F77NAME(dgeqr2)(&ncv, &nconv, QQ->data, &ncv, workev, workev+ncv,
		     &ierr);
#endif
      
      /*--------------------------------------------------------%
      | * Postmultiply V by QQ using dorm2r.                     |
      | * Copy the first NCONV columns of VQQ into Z.            |
      | * Postmultiply Z by R.                                  |
      | The N by NCONV matrix Z is now a matrix representation  |
      | of the approximate invariant subspace associated with   |
      | the Ritz values in workl(iheigr) and workl(iheigi)      |
      | The first NCONV columns of V are now approximate Schur  |
      | vectors associated with the real upper quasi-triangular |
      | matrix of order NCONV in workl(iuptri)                  |
      %--------------------------------------------------------*/
         
      
      for (i = 1; i <= nconv; ++i) {
            
	 // V = V*H_i
   
	 if (workev[i-1] != 0) {
	    
	    // t = V*v
	    
	    int k;
	    tmpvec->Copy(*V[i-1]);
	    for (k = i+1; k <= ncv; ++k)
	       tmpvec->AddScale((*QQ)(k, i), *V[k-1]);
     
	    // V = V - tau*t * v'
   
	    V[i-1]->AddScale(-workev[i-1], *tmpvec);
	    for (k = i+1; k <= ncv; ++k)
	       V[k-1]->AddScale(-(*QQ)(k, i)*workev[i-1], *tmpvec);
	    
	 }
      }
     
	 
      for (i = 1; i <= nconv; ++i) {
	 
         /*--------------------------------------------------%
         | Perform both a column and row scaling if the      |
         | diagonal element of workl(invsub,ldq) is negative |
         | I'm lazy and don't take advantage of the upper    |
         | quasi-triangular form of workl(iuptri,ldq)        |
         | Note that since QQ is orthogonal, R is a diagonal  |
         | matrix consisting of plus or minus ones           |
         %--------------------------------------------------*/

	 if ((*QQ)(i,i) < 0) {
	    double n_one = -1;
#ifdef SINGLE_PRECISION
	    F77NAME(sscal)(&nconv, &n_one, R->data+i-1, &ncv);
	    F77NAME(sscal)(&nconv, &n_one, R->data+(i-1)*ncv, &one);
#else
	    F77NAME(dscal)(&nconv, &n_one, R->data+i-1, &ncv);
	    F77NAME(dscal)(&nconv, &n_one, R->data+(i-1)*ncv, &one);
#endif
	 }
      }
      if (EigVectors) {

         /*-------------------------------------------%
         | Compute the NCONV wanted eigenvectors of T |
         | located in workl(iuptri,ldq).              |
         %-------------------------------------------*/

	 for (j = 0; j < ncv; ++j) {
	    if (j < nconv)
	       select[j] = 1;
	    else
	       select[j] = 0;
	 }

         int ierr = 0, outncv;
#ifdef SINGLE_PRECISION
	 F77NAME(strevc)("Right", "Select", select, &ncv, R->data,
			 &ncv, NULL, &one, QQ->data, &ncv, &ncv, &outncv, 
			 workev, &ierr);
#else
	 F77NAME(dtrevc)("Right", "Select", select, &ncv, R->data,
			 &ncv, NULL, &one, QQ->data, &ncv, &ncv, &outncv, 
			 workev, &ierr);
#endif
	 
	 if (ierr != 0) {
	    return ComputeEigVectsFailed;
	 }

         /*-----------------------------------------------%
         | Scale the returning eigenvectors so that their |
         | Euclidean norms are all one. LAPACK subroutine |
         | dtrevc returns each eigenvector normalized so  |
         | that the element of largest magnitude has      |
         | magnitude 1;                                   |
         %-----------------------------------------------*/

	 int iconj = 0;
         
	 for (j = 1; j <= nconv; ++j) {
	    
	    double scale;
	    if (heigi[j-1] == 0) {
	       
               /*---------------------%
               | real eigenvalue case |
               %---------------------*/
	       
#ifdef SINGLE_PRECISION
	       scale = 1/F77NAME(snrm2)(&ncv, QQ->data+(j-1)*ncv, &one);
	       F77NAME(sscal)(&ncv, &scale, QQ->data+(j-1)*ncv, &one);
#else
	       scale = 1/F77NAME(dnrm2)(&ncv, QQ->data+(j-1)*ncv, &one);
	       F77NAME(dscal)(&ncv, &scale, QQ->data+(j-1)*ncv, &one);
#endif
	    }
	    else {

               /*------------------------------------------%
               | Complex conjugate pair case. Note that    |
               | since the real and imaginary part of      |
               | the eigenvector are stored in consecutive |
               | columns, we further normalize by the      |
               | square root of two.                       |
               %------------------------------------------*/

	       if (iconj == 0) {
#ifdef SINGLE_PRECISION
		  double n1 = F77NAME(snrm2)(&ncv, QQ->data+(j-1)*ncv, &one);
		  double n2 = F77NAME(snrm2)(&ncv, QQ->data+j*ncv, &one);
		  scale = 1/F77NAME(slapy2)(&n1, &n2);

		  F77NAME(sscal)(&ncv, &scale, QQ->data+(j-1)*ncv, &one);
		  F77NAME(sscal)(&ncv, &scale, QQ->data+j*ncv, &one);
#else
		  double n1 = F77NAME(dnrm2)(&ncv, QQ->data+(j-1)*ncv, &one);
		  double n2 = F77NAME(dnrm2)(&ncv, QQ->data+j*ncv, &one);
		  scale = 1/F77NAME(dlapy2)(&n1, &n2);

		  F77NAME(dscal)(&ncv, &scale, QQ->data+(j-1)*ncv, &one);
		  F77NAME(dscal)(&ncv, &scale, QQ->data+j*ncv, &one);
#endif
		  iconj = 1;
	       }
	       else
		  iconj = 0;
	       
	    }
	 }

         double dbl_one = 1.0;
         double dbl_zero = 0.0;
#ifdef SINGLE_PRECISION
	 F77NAME(sgemv)("T", &ncv, &nconv, &dbl_one, QQ->data, &ncv, hbds, 
			&one, &dbl_zero, workev, &one);
#else
	 F77NAME(dgemv)("T", &ncv, &nconv, &dbl_one, QQ->data, &ncv, hbds, 
			&one, &dbl_zero, workev, &one);
#endif
	 
	 iconj = 0;
	 for (j = 1; j <= nconv; ++j) {
	    if (heigi[j-1] != 0) {
	       
	       if (iconj == 0) {
#ifdef SINGLE_PRECISION
		  workev[j-1] = F77NAME(slapy2)(&workev[j-1], &workev[j]);
#else
		  workev[j-1] = F77NAME(dlapy2)(&workev[j-1], &workev[j]);
#endif
		  workev[j] = workev[j-1];
		  iconj = 1;
	       }
	       else 
		  iconj = 0;
	    }
	 }

	 if (DispFlag > 3) {
	    cout << "HCL_IRArnoldi_d::Solve: The eigenvector matrix for T:" 
		 << endl;
            cout << *QQ << endl;
	 }

         /*--------------------------------------%
         | Copy Ritz estimates into workl(ihbds) |
         %--------------------------------------*/


#ifdef SINGLE_PRECISION
	 F77NAME(scopy)(&nconv, workev, &one, hbds, &one);
#else
	 F77NAME(dcopy)(&nconv, workev, &one, hbds, &one);
#endif

         /*--------------------------------------------------------%
         | Compute the QR factorization of the eigenvector matrix  |
         | associated with leading portion of T in the first NCONV |
         | columns of workl(invsub,ldq).                           |
         %--------------------------------------------------------*/

#ifdef SINGLE_PRECISION
	 F77NAME(sgeqr2)(&ncv, &nconv, QQ->data, &ncv, workev, workev+ncv,
			 &ierr);
#else
	 F77NAME(dgeqr2)(&ncv, &nconv, QQ->data, &ncv, workev, workev+ncv,
			 &ierr);
#endif
	 
         /*---------------------------------------------%
         | * Postmultiply Z by QQ.                       |
         | * Postmultiply Z by R.                       |
         | The N by NCONV matrix Z is now contains the  |
         | Ritz vectors associated with the Ritz values |
         | in workl(iheigr) and workl(iheigi).          |
         %---------------------------------------------*/
	 
   
	 for (i = 1; i <= nconv; ++i) {
	       
	    // V = V*H_i
      
	    if (workev[i-1] != 0) {
	       
	       // t = V*v
	       
	       int k;
	       tmpvec->Copy(*V[i-1]);
	       for (k = i+1; k <= ncv; ++k)
	          tmpvec->AddScale((*QQ)(k, i), *V[k-1]);
	
	       // V = V - tau*t * v'
      
	       V[i-1]->AddScale(-workev[i-1], *tmpvec);
	       for (k = i+1; k <= ncv; ++k)
		  V[k-1]->AddScale(-(*QQ)(k, i)*workev[i-1], *tmpvec);
	    
	    }
	 }

	 for (i = nconv; i >= 1; --i) {
	    V[i-1]->Mul((*QQ)(i,i));
	    for (j = 1; j < i; ++j) {
	       V[i-1]->AddScale((*QQ)(j,i), *V[j-1]);
	    }
	 }

      } /* if EigVectors */

   } /* if rvec */

   else {  
      
      /*-----------------------------------------------------%
      | An approximate invariant subspace is not needed.     |
      | Place the Ritz values computed DNAUPD into DR and DI |
      %-----------------------------------------------------*/
      
#ifdef SINGLE_PRECISION
      F77NAME(scopy)(&nconv, ritzr, &one, dr, &one);
      F77NAME(scopy)(&nconv, ritzi, &one, di, &one);
      F77NAME(scopy)(&nconv, ritzr, &one, heigr, &one);
      F77NAME(scopy)(&nconv, ritzi, &one, heigi, &one);
      F77NAME(scopy)(&nconv, bounds, &one, hbds, &one);
#else
      F77NAME(dcopy)(&nconv, ritzr, &one, dr, &one);
      F77NAME(dcopy)(&nconv, ritzi, &one, di, &one);
      F77NAME(dcopy)(&nconv, ritzr, &one, heigr, &one);
      F77NAME(dcopy)(&nconv, ritzi, &one, heigi, &one);
      F77NAME(dcopy)(&nconv, bounds, &one, hbds, &one);
#endif

   }

   /*-----------------------------------------------%
   | Transform the Ritz values and possibly vectors |
   | and corresponding error bounds of OP to those  |
   | of A*x = lambda*B*x.                           |
   %-----------------------------------------------*/

   if (type == Regular) {
#ifdef SINGLE_PRECISION
      F77NAME(sscal)(&ncv, &rnorm, hbds, &one);
#else
      F77NAME(dscal)(&ncv, &rnorm, hbds, &one);
#endif
   }
   else {
      
      /*--------------------------------------%
      |   A spectral transformation was used. |
      | * Determine the Ritz estimates of the |
      |   Ritz values in the original system. |
      %--------------------------------------*/

      if (type == ShiftI) {
	 if (EigVectors || SchurVectors) {
#ifdef SINGLE_PRECISION
	    F77NAME(sscal)(&ncv, &rnorm, hbds, &one);
#else
	    F77NAME(dscal)(&ncv, &rnorm, hbds, &one);
#endif
	 }
	 
	 double temp;
	 for (k = 0; k < ncv; ++k) {
#ifdef SINGLE_PRECISION
	    temp = F77NAME(slapy2)(&heigr[k], &heigi[k]);
#else
	    temp = F77NAME(dlapy2)(&heigr[k], &heigi[k]);
#endif
	    hbds[k] = abs(hbds[k])/temp/temp;
	 }
      }

      /*----------------------------------------------------------%
      | *  Transform the Ritz values back to the original system. |
      |    For TYPE = 'SHIFTI' the transformation is              |
      |             lambda = 1/theta + sigma                      |
      |    For TYPE = 'REALPT' or 'IMAGPT' the user must form     |
      |    Rayleigh quotients or a projection. See remark 3 above.|
      | NOTES:                                                    |
      | *The Ritz vectors are not affected by the transformation. |
      %----------------------------------------------------------*/

      if (type == ShiftI) {
	 double temp;
	 for (k = 0; k < ncv; ++k) {
#ifdef SINGLE_PRECISION
	    temp = F77NAME(slapy2)(&heigr[k], &heigi[k]);
#else
	    temp = F77NAME(dlapy2)(&heigr[k], &heigi[k]);
#endif
	    heigr[k] = heigr[k]/temp/temp + sigmar;
	    heigi[k] = heigi[k]/temp/temp;
	 }

#ifdef SINGLE_PRECISION
         F77NAME(scopy)(&nconv, heigr, &one, dr, &one);
         F77NAME(scopy)(&nconv, heigi, &one, di, &one);
#else
         F77NAME(dcopy)(&nconv, heigr, &one, dr, &one);
         F77NAME(dcopy)(&nconv, heigi, &one, di, &one);
#endif
      }
      else if (type == RealPt || type == ImagPt) {
#ifdef SINGLE_PRECISION
         F77NAME(scopy)(&nconv, heigr, &one, dr, &one);
         F77NAME(scopy)(&nconv, heigi, &one, di, &one);
#else
         F77NAME(dcopy)(&nconv, heigr, &one, dr, &one);
         F77NAME(dcopy)(&nconv, heigi, &one, di, &one);
#endif
      }
   }
      
   if (DispFlag > 2) {
	    
      cout << "HCL_IRArnoldi_d::Solve: Real part of the Ritz values:" << endl;
      HCL_RnVector_d t1(nconv, dr); 
      t1.Write(cout);
      
      cout << "HCL_IRArnoldi_d::Solve: Imaginary part of the Ritz values:" 
	   << endl;
      HCL_RnVector_d t2(nconv, di); 
      t2.Write(cout);
   
   }

   /*------------------------------------------------%
   | Eigenvector Purification step. Formally perform |
   | one of inverse subspace iteration. Only used    |
   | for MODE = 2.                                   |
   %------------------------------------------------*/

   if (EigVectors && type == ShiftI) {

      /*-----------------------------------------------%
      | Purify the computed Ritz vectors by adding a   |
      | little bit of the residual vector:             |
      |                      T                         |
      |          resid(:)*( e    s ) / theta           |
      |                      NCV                       |
      | where H s = s theta. Remember that when theta  |
      | has nonzero imaginary part, the corresponding  |
      | Ritz vector is stored across two columns of Z. |
      %-----------------------------------------------*/

      
      int iconj = 0;
      for (j = 0; j < nconv; ++j) {
	 if (heigi[j] == 0) {
	    workev[j] = (*QQ)(ncv,j+1) / heigr[j];
	 }
	 else if (iconj == 0) {
#ifdef SINGLE_PRECISION
	    double temp = F77NAME(slapy2)(&heigr[j],&heigi[j]);
#else
	    double temp = F77NAME(dlapy2)(&heigr[j],&heigi[j]);
#endif
	    workev[j] = ((*QQ)(ncv,j+1)*heigr[j] + 
				    (*QQ)(ncv,j+2)*heigi[j])/temp/temp;
	    workev[j+1] = ((*QQ)(ncv,j+2)*heigr[j] + 
				    (*QQ)(ncv,j+1)*heigi[j])/temp/temp;
	    iconj = 1;
	 }
	 else
	    iconj = 0;
      }

      /*--------------------------------------%
      | Perform a rank one update to Z and    |
      | purify all the Ritz vectors together. |
      %--------------------------------------*/

      for (k = 0; k < nconv; ++k) {
	 V[k]->AddScale(workev[k], *resid);
      }
   }

   return Success;
}
	 


int
HCL_IRArnoldi_d::Arnoldi(HCL_LinearOp_d & A, HCL_LinearOp_d * B, 
	GeneralMatrix_d & HH, HCL_Vector_d **VV, 
	int k, int np, int &npout,
	HCL_Vector_d & resid, double & rrnorm,
	HCL_Vector_d * tempvec, int ttype)
{
#ifdef SINGLE_PRECISION
   float unfl = F77NAME(slamch)( "s" );  // safe minimum
   float ovfl = 1/unfl;
   F77NAME(slabad)(&unfl, &ovfl);
   float ulp = F77NAME(slamch)("p");     // precision
#else
   double unfl = F77NAME(dlamch)( "s" );  // safe minimum
   double ovfl = 1/unfl;
   F77NAME(dlabad)(&unfl, &ovfl);
   double ulp = F77NAME(dlamch)("p");     // precision
#endif
   double smlnum = unfl*( 1 / ulp );

   double betaj, wnorm;

   timer_d tme(&Factor_time);

   if (DispFlag>3) {
      cout << endl << "===========HCL_IRArnoldi_d::Arnoldi===========" << endl;
   }
   
   if (B != NULL && ttype == ShiftI) {
      {
        timer_d tme(&B_time);
        B->Image(resid, *tempvec);
      }
      B_count++;
   }

    /*-------------------------------------------------------------%
    |                                                              |
    |        A R N O L D I     I T E R A T I O N     L O O P       |
    |                                                              |
    %-------------------------------------------------------------*/

   int i, j;
   for (j = k; j < k+np; ++j) {
      
      if (DispFlag>3) {
	 cout << "HCL_IRArnoldi_d::Arnoldi: Generating Arnoldi vector "
                 "number : " << j+1 << endl;
	 cout << "HCL_IRArnoldi_d::Arnoldi: B-norm of the current residual: " 
	      << rrnorm << endl;
      }


      /*--------------------------------------------------%
      | STEP 1: Check if the norm of j-th residual        |
      | vector is zero. Equivalent to determing whether   |
      | an exact j-step Arnoldi factorization is present. |
      %--------------------------------------------------*/

      betaj = rrnorm;

      if (rrnorm == 0) {

         /*--------------------------------------------------%
         | Invariant subspace found. Return the size of the  |
	 | subspace already computed.                        |
         %--------------------------------------------------*/
	 
	 if (DispFlag >1) {
            cout <<"HCL_IRArnoldi_d::Arnoldi: ****** Invariant subspace found at step ******"
	         << j << "." << endl;
	 }
	 npout = j-k;
	 return 0;
      }
   
      /*-------------------------------%
      | STEP 2:  v_{j} = r_{j-1}/rrnorm |
      %-------------------------------*/

      VV[j]->Mul(1.0/rrnorm, resid);
      tempvec->Mul(1.0/rrnorm);

      /*-----------------------------------------------------%
      | STEP 3:  r_{j} = A*v_{j};                           |
      | Note that this is not quite yet r_{j}. See STEP 4    |
      %-----------------------------------------------------*/

      if (ttype == Regular || B == NULL) {
	 {
	   timer_d tme(&OP_time);
           A.Image(*VV[j], resid);
	 }
	 OP_count++;
      }
      else {
	 // shift and invert mode 
	 // tempvec contains B*resid
	 {
	   timer_d tme(&OP_time);
	   A.Image(*tempvec, resid);
	 }
	 OP_count++;
      }
	 

      /*--------------------------------------%
      | STEP 4:  Finish extending the Arnoldi |
      |          factorization to length j.   |
      %--------------------------------------*/

      if (B != NULL) {
	 {
	   timer_d tme(&B_time);
	   B->Image(resid, *tempvec);
	 }
	 B_count++;
      }
      else {
	 tempvec->Copy(resid);
      }

      /*------------------------------------%
      | The following is needed for STEP 5. |
      | Compute the norm of A*v_{j}.        |
      %------------------------------------*/

      if (B != NULL) {
	 wnorm = sqrt(abs(resid.Inner(*tempvec)));
      }
      else {
	 wnorm = resid.Norm();
      }

      /*----------------------------------------%
      | Compute the j-th residual corresponding |
      | to the j step factorization.            |
      | Use Classical Gram Schmidt and compute: |
      | w_{j} <-  V_{j}^T * B * A * v_{j}      |
      | r_{j} <-  A*v_{j} - V_{j} * w_{j}      |
      %----------------------------------------*/

      for (i = 0; i <= j; ++i) {
	 HH(i+1, j+1) = VV[i]->Inner(*tempvec);
//	HH(i+1, j+1) = VV[i]->Inner(resid); // Changed by MSG to do MGS
	 resid.AddScale(-HH(i+1, j+1), *VV[i]);
      }

      if (j > 0) {
	 HH(j+1, j) = betaj;
      }

      /*---------------------------%
      | Compute the norm of r_{j}. |
      %---------------------------*/

      if (B != NULL) {
	 {
	   timer_d tme(&B_time);
	   B->Image(resid, *tempvec);
	 }
	 B_count++;
	 rrnorm = sqrt(abs(resid.Inner(*tempvec)));
      }
      else {
	 tempvec->Copy(resid);
	 rrnorm = resid.Norm();
      }
//      cout << "residual:" << endl;
//      resid.Write(cout);
	
      /*----------------------------------------------------------%
      | STEP 5: Re-orthogonalization / Iterative refinement phase |
      | Maximum one tries.                                        |
      |                                                           |
      |          s      = V_{j}^T * B * r_{j}                     |
      |          r_{j}  = r_{j} - V_{j}*s                         |
      |          alphaj = alphaj + s_{j}                          |
      |                                                           |
      | The stopping criteria used for iterative refinement is    |
      | discussed in Parlett's book SEP, page 107 and in Gragg &  |
      | Reichel ACM TOMS paper; Algorithm 686, Dec. 1990.         |
      | Determine if we need to correct the residual. The goal is |
      | to enforce ||v(:,1:j)^T * r_{j}|| .le. eps * || r_{j} ||  |
      | The following test determines whether the sine of the     |
      | angle between  A*x and the computed residual is less     |
      | than or equal to 0.717.                                   |
      %----------------------------------------------------------*/

      // Changed by MSG to use a local variable rather than iter
      int lociter = 0;
      while (rrnorm <= 0.717*wnorm) {

         timer_d tme(&Reorthog_time);
         Reorthog_count++;

	 if (DispFlag>3) {
	    cout << "HCL_IRArnoldi_d::Arnoldi: Re-orthogonalization phase" 
		 << endl;
	    cout << "HCL_IRArnoldi_d::Arnoldi: wnorm = " << wnorm << "   rrnorm = "
		 << rrnorm << endl;
	    
	    cout << "HCL_IRArnoldi_d::Arnoldi: j-th column of HH:" << endl;
	    for (i = 0; i <= j; ++i)
	       cout << "     " << HH(i+1, j+1) << endl;
	 }

         /*--------------------------------------------------%
         | Enter the Iterative refinement phase. If further  |
         | refinement is necessary, loop back here. The loop |
         | variable is LOCITER. Perform a step of Classical  |
         | Gram-Schmidt using all the Arnoldi vectors V_{j}  |
         %--------------------------------------------------*/

         if (lociter == 2) {

            if (DispFlag>1) {
	       cout << "HCL_IRArnoldi_d::Arnoldi: Failed; resid is numerically"
		    << " in the span of VV" << endl;
	    }

            /*--------------------------------------%
            | RESID is numerically in the span of VV |
            %--------------------------------------*/
             
	    resid.Zero();
	    rrnorm = 0;
            
	    npout = j;
	    return 0;
	 }

	 wnorm = rrnorm;

//         if (DispFlag>3) {
//	    cout << "HCL_IRArnoldi_d::Arnoldi: original residual" << endl;
//	    resid.Write(cout);
//	    cout << "HCL_IRArnoldi_d::Arnoldi: original h" << endl;
//	    HH.Write(cout);
//	 }

         /*---------------------------------------------------%
         | Compute V_{j}^T * B * r_{j}.                       |
         | WORKD(IRJ:IRJ+J-1) = v(:,1:J)'*WORKD(IPJ:IPJ+N-1). |
         %---------------------------------------------------*/

         /*--------------------------------------------%
         | Compute the correction to the residual:     |
         | r_{j} = r_{j} - V_{j} * WORKD(IRJ:IRJ+J-1). |
         | The correction to HH is v(:,1:J)*HH(1:J,1:J)  |
         | + v(:,1:J)*WORKD(IRJ:IRJ+J-1)*e'_j.         |
         %--------------------------------------------*/

	    
	 for (i = 0; i <= j; ++i) {
	    double c = VV[i]->Inner(*tempvec);
	    HH(i+1, j+1) += c;
	    resid.AddScale(-c, *VV[i]);
	 }
	 
//	 if (DispFlag>3) {
//	    cout << "HCL_IRArnoldi_d::Arnoldi: corrected residual" << endl;
//	    resid.Write(cout);
//	    cout << "HCL_IRArnoldi_d::Arnoldi: corrected h" << endl;
//	    HH.Write(cout);
//	 }

         /*----------------------------------------------------%
         | Compute the norm of the corrected residual r_{j}.   |
         %----------------------------------------------------*/
            

	 if (B != NULL) {
	    {
	      timer_d tme(&B_time);
	      B->Image(resid, *tempvec);
	    }
	    B_count++;
	    rrnorm = sqrt(abs(resid.Inner(*tempvec)));
	 }
	 else {
            tempvec->Copy( resid );
	    rrnorm = resid.Norm();
	 }

         if (DispFlag>3) {
	    cout << "HCL_IRArnoldi_d::Arnoldi: Iterative refinement for Arnoldi residual: " 
		 << wnorm << endl;
	    cout << "HCL_IRArnoldi_d::Arnoldi: rrnorm = " << wnorm << "  rrnorm1 = "
		 << rrnorm << endl;
	 }

	 lociter++;
      }

      /*--------------------------------------%
      | No need for further refinement.       |
      | The cosine of the angle between the   |
      | corrected residual vector and the old |
      | residual vector is greater than 0.717 |
      | In other words the corrected residual |
      | and the old residual vector share an  |
      | angle of less than arcCOS(0.717)      |
      %--------------------------------------*/

//      cout << "corrected residual:" << endl;
//      resid.Write(cout);


      /*-----------------------------------%
      | STEP 6: Update  j = j+1;  Continue |
      %-----------------------------------*/
   }

   /*-------------------------------------------%
   | Check for splitting and deflation.         |
   | Use a standard test as in the QR algorithm |
   | REFERENCE: LAPACK subroutine dlahqr        |
   %-------------------------------------------*/
   for (i = (k>1)?k:1 ; i <= k+np-1; ++i) {
	      
      double tst1 = abs( HH( i, i ) ) + abs( HH( i+1, i+1 ) );
      if (tst1 == 0) {
	 int kplusnp = k+np;
	 // compute the 1 Norm of HH
	 int ldh = HH.numRows();
#ifdef SINGLE_PRECISION
	 tst1 = F77NAME(slanhs)("1", &kplusnp, HH.data, &ldh, NULL);
#else
	 tst1 = F77NAME(dlanhs)("1", &kplusnp, HH.data, &ldh, NULL);
#endif
      }
      
      if (abs( HH( i+1,i ) ) <= max( ulp*tst1, smlnum ))
	 HH(i+1,i) = 0;
   }

   if (DispFlag>5) {
      cout << "HCL_IRArnoldi_d::Arnoldi: Returning Hessenberg matrix HH "
              "of order " << k+np << endl;
      cout << HH << endl;
   }

   npout = k+np;
   
   return 0;
}

ostream & HCL_IRArnoldi_d::Write(ostream & str) const
{
   str << "HCL_IRArnoldi_d eigenvalue solver" << endl;
   str << "Parameter table is" << endl;
   ParamTable->Write( str );
   return str;
}


