/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_LineSearch_d.h"

HCL_LineSearch_MT_d::HCL_LineSearch_MT_d( const HCL_LineSearch_d & x )
{
   cerr << "Error: HCL_LineSearch_MT_d copy constructor called" << endl;
   exit(1);
}

HCL_LineSearch_MT_d::HCL_LineSearch_MT_d( char *fname )
{
   if( fname != NULL )
      ParamTable = new Table( fname,"LineSearchMT LineSearch",20 );
   else
      ParamTable = new Table;
   S = NULL;
   dump = NULL;
}

HCL_LineSearch_MT_d::~HCL_LineSearch_MT_d()
{
   if( dump != NULL )
      delete dump;
   HCL_delete( S );
   HCL_delete( ParamTable );
}

void HCL_LineSearch_MT_d::Initialize( HCL_VectorSpace_d & dom )
{

   // Read the parameters
   int MinFlag = ParamTable->GetValue( "MinStep",MinStep );
   int MaxFlag = ParamTable->GetValue( "MaxStep",MaxStep );
   if( MinFlag && MaxFlag || (!MinFlag && !MaxFlag && MinStep >= MaxStep) )
   {
      MaxStep = 1.0e20;
      MinStep = 1.0e-20;
   }
   else if( MinFlag && !MaxFlag )
      MinStep = 1.0e-40*MaxStep;
   else if (MaxFlag && !MinFlag )
      MaxStep = 1.0e40*MinStep;
   if( ParamTable->GetValue( "MaxSample",MaxSample ) )
      MaxSample = 8;
   if( ParamTable->GetValue( "FcnDecreaseTol",FcnDecreaseTol ) )
      FcnDecreaseTol = 1e-4;
   if( ParamTable->GetValue( "SlopeDecreaseTol",SlopeDecreaseTol ) )
      SlopeDecreaseTol = 0.9;
#ifdef SINGLE_PRECISION
   HCL_FPArch_s arch;
#else
   HCL_FPArch_d arch;
#endif
   if( ParamTable->GetValue( "IntervalTol",IntervalTol ) )
      IntervalTol = sqrt(arch.eps);
   if( ParamTable->GetValue( "BracketIncrease",BracketIncrease ) )
      BracketIncrease = 4.0;
   if( ParamTable->GetValue( "ZeroDivideTol",ZeroDivideTol ) )
      ZeroDivideTol = 0.0;
   if( ParamTable->GetValue( "DispFlag",DispFlag ) )
      DispFlag = 0;
   if( ParamTable->GetValue( "DumpFlag",DumpFlag ) )
      DumpFlag = 0;
   if( ParamTable->GetValue( "DispPrecision",DispPrecision ) )
      DispPrecision = 6;
   OldDispPrec = cout.precision( DispPrecision );
   if( ParamTable->GetValue( "DumpPrecision",DumpPrecision ) )
      DumpPrecision = DispPrecision;
   if( DumpFlag && dump == NULL )
   {
      if( ParamTable->GetValue( "DumpFile",DumpFile ) )
         strcpy( DumpFile,"HCL_LineSearch_MT_d.DumpFile" );
      dump = new ofstream( DumpFile,ios::out );
      dump->precision( DumpPrecision );
   }
   else
      dump = NULL;

   // Check that S is consistent

   if( S != NULL )
   {
      if( S->Domain() != dom )
      {
         cerr << "Error in HCL_LineSearch_DS_d::Initialize: domain does "
                 "not match preconditioner" << endl;
         exit(1);
      }
      tmpvec = dom.Member();
   }
   else
      tmpvec = NULL;

   Itn = 0;
}

// Deallocate vectors and fill in the output parameters
void HCL_LineSearch_MT_d::Clean()
{
   cout.precision( OldDispPrec );
   
   ParamTable->PutValue("TermCode", TermCode);
   ParamTable->PutValue("MaxTkn", MaxTkn);
}

void HCL_LineSearch_MT_d::DisplayResults( int info )
{
   if( DispFlag )
   {
      if( TermCode == Succeed )
         cout << "Line search succeeded" << endl;
      else
         switch( info )
         {
            case 2:
               cout << "Line search halted: interval of uncertainty "
                       "is below tolerance" << endl;
               break;
            case 3:
               cout << "LineSearch halted: maximum number of iterations "
                       "reached" << endl;
               break;
            case 4:
               cout << "Line Search halted: minimum step length reached"
                    << endl;
               break;
            case 5:
               cout << "Line Search halted: maximum step length reached"
                    << endl;
               break;
            case 6:
               cout << "Line Search halted: failure to make progress; "
                       "tolerances may be too small" << endl;
               break;
            default:
               cout << "Error in HCL_LineSearch_MT_d::DisplayResults: "
                       "this default case should not occur" << endl;
         }
      if( TermCode == NoReduction )
         cout << "   Function value not reduced" << endl;
   }
   if( DumpFlag && dump != NULL )
   {
      if( TermCode == Succeed )
         (*dump) << "Line search succeeded" << endl;
      else
         switch( info )
         {
            case 2:
               (*dump) << "Line search halted: interval of uncertainty "
                          "is below tolerance" << endl;
               break;
            case 3:
               (*dump) << "LineSearch halted: maximum number of iterations "
                          "reached" << endl;
               break;
            case 4:
               (*dump) << "Line Search halted: minimum step length reached"
                       << endl;
               break;
            case 5:
               (*dump) << "Line Search halted: maximum step length reached"
                       << endl;
               break;
            case 6:
               (*dump) << "Line Search halted: failure to make progress; "
                          "tolerances may be too small" << endl;
               break;
            default:
               (*dump) << "Error in HCL_LineSearch_MT_d::DisplayResults: "
                          "this default case should not occur" << endl;
         }
      if( TermCode == NoReduction )
         (*dump) << "   Function value not reduced" << endl;
   }
}
      
void HCL_LineSearch_MT_d::SetScaling( const HCL_LinearOp_d * SS )
{
   HCL_delete( S );
   S = (HCL_LinearOp_d*)SS;
   S->IncCount();
}

void HCL_LineSearch_MT_d::UnSetScaling()
{
   HCL_delete( S );
   S = NULL;
}

HCL_EvaluateFunctional_d *
HCL_LineSearch_MT_d::Search(HCL_Functional_d & f,const HCL_Vector_d & xcur,
           HCL_Vector_d & xnext,HCL_Vector_d & pdir,
           HCL_EvaluateFunctional_d * xcur_eval )
{
   Initialize(f.Domain());

   if( DispFlag )
       cout << "===============HCL_LineSearch_MT_d===============" << endl;
   if( DumpFlag && dump != NULL )
       (*dump) << "===============HCL_LineSearch_MT_d===============" << endl;

   // Compute the length of the Newton step, and scale
   // the search direction (if necessary)

//   double newtlen = pdir.Norm();
   double newtlen;
   if( S != NULL )
   {
      S->Image( pdir,*tmpvec );
      newtlen = sqrt(pdir.Inner( *tmpvec ));
   }
   else
      newtlen = pdir.Norm();

   double t = f.MaxStep( xcur,pdir ),fmaxs;
   if( t <= 0.0 )
   {
      if( DispFlag )
         cout << "Error in HCL_LineSearch_Fl_d: vector not in domain"
                 " (fmaxs == 0.0)" << endl;
      if( DumpFlag && dump != NULL )
         (*dump) << "Error in HCL_LineSearch_Fl_d: vector not in domain"
                 " (fmaxs == 0.0)" << endl;
      TermCode = NoReduction;
      DisplayResults( 6 );
      Clean();
      return NULL;
   }
   HCL_FPArch_d arch;
   if( t == arch.large )
      fmaxs = t;
   else
      fmaxs = 0.99*newtlen*t;
   double fcnmaxstep = min( fmaxs,MaxStep );

   if( newtlen > fcnmaxstep )    // ||dir|| > MaxStep
   {
      double scale;
      if( ProtectedDivision( fcnmaxstep,newtlen,scale,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "ProtectedDivision error: newtlen too small" 
                 << "   " << fcnmaxstep << "/" << newtlen << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "ProtectedDivision error: newtlen too small"
                    << "   " << fcnmaxstep << "/" << newtlen << endl;

         TermCode = NoReduction;
         DisplayResults( 6 );
         Clean();
         return NULL;
      }
      pdir.Mul( scale );
      newtlen = fcnmaxstep;
      if( DispFlag > 1 )
         cout << "Vector dir is too long; multiply by " << scale << endl;
      if( DumpFlag > 1 && dump != NULL )
         (*dump) << "Vector dir is too long; multiply by " << scale << endl;
   }

   // Compute the initial slope; exit if not negative

   double dginit = xcur_eval->GradientRef().Inner( pdir );
   
   if( dginit >= 0.0 )
   {
      if( DispFlag )
      {
         cout << "Initial slope nonnegative in line search" << endl;
         cout << "dginit: " << dginit << endl;
      }
      if( DumpFlag && dump != NULL )
      {
         (*dump) << "Initial slope nonnegative in line search" << endl;
         (*dump) << "dginit: " << dginit << endl;
      }
      TermCode = InitialSlopePositive;
      DisplayResults( 6 );
      Clean();
      return NULL;
   }
   double MinMu,MaxMu;
   if( ProtectedDivision( MinStep,newtlen,MinMu,ZeroDivideTol ) ||
       ProtectedDivision( fcnmaxstep,newtlen,MaxMu,ZeroDivideTol ) )
   {
      if( DispFlag > 1 )
         cout << "Protected division failed (computing MinMu and MaxMu)"
              << "   " << MinStep << "/" << newtlen << ", "
              << fcnmaxstep << "/" << newtlen << endl;
      if( DumpFlag && dump != NULL )
         (*dump) << "Protected division failed (computing MinMu and MaxMu)"
                 << "   " << MinStep << "/" << newtlen << ", "
                 << fcnmaxstep << "/" << newtlen << endl;

      TermCode = NoReduction;
      DisplayResults( 6 );
      Clean();
      return NULL;
   }
   

   double xnorm;
   
//   if (DispFlag > 1 || (DumpFlag > 1 && dump != NULL))
//      xnorm = xcur.Norm();
   if (DispFlag > 1 || (DumpFlag && dump != NULL))
      if( S != NULL )
      {
         S->Image( xcur,*tmpvec );
         xnorm = sqrt(xcur.Inner( *tmpvec ));
      }
      else 
         xnorm = xcur.Norm();

   if( DispFlag > 1 )
   {
      cout << "f = " << xcur_eval->ValueRef() << endl;
      cout << "||x|| = " << xnorm << endl;
      cout << "||dir|| = " << newtlen << endl;
      cout << "Initial slope = " << dginit << endl;
      cout << "Maximum step to boundary: " << fmaxs << endl;
      cout << "Minimum mu: " << MinMu << endl;
      cout << "Maximum mu: " << MaxMu << endl;
   }
   if( DumpFlag > 1 && dump != NULL )
   {
      (*dump) << "f = " << xcur_eval->ValueRef() << endl;
      (*dump) << "||x|| = " << xnorm << endl;
      (*dump) << "||dir|| = " << newtlen << endl;
      (*dump) << "Initial slope: " << dginit << endl;
      (*dump) << "Maximum step to boundary: " << fmaxs << endl;
      (*dump) << "Minimum mu: " << MinMu << endl;
      (*dump) << "Maximum mu: " << MaxMu << endl;
   }

   int bracket = 0;
   int stage1 = 1;
   int NumSamples = 0;
   double finit = xcur_eval->ValueRef();
   double dgtest = FcnDecreaseTol*dginit;
   double width = MaxStep - MinStep;
   double width1 = width*2.0;

   double mux = 0.0;
   double fx = finit;
   double dgx = dginit;
   double muy = 0.0;
   double fy = finit;
   double dgy = dginit;
   double mu = 1.0;
   mu = min( mu,MaxMu );
   mu = max( mu,MinMu );
   double mumin,mumax;

   MaxTkn = 0;
   int First = 1,infoc;
   HCL_EvaluateFunctional_d * eval = xcur_eval;
   double fv,dg;
   for(;;)
   {
      // Set the minimum and maximum steps to correspond to the present
      // interval of uncertainty.

      if( bracket )
      {
         mumin = min( mux,muy );
         mumax = max( mux,muy );
      }
      else
      {
         mumin = mux;
         mumax = mu + BracketIncrease*(mu-mux);
      }

      // Force the step to be within the bounds

      mu = max( mu,MinMu );
      mu = min( mu,MaxMu );

      // In case of failure, let mu correspond to the best point
      // Note: This is slightly different from the More'-Thuente
      // code.  In case the interval of uncertainty becomes too
      // small, we accept the new mu as the "best point", since
      // any point in the small interval should equally acceptable,
      // and it it likely that the new mu is better than the
      // previous best point.

//      if( (bracket && (mu <= mumin || mu >= mumax)) ||
//                NumSamples >= MaxSample-1 ||
//                (bracket && mumax-mumin <= IntervalTol*mumax) )
      if( (bracket && (mu <= mumin || mu >= mumax)) ||
           NumSamples >= MaxSample-1 )
         mu = mux;

      // Evaluate the function and gradient at mu

      xnext.AddScale( mu,xcur,pdir );
      if( !First )
         HCL_delete( eval );
      eval = f.Evaluate( xnext );
      First = 0;
      NumSamples++;
      fv = eval->ValueRef();
      dg = eval->GradientRef().Inner( pdir );
      double ftest1 = finit + mu*dgtest;
      if( DispFlag > 1 )
         cout << NumSamples << " mu = " << mu << " f = " << fv << " slope = " 
              << dg << endl;
      if( DumpFlag > 1 && dump != NULL )
         (*dump) << NumSamples << " mu = " << mu << " f = " << fv
                 << " slope = " << dg << endl;

      // Check for convergence

      int info = 0;
      if( (bracket && (mu <= mumin || mu >= mumax)) )
         info = 6;
      else if( mu == MaxMu && fv <= ftest1 && dg <= dgtest )
      {
         info = 5;
         MaxTkn = 1;
      }
      else if( mu == MinMu && (fv > ftest1 || dg >= dgtest) )
         info = 4;
      else if( NumSamples >= MaxSample )
         info = 3;
      else if( bracket && mumax-mumin <= IntervalTol*mumax )
         info = 2;

      if( info )
      {
         if( fv >= finit )
            TermCode = NoReduction;
         else if( info == 3 )
            TermCode = TooManyFunctionEvals;
         else
            TermCode = Fail;

         DisplayResults( info );
         Clean();
         if( TermCode == NoReduction )
         {
            HCL_delete( eval );
            return NULL;
         }
         else
            return eval;
      }
      if( fv <= ftest1 && abs(dg) <= SlopeDecreaseTol*(-dginit) )
      {
         TermCode = Succeed;
         DisplayResults( 1 );
         Clean();
         return eval;
      }

      // In the first stage, we seek a step for which the modified
      // function has a nonpositive value and nonnegative derivative.

      if( stage1 && fv <= ftest1 &&
          dg >= min( FcnDecreaseTol,SlopeDecreaseTol)*dginit )
         stage1 = 0;

      // We use the modified function to predict the step only if
      // we do not have a step for which the modified function has
      // a nonpositive function value and nonnegative derivative,
      // and if a lower function value has been obtained but the
      // decrease is not sufficient.

      if( stage1 && fv <= fx && fv >= ftest1 )
      {
         // Define the modified function and derivative values.

         double fm = fv - mu*dgtest;
         double fxm = fx - mux*dgtest;
         double fym = fy - muy*dgtest;
         double dgm = dg - dgtest;
         double dgxm = dgx - dgtest;
         double dgym = dgy - dgtest;

         // Update the interval of uncertainty and compute the next step.

         infoc = cstep( mux,fxm,dgxm,muy,fym,dgym,mu,fm,dgm,bracket,
                        mumin,mumax );

         // Reset the function and gradient values.

         fx = fxm + mux*dgtest;
         fy = fym + muy*dgtest;
         dgx = dgxm + dgtest;
         dgy = dgym + dgtest;
      }
      else  // Update the interval of uncertainty and compute the new step.
         infoc = cstep( mux,fx,dgx,muy,fy,dgy,mu,fv,dg,bracket,mumin,mumax );

      // Force a sufficient decrease in the size of the interval of
      // uncertainty.

      if( bracket )
      {
         if( abs(muy-mux) >= 0.66*width1 )
            mu = mux + 0.5*(muy-mux);
         width1 = width;
         width = abs(muy-mux);
      }
   }
}

int HCL_LineSearch_MT_d::cstep( double & mux,double & fx,double & dgx,
                                double & muy,double & fy,double & dgy,
                                double & mu,double & fv,double & dg,
                                int & bracket,double & mumin,double & mumax )
{
   // Check the input for errors

   if( bracket && (mu <= min(mux,muy) || mu >= max(mux,muy)) ||
       dgx*(mu-mux) >= 0.0 || mumax < mumin )
   {
      cerr << "Error in HCL_LineSearch_MT_d::cstep: error in input" << endl;
      exit(1);
   }

   // Determine if derivatives have opposite signs.

   double sgnd = dg;
   if( dgx < 0.0 ) sgnd = -dg;
   int infoc,bound;
   double muf;
   int FPErr = 0;

   // First case: a higher function value.
   // The minimum is bracketed.  If the cubic step is closer
   // to mux than the quadratic step, the cubic step is taken,
   // otherwise the average of the cubic and quadratic steps is taken.

   if( fv > fx )
   {
      if( DispFlag > 1 )
         cout << "   cstep: Case 1" << endl;
      if( DumpFlag > 1 && dump )
         cout << "   cstep: Case 1" << endl;

      infoc = 1;
      bracket = 1;
      bound = 1;
      double theta;
      if(FPErr = ProtectedDivision(3.0*(fx-fv),mu-mux,theta,ZeroDivideTol))
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 1) theta"
                 << endl << "   " << 3.0*(fx-fv) << "/" << mu-mux << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 1) theta"
                    << endl << "   " << 3.0*(fx-fv) << "/" << mu-mux << endl;
         goto FPErrTarget;
      }
      theta = theta + dgx + dg;
      double s = max(max(abs(theta),abs(dgx)),abs(dg));
      double t1,t2,t3;
      if( FPErr = ProtectedDivision( theta,s,t1,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 1) gamma (t1)"
                 << endl << "   " << theta << "/" << s << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 1) gamma (t1)"
                    << endl << "   " << theta << "/" << s << endl;
         goto FPErrTarget;
      }
      if( FPErr = ProtectedDivision( dgx,s,t2,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 1) gamma (t2)"
                 << endl << "   " << dgx << "/" << s << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 1) gamma (t2)"
                    << endl << "   " << dgx << "/" << s << endl;
         goto FPErrTarget;
      }
      if( FPErr = ProtectedDivision( dg,s,t3,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 1) gamma (t3)"
                 << endl << "   " << dg << "/" << s << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 1) gamma (t3)"
                    << endl << "   " << dg << "/" << s << endl;
         goto FPErrTarget;
      }
      double gamma = s*sqrt(t1*t1-t2*t3);
      if( mu < mux ) gamma = -gamma;
      double p = (gamma-dgx)+theta;
      double q = ((gamma-dgx)+gamma)+dg;
      double r;
      if( FPErr = ProtectedDivision( p,q,r,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 1) r"
                 << endl << "   " << p << "/" << q << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 1) r"
                    << endl << "   " << p << "/" << q << endl;
         goto FPErrTarget;
      }
      double muc = mux + r*(mu-mux);
      if( FPErr = ProtectedDivision( fx-fv,mu-mux,t1,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 1) muq (t1)"
                 << endl << "   " << fx-fv << "/" << mu-mux << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 1) muq (t1)"
                    << endl << "   " << fx-fv << "/" << mu-mux << endl;
         goto FPErrTarget;
      }
      t1 += dgx;
      if( FPErr = ProtectedDivision( dgx,t1,t2,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 1) muq (t2)"
                 << endl << "   " << dgx << "/" << t1 << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 1) muq (t2)"
                    << endl << "   " << dgx << "/" << t1 << endl;
         goto FPErrTarget;
      }
      double muq = mux + (t2/2.0)*(mu-mux);
      if( abs( muc-mux) < abs(muq-mux) )
         muf = muc;
      else
         muf = muc + (muq-muc)/2.0;
   }

   // The second case: a lower function value and derivatives of opposite
   // sign.  The minimum is bracketed.  If the cubic step is
   // closer to mux than the quadratic (secant) step, then the
   // cubic step is taken.  Otherwise the quadratic step is taken.

   else if( sgnd < 0.0 )
   {
      if( DispFlag > 1 )
         cout << "   cstep: Case 2" << endl;
      if( DumpFlag > 1 && dump )
         cout << "   cstep: Case 2" << endl;

      infoc = 2;
      bracket = 1;
      bound = 0;
      double theta;
      if( FPErr = ProtectedDivision( 3.0*(fx-fv),mu-mux,theta,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 2) theta"
                 << endl << "   " << 3.0*(fx-fv) << "/" << mu-mux << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 2) theta"
                    << endl << "   " << 3.0*(fx-fv) << "/" << mu-mux << endl;
         goto FPErrTarget;
      }
      theta = theta + dgx + dg;
      double s = max(max(abs(theta),abs(dgx)),abs(dg));
      double t1,t2,t3;
      if( FPErr = ProtectedDivision( theta,s,t1,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 2) gamma (t1)"
                 << endl << "   " << theta << "/" << s << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 2) gamma (t1)"
                    << endl << "   " << theta << "/" << s << endl;
         goto FPErrTarget;
      }
      if( FPErr = ProtectedDivision( dgx,s,t2,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 2) gamma (t2)"
                 << endl << "   " << dgx << "/" << s << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 2) gamma (t2)"
                    << endl << "   " << dgx << "/" << s << endl;
         goto FPErrTarget;
      }
      if( FPErr = ProtectedDivision( dg,s,t3,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 2) gamma (t3)"
                 << endl << "   " << dg << "/" << s << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 2) gamma (t3)"
                    << endl << "   " << dg << "/" << s << endl;
         goto FPErrTarget;
      }
      double gamma = s*sqrt(t1*t1-t2*t3);
      if( mu > mux ) gamma = -gamma;
      double p = (gamma-dg)+theta;
      double q = ((gamma-dg)+gamma)+dgx;
      double r;
      if( FPErr = ProtectedDivision( p,q,r,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 2) r"
                 << endl << "   " << p << "/" << q << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 2) r"
                    << endl << "   " << p << "/" << q << endl;
         goto FPErrTarget;
      }
      double muc = mu + r*(mux-mu);
      if( FPErr = ProtectedDivision( dg,dg-dgx,t1,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 2) muq (t1)"
                 << endl << "   " << dg << "/" << dg-dgx << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 2) muq (t1)"
                    << endl << "   " << dg << "/" << dg-dgx << endl;
         goto FPErrTarget;
      }
      double muq = mu + t1*(mux-mu);
      if( abs(muc-mu) > abs(muq-mu) )
         muf = muc;
      else
         muf = muq;
   }

   // Third case: a lower function value, derivatives of the same sign,
   // and the magnitude of the derivative decreases.  The cubic step
   // is only used if the cubic tends to infinity in the direction
   // of the step or if the minimum of the cubic is beyond mu.
   // Otherwise the cubic step is defined to be either mumin or mumax.
   // The quadratic (secant) step is also computed and if the minimum
   // is bracketed, then the step closest to mu x is taken, otherwise
   // the step farthest away is taken.

   else if( abs(dg) < abs(dgx) )
   {
      if( DispFlag > 1 )
         cout << "   cstep: Case 3" << endl;
      if( DumpFlag > 1 && dump )
         cout << "   cstep: Case 3" << endl;

      infoc = 3;
      bound = 1;
      double theta;
      if( FPErr = ProtectedDivision( 3.0*(fx-fv),mu-mux,theta,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 3) theta"
                 << endl << "   " << 3.0*(fx-fv) << "/" << mu-mux << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 3) theta"
                    << endl << "   " << 3.0*(fx-fv) << "/" << mu-mux << endl;
         goto FPErrTarget;
      }
      theta = theta + dgx + dg;
      double s = max(max(abs(theta),abs(dgx)),abs(dg));
      double t1,t2,t3;
      if( FPErr = ProtectedDivision( theta,s,t1,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 3) gamma (t1)"
                 << endl << "   " << theta << "/" << s << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 3) gamma (t1)"
                    << endl << "   " << theta << "/" << s << endl;
         goto FPErrTarget;
      }
      if( FPErr = ProtectedDivision( dgx,s,t2,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 3) gamma (t2)"
                 << endl << "   " << dgx << "/" << s << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 3) gamma (t2)"
                    << endl << "   " << dgx << "/" << s << endl;
         goto FPErrTarget;
      }
      if( FPErr = ProtectedDivision( dg,s,t3,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 3) gamma (t3)"
                 << endl << "   " << dg << "/" << s << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 3) gamma (t3)"
                    << endl << "   " << dg << "/" << s << endl;
         goto FPErrTarget;
      }
      double gamma = s*sqrt(max((double)0.0,(double)(t1*t1-t2*t3)));
      if( mu > mux ) gamma = -gamma;
      double p = (gamma-dg)+theta;
      double q = (gamma+(dgx-dg))+gamma;
      double r;
      if( FPErr = ProtectedDivision( p,q,r,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 3) r"
                 << endl << "   " << p << "/" << q << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 3) r"
                    << endl << "   " << p << "/" << q << endl;
         goto FPErrTarget;
      }
      double muc;
      if( r < 0.0 && gamma != 0.0 )
         muc = mu + r*(mux-mu);
      else if( mu > mux )
         muc = mumax;
      else
         muc = mumin;
      if( FPErr = ProtectedDivision( dg,dg-dgx,t1,ZeroDivideTol ) )
      {
         if( DispFlag > 1 )
            cout << "Protected division failed: cstep (case 3) muq (t1)"
                 << endl << "   " << dg << "/" << dg-dgx << endl;
         if( DumpFlag > 1 && dump != NULL )
            (*dump) << "Protected division failed: cstep (case 3) muq (t1)"
                    << endl << "   " << dg << "/" << dg-dgx << endl;
         goto FPErrTarget;
      }
      double muq = mu + t1*(mux-mu);
      if( bracket )
      {
         if( abs(muc-mu) < abs(muq-mu) )
            muf = muc;
         else
            muf = muq;
      }
      else if( abs(mu-muc) > abs(mu-muq) )
         muf = muc;
      else
         muf = muq;
   }

   // Fourth case. a lower function value, derivatives of the same
   // sign, and the magnitude of the derivative does not decrease.
   // If the minimum is not bracketed, the step is either mumin or
   // mumax.  Otherwise the cubic step is taken.

   else
   {
      if( DispFlag > 1 )
         cout << "   cstep: Case 4" << endl;
      if( DumpFlag > 1 && dump )
         cout << "   cstep: Case 4" << endl;
      infoc = 4;
      bound = 0;
      if( bracket )
      {
         double theta;
         if(FPErr = ProtectedDivision(3.0*(fv-fy),muy-mu,theta,ZeroDivideTol))
         {
            if( DispFlag > 1 )
               cout << "Protected division failed: cstep (case 4) theta"
                    << endl << "   " << 3.0*(fv-fy) << "/" << muy-mu << endl;
            if( DumpFlag > 1 && dump != NULL )
               (*dump) << "Protected division failed: cstep (case 4) theta"
               << endl << "   " << 3.0*(fv-fy) << "/" << muy-mu << endl;
            goto FPErrTarget;
         }
         theta = theta + dgy + dg;
         double s = max(max(abs(theta),abs(dgy)),abs(dg));
         double t1,t2,t3;
         if( FPErr = ProtectedDivision( theta,s,t1,ZeroDivideTol ) )
         {
            if( DispFlag > 1 )
               cout << "Protected division failed: cstep (case 4) gamma (t1)"
                    << endl << "   " << theta << "/" << s << endl;
            if( DumpFlag > 1 && dump != NULL )
              (*dump) << "Protected division failed: cstep (case 4) gamma (t1)"
                      << endl << "   " << theta << "/" << s << endl;
            goto FPErrTarget;
         }
         if( FPErr = ProtectedDivision( dgy,s,t2,ZeroDivideTol ) )
         {
            if( DispFlag > 1 )
               cout << "Protected division failed: cstep (case 4) gamma (t2)"
                    << endl << "   " << dgy << "/" << s << endl;
            if( DumpFlag > 1 && dump != NULL )
             (*dump) << "Protected division failed: cstep (case 4) gamma (t2)"
                     << endl << "   " << dgy << "/" << s << endl;
            goto FPErrTarget;
         }
         if( FPErr = ProtectedDivision( dg,s,t3,ZeroDivideTol ) )
         {
            if( DispFlag > 1 )
               cout << "Protected division failed: cstep (case 4) gamma (t3)"
                    << endl << "   " << dg << "/" << s << endl;
            if( DumpFlag > 1 && dump != NULL )
             (*dump) << "Protected division failed: cstep (case 4) gamma (t3)"
                     << endl << "   " << dg << "/" << s << endl;
            goto FPErrTarget;
         }
         double gamma = s*sqrt(t1*t1-t2*t3);
         if( mu > muy ) gamma = -gamma;
         double p = (gamma-dg)+theta;
         double q = ((gamma-dg)+gamma)+dgy;
         double r;
         if( FPErr = ProtectedDivision( p,q,r,ZeroDivideTol ) )
         {
            if( DispFlag > 1 )
               cout << "Protected division failed: cstep (case 4) r"
                    << endl << "   " << p << "/" << q << endl;
            if( DumpFlag > 1 && dump != NULL )
               (*dump) << "Protected division failed: cstep (case 4) r"
                       << endl << "   " << p << "/" << q << endl;
            goto FPErrTarget;
         }
         double muc = mu + r*(muy-mu);
         muf = muc;
      }
      else if( mu > mux )
         muf = mumax;
      else
         muf = mumin;
   }

   FPErrTarget: ;

   // Update the interval of uncertainty.  This update does not
   // depend on the new step or the case analysis above.

   if( fv > fx )
   {
      muy = mu;
      fy = fv;
      dgy = dg;
   }
   else
   {
      if( sgnd < 0.0 )
      {
         muy = mux;
         fy = fx;
         dgy = dgx;
      }
      mux = mu;
      fx = fv;
      dgx = dg;
   }

   // Bisect the interval if there was a floating point error in
   // the calculation of muf

   if( FPErr && bracket )
      muf = 0.5*(mux+muy);
   else if( FPErr )
      muf = mumax;

   // Compute the new step and safeguard it.
  
   muf = min((double)mumax,(double)muf);
   muf = max((double)mumin,(double)muf);
   mu = muf;
   if( bracket && bound )
      if( muy > mux )
         mu = min((double)(mux+0.66*(muy-mux)),(double)mu);
      else
         mu = max((double)(mux+0.66*(muy-mux)),(double)mu);
   return infoc;
}

ostream & HCL_LineSearch_MT_d::Write( ostream & str ) const
{
   str << "HCL_LineSearch_MT_d More-Thuente line search algorithm"
       << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}
