/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#include "HCL_TrustRegion_SS_d.h"

HCL_TRSSOp_d::HCL_TRSSOp_d()
{
   cerr << "Error: HCL_TRSSOp_d default constructor called" << endl;
   exit(1);
}

HCL_TRSSOp_d::HCL_TRSSOp_d( const HCL_TRSSOp_d & )
{
   cerr << "Error: HCL_TRSSOp_d copy constructor called" << endl;
   exit(1);
}

HCL_TRSSOp_d::HCL_TRSSOp_d( const HCL_LinearOp_d * AA, const HCL_Vector_d * gg,
                            double a,HCL_LinearOp_d * CC,
                            HCL_LinearSolver_d * ls )
{
   if( AA->Range() != AA->Domain() )
   {
      cerr << "Error in HCL_TRSSOp_d constructor: domain and range of operator"
              " must agree" << endl;
      exit(1);
   }
   if( gg->Space() != AA->Domain() )
   {
      cerr << "Error in HCL_TRSSOp_d constructor: vector must lie in domain "
              "of operator" << endl;
      exit(1);
   }
   HCL_RnSpace_d * R1 = new HCL_RnSpace_d( 1 );
   dom = new HCL_GenericProductSpace_d( *R1,AA->Domain() );
   HCL_delete( R1 );
   ParamTable = new Table;
   ParamTable->PutValue( "alpha",a );
   A = (HCL_LinearOp_d*)AA;A->IncCount();
   g = (HCL_Vector_d*)gg;g->IncCount();
   if( CC != NULL )
   {
      C = CC;
      C->IncCount();
   }
   else
      C = NULL;
   if( ls && CC )
   {
      lsolver = ls;
      ls->IncCount();
   }
   else
      lsolver = NULL;
}

HCL_TRSSOp_d::~HCL_TRSSOp_d()
{
   HCL_delete( lsolver );
   HCL_delete( C );
   HCL_delete( g );
   HCL_delete( A );
   HCL_delete( dom );
   HCL_delete( ParamTable );
}

Table & HCL_TRSSOp_d::Parameters() const
{
   return *ParamTable;
}

HCL_VectorSpace_d & HCL_TRSSOp_d::Domain() const
{
   return *dom;
}

HCL_VectorSpace_d & HCL_TRSSOp_d::Range() const
{
   return *dom;
}

void HCL_TRSSOp_d::Image( const HCL_Vector_d & xx,
                                HCL_Vector_d & yy ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_TRSSOp_d::Image: vector does not lie in domain"
           << endl;
      exit(1);
   }
   HCL_GenericProductVector_d & x = (HCL_GenericProductVector_d&)xx;
   HCL_RnVector_d & gamma = (HCL_RnVector_d&)x(1);
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_TRSSOp_d::Image: vector does not lie in range"
           << endl;
      exit(1);
   }
   HCL_GenericProductVector_d & y = (HCL_GenericProductVector_d&)yy;
   HCL_RnVector_d & gamma1 = (HCL_RnVector_d&)y(1);

   double a;
   ParamTable->GetValue( "alpha",a );
   gamma1(1) = a*gamma(1) + g->Inner( x(2) );

   if( C == NULL )
   {
      A->Image( x(2),y(2) );
      y(2).AddScale( gamma(1),*g );
   }
   else if( lsolver )
   {
      HCL_Vector_d * tmp = A->Range().Member();
      A->Image( x(2),*tmp );
      tmp->AddScale( gamma(1),*g );
      lsolver->Solve( *C,*tmp,y(2) );
      HCL_delete( tmp );
   }
   else
   {
      HCL_Vector_d * tmp = A->Range().Member();
      A->Image( x(2),*tmp );
      tmp->AddScale( gamma(1),*g );
      C->InvImage( *tmp,y(2) );
      HCL_delete( tmp );
   }
}

void HCL_TRSSOp_d::AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const
{
   Image( y,x );
}

void HCL_TRSSOp_d::InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const
{
   cerr << "Error: HCL_TRSSOp_d::InvImage is not implemented" << endl;
   exit(1);
}

void HCL_TRSSOp_d::InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const
{
   cerr << "Error: HCL_TRSSOp_d::InvAdjImage is not implemented" << endl;
   exit(1);
}

ostream & HCL_TRSSOp_d::Write( ostream & str ) const
{
   double a;
   ParamTable->GetValue( "alpha",a );
   str << "HCL_TRSSOp_d defined by alpha = " << a << endl;
   str << "g = " << endl;
   g->Write( str );
   str << "A = " << endl;
   A->Write( str );
   return str;
}

HCL_TrustRegion_SS_d::HCL_TrustRegion_SS_d( const HCL_TrustRegion_SS_d & x )
{
   cerr << "Error: HCL_TrustRegion_SS_d copy constructor called" << endl;
   exit(1);
}

HCL_TrustRegion_SS_d::HCL_TrustRegion_SS_d( char * fname )
{
   if( fname != NULL )
      ParamTable = new Table( fname,"TRSolver" );
   else
      ParamTable = new Table;
   C = NULL;
   lsolver = NULL;
   eigsolver = new HCL_IRArnoldi_d( fname );
   eigsolver->Parameters().PutValue( "Which","SR" );
   eigsolver->Parameters().PutValue( "SubspaceDim",10 );
}

HCL_TrustRegion_SS_d::~HCL_TrustRegion_SS_d()
{
   HCL_delete( C );
   HCL_delete( lsolver );
   HCL_delete( eigsolver );
   HCL_delete( ParamTable );
}

Table & HCL_TrustRegion_SS_d::Parameters() const
{
   return *ParamTable;
}

void HCL_TrustRegion_SS_d::Initialize( HCL_VectorSpace_d & dom )
{
   if( Parameters().GetValue( "Delta",Delta ) )
      Delta = 1.0;
   if( Parameters().GetValue( "MaxItn",MaxItn ) )
      MaxItn = 20;
   if( Parameters().GetValue( "DispFreq",DispFreq ) )
      DispFreq =  MaxItn/5;
   if( Parameters().GetValue( "DispFlag",DispFlag ) )
      DispFlag = 0;
   if( Parameters().GetValue( "DispPrecision",DispPrecision ) )
      DispPrecision = 6;
   OldDispPrec = cout.precision( DispPrecision );
   if( Parameters().GetValue( "HardCaseTol",HardCaseTol ) )
      HardCaseTol = 0.01;
   if( Parameters().GetValue( "Tol",Tol ) )
      Tol = 0.01;

   double eigtol;
   if( eigsolver->Parameters().GetValue( "Tol",eigtol ) )
   {
      eigtol = Tol/10.0;
      eigsolver->Parameters().PutValue( "Tol",eigtol );
   }
   ZeroTol = eigtol;

   if( C != NULL )
      if( C->Domain() != dom )
      {
         cerr << "Error in HCL_TrustRegion_SS_d::Initialize: domain does "
                 "not match preconditioner" << endl;
         exit(1);
      }

   gammaVec = new HCL_RnVector_d( 1 );
   y = NULL;   
   Balpha = NULL;
   xtemp = dom.Member();
}

void HCL_TrustRegion_SS_d::DumpProblem( const HCL_LinearOp_d & A,
                                        const HCL_Vector_d & g,
                                              HCL_Vector_d & x )
{
      cout << "HCL_TrustRegion_SS_d" << endl;
      cout << "Delta = " << Delta << endl;
      cout << "A = " << endl;
      A.Write( cout );
      cout << "g = " << endl;
      g.Write( cout );
      cout << "alpha = " << alpha << endl;
      if( C != NULL )
      {
         cout << "Preconditioner is " << endl;
         C->Write( cout );
      }
      cout << "lambda = " << lam1 << endl;
      cout << "x = " << endl;
      x.Write( cout );
}

void HCL_TrustRegion_SS_d::Clean()
{
   cout.precision( OldDispPrec );
   HCL_delete( xtemp );
   HCL_delete( Balpha );
   HCL_delete( y );
   HCL_delete( gammaVec );
}

void HCL_TrustRegion_SS_d::SetScaling( HCL_LinearOp_d * S,
                                       HCL_LinearSolver_d * ls )
{
   HCL_delete( C );
   HCL_delete( lsolver );
   C = S;S->IncCount();
   lsolver = ls;if( ls ) ls->IncCount();
}
   
void HCL_TrustRegion_SS_d::UnSetScaling()
{
   HCL_delete( C );
   HCL_delete( lsolver );
   C = NULL;
   lsolver = NULL;
}

int HCL_TrustRegion_SS_d::GetInteriorSolution( const HCL_LinearOp_d & A,
                                               const HCL_Vector_d & g,
                                                     HCL_Vector_d & p )
{
   HCL_TRCG_d cgsolver;
   if( C != NULL )
      cgsolver.SetScaling( C,lsolver );
   cgsolver.Parameters().PutValue( "Tol",Tol );
   cgsolver.Parameters().PutValue( "MaxItn",100 );
   cgsolver.Parameters().PutValue( "DispFlag",DispFlag );
   cgsolver.Parameters().PutValue( "Delta",Delta );
   cgsolver.Solve( A,g,p );
   int flag;
   cgsolver.Parameters().GetValue( "TermCode",flag );
   return flag;
}

double HCL_TrustRegion_SS_d::GetxNorm( const HCL_Vector_d & x ) const
{
   double v;
   if( C != NULL )
   {
      C->Image( x,*xtemp );
      v = sqrt( x.Inner( *xtemp ) );
   }
   else
      v = x.Norm();
   return v;
}

void HCL_TrustRegion_SS_d::Solve( const HCL_LinearOp_d & A,
                                  const HCL_Vector_d & g,HCL_Vector_d & x )
{
   if( x.Space() != A.Domain() )
   {
      cerr << "Error in HCL_TrustRegion_SS_d::Solve: vector does not belong "
              "to domain" << endl;
      exit(1);
   }
   if( g.Space() != A.Domain() )
   {
      cerr << "Error in HCL_TrustRegion_SS_d::Solve: gradient does not belong "
              "to domain" << endl;
      exit(1);
   }

   // Initialize

   double lam_hat=0.0;
   Initialize( A.Domain() );
   if( lsolver )
   {
      lsolver->Solve( *C,g,*xtemp );
      gnorm = sqrt( g.Inner( *xtemp ) );
   }
   else if( C )
   {
      C->InvImage( g,*xtemp );
      gnorm = sqrt( g.Inner( *xtemp ) );
   }
   else
      gnorm = g.Norm();

   y = new HCL_GenericProductVector_d( gammaVec,&x );

   // Compute an upper bound for alpha
   // ??? Need to adjust this for a scaled norm ???

   y->Random();
   A.Image( (*y)(2),*xtemp );
   double delta_S = (xtemp->Inner( (*y)(2) ))/((*y)(2).Norm2());
   alpha_u = delta_S + Delta*gnorm;

   // Get the initial value of alpha

   if( Parameters().GetValue( "alpha",alpha ) )
      alpha = min( 0.0,alpha_u );

   // Creat Balpha, the bordered matrix parameterized by alpha

   if( C )
      Balpha = new HCL_TRSSOp_d( &A,&g,alpha,C,lsolver );
   else
      Balpha = new HCL_TRSSOp_d( &A,&g,alpha );

   // Solve the eigenvalue problem to find the two smallest
   // eigenvalues of Balpha

   int nev = 5;
   if( (eigsolver->Solve( nev,*Balpha,*y )) )
   {
      if( DispFlag )
         cout << "Error in HCL_TrustRegion_SS_d::Solve: Eigenvalue solver "
                 "failed on iteration 0" << endl;
      if( DispFlag > 3 )
         DumpProblem( A,g,x );
      Parameters().PutValue( "Reduction",0.0 );
      Parameters().PutValue( "StepLength",0.0 );
      Parameters().PutValue( "TermCode",FAILED );
      Clean();
      return;
   }
   eigsolver->Eigenvalue( 1,lam1 );
   eigsolver->Eigenvalue( 2,lam2 );
   int oneptr = 1,twoptr = 2;
   if( lam2 < lam1 )
   {
      lam2 = lam1;
      eigsolver->EigPair( 2,lam1,y );
      oneptr = 2;
      twoptr = 2;
   }
   else
      eigsolver->EigPair( 1,lam1,y );

   // Initialize the safeguard bounds
   // ??? Need to adjust this for a scaled norm ???

   alpha_l = lam1-gnorm/Delta;
   lam_l = lam1-gnorm*(1.0/Delta+Delta);
   lam_u = min( delta_S,0.0 );

   double nu1 = (*gammaVec)(1);
   xnorm = GetxNorm( (*y)(2) );

   if( DispFlag > 1 )
      cout << "HCL_TrustRegion_SS_d" << endl
           << "   Iteration: 0" << endl
           << "      alpha = " << alpha << endl
           << "      Smallest eigenvalues are " 
           << lam1  << " and " << lam2 << endl
           << "      nu1 = " << nu1 << endl
           << "      Current safeguards: " << endl
           << "         [lam_l,lam_u] = [" << lam_l << ","
           << lam_u << "]" << endl
           << "         [alpha_l,alpha_u] = [" << alpha_l << ","
           << alpha_u << "]" << endl;

   // Check for the hard case

   if( max( gnorm*abs(nu1),abs(nu1) ) <= HardCaseTol*xnorm )
   {
      if( DispFlag > 1 )
         cout << "   Hard case encountered;   nu1 = " << nu1 << endl;
      eigsolver->EigPair( twoptr,lam1,y );
      nu1 = (*gammaVec)(1);
      xnorm = GetxNorm( (*y)(2) );
   }

   // Update alpha, checking for interior solution

   y->Mul( 1/nu1 );
   xnorm /= abs(nu1);

   if( (xnorm < Delta && lam1 >= -ZeroTol) || lam_l >= 0.0 )
   {
      int flag = GetInteriorSolution( A,g,x );
      xnorm = GetxNorm( x );
      A.Image( x,*xtemp );
      double red = -(0.5*(x.Inner( *xtemp )) + g.Inner( x ));
      Parameters().PutValue( "Reduction",red );
      Parameters().PutValue( "StepLength",xnorm );
      Parameters().PutValue( "alpha",alpha );
      Parameters().PutValue( "Multiplier",-lam_hat );
      Parameters().PutValue( "TermCode",INTERIOR_SOLUTION );
      Clean();
      if( DispFlag > 0 )
         cout << "HCL_TrustRegion_SS_d" << endl
              << "   Iteration: 0" << endl
              << "      Interior solution found" << endl
              << "      Reduction = " << red << endl
              << "      StepLength = " << xnorm << endl;
      return;
   }

   phi = -g.Inner( x );
   dphi = xnorm*xnorm;

   if( DispFlag > 1 )
      cout << "      xnorm = " << xnorm << endl
           << "      phi = " << phi << " dphi = " << dphi << endl;

   if( xnorm < Delta )
   {
      lam_l = lam1;
      alpha_l = alpha;
   }
   else
   {
      lam_u = min( lam1,0.0 );
      alpha_u = alpha;
   }

   alpha = alpha + (alpha-lam1)/xnorm*((Delta-xnorm)/Delta)*
                                      (Delta+1.0/xnorm);

   if( alpha > alpha_u || alpha < alpha_l )
   {
      if( DispFlag > 2 )
         cout << "   alpha exceeds safeguards; safeguarding (alpha = "
              << alpha << ")" << endl;
      alpha = (alpha_l+alpha_u)/2.0;
   }

   Balpha->Parameters().PutValue( "alpha",alpha ); 
   phi_prev = phi;
   dphi_prev = dphi;
   lam1_prev = lam1;
   xnorm_prev = xnorm;
   
   // Compute the two smallest eigenvalues of Balpha

   int itn = 1;
   y->Random();
   nev = 5;
   if( (eigsolver->Solve( nev,*Balpha,*y )) )
   {
      if( DispFlag )
         cout << "Error in HCL_TrustRegion_SS_d::Solve: Eigenvalue solver "
                 "failed on iteration " << itn << endl;
      if( DispFlag > 3 )
         DumpProblem( A,g,x );
      Parameters().PutValue( "Reduction",0.0 );
      Parameters().PutValue( "StepLength",0.0 );
      Parameters().PutValue( "TermCode",FAILED );
      Clean();
      return;
   }
   eigsolver->Eigenvalue( 1,lam1 );
   eigsolver->Eigenvalue( 2,lam2 );
   oneptr = 1;
   twoptr = 2;
   if( lam2 < lam1 )
   {
      lam2 = lam1;
      eigsolver->EigPair( 2,lam1,y );
      oneptr = 2;
      twoptr = 1;
   }
   else
      eigsolver->EigPair( 1,lam1,y );

   // Update safeguards (PROTECTED DIVISION????)
   // ??? Need to adjust this for a scaled norm ???

   nu1 = (*gammaVec)(1);
   xnorm = GetxNorm( (*y)(2) );
   double test = xnorm/abs(nu1) - Delta;

   if(  test < 0.0 )
   {
      lam_l = lam1;
      alpha_l = alpha;
   }
   else
   {
      lam_u = min( lam1,0.0 );
      alpha_u = alpha;
   }
   int done = 0;
   if( abs( test ) < Tol*Delta )
      done = 1;

   while( !done )
   {
      if( DispFlag > 1 )
         cout << "HCL_TrustRegion_SS_d" << endl
              << "   Iteration: " << itn  << endl
              << "      Relative error in ||x|| = " << abs(test)/Delta << endl
              << "      alpha = " << alpha << endl
              << "      Smallest eigenvalues are " 
              << lam1  << " and " << lam2 << endl
              << "      nu1 = " << nu1 << endl
              << "      Current safeguards: " << endl
              << "         [lam_l,lam_u] = [" << lam_l << ","
              << lam_u << "]" << endl
              << "         [alpha_l,alpha_u] = [" << alpha_l << ","
              << alpha_u << "]" << endl;

      // Check for hard case

      if( max( gnorm*abs(nu1),abs(nu1) ) <= HardCaseTol*xnorm )
      {
         if( DispFlag > 1 )
            cout << "   Hard case encountered;   nu1 = " << nu1 << endl;
         eigsolver->EigPair( twoptr,lam1,y );
         nu1 = (*gammaVec)(1);
         xnorm = GetxNorm( (*y)(2) );
      }

      // Update alpha, checking for interior solution

      y->Mul( 1/nu1 );
      xnorm /= abs(nu1);

      if( (xnorm < Delta && lam1 >= -ZeroTol) || lam_l >= 0.0 )
      {
         int flag = GetInteriorSolution( A,g,x );
         xnorm = GetxNorm( x );
         A.Image( x,*xtemp );
         double red = -(0.5*(x.Inner( *xtemp )) + g.Inner( x ));
         Parameters().PutValue( "Reduction",red );
         Parameters().PutValue( "StepLength",xnorm );
         Parameters().PutValue( "alpha",alpha );
         Parameters().PutValue( "Multiplier",-lam_hat );
         Parameters().PutValue( "TermCode",INTERIOR_SOLUTION );
         Clean();
         if( DispFlag > 0 )
            cout << "HCL_TrustRegion_SS_d" << endl
                 << "   Iteration: " << itn << endl
                 << "      Interior solution found" << endl
                 << "      Reduction = " << red << endl
                 << "      Multiplier = " << -lam_hat << endl
                 << "      StepLength = " << xnorm << endl;
         return;
      }

      phi = -g.Inner( x );
      dphi = xnorm*xnorm;
      delta_S = min( delta_S,lam1+phi/dphi );
      lam_hat = (lam1_prev*xnorm_prev*(xnorm-Delta)+
                 lam1*xnorm*(Delta-xnorm_prev))/(Delta*(xnorm-xnorm_prev));
      if( DispFlag > 1 )
         cout << "      xnorm = " << xnorm << endl
              << "      phi = " << phi << " dphi = " << dphi << endl
              << "      lam_hat (before safeguarding) = " << lam_hat << endl;
//      lam_hat = min( min( lam_hat,delta_S ),0.0 );
      if( lam_hat > lam_u ) lam_hat = lam_u;
      else if( lam_hat < lam_l ) lam_hat = lam_l;

      double w = (lam1-lam_hat)/(lam1-lam1_prev);
      double alpha1 = lam_hat + w*phi_prev+(1-w)*phi+
                      (xnorm_prev*xnorm*(xnorm-xnorm_prev))/
                      (w*xnorm+(1-w)*xnorm_prev)*
                      (lam1_prev-lam_hat)*(lam1-lam_hat)/
                      (lam1-lam1_prev);

/*      double gamma2 = (lam1-lam1_prev)*xnorm_prev*xnorm/(xnorm-xnorm_prev);
      gamma2 = gamma2*gamma2;
      double delta = (lam1*xnorm-lam1_prev-xnorm_prev)/(xnorm-xnorm_prev);
      double etat = phi-gamma2/(delta-lam1);
      double etat_prev = phi_prev-gamma2/(delta-lam1_prev);
      double eta = w*etat_prev + (1.0-w)*etat;
      double alpha1 = lam_hat + gamma2/(delta-lam_hat)+eta;*/

      // Check safeguards on alpha

      if( alpha1 < alpha_l || alpha1 > alpha_u )
      {
         if( DispFlag > 2 )
            cout << "   alpha1 exceeds safeguards; safeguarding (alpha1 = "
                 << alpha1 << ")" << endl;
         alpha = (alpha_l+alpha_u)/2.0;
      }
      else
      {
         if( DispFlag > 2 )
            cout << "   Update alpha in standard fashion: alpha = "
                 << alpha1 << endl;
         alpha = alpha1;
      }

      // Update alpha in the bordered matrix

      Balpha->Parameters().PutValue( "alpha",alpha );
      phi_prev = phi;
      dphi_prev = dphi;
      lam1_prev = lam1;
      xnorm_prev = xnorm;
      itn++;

      // Compute the two smallest eigenvalues of Balpha

      y->Random();
      nev = 5;
      if( (eigsolver->Solve( nev,*Balpha,*y )) )
      {
         if( DispFlag )
            cout << "Error in HCL_TrustRegion_SS_d::Solve: Eigenvalue solver "
                    "failed on iteration " << itn << endl;
         if( DispFlag > 3 )
            DumpProblem( A,g,x );
         Parameters().PutValue( "Reduction",0.0 );
         Parameters().PutValue( "StepLength",0.0 );
         Parameters().PutValue( "TermCode",FAILED );
         Clean();
         return;
      }
      eigsolver->Eigenvalue( 1,lam1 );
      eigsolver->Eigenvalue( 2,lam2 );
      oneptr = 1;
      twoptr = 2;
      if( lam2 < lam1 )
      {
         lam2 = lam1;
         eigsolver->EigPair( 2,lam1,y );
         oneptr = 2;
         twoptr = 1;
      }
      else
         eigsolver->EigPair( 1,lam1,y );

      // Update safeguards (PROTECTED DIVISION????)
      // ??? Adjust this for scaled norm ???

      nu1 = (*gammaVec)(1);
      xnorm = GetxNorm( (*y)(2) );
      test = xnorm/abs(nu1) - Delta;
      if( test < 0.0 )
      {
         lam_l = lam1;
         alpha_l = alpha;
      }
      else
      {
         lam_u = min( lam1,0.0 );
         alpha_u = alpha;
      }

      if( abs( test ) < Tol*Delta )
         done = 1;
      else if( itn > MaxItn )
         done = 1;
   }
   y->Mul( 1/nu1 );
   xnorm /= abs(nu1);
   phi = -g.Inner( x );
   dphi = xnorm*xnorm;
   A.Image( x,*xtemp );
   double red = -(0.5*(x.Inner( *xtemp )) + g.Inner( x ));
   if( DispFlag )
      cout << "HCL_TrustRegion_SS_d" << endl
           << "   Iteration: " << itn  << endl
           << "      Relative error in ||x|| = " << abs(test)/Delta << endl
           << "      alpha = " << alpha << endl
           << "      Smallest eigenvalues are " 
           << lam1  << " and " << lam2 << endl
           << "      nu1 = " << nu1 << endl
           << "      Current safeguards: " << endl
           << "         [lam_l,lam_u] = [" << lam_l << ","
           << lam_u << "]" << endl
           << "         [alpha_l,alpha_u] = [" << alpha_l << ","
           << alpha_u << "]" << endl << endl
           << "Reduction = " << red << endl
           << "StepLength = " << xnorm << endl;
   
   if( red <= 0.0 )
   {
      cout << "Error in HCL_TrustRegion_SS_d: red = " << red << endl;
      if( DispFlag > 3 )
         DumpProblem( A,g,x );
   }
   Parameters().PutValue( "Reduction",red );
   Parameters().PutValue( "StepLength",xnorm );
   Parameters().PutValue( "alpha",alpha );
   Parameters().PutValue( "Multiplier",-lam_hat );
   if( red <= 0.0 )
      Parameters().PutValue( "TermCode",FAILED );
   if( abs( test ) < Tol*Delta )
      Parameters().PutValue( "TermCode",BOUNDARY_SOLUTION );
   else if( itn > MaxItn )
      Parameters().PutValue( "TermCode",BOUNDARY_SOLUTION ); // ??? Change this
   Clean();
}

ostream & HCL_TrustRegion_SS_d::Write( ostream & str ) const
{
   str << "HCL_TrustRegion_SS_d Santosa-Sorensen trust region solver"
       << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}
