/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#include "HCL_UMinAlternation_d.h"

HCL_UMinAlternation_d::HCL_UMinAlternation_d()
{
   cerr << "Error: HCL_UMinAlternation_d default constructor called"
        << endl;
   exit(1);
}

HCL_UMinAlternation_d::HCL_UMinAlternation_d(
                           const HCL_UMinAlternation_d & )
{
   cerr << "Error: HCL_UMinAlternation_d copy constructor called"
        << endl;
   exit(1);
}

HCL_UMinAlternation_d::HCL_UMinAlternation_d( int n,
                                                      char * fname )
{
   M = n;
   eval = NULL;
   if( fname )
      ParamTable = new Table( fname,"UMinAlternation UMinProductDomain" );
   else
      ParamTable = new Table;

   umin = new HCL_UMin_d*[n];
   int i;
   for( i=0;i<n;i++ ) umin[i] = NULL;
   numinit = 0;
}

HCL_UMinAlternation_d::~HCL_UMinAlternation_d()
{
   int i;
   for( i=0;i<M;i++ ) HCL_delete( umin[i] );
   delete[] umin;
   HCL_delete( ParamTable );
   HCL_delete( eval );
}

Table & HCL_UMinAlternation_d::Parameters() const
{
   return *ParamTable;
}

HCL_EvalFunctionalProductDomain_d & HCL_UMinAlternation_d::LastEval()
                                                                   const
{
   if( !eval )
   {
      cerr << "Error in HCL_UMinAlternation_d::LastEval: not set" << endl;
      exit(1);
   }
   return *eval;
}

void HCL_UMinAlternation_d::SetScaling( HCL_LinearOp_d * SS,
                                            HCL_LinearSolver_d * llsolver )
{
}

void HCL_UMinAlternation_d::UnSetScaling()
{
}

void HCL_UMinAlternation_d::Initialize( HCL_VectorSpace_d & dom )
{
   if( ParamTable->GetValue( "MaxItn",MaxItn ) )
   {
      MaxItn = 100;
      ParamTable->PutValue( "MaxItn",MaxItn );
   }
   if( ParamTable->GetValue( "Typf",Typf ) )
   {
      Typf = 1.0;
      ParamTable->PutValue( "Typf",Typf );
   }
   if( ParamTable->GetValue( "TypxNorm",TypxNorm ) )
   {
      TypxNorm = 1.0;
      ParamTable->PutValue( "TypxNorm",TypxNorm );
   }
   if( ParamTable->GetValue( "GradTol",GradTol ) )
   {
      GradTol = 1.0e-2;
      ParamTable->PutValue( "GradTol",GradTol );
   }
   int MinFlag = ParamTable->GetValue( "MinStep",MinStep );
   int MaxFlag = ParamTable->GetValue( "MaxStep",MaxStep );
   if( MinFlag && MaxFlag || (!MinFlag && !MaxFlag && MinStep >= MaxStep) )
   {
      MaxStep = 1.0e20;
      MinStep = 1.0e-20;
      ParamTable->PutValue( "MaxStep",MaxStep );
      ParamTable->PutValue( "MinStep",MinStep );
   }
   else if( MinFlag && !MaxFlag )
   {
      MinStep = 1.0e-40*MaxStep;
      ParamTable->PutValue( "MinStep",MinStep );
   }
   else if (MaxFlag && !MinFlag )
   {
      MaxStep = 1.0e40*MinStep;
      ParamTable->PutValue( "MaxStep",MaxStep );
   }
   if( ParamTable->GetValue( "CscMaxLimit",CscMaxLimit ) )
   {
      CscMaxLimit = 5;
      ParamTable->PutValue( "CscMaxLimit",CscMaxLimit );
   }
   if( ParamTable->GetValue( "DispFlag",DispFlag ) )
   {
      DispFlag = 0;
      ParamTable->PutValue( "DispFlag",DispFlag );
   }
   if( ParamTable->GetValue( "DumpFlag",DumpFlag ) )
   {
      DumpFlag = 0;
      ParamTable->PutValue( "DumpFlag",DumpFlag );
   }
   if( ParamTable->GetValue( "DispPrecision",DispPrecision ) )
   {
      DispPrecision = 6;
      ParamTable->PutValue( "DispPrecision",DispPrecision );
   }
   OldDispPrec = cout.precision( DispPrecision );
   if( ParamTable->GetValue( "DumpPrecision",DumpPrecision ) )
   {
      DumpPrecision = DispPrecision;
      ParamTable->PutValue( "DumpPrecision",DumpPrecision );
   }
   if( DumpFlag )
   {
      if( ParamTable->GetValue( "DumpFile",DumpFile ) )
      {
         strcpy( DumpFile,"HCL_UMinAlternation.DumpFile" );
         ParamTable->PutValue( "DumpFile",DumpFile );
      }
      dump = new ofstream( DumpFile,ios::out );
      dump->precision( DumpPrecision );
   }
   else
      dump = NULL;
   if( ParamTable->GetValue( "TraceSteps",TraceSteps ) )
   {
      TraceSteps = 0;
      ParamTable->PutValue( "TraceSteps",TraceSteps );
   }
   if( TraceSteps )
   {
      if( ParamTable->GetValue( "StepFile",StepFile ) )
      {
         strcpy( StepFile,"HCL_UMinAlternation.StepFile" );
         ParamTable->PutValue( "StepFile",StepFile );
      }
      Steps = new ofstream( StepFile,ios::out );
   }
   else
      Steps = NULL;

   S = NULL;
   lsolver = NULL;
   tmpvec = NULL;

   xnext = (HCL_ProductVector_d*)dom.Member();
   step = (HCL_ProductVector_d*)dom.Member();
   Itn = 0;
   CscMax = 0;
}

void HCL_UMinAlternation_d::SetNext( HCL_UMin_d * um )
{
   if( numinit == M )
   {
      cerr << "Error in HCL_UMinAlternation_d::SetNext: already "
              "initialized" << endl;
      exit(1);
   }
   umin[numinit] = um;um->IncCount();
   numinit++;
}

int HCL_UMinAlternation_d::StoppingTest( HCL_ProductVector_d & x,
                                             HCL_ProductVector_d * x1 )
{
   fx = eval->ValueRef();

   // Compute the norm of the gradient

   if( lsolver != NULL )
   {
      lsolver->Solve( *S,eval->GradientRef(),*tmpvec );
      gnorm = sqrt( eval->GradientRef().Inner( *tmpvec ) );
   }
   else if( S != NULL )
   {
      S->InvImage( eval->GradientRef(),*tmpvec );
      gnorm = sqrt( eval->GradientRef().Inner( *tmpvec ) );
   }
   else
      gnorm = eval->GradientRef().Norm();

   // Check for an approximate stationary point
   
   if( S!= NULL )
   {
      S->Image( x,*tmpvec );
      xnorm = sqrt( x.Inner( *tmpvec ) );
   }
   else
      xnorm = x.Norm();

   double RelGradNorm = max( xnorm,TypxNorm )*gnorm/
                        max( abs(eval->ValueRef()),Typf );

   if( Itn == 0 )
      if( RelGradNorm < GradTol/1000.0 )
         return PossibleMinimizer;
      else if( Itn >= MaxItn )
         return IterationLimit;
      else
         return 0;

   if( RelGradNorm < GradTol )
      return PossibleMinimizer;

   // Compute the norm of the step

   if( x1 )
   {
      step->Sub( *x1,x );
      if( S != NULL )
      {
         S->Image( *step,*tmpvec );
         stepnorm = sqrt(step->Inner( *tmpvec ));
      }
      else
         stepnorm = step->Norm();

      // Check for convergence

      if( stepnorm/max( TypxNorm,xnorm) < MinStep )
         return PossibleConvergence;
   }

   // Check for iteration limit

   if( Itn >= MaxItn )
      return IterationLimit;

   // Check for divergence

   if( stepnorm >= MaxStep )
   {
      CscMax++;
      if( CscMax == CscMaxLimit )
         return PossibleDivergence;
   }
   else
      CscMax = 0;

   // No stopping criterion satisfied

   return 0;
}

void HCL_UMinAlternation_d::Display( const HCL_Functional_d & f,
                                     const HCL_Vector_d & x )
{
   if( DispFlag > 1 || (DispFlag && TermCode) )
   {
      cout << endl << "===========HCL_UMinAlternation_d============="
           << endl;
      cout << "iteration: " << Itn << endl;
      cout << " f = " << fx << endl;
      cout << "||g|| = " << gnorm << endl;
      if( Itn ) cout << "||step|| = " << stepnorm << endl;
      cout << endl << "================================================="
           << endl;
   }
   if( (DumpFlag > 1 || (DumpFlag && TermCode)) && dump != NULL )
   {
      if( Itn == 0 )
      {
         *dump << "Summary of HCL_UMinAlternation_d run" << endl;
         *dump << "----------------------------------------" << endl << endl;
         *dump << "f = " << endl;
         f.Write( *dump );
         *dump << endl << "ParamTable = " << endl;
         Parameters().Write( *dump );
         *dump << endl << endl;
      }
      (*dump) << endl << "======================" << endl;
      (*dump) << "iteration: " << Itn << endl;
      (*dump) << " f = " << fx << endl;
      (*dump) << "||g|| = " << gnorm << endl;
      if( Itn ) (*dump) << "||step|| = " << stepnorm << endl;
      (*dump) << endl << "======================" << endl;
   }
   if( TermCode )
   {
      if( DispFlag )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               cout << "HCL_UMinAlternation_d terminated: Possible "
                       "minimizer" << endl;
               break;
            case PossibleConvergence:
               cout << "HCL_UMinAlternation_d terminated: Possible "
                       "convergence" << endl;
               break;
            case LineSearchFailed:
               cout << "HCL_UMinAlternation_d terminated: Line search "
                       "failed" << endl;
               break;
            case PossibleDivergence:
               cout << "HCL_UMinAlternation_d terminated: Possible "
                       "divergence" << endl;
               break;
            case IterationLimit:
               cout << "HCL_UMinAlternation_d terminated: Iteration limit "
                       "reached" << endl;
               break;
            default:
               cout << "HCL_UMinAlternation_d terminated: Unknown "
                       "termination code" << endl;
               break;
         }
      }
      if( DumpFlag && dump != NULL )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               (*dump) << "Terminated: Possible minimizer" << endl;
               break;
            case PossibleConvergence:
               (*dump) << "Terminated: Possible convergence" << endl;
               break;
            case LineSearchFailed:
               (*dump) << "Terminated: Line search failed" << endl;
               break;
            case PossibleDivergence:
               (*dump) << "Terminated: Possible divergence" << endl;
               break;
            case IterationLimit:
               (*dump) << "Terminated: Iteration limit reached" << endl;
               break;
            case InaccurateGradient:
               (*dump) << "Terminated: Line search failed due to "
                          "inaccurate gradient" << endl;
               break;
            default:
               (*dump) << "Terminated: Unknown termination code" << endl;
               break;
         }
      }
   }
   if( TraceSteps && Steps != NULL )
      x.Write( *Steps );
}

void HCL_UMinAlternation_d::Clean()
{
   cout.precision( OldDispPrec );
   HCL_delete( step );
   HCL_delete( xnext );
   HCL_delete( tmpvec );
   if( TraceSteps )
      delete Steps;
   if( DumpFlag )
      delete dump;
}

int HCL_UMinAlternation_d::Minimize( HCL_FunctionalProductDomain_d & f,
                                         HCL_ProductVector_d & x )
{
   // M is the number of factors in the domain

   if( numinit < M )
   {
      cerr << "Error in HCL_UMinAlternation_d::Minimize: not initialized"
           << endl;
      exit(1);
   }
   if( x.Number() != M )
   {
      cerr << "Error in HCL_UMinAlternation_d::Minimize: vector has the "
              "wrong number of factors" << endl;
      exit(1);
   }
   if( x.Space() != f.Domain() )
   {
      cerr << "Error in HCL_UMinAlternation_d::Minimize: vector is "
              "not in domain" << endl;
      exit(1);
   }

   // initialize parameters

   Initialize( f.Domain() );

   // Evaluate f(x0) and grad f(x0)

   if( eval ) HCL_delete( eval );
   eval = (HCL_EvalFunctionalProductDomain_d*)f.Evaluate( x );

   // Check to see if initial point is stationary

   TermCode = StoppingTest( x );
   Display( f,x );
   xnext->Copy( x );      

   while( !TermCode )
   {

      // Perform one alternation sweep

      int i;
      for( i=1;i<=M;i++ )
      {
         HCL_RestrictedFunctional_d g( &f,*xnext,i );
         HCL_Vector_d & y = (*xnext)(i);

         umin[i-1]->Minimize( g,y );
      }
      HCL_delete( eval );
      eval = &(((HCL_EvalRestrictedFunctional_d&)
                 (umin[M-1])->LastEval()).UnrestrictedEval());
      eval->IncCount();

      Itn++;
      TermCode = StoppingTest( x,xnext );
      Display( f,x );
      x.Copy( *xnext );
   }
   Clean();
   return TermCode;
}

ostream & HCL_UMinAlternation_d::Write( ostream & str ) const
{
   str << "HCL_UMinAlternation_d alteration minimization algorithm"
       << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}
