#include "HCL_UMinSubspace_d.h"

HCL_UMinSubspace_d::HCL_UMinSubspace_d(
                           const HCL_UMinSubspace_d & )
{
   cerr << "Error: HCL_UMinSubspace_d copy constructor called"
        << endl;
   exit(1);
}

HCL_UMinSubspace_d::HCL_UMinSubspace_d( char * fname )
{
   eval = NULL;
   if( fname )
      ParamTable = new Table( fname,"UMinSubspace UMinProductDomain" );
   else
      ParamTable = new Table;
   trsolver = new HCL_MatTR_d( fname );
}

HCL_UMinSubspace_d::~HCL_UMinSubspace_d()
{
   HCL_delete( trsolver );
   HCL_delete( ParamTable );
   HCL_delete( eval );
}

Table & HCL_UMinSubspace_d::Parameters() const
{
   return *ParamTable;
}

HCL_EvalFunctionalProductDomain_d & HCL_UMinSubspace_d::LastEval()
                                                                   const
{
   if( !eval )
   {
      cerr << "Error in HCL_UMinSubspace_d::LastEval: not set" << endl;
      exit(1);
   }
   return *eval;
}

void HCL_UMinSubspace_d::SetScaling( HCL_LinearOp_d * SS,
                                     HCL_LinearSolver_d * llsolver )
{
}


void HCL_UMinSubspace_d::UnSetScaling()
{
}

void HCL_UMinSubspace_d::Initialize( HCL_VectorSpace_d & dom )
{
   if( ParamTable->GetValue( "MaxItn",MaxItn ) )
   {
      MaxItn = 100;
      ParamTable->PutValue( "MaxItn",MaxItn );
   }
   if( ParamTable->GetValue( "Typf",Typf ) )
   {
      Typf = 1.0;
      ParamTable->PutValue( "Typf",Typf );
   }
   if( ParamTable->GetValue( "TypxNorm",TypxNorm ) )
   {
      TypxNorm = 1.0;
      ParamTable->PutValue( "TypxNorm",TypxNorm );
   }
   if( ParamTable->GetValue( "GradTol",GradTol ) )
   {
      GradTol = 1.0e-2;
      ParamTable->PutValue( "GradTol",GradTol );
   }
   int MinFlag = ParamTable->GetValue( "MinStep",MinStep );
   int MaxFlag = ParamTable->GetValue( "MaxStep",MaxStep );
   if( MinFlag && MaxFlag || (!MinFlag && !MaxFlag && MinStep >= MaxStep) )
   {
      MaxStep = 1.0e20;
      MinStep = 1.0e-20;
      ParamTable->PutValue( "MaxStep",MaxStep );
      ParamTable->PutValue( "MinStep",MinStep );
   }
   else if( MinFlag && !MaxFlag )
   {
      MinStep = 1.0e-40*MaxStep;
      ParamTable->PutValue( "MinStep",MinStep );
   }
   else if (MaxFlag && !MinFlag )
   {
      MaxStep = 1.0e40*MinStep;
      ParamTable->PutValue( "MaxStep",MaxStep );
   }
   if( ParamTable->GetValue( "CscMaxLimit",CscMaxLimit ) )
   {
      CscMaxLimit = 5;
      ParamTable->PutValue( "CscMaxLimit",CscMaxLimit );
   }
   if( ParamTable->GetValue( "KrylovSubspaceLevel",KrylovSubspaceLevel ) )
   {
      KrylovSubspaceLevel = 2;
      ParamTable->PutValue( "KrylovSubspaceLevel",KrylovSubspaceLevel );
   }
   if( ParamTable->GetValue( "Delta",Delta ) )
   {
      Delta = 1.0;
      ParamTable->PutValue( "Delta",Delta );
   }
   if( Parameters().GetValue( "LowerTol",LowerTol ) )
   {
      LowerTol = 0.1;
      ParamTable->PutValue( "LowerTol",LowerTol );
   }
   if( Parameters().GetValue( "UpperTol",UpperTol ) )
   {
      UpperTol = 0.9;
      ParamTable->PutValue( "UpperTol",UpperTol );
   }
   if( Parameters().GetValue( "Reorthogonalize",Reorthogonalize ) )
   {
      Reorthogonalize = 0;
      ParamTable->PutValue( "Reorthogonalize",Reorthogonalize );
   }
   if( ParamTable->GetValue( "DispFlag",DispFlag ) )
   {
      DispFlag = 0;
      ParamTable->PutValue( "DispFlag",DispFlag );
   }
   if( ParamTable->GetValue( "DumpFlag",DumpFlag ) )
   {
      DumpFlag = 0;
      ParamTable->PutValue( "DumpFlag",DumpFlag );
   }
   if( ParamTable->GetValue( "DispPrecision",DispPrecision ) )
   {
      DispPrecision = 6;
      ParamTable->PutValue( "DispPrecision",DispPrecision );
   }
   OldDispPrec = cout.precision( DispPrecision );
   if( ParamTable->GetValue( "DumpPrecision",DumpPrecision ) )
   {
      DumpPrecision = DispPrecision;
      ParamTable->PutValue( "DumpPrecision",DumpPrecision );
   }
   if( DumpFlag )
   {
      if( ParamTable->GetValue( "DumpFile",DumpFile ) )
      {
         strcpy( DumpFile,"HCL_UMinSubspace.DumpFile" );
         ParamTable->PutValue( "DumpFile",DumpFile );
      }
      dump = new ofstream( DumpFile,ios::out );
      dump->precision( DumpPrecision );
   }
   else
      dump = NULL;
   if( ParamTable->GetValue( "TraceSteps",TraceSteps ) )
   {
      TraceSteps = 0;
      ParamTable->PutValue( "TraceSteps",TraceSteps );
   }
   if( TraceSteps )
   {
      if( ParamTable->GetValue( "StepFile",StepFile ) )
      {
         strcpy( StepFile,"HCL_UMinSubspace.StepFile" );
         ParamTable->PutValue( "StepFile",StepFile );
      }
      Steps = new ofstream( StepFile,ios::out );
   }
   else
      Steps = NULL;

   S = NULL;
   lsolver = NULL;
   tmpvec = (HCL_ProductVector_d*)dom.Member();

   xnext = (HCL_ProductVector_d*)dom.Member();
   step = (HCL_ProductVector_d*)dom.Member();
   Itn = 0;
   CscMax = 0;
}

int HCL_UMinSubspace_d::StoppingTest( HCL_ProductVector_d & x,
                                             HCL_ProductVector_d * x1 )
{
   fx = eval->ValueRef();

   // Compute the norm of the gradient

   if( lsolver != NULL )
   {
      lsolver->Solve( *S,eval->GradientRef(),*tmpvec );
      gnorm = sqrt( eval->GradientRef().Inner( *tmpvec ) );
   }
   else if( S != NULL )
   {
      S->InvImage( eval->GradientRef(),*tmpvec );
      gnorm = sqrt( eval->GradientRef().Inner( *tmpvec ) );
   }
   else
      gnorm = eval->GradientRef().Norm();

   // Check for an approximate stationary point
   
   if( S!= NULL )
   {
      S->Image( x,*tmpvec );
      xnorm = sqrt( x.Inner( *tmpvec ) );
   }
   else
      xnorm = x.Norm();

   double RelGradNorm = max( xnorm,TypxNorm )*gnorm/
                        max( abs(eval->ValueRef()),Typf );

   if( Itn == 0 )
      if( RelGradNorm < GradTol/1000.0 )
         return PossibleMinimizer;
      else if( Itn >= MaxItn )
         return IterationLimit;
      else
         return 0;

   if( RelGradNorm < GradTol )
      return PossibleMinimizer;

   // Compute the norm of the step

   if( x1 )
   {
      step->Sub( *x1,x );
      if( S != NULL )
      {
         S->Image( *step,*tmpvec );
         stepnorm = sqrt(step->Inner( *tmpvec ));
      }
      else
         stepnorm = step->Norm();

      // Check for convergence

      if( stepnorm/max( TypxNorm,xnorm) < MinStep )
         return PossibleConvergence;
   }

   // Check for iteration limit

   if( Itn >= MaxItn )
      return IterationLimit;

   // Check for divergence

   if( stepnorm >= MaxStep )
   {
      CscMax++;
      if( CscMax == CscMaxLimit )
         return PossibleDivergence;
   }
   else
      CscMax = 0;

   // No stopping criterion satisfied

   return 0;
}

void HCL_UMinSubspace_d::Display0( const HCL_Functional_d & f,
                                       const HCL_Vector_d & x )
{
   if( DispFlag > 1 || (DispFlag && TermCode) )
   {
      cout << endl << "=============HCL_UMinSubspace_d============="
           << endl;
      cout << "iteration: " << Itn << endl;
      cout << " f = " << fx << endl;
      cout << "||g|| = " << gnorm << endl;
      cout << "Delta: " << Delta << endl;
      cout << endl << "======================================" << endl;
   }
   if( (DumpFlag > 1 || (DumpFlag && TermCode)) && dump != NULL )
   {
      *dump << "Summary of HCL_UMinSubspace_d run" << endl;
      *dump << "-------------------------" << endl << endl;
      *dump << "f = " << endl;
      f.Write( *dump );
      *dump << endl << "ParamTable = " << endl;
      ((HCL_UMinSubspace_d*)this)->Parameters().Write( *dump );
      *dump << endl << endl;
      *dump << endl << "=============HCL_UMinSubspace_d============="
            << endl;
      (*dump) << "iteration: " << Itn << endl;
      (*dump) << " f = " << fx << endl;
      (*dump) << "||g|| = " << gnorm << endl;
      *dump << "Delta: " << Delta << endl;
      *dump << endl << "======================================" << endl;
   }
   if( TermCode )
   {
      if( DispFlag )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               cout << "HCL_UMinSubspace_d: Terminated: Possible "
                       "minimizer" << endl;
               break;
            case PossibleConvergence:
               cout << "HCL_UMinSubspace_d: Terminated: Possible "
                       "convergence" << endl;
               break;
            case IterationLimit:
               cout << "HCL_UMinSubspace_d: Terminated: Iteration "
                       "limit reached" << endl;
               break;
            default:
               cout << "HCL_UMinSubspace_d: Terminated: Unknown "
                       "termination code" << endl;
               break;
         }
      }
      if( DumpFlag && dump != NULL )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               *dump << "HCL_UMinSubspace_d: Terminated: Possible "
                        "minimizer" << endl;
               break;
            case PossibleConvergence:
               *dump << "HCL_UMinSubspace_d: Terminated: Possible "
                        "convergence"
                     << endl;
               break;
            case IterationLimit:
               *dump << "HCL_UMinSubspace_d: Terminated: Iteration "
                        "limit reached" << endl;
               break;
            default:
               *dump << "HCL_UMinSubspace_d: Terminated: Unknown "
                        "termination code" << endl;
               break;
         }
      }
   }
}

void HCL_UMinSubspace_d::Display( const HCL_Functional_d & f,
                                      const HCL_Vector_d & x )
{
   if( DispFlag > 1 || (DispFlag && TermCode) )
   {
      cout << endl << "===========HCL_UMinSubspace_d============="
           << endl;
      cout << "iteration: " << Itn << endl;
      cout << " f = " << fx << endl;
      cout << "||g|| = " << gnorm << endl;
      if( Itn ) cout << "||step|| = " << stepnorm << endl;
      cout << "pred: " << pred << " ared: " << ared << " ratio: " << ratio
           << endl;
      cout << "Delta: " << Delta << endl;
      cout << endl << "=============================================="
           << endl;
   }
   if( (DumpFlag > 1 || (DumpFlag && TermCode)) && dump != NULL )
   {
      (*dump) << endl << "======================" << endl;
      (*dump) << "iteration: " << Itn << endl;
      (*dump) << " f = " << fx << endl;
      (*dump) << "||g|| = " << gnorm << endl;
      if( Itn ) (*dump) << "||step|| = " << stepnorm << endl;
      (*dump) << "pred: " << pred << " ared: " << ared << " ratio: " << ratio
              << endl;
      (*dump) << "Delta: " << Delta << endl;
      (*dump) << endl << "======================" << endl;
   }
   if( TermCode )
   {
      if( DispFlag )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               cout << "HCL_UMinSubspace_d terminated: Possible "
                       "minimizer" << endl;
               break;
            case PossibleConvergence:
               cout << "HCL_UMinSubspace_d terminated: Possible "
                       "convergence" << endl;
               break;
            case PossibleDivergence:
               cout << "HCL_UMinSubspace_d terminated: Possible "
                       "divergence" << endl;
               break;
            case IterationLimit:
               cout << "HCL_UMinSubspace_d terminated: Iteration limit "
                       "reached" << endl;
               break;
            default:
               cout << "HCL_UMinSubspace_d terminated: Unknown "
                       "termination code" << endl;
               break;
         }
      }
      if( DumpFlag && dump != NULL )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               (*dump) << "Terminated: Possible minimizer" << endl;
               break;
            case PossibleConvergence:
               (*dump) << "Terminated: Possible convergence" << endl;
               break;
            case LineSearchFailed:
               (*dump) << "Terminated: Line search failed" << endl;
               break;
            case PossibleDivergence:
               (*dump) << "Terminated: Possible divergence" << endl;
               break;
            case IterationLimit:
               (*dump) << "Terminated: Iteration limit reached" << endl;
               break;
            case InaccurateGradient:
               (*dump) << "Terminated: Line search failed due to "
                          "inaccurate gradient" << endl;
               break;
            default:
               (*dump) << "Terminated: Unknown termination code" << endl;
               break;
         }
      }
   }
   if( TraceSteps && Steps != NULL )
      x.Write( *Steps );
}

void HCL_UMinSubspace_d::Clean()
{
   cout.precision( OldDispPrec );
   HCL_delete( step );
   HCL_delete( xnext );
   HCL_delete( tmpvec );
   if( TraceSteps )
      delete Steps;
   if( DumpFlag )
      delete dump;
}

int HCL_UMinSubspace_d::Minimize( HCL_FunctionalProductDomain_d & f,
                                         HCL_ProductVector_d & x )
{
   // M is the number of factors in the domain

   if( x.Space() != f.Domain() )
   {
      cerr << "Error in HCL_UMinSubspace_d::Minimize: vector is "
              "not in domain" << endl;
      exit(1);
   }

   // initialize parameters

   Initialize( f.Domain() );

   // Evaluate f(x0) and grad f(x0)

   if( eval ) HCL_delete( eval );
   eval = (HCL_EvalFunctionalProductDomain_d*)f.Evaluate( x );

   // Check to see if initial point is stationary

   TermCode = StoppingTest( x );
   Display0( f,x );

   int newflag = 1;
   HCL_SymMatOp_d * B=NULL;
   HCL_RnVector_d * g=NULL;
   HCL_RnVector_d * p=NULL;
   HCL_BasisProductSpace_d * basis=NULL;
   int n;
   while( !TermCode )
   {

      if( newflag )
      {
         HCL_delete( p );
         HCL_delete( g );
         HCL_delete( B );
         HCL_delete( basis );

         // Compute the vectors in the subspace

         basis = Basis();
         n = (*basis).Dim();
         if( n <= 0 )
         {
            cerr << "Error in HCL_UMinSubspace_d::Minimize: the basis "
                    "has dimension zero" << endl;
            exit(1);
         }
         if( DispFlag > 1 )
            cout << "Dimension of subspace to search: " << n << endl;

         p = new HCL_RnVector_d( n );
         // Compute the projected operator B and the projected gradient g

         ProjectedProblem( *basis,B,g );
      }

      // Solve the trust region subproblem on the subspace

      trsolver->Parameters().PutValue( "Delta",Delta );
      if( Itn )
         trsolver->Parameters().PutValue( "UseLastLam",1 );
      else
         trsolver->Parameters().PutValue( "UseLastLam",0 );
      trsolver->Solve( *B,*g,*p );

      if( trsolver->Parameters().GetValue( "Reduction",pred ) )
      {
         cerr << "Error in HCL_UMinSubspace_d::Minimize: HCL_MatTR_d "
                 "did not set 'Reduction'" << endl;
         HCL_delete( basis );
         HCL_delete( eval );
         Clean();
         exit(1);
      }
      if( trsolver->Parameters().GetValue( "TermCode",trcode ) )
      {
         cerr << "Error in HCL_UMinSubspace_d::Minimize: trust region "
                 "solver did not set TermCode" << endl;
         HCL_delete( basis );
         HCL_delete( eval );
         Clean();
         exit(1);
      }

      if( pred > 0.0 )
      {
         (*basis).LinearCombination( *p,*xnext );

         if( !Reorthogonalize )
            StepLength = xnext->Norm();
         else if(trsolver->Parameters().GetValue("StepLength",StepLength))
            StepLength = xnext->Norm();

         xnext->Add( x );
         HCL_EvalFunctionalProductDomain_d * evalnext = 
             (HCL_EvalFunctionalProductDomain_d*)f.Evaluate( *xnext );
         ared = eval->ValueRef()-evalnext->ValueRef();
         ratio = ared/pred;
         int cflag = 0;
         if( ared > 0.0 )
         {
            HCL_delete( eval );
            eval = evalnext;
            cflag = 1;
            Itn++;
            newflag = 1;
         }
         else
         {
            HCL_delete( evalnext );
            newflag = 0;
         }

         if( ratio >= UpperTol && ared > 0.0 )
            Delta = max( Delta,2.0*StepLength );
         else if( ratio < LowerTol || ared <= 0.0 || trcode == FAILED )
            Delta = min( Delta,StepLength )/2.0;

         if( Delta > MaxStep ) Delta = MaxStep;

         TermCode = StoppingTest( x,xnext );
         if( cflag ) x.Copy( *xnext );
         Display( f,x );
      }
      else
         Delta *= 0.5;
   }
   HCL_delete( p );
   HCL_delete( g );
   HCL_delete( B );
   HCL_delete( basis );
   Clean();
   return TermCode;
}

HCL_BasisProductSpace_d * HCL_UMinSubspace_d::Basis()
{
   int M = eval->ProductDomain().Number();
   int n = 0,t=1,i,j;
   for( i=0;i<KrylovSubspaceLevel;i++ )
   {
      t *= M;
      n += t;
   }

   HCL_FPArch_d arch;
   HCL_BasisProductSpace_d * basis = new HCL_BasisProductSpace_d(
                             &(eval->ProductDomain()),n );
   HCL_BasisProductSpace_d & B = *basis;
   int k,l,q;
   if( !S )
   {
      int * flag = new int[n+1];
      for( j=1;j<=M;j++ )
      {
         B(j).Zero();
         B(j)(j).Copy( eval->PartialGradientRef( j ) );
         flag[j] = j;
      }
      k = M;
      l = 0;
      for( i=2;i<=KrylovSubspaceLevel;i++ )
      {
         int tl = k;
         for( j=1;j<=pow(M,i-1);j++ )
         {
            l++;
            for( q=1;q<=M;q++ )
            {
               k++;
               B(k).Zero();
               eval->HessianBlockRef(q,flag[l]).Image( B(l)(flag[l]),B(k)(q) );
               if( Reorthogonalize )
               {
                  double vn = B(k)(q).Norm();
                  if( vn > sqrt(arch.small) )
                     B(k)(q).Mul( 1.0/vn );
               }
               flag[k] = q;
            }
         }
         l = tl;
      }
      delete[] flag;
   }
   else
   {
      cerr << "Error in HCL_UMinSubspace_d::Basis: scaling not "
              "implemented" << endl;
      exit(1);
   }
   B.ModifiedGramSchmidtWithReduction();
   if( Reorthogonalize )
      B.ModifiedGramSchmidtWithReduction();
   return basis;
}

void HCL_UMinSubspace_d::ProjectedProblem( HCL_BasisProductSpace_d & Bas,
                             HCL_SymMatOp_d* & B,HCL_RnVector_d* & g ) const
{
   int n = Bas.Dim();
   GenSymMatrix_d * C = new GenSymMatrix_d( n );
   g = new HCL_RnVector_d( n );
   int i,j;
   for( j=1;j<=n;j++ )
   {
      (*g)(j) = eval->GradientRef().Inner( Bas(j) );
      eval->HessianRef().Image( Bas(j),*tmpvec );
      for( i=1;i<=j;i++ )
         C->set(i,j,Bas(i).Inner( *tmpvec ) );
   }
   B = new HCL_SymMatOp_d( C,1,&(g->Space()),&(g->Space()) );
}

ostream & HCL_UMinSubspace_d::Write( ostream & str ) const
{
   str << "HCL_UMinSubspace_d: unconstrained minimization based on choosing"
       << "search directions from a subspace" << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}

