/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "maxmin_d.h"
#include "HCL_UMinTR_d.h"

HCL_UMinTR_d::HCL_UMinTR_d()
{
   cerr << "Error: HCL_UMinTR_d default constructor called" << endl;
   exit(1);
}

HCL_UMinTR_d::HCL_UMinTR_d( const HCL_UMinTR_d & xx )
{
   cerr << "Error: HCL_UMinTR_d copy constructor called" << endl;
   exit(1);
}

HCL_UMinTR_d::HCL_UMinTR_d( HCL_TrustRegionSolver_d * trs,char * fname )
{
   if( fname )
      ParamTable = new Table( fname,"UMinTR UMin" );
   else
      ParamTable = new Table;
   if( trs )
   {
      trsolver = trs;
      trs->IncCount();
   }
   else
      trsolver = new HCL_TRCG_d( fname );
   C = NULL;
   lsolver = NULL;
   dump = NULL;
   fevalFlag = 0;
   feval = NULL;
   tmpvec = NULL;
}

HCL_UMinTR_d::~HCL_UMinTR_d()
{
   if( fevalFlag ) HCL_delete( feval );
   HCL_delete( ParamTable );
   HCL_delete( C );
   HCL_delete( lsolver );
   HCL_delete( trsolver );
}

void HCL_UMinTR_d::SetScaling( HCL_LinearOp_d * S,
                               HCL_LinearSolver_d * ls )
{
   HCL_delete( C );
   HCL_delete( lsolver );
   C = S;S->IncCount();
   lsolver = ls;if( ls ) ls->IncCount();
}
   
void HCL_UMinTR_d::UnSetScaling()
{
   HCL_delete( C );
   HCL_delete( lsolver );
   C = NULL;
   lsolver = NULL;
}

void HCL_UMinTR_d::Initialize( HCL_VectorSpace_d & dom )
{
   if( fevalFlag )
      HCL_delete( feval );
   fevalFlag = 0;
   if( Parameters().GetValue( "MaxItn",MaxItn ) )
   {
      MaxItn = 100;
      Parameters().PutValue( "MaxItn",MaxItn );
   }
   if( ParamTable->GetValue( "Typf",Typf ) )
   {
      Typf = 1.0;
      Parameters().PutValue( "Typf",Typf );
   }
   if( ParamTable->GetValue( "TypxNorm",TypxNorm ) )
   {
      TypxNorm = 1.0;
      Parameters().PutValue( "TypxNorm",TypxNorm );
   }
   if( Parameters().GetValue( "LowerTol",LowerTol ) )
   {
      LowerTol = 0.1;
      Parameters().PutValue( "LowerTol",LowerTol );
   }
   if( Parameters().GetValue( "UpperTol",UpperTol ) )
   {
      UpperTol = 0.9;
      Parameters().PutValue( "UpperTol",UpperTol );
   }
   if( Parameters().GetValue( "Delta",Delta ) )
   {
      Delta = 1.0;
      Parameters().PutValue( "Delta",Delta );
   }
   if( Parameters().GetValue( "GradTol",GradTol ) )
   {
      GradTol = 1.0e-2;
      Parameters().PutValue( "GradTol",GradTol );
   }
   int MinFlag = ParamTable->GetValue( "MinStep",MinStep );
   int MaxFlag = ParamTable->GetValue( "MaxStep",MaxStep );
   if( MinFlag && MaxFlag || (!MinFlag && !MaxFlag && MinStep >= MaxStep) )
   {
      MaxStep = 1.0e20;
      MinStep = 1.0e-20;
      Parameters().PutValue( "MinStep",MinStep );
      Parameters().PutValue( "MaxStep",MaxStep );
   }
   else if( MinFlag && !MaxFlag )
   {
      MinStep = 1.0e-40*MaxStep;
      Parameters().PutValue( "MinStep",MinStep );
   }
   else if (MaxFlag && !MinFlag )
   {
      MaxStep = 1.0e40*MinStep;
      Parameters().PutValue( "MaxStep",MaxStep );
   }
   trsolver->Parameters().PutValue( "MaxStep",MaxStep );
   trsolver->Parameters().PutValue( "MinStep",MinStep );
   if( ParamTable->GetValue( "CscMaxLimit",CscMaxLimit ) )
   {
      CscMaxLimit = 5;
      Parameters().PutValue( "CscMaxLimit",CscMaxLimit );
   }
   if( ParamTable->GetValue( "DispFlag",DispFlag ) )
   {
      DispFlag = 0;
      Parameters().PutValue( "DispFlag",DispFlag );
   }
   if( ParamTable->GetValue( "DumpFlag",DumpFlag ) )
   {
      DumpFlag = 0;
      Parameters().PutValue( "DumpFlag",DumpFlag );
   }
   if( ParamTable->GetValue( "DispPrecision",DispPrecision ) )
   {
      DispPrecision = 6;
      Parameters().PutValue( "DispPrecision",DispPrecision );
   }
   OldDispPrec = cout.precision( DispPrecision );
   if( ParamTable->GetValue( "DumpPrecision",DumpPrecision ) )
   {
      DumpPrecision = DispPrecision;
      Parameters().PutValue( "DispPrecision",DispPrecision );
   }
   if( DumpFlag )
   {
      if( ParamTable->GetValue( "DumpFile",DumpFile ) )
      {
         strcpy( DumpFile,"HCL_UMinTR.DumpFile" );
         Parameters().PutValue( "DumpFile",DumpFile );
      }
      dump = new ofstream( DumpFile,ios::out );
      dump->precision( DumpPrecision );
   }
   else
      dump = NULL;
   if( ParamTable->GetValue( "TraceSteps",TraceSteps ) )
   {
      TraceSteps = 0;
      Parameters().PutValue( "TraceSteps",TraceSteps );
   }
   if( TraceSteps )
   {
      if( ParamTable->GetValue( "StepFile",StepFile ) )
      {
         strcpy( StepFile,"HCL_UMinTR.StepFile" );
         Parameters().PutValue( "StepFile",StepFile );
      }
      Steps = new ofstream( StepFile,ios::out );
   }
   else
      Steps = NULL;

   step = dom.Member();
   xnext = dom.Member();

   if( C != NULL )
   {
      if( C->Domain() != dom )
      {
         cerr << "Error in HCL_UMinTR_d::Initialize: domain does not match"
                 " preconditioner" << endl;
         exit(1);
      }
      trsolver->SetScaling( C,lsolver );
      tmpvec = dom.Member();
   }
   else
      tmpvec = NULL;

   Itn = 0;
   CscMax = 0;
}

void HCL_UMinTR_d::Clean()
{
   cout.precision( OldDispPrec );
   HCL_delete( xnext );
   HCL_delete( step );
   HCL_delete( tmpvec );
   if( TraceSteps )
      delete Steps;
   if( DumpFlag )
      delete dump;
}

HCL_EvaluateFunctional_d & HCL_UMinTR_d::LastEval() const
{
   if( !fevalFlag )
   {
      cerr << "Error in HCL_UMinTR_d::LastEval: attempt to access last "
              "evaluation object before it was set." << endl;
      exit(1);
   }
   return *feval;
}

int HCL_UMinTR_d::Minimize( HCL_Functional_d & f,HCL_Vector_d & xcur )
{

   // Initialize

   Initialize( f.Domain() );

   feval = f.Evaluate( xcur );
   fevalFlag = 1;
   fcur = feval->Value();

   // Test for minimizer

   StepLength = -1.0;
   TermCode = StoppingTest( xcur );

   // Main loop

   Display0( f,xcur );
   while( !TermCode )
   {

      // Solve the trust region subproblem

      trsolver->Parameters().PutValue( "Delta",Delta );
      trsolver->Parameters().PutValue( "MaxStep",MaxStep*max(xnorm,TypxNorm) );
      trsolver->Parameters().PutValue( "MinStep",MinStep*max(xnorm,TypxNorm) );
      trsolver->Solve( feval->HessianRef(),feval->GradientRef(),*step );

      // Compute the predicted and actual reductions

      xnext->Add( xcur,*step );
      HCL_EvaluateFunctional_d * fevalnext = f.Evaluate( *xnext );
      fnext = fevalnext->Value();
      ared = fcur - fnext;
      if( trsolver->Parameters().GetValue( "Reduction",pred ) )
      {
         cerr << "Error in HCL_UMinTR_d::Minimize: trust region solver "
                 "did not set TRSolver::Reduction" << endl;
         HCL_delete( fevalnext );
         HCL_delete( feval );
         fevalFlag = 0;
         Clean();
         exit(1);
      }
      int trcode;
      if( trsolver->Parameters().GetValue( "TermCode",trcode ) )
      {
         cerr << "Error in HCL_UMinTR_d::Minimize: trust region solver "
                 "did not set TRSolver::TermCode" << endl;
         HCL_delete( fevalnext );
         HCL_delete( feval );
         fevalFlag = 0;
         Clean();
         exit(1);
      }

      // take step if ratio is sufficiently large

      ratio = ared/pred;
      if( ratio >= LowerTol && ared > 0.0 && trcode != FAILED )
      {
         HCL_delete( feval );
         feval = fevalnext;
         xcur.Copy( *xnext );
         fcur = fnext;
         Itn++;
      }
      else
         HCL_delete( fevalnext );

      // Update the trust region radius

      if(trsolver->Parameters().GetValue("StepLength",StepLength))
      {
         cerr << "Error in HCL_UMinTR_d::Minimize: trust region solver "
                 "did not set TRSolver::StepLength" << endl;
         HCL_delete( feval );
         fevalFlag = 0;
         Clean();
         exit(1);
      }
      if( ratio >= UpperTol && ared > 0.0 )
         Delta = max( Delta,2.0*StepLength );
      else if( ratio < LowerTol || ared <= 0.0 || trcode == FAILED )
         Delta = min( Delta,StepLength )/2.0;

      if( Delta > MaxStep ) Delta = MaxStep;

      // Test stopping criteria

      if( trcode != FAILED )
         TermCode = StoppingTest( xcur );
      Display( xcur );
   }
   Clean();
   return TermCode;
}

void HCL_UMinTR_d::Display( const HCL_Vector_d & x )
{
   if( DispFlag > 1 || (DispFlag && TermCode) )
   {
      cout << endl << "=============HCL_UMinTR_d=============" << endl;
      cout << "HCL_UMinTR_d: iteration: " << Itn << endl;
      cout << " f = " << fcur << endl;
      cout << "||g|| = " << gnorm << endl;
      cout << "pred: " << pred << " ared: " << ared << " ratio: " << ratio
           << endl;
      cout << "Delta: " << Delta << endl;
      cout << endl << "======================================" << endl;
   }
   if( (DumpFlag > 1 || DumpFlag && TermCode) && dump != NULL )
   {
      *dump << endl << "=============HCL_UMinTR_d=============" << endl;
      *dump << "HCL_UMinTR_d: iteration: " << Itn << endl;
      *dump << " f = " << fcur << endl;
      *dump << "||g|| = " << gnorm << endl;
      *dump << "pred: " << pred << " ared: " << ared << " ratio: " << ratio
           << endl;
      *dump << "Delta: " << Delta << endl;
      *dump << endl << "======================================" << endl;
   }
   if( TermCode )
   {
      if( DispFlag )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               cout << "HCL_UMinTR_d: Terminated: Possible minimizer" << endl;
               break;
            case PossibleConvergence:
               cout << "HCL_UMinTR_d: Terminated: Possible convergence"
                    << endl;
               break;
            case IterationLimit:
               cout << "HCL_UMinTR_d: Terminated: Iteration limit reached"
                    << endl;
               break;
            case PossibleDivergence:
               cout << "HCL_UMinTR_d: Terminated: Possible divergence"
                    << endl;
               break;
            default:
               cout << "HCL_UMinTR_d: Terminated: Unknown termination code"
                    << endl;
               break;
         }
      }
      if( DumpFlag && dump != NULL )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               *dump << "HCL_UMinTR_d: Terminated: Possible minimizer" << endl;
               break;
            case PossibleConvergence:
               *dump << "HCL_UMinTR_d: Terminated: Possible convergence"
                     << endl;
               break;
            case IterationLimit:
               *dump << "HCL_UMinTR_d: Terminated: Iteration limit reached"
                     << endl;
               break;
            case PossibleDivergence:
               *dump << "HCL_UMinTR_d: Terminated: Possible divergence"
                    << endl;
               break;
            default:
               *dump << "HCL_UMinTR_d: Terminated: Unknown termination code"
                     << endl;
               break;
         }
      }
   }
   if( TraceSteps && Steps != NULL )
      x.Write( *Steps );
}

void HCL_UMinTR_d::Display0( const HCL_Functional_d & f,
                             const HCL_Vector_d & x )
{
   if( DispFlag > 1 || (DispFlag && TermCode) )
   {
      cout << endl << "=============HCL_UMinTR_d=============" << endl;
      cout << "iteration: " << Itn << endl;
      cout << " f = " << fcur << endl;
      cout << "||g|| = " << gnorm << endl;
      if( Itn > 0 )
         cout << "pred: " << pred << " ared: " << ared << " ratio: " << ratio
              << endl;
      cout << "Delta: " << Delta << endl;
      cout << endl << "======================================" << endl;
   }
   if( (DumpFlag > 1 || (DumpFlag && TermCode)) && dump != NULL )
   {
      *dump << "Summary of HCL_UMinTR_d run" << endl;
      *dump << "-------------------------" << endl << endl;
      *dump << "f = " << endl;
      f.Write( *dump );
      *dump << endl << "ParamTable = " << endl;
      ((HCL_UMinTR_d*)this)->Parameters().Write( *dump );
      *dump << endl << endl;
      *dump << endl << "=============HCL_UMinTR_d=============" << endl;
      (*dump) << "iteration: " << Itn << endl;
      (*dump) << " f = " << fcur << endl;
      (*dump) << "||g|| = " << gnorm << endl;
      if( Itn > 0 )
         *dump << "pred: " << pred << " ared: " << ared << " ratio: " << ratio
               << endl;
      *dump << "Delta: " << Delta << endl;
      *dump << endl << "======================================" << endl;
   }
   if( TermCode )
   {
      if( DispFlag )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               cout << "HCL_UMinTR_d: Terminated: Possible minimizer" << endl;
               break;
            case PossibleConvergence:
               cout << "HCL_UMinTR_d: Terminated: Possible convergence"
                    << endl;
               break;
            case IterationLimit:
               cout << "HCL_UMinTR_d: Terminated: Iteration limit reached"
                    << endl;
               break;
            default:
               cout << "HCL_UMinTR_d: Terminated: Unknown termination code"
                    << endl;
               break;
         }
      }
      if( DumpFlag && dump != NULL )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               *dump << "HCL_UMinTR_d: Terminated: Possible minimizer" << endl;
               break;
            case PossibleConvergence:
               *dump << "HCL_UMinTR_d: Terminated: Possible convergence"
                     << endl;
               break;
            case IterationLimit:
               *dump << "HCL_UMinTR_d: Terminated: Iteration limit reached"
                     << endl;
               break;
            default:
               *dump << "HCL_UMinTR_d: Terminated: Unknown termination code"
                     << endl;
               break;
         }
      }
   }
}

int HCL_UMinTR_d::StoppingTest( const HCL_Vector_d & x )
{
   // Compute the norm of the gradient

   if( lsolver != NULL )
   {
      lsolver->Solve( *C,feval->GradientRef(),*tmpvec );
      gnorm = sqrt( feval->GradientRef().Inner( *tmpvec ) );
   }
   else if( C != NULL )
   {
      C->InvImage( feval->GradientRef(),*tmpvec );
      gnorm = sqrt( feval->GradientRef().Inner( *tmpvec ) );
   }
   else
      gnorm = feval->GradientRef().Norm();

   // Check for an approximate stationary point
   
   // Compute the norm of x

   if( C != NULL )
   {
      C->Image( x,*tmpvec );
      xnorm = sqrt( x.Inner( *tmpvec ) );
   }
   else
      xnorm = x.Norm();

   double RelGradNorm = max( xnorm,TypxNorm )*gnorm/max( abs(fcur),Typf );

   if( (RelGradNorm < GradTol && Itn > 0) ||
       (RelGradNorm < GradTol/1000.0 && Itn == 0) )
      return PossibleMinimizer;

   // Note: StepLength = -1.0 indicates no step has been taken yet

   if( StepLength != -1.0 && StepLength/max( xnorm,TypxNorm )  <= MinStep )
      return PossibleConvergence;

   if( StepLength >= 0.99*MaxStep )
   {
      CscMax++;
      if( CscMax == CscMaxLimit )
         return PossibleDivergence;
   }
   else
      CscMax = 0;

   if( Itn >= MaxItn )
      return IterationLimit;
   return 0;
}

ostream & HCL_UMinTR_d::Write( ostream & str ) const
{
   str << "HCL_UMinTR_d trust region unconstrained minimization algorithm"
       << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}
