/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include <iostream.h>
#include <stdio.h>
#include <float.h>
#include "maxmin_d.h"

#include "HCL_UMin_lbfgs_d.h"
#include "HCL_Rn_d.h"

HCL_UMin_lbfgs_d::HCL_UMin_lbfgs_d( const HCL_UMin_lbfgs_d & um )
{
   cerr << "Error: HCL_UMin_lbfgs_d copy constructor called" << endl;
   exit(1);
}

HCL_UMin_lbfgs_d::HCL_UMin_lbfgs_d(HCL_LineSearch_d *line_search,char * fname)
{
   if( fname != NULL )
      ParamTable = new Table( fname,"UMinlbfgs UMin" );
   else
      ParamTable = new Table;
   if( line_search )
   {
      lsearch = line_search;
      lsearch->IncCount();
   }
   else
      lsearch = new HCL_LineSearch_DS_d( fname );
   S = NULL;
   lsolver = NULL;
   dump = NULL;
   Steps = NULL;
   feval = NULL;
   fevalFlag = 0;
}

HCL_UMin_lbfgs_d::~HCL_UMin_lbfgs_d()
{
   if( fevalFlag ) HCL_delete( feval );
   HCL_delete( lsolver );
   HCL_delete( S );
   HCL_delete( lsearch );
   HCL_delete( ParamTable );
}

void HCL_UMin_lbfgs_d::Initialize( HCL_VectorSpace_d & dom )
{
   if( ParamTable->GetValue( "MaxItn",MaxItn ) )
   {
      MaxItn = 100;
      ParamTable->PutValue( "MaxItn",MaxItn );
   }
   if( ParamTable->GetValue( "Typf",Typf ) )
   {
      Typf = 1.0;
      ParamTable->PutValue( "Typf",Typf );
   }
   if( ParamTable->GetValue( "TypxNorm",TypxNorm ) )
   {
      TypxNorm = 1.0;
      ParamTable->PutValue( "TypxNorm",TypxNorm );
   }
   if( ParamTable->GetValue( "MaxUpdates",MaxUpdates ) )
   {
      MaxUpdates = 4;
      ParamTable->PutValue( "MaxUpdates",MaxUpdates );
   }
   if( ParamTable->GetValue( "GradTol",GradTol ) )
   {
      GradTol = 1.0e-2;
      ParamTable->PutValue( "GradTol",GradTol );
   }
   int MinFlag = ParamTable->GetValue( "MinStep",MinStep );
   int MaxFlag = ParamTable->GetValue( "MaxStep",MaxStep );
   if( MinFlag && MaxFlag || (!MinFlag && !MaxFlag && MinStep >= MaxStep) )
   {
      MaxStep = 1.0e20;
      MinStep = 1.0e-20;
      ParamTable->PutValue( "MaxStep",MaxStep );
      ParamTable->PutValue( "MinStep",MinStep );
   }
   else if( MinFlag && !MaxFlag )
   {
      MinStep = 1.0e-40*MaxStep;
      ParamTable->PutValue( "MinStep",MinStep );
   }
   else if (MaxFlag && !MinFlag )
   {
      MaxStep = 1.0e40*MinStep;
      ParamTable->PutValue( "MaxStep",MaxStep );
   }
   if( ParamTable->GetValue( "CscMaxLimit",CscMaxLimit ) )
   {
      CscMaxLimit = 5;
      ParamTable->PutValue( "CscMaxLimit",CscMaxLimit );
   }
   if( ParamTable->GetValue( "DispFlag",DispFlag ) )
   {
      DispFlag = 0;
      ParamTable->PutValue( "DispFlag",DispFlag );
   }
   if( ParamTable->GetValue( "DumpFlag",DumpFlag ) )
   {
      DumpFlag = 0;
      ParamTable->PutValue( "DumpFlag",DumpFlag );
   }
   if( ParamTable->GetValue( "DispPrecision",DispPrecision ) )
   {
      DispPrecision = 6;
      ParamTable->PutValue( "DispPrecision",DispPrecision );
   }
   OldDispPrec = cout.precision( DispPrecision );
   if( ParamTable->GetValue( "DumpPrecision",DumpPrecision ) )
   {
      DumpPrecision = DispPrecision;
      ParamTable->PutValue( "DumpPrecision",DumpPrecision );
   }
   if( DumpFlag )
   {
      if( ParamTable->GetValue( "DumpFile",DumpFile ) )
      {
         strcpy( DumpFile,"HCL_UMin_lbfgs.DumpFile" );
         ParamTable->PutValue( "DumpFile",DumpFile );
      }
      dump = new ofstream( DumpFile,ios::out );
      dump->precision( DumpPrecision );
   }
   else
      dump = NULL;
   if( ParamTable->GetValue( "TraceSteps",TraceSteps ) )
   {
      TraceSteps = 0;
      ParamTable->PutValue( "TraceSteps",TraceSteps );
   }
   if( TraceSteps )
   {
      if( ParamTable->GetValue( "StepFile",StepFile ) )
      {
         strcpy( StepFile,"HCL_UMin_lbfgs.StepFile" );
         ParamTable->PutValue( "StepFile",StepFile );
      }
      Steps = new ofstream( StepFile,ios::out );
   }
   else
      Steps = NULL;

   if( ParamTable->GetValue( "InvHessianScale",InvHessianScale ) )
   {
      InvHessianScale = 1.0;
      ParamTable->PutValue( "InvHessianScale",InvHessianScale );
   }

   if( S )
   {
      if( S->Domain() != dom )
      {
         cerr << "Error in HCL_UMin_lbfgs_d::Initialize: domain does not match"
                 " preconditioner" << endl;
         exit(1);
      }
      tmpvec = dom.Member();
      lsearch->SetScaling( S );
   }
   else
   {
      tmpvec = NULL;
      lsearch->UnSetScaling();
   }

   H0 = new HCL_ScaleLinearOp_d( &dom,InvHessianScale );
   H = new HCL_lmbfgsOp1_d( H0,MaxUpdates,S,lsolver );
   xnext = dom.Member();
   dir = dom.Member();
   Itn = 0;
   CscMax = 0;
}

void HCL_UMin_lbfgs_d::Clean()
{
   cout.precision( OldDispPrec );
   HCL_delete( dir );
   HCL_delete( xnext );
   HCL_delete( H );
   HCL_delete( H0 );
   HCL_delete( tmpvec );
   if( TraceSteps )
      delete Steps;
   if( DumpFlag )
      delete dump;
}

HCL_EvaluateFunctional_d & HCL_UMin_lbfgs_d::LastEval() const
{
   if( !fevalFlag )
   {
      cerr << "Error in HCL_UMin_lbfgs_d::LastEval: no evaluation "
              "object" << endl;
      exit(1);
   }
   return *feval;
}

void HCL_UMin_lbfgs_d::SetScaling( HCL_LinearOp_d * SS,
                                   HCL_LinearSolver_d * ls )
{
   HCL_delete( S );
   HCL_delete( lsolver );
   S = SS;if( SS ) SS->IncCount();
   lsolver = ls;if( ls ) ls->IncCount();
}
   
void HCL_UMin_lbfgs_d::UnSetScaling()
{
   HCL_delete( S );
   HCL_delete( lsolver );
   S = NULL;
   lsolver = NULL;
}

int HCL_UMin_lbfgs_d::Minimize( HCL_Functional_d & f,HCL_Vector_d & x )
{
   int ls_result, MaxTkn;

   if( x.Space() != f.Domain() )
   {
      cerr << "Error in HCL_UMin_lbfgs_d::Minimize: vector is not in domain"
           << endl;
      exit(1);
   }

   // initialize parameters and temporary vectors

   Initialize( f.Domain() );

   // Evaluate f(x0) and grad f(x0)

   if( fevalFlag ) HCL_delete( feval );
   feval = f.Evaluate( x );
   fevalFlag = 1;
   f_x = feval->ValueRef();

   // Check to see if initial point is stationary

   StDescDir = 1;
   TermCode = StoppingTest( x,*feval,0, 0 );

   while( !TermCode )
   {
      Display( f,x );

      // Compute the lmBFGS search direction

      H->Image( feval->GradientRef(),*dir );
      dir->Neg();

      // Perform line search

      HCL_EvaluateFunctional_d *tmp_eval = feval;
      lsearch->Parameters().PutValue( "MaxStep",MaxStep*max(xnorm,TypxNorm) );
      lsearch->Parameters().PutValue( "MinStep",MinStep*max(xnorm,TypxNorm) );
      feval = lsearch->Search( f,x,*xnext,*dir,tmp_eval );
      if( feval == NULL ) feval = tmp_eval;
      if( (lsearch->Parameters()).GetValue("TermCode", ls_result) )
      {
         cerr << "Error in HCL_UMin_lbfgs_d::Minimize: line search failed "
                 "to set LineSearch::TermCode" << endl;
         exit(1);
      }
      if( (lsearch->Parameters()).GetValue("MaxTkn", MaxTkn) )
      {
         cerr << "Error in HCL_UMin_lbfgs_d::Minimize: line search failed "
                 "to set LineSearch::MaxTkn" << endl;
         exit(1);
      }

      Itn++;
      TermCode = StoppingTest(x,*feval,ls_result, MaxTkn);

      if( TermCode )
      {
         if( ls_result >= 0 )
         {
            HCL_delete( tmp_eval );
            x.Copy( *xnext );
            f_x = feval->ValueRef();
         }
         Display( f,x );
         Clean();
         return TermCode;
      }

      // Update lmBFGS inverse Hessian approximation

      if( ls_result )
      {
         H->Reset();
         StDescDir = 1;
      }
      else
      {
         H->Update( x,*xnext,tmp_eval->GradientRef(),feval->GradientRef() );
         StDescDir = 0;
      }

      // Prepare for next iteration

      if( ls_result >= 0 )
      {
         HCL_delete( tmp_eval );
         x.Copy( *xnext );
         f_x = feval->ValueRef();
      }
   }
   Clean();
   return TermCode;
}

void HCL_UMin_lbfgs_d::Display( const HCL_Functional_d & f,
                                const HCL_Vector_d & x )
{
   if( DispFlag > 1 || (DispFlag && TermCode) )
   {
      cout << endl << "===========HCL_UMin_lbfgs_d=============" << endl;
      cout << "iteration: " << Itn << endl;
      cout << " f = " << f_x << endl;
      cout << "||g|| = " << gnorm << endl;
      if( StDescDir )
         cout << "Steepest descent direction" << endl;
      cout << endl << "========================================" << endl;
   }
   if( (DumpFlag > 1 || (DumpFlag && TermCode)) && dump != NULL )
   {
      if( Itn == 0 )
      {
         *dump << "Summary of HCL_UMin_lbfgs_d run" << endl;
         *dump << "-------------------------" << endl << endl;
         *dump << "f = " << endl;
         f.Write( *dump );
         *dump << endl << "ParamTable = " << endl;
         Parameters().Write( *dump );
         *dump << endl << endl;
      }
      (*dump) << endl << "======================" << endl;
      (*dump) << "iteration: " << Itn << endl;
      (*dump) << " f = " << f_x << endl;
      (*dump) << "||g|| = " << gnorm << endl;
      if( StDescDir )
         (*dump) << "Steepest descent direction" << endl;
      (*dump) << endl << "======================" << endl;
   }
   if( TermCode )
   {
      if( DispFlag )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               cout << "HCL_UMin_lbfgs_d terminated: Possible minimizer" 
                    << endl;
               break;
            case PossibleConvergence:
               cout << "HCL_UMin_lbfgs_d terminated: Possible convergence"
                    << endl;
               break;
            case LineSearchFailed:
               cout << "HCL_UMin_lbfgs_d terminated: Line search failed"
                    << endl;
               break;
            case PossibleDivergence:
               cout << "HCL_UMin_lbfgs_d terminated: Possible divergence"
                    << endl;
               break;
            case IterationLimit:
               cout << "HCL_UMin_lbfgs_d terminated: Iteration limit reached"
                    << endl;
               break;
            case InaccurateGradient:
               cout << "HCL_UMin_lbfgs_d terminated: Line search failed "
                       "because of inaccurate gradient" << endl;
               break;
            default:
               cout << "HCL_UMin_lbfgs_d terminated: Unknown termination code"
                    << endl;
               break;
         }
      }
      if( DumpFlag && dump != NULL )
      {
         switch( TermCode )
         {
            case PossibleMinimizer:
               (*dump) << "Terminated: Possible minimizer" << endl;
               break;
            case PossibleConvergence:
               (*dump) << "Terminated: Possible convergence" << endl;
               break;
            case LineSearchFailed:
               (*dump) << "Terminated: Line search failed" << endl;
               break;
            case PossibleDivergence:
               (*dump) << "Terminated: Possible divergence" << endl;
               break;
            case IterationLimit:
               (*dump) << "Terminated: Iteration limit reached" << endl;
               break;
            case InaccurateGradient:
               (*dump) << "Terminated: Line search failed due to "
                          "inaccurate gradient" << endl;
               break;
            default:
               (*dump) << "Terminated: Unknown termination code" << endl;
               break;
         }
      }
   }
   if( TraceSteps && Steps != NULL )
      x.Write( *Steps );
}

int HCL_UMin_lbfgs_d::StoppingTest( const HCL_Vector_d & x,
                                     HCL_EvaluateFunctional_d & eval,
                                     int ls_result, int MaxTkn)
{
   if( StDescDir && ls_result == HCL_LineSearch_d::InitialSlopePositive )
      return InaccurateGradient;
   else if( StDescDir && ls_result < 0 )
      return LineSearchFailed;
   else if( ls_result < 0 )
      return 0;

   // Compute the norm of the gradient

   if( S )
   {
      if( lsolver )
         lsolver->Solve( *S,eval.GradientRef(),*tmpvec );
      else
         S->InvImage( eval.GradientRef(),*tmpvec );
      gnorm = sqrt( eval.GradientRef().Inner( *tmpvec ) );
   }
   else
      gnorm = eval.GradientRef().Norm();

   // Check for an approximate stationary point
   
   if( S )
   {
      S->Image( x,*tmpvec );
      xnorm = sqrt( x.Inner( *tmpvec ) );
   }
   else
      xnorm = x.Norm();

   double RelGradNorm = max( xnorm,TypxNorm )*gnorm/
                        max( abs(eval.ValueRef()),Typf );


   if( Itn == 0 )
      if( RelGradNorm < GradTol/1000.0 )
         return PossibleMinimizer;
      else if( Itn >= MaxItn )
         return IterationLimit;
      else
         return 0;

   if( RelGradNorm < GradTol )
      return PossibleMinimizer;

   // Compute the norm of the step

   dir->Sub( *xnext,x );
   if( S )
   {
      S->Image( *dir,*tmpvec );
      snorm = sqrt(dir->Inner( *tmpvec ));
   }
   else
      snorm = dir->Norm();

   // Check for convergence

   if( snorm/max( TypxNorm,xnorm) < MinStep )
      return PossibleConvergence;

   // Check for iteration limit

   if( Itn >= MaxItn )
      return IterationLimit;

   // Check for divergence

   if( MaxTkn )
   {
      CscMax++;
      if( CscMax == CscMaxLimit )
         return PossibleDivergence;
   }
   else
      CscMax = 0;

   // No stopping criterion satisfied

   return 0;
}

ostream & HCL_UMin_lbfgs_d::Write( ostream & str ) const
{
   str << "HCL_UMin_lbfgs_d limited memory BFGS algorithm"
       << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}
