/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#include "HCL_PartitionedRnVector_d.H"

HCL_PartitionedRnSpace_d::HCL_PartitionedRnSpace_d(){
  cerr<<"Error: no default constructor for HCL_PartitionedRnSpace_d"<<endl;
  exit(1);
}

HCL_PartitionedRnSpace_d::HCL_PartitionedRnSpace_d(
                          const HCL_PartitionedRnSpace_d & x)
{
   n = x.n;
   IdenticalFlag = x.IdenticalFlag;
   if( IdenticalFlag )
   {
      rnsp = new HCL_RnSpace_d*[1];
      rnsp[0] = new HCL_RnSpace_d( *(x.rnsp[0]) );
      numinit = n;
   }
   else
   {
      rnsp = new HCL_RnSpace_d*[n];
      int i;
      for( i=0;i<n;i++ )
         rnsp[i] = new HCL_RnSpace_d( *(x.rnsp[i]) );
      numinit = x.numinit;
   }
}

HCL_PartitionedRnSpace_d::HCL_PartitionedRnSpace_d(int mm, int nn)
{
   n=nn;
   numinit = n;
   rnsp = new HCL_RnSpace_d*[1];
   rnsp[0] = new HCL_RnSpace_d( mm );
   IdenticalFlag = 1;
}

HCL_PartitionedRnSpace_d::HCL_PartitionedRnSpace_d( int nn )
{
   n=nn;
   numinit = 0;
   rnsp = new HCL_RnSpace_d*[n];
   IdenticalFlag = 0;
}

HCL_PartitionedRnSpace_d::~HCL_PartitionedRnSpace_d()
{
   if( IdenticalFlag )
      HCL_delete( rnsp[0] );
   else
   {
      int i;
      for( i=0;i<numinit;i++ )
         HCL_delete( rnsp[i] );
   }
   delete[] rnsp;
}

void * HCL_PartitionedRnSpace_d::GetType() const {
  return (void *)(&Type);
}

HCL_Vector_d * HCL_PartitionedRnSpace_d::Member() const
{
   if( numinit < n )
   {
      cerr << "Error in HCL_PartitionedRnSpace_d::Member: space is not"
           << endl << "initialized" << endl;
      exit(1);
   }
   return new HCL_PartitionedRnVector_d( (HCL_PartitionedRnSpace_d*)this );
}

HCL_VectorSpace_d * HCL_PartitionedRnSpace_d::Replicate() const
{
   if( numinit < n )
   {
      cerr << "Error in HCL_PartitionedRnSpace_d::Replicate: space is not"
           << endl << "initialized" << endl;
      exit(1);
   }
   HCL_PartitionedRnSpace_d * v;
   if( IdenticalFlag )
      v = new HCL_PartitionedRnSpace_d( rnsp[0]->Dim(),n );
   else
   {
      v = new HCL_PartitionedRnSpace_d( n );
      int i;
      for( i=0;i<n;i++ )
         v->SetNext( rnsp[i] );
   }
   return v;
}

int HCL_PartitionedRnSpace_d::Number() const
{
  return n;
}

HCL_VectorSpace_d & HCL_PartitionedRnSpace_d::operator()( int i ) const
{
   if( IdenticalFlag )
      return *(rnsp[0]);
   else
      return *(rnsp[i-1]);
}

void HCL_PartitionedRnSpace_d::SetNext( int m )
{
   if( numinit==n )
   {
      cerr << "Error in HCL_PartitionedRnSpace_d::SetNext: space is already"
           << endl << "initialized" << endl;
      exit(1);
   }
   rnsp[numinit] = new HCL_RnSpace_d( m );
   numinit++;
}

void HCL_PartitionedRnSpace_d::SetNext( HCL_RnSpace_d * s )
{
   if( numinit==n )
   {
      cerr << "Error in HCL_PartitionedRnSpace_d::SetNext: space is already"
           << endl << "initialized" << endl;
      exit(1);
   }
   rnsp[numinit] = s;s->IncCount();
   numinit++;
}

HCL_PartitionedRnVector_d::HCL_PartitionedRnVector_d()
{
   cerr<<"Error: no default constructor for HCL_PartitionedRnVector_d"<<endl;
   exit(1);
}

HCL_PartitionedRnVector_d::HCL_PartitionedRnVector_d(
                           const HCL_PartitionedRnVector_d & p)
{
  cerr<<"Error: no copy constructor for HCL_PartitionedRnVector_d"<<endl;
  exit(1);
}

HCL_PartitionedRnVector_d::HCL_PartitionedRnVector_d(int mm, int nn)
{
   sp = new HCL_PartitionedRnSpace_d(mm,nn);
   n=nn;
   rnvec = new HCL_RnVector_d*[n];
   data = new double[mm*nn];
   for (int i=0;i<n;i++)
   {
     rnvec[i]=new HCL_RnVector_d(mm,&(data[i*mm]));
   }
}

HCL_PartitionedRnVector_d::HCL_PartitionedRnVector_d(
                           HCL_PartitionedRnSpace_d * s )
{
   sp = s;s->IncCount();
   n = sp->Number();
   rnvec = new HCL_RnVector_d*[n];
   int * D = new int[n];
   int i;
   dim = 0;
   for( i=1;i<=n;i++ )
   {
      D[i-1] = (*sp)(i).Dim();
      dim += D[i-1];
   }
   data = new double[dim];
   int p=0;
   for( i=1;i<=n;i++ )
   {
      rnvec[i-1] = new HCL_RnVector_d( D[i-1],&(data[p]) );
      p += D[i-1];
   }
   delete[] D;
}
   
HCL_PartitionedRnVector_d::HCL_PartitionedRnVector_d( char * fname )
{
   ifstream in( fname );
   if( !in )
   {
      cerr << "Error in HCL_PartitionedRnVector_d constructor: file "
           << fname << " cannot be opened" << endl;
      exit(1);
   }
   in >> n;
   sp = new HCL_PartitionedRnSpace_d( n );
   dim=0;
   int D,i,j;
   double junk;
   for( i=0;i<n;i++ )
   {
      in >> D;
      dim += D;
      for( j=0;j<D;j++ ) in >> junk;
      sp->SetNext( D );
   }
   rnvec = new HCL_RnVector_d*[n];
   data = new double[dim];
   ifstream in1( fname );
   int p=0;
   in1 >> n;
   for( i=0;i<n;i++ )
   {
      in1 >> D;
      rnvec[i]=new HCL_RnVector_d( D,&(data[p]));
      for( j=p;j<p+D;j++ )
         in1 >> data[j];
      p += D;
   }
}

HCL_PartitionedRnVector_d::~HCL_PartitionedRnVector_d()
{
   for (int i=0;i<n;i++)
      HCL_delete(rnvec[i]);
   delete [] rnvec;
   delete [] data;
   HCL_delete(sp);
}

HCL_Vector_d & HCL_PartitionedRnVector_d::operator()(int i) const
{ 
   if (i<1 || i>n)
   {
      cerr<<"Error: attempt to access component "<<i<<" of"<<endl;
      cerr<<"HCL_PartitionedRnVector_d with "<<n<<" components"<<endl;
      exit(1);
   }
   return *(rnvec[i-1]);
}

HCL_VectorSpace_d & HCL_PartitionedRnVector_d::Space() const
{
   return *sp;
}

void HCL_PartitionedRnVector_d::Save( char * fname ) const
{
   ofstream file( fname,ios::out );
   file.precision( 16 );
   file << n << endl;
   int i,j,p=0;
   for( i=0;i<n;i++ )
   {
      int D = rnvec[i]->Dim();
      file << D << endl;
      for( j=p;j<p+D;j++ )
         file << data[j] << endl;
      p += D;
   }
}
