#include "HCL_RnRecorder_d.h"

HCL_RnRecorder_d::HCL_RnRecorder_d()
{
   cerr << "The HCL_RnRecorder_d default constructor is not implemented"
        << endl;
   exit(1);
}

HCL_RnRecorder_d::HCL_RnRecorder_d( const HCL_RnRecorder_d & )
{
   cerr << "The HCL_RnRecorder_d copy constructor is not implemented"
        << endl;
   exit(1);
}

HCL_RnRecorder_d::HCL_RnRecorder_d( int dim,int m,char * fname )
{
   n = dim;
   maxnv = m;
   sp = new HCL_RnSpace_d( n );
   V = new GeneralMatrix_d( n,maxnv );
   nv = 0;
   ptr = 0;
   if( fname )
   {
      filename = new char[strlen(fname)+1];
      strcpy( filename,fname );
   }
   else
      filename = NULL;
}

HCL_RnRecorder_d::~HCL_RnRecorder_d()
{
   if( filename )
   {
      Save( filename );
      delete[] filename;
   }
   delete V;
   HCL_delete( sp );
}

void HCL_RnRecorder_d::Record( const HCL_Vector_d & Input )
{
   if( Input.Space() != *sp )
   {
      cerr << "Error in HCL_RnRecorder_d::Record: vector is not in the "
           << endl << "correct space" << endl;
      exit(1);
   }
   HCL_RnVector_d & x = (HCL_RnVector_d&)Input;
   if( nv >= maxnv )
   {
      cerr << "Error in HCL_RnRecorder_d::Record: too many vectors" << endl;
      exit(1);
   }
   
   nv++;ptr = nv;
   int i;
   for( i=1;i<=n;i++ )
      (*V)(i,nv) = x(i);
}

void HCL_RnRecorder_d::Rewind() const
{
   HCL_RnRecorder_d * t = (HCL_RnRecorder_d*)this; // casting away const
   if( nv )
      t->ptr = 1;
   else
      t->ptr = 0;
}

void HCL_RnRecorder_d::PlayBack( HCL_Vector_d & Output ) const
{
   if( Output.Space() != *sp )
   {
      cerr << "Error in HCL_RnRecorder_d::PlayBack: vector is not in the "
              "correct space" << endl;
      exit(1);
   }
   HCL_RnVector_d & x = (HCL_RnVector_d&)Output;
   if( nv == 0 )
   {
      cerr << "Error in HCL_RnRecorder_d::PlayBack: no vectors have been "
              "recorded" << endl;
      exit(1);
   }
   HCL_RnRecorder_d * t = (HCL_RnRecorder_d*)this; // casting away const
   if( ptr == 0 || ptr > nv ) (t->ptr) = nv;
   int i;
   for( i=1;i<=n;i++ ) x(i) = (*V)(i,ptr);
}

int HCL_RnRecorder_d::Next() const
{
   HCL_RnRecorder_d * t = (HCL_RnRecorder_d*)this; // casting away const
   if( ptr == nv )
      return 0;
   else
   {
      (t->ptr)++;
      return ptr;
   }
}

int HCL_RnRecorder_d::Number() const
{
   return nv;
}

void HCL_RnRecorder_d::Erase()
{
   nv = 0;ptr = 0;
}

void HCL_RnRecorder_d::Save( char * fname ) const
{
   ofstream out( fname,ios::out );
#if SINGLE_PRECISION
   out << setprecision(8);
#else
   out << setprecision(16);
#endif
   out << n << " " << nv << endl;
   int i,j;
   for( j=1;j<=nv;j++ )
      for( i=1;i<=n;i++ )
         out << (*V)(i,j) << endl;
}

ostream & HCL_RnRecorder_d::Write( ostream & str ) const
{
   str << "HCL_RnRecorder_d vector recorder object"
       << endl;
   return str;
}
