/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_Rn_d.h"

#ifndef RAND_MAX
#define RAND_MAX 32767
#endif

HCL_RnSpace_d::HCL_RnSpace_d()
{
   cerr << "Error: HCL_RnSpace_d default constructor called" << endl;
   exit(1);
}

int HCL_RnSpace_d::operator == ( const HCL_VectorSpace_d & xx ) const
{
   if( GetType() != xx.GetType() )
      return 0;
   HCL_RnSpace_d & x = (HCL_RnSpace_d&)xx;
   return (Dim() == x.Dim());
}

int HCL_RnSpace_d::operator != ( const HCL_VectorSpace_d & x ) const
{
   return !operator==( x );
}

HCL_Vector_d * HCL_RnSpace_d::Member() const { 
   return new HCL_RnVector_d( this->Dim() );
}

HCL_VectorSpace_d * HCL_RnSpace_d::Replicate() const
{
   return new HCL_RnSpace_d( this->Dim() );
}

ostream & HCL_RnSpace_d::Write( ostream & str ) const 
{
   str << "HCL_RnSpace_d with dimension " << dimension << endl;
   return str;
}

HCL_RnVector_d::HCL_RnVector_d()
{
   cerr << "Error: HCL_RnVector_d default constructor called" << endl;
   exit(1);
}

HCL_RnVector_d::HCL_RnVector_d( const HCL_RnVector_d & x )
{
  // no automatic save
  savename = NULL;
   sp = x.sp;sp->IncCount();
   int dim = sp->Dim();
   data = new double[dim];
   for( int i=0;i<dim;i++ ) data[i] = x.data[i];
   newdata=1;
}

HCL_RnVector_d::HCL_RnVector_d( HCL_RnSpace_d * s )
{
  // no automatic save
  savename = NULL;
   sp = s;s->IncCount();
   data = new double[s->Dim()];
   newdata=1;
}

HCL_RnVector_d::HCL_RnVector_d( const HCL_RnSpace_d & s, char * fname )
{
  sp = new HCL_RnSpace_d(s);
  data = new double[sp->Dim()];
   newdata=1;
  if ( fname != NULL ) {
    savename = new char[strlen(fname)+1];
    int i;
    for (i=0;i<strlen(fname)+1;i++) {
      savename[i]='\0';
    }
    strcpy(savename,fname);
  }
  else {
    savename = NULL;
  }
}

HCL_RnVector_d::HCL_RnVector_d( int n , double *d)
{
  // no automatic save
  savename = NULL;
   sp = new HCL_RnSpace_d( n );
   if (d) {
     data = d;
     newdata=0;
   }
   else {
     data = new double[n];
     newdata=1;
   }
}

HCL_RnVector_d::HCL_RnVector_d( char * fname,int save )
{
   ifstream filein( fname,ios::in | ios::nocreate );
   if( filein == NULL )
   {
      cerr << "Error in HCL_RnVector_d constructor: file " << fname
           << " does not exist" << endl;
      exit(1);
   }

   // NOTE BENE: automatic save feature added!
   int i;
   if (save) {
      savename = new char[strlen(fname)+1];
      for (i=0;i<strlen(fname)+1;i++) {
        savename[i]='\0';
      }
      strcpy(savename,fname);
   }
   else
     savename = NULL;

   int n;
   filein >> n;
   sp = new HCL_RnSpace_d( n );
   data = new double[n];
   newdata=1;
   for( i=0;i<n;i++ ) filein >> data[i];
}

HCL_RnVector_d::~HCL_RnVector_d()
{
  // automatic save feature
  if ( savename != NULL ) {
    Save(savename);
  }
  if (newdata) delete[] data;
  HCL_delete( sp );
}

void HCL_RnVector_d::Copy( const HCL_Vector_d & xx )
{
   // Check to verify that vectors belong to the same space

   if( Space() != xx.Space() )
   {
      cerr << "Error in HCL_RnVector_d::Copy: "
              "vectors do not belong to the same space" << endl;
      exit(1);
   }
   HCL_RnVector_d & x = (HCL_RnVector_d&)xx;

   // copy the data

   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] = x.data[i];
}

void HCL_RnVector_d::Neg()
{
   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] = -data[i];
}

void HCL_RnVector_d::Mul( const double & a )
{
   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] *= a;
}

void HCL_RnVector_d::Mul( const double & a,const HCL_Vector_d & xx )
{
 if( Space() != xx.Space() )
 {
   cerr << "Error in HCL_RnVector_d::Mul: vectors are not from the same space"
        << endl;
   exit(1);
 }
 HCL_RnVector_d & x = (HCL_RnVector_d&)xx;
 int n = Dim();
 for( int i=0;i<n;i++ ) data[i] = x.data[i]*a;
}

void HCL_RnVector_d::Add( const HCL_Vector_d & xx )
{
   if( Space() != xx.Space() )
   {
      cerr << "Error in HCL_RnVector_d::Add: vectors are not from the "
              "same space" << endl;
      exit(1);
   }
   HCL_RnVector_d & x = (HCL_RnVector_d&)xx;
   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] += x.data[i];
}

void HCL_RnVector_d::Add( const HCL_Vector_d & xx,const HCL_Vector_d & yy )
{
   if( Space() != xx.Space() )
   {
     cerr<<"Error in HCL_RnVector_d::Add: vectors are not from the same space" 
         << endl;
      exit(1);
   }
   HCL_RnVector_d & x = (HCL_RnVector_d&)xx;
   if( Space() != yy.Space() )
   {
      cerr <<"Error in HCL_RnVector_d::Add: vectors are not from the same space"
          << endl;
      exit(1);
   }
   HCL_RnVector_d & y = (HCL_RnVector_d&)yy;
   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] = x.data[i] + y.data[i];
}

double HCL_RnVector_d::Inner( const HCL_Vector_d & xx ) const
{
   if( Space() != xx.Space() )
   {
      cerr << "Error in HCL_RnVector_d::Inner: vectors are not from the "
               "same space" << endl;
      exit(1);
   }
   HCL_RnVector_d & x = (HCL_RnVector_d&)xx;
   int n = Dim();
   double a = 0.0;
   for( int i=0;i<n;i++ ) a += data[i]*x.data[i];
   return a;
}

void HCL_RnVector_d::Sub( const HCL_Vector_d & xx )
{
  if( Space() != xx.Space() )
  {
    cerr << "Error in HCL_RnVector_d::Sub: vectors are not from the same space"
         << endl;
     exit(1);
  }
  HCL_RnVector_d & x = (HCL_RnVector_d&)xx;
  int n = Dim();
  for( int i=0;i<n;i++ ) data[i] -= x.data[i];
}

void HCL_RnVector_d::Sub( const HCL_Vector_d & xx,
                    const HCL_Vector_d & yy )
{
  if( Space() != xx.Space() )
  {
    cerr << "Error in HCL_RnVector_d::Sub: vectors are not from the same space"
         << endl;
     exit(1);
  }
  HCL_RnVector_d & x = (HCL_RnVector_d&)xx;
  if( Space() != yy.Space() )
  {
    cerr << "Error in HCL_RnVector_d::Sub: vectors are not from the same space"
         << endl;
     exit(1);
  }
  HCL_RnVector_d & y = (HCL_RnVector_d&)yy;
  int n = Dim();
  for( int i=0;i<n;i++ ) data[i] = x.data[i] - y.data[i];
}

void HCL_RnVector_d::ScaleAdd( const double  & a,
                         const HCL_Vector_d & xx )
{
   if( Space() != xx.Space() )
   {
      cerr << "Error in HCL_RnVector_d::ScaleAdd: "
              "vectors are not from the same space" << endl;
      exit(1);
   }
   HCL_RnVector_d & x = (HCL_RnVector_d&)xx;
   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] = a*data[i] + x.data[i];
}

void HCL_RnVector_d::AddScale( const double  & a,
                         const HCL_Vector_d & xx )
{
   if( Space() != xx.Space() )
   {
      cerr << "Error in HCL_RnVector_d::AddScale: "
              "vectors are not from the same space" << endl;
      exit(1);
   }
   HCL_RnVector_d & x = (HCL_RnVector_d&)xx;
   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] = data[i] + x.data[i]*a;
}

void HCL_RnVector_d::AddScale( const double  & a,
                         const HCL_Vector_d & xx,
                         const HCL_Vector_d & yy )
{
   if( Space() != xx.Space() )
   {
      cerr << "Error in HCL_RnVector_d::AddScale: "
              "vectors are not from the same space" << endl;
      exit(1);
   }
   HCL_RnVector_d & x = (HCL_RnVector_d&)xx;
   if( Space() != yy.Space() )
   {
      cerr << "Error in HCL_RnVector_d::AddScale: "
              "vectors are not from the same space" << endl;
      exit(1);
   }
   HCL_RnVector_d & y = (HCL_RnVector_d&)yy;
   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] = x.data[i] + y.data[i]*a;
}

double HCL_RnVector_d::Norm() const
{
   double a = 0.0;
   int n = Dim();
   for( int i=0;i<n;i++ ) a += data[i]*data[i];
   return sqrt(a);
}

double HCL_RnVector_d::Norm2() const
{
   int n = Dim();
   double a = 0.0;
   for( int i=0;i<n;i++ ) a += data[i]*data[i];
   return a;
}

void HCL_RnVector_d::Fill( double a )
{
   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] = a;
}

void HCL_RnVector_d::Zero()
{
   int n = Dim();
   for( int i=0;i<n;i++ ) data[i] = 0.0;
}

void HCL_RnVector_d::Random()
{
   int d = Dim();

   int DIST = 2;  // uniform distribution in (-1,1)
#ifdef SINGLE_PRECISION
   F77NAME(slarnv)(&DIST, rnd_seed, &d, data);
#else
   F77NAME(dlarnv)(&DIST, rnd_seed, &d, data);
#endif

#if 0
   double imax = 1.0/(double)RAND_MAX;
   for( int i=0;i<d;i++ )
   {
      double rn = (double)(rand());
      data[i] = rn*imax - 0.5;
   }
#endif
}

void HCL_RnVector_d::Save( char * fname )
{
   ofstream str( fname,ios::out );
   int n = Dim();
   str << n << endl;
   str << setprecision(16);
   for( int i=0;i<n;i++ ) str << "      " << data[i] << endl;
}

void HCL_RnVector_d::DiagScale( const HCL_Vector_d & x ) {
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::DiagScale"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  int i;
  for (i=0;i<Dim();i++ ) {
    data[i] *= xx.data[i];
  }
}

void HCL_RnVector_d::DiagScale( const HCL_Vector_d & x,
			  const HCL_Vector_d & y) {
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::DiagScale"<<endl;
    cerr<<"first parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"first parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }
  if (Space() != y.Space()) {
    cerr<<"Error: HCL_RnVector_d::DiagScale"<<endl;
    cerr<<"second parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"second parameter vector:"<<endl;
    y.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  HCL_RnVector_d & yy = (HCL_RnVector_d &)y;
  int i;
  for (i=0;i<Dim();i++ ) {
    data[i] = xx.data[i]*yy.data[i];
  }
}

void HCL_RnVector_d::DiagRecipScale( const HCL_Vector_d & x,
			       double tol) {
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::DiagRecipScale"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  int i;
  HCL_FPArch_d FP;
  double thistol=max(tol,FP.sfmin);
  
  for (i=0;i<Dim();i++ ) {
    if (xx.data[i] >= 0.0) {
      data[i] = data[i]/max(xx.data[i],thistol);
    }
    else {
      data[i] = data[i]/min(xx.data[i],-thistol);
    }
  }
}

void HCL_RnVector_d::DiagRecipScale( const HCL_Vector_d & x,
			       const HCL_Vector_d & y,
			       double tol) {
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::DiagRecipScale"<<endl;
    cerr<<"first parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"first parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }
  if (Space() != y.Space()) {
    cerr<<"Error: HCL_RnVector_d::DiagRecipScale"<<endl;
    cerr<<"second parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"second parameter vector:"<<endl;
    y.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  HCL_RnVector_d & yy = (HCL_RnVector_d &)y;
  int i;
  HCL_FPArch_d FP;
  double thistol=max(tol,FP.sfmin);
  
  for (i=0;i<Dim();i++ ) {
    if (xx.data[i] >= 0.0) {
      data[i] = yy.data[i]/max(xx.data[i],thistol);
    }
    else {
      data[i] = yy.data[i]/min(xx.data[i],-thistol);
    }
  }
}

double HCL_RnVector_d::Max() const {
  double M = -FLT_MAX;
  int i;
   
  for (i=0;i<Dim();i++ )
    M = (data[i]>M?data[i]:M);
  return M;
}

void HCL_RnVector_d::Max( const HCL_Vector_d & x ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Max"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;

  int i;
  for (i=0;i<Dim();i++ )

    data[i] = (data[i]<xx.data[i]?xx.data[i]:data[i]);
}

void HCL_RnVector_d::Max( const HCL_Vector_d & x,
		    const HCL_Vector_d & y ) {
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Max"<<endl;
    cerr<<"first parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"first parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }
  if (Space() != y.Space()) {
    cerr<<"Error: HCL_RnVector_d::Max"<<endl;
    cerr<<"second parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"second parameter vector:"<<endl;
    y.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  HCL_RnVector_d & yy = (HCL_RnVector_d &)y;

  int i;
  for (i=0;i<Dim();i++ )
    data[i] = (yy.data[i]<xx.data[i]?xx.data[i]:yy.data[i]);
}

double HCL_RnVector_d::Min() const {
  double M = FLT_MAX;
  int i;
   
  for (i=0;i<Dim();i++ )
    M = (data[i]<M?data[i]:M);
  return M;
}


void HCL_RnVector_d::Min( const HCL_Vector_d & x ) {
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Min"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  int i;
  for (i=0;i<Dim();i++ )
    data[i] = (data[i]>xx.data[i]?xx.data[i]:data[i]);
}

void HCL_RnVector_d::Min( const HCL_Vector_d & x,
		    const HCL_Vector_d & y ) {
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Min"<<endl;
    cerr<<"first parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"first parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }
  if (Space() != y.Space()) {
    cerr<<"Error: HCL_RnVector_d::Min"<<endl;
    cerr<<"second parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"second parameter vector:"<<endl;
    y.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  HCL_RnVector_d & yy = (HCL_RnVector_d &)y;

  int i;
  for (i=0;i<Dim();i++ )
    data[i] = (yy.data[i]>xx.data[i]?xx.data[i]:yy.data[i]);
}

void HCL_RnVector_d::Greater( double t ) {
  int i;
  for (i=0;i<Dim();i++ )
    data[i] = (data[i]>t?1.0:0.0);
}

void HCL_RnVector_d::Greater( const HCL_Vector_d & x ) {
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Greater"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;

  int i;
  for (i=0;i<Dim();i++ )
    data[i] = (data[i]>xx.data[i]?1.0:0.0);
}

void HCL_RnVector_d::Greater( const HCL_Vector_d & x,
			const HCL_Vector_d & y )
{
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Greater"<<endl;
    cerr<<"first parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"first parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }
  if (Space() != y.Space()) {
    cerr<<"Error: HCL_RnVector_d::Greater"<<endl;
    cerr<<"second parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"second parameter vector:"<<endl;
    y.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  HCL_RnVector_d & yy = (HCL_RnVector_d &)y;

  int i;
  for (i=0;i<Dim();i++ )
    data[i] = (xx.data[i]>yy.data[i]?1.0:0.0);
}

void HCL_RnVector_d::Greater( const HCL_Vector_d & x,
			double t ) {
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Greater"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;

  int i;
  for (i=0;i<Dim();i++ )
    data[i] = (xx.data[i]>t?1.0:0.0);
}

void HCL_RnVector_d::Greater( double t,const HCL_Vector_d & x ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Greater"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;

  int i;
  for (i=0;i<Dim();i++ )
    data[i] = (t>xx.data[i]?1.0:0.0);
}

double HCL_RnVector_d::Sum() const {

  double S = 0.0;
  int i;
   
  for (i=0;i<Dim();i++ )
    S += data[i];
  return S;
}

void HCL_RnVector_d::Add( double a ) {

  int i;

  for (i=0;i<Dim();i++ )
    data[i] += a;
}

void HCL_RnVector_d::Add( const HCL_Vector_d & x, double a ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Add"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  HCL_RnVector_d xx = (HCL_RnVector_d &)x;
  int i;

  for (i=0;i<Dim();i++ )
    data[i] = xx.data[i]+a;
}

void HCL_RnVector_d::Abs() {

  int i;
   
  for (i=0;i<Dim();i++ )
    data[i] = (data[i]>=0?data[i]:-data[i]);
}

void HCL_RnVector_d::Abs( const HCL_Vector_d & x ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Abs"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  int i;
  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
   
  for (i=0;i<Dim();i++ )
    data[i] = (xx.data[i]>=0.0?xx.data[i]:-xx.data[i]);
}

void HCL_RnVector_d::Sign() {

  int i;
   
  for (i=0;i<Dim();i++ )
    data[i] = (data[i]==0.0?0.0:(data[i]>0?1.0:-1.0));
}

void HCL_RnVector_d::Sign( double t ) {
  int i;
   
  for (i=0;i<Dim();i++ )
    data[i] = (data[i]==t?0.0:(data[i]>t?1.0:-1.0));
}

void HCL_RnVector_d::Sign( const HCL_Vector_d & x ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Sign"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  int i;
   
  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;

  for (i=0;i<Dim();i++ )
    data[i] = (xx.data[i]==data[i]?0.0:(data[i]>xx.data[i]?1.0:-1.0));
}

void HCL_RnVector_d::Sign( const HCL_Vector_d & x,
		     const HCL_Vector_d & y ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Sign"<<endl;
    cerr<<"first parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"first parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }
  if (Space() != y.Space()) {
    cerr<<"Error: HCL_RnVector_d::DiagRecipScale"<<endl;
    cerr<<"second parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    y.Write(cerr);
    exit(1);
  }

  int i;
   
  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  HCL_RnVector_d & yy = (HCL_RnVector_d &)y;

  for (i=0;i<Dim();i++ )
    data[i] = (xx.data[i]==yy.data[i]?0.0:(xx.data[i]>yy.data[i]?1.0:-1.0));
}

void HCL_RnVector_d::Sign( const HCL_Vector_d & x,
		     double t )
{
  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Sign"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  int i;
  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
   
  for (i=0;i<Dim();i++ )
    data[i] = (xx.data[i]==t?0.0:(xx.data[i]>t?1.0:-1.0));
}

void HCL_RnVector_d::Sign( double t,const HCL_Vector_d & x ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Sign"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  int i;
  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
   
  for (i=0;i<Dim();i++ )
    data[i] = (xx.data[i]==t?0.0:(xx.data[i]<t?1.0:-1.0));
}

void HCL_RnVector_d::Power( double p ) {

  int i;
  for (i=0;i<Dim();i++ )
    data[i] = pow( data[i],p );
}

void HCL_RnVector_d::Power( double p, const HCL_Vector_d & x ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Power"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  int i;
  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  
  for (i=0; i<Dim(); i++) {
    data[i] = pow( xx.data[i], p );
  }
}

void HCL_RnVector_d::Sqrt() {

  int i;

  for (i=0; i<Dim(); i++) {
    if (data[i] < 0.0) {
      cerr<<"Error: HCL_RnVector_d::Sqrt"<<endl;
      cerr<<"entry "<<i<<" = "<<data[i]<<" negative"<<endl;
      exit(1);
    }
    data[i] = sqrt( data[i] );
  }
}

void HCL_RnVector_d::Sqrt( const HCL_Vector_d & x ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Sqrt"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  int i;
  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  
  for (i=0; i<Dim(); i++) {
    if (xx.data[i] < 0.0) {
      cerr<<"Error: HCL_RnVector_d::Sqrt"<<endl;
      cerr<<"entry "<<i<<" = "<<xx.data[i]<<" negative"<<endl;
      exit(1);
    }
    data[i] = sqrt( xx.data[i] );
  }
}

void HCL_RnVector_d::Exp() {

  int i;
  for (i=0;i<Dim();i++ )
    data[i] = exp( data[i] );
}

void HCL_RnVector_d::Exp( const HCL_Vector_d & x ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Exp"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  int i;
  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  
  for (i=0; i<Dim(); i++) {
    data[i] = exp( xx.data[i] );
  }
}

void HCL_RnVector_d::Log() {

  int i;

  HCL_FPArch_d FP;
  for (i=0; i<Dim(); i++) {
    if (data[i] <= FP.sfmin) {
      cerr<<"Error: HCL_RnVector_d::Log"<<endl;
      cerr<<"entry "<<i<<" = "<<data[i]<<" too small to take log"<<endl;
      exit(1);
    }
    data[i] = log( data[i] );
  }
}

void HCL_RnVector_d::Log( const HCL_Vector_d & x ) {

  if (Space() != x.Space()) {
    cerr<<"Error: HCL_RnVector_d::Log"<<endl;
    cerr<<"parameter HCL_Vector not in same space as object"<<endl;
    cerr<<"This HCL_RnVector_d:"<<endl;
    Write(cerr);
    cerr<<"parameter vector:"<<endl;
    x.Write(cerr);
    exit(1);
  }

  int i;
  HCL_RnVector_d & xx = (HCL_RnVector_d &)x;
  HCL_FPArch_d FP;

  for (i=0; i<Dim(); i++) {
    if (xx.data[i] <= FP.sfmin) {
      cerr<<"Error: HCL_RnVector_d::Log"<<endl;
      cerr<<"entry "<<i<<" = "<<xx.data[i]<<" too small to take log"<<endl;
      exit(1);
    }
    data[i] = log( xx.data[i] );
  }
}

ostream & HCL_RnVector_d::Write( ostream & str ) const {
  int n = Dim();
  str << setprecision(16);
  if( n<=10 )
  {
     str << "HCL_RnSpace_d with dimension " << n << endl;
     str << "Components are" << endl;
     for( int i=0;i<n;i++ ) str << "      " << data[i] << endl;
  }
  else
  {
     str << "HCL_RnSpace_d with dimension " << n << endl;
     str << "First 10 components are" << endl;
     for( int i=0;i<10;i++ ) str << "      " << data[i] << endl;
  }
  return str;
}

