/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_BiLinearOp_d.h"

void HCL_BiLinearOp_d::PartialAdjImage( int i,const HCL_Vector_d & xy,
                                              const HCL_Vector_d & z,
                                                    HCL_Vector_d & yx ) const
{
   cerr << "Error: HCL_BiLinearOp_d::PartialAdjImage is not implemented"
        << endl;
   exit(1);
}

HCL_LinearOp_d * HCL_BiLinearOp_d::Op1( const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain1() )
   {
      cerr << "Error in HCL_BiLinearOp_d::Op1: input vector is not in "
              "domain 1" << endl;
      cerr << "Domain 1:" << endl;
      Domain1().Write( cerr );
      cerr << "Space of input vector:" << endl;
      x.Space().Write( cerr );
      exit(1);
   }
   return new HCL_RestrictedBiLinearOp_d( this,1,x );
}

HCL_LinearOp_d * HCL_BiLinearOp_d::Op2( const HCL_Vector_d & y ) const
{
   if( y.Space() != Domain2() )
   {
      cerr << "Error in HCL_BiLinearOp_d::Op2: input vector is not in "
              "domain 2" << endl;
      cerr << "Domain 2:" << endl;
      Domain2().Write( cerr );
      cerr << "Space of input vector:" << endl;
      y.Space().Write( cerr );
      exit(1);
   }
   return new HCL_RestrictedBiLinearOp_d( this,2,y );
}

HCL_LinearOp_d * HCL_BiLinearOp_d::Op3( const HCL_Vector_d & z ) const
{
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_BiLinearOp_d::Op3: input vector is not in "
              "range" << endl;
      cerr << "Range:" << endl;
      Range().Write( cerr );
      cerr << "Space of input vector:" << endl;
      z.Space().Write( cerr );
      exit(1);
   }
   return new HCL_RestrictedBiLinearOp_d( this,3,z );
}

int HCL_BiLinearOp_d::CheckAdj( int Display,double tol )
{
   int err = 0;

   if( Display )
      cout << "HCL_BiLinearOp_d::CheckAdj: Checking Op1" << endl;
   HCL_Vector_d * x = Domain1().Member();
   x->Random();
   HCL_LinearOp_d  * op1 = Op1( *x );
   err += op1->CheckAdj( Display,tol );
   HCL_delete( op1 );
   HCL_delete( x );

   if( Display )
      cout << "HCL_BiLinearOp_d::CheckAdj: Checking Op2" << endl;
   HCL_Vector_d * y = Domain2().Member();
   y->Random();
   HCL_LinearOp_d  * op2 = Op2( *y );
   err += op2->CheckAdj( Display,tol );
   HCL_delete( op2 );
   HCL_delete( y );

   if( Display )
      cout << "HCL_BiLinearOp_d::CheckAdj: Checking Op3" << endl;
   HCL_Vector_d * z = Range().Member();
   z->Random();
   HCL_LinearOp_d  * op3 = Op3( *z );
   err += op3->CheckAdj( Display,tol );
   HCL_delete( op3 );
   HCL_delete( z );

   return err;
}

HCL_RestrictedBiLinearOp_d::HCL_RestrictedBiLinearOp_d()
{
   cerr << "Error: HCL_RestrictedBiLinearOp_d default constructor not "
           "implemented" << endl;
   exit(1);
}

HCL_RestrictedBiLinearOp_d::HCL_RestrictedBiLinearOp_d(
                            const HCL_RestrictedBiLinearOp_d & )
{
   cerr << "Error: HCL_RestrictedBiLinearOp_d copy constructor not "
           "implemented" << endl;
   exit(1);
}

HCL_RestrictedBiLinearOp_d::HCL_RestrictedBiLinearOp_d( 
                            const HCL_BiLinearOp_d * t,
                            int of,const HCL_Vector_d & v )
{
   Bop = t;t->IncCount();
   if( of<1 || of>3 )
   {
      cerr << "Error in HCL_RestrictedBiLinearOp_d constructor: "
              "invalid operator flag (must be 1, 2, or 3); " << endl;
      cerr << "input was " << of << endl;
      exit(1);
   }
   OpFlag = of;
   fixed = v.Space().Member();
   fixed->Copy( v );
}

HCL_RestrictedBiLinearOp_d::~HCL_RestrictedBiLinearOp_d()
{
   HCL_delete( fixed );
   HCL_delete( (HCL_BiLinearOp_d*)Bop );
}

HCL_VectorSpace_d & HCL_RestrictedBiLinearOp_d::Domain() const
{
   switch( OpFlag )
   {
      case 1:
         return Bop->Domain2();
         break;
      case 2:
         return Bop->Domain1();
         break;
      case 3:
         return Bop->Domain1();
         break;
   }
}

HCL_VectorSpace_d & HCL_RestrictedBiLinearOp_d::Range() const
{
   switch( OpFlag )
   {
      case 1:
         return Bop->Range();
         break;
      case 2:
         return Bop->Range();
         break;
      case 3:
         return Bop->Domain2();
         break;
   }
}

void HCL_RestrictedBiLinearOp_d::Image( const HCL_Vector_d & x,
                                              HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_RestrictedBiLinearOp_d::Image: input vector "
              "is not in domain" << endl;
      cerr << "Domain:" << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector:" << endl;
      x.Space().Write( cerr );
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_RestrictedBiLinearOp_d::Image: output vector "
              "is not in range" << endl;
      cerr << "Range:" << endl;
      Range().Write( cerr );
      cerr << "Space of output vector:" << endl;
      y.Space().Write( cerr );
      exit(1);
   }

   switch( OpFlag )
   {
      case 1:
         Bop->Image( *fixed,x,y );
         break;
      case 2:
         Bop->Image( x,*fixed,y );
         break;
      case 3:
         Bop->PartialAdjImage( 1,x,*fixed,y );
         break;
   }
}
   
void HCL_RestrictedBiLinearOp_d::AdjImage( const HCL_Vector_d & y,
                                                 HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_RestrictedBiLinearOp_d::AdjImage: output vector "
              "is not in domain" << endl;
      cerr << "Domain:" << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector:" << endl;
      x.Space().Write( cerr );
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_RestrictedBiLinearOp_d::AdjImage: input vector "
              "is not in range" << endl;
      cerr << "Range:" << endl;
      Range().Write( cerr );
      cerr << "Space of input vector:" << endl;
      y.Space().Write( cerr );
      exit(1);
   }

   switch( OpFlag )
   {
      case 1:
         Bop->PartialAdjImage( 1,*fixed,y,x );
         break;
      case 2:
         Bop->PartialAdjImage( 2,*fixed,y,x );
         break;
      case 3:
         Bop->PartialAdjImage( 2,y,*fixed,x );
         break;
   }
}
   
void HCL_RestrictedBiLinearOp_d::InvImage(const HCL_Vector_d & y,
                                                HCL_Vector_d & x) const
{
   cerr << "Error: HCL_RestrictedBiLinearOp_d::InvImage is not implemented"
        << endl;
   exit(1);
}

void HCL_RestrictedBiLinearOp_d::InvAdjImage(const HCL_Vector_d & x,
                                                   HCL_Vector_d & y) const
{
   cerr << "Error: HCL_RestrictedBiLinearOp_d::InvAdjImage is not implemented"
        << endl;
   exit(1);
}

ostream & HCL_RestrictedBiLinearOp_d::Write( ostream & str ) const
{
   str << "HCL_RestrictedBiLinearOp_d (Op" << OpFlag << ") based on "
          "the bilinear operator: " <<endl;
   Bop->Write( str );
   str << endl << "Domain:" << endl;
   Domain().Write( str );
   str << "Range:" << endl;
   Range().Write( str );
   return str;
}
