/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_Functional_d.h"

HCL_EvaluateFunctional_d::HCL_EvaluateFunctional_d()
{
   valFlag = 0;
   SavedGradient = NULL;
   SavedHessian = NULL;
}

HCL_EvaluateFunctional_d::HCL_EvaluateFunctional_d(
                        const HCL_EvaluateFunctional_d & x )
{
   valFlag = x.valFlag;
   SavedValue = x.SavedValue;
   SavedGradient = x.SavedGradient;
   if( SavedGradient != NULL ) SavedGradient->IncCount();
   SavedHessian = x.SavedHessian;
   if( SavedHessian != NULL ) SavedHessian->IncCount();
}

HCL_EvaluateFunctional_d::~HCL_EvaluateFunctional_d()
{
   HCL_delete( SavedHessian );
   HCL_delete( SavedGradient );
}

void HCL_EvaluateFunctional_d::HessianImage( const HCL_Vector_d & dx,
                                             HCL_Vector_d & dy ) const
{
   cerr << "Error: HCL_EvaluateFunctional_d::HessianImage is not "
           "implemented" << endl;
   exit(1);
}

void HCL_EvaluateFunctional_d::HessianInvImage( const HCL_Vector_d & dy,
                                                HCL_Vector_d & dx ) const
{
   cerr << "Error: HCL_EvaluateFunctional_d::HessianInvImage is not "
           "implemented" << endl;
   exit(1);
}

double HCL_EvaluateFunctional_d::ValueRef() const
{
   if( !valFlag )
   {
      // Casting away const
      HCL_EvaluateFunctional_d * t = (HCL_EvaluateFunctional_d*)this;
      t->SavedValue = Value();
      t->valFlag = 1;
   }
   return SavedValue;
}

const HCL_Vector_d & HCL_EvaluateFunctional_d::GradientRef() const
{
   if( !SavedGradient )
   {
      // Casting away const
      HCL_EvaluateFunctional_d * t = (HCL_EvaluateFunctional_d*)this;
      t->SavedGradient = Domain().Member();
      Gradient( *(t->SavedGradient) );
   }
   return *SavedGradient;
}

HCL_LinearOp_d * HCL_EvaluateFunctional_d::Hessian() const
{
   return new HCL_FunctionalDefaultHessian_d( this );
}


const HCL_LinearOp_d & HCL_EvaluateFunctional_d::HessianRef() const
{
   if( SavedHessian == NULL )
   {
      // Casting away const
      HCL_EvaluateFunctional_d * t = (HCL_EvaluateFunctional_d*)this;
      t->SavedHessian = Hessian();
   }
   return *SavedHessian;
}
double HCL_Functional_d::Value1( const HCL_Vector_d & x ) const
{
   cerr << "Error: HCL_Functional_d::Value1 is not implemented" << endl;
   exit(1);
   return 0;
}

void HCL_Functional_d::Gradient1( const HCL_Vector_d & x,
                                        HCL_Vector_d & g ) const
{
   cerr << "Error: HCL_Functional_d::Gradient1 is not implemented" << endl;
   exit(1);
}

void HCL_Functional_d::HessianImage( const HCL_Vector_d & x,
                                     const HCL_Vector_d & dx,
                                           HCL_Vector_d & dy ) const
{
   cerr << "Error: HCL_Functional_d::HessianImage is not implemented" << endl;
   exit(1);
}

void HCL_Functional_d::HessianInvImage( const HCL_Vector_d & x,
                                        const HCL_Vector_d & dy,
                                              HCL_Vector_d & dx ) const
{
   cerr << "Error: HCL_Functional_d::HessianInvImage is not implemented"
        << endl;
   exit(1);
}

HCL_LinearOp_d * HCL_Functional_d::Hessian1( const HCL_Vector_d & x ) const
{
   return (HCL_LinearOp_d*)NULL;
}

HCL_EvaluateFunctional_d * HCL_Functional_d::Evaluate(
                           const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_Functional_d::Evaluate: input vector "
              "is not in domain" << endl;
      cerr << "Domain:" << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector:" << endl;
      x.Space().Write( cerr );
      exit(1);
   }
   return new HCL_FunctionalDefaultEval_d( this,x );
}

double HCL_Functional_d::Value( const HCL_Vector_d & x ) const
{
   HCL_EvaluateFunctional_d * eval = Evaluate( x );
   double v = eval->Value();
   HCL_delete( eval );
   return v;
}

void HCL_Functional_d::Gradient( const HCL_Vector_d & x,
                                       HCL_Vector_d & g ) const
{
   HCL_EvaluateFunctional_d * eval = Evaluate( x );
   eval->Gradient( g );
   HCL_delete( eval );
}

HCL_LinearOp_d * HCL_Functional_d::Hessian( const HCL_Vector_d & x ) const
{
   HCL_EvaluateFunctional_d * eval = Evaluate( x );
   HCL_LinearOp_d * H = eval->Hessian();
   HCL_delete( eval );
   return H;
}

int HCL_Functional_d::CheckGrad( const HCL_Vector_d & y,
                                 const HCL_Vector_d & p,ostream & str,int n,
                                 double hmin,double hmax )
{
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_Functional_d::CheckGrad: "
               "vector y is not in Domain" << endl;
      exit(1);
   }
   if( p.Space() != Domain() )
   {
      cerr << "Error in HCL_Functional_d::CheckGrad: "
              "vector p is not in Domain" << endl;
      exit(1);
   }

   if( hmax <= hmin )
   {
      double temp = hmax;
      hmax = hmin;
      hmin = temp;
   }
   if( hmin <= 0.0 )
   {
      hmin = 0.1;
      hmax = 1.0;
   }
   if( n <= 0 )
      n = 10;

   double hlimit1 = MaxStep( y,p );
   HCL_Vector_d * ptemp = p.Space().Member();
   ptemp->Mul( -1.0,p );
   double hlimit2 = MaxStep( y,*ptemp );
   HCL_delete( ptemp );
   double hlimit = min( hlimit1,hlimit2 );
   if( hlimit <= 0.0 )
   {
      cerr << "Error in HCL_Functional_d::CheckGrad: direction is not "
              "feasible" << endl;
      exit(1);
   }
   if( hmax >= hlimit )
   {
      hmax = 0.99*hlimit;
      if( hmin >= hmax )
         hmin = hmax/n;
   }

   HCL_Vector_d * x = Domain().Member();
   HCL_Vector_d * g = Domain().Member();
   Gradient( y,*g );

#if SINGLE_PRECISION
   int nd = 8;
#else
   int nd = 16;
#endif
   int oldprecision = str.precision( nd );

   double d2 = g->Inner( p );
   double ad2 = abs( d2 );
   int rflag = 1;
   HCL_FPArch_d arch;
   if( ad2 < arch.sfmin )
   {
      rflag = 0;
      str << "HCL_FunctionalHess_d::CheckGrad: norm of gradient is too "
          << endl << "small; displaying absolute error" << endl;
   }

   int i;
   double hstep = (hmax-hmin)/(n-1);
   if( rflag )
      str << setw(5) << "h" << setw(nd+1) << "slope" << setw(nd+11)
           << "cent. diff." << setw(nd+4) << "Rel. Err." << endl;
   else
      str << setw(5) << "h" << setw(nd+1) << "slope" << setw(nd+11)
           << "cent. diff." << setw(nd+4) << "Error" << endl;

   for( i=n-1;i>=0;i-- )
   {
      double h = hmin+i*hstep;
      x->Copy( y );
      x->AddScale( -h,p );
      double f1 = Value( *x );
      x->Copy( y );
      x->AddScale( h,p );
      double f2 = Value( *x );
      double d1 = (f2-f1)/(2.0*h);
      if( rflag )
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << d2 << " " << setw(nd+6)
             << d1 << " " << setw(nd+6) << abs( d1-d2 )/ad2 << endl;
      else
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << d2 << " " << setw(nd+6)
             << d1 << " " << setw(nd+6) << abs( d1-d2 ) << endl;
   }
   HCL_delete( g );
   HCL_delete( x );
   str.precision( oldprecision );
   return 0;
}

int HCL_Functional_d::CheckHess( const HCL_Vector_d & y,
                                 const HCL_Vector_d & p,ostream & str,
                                 int n,double hmin,double hmax )
{
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_Functional_d::CheckHess: "
               "vector y is not in Domain" << endl;
      exit(1);
   }
   if( p.Space() != Domain() )
   {
      cerr << "Error in HCL_Functional_d::CheckHess: "
              "vector p is not in Domain" << endl;
      exit(1);
   }

   if( hmax <= hmin )
   {
      double temp = hmax;
      hmax = hmin;
      hmin = temp;
   }
   if( hmin <= 0.0 )
   {
      hmin = 0.1;
      hmax = 1.0;
   }
   if( n <= 0 )
      n = 10;

   double hlimit1 = MaxStep( y,p );
   HCL_Vector_d * ptemp = p.Space().Member();
   ptemp->Mul( -1.0,p );
   double hlimit2 = MaxStep( y,*ptemp );
   HCL_delete( ptemp );
   double hlimit = min( hlimit1,hlimit2 );
   if( hlimit <= 0.0 )
   {
      cerr << "Error in HCL_Functional_d::CheckHess: direction is not "
              "feasible" << endl;
      exit(1);
   }
   if( hmax >= hlimit )
   {
      hmax = 0.99*hlimit;
      if( hmin >= hmax )
         hmin = hmax/n;
   }

   HCL_Vector_d * x1 = Domain().Member();
   HCL_Vector_d * x2 = Domain().Member();
   HCL_Vector_d * g1 = Domain().Member();
   HCL_Vector_d * g2 = Domain().Member();
   HCL_Vector_d * dg = Domain().Member();

   HCL_EvaluateFunctional_d * eval = Evaluate( y );
   HCL_LinearOp_d * hess = eval->Hessian();
   eval->Gradient( *dg );
   hess->Image( p,*dg );
   HCL_delete( hess );
   HCL_delete( eval );

#if SINGLE_PRECISION
   int nd = 8;
#else
   int nd = 16;
#endif
   int oldprecision = str.precision( nd );

   double dgnorm = dg->Norm();
   int rflag = 1;
   HCL_FPArch_d arch;
   if( dgnorm < arch.sfmin )
   {
      rflag = 0;
      str << "HCL_Functional_d::CheckHess: norm of gradient is too "
          << endl << "small; displaying absolute error" << endl;
   }

   if( rflag )
      str << setw(5) << "h" << setw(nd+5) << "norm of diff." << setw(nd+6)
           << "norm" << setw(nd+6) << "Rel. Err." << endl;
   else
      str << setw(5) << "h" << setw(nd+5) << "norm of diff." << setw(nd+6)
           << "norm" << endl;
   int i;
   double hstep = (hmax-hmin)/(n-1);
   for( i=n-1;i>=0;i-- )
   {
      double h = hmin+i*hstep;
      x1->AddScale( -h,y,p );
      Gradient( *x1,*g1 );
      x2->AddScale( h,y,p );
      Gradient( *x2,*g2 );
      g2->Sub( *g1 );
      g2->Mul( 1.0/(2.0*h) );
      g2->Sub( *dg );
      double n1 = g2->Norm();
      if( rflag )
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << n1 << " " 
             << setw(nd+6) << dgnorm << " " << setw(nd+6) << n1/dgnorm << endl;
      else
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << n1 << " " 
             << setw(nd+6) << dgnorm << " " << endl;
   }
   HCL_delete( dg );
   HCL_delete( g2 );
   HCL_delete( g1 );
   HCL_delete( x2 );
   HCL_delete( x1 );
   str.precision( oldprecision );
   return 0;
}

void HCL_Functional_d::Scan( const HCL_Vector_d & x,const HCL_Vector_d & dx,
                             int N, double hmin, double hmax,
                             char * fname)
{
   if (x.Space() != Domain())
   {
      cerr<<"Error: HCL_Functional_d::Scan"<<endl;
      cerr<<"1st arg not member of domain"<<endl;
      cerr<<"Domain space:"<<endl;
      Domain().Write(cerr);
      cerr<<"input vector space:"<<endl;
      (x.Space()).Write(cerr);
      exit(1);
   }

   if (dx.Space() != Domain())
   {
      cerr<<"Error: HCL_Functional_d::Scan"<<endl;
      cerr<<"2nd arg not member of domain"<<endl;
      cerr<<"Domain space:"<<endl;
      Domain().Write(cerr);
      cerr<<"input vector space:"<<endl;
      (dx.Space()).Write(cerr);
      exit(1);
   }

   if (N <= 0)
   {
      cerr<<"Error: HCL_Functional_d::Scan"<<endl;
      cerr<<"Requested number of scan points = "<<N<<" not positive"<<endl;
      exit(1);
   }

    ostream * str;
    if( fname )
    {
       str = new ofstream(fname);
// The following command commented out because it won't compile
// under CC (SC5.0)
//       (*str)<<setiosflags(ios::showpoint);
    }
    else
       str = &cout;

#if SINGLE_PRECISION
   int nd = 8;
#else
   int nd = 16;
#endif
    int oldprecision = str->precision( nd );

    if( hmax <= hmin )
    {
       double temp = hmax;
       hmax = hmin;
       hmin = temp;
    }
    if( hmax > 0.0 )
    {
       double hlimit = MaxStep( x,dx );
       if( hlimit <= 0.0 )
       {
          cerr << "Error in HCL_Functional_d::Scan: direction is not "
                  "feasible" << endl;
          exit(1);
       }       
       if( hmax >= hlimit )
       {
          hmax = 0.99*hlimit;
          if( hmin >= hmax )
             hmin = hmax/N;
       }
    }
    if( hmin < 0.0 )
    {
       HCL_Vector_d * ptemp = dx.Space().Member();
       ptemp->Mul( -1.0,dx );
       double hlimit = MaxStep( x,*ptemp );
       HCL_delete( ptemp );
       if( hlimit <= 0.0 )
       {
          cerr << "Error in HCL_Functional_d::Scan: direction is not "
                  "feasible" << endl;
          exit(1);
       }       
       if( hmin <= -hlimit )
       {
          hmin = -0.99*hlimit;
          if( hmin >= hmax )
             hmax = hmin/N;
       }
    }

    double dh;
    if (N == 1)
      dh=1.0;
    else
      dh=(hmax-hmin)/(N-1);
    double h = hmin;
    double v = 0.0;
    int i;

    HCL_Vector_d * z = Domain().Member();

    (*str)<<endl;
    (*str)<<"/////////////////////////////////////////////////////"<<endl;
    (*str)<<"//////////// HCL_Functional_d: SCAN RECORD ////////////"<<endl;
    (*str)<<"/////////////////////////////////////////////////////"<<endl;
    (*str)<<endl;

    for (i=0;i<N;i++) {
      h = hmin + i*dh;
      z->AddScale(h,x,dx);
      v = Value(*z);
      (*str) << setprecision(6) << setw(8) << h
             << "      " << setprecision(nd) << v << endl;
    }
    HCL_delete( z );
    if( fname ) delete str;
    str->precision( oldprecision );
}

double HCL_Functional_d::MaxStep( const HCL_Vector_d & x,
                                  const HCL_Vector_d & dir) const
{
   HCL_FPArch_d arch;
   return arch.large;
}

HCL_FunctionalDefaultEval_d::HCL_FunctionalDefaultEval_d()
{
   cerr << "Error: HCL_FunctionalDefaultEval_d default constructor "
           "not implemented" << endl;
   exit(1);
}

HCL_FunctionalDefaultEval_d::HCL_FunctionalDefaultEval_d(
                             const HCL_FunctionalDefaultEval_d & )
{
   cerr << "Error: HCL_FunctionalDefaultEval_d copy constructor "
           "not implemented" << endl;
   exit(1);
}

HCL_FunctionalDefaultEval_d::HCL_FunctionalDefaultEval_d(
                             const HCL_Functional_d * ff,
                             const HCL_Vector_d & in )
{
   f = ff;f->IncCount();
   x = in.Space().Member();
   x->Copy( in );
}

HCL_FunctionalDefaultEval_d::~HCL_FunctionalDefaultEval_d()
{
   HCL_delete( x );
   HCL_delete( (HCL_Functional_d*)f );
}

double HCL_FunctionalDefaultEval_d::Value() const
{
   return f->Value1( *x );
}

void HCL_FunctionalDefaultEval_d::Gradient( HCL_Vector_d & y ) const
{
   f->Gradient1( *x,y );
}

HCL_LinearOp_d * HCL_FunctionalDefaultEval_d::Hessian() const
{
   HCL_LinearOp_d * H = f->Hessian1( *x );
   if( H )
      return H;
   else
      return new HCL_FunctionalDefaultHessian_d( this );
}

ostream & HCL_FunctionalDefaultEval_d::Write( ostream & str ) const
{
   str << "HCL_FunctionalDefaultEval_d based on the following "
          "functional" << endl;
   return f->Write( str );
}

HCL_FunctionalDefaultHessian_d::HCL_FunctionalDefaultHessian_d()
{
   cerr << "Error: HCL_FunctionalDefaultHessian_d default "
           "constructor not implemented"
        << endl;
   exit(1);
}

HCL_FunctionalDefaultHessian_d::HCL_FunctionalDefaultHessian_d(
              const HCL_FunctionalDefaultHessian_d & L )
{
   cerr << "Error: HCL_FunctionalDefaultHessian_d copy "
           "constructor not implemented"
        << endl;
   exit(1);
}

HCL_FunctionalDefaultHessian_d::HCL_FunctionalDefaultHessian_d(
                     const HCL_EvaluateFunctional_d * ev )
{
   eval = ev;ev->IncCount();
}

HCL_FunctionalDefaultHessian_d::~HCL_FunctionalDefaultHessian_d()
{
   HCL_delete( (HCL_EvaluateFunctional_d*)eval );
}

HCL_VectorSpace_d & HCL_FunctionalDefaultHessian_d::Domain() const
{
   return eval->Domain();
}

HCL_VectorSpace_d & HCL_FunctionalDefaultHessian_d::Range() const
{
   return eval->Domain();
}

void HCL_FunctionalDefaultHessian_d::Image( const HCL_Vector_d & x,
                                                  HCL_Vector_d & y ) const
{
   eval->HessianImage( x,y );
}

void HCL_FunctionalDefaultHessian_d::AdjImage( const HCL_Vector_d & y,
                                                     HCL_Vector_d & x ) const
{
   eval->HessianImage( y,x );
}

void HCL_FunctionalDefaultHessian_d::InvImage(const HCL_Vector_d & y,
                                                    HCL_Vector_d & x) const
{
   eval->HessianInvImage( y,x );
}

void HCL_FunctionalDefaultHessian_d::InvAdjImage(const HCL_Vector_d & x,
                                                 HCL_Vector_d & y) const
{
   eval->HessianInvImage( x,y );
}

ostream & HCL_FunctionalDefaultHessian_d::Write( ostream & str ) const
{
   str << "HCL_FunctionalDefaultHessian_d based on the following "
          "evaluation object"
       << endl;
   return eval->Write( str );
}
