#include "HCL_Arch_d.h"
#include "HCL_LinearOp_d.h"

int HCL_LinearOp_d::CheckAdj( int Display,double t )
{
   HCL_FPArch_d arch;
   double Inner1, Inner2;
   double tol = t * arch.eps;
   HCL_Vector_d *x = Domain().Member();
   HCL_Vector_d *y = Range().Member();
   HCL_Vector_d *Ax = Range().Member();
   HCL_Vector_d *Aty = Domain().Member();
   x->Random();
   y->Random();
   Ax->Random();
   Aty->Random();

   Image(*x,*Ax);
   Inner2 = y->Inner(*Ax);
   AdjImage(*y,*Aty);
   Inner1 = x->Inner(*Aty);
   HCL_delete( Aty );
   HCL_delete( Ax );
   HCL_delete( y );
   HCL_delete( x );
   
   if (fabs(Inner1-Inner2) > (fabs(Inner1) * tol))
   {
      if (Display)
      cout << "Error: (Av, w) differs from (v, A'w): " << endl
           << "   (Av, w)-(v,A'w) = " << Inner2-Inner1 << endl
           << "   (Av, w) = " << Inner2 << endl;
      return 1;
   }
   else
   {
      if (Display > 1)
         cout << "(Av, w) agrees with (v, A'w): " << endl
              << "   (Av, w) = (v, A'w) = " << Inner1 << endl;
      return 0;
   }
}

void HCL_LinearOp_d::NormalImage( const HCL_Vector_d & x,
                                        HCL_Vector_d & z ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_LinearOp_d::NormalImage: vector is not in domain"
           << endl;
      exit(1);
   }
   if( z.Space() != Domain() )
   {
      cerr << "Error in HCL_LinearOp_d::NormalImage: vector is not in range "
              "of normal operator (i.e. domain of operator)" << endl;
      exit(1);
   }
   HCL_Vector_d * y = Range().Member();
   Image( x,*y );
   AdjImage( *y,z );
   HCL_delete( y );
}

int HCL_LinearOp_d::CheckNormal( int Display,double t )
{
   HCL_FPArch_d arch;
   double tol = t*arch.eps;
   double Norm, ndiff;
   HCL_Vector_d *x = Domain().Member();
   HCL_Vector_d *Ax = Range().Member();
   HCL_Vector_d *AtAx = Domain().Member();
   HCL_Vector_d *AtAx1 = Domain().Member();

   x->Random();

   Image(*x, *Ax);
   AdjImage(*Ax,*AtAx);
   NormalImage(*x, *AtAx1);

   AtAx1->Sub(*AtAx);

   ndiff = AtAx1->Norm();
   Norm = AtAx->Norm();

   HCL_delete( AtAx1 );
   HCL_delete( AtAx );
   HCL_delete( Ax );
   HCL_delete( x );

   if (ndiff > Norm*tol )
   {
      if(Display)
         cout << "Error: (A'A)x differs from A'(Ax)." << endl
              << "   ||(A'A)x-A'(Ax)|| = " << ndiff << endl
              << "   ||A'(Ax)|| = " << Norm << endl;
       return 1;
   }
   else
   {
      if(Display > 1)
         cout << "(A'A)x agrees with A'(Ax)." << endl
              << "   ||(A'A)x-A'(Ax)|| = " << ndiff << endl
              << "   ||A'(Ax)|| = " << Norm << endl;
      return 0;
   }
}

void HCL_LinearOp_d::AdjImage( const HCL_Vector_d & y,
                                     HCL_Vector_d & x ) const
{
   cerr << "Error: HCL_LinearOp_d::AdjImage is not implemented" << endl;
   exit(1);
}

void HCL_LinearOp_d::InvImage(const HCL_Vector_d & y,
                                    HCL_Vector_d & x) const
{
   cerr << "Error: HCL_LinearOp_d::InvImage is not implemented" << endl;
   exit(1);
}

void HCL_LinearOp_d::InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const
{
   cerr << "Error: HCL_LinearOp_d::InvAdjImage is not implemented" << endl;
   exit(1);
}

void HCL_LinearOp_d::ImageAdd(const HCL_Vector_d & x,
			      const HCL_Vector_d & z,
			      HCL_Vector_d & y,
			      double a,
			      double b) const {

  if( x.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::ImageAdd: first input vector is not "
      "in domain" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( z.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::ImageAdd: second input vector is not "
      "in range" << endl;
    cerr << "Space of input vector:" << endl;
    z.Space().Write( cerr );
    cerr << "Range:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  if( y.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::ImageAdd: output vector is not "
      "in range" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Range:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  Image(x,y);
  if (fabs(a-1.0) > Arch.eps) {
    y.Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.AddScale(b,z);
  }
  else {
    y.Add(z);
  }
}

void HCL_LinearOp_d::AdjImageAdd(const HCL_Vector_d & x,
				 const HCL_Vector_d & z,
				 HCL_Vector_d & y,
				 double a,
				 double b) const {

  if( x.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::AdjImageAdd: first input vector is not "
      "in range" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Range:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  if( z.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::AdjImageAdd: second input vector is not "
      "in domain" << endl;
    cerr << "Space of input vector:" << endl;
    z.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( y.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::AdjImageAdd: output vector is not "
      "in domain" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  AdjImage(x,y);
  if (fabs(a-1.0) > Arch.eps) {
    y.Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.AddScale(b,z);
  }
  else {
    y.Add(z);
  }
}

void HCL_LinearOp_d::NormalImageAdd(const HCL_Vector_d & x,
				    const HCL_Vector_d & z,
				    HCL_Vector_d & y,
				    double a,
				    double b) const {

  if( x.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::NormalImageAdd: first input "
      "vector is not in domain" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( z.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::NormalImageAdd: second input "
      "vector is not in domain" << endl;
    cerr << "Space of input vector:" << endl;
    z.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( y.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::NormalImageAdd: output vector is not "
      "in domain" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  NormalImage(x,y);
  if (fabs(a-1.0) > Arch.eps) {
    y.Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.AddScale(b,z);
  }
  else {
    y.Add(z);
  }
}

void HCL_LinearOp_d::InvImageAdd(const HCL_Vector_d & x,
				 const HCL_Vector_d & z,
				 HCL_Vector_d & y,
				 double a,
				 double b) const {

  if( x.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::InvImageAdd: first input vector is not "
      "in range" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Range:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  if( z.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::InvImageAdd: second input vector is not "
      "in domain" << endl;
    cerr << "Space of input vector:" << endl;
    z.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( y.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::InvImageAdd: output vector is not "
      "in domain" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  InvImage(x,y);
  if (fabs(a-1.0) > Arch.eps) {
    y.Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.AddScale(b,z);
  }
  else {
    y.Add(z);
  }
}

void HCL_LinearOp_d::InvAdjImageAdd(const HCL_Vector_d & x,
				    const HCL_Vector_d & z,
				    HCL_Vector_d & y,
				    double a,
				    double b) const {

  if( x.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::InvAdjImageAdd: first input "
      "vector is not in domain" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( z.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::InvAdjImageAdd: second input "
      "vector is not in range" << endl;
    cerr << "Space of input vector:" << endl;
    z.Space().Write( cerr );
    cerr << "Range:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  if( y.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::InvAdjImageAdd: output vector is not "
      "in range" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Range:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  InvAdjImage(x,y);
  if (fabs(a-1.0) > Arch.eps) {
    y.Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.AddScale(b,z);
  }
  else {
    y.Add(z);
  }
}

void HCL_LinearOp_d::ImageAdd(const HCL_Vector_d & x,
			      HCL_Vector_d & y,
			      double a,
			      double b) const {

  if( x.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::ImageAdd: input vector is not "
      "in domain" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( y.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::ImageAdd: output vector is not "
      "in range" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Range:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  HCL_Vector_d * tmp = y.Space().Member();
  Image(x,*tmp);
  if (fabs(a-1.0) > Arch.eps) {
    tmp->Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.ScaleAdd( b,*tmp);
  }
  else {
    y.Add(*tmp);
  }
  HCL_delete( tmp );
}

void HCL_LinearOp_d::AdjImageAdd(const HCL_Vector_d & x,
				 HCL_Vector_d & y,
				 double a,
				 double b) const {

  if( x.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::AdjImageAdd: first input "
      "vector is not in range" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( y.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::AdjImageAdd: output vector is not "
      "in domain" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  HCL_Vector_d * tmp= y.Space().Member();
  AdjImage(x,*tmp);
  if (fabs(a-1.0) > Arch.eps) {
    tmp->Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.ScaleAdd(b,*tmp);
  }
  else {
    y.Add(*tmp);
  }
  HCL_delete( tmp );
}

void HCL_LinearOp_d::NormalImageAdd(const HCL_Vector_d & x,
				    HCL_Vector_d & y,
				    double a,
				    double b) const {

  if( x.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::NormalImageAdd: first input "
      "vector is not in domain" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( y.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::NormalImageAdd: output vector is not "
      "in domain" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  HCL_Vector_d * tmp = y.Space().Member();
  NormalImage(x,*tmp);
  if (fabs(a-1.0) > Arch.eps) {
    tmp->Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.ScaleAdd(b,*tmp);
  }
  else {
    y.Add(*tmp);
  }
  HCL_delete(tmp);
}

void HCL_LinearOp_d::InvImageAdd(const HCL_Vector_d & x,
				 HCL_Vector_d & y,
				 double a,
				 double b) const {

  if( x.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::InvImageAdd: first input vector is not "
      "in range" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Range:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  if( y.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::InvImageAdd: output vector is not "
      "in domain" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  HCL_Vector_d * tmp = y.Space().Member();
  InvImage(x,*tmp);
  if (fabs(a-1.0) > Arch.eps) {
    tmp->Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.ScaleAdd(b,*tmp);
  }
  else {
    y.Add(*tmp);
  }
  HCL_delete(tmp);
}

void HCL_LinearOp_d::InvAdjImageAdd(const HCL_Vector_d & x,
				    HCL_Vector_d & y,
				    double a,
				    double b) const {

  if( x.Space() != Domain() ) {
    cerr << "Error in HCL_LinearOp_d::InvAdjImageAdd: first input "
      "vector is not in domain" << endl;
    cerr << "Space of input vector:" << endl;
    x.Space().Write( cerr );
    cerr << "Domain:" << endl;
    Domain().Write( cerr );
    exit(1);
  }
  if( y.Space() != Range() ) {
    cerr << "Error in HCL_LinearOp_d::InvAdjImageAdd: output vector is not "
      "in range" << endl;
    cerr << "Space of output vector:" << endl;
    y.Space().Write( cerr );
    cerr << "Range:" << endl;
    Range().Write( cerr );
    exit(1);
  }
  HCL_FPArch_d Arch;
  HCL_Vector_d * tmp = y.Space().Member();
  InvAdjImage(x,*tmp);
  if (fabs(a-1.0) > Arch.eps) {
    tmp->Mul(a);
  }  
  if (fabs(b-1.0) > Arch.eps) {
    y.ScaleAdd(b,*tmp);
  }
  else {
    y.Add(*tmp);
  }
  HCL_delete(tmp);
}


