/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_Op_d.h"

HCL_EvaluateOp_d::HCL_EvaluateOp_d()
{
   SavedImage = NULL;
   SavedDeriv = NULL;
   Saved2ndDeriv = NULL;
}

HCL_EvaluateOp_d::HCL_EvaluateOp_d(
             const HCL_EvaluateOp_d & e )
{
   if( e.SavedImage )
   {
      SavedImage = e.SavedImage;
      SavedImage->IncCount();
   }
   if( e.SavedDeriv )
   {
      SavedDeriv = e.SavedDeriv;
      SavedDeriv->IncCount();
   }
   if( e.Saved2ndDeriv )
   {
      Saved2ndDeriv = e.Saved2ndDeriv;
      Saved2ndDeriv->IncCount();
   }
}

HCL_EvaluateOp_d::~HCL_EvaluateOp_d()
{
   HCL_delete( Saved2ndDeriv );
   HCL_delete( SavedDeriv );
   HCL_delete( SavedImage );
}

const HCL_Vector_d & HCL_EvaluateOp_d::ImageRef() const
{
   HCL_EvaluateOp_d * t = (HCL_EvaluateOp_d*)this;
   if( SavedImage == NULL )
   {
      t->SavedImage = Range().Member();
      Image( *(t->SavedImage) );
   }
   return *SavedImage;
}

HCL_LinearOp_d * HCL_EvaluateOp_d::Deriv() const
{
   return new HCL_OpDefaultDeriv_d( this );
}

HCL_BiLinearOp_d * HCL_EvaluateOp_d::SecondDeriv() const
{
   return new HCL_OpDefaultSecondDeriv_d( this );
}

void HCL_EvaluateOp_d::DerivImage( const HCL_Vector_d & dx,
                                         HCL_Vector_d & dy ) const
{
   cerr << "Error: HCL_EvaluateOp_d::DerivImage not implemented" << endl;
   exit(1);
}

void HCL_EvaluateOp_d::DerivAdjImage( const HCL_Vector_d & dy,
                                            HCL_Vector_d & dx ) const
{
   cerr << "Error: HCL_EvaluateOp_d::DerivAdjImage not implemented" << endl;
   exit(1);
}

void HCL_EvaluateOp_d::DerivInvImage( const HCL_Vector_d & dy,
                                            HCL_Vector_d & dx ) const
{
   cerr << "Error: HCL_EvaluateOp_d::DerivInvImage not implemented" << endl;
   exit(1);
}

void HCL_EvaluateOp_d::DerivInvAdjImage( const HCL_Vector_d & dx,
                                             HCL_Vector_d & dy ) const
{
   cerr << "Error: HCL_EvaluateOp_d::DerivInvAdjImage not implemented" << endl;
   exit(1);
}

void HCL_EvaluateOp_d::SecondDerivImage( const HCL_Vector_d & dx1,
                                         const HCL_Vector_d & dx2,
                                               HCL_Vector_d & dy ) const
{
   cerr << "Error: HCL_EvaluateOp_d::SecondDerivImage not implemented" << endl;
   exit(1);
}

void HCL_EvaluateOp_d::SecondDerivPartialAdjImage( const HCL_Vector_d & dx1,
                                         const HCL_Vector_d & dy,
                                               HCL_Vector_d & dx2 ) const
{
   cerr << "Error: HCL_EvaluateOp_d::SecondDerivPartialAdjImage not "
           "implemented" << endl;
   exit(1);
}

void HCL_EvaluateOp_d::ImageAdd( HCL_Vector_d & y,double a,double b ) const
{
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_EvaluateOp_d::ImageAdd: " << endl
           << "output vector is not in range" << endl;
      cerr << "Space of output vector:" << endl;
      y.Space().Write( cerr );
      cerr << "Range:" << endl;
      Range().Write( cerr );
      exit(1);
   }
   HCL_FPArch_d arch;
   HCL_Vector_d * z = Range().Member();
   Image( *z );
   if( fabs(a-1.0) > arch.eps )
      z->Mul( a );
   if( fabs(b-1.0) > arch.eps )
      y.ScaleAdd( b,*z );
   else
      y.Add( *z );
   HCL_delete( z );
}

void HCL_EvaluateOp_d::ImageAdd( const HCL_Vector_d & z,
                                 HCL_Vector_d & y,double a,double b ) const
{
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_EvaluateOp_d::ImageAdd: " << endl
           << "output vector is not in range" << endl;
      cerr << "Space of output vector:" << endl;
      y.Space().Write( cerr );
      cerr << "Range:" << endl;
      Range().Write( cerr );
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_EvaluateOp_d::ImageAdd: " << endl
           << "input vector is not in range" << endl;
      cerr << "Space of input vector:" << endl;
      z.Space().Write( cerr );
      cerr << "Range:" << endl;
      Range().Write( cerr );
      exit(1);
   }
   HCL_FPArch_d arch;
   Image( y );
   if( fabs(a-1.0) > arch.eps )
      y.Mul( a );
   if( fabs(b-1.0) > arch.eps )
      y.ScaleAdd( b,z );
   else
      y.Add( z );
}

const HCL_LinearOp_d & HCL_EvaluateOp_d::DerivRef() const
{
   HCL_EvaluateOp_d * t = (HCL_EvaluateOp_d*)this;
   if( SavedDeriv == NULL )
      t->SavedDeriv = Deriv();
   return *SavedDeriv;
}

const HCL_BiLinearOp_d & HCL_EvaluateOp_d::SecondDerivRef() const
{
   HCL_EvaluateOp_d * t = (HCL_EvaluateOp_d*)this;
   if( Saved2ndDeriv == NULL )
      t->Saved2ndDeriv = SecondDeriv();
   return *Saved2ndDeriv;
}

HCL_Op_d::HCL_Op_d(){}
HCL_Op_d::HCL_Op_d( const HCL_Op_d & ){}
HCL_Op_d::~HCL_Op_d(){}

double HCL_Op_d::MaxStep( const HCL_Vector_d & x, 
                          const HCL_Vector_d &dir) const
{
   HCL_FPArch_d arch;
   return arch.large;
}

void HCL_Op_d::Image( const HCL_Vector_d & x,HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::Image: vector is not in domain"
           << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_Op_d::Image: vector is not in range"
           << endl;
      exit(1);
   }
   HCL_EvaluateOp_d * eval = Evaluate( x );
   eval->Image( y );
   HCL_delete( eval );
}

void HCL_Op_d::ImageAdd( const HCL_Vector_d & x,HCL_Vector_d & y,
                         double a,double b ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::Image: vector is not in domain"
           << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_Op_d::Image: vector is not in range"
           << endl;
      exit(1);
   }
   HCL_EvaluateOp_d * eval = Evaluate( x );
   eval->ImageAdd( y,a,b );
   HCL_delete( eval );
}

void HCL_Op_d::ImageAdd( const HCL_Vector_d & x,const HCL_Vector_d & z,
                         HCL_Vector_d & y,
                         double a,double b ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::Image: vector is not in domain"
           << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_Op_d::Image: vector is not in range"
           << endl;
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_Op_d::Image: vector is not in range"
           << endl;
      exit(1);
   }
   HCL_EvaluateOp_d * eval = Evaluate( x );
   eval->ImageAdd( z,y,a,b );
   HCL_delete( eval );
}

HCL_LinearOp_d * HCL_Op_d::Deriv( const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::Deriv: vector is not in domain" << endl;
      exit(1);
   }
   HCL_EvaluateOp_d * eval = Evaluate( x );
   HCL_LinearOp_d * D = eval->Deriv();
   HCL_delete( eval );
   return D;
}

int HCL_Op_d::CheckDeriv( const HCL_Vector_d & y,
                               const HCL_Vector_d & p,ostream & str,
                               int n,double hmin,double hmax ) const
{
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::CheckDeriv: "
               "vector y is not in Domain" << endl;
      exit(1);
   }
   if( p.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::CheckDeriv: "
              "vector p is not in Domain" << endl;
      exit(1);
   }

   if( hmax <= hmin )
   {
      double temp = hmax;
      hmax = hmin;
      hmin = temp;
   }
   if( hmin <= 0.0 )
   {
      hmin = 0.1;
      hmax = 1.0;
   }
   if( n <= 0 )
      n = 10;

   double hlimit1 = MaxStep( y,p );
   HCL_Vector_d * ptemp = p.Space().Member();
   ptemp->Mul( -1.0,p );
   double hlimit2 = MaxStep( y,*ptemp );
   HCL_delete( ptemp );
   double hlimit = min( hlimit1,hlimit2 );
   if( hlimit <= 0.0 )
   {
      cerr << "Error in HCL_Op_d::CheckDeriv: direction is not "
              "feasible" << endl;
      exit(1);
   }
   if( hmax >= hlimit )
   {
      hmax = 0.99*hlimit;
      if( hmin >= hmax )
         hmin = hmax/n;
   }

   HCL_Vector_d * x1 = Domain().Member();
   HCL_Vector_d * x2 = Domain().Member();
   HCL_Vector_d * g1 = Range().Member();
   HCL_Vector_d * g2 = Range().Member();
   HCL_Vector_d * dg = Range().Member();

   HCL_LinearOp_d * DFy = Deriv( y );
   DFy->Image( p,*dg );
   HCL_delete( DFy );

#if SINGLE_PRECISION
   int nd = 8;
#else
   int nd = 16;
#endif
   int oldprecision = str.precision( nd );

   double dgnorm = dg->Norm();
   int rflag = 1;
   HCL_FPArch_d arch;
   if( dgnorm < arch.sfmin )
   {
      rflag = 0;
      str << "HCL_Op_d::CheckDeriv: norm of first variation is too "
          << endl << "small; displaying absolute error" << endl;
   }

   if( rflag )
      str << setw(5) << "h" << setw(nd+1) << " norm of diff." << setw(nd+11)
           << "norm" << setw(nd+4) << "Rel. Err." << endl;
   else
      str << setw(5) << "h" << setw(nd+1) << " norm of diff." << setw(nd+11)
           << "norm" << endl;
   int i;
   double hstep = (hmax-hmin)/(n-1);
   for( i=n-1;i>=0;i-- )
   {
      double h = hmin+i*hstep;
      x1->AddScale( -h,y,p );
      Image( *x1,*g1 );
      x2->AddScale( h,y,p );
      Image( *x2,*g2 );
      g2->Sub( *g1 );
      g2->Mul( 1.0/(2.0*h) );
      g2->Sub( *dg );
      double n1 = g2->Norm();
      if( rflag )
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << n1 << " " << setw(nd+6)
             << dgnorm << " " << setw(nd+6) << n1/dgnorm << endl;
      else
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << n1 << " " << setw(nd+6)
             << dgnorm << endl;
   }
   HCL_delete( dg );
   HCL_delete( g2 );
   HCL_delete( g1 );
   HCL_delete( x2 );
   HCL_delete( x1 );
   str.precision( oldprecision );
   return 0;
}

int HCL_Op_d::CheckDerivAdj( const HCL_Vector_d & y,
                             const HCL_Vector_d & p,ostream & str,
                             int n,double hmin,double hmax ) const
{
   CheckDeriv( y,p,str,n,hmin,hmax );
   HCL_LinearOp_d * D = Deriv( y );
   D->CheckAdj( 2 );
   HCL_delete( D );
   return 0;
}

int HCL_Op_d::CheckSecondDeriv( const HCL_Vector_d & y,
                               const HCL_Vector_d & p,ostream & str,
                               int n,double hmin,double hmax ) const
{
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::CheckSecondDeriv: "
               "vector y is not in Domain" << endl;
      exit(1);
   }
   if( p.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::CheckSecondDeriv: "
              "vector p is not in Domain" << endl;
      exit(1);
   }

   if( hmax <= hmin )
   {
      double temp = hmax;
      hmax = hmin;
      hmin = temp;
   }
   if( hmin <= 0.0 )
   {
      hmin = 0.1;
      hmax = 1.0;
   }
   if( n <= 0 )
      n = 10;

   double hlimit1 = MaxStep( y,p );
   HCL_Vector_d * ptemp = p.Space().Member();
   ptemp->Mul( -1.0,p );
   double hlimit2 = MaxStep( y,*ptemp );
   HCL_delete( ptemp );
   double hlimit = min( hlimit1,hlimit2 );
   if( hlimit <= 0.0 )
   {
      cerr << "Error in HCL_Op_d::CheckSecondDeriv: direction is not "
              "feasible" << endl;
      exit(1);
   }
   if( hmax >= hlimit )
   {
      hmax = 0.99*hlimit;
      if( hmin >= hmax )
         hmin = hmax/n;
   }

   HCL_Vector_d * x1 = Domain().Member();
   HCL_Vector_d * x2 = Domain().Member();
   HCL_Vector_d * g0 = Range().Member();
   HCL_Vector_d * g1 = Range().Member();
   HCL_Vector_d * g2 = Range().Member();
   HCL_Vector_d * d2g = Range().Member();

   HCL_BiLinearOp_d * D2Fy = SecondDeriv( y );
   D2Fy->Image( p,p,*d2g );
   HCL_delete( D2Fy );

#if SINGLE_PRECISION
   int nd = 8;
#else
   int nd = 16;
#endif
   int oldprecision = str.precision( nd );

   double d2gnorm = d2g->Norm();
   int rflag = 1;
   HCL_FPArch_d arch;
   if( d2gnorm < arch.sfmin )
   {
      rflag = 0;
      str << "HCL_Op_d::CheckSecondDeriv: norm of second variation is too "
          << endl << "small; displaying absolute error" << endl;
   }

   if( rflag )
      str << setw(5) << "h" << setw(nd+1) << " norm of diff." << setw(nd+11)
           << "norm" << setw(nd+4) << "Rel. Err." << endl;
   else
      str << setw(5) << "h" << setw(nd+1) << " norm of diff." << setw(nd+11)
           << "norm" << endl;
   int i;
   double hstep = (hmax-hmin)/(n-1);
   Image( y,*g0 );
   for( i=n-1;i>=0;i-- )
   {
      double h = hmin+i*hstep;
      x1->AddScale( -h,y,p );
      Image( *x1,*g1 );
      x2->AddScale( h,y,p );
      Image( *x2,*g2 );
      g2->Add( *g1 );
      g2->AddScale( -2.0,*g0 );
      g2->Mul( 1.0/(h*h) );
      g2->Sub( *d2g );
      double n1 = g2->Norm();
      if( rflag )
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << n1 << " " << setw(nd+6)
             << d2gnorm << " " << setw(nd+6) << n1/d2gnorm << endl;
      else
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << n1 << " " << setw(nd+6)
             << d2gnorm << endl;
   }
   HCL_delete( d2g );
   HCL_delete( g2 );
   HCL_delete( g1 );
   HCL_delete( g0 );
   HCL_delete( x2 );
   HCL_delete( x1 );
   str.precision( oldprecision );
   return 0;
}

int HCL_Op_d::CheckSecondDerivAdj( const HCL_Vector_d & y,
                                   const HCL_Vector_d & p,ostream & str,
                                   int n,double hmin,double hmax ) const
{
   CheckSecondDeriv( y,p,str,n,hmin,hmax );
   HCL_BiLinearOp_d * B = SecondDeriv( y );
   B->CheckAdj( 2 );
   HCL_delete( B );
   return 0;
}

HCL_BiLinearOp_d * HCL_Op_d::SecondDeriv( const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::SecondDeriv: vector is not in "
              "domain" << endl;
      exit(1);
   }
   HCL_EvaluateOp_d * eval = Evaluate( x );
   HCL_BiLinearOp_d * d = eval->SecondDeriv();
   HCL_delete( eval );
   return d;
}

void HCL_Op_d::DerivImage( const HCL_Vector_d & x,const HCL_Vector_d & dx,
                                 HCL_Vector_d & dy ) const
{
   cerr << "Error: HCL_Op_d::DerivImage not implemented" << endl;
   exit(1);
}

void HCL_Op_d::DerivAdjImage( const HCL_Vector_d & x,const HCL_Vector_d & dy,
                                    HCL_Vector_d & dx ) const
{
   cerr << "Error: HCL_Op_d::DerivAdjImage not implemented" << endl;
   exit(1);
}

void HCL_Op_d::DerivInvImage( const HCL_Vector_d & x,const HCL_Vector_d & dy,
                                            HCL_Vector_d & dx ) const
{
   cerr << "Error: HCL_Op_d::DerivInvImage not implemented" << endl;
   exit(1);
}

void HCL_Op_d::DerivInvAdjImage( const HCL_Vector_d & x,
                                 const HCL_Vector_d & dx,
                                       HCL_Vector_d & dy ) const
{
   cerr << "Error: HCL_Op_d::DerivInvAdjImage not implemented" << endl;
   exit(1);
}

void HCL_Op_d::SecondDerivImage( const HCL_Vector_d & x,
                                 const HCL_Vector_d & dx1,
                                 const HCL_Vector_d & dx2,
                                       HCL_Vector_d & dy ) const
{
   cerr << "Error: HCL_Op_d::SecondDerivImage not implemented" << endl;
   exit(1);
}

void HCL_Op_d::SecondDerivPartialAdjImage( const HCL_Vector_d & x,
                                           const HCL_Vector_d & dx1,
                                           const HCL_Vector_d & dy,
                                                 HCL_Vector_d & dx2 ) const
{
   cerr << "Error: HCL_Op_d::SecondDerivPartialAdjImage not implemented"
        << endl;
   exit(1);
}

HCL_EvaluateOp_d * HCL_Op_d::Evaluate( const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_Op_d::Evaluate: input vector is not in domain"
           << endl;
      cerr << "Domain:" << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector:" << endl;
      x.Space().Write( cerr );
      exit(1);
   }
   return new HCL_OpDefaultEval_d( this,x );
}

void HCL_Op_d::Image1( const HCL_Vector_d & x,HCL_Vector_d & y ) const
{
   cerr << "Error: HCL_Op_d::Image1 not implemented" << endl;
   exit(1);
}

HCL_LinearOp_d * HCL_Op_d::Deriv1( const HCL_Vector_d & x ) const
{
   return (HCL_LinearOp_d*)NULL;
}

HCL_BiLinearOp_d * HCL_Op_d::SecondDeriv1( const HCL_Vector_d & x ) const
{
   return (HCL_BiLinearOp_d*)NULL;
}

HCL_OpDefaultEval_d::HCL_OpDefaultEval_d()
{
   cerr << "Error: HCL_OpDefaultEval_d default constructor not implemented"
        << endl;
   exit(1);
}

HCL_OpDefaultEval_d::HCL_OpDefaultEval_d( const HCL_OpDefaultEval_d & L )
{
   cerr << "Error: HCL_OpDefaultEval_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_OpDefaultEval_d::HCL_OpDefaultEval_d( const HCL_Op_d * op_,
                                          const HCL_Vector_d & in )
{
   op = op_;op->IncCount();
   x = in.Space().Member();
   x->Copy( in );
}

HCL_OpDefaultEval_d::~HCL_OpDefaultEval_d()
{
   HCL_delete( x );
   HCL_delete( (HCL_Op_d*)op );
}

HCL_LinearOp_d * HCL_OpDefaultEval_d::Deriv() const
{
   HCL_LinearOp_d * DFx = op->Deriv1( *x );
   if( DFx )
      return DFx;
   else
      return new HCL_OpDefaultDeriv_d( this );
}

HCL_BiLinearOp_d * HCL_OpDefaultEval_d::SecondDeriv() const
{
   HCL_BiLinearOp_d * D2Fx = op->SecondDeriv1( *x );
   if( D2Fx )
      return D2Fx;
   else
      return new HCL_OpDefaultSecondDeriv_d( this );
}

ostream & HCL_OpDefaultEval_d::Write( ostream & str ) const
{
   str << "HCL_OpPDDefaultEval_d" << endl;
   str << "Underlying operator is: " << endl;
   op->Write( str );
   return str;
}

HCL_OpDefaultDeriv_d::HCL_OpDefaultDeriv_d()
{
   cerr << "Error: HCL_OpDefaultDeriv_d default constructor not implemented"
        << endl;
   exit(1);
}

HCL_OpDefaultDeriv_d::HCL_OpDefaultDeriv_d( const HCL_OpDefaultDeriv_d & L )
{
   cerr << "Error: HCL_OpDefaultDeriv_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_OpDefaultDeriv_d::HCL_OpDefaultDeriv_d( const HCL_EvaluateOp_d * ev )
{
   eval = ev;ev->IncCount();
}

HCL_OpDefaultDeriv_d::~HCL_OpDefaultDeriv_d()
{
   HCL_delete( (HCL_EvaluateOp_d*)eval );
}

HCL_VectorSpace_d & HCL_OpDefaultDeriv_d::Domain() const
{
   return eval->Domain();
}

HCL_VectorSpace_d & HCL_OpDefaultDeriv_d::Range() const
{
   return eval->Range();
}

void HCL_OpDefaultDeriv_d::Image( const HCL_Vector_d & x,
                                        HCL_Vector_d & y ) const
{
   eval->DerivImage( x,y );
}

void HCL_OpDefaultDeriv_d::AdjImage( const HCL_Vector_d & y,
                                           HCL_Vector_d & x ) const
{
   eval->DerivAdjImage( y,x );
}

void HCL_OpDefaultDeriv_d::InvImage(const HCL_Vector_d & y,
                                          HCL_Vector_d & x) const
{
   eval->DerivInvImage( y,x );
}

void HCL_OpDefaultDeriv_d::InvAdjImage(const HCL_Vector_d & x,
                                             HCL_Vector_d & y) const
{
   eval->DerivInvAdjImage( x,y );
}

ostream & HCL_OpDefaultDeriv_d::Write( ostream & str ) const
{
   str << "HCL_OpDefaultDeriv_d based on the following evaluation object"
       << endl;
   return eval->Write( str );
}

HCL_OpDefaultSecondDeriv_d::HCL_OpDefaultSecondDeriv_d()
{
   cerr << "Error: HCL_OpDefaultSecondDeriv_d default constructor not "
           "implemented" << endl;
   exit(1);
}

HCL_OpDefaultSecondDeriv_d::HCL_OpDefaultSecondDeriv_d(
                            const HCL_OpDefaultSecondDeriv_d & L )
{
   cerr << "Error: HCL_OpDefaultSecondDeriv_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_OpDefaultSecondDeriv_d::HCL_OpDefaultSecondDeriv_d(
                            const HCL_EvaluateOp_d * ev )
{
   eval = ev;ev->IncCount();
}

HCL_OpDefaultSecondDeriv_d::~HCL_OpDefaultSecondDeriv_d()
{
   HCL_delete( (HCL_EvaluateOp_d*)eval );
}

void HCL_OpDefaultSecondDeriv_d::Image( const HCL_Vector_d & dx1,
                                        const HCL_Vector_d & dx2,
                                              HCL_Vector_d & dy ) const
{
   eval->SecondDerivImage( dx1,dx2,dy );
}

void HCL_OpDefaultSecondDeriv_d::PartialAdjImage( int i,
                                              const HCL_Vector_d & xy,
                                              const HCL_Vector_d & z,
                                                    HCL_Vector_d & yx ) const
{
   eval->SecondDerivPartialAdjImage( xy,z,yx );
}

ostream & HCL_OpDefaultSecondDeriv_d::Write( ostream & str ) const
{
   str << "HCL_OpDefaultSecondDeriv_d based on the following evaluation object"
       << endl;
   return eval->Write( str );
}
