/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

// For random numbers
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "HCL_Vector_d.h"
#include "HCL_Arch_d.h"

int HCL_VectorSpace_d::Test( int display,double tolerance ) const
{

   HCL_FPArch_d arch;
   
   // Set up random number generator
   srand( (unsigned)time(NULL));

   double tol = tolerance * arch.eps;
   int errs = 0;

   HCL_Vector_d * x = Member();
   HCL_Vector_d * y = Member();
   HCL_Vector_d * z = Member();
   
   x->Random();
   y->Random();
   z->Random();

   errs += TestAdd(*x,*y, tol, display);
   errs += TestSub(*x, *y, tol, display);
   errs += TestMulDist(*x, *y, tol, display);
   errs += TestAddAssoc(*x, *y, *z,tol, display);
   errs += TestAddComm(*x, *y, tol, display);
   errs += TestSubAddNeg(*x, *y, tol, display);
   errs += TestScaleAdd(*x, *y, tol, display);
   errs += TestNegMul(*x, tol, display);
   errs += TestSubDist(*x, *y, *z, tol, display);
   errs += TestIdentity(*x, tol, display);
   errs += TestZeroMul0(*x, tol, display);
   errs += TestAddZero(*x, tol, display);
   errs += TestNegZero(tol, display);
   errs += TestAddNeg(*x, tol, display);
   errs += TestMul(*x, tol, display);
   errs += TestSAddAddS(*x, *y, tol, display);
   errs += TestNormNorm2(*x, tol, display);
   errs += TestNormInner(*x, tol, display);

   if( display )
      cout << endl << "HCL_VectorSpace_d::Test: there were " << errs 
           << " errors detected." << endl;
   return errs;

   HCL_delete(z);
   HCL_delete(y);
   HCL_delete(x);
}


int HCL_VectorSpace_d::TestAdd( HCL_Vector_d& x, HCL_Vector_d& y,
                                double tol, int display ) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;

   z1->Add(x,y);

   z2->Copy(x);
   z2->Add(y);

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x+y","z <- x, z <- z+y",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x+y","z <- x, z <- z+y",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestSub(HCL_Vector_d& x, HCL_Vector_d& y, 
   double tol, int display) const 
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1,ndiff;
   z1->Sub(x,y);

   z2->Copy(x);
   z2->Sub(y);
   
   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x-y","z <- x, z <- z-y",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x-y","z <- x, z <- z-y",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestMulDist(HCL_Vector_d& x, HCL_Vector_d& y,
double tol, int display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   double Scalar = (double)((double)rand()/(double)rand());;
   z1->Add(x,y);
   z1->Mul(Scalar);
   z2->Copy(x);
   z2->Mul(Scalar);
   z2->AddScale(Scalar, y);
   
   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x+y, z <- a*z","z <- a*x, z <- z+a*y",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x+y, z <- a*z","z <- a*x, z <- z+a*y",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}


int HCL_VectorSpace_d::TestAddComm(HCL_Vector_d& x, HCL_Vector_d& y,
double tol, int display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   z1->Add(x,y);
   z2->Add(y,x);

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x+y","z <- y+x",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x+y","z <- y+x",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}


int HCL_VectorSpace_d::TestAddAssoc(HCL_Vector_d& x, HCL_Vector_d& y,
HCL_Vector_d& z, double tol, int display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   z1->Add(x,y);
   z1->Add(z);
   z2->Add(y,z);
   z2->Add(x);
   
   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "w <- x+y, w <- w+z","w <- y+z, w <- w+x",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "w <- x+y, w <- w+z","w <- y+z, w <- w+x",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestSubAddNeg(HCL_Vector_d& x, HCL_Vector_d& y,
double tol, int display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   z1->Copy(y);
   z1->Neg();
   z1->Add(x);
   
   z2->Sub(x,y);

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- y,z <- -z, z <- z+x","z <- x-y",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- y,z <- -z, z <- z+x","z <- x-y",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}


int HCL_VectorSpace_d::TestScaleAdd(HCL_Vector_d& x, HCL_Vector_d&
y, double tol, int display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   double Scalar = (double)(double)rand()/(double)rand();
   // z1 = x + ky
   z1->AddScale(Scalar, x, y);
   z2->Copy(x);
   z2->AddScale(Scalar, y); 
   
   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x+a*y","z <- x,z <- z+a*y",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x+a*y","z <- x,z <- z+a*y",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestNegMul(HCL_Vector_d& x, double tol, int
display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   z1->Copy(x);
   z1->Neg();

   z2->Copy(x);
   z2->Mul(-1.0);

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x,z <- -z","z <- -1*x",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x,z <- -z","z <- -1*x",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestSubDist(HCL_Vector_d& x, HCL_Vector_d& y,
HCL_Vector_d& z, double tol, int display) const
{

   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   HCL_Vector_d *Temp = Member();
   double n1, ndiff;
   Temp->Add(y,z);
   z1->Sub(x, *Temp);

   z2->Sub(x,y);
   z2->Sub(z);

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "w <- x-(y+z)","w <- x-y,w <- w-z",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "w <- x-(y+z)","w <- x-y,w <- w-z",ndiff,n1 );
   HCL_delete(Temp);
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestIdentity(HCL_Vector_d& x, double tol, int
display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   z1->Copy(x);
   z2->Copy(x);
   z2->Mul(1.0);

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x","z <- 1*x",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x","z <- 1*x",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}   

int HCL_VectorSpace_d::TestZeroMul0(HCL_Vector_d& x, double tol, int
display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   z1->Zero();
   z2->Copy(x);
   z2->Mul(0.0);

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- 0","z <- 0*x",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- 0","z <- 0*x",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestAddZero(HCL_Vector_d& x, double tol, int
display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   
   z1->Copy(x);
   z2->Zero();
   z2->Add(x);
   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x","z <- 0,z <- z+x",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x","z <- 0,z <- z+x",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestNegZero(double tol, int display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double n1, ndiff;
   z1->Zero();
   z2->Zero();
   z2->Neg();

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- 0","z <- 0,z <- -z",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- 0","z <- 0,z <- -z",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}


int HCL_VectorSpace_d::TestAddNeg(HCL_Vector_d& x, double tol, int 
display)
const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double ndiff,n1;

   z1->Sub( x,x );
   z2->Zero();

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x-x","z <- 0",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x-x","z <- 0",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestMul(HCL_Vector_d& x, double tol, int display)
const
{

   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double Scalar = (double)(double)rand()/(double)rand();
   double n1,ndiff;
   z1->Mul(Scalar, x);
   z2->Copy(x);
   z2->Mul(Scalar);

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- a*x","z <- x,z <- a*z",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- a*x","z <- x,z <- a*z",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestSAddAddS(HCL_Vector_d &x, HCL_Vector_d &y,
double tol, int display) const
{
   HCL_Vector_d *z1 = Member();
   HCL_Vector_d *z2 = Member();
   double ndiff,n1;
   double Scalar = (double)(double)rand()/(double)rand();
   
   z1->Copy(x);
   z1->ScaleAdd(Scalar, y);
   z2->Copy(y);
   z2->AddScale(Scalar, x);

   int flag;
   if( (flag=Compare( *z1,*z2,tol,&n1,&ndiff )) )
   {
      if( display )
         Message( "z <- x,z <- a*z+y","z <- y,z <- z+a*x",ndiff,n1 );
   }
   else
      if( display > 1 )
         Message1( "z <- x,z <- a*z+y","z <- y,z <- z+a*x",ndiff,n1 );
   HCL_delete( z2 );
   HCL_delete( z1 );
   return flag;
}

int HCL_VectorSpace_d::TestNormNorm2(HCL_Vector_d& x, double tol,
                                     int display) const
{
   double Norm1, Norm2;
   Norm1 = x.Norm();
   Norm1 *= Norm1;
   Norm2 = x.Norm2();
   
   if (fabs(Norm2-Norm1) > fabs(tol * Norm2))   
   {
      if( display )
         cout << endl << "Error detected while testing vector space: "
              << endl
              << "   x.Norm()*x.Norm() differs from x.Norm2()" << endl
              << "   Difference: " << Norm1-Norm2 << endl
              << "   First result: " << Norm1 << endl;
      return 1;
   }
   else
   {
      if( display > 1 )
         cout << endl << "No error detected while testing vector space: "
              << endl
              << "   x.Norm()*x.Norm() agrees with x.Norm2()" << endl
              << "   Difference: " << Norm1-Norm2 << endl
              << "   First result: " << Norm1 << endl;
      return 0;
   }
}

int HCL_VectorSpace_d::TestNormInner(HCL_Vector_d& x, double tol,
                                     int display) const
{
   double Inner;
   double Norm2;
   Norm2 = x.Norm2();
   Inner = x.Inner(x);
        
   if (fabs(Inner-Norm2) > fabs(Inner * tol))
   {
      if( display )
        cout << endl << "Error detected while testing vector space: "
             << endl
             << "   x.Norm2() differs from x.Inner(x)" << endl
             << "   Difference: " << Norm2-Inner << endl
             << "   First result: " << Norm2 << endl;
      return 1;
   }
   else
   {
      if( display > 1 )
        cout << endl << "No error detected while testing vector space: "
             << endl
             << "   x.Norm2() agrees with x.Inner(x)" << endl
             << "   Difference: " << Norm2-Inner << endl
             << "   First result: " << Norm2 << endl;
      return 0;
   }
}

// Returns 1 if same, 0 if different
int HCL_VectorSpace_d::Compare(HCL_Vector_d& x, HCL_Vector_d& y, double tol,
                               double *n, double *ndiff) const
{   y.Sub(x);
   *ndiff = y.Norm();
   *n = x.Norm();
   
   if (*ndiff > tol * (*n))
      return 1;
   else
      return 0;
}

void HCL_VectorSpace_d::Message( const char * str1,const char * str2,
                                 double ndiff,double n1 ) const
{
   cout << endl << "Error detected while testing vector space: " << endl
        << "   " << str1 << " differs from " << str2 << endl
        << "   Norm of difference: " << ndiff << endl
        << "   Norm of first result: " << n1 << endl;
}

void HCL_VectorSpace_d::Message1( const char * str1,const char * str2,
                                 double ndiff,double n1 ) const
{
   cout << endl << "No error detected while testing vector space: " << endl
        << "   " << str1 << " agrees with " << str2 << endl
        << "   Norm of difference: " << ndiff << endl
        << "   Norm of first result: " << n1 << endl;
}

double * HCL_Vector_d::Data()
{
   return (double*)NULL;
}

