/*
                 THE HILBERT CLASS LIBRARY                        
                Release 0.8, September 1997                       
                                                                  
       copyright (C) 1995, 1996, 1997 by Mark S. Gockenbach,      
              Matthew Petro, and William W. Symes                 
                                                                  
The source code contained in this file FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_ALFcnl_d.h"

HCL_EvalALFcnl_d::HCL_EvalALFcnl_d()
{
   cerr << "Error: HCL_EvalALFcnl_d default constructor called"
        << endl;
   exit(1);
}

HCL_EvalALFcnl_d::HCL_EvalALFcnl_d( const HCL_EvalALFcnl_d & )
{
   cerr << "Error: HCL_EvalALFcnl_d copy constructor called"
        << endl;
   exit(1);
}

HCL_EvalALFcnl_d::HCL_EvalALFcnl_d( HCL_EvaluateFunctional_d * fev,
                HCL_Vector_d * l,HCL_EvaluateOp_d * Hev,double r,
                HCL_LinearOp_d * SS )
{
   // No error checking is done for consistency of domains, etc.
   // This is already done in the functional class.

   feval = fev;fev->IncCount();
   Heval = Hev;Hev->IncCount();
   lambda = l;l->IncCount();
   S = SS;if( S ) S->IncCount();

   rho = r;
   Slambda = NULL;
   Hx = NULL;
   SHx = NULL;
   DHx = NULL;
   D2Hx = NULL;
   LamPlusRhoHx = NULL;
   SLamPlusRhoHx = NULL;
}


HCL_EvalALFcnl_d::~HCL_EvalALFcnl_d()
{
   HCL_delete( Slambda );
   HCL_delete( SHx );
   HCL_delete( S );
   HCL_delete( SLamPlusRhoHx );
   HCL_delete( LamPlusRhoHx );
   HCL_delete( DHx );
   HCL_delete( Hx );
   HCL_delete( lambda );
   HCL_delete( Heval );
   HCL_delete( feval );
}

double HCL_EvalALFcnl_d::Value() const
{
  // Casting away const
   HCL_EvalALFcnl_d * t = (HCL_EvalALFcnl_d*)this;
   if( !Hx )
   {
      t->Hx = Heval->Range().Member();
      Heval->Image( *(t->Hx) );
   }

   if( S && !SHx )
   {
      t->SHx = S->Range().Member();
      S->Image( *Hx,*(t->SHx) );
   }      

   double value = feval->ValueRef();
   if( SHx )
      value += lambda->Inner( *SHx );
   else
      value += lambda->Inner( *Hx );
   if( SHx )
      value += 0.5*rho*Hx->Inner( *SHx );
   else
      value += 0.5*rho*Hx->Norm2();

   return value;
}

double HCL_EvalALFcnl_d::fValue() const
{
   return feval->ValueRef();
}

void HCL_EvalALFcnl_d::ConstraintViolation( HCL_Vector_d & v )
{
   if( !Hx )
   {
      Hx = Heval->Range().Member();
      Heval->Image( *Hx );
   }
   v.Copy( *Hx );
}

void HCL_EvalALFcnl_d::Gradient( HCL_Vector_d & gg ) const
{
   // Casting away const
   HCL_EvalALFcnl_d * t = (HCL_EvalALFcnl_d*)this;
   if( !Hx )
   {
      t->Hx = Heval->Range().Member();
      Heval->Image( *(t->Hx) );
   }
   if( !LamPlusRhoHx && !S )
   {
      t->LamPlusRhoHx = lambda->Space().Member();
      (t->LamPlusRhoHx)->AddScale( rho,*lambda,*Hx );
   }
   if( !SHx && S )
   {
      t->SHx = S->Range().Member();
      S->Image( *Hx,*(t->SHx) );
   }
   if( !Slambda && S )
   {
      t->Slambda = S->Range().Member();
      S->Image( *lambda,*(t->Slambda) );
   }
   if( !SLamPlusRhoHx && S )
   {
      t->SLamPlusRhoHx = S->Range().Member();
      (t->SLamPlusRhoHx)->AddScale( rho,*Slambda,*SHx );
   }
   if( !DHx )
      t->DHx = Heval->Deriv();
   if( SLamPlusRhoHx )
      DHx->AdjImage( *SLamPlusRhoHx,gg );
   else
      DHx->AdjImage( *LamPlusRhoHx,gg );
   gg.Add( feval->GradientRef() );
}

void HCL_EvalALFcnl_d::GradientLag( HCL_Vector_d & gg )
{
   if( !DHx )
      DHx = Heval->Deriv();
   if( !Slambda && S )
   {
      Slambda = S->Range().Member();
      S->Image( *lambda,*Slambda );
   }
   if( Slambda )
      DHx->AdjImage( *Slambda,gg );
   else
      DHx->AdjImage( *lambda,gg );
   gg.Add( feval->GradientRef() );
}

HCL_LinearOp_d * HCL_EvalALFcnl_d::Hessian() const
{
  // Casting away const
   HCL_EvalALFcnl_d * t = (HCL_EvalALFcnl_d*)this;

   HCL_LinearOp_d * hessf = feval->Hessian();
   if( !DHx )
      t->DHx = Heval->Deriv();
   HCL_LinearOp_d * T;
   if( !S )
      T = new HCL_Normal_d( DHx );
   else
   {
      HCL_CompLinearOp_d * t1 = new HCL_CompLinearOp_d(
                                    &(Domain()),&(Domain()),3 );
      t1->SetNext( DHx );
      t1->SetNext( S );
      HCL_Adjoint_d * t2 = new HCL_Adjoint_d( DHx );
      t1->SetNext( t2 );
      HCL_delete( t2 );
      T = t1;
   }
   if( !D2Hx )
      t->D2Hx = Heval->SecondDeriv();
   if( !LamPlusRhoHx && !S )
   {
      if( !Hx )
      {
         t->Hx = Heval->Range().Member();
         Heval->Image( *(t->Hx) );
      }
      t->LamPlusRhoHx = lambda->Space().Member();
      (t->LamPlusRhoHx)->AddScale( rho,*lambda,*Hx );
   }
   else if( !SLamPlusRhoHx && S )
   {
      if( !Hx )
      {
         t->Hx = Heval->Range().Member();
         Heval->Image( *(t->Hx) );
      }
      if( !SHx )
      {
         t->SHx = S->Range().Member();
         S->Image( *Hx,*(t->SHx) );
      }
      if( !Slambda )
      {
         t->Slambda = S->Range().Member();
         S->Image( *lambda,*(t->Slambda) );
      }
      t->SLamPlusRhoHx = S->Range().Member();
      (t->SLamPlusRhoHx)->AddScale( rho,*Slambda,*SHx );
   }
   HCL_LinCombLinearOp_d * Hess = new HCL_LinCombLinearOp_d(
                                    &(Domain()),&(Domain()),3 );
   Hess->SetNext( hessf,1.0 );
   HCL_delete( hessf );
   Hess->SetNext( T,rho );
   HCL_delete( T );
   HCL_LinearOp_d * tmp;
   if( S )
      tmp = D2Hx->Op3( *SLamPlusRhoHx );
   else
      tmp = D2Hx->Op3( *LamPlusRhoHx );
   Hess->SetNext( tmp,1.0 );
   HCL_delete( tmp );
   return Hess;
}

void HCL_EvalALFcnl_d::MultiplierUpdate( HCL_Vector_d & l )
{
   if( SLamPlusRhoHx )
      l.Copy( *SLamPlusRhoHx );
   else if( LamPlusRhoHx && !S )
      l.Copy( *LamPlusRhoHx );
   else
   {
      if( !Hx )
      {
         Hx = Heval->Range().Member();
         Heval->Image( *Hx );
      }
      if( S )
      {
         if( !SHx )
         {
            SHx = S->Range().Member();
            S->Image( *Hx,*SHx );
         }
         if( !Slambda )
         {
            Slambda = S->Range().Member();
            S->Image( *lambda,*Slambda );
         }
         if( !SLamPlusRhoHx )
         {
            SLamPlusRhoHx = S->Range().Member();
            SLamPlusRhoHx->AddScale( rho,*Slambda,*SHx );
         }
         l.Copy( *SLamPlusRhoHx );
      }
      else
      {         
         if( !LamPlusRhoHx )
         {
            LamPlusRhoHx = lambda->Space().Member();
            LamPlusRhoHx->AddScale( rho,*lambda,*Hx );
         }
         l.Copy( *LamPlusRhoHx );
      }
   }
}

ostream & HCL_EvalALFcnl_d::Write( ostream & str ) const
{
   str << "HCL_EvalALFcnl_d with: " << endl;
   str << "feval = " << endl;
   feval->Write( str );
   str << "lambda = " << endl;
   lambda->Write( str );
   str << "Heval = " << endl;
   Heval->Write( str );
   return str;
}

HCL_ALFcnl_d::HCL_ALFcnl_d()
{
   cerr << "Error: HCL_ALFcnl_d default constructor called"
        << endl;
   exit(1);
}

HCL_ALFcnl_d::HCL_ALFcnl_d( const HCL_ALFcnl_d & F )
{
   cerr << "Error: HCL_ALFcnl_d copy constructor called"
        << endl;
   exit(1);
}

HCL_ALFcnl_d::HCL_ALFcnl_d( HCL_Functional_d * ff,
                            HCL_Op_d * HH,HCL_Vector_d * l,double rho,
                            HCL_LinearOp_d * SS )
{
   if( ff->Domain() != HH->Domain() )
   {
      cerr << "Error in HCL_ALFcnl_d constructor: objective "
              "function and constraint operator must have same domain"
           << endl;
      exit(1);
   }
   if( l->Space() != HH->Range() )
   {
      cerr << "Error in HCL_ALFcnl_d constructor: multiplier "
              "must belong to range of constraint operator"
           << endl;
      exit(1);
   }
   if( SS )
      if( SS->Range() != HH->Range() )
      {
         cerr << "Error in HCL_ALFcnl_d constructor: ranges of "
                 "operator and preconditioner differ" << endl;
         exit(1);
      }
      else if( SS->Domain() != SS->Range() )
      {
         cerr << "Error in HCL_ALFcnl_d constructor: domain and "
                 "range of preconditioner differ" << endl;
         exit(1);
      }
      
   f = ff;ff->IncCount();
   H = HH;HH->IncCount();
   ParamTable = new Table;
   ParamTable->PutValue( "PenParam",rho );
   lam = l;l->IncCount();
   S = SS;
   if( S ) S->IncCount();
}

HCL_ALFcnl_d::~HCL_ALFcnl_d()
{
   HCL_delete( S );
   HCL_delete( lam );
   HCL_delete( ParamTable );
   HCL_delete( H );
   HCL_delete( f );
}

double HCL_ALFcnl_d::MaxStep( const HCL_Vector_d & x, 
                              const HCL_Vector_d &dir) const
{
   double s = min( f->MaxStep( x,dir ),H->MaxStep( x,dir ) );
   return s;
}

HCL_EvaluateFunctional_d * 
           HCL_ALFcnl_d::Evaluate( const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ALFcnl_d::Evaluate: vector is not in "
           << "domain" << endl;
      exit(1);
   }
   double rho;
   ParamTable->GetValue( "PenParam",rho );
   HCL_EvaluateFunctional_d * feval = f->Evaluate( x );
   HCL_EvaluateOp_d * Heval = H->Evaluate( x );
   HCL_EvaluateFunctional_d * Feval;
   Feval = new HCL_EvalALFcnl_d( feval,lam,Heval,rho,S );
   HCL_delete( Heval );
   HCL_delete( feval );
   return Feval;
}

ostream & HCL_ALFcnl_d::Write( ostream & str ) const
{
   str << "HCL_ALFcnl_d defined by:" << endl;
   str << "f (objective function):" << endl;
   f->Write( str );
   str << "H (equality constraint operator): " << endl;
   H->Write( str );
   return str;
}
