#include "HCL_Adjoint_d.h"

HCL_Adjoint_d::HCL_Adjoint_d()
{
   cerr << "Error: HCL_Adjoint_d default constructor called" << endl;
   exit(1);
}

HCL_Adjoint_d::HCL_Adjoint_d( const HCL_Adjoint_d & x )
{
   cerr << "Error: HCL_Adjoint_d copy constructor called" << endl;
   exit(1);
}

HCL_Adjoint_d::HCL_Adjoint_d( HCL_LinearOp_d * l )
{
   L = l;l->IncCount();
}

HCL_Adjoint_d::~HCL_Adjoint_d()
{
   HCL_delete( L );
}

HCL_VectorSpace_d & HCL_Adjoint_d::Domain() const { return L->Range(); }
HCL_VectorSpace_d & HCL_Adjoint_d::Range() const { return L->Domain(); }

void HCL_Adjoint_d::Image( const HCL_Vector_d & x,
                                 HCL_Vector_d & y ) const
{
   L->AdjImage( x,y );
}

void HCL_Adjoint_d::AdjImage( const HCL_Vector_d & y,
                                    HCL_Vector_d & x ) const
{
   L->Image( y,x );
}

void HCL_Adjoint_d::InvImage(const HCL_Vector_d & y,HCL_Vector_d & x) const
{
   L->InvAdjImage( y,x );
}

void HCL_Adjoint_d::InvAdjImage(const HCL_Vector_d & x,HCL_Vector_d & y) const
{
   L->InvImage( x,y );
}

void HCL_Adjoint_d::ImageAdd( const HCL_Vector_d & x,const HCL_Vector_d & z,
                                    HCL_Vector_d & y,double a,
                                    double b ) const
{
   L->AdjImageAdd( x,z,y,a,b );
}

void HCL_Adjoint_d::AdjImageAdd( const HCL_Vector_d & x,const HCL_Vector_d & z,
                                       HCL_Vector_d & y,double a,
                                       double b) const
{
   L->ImageAdd( x,z,y,a,b );
}

void HCL_Adjoint_d::InvImageAdd( const HCL_Vector_d & x,const HCL_Vector_d & z,
                                       HCL_Vector_d & y,double a,
                                       double b ) const
{
   L->InvAdjImageAdd( x,z,y,a,b );
}

void HCL_Adjoint_d::InvAdjImageAdd( const HCL_Vector_d & x,
                                    const HCL_Vector_d & z,
                                          HCL_Vector_d & y,
                                          double a,double b) const
{
   L->InvImageAdd( x,z,y,a,b );
}

void HCL_Adjoint_d::ImageAdd( const HCL_Vector_d & x,HCL_Vector_d & y,
                              double a,double b ) const
{
   L->AdjImageAdd( x,y,a,b );
}

void HCL_Adjoint_d::AdjImageAdd( const HCL_Vector_d & x,HCL_Vector_d & y,
                                 double a,double b) const
{
   L->ImageAdd( x,y,a,b );
}

void HCL_Adjoint_d::InvImageAdd( const HCL_Vector_d & x,
                                       HCL_Vector_d & y,double a,
                                       double b ) const
{
   L->InvAdjImageAdd( x,y,a,b );
}

void HCL_Adjoint_d::InvAdjImageAdd( const HCL_Vector_d & x,
                                          HCL_Vector_d & y,
                                          double a,double b) const
{
   L->InvImageAdd( x,y,a,b );
}

ostream & HCL_Adjoint_d::Write( ostream & str ) const
{
   str << "Adjoint operator of" << endl;
   return L->Write( str );
}

HCL_Normal_d::HCL_Normal_d()
{
   cerr << "Error: HCL_Normal_d default constructor called" << endl;
   exit(1);
}

HCL_Normal_d::HCL_Normal_d( const HCL_Normal_d & x )
{
   cerr << "Error: HCL_Normal_d copy constructor called" << endl;
   exit(1);
}

HCL_Normal_d::HCL_Normal_d( HCL_LinearOp_d * l )
{
   L = l;l->IncCount();
}

HCL_Normal_d::~HCL_Normal_d()
{
   HCL_delete( L );
}

HCL_VectorSpace_d & HCL_Normal_d::Domain() const
{ 
   return L->Domain();
}

HCL_VectorSpace_d & HCL_Normal_d::Range() const
{
   return Domain();
}

void HCL_Normal_d::Image( const HCL_Vector_d & x,
                                HCL_Vector_d & z ) const
{
   L->NormalImage( x,z );
}

void HCL_Normal_d::AdjImage( const HCL_Vector_d & x,
                                   HCL_Vector_d & z ) const
{
   L->NormalImage( x,z );
}

void HCL_Normal_d::InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const
{
   if( y.Space() != L->Domain() )
   {
      cerr << "Error in HCL_Normal_d::InvImage: vector is not in range"
           << endl;
      exit(1);
   }
   if( x.Space() != L->Domain() )
   {
      cerr << "Error in HCL_Normal_d::InvImage: vector is not in domain"
           << endl;
      exit(1);
   }
   HCL_Vector_d * tempvec = L->Range().Member();
   L->InvAdjImage( y,*tempvec );
   L->InvImage( *tempvec,x );
   HCL_delete( tempvec );
}

void HCL_Normal_d::InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const
{
   InvImage( x,y );
}

void HCL_Normal_d::ImageAdd( const HCL_Vector_d & x,const HCL_Vector_d & z,
                                    HCL_Vector_d & y,double a,
                                    double b ) const
{
   L->NormalImageAdd( x,z,y,a,b );
}

void HCL_Normal_d::AdjImageAdd( const HCL_Vector_d & x,const HCL_Vector_d & z,
                                       HCL_Vector_d & y,double a,
                                       double b) const
{
   L->NormalImageAdd( x,z,y,a,b );
}

void HCL_Normal_d::ImageAdd( const HCL_Vector_d & x,HCL_Vector_d & y,
                              double a,double b ) const
{
   L->NormalImageAdd( x,y,a,b );
}

void HCL_Normal_d::AdjImageAdd( const HCL_Vector_d & x,HCL_Vector_d & y,
                                 double a,double b) const
{
   L->NormalImageAdd( x,y,a,b );
}

ostream & HCL_Normal_d::Write( ostream & str ) const
{
   str << "Normal operator of" << endl;
   return L->Write( str );
}

HCL_InvLinearOp_d::HCL_InvLinearOp_d()
{
   cerr << "Error: HCL_InvLinearOp_d default constructor called" << endl;
   exit(1);
}

HCL_InvLinearOp_d::HCL_InvLinearOp_d( const HCL_InvLinearOp_d & x )
{
   cerr << "Error: HCL_InvLinearOp_d copy constructor called" << endl;
   exit(1);
}

HCL_InvLinearOp_d::HCL_InvLinearOp_d( HCL_LinearOp_d * l,
                                      HCL_LinearSolver_d *ls )
{
   L = l; L->IncCount();
   lsolver = ls; lsolver->IncCount();
}

HCL_InvLinearOp_d::HCL_InvLinearOp_d( HCL_LinearOp_d * l )
{
   L = l;L->IncCount();
   lsolver = NULL;
}

HCL_InvLinearOp_d::~HCL_InvLinearOp_d()
{
   HCL_delete (lsolver);
   HCL_delete (L);
}

HCL_VectorSpace_d & HCL_InvLinearOp_d::Domain() const
{
   return L->Range();
}

HCL_VectorSpace_d & HCL_InvLinearOp_d::Range() const
{
   return L->Domain();
}

void HCL_InvLinearOp_d::Image( const HCL_Vector_d & x,
                                     HCL_Vector_d & y ) const
{
   if( lsolver )
      lsolver->Solve(*L, x, y);
   else 
      L->InvImage( x,y );
}

void HCL_InvLinearOp_d::InvImage( const HCL_Vector_d & y,
                                        HCL_Vector_d & x ) const
{
   L->Image( y,x );
}

void HCL_InvLinearOp_d::AdjImage( const HCL_Vector_d & x,
                                        HCL_Vector_d & z ) const
{
   L->InvAdjImage( x,z );
}

void HCL_InvLinearOp_d::InvAdjImage(const HCL_Vector_d & x,
                                          HCL_Vector_d & y) const
{
   L->AdjImage( x,y );
}

void HCL_InvLinearOp_d::ImageAdd( const HCL_Vector_d & x,
                                  const HCL_Vector_d & z,
                                        HCL_Vector_d & y,double a,
                                        double b ) const
{
   L->InvImageAdd( x,z,y,a,b );
}

void HCL_InvLinearOp_d::AdjImageAdd( const HCL_Vector_d & x,
                                     const HCL_Vector_d & z,
                                           HCL_Vector_d & y,double a,
                                           double b) const
{
   L->InvAdjImageAdd( x,z,y,a,b );
}

void HCL_InvLinearOp_d::InvImageAdd( const HCL_Vector_d & x,
                                     const HCL_Vector_d & z,
                                           HCL_Vector_d & y,
                                           double a,
                                           double b ) const
{
   L->ImageAdd( x,z,y,a,b );
}

void HCL_InvLinearOp_d::InvAdjImageAdd( const HCL_Vector_d & x,
                                        const HCL_Vector_d & z,
                                              HCL_Vector_d & y,
                                              double a,
                                              double b ) const
{
   L->AdjImageAdd( x,z,y,a,b );
}

void HCL_InvLinearOp_d::ImageAdd( const HCL_Vector_d & x,HCL_Vector_d & y,
                                  double a,double b ) const
{
   L->InvImageAdd( x,y,a,b );
}

void HCL_InvLinearOp_d::AdjImageAdd( const HCL_Vector_d & x,HCL_Vector_d & y,
                                     double a,double b) const
{
   L->InvAdjImageAdd( x,y,a,b );
}

void HCL_InvLinearOp_d::InvImageAdd( const HCL_Vector_d & x,
                                           HCL_Vector_d & y,
                                           double a,
                                           double b ) const
{
   L->ImageAdd( x,y,a,b );
}

void HCL_InvLinearOp_d::InvAdjImageAdd( const HCL_Vector_d & x,
                                              HCL_Vector_d & y,
                                              double a,
                                              double b ) const
{
   L->AdjImageAdd( x,y,a,b );
}

ostream & HCL_InvLinearOp_d::Write( ostream & str ) const
{
   str << "Inverse operator of" << endl;
   L->Write( str );
   return str;
}
