/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_BigMFcnl_d.h"

HCL_BigMFcnlHessian_d::HCL_BigMFcnlHessian_d()
{
   cerr << "Error: HCL_BigMFcnlHessian_d default constructor called" << endl;
   exit(1);
}

HCL_BigMFcnlHessian_d::HCL_BigMFcnlHessian_d(
                       const HCL_BigMFcnlHessian_d & L )
{
   cerr << "Error: HCL_BigMFcnlHessian_d copy constructor called" << endl;
   exit(1);
}

HCL_BigMFcnlHessian_d::HCL_BigMFcnlHessian_d( HCL_GenericProductSpace_d * d,
                                              HCL_LinearOp_d * AA,
                                              double MM,
                                              HCL_Vector_d * msk )
{
   dom = d;dom->IncCount();
   M = MM;
   A = AA;A->IncCount();
   mask = msk;mask->IncCount();
   msum = mask->Sum();
   temp = (A->Range().Member());
   Ate = A->Domain().Member();
   A->AdjImage( *mask,*Ate );
}

HCL_BigMFcnlHessian_d::~HCL_BigMFcnlHessian_d()
{
   HCL_delete( Ate );
   HCL_delete( temp );
   HCL_delete( mask );
   HCL_delete( A );
   HCL_delete( dom );
}

HCL_VectorSpace_d & HCL_BigMFcnlHessian_d::Domain() const
{
   return *dom;
}

HCL_VectorSpace_d & HCL_BigMFcnlHessian_d::Range() const
{
   return *dom;
}

void HCL_BigMFcnlHessian_d::Image( const HCL_Vector_d & xx,
                                         HCL_Vector_d & yy ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_BigMFcnlHessian_d::Image: vector is not in domain"
           << endl;
      exit(1);
   }
   HCL_GenericProductVector_d & x = (HCL_GenericProductVector_d&)xx;
   HCL_RnVector_d & tau1 = (HCL_RnVector_d&)x(2);
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_BigMFcnlHessian_d::Image: vector is not in range"
           << endl;
      exit(1);
   }
   HCL_GenericProductVector_d & y = (HCL_GenericProductVector_d&)yy;
   HCL_RnVector_d & tau2 = (HCL_RnVector_d&)y(2);

   A->Image( x(1),*temp );
   temp->DotStar( *mask );
   A->AdjImage( *temp,y(1) );
   y(1).AddScale( -tau1(1),*Ate );
   tau2(1) = (M + msum)*tau1(1) - x(1).Inner( *Ate );
}

ostream & HCL_BigMFcnlHessian_d::Write( ostream & str ) const
{
   str << "HCL_BigMFcnlHessian_d defined by M = " << M << " and" << endl;
   return A->Write( str );
}

HCL_EvalBigMFcnl_d::HCL_EvalBigMFcnl_d()
{
   cerr << "Error: HCL_EvalBigMFcnl_d default constructor called" << endl;
   exit(1);
}

HCL_EvalBigMFcnl_d::HCL_EvalBigMFcnl_d( const HCL_EvalBigMFcnl_d & )
{
   cerr << "Error: HCL_EvalBigMFcnl_d copy constructor called" << endl;
   exit(1);
}

HCL_EvalBigMFcnl_d::HCL_EvalBigMFcnl_d( const HCL_GenericProductVector_d & y,
                                        HCL_LinearOp_d * AA,
                                        HCL_Vector_d * b,
                                        double MM,double ee )
{
   x = y(1).Space().Member();
   x->Copy( y(1) );
   tau = ((HCL_RnVector_d&)(y(2)))(1);
   dom = (HCL_GenericProductSpace_d*)(&y.Space());dom->IncCount();
   A = AA;A->IncCount();
   M = MM;
   eps = ee;

   res = (b->Space().Member());
   A->Image( *x,*res );
   res->Sub( *b );
   res->Add( -(tau-eps) );

   mask = (b->Space().Member());
   mask->Greater( *res,0.0 );
   res->DotStar( *mask );
}
   
HCL_EvalBigMFcnl_d::~HCL_EvalBigMFcnl_d()
{
   HCL_delete( mask );
   HCL_delete( res );
   HCL_delete( A );
   HCL_delete( dom );
   HCL_delete( x );
}

HCL_VectorSpace_d & HCL_EvalBigMFcnl_d::Domain() const
{
   return *dom;
}

double HCL_EvalBigMFcnl_d::Value() const
{
  HCL_Vector_d * res2 = (A->Range().Member());
  res2->DotStar( *res,*res );

  double val = 0.5*(M*tau*tau + res2->Sum());
  HCL_delete( res2 );
  return val;
}

void HCL_EvalBigMFcnl_d::Gradient( HCL_Vector_d & gg ) const
{
   if( gg.Space() != Domain() )
   {
      cerr << "Error in HCL_EvalBigMFcnl_d::Gradient: vector is not in domain"
           << endl;
      exit(1);
   }
   HCL_GenericProductVector_d & g = (HCL_GenericProductVector_d&)gg;
   HCL_RnVector_d & t = (HCL_RnVector_d&)g(2);

   A->AdjImage( *res,g(1) );
   t(1) = M*tau-res->Sum();
}

HCL_LinearOp_d * HCL_EvalBigMFcnl_d::Hessian() const
{
   return new HCL_BigMFcnlHessian_d( dom,A,M,mask );
}

ostream & HCL_EvalBigMFcnl_d::Write( ostream & str ) const
{
   str << "HCL_EvalBigMFcnl_d object with: " << endl;
   str << "tau = " << tau << " M = " << M << endl;
   str << "A = " << endl;
   return A->Write( str );
}

HCL_BigMFcnl_d::HCL_BigMFcnl_d()
{
   cerr << "Error: HCL_BigMFcnl_d default constructor called" << endl;
   exit(1);
}

HCL_BigMFcnl_d::HCL_BigMFcnl_d( const HCL_BigMFcnl_d & )
{
   cerr << "Error: HCL_BigMFcnl_d copy constructor called" << endl;
   exit(1);
}

HCL_BigMFcnl_d::HCL_BigMFcnl_d( HCL_LinearOp_d * AA,
                                HCL_Vector_d * bb,
                                double MM,double eeps )
{
   if( bb->Space() != AA->Range() )
   {
      cerr << "Error in HCL_BigMFcnl_d constructor: vector must lie in "
              "range" << endl;
      exit(1);
   }
   A = AA;A->IncCount();
   b = bb;b->IncCount();
   dom = new HCL_GenericProductSpace_d( 2 );
   HCL_RnSpace_d * R1 = new HCL_RnSpace_d( 1 );
   dom->SetNext( &(A->Domain()) );
   dom->SetNext( R1 );
   HCL_delete( R1 );
   M = MM;
   eps = eeps;
}

HCL_BigMFcnl_d::~HCL_BigMFcnl_d()
{
   HCL_delete( dom );
   HCL_delete( b );
   HCL_delete( A );
}

HCL_VectorSpace_d & HCL_BigMFcnl_d::Domain() const
{
   return *dom;
}

HCL_EvaluateFunctional_d * HCL_BigMFcnl_d::Evaluate(
                               const HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_BigMFcnl_d::Evaluate: vector must lie in "
              "domain" << endl;
      exit(1);
   }
   HCL_GenericProductVector_d & x = (HCL_GenericProductVector_d&)xx;
   return new HCL_EvalBigMFcnl_d( x,A,b,M,eps );
}

ostream & HCL_BigMFcnl_d::Write( ostream & str ) const
{
   str << "HCL_BigMFcnl_d object with: " << endl;
   str << " M = " << M << endl;
   str << "A = " << endl;
   A->Write( str );
   str << "b = " << endl;
   return b->Write( str );
}

