/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_BlockOp_d.h"

HCL_BlockLinearOp_d::HCL_BlockLinearOp_d()
{
   cerr << "Error: HCL_BlockLinearOp_d default constructor called"
        << endl;
         exit(1);
}

HCL_BlockLinearOp_d::HCL_BlockLinearOp_d( const HCL_BlockLinearOp_d & L )
{
   cerr << "Error: HCL_BlockLinearOp_d copy constructor called"
        << endl;
   exit(1);
}

HCL_BlockLinearOp_d::HCL_BlockLinearOp_d( HCL_ProductSpace_d * d,
                                          HCL_ProductSpace_d * r )
{
   dom = d;d->IncCount();
   range = r;r->IncCount();
   dom1 = range1 = NULL;
   rows = d->Number();
   cols = r->Number();
   RowOp = 0;
   ColOp = 0;
   num = rows*cols;
   blist = new HCL_LinearOp_d*[num];
   int i;
   for( i=0;i<num;i++ ) blist[i] = NULL;
}

HCL_BlockLinearOp_d::HCL_BlockLinearOp_d( HCL_ProductSpace_d * d,
                                          HCL_VectorSpace_d * r,int flag )
{
   if( flag==ROWOP )
   {
      dom = d;d->IncCount();
      range1 = r;r->IncCount();
      dom1 = NULL;
      range = NULL;
      rows = 1;
      cols = dom->Number();
      RowOp = 1;
      ColOp = 0;
      num = cols;
   }
   else if( flag==COLUMNOP )
   {
      dom1 = r;r->IncCount();
      range = d;d->IncCount();
      dom = NULL;
      range1 = NULL;
      rows = range->Number();
      cols = 1;
      num = rows;
      RowOp = 0;
      ColOp = 1;
   }
   blist = new HCL_LinearOp_d*[num];
   int i;
   for( i=0;i<num;i++ ) blist[i] = NULL;
}

HCL_BlockLinearOp_d::~HCL_BlockLinearOp_d()
{
   int i;
   for( i=0;i<num;i++ )
      HCL_delete( blist[i] );
   delete[] blist;
   HCL_delete( range1 );
   HCL_delete( range );
   HCL_delete( dom1 );
   HCL_delete( dom );
}

HCL_VectorSpace_d & HCL_BlockLinearOp_d::Domain() const
{
   if( ColOp )
      return *dom1;
   else
      return *dom;
}

HCL_VectorSpace_d & HCL_BlockLinearOp_d::Range() const
{
   if( RowOp )
      return *range1;
   else
      return *range;
}

int HCL_BlockLinearOp_d::Initialized() const
{
   int i;
   for( i=0;i<num;i++ )
      if( !blist[i] )
         return 0;
   return 1;
}

int HCL_BlockLinearOp_d::IsSet( int i,int j ) const
{
   int n = (i-1)*cols+j-1;
   if( blist[n] )
      return 1;
   else
      return 0;
}

void HCL_BlockLinearOp_d::SetNext( HCL_LinearOp_d * L )
{
   if( Initialized() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::SetNext: too many blocks"
           << endl;
      exit(1);
   }
      
   int i,j,n=0;;
   for( i=1;i<=rows;i++ )
      for( j=1;j<=cols;j++ )
      {
         if( !blist[n] )
         {
            if( ColOp )
            {
               if( L->Domain() != *dom1 )
               {
                  cerr << "Error in HCL_BlockLinearOp_d::SetNext: domains are "
                          "incompatible" << endl;
                  exit(1);
               }
            }
            else
            {
               if( L->Domain() != (*dom)(j) )
               {
                  cerr << "Error in HCL_BlockLinearOp_d::SetNext: domains are "
                          "incompatible" << endl;
                  exit(1);
               }
            }

            if( RowOp )
            {
               if( L->Range() != *range1 )
               {
                  cerr << "Error in HCL_BlockLinearOp_d::SetNext: ranges are "
                          "incompatible" << endl;
                  exit(1);
               }
            }
            else
            {
               if( L->Range() != (*range)(i) )
               {
                  cerr << "Error in HCL_BlockLinearOp_d::SetNext: ranges are "
                          "incompatible" << endl;
                  exit(1);
               }
            }

            blist[n] = L;
            L->IncCount();
            return;
         }
         n++;
      }
}            

void HCL_BlockLinearOp_d::Set( int i,int j,HCL_LinearOp_d * L )
{
   if( i < 1 || i > rows )
   {
      cerr << "Error in HCL_BlockLinearOp_d::Set: invalid row index"
           << endl;
      exit(1);
   }
   if( j < 1 || j > cols )
   {
      cerr << "Error in HCL_BlockLinearOp_d::Set: invalid column index"
           << endl;
      exit(1);
   }

   int n = (i-1)*cols+j-1;

   if( blist[n] )
      HCL_delete( blist[n] );
   
   if( ColOp )
   {
      if( L->Domain() != *dom1 )
      {
         cerr << "Error in HCL_BlockLinearOp_d::Set: domains are "
                 "incompatible" << endl;
         exit(1);
      }
   }
   else
   {
      if( L->Domain() != (*dom)(j) )
      {
         cerr << "Error in HCL_BlockLinearOp_d::Set: domains are "
                 "incompatible" << endl;
         exit(1);
      }
   }

   if( RowOp )
   {
      if( L->Range() != *range1 )
      {
         cerr << "Error in HCL_BlockLinearOp_d::Set: ranges are "
                 "incompatible" << endl;
         exit(1);
      }
   }
   else
   {
      if( L->Range() != (*range)(i) )
      {
         cerr << "Error in HCL_BlockLinearOp_d::Set: ranges are "
                 "incompatible" << endl;
         exit(1);
      }
   }

   blist[n] = L;
   L->IncCount();
}

void HCL_BlockLinearOp_d::Image( const HCL_Vector_d & xx,
                                       HCL_Vector_d & yy ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::Image: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::Image: vector is not in domain"
           << endl;
      exit(1);
   }
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::Image: vector is not in range"
           << endl;
      exit(1);
   }

   if( RowOp )
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      (*this)(1,1).Image( x(1),yy );
      int j;
      for( j=2;j<=cols;j++ )
         (*this)(1,j).ImageAdd( x(j),yy );
   }
   else if( ColOp )
   {
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      int i;
      for( i=1;i<=rows;i++ )
         (*this)(i,1).Image( xx,y(i) );
   }
   else
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      int i;
      for( i=1;i<=rows;i++ )
      {
         (*this)(i,1).Image( x(1),y(i) );
         int j;
         for( j=2;j<=cols;j++ )
            (*this)(i,j).ImageAdd( x(j),y(i) );
      }
   }
}

void HCL_BlockLinearOp_d::AdjImage( const HCL_Vector_d & xx,
                                          HCL_Vector_d & yy ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImage: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Range() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImage: vector is "
              "not in domain" << endl;
      exit(1);
   }
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImage: vector is "
              "not in range" << endl;
      exit(1);
   }

   if( ColOp )
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      (*this)(1,1).AdjImage( x(1),yy );
      int j;
      for( j=2;j<=rows;j++ )
         (*this)(j,1).AdjImageAdd( x(j),yy );
   }
   else if( RowOp )
   {
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      int i;
      for( i=1;i<=cols;i++ )
         (*this)(1,i).AdjImage( xx,y(i) );
   }
   else
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      int i;
      for( i=1;i<=cols;i++ )
      {
         (*this)(1,i).AdjImage( x(1),y(i) );
         int j;
         for( j=2;j<=rows;j++ )
            (*this)(j,i).AdjImageAdd( x(j),y(i) );
      }
   }
}

void HCL_BlockLinearOp_d::InvImage(const HCL_Vector_d & y,
                                         HCL_Vector_d & x) const
{
   cerr << "HCL_BlockLinearOp_d::InvImage is not implemented" << endl;
   exit(1);
}

void HCL_BlockLinearOp_d::InvAdjImage(const HCL_Vector_d & x,
                                            HCL_Vector_d & y) const
{
   cerr << "HCL_BlockLinearOp_d::InvAdjImage is not implemented" << endl;
   exit(1);
}

void HCL_BlockLinearOp_d::ImageAdd( const HCL_Vector_d & xx,
                                    const HCL_Vector_d & zz,
                                    HCL_Vector_d & yy,double a,
                                    double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::ImageAdd: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::ImageAdd: input vector "
              "is not in domain"
           << endl;
      exit(1);
   }
   if( zz.Space() != Range() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::ImageAdd: second input vector "
              "is not in range"
           << endl;
      exit(1);
   }
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::ImageAdd: output vector "
              "is not in range"
           << endl;
      exit(1);
   }

   if( RowOp )
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      (*this)(1,1).ImageAdd( x(1),zz,yy,a,b );
      int j;
      for( j=2;j<=cols;j++ )
         (*this)(1,j).ImageAdd( x(j),yy,a );
   }
   else if( ColOp )
   {
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;
      int i;
      for( i=1;i<=rows;i++ )
         (*this)(i,1).ImageAdd( xx,z(i),y(i),a,b );
   }
   else
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;
      int i;
      for( i=1;i<=rows;i++ )
      {
         (*this)(i,1).ImageAdd( x(1),z(i),y(i),a,b );
         int j;
         for( j=2;j<=cols;j++ )
            (*this)(i,j).ImageAdd( x(j),y(i),a );
      }
   }
}

void HCL_BlockLinearOp_d::AdjImageAdd( const HCL_Vector_d & xx,
                                       const HCL_Vector_d & zz,
                                       HCL_Vector_d & yy, double a,
                                       double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImageAdd: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Range() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImageAdd: input vector is "
              "not in range" << endl;
      exit(1);
   }
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImageAdd: output vector is "
              "not in domain" << endl;
      exit(1);
   }
   if( zz.Space() != Domain() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImageAdd: second input "
           << endl << "vector is not in domain" << endl;
      exit(1);
   }

   if( ColOp )
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      (*this)(1,1).AdjImageAdd( x(1),zz,yy,a,b );
      int j;
      for( j=2;j<=rows;j++ )
         (*this)(j,1).AdjImageAdd( x(j),yy,a );
   }
   else if( RowOp )
   {
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;
      int i;
      for( i=1;i<=cols;i++ )
         (*this)(1,i).AdjImageAdd( xx,z(i),y(i),a,b );
   }
   else
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;
      int i;
      for( i=1;i<=cols;i++ )
      {
         (*this)(1,i).AdjImageAdd( x(1),z(i),y(i),a,b );
         int j;
         for( j=2;j<=rows;j++ )
            (*this)(j,i).AdjImageAdd( x(j),y(i),a );
      }
   }
}

void HCL_BlockLinearOp_d::ImageAdd( const HCL_Vector_d & xx,
                                    HCL_Vector_d & yy,double a,
                                    double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::ImageAdd: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::ImageAdd: input vector "
              "is not in domain"
           << endl;
      exit(1);
   }
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::ImageAdd: output vector "
              "is not in range"
           << endl;
      exit(1);
   }

   if( RowOp )
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      (*this)(1,1).ImageAdd( x(1),yy,a,b );
      int j;
      for( j=2;j<=cols;j++ )
         (*this)(1,j).ImageAdd( x(j),yy,a );
   }
   else if( ColOp )
   {
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      int i;
      for( i=1;i<=rows;i++ )
         (*this)(i,1).ImageAdd( xx,y(i),a,b );
   }
   else
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      int i;
      for( i=1;i<=rows;i++ )
      {
         (*this)(i,1).ImageAdd( x(1),y(i),a,b );
         int j;
         for( j=2;j<=cols;j++ )
            (*this)(i,j).ImageAdd( x(j),y(i),a );
      }
   }
}

void HCL_BlockLinearOp_d::AdjImageAdd( const HCL_Vector_d & xx,
                                       HCL_Vector_d & yy, double a,
                                       double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImageAdd: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Range() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImageAdd: input vector is "
              "not in range" << endl;
      exit(1);
   }
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_BlockLinearOp_d::AdjImageAdd: output vector is "
              "not in domain" << endl;
      exit(1);
   }

   if( ColOp )
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      (*this)(1,1).AdjImageAdd( x(1),yy,a,b );
      int j;
      for( j=2;j<=rows;j++ )
         (*this)(j,1).AdjImageAdd( x(j),yy,a );
   }
   else if( RowOp )
   {
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      int i;
      for( i=1;i<=cols;i++ )
         (*this)(1,i).AdjImageAdd( xx,y(i),a,b );
   }
   else
   {
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
      int i;
      for( i=1;i<=cols;i++ )
      {
         (*this)(1,i).AdjImageAdd( x(1),y(i),a,b );
         int j;
         for( j=2;j<=rows;j++ )
            (*this)(j,i).AdjImageAdd( x(j),y(i),a );
      }
   }
}

ostream & HCL_BlockLinearOp_d::Write( ostream & str ) const
{
   str << "HCL_BlockLinearOp_d (" << rows << "," << cols << ") with "
          "blocks (by row):" << endl;
   int i;
   for( i=0;i<num;i++ )
      if( blist[i] )
         blist[i]->Write( str );
      else
         str << "This block not set yet" << endl;
   return str;
}

HCL_DiagBlockLinearOp_d::HCL_DiagBlockLinearOp_d()
{
   cerr << "Error: HCL_DiagBlockLinearOp_d default constructor called"
        << endl;
   exit(1);
}

HCL_DiagBlockLinearOp_d::HCL_DiagBlockLinearOp_d(
                         const HCL_DiagBlockLinearOp_d & L )
{
   cerr << "Error: HCL_DiagBlockLinearOp_d copy constructor called"
        << endl;
   exit(1);
}

HCL_DiagBlockLinearOp_d::HCL_DiagBlockLinearOp_d(
                                      HCL_ProductSpace_d * d,
                                      HCL_ProductSpace_d * r )
{
   if( d->Number() != r->Number() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d constructor: domain "
              "and range must have same number of factors" << endl;
      exit(1);
   }
   num = d->Number();
   dom = d;d->IncCount();
   range = r;r->IncCount();
   blist = new HCL_LinearOp_d*[num];
   int i;
   for( i=0;i<num;i++ ) blist[i] = NULL;
}

HCL_DiagBlockLinearOp_d::~HCL_DiagBlockLinearOp_d()
{
   int i;
   for( i=0;i<num;i++ )
      HCL_delete( blist[i] );
   delete[] blist;
   HCL_delete( range );
   HCL_delete( dom );
}

int HCL_DiagBlockLinearOp_d::Initialized() const
{
   int i;
   for( i=0;i<num;i++ )
      if( !blist[i] )
         return 0;
   return 1;
}

int HCL_DiagBlockLinearOp_d::IsSet( int i ) const
{
   if( blist[i-1] )
      return 1;
   else
      return 0;
}

void HCL_DiagBlockLinearOp_d::SetNext( HCL_LinearOp_d * L )
{
   if( Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::SetNext: too many "
              "blist" << endl;
      exit(1);
   }
   int i=0,j;
   for( j=0;j<num;j++ )
      if( blist[j] ) i++;
   if( L->Domain() != (*dom)(i+1) || L->Range() != (*range)(i+1) )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::SetNext: "
              "spaces do not agree" << endl;
      exit(1);
   }
   blist[i] = L;
   L->IncCount();
}

void HCL_DiagBlockLinearOp_d::Set( int i,HCL_LinearOp_d * L )
{
   if( i < 1 || i > num )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::Set: invalid index"
           << endl;
      exit(1);
   }

   if( blist[i-1] )
      HCL_delete( blist[i-1] );
   
   if( L->Domain() != (*dom)(i) )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::Set: domains are "
              "incompatible" << endl;
      exit(1);
   }

   if( L->Range() != (*range)(i) )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::Set: ranges are "
              "incompatible" << endl;
      exit(1);
   }

   blist[i-1] = L;
   L->IncCount();
}

void HCL_DiagBlockLinearOp_d::Image( const HCL_Vector_d & xx,
                                           HCL_Vector_d & yy ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::Image: not initialized"
           << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::Image: vector is not "
              "in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::Image: vector is not "
              "in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;

   int i;
   for( i=1;i<=num;i++ )
      (*this)(i).Image( x(i),y(i) );
}

void HCL_DiagBlockLinearOp_d::AdjImage( const HCL_Vector_d & xx,
                                              HCL_Vector_d & yy ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImage: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImage: vector is "
              "not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImage: vector is "
              "not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).AdjImage( x(i),y(i) );
}

void HCL_DiagBlockLinearOp_d::NormalImage( const HCL_Vector_d & xx,
                                                 HCL_Vector_d & zz ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImage: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImage: input vector is "
              "not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( zz.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImage: output "
              "vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).NormalImage( x(i),z(i) );
}

void HCL_DiagBlockLinearOp_d::InvAdjImage( const HCL_Vector_d & xx,
                                                 HCL_Vector_d & yy ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImage: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImage: vector "
              "is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImage: vector "
              "is not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).InvAdjImage( x(i),y(i) );
}

void HCL_DiagBlockLinearOp_d::InvImage( const HCL_Vector_d & xx,
                                              HCL_Vector_d & yy ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImage: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImage: vector is "
              "not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImage: vector is "
              "not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).InvImage( x(i),y(i) );
}

void HCL_DiagBlockLinearOp_d::ImageAdd(const HCL_Vector_d & xx,
                                       const HCL_Vector_d & zz,
                                             HCL_Vector_d & yy,
                                       double a,double b) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::ImageAdd: not initialized"
           << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::ImageAdd: vector is not "
              "in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::ImageAdd: vector is not "
              "in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   if( zz.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::ImageAdd: second " << endl
           << "input vector is not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;

   int i;
   for( i=1;i<=num;i++ )
      (*this)(i).ImageAdd( x(i),z(i),y(i),a,b );
}

void HCL_DiagBlockLinearOp_d::AdjImageAdd( const HCL_Vector_d & xx,
                                           const HCL_Vector_d & zz,
                                           HCL_Vector_d & yy,
                                           double a,double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImageAdd: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImageAdd: input vector "
           << endl << "is not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImageAdd: output vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   if( zz.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImageAdd: second "
           << endl << "input vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).AdjImageAdd( x(i),z(i),y(i),a,b );
}

void HCL_DiagBlockLinearOp_d::NormalImageAdd( const HCL_Vector_d & xx,
                                              const HCL_Vector_d & zz,
                                              HCL_Vector_d & yy,
                                              double a,double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImageAdd: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImageAdd: input "
           << endl << "vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImageAdd: output "
              "vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   if( zz.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImageAdd: second "
           << endl << "input vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).NormalImageAdd( x(i),z(i),y(i),a,b );
}

void HCL_DiagBlockLinearOp_d::InvImageAdd( const HCL_Vector_d & xx,
                                           const HCL_Vector_d & zz,
                                           HCL_Vector_d & yy,
                                           double a,double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImageAdd: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImageAdd: input vector"
           << endl << "is not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImageAdd: output vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   if( zz.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImageAdd: second"
           << endl << "input vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).InvImageAdd( x(i),z(i),y(i),a,b );
}

void HCL_DiagBlockLinearOp_d::InvAdjImageAdd( const HCL_Vector_d & xx,
                                              const HCL_Vector_d & zz,
                                              HCL_Vector_d & yy,
                                              double a,double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImageAdd: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImageAdd: input vector"
           << endl << "is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImageAdd: output "
           << endl << "vector is not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   if( zz.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImageAdd: second"
           << endl << "input vector is not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & z = (HCL_ProductVector_d&)zz;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).InvAdjImageAdd( x(i),z(i),y(i),a,b );
}

void HCL_DiagBlockLinearOp_d::ImageAdd(const HCL_Vector_d & xx,
                                             HCL_Vector_d & yy,
                                       double a,double b) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::ImageAdd: not initialized"
           << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::ImageAdd: vector is not "
              "in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::ImageAdd: vector is not "
              "in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;

   int i;
   for( i=1;i<=num;i++ )
      (*this)(i).ImageAdd( x(i),y(i),a,b );
}

void HCL_DiagBlockLinearOp_d::AdjImageAdd( const HCL_Vector_d & xx,
                                           HCL_Vector_d & yy,
                                           double a,double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImageAdd: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImageAdd: input vector "
           << endl << "is not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::AdjImageAdd: output vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).AdjImageAdd( x(i),y(i),a,b );
}

void HCL_DiagBlockLinearOp_d::NormalImageAdd( const HCL_Vector_d & xx,
                                              HCL_Vector_d & yy,
                                              double a,double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImageAdd: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImageAdd: input "
           << endl << "vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::NormalImageAdd: output "
              "vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).NormalImageAdd( x(i),y(i),a,b );
}

void HCL_DiagBlockLinearOp_d::InvImageAdd( const HCL_Vector_d & xx,
                                           HCL_Vector_d & yy,
                                           double a,double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImageAdd: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImageAdd: input vector"
           << endl << "is not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvImageAdd: output vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).InvImageAdd( x(i),y(i),a,b );
}

void HCL_DiagBlockLinearOp_d::InvAdjImageAdd( const HCL_Vector_d & xx,
                                              HCL_Vector_d & yy,
                                              double a,double b ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImageAdd: not "
              "initialized" << endl;
      exit(1);
   }
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImageAdd: input vector"
           << endl << "is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_DiagBlockLinearOp_d::InvAdjImageAdd: output "
           << endl << "vector is not in range" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;

   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).InvAdjImageAdd( x(i),y(i),a,b );
}

ostream & HCL_DiagBlockLinearOp_d::Write( ostream & str ) const
{
   str << "DiagBlockLinearOp mapping the space" << endl;
   Domain().Write( str );
   str << "to the space" << endl;
   Range().Write( str );
   str << "The blocks are: " << endl;
   int i;
   for( i=0;i<Number();i++ )
      if( blist[i] )
      {
         str << "Block " << i+1 << ":" << endl;
         blist[i]->Write( str );
      }
   return str;
}

