#include "HCL_BlockBiLinearOp_d.h"

HCL_BlockBiLinearOp_d::HCL_BlockBiLinearOp_d()
{
   cerr << "Error: HCL_BlockBiLinearOp_d default constructor is not "
           "implemented" << endl;;
   exit(1);
}

HCL_BlockBiLinearOp_d::HCL_BlockBiLinearOp_d( const HCL_BlockBiLinearOp_d & )
{
   cerr << "Error: HCL_BlockBiLinearOp_d copy constructor is not "
           "implemented" << endl;;
   exit(1);
}

HCL_BlockBiLinearOp_d::HCL_BlockBiLinearOp_d( HCL_ProductSpace_d * d1,
                                              HCL_ProductSpace_d * d2,
                                              HCL_VectorSpace_d * r )
{
   X = d1;X->IncCount();
   Y = d2;Y->IncCount();
   Z = r;Z->IncCount();
   I = Y->Number();
   J = X->Number();
   blocks = new HCL_BiLinearOp_d*[I*J];
   int k;
   for( k=0;k<I*J;k++ ) blocks[k] = NULL;
}


HCL_BlockBiLinearOp_d::~HCL_BlockBiLinearOp_d()
{
   int i,j;
   for( i=0;i<I;i++ )
      for( j=0;j<J;j++ )
         HCL_delete( blocks[i*J+j] );
   delete[] blocks;
   HCL_delete( Z );
   HCL_delete( Y );
   HCL_delete( X );
}

void HCL_BlockBiLinearOp_d::Set( int i,int j,HCL_BiLinearOp_d * Bij )
{
   if( Bij->Domain1() != (*X)(j) )
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::Set: Domain1 of input "
           << endl << "operator does not match factor " << j << " of Domain1"
           << endl;
      cerr << "Domain1 of input operator:" << endl;
      Bij->Domain1().Write( cerr );
      cerr << "Factor " << j << " of Domain1:" << endl;
      (*X)(j).Write( cerr );
      exit(1);
   }
   if( Bij->Domain2() != (*Y)(i) )
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::Set: Domain2 of input "
           << endl << "operator does not match factor " << i << " of Domain2"
           << endl;
      cerr << "Domain2 of input operator:" << endl;
      Bij->Domain2().Write( cerr );
      cerr << "Factor " << i << " of Domain2:" << endl;
      (*Y)(i).Write( cerr );
      exit(1);
   }
   if( Bij->Range() != *Z )
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::Set: range of input "
           << endl << "operator does not range" << endl;
      cerr << "Range of input operator:" << endl;
      Bij->Range().Write( cerr );
      cerr << "Range:" << endl;
      (*Z).Write( cerr );
      exit(1);
   }

   if( blocks[(i-1)*J+j-1] )
      HCL_delete( blocks[(i-1)*J+j-1] );
   blocks[(i-1)*J+j-1] = Bij;
   Bij->IncCount();
}

int HCL_BlockBiLinearOp_d::Initialized() const
{
   int k;
   for( k=0;k<I*J;k++ )
      if( !blocks[k] )
         return 0;
   return 1;
}

int HCL_BlockBiLinearOp_d::IsSet( int i,int j ) const
{
   if( blocks[(i-1)*J+j-1] )
      return 1;
   return 0;
}

void HCL_BlockBiLinearOp_d::Image( const HCL_Vector_d & xx,
                                   const HCL_Vector_d & yy,
                                         HCL_Vector_d & z ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::Image: not initialized"
           << endl;
      exit(1);
   }
   if( xx.Space() != Domain1() )
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::Image: input vector "
              "is not in Domain1" << endl;
      cerr << "Space of input vector:" << endl;
      xx.Space().Write( cerr );
      cerr << "Domain1:" << endl;
      Domain1().Write( cerr );
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Domain2() )
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::Image: input vector "
              "is not in Domain2" << endl;
      cerr << "Space of input vector:" << endl;
      yy.Space().Write( cerr );
      cerr << "Domain2:" << endl;
      Domain2().Write( cerr );
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::Image: output vector "
              "is not in Range" << endl;
      cerr << "Space of output vector:" << endl;
      z.Space().Write( cerr );
      cerr << "Range:" << endl;
      Range().Write( cerr );
      exit(1);
   }
   HCL_Vector_d * temp = z.Space().Member();
   int i,j;
   (*this)( 1,1 ).Image( x(1),y(1),z );
   for( j=2;j<=J;j++ )
   {
      (*this)( 1,j ).Image( x(j),y(1),*temp );
      z.Add( *temp );
   }
   for( i=2;i<=I;i++ )
      for( j=1;j<=J;j++ )
      {
         (*this)( i,j ).Image( x(j),y(i),*temp );
         z.Add( *temp );
      }
   HCL_delete( temp );
}

ostream & HCL_BlockBiLinearOp_d::Write( ostream & str ) const
{
   str << "HCL_BlockBiLinearOp_d" << endl;
   str << "Domain1:" << endl;
   Domain1().Write( str );
   str << "Domain2:" << endl;
   Domain2().Write( str );
   str << "Range:" << endl;
   Range().Write( str );
   int i,j;
   for( i=1;i<=I;i++ )
      for( j=1;j<=J;j++ )
         if( blocks[(i-1)*J+j-1] )
         {
            str << "Block (" << i << "," << j << "):" << endl;
            (*this)( i,j ).Write( str );
         }
         else
            str << "Block (" << i << "," << j << ") not set" << endl;

   return str;
}

void HCL_BlockBiLinearOp_d::PartialAdjImage( int i,const HCL_Vector_d & xy,
                                                       const HCL_Vector_d & z,
                                                    HCL_Vector_d & yx ) const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::PartialAdjImage: "
              "not initialized" << endl;
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::PartialAdjImage: "
              "input vector is not in Range" << endl;
      cerr << "Space of intput vector:" << endl;
      z.Space().Write( cerr );
      cerr << "Range:" << endl;
      Range().Write( cerr );
      exit(1);
   }
   if( i==1 )
   {
      if( xy.Space() != Domain1() )
      {
         cerr << "Error in HCL_BlockBiLinearOp_d::PartialAdjImage: "
                 "input vector is not in Domain1" << endl;
         cerr << "Space of input vector:" << endl;
         xy.Space().Write( cerr );
         cerr << "Domain1:" << endl;
         Domain1().Write( cerr );
         exit(1);
      }
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)xy;
      if( yx.Space() != Domain2() )
      {
         cerr << "Error in HCL_BlockBiLinearOp_d::"
                 "PartialAdjImage: output vector "
                 "is not in Domain2" << endl;
         cerr << "Space of output vector:" << endl;
         yx.Space().Write( cerr );
         cerr << "Domain2:" << endl;
         Domain2().Write( cerr );
         exit(1);
      }
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)yx;

      int i,j;
      for( i=1;i<=I;i++ )
      {
         HCL_Vector_d * temp = y(i).Space().Member();
         HCL_LinearOp_d * L = (*this)(i,1).Op1( x(1) );
         L->AdjImage( z,y(i) );
         HCL_delete( L );
         for( j=2;j<=J;j++ )
         {
            L = (*this)(i,j).Op1( x(j) );
            L->AdjImage( z,*temp );
            HCL_delete( L );
            y(i).Add( *temp );
         }
         HCL_delete( temp );
      }
   }
   else if( i==2 )
   {
      if( xy.Space() != Domain2() )
      {
         cerr << "Error in HCL_BlockBiLinearOp_d::PartialAdjImage: "
                 "input vector is not in Domain2" << endl;
         cerr << "Space of input vector:" << endl;
         xy.Space().Write( cerr );
         cerr << "Domain2:" << endl;
         Domain2().Write( cerr );
         exit(1);
      }
      HCL_ProductVector_d & y = (HCL_ProductVector_d&)xy;
      if( yx.Space() != Domain1() )
      {
         cerr << "Error in HCL_BlockBiLinearOp_d::"
                 "PartialAdjImage: output vector "
                 "is not in Domain1" << endl;
         cerr << "Space of output vector:" << endl;
         yx.Space().Write( cerr );
         cerr << "Domain1:" << endl;
         Domain1().Write( cerr );
         exit(1);
      }
      HCL_ProductVector_d & x = (HCL_ProductVector_d&)yx;

      int i,j;
      for( j=1;j<=J;j++ )
      {
         HCL_Vector_d * temp = x(j).Space().Member();
         HCL_LinearOp_d * L = (*this)( 1,j ).Op2( y(1) );
         L->AdjImage( z,x(j) );
         HCL_delete( L );
         for( i=2;i<=I;i++ )
         {
            L = (*this)(i,j).Op2( y(i) );
            L->AdjImage( z,*temp );
            HCL_delete( L );
            x(j).Add( *temp );
         }
         HCL_delete( temp );
      }
   }
   else
   {
      cerr << "Error in HCL_BlockBiLinearOp_d::PartialAdjImage: "
              "invalid flag (" << i << "); must be 1 or 2" << endl;
      exit(1);
   }
}
