/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_CompLinearOp_d.h"

HCL_CompLinearOp_d::HCL_CompLinearOp_d()
{
   cerr << "Error: HCL_CompLinearOp_d default constructor is not implemented"
        << endl;
   exit(1);
}

HCL_CompLinearOp_d::HCL_CompLinearOp_d( const HCL_CompLinearOp_d & L )
{
   cerr << "Error: HCL_CompLinearOp_d copy constructor is not implemented"
        << endl;
   exit(1);
}

HCL_CompLinearOp_d::HCL_CompLinearOp_d( HCL_VectorSpace_d * d,
                                        HCL_VectorSpace_d * r,int n )
{
   num = n;
   dom = d;d->IncCount();
   range = r;r->IncCount();
   ops = new HCL_LinearOp_d *[num];
   numinit = 0;
}

HCL_CompLinearOp_d::HCL_CompLinearOp_d( HCL_LinearOp_d * op1,
                                        HCL_LinearOp_d * op2 )
{
   if( op1->Range() != op2->Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d constructor: operators are "
              "incompatible" << endl;
      exit(1);
   }
   num = 2;
   numinit = 2;
   ops = new HCL_LinearOp_d*[2];
   ops[0] = op1;op1->IncCount();
   ops[1] = op2;op2->IncCount();
   dom = &(op1->Domain());dom->IncCount();
   range = &(op2->Range());range->IncCount();
}

HCL_CompLinearOp_d::~HCL_CompLinearOp_d()
{
   HCL_delete( range );
   HCL_delete( dom );
   int i;
   for( i=0;i<numinit;i++ )
      HCL_delete( ops[i] );
   delete[] ops;
}

void HCL_CompLinearOp_d::SetNext( HCL_LinearOp_d * L )
{
   if( numinit == num )
   {
      cerr << "Error in HCL_CompLinearOp_d::SetNext: "
              "too many factors" << endl;
      exit(1);
   }
   if( numinit == 0 )
   {
      if( L->Domain() != Domain() )
      {
         cerr << "Error in HCL_CompLinearOp_d::SetNext: "
                 "Domain and Range do not agree" << endl;
         exit(1);
      }
   }
   else if( L->Domain() != ops[numinit-1]->Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::SetNext: "
              "Domain and Range do not agree" << endl;
      exit(1);
   }
   if( numinit == num-1 && L->Range() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::SetNext: "
              "Domain and Range do not agree" << endl;
      exit(1);
   }
   ops[numinit] = L;L->IncCount();
   numinit++;
}

void HCL_CompLinearOp_d::Image( const HCL_Vector_d & x,
                                      HCL_Vector_d & y ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::Image: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::Image: vector is "
              "not in the Domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::Image: vector is "
              "not in the Range" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=0;i<num-1;i++ )
   {
      L = ops[i];
      out = L->Range().Member();
      L->Image( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[num-1];
   L->Image( *in,y );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::AdjImage( const HCL_Vector_d & x,
                                         HCL_Vector_d & y ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImage: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImage: vector is "
              "not in the Domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImage: vector is "
              "not in the Range" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=num-1;i>0;i-- )
   {
      L = ops[i];
      out = L->Domain().Member();
      L->AdjImage( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[0];
   L->AdjImage( *in,y );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::NormalImage( const HCL_Vector_d & x,
                                            HCL_Vector_d & y ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImage: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImage: input vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImage: output vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( num==1 )
   {
      ops[0]->NormalImage( x,y );
      return;
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=0;i<num-1;i++ )
   {
      L = ops[i];
      out = L->Range().Member();
      L->Image( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[num-1];
   out = L->Domain().Member();
   L->NormalImage( *in,*out );
   if( in != &x ) HCL_delete( in );
   in = out;
   for( i=num-2;i>0;i-- )
   {
      L = ops[i];
      out = L->Domain().Member();
      L->AdjImage( *in,*out );
      HCL_delete( in );
      in = out;
   }
   L = ops[0];
   L->AdjImage( *in,y );
   HCL_delete( in );
}

void HCL_CompLinearOp_d::InvImage( const HCL_Vector_d & x,
                                         HCL_Vector_d & y ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImage: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImage: vector is "
              "not in the Domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImage: vector is "
              "not in the Range" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=num-1;i>0;i-- )
   {
      L = ops[i];
      out = L->Domain().Member();
      L->InvImage( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[0];
   L->InvImage( *in,y );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::InvAdjImage( const HCL_Vector_d & x,
                                            HCL_Vector_d & y ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImage: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImage: vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImage: vector is "
              "not in the range" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=0;i<num-1;i++ )
   {
      L = ops[i];
      out = L->Range().Member();
      L->InvAdjImage( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[num-1];
   L->InvAdjImage( *in,y );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::ImageAdd( const HCL_Vector_d & x,
                                   const HCL_Vector_d & z,HCL_Vector_d & y,
                                   double a,double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::ImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::ImageAdd: input vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::ImageAdd: output vector is "
              "not in the range" << endl;
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::ImageAdd: second input vector is "
              "not in the range" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=0;i<num-1;i++ )
   {
      L = ops[i];
      out = L->Range().Member();
      L->Image( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[num-1];
   L->ImageAdd( *in,z,y,a,b );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::AdjImageAdd( const HCL_Vector_d & x,
                                      const HCL_Vector_d & z,HCL_Vector_d & y,
                                      double a,double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImageAdd: input vector is "
              "not in the range" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImageAdd: output vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( z.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImageAdd: second input "
              "vector is not in the domain" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=num-1;i>0;i-- )
   {
      L = ops[i];
      out = L->Domain().Member();
      L->AdjImage( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[0];
   L->AdjImageAdd( *in,z,y,a,b );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::NormalImageAdd( const HCL_Vector_d & x,
                                         const HCL_Vector_d & z,
                                         HCL_Vector_d & y,double a,
                                         double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImageAdd: input vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImageAdd: output vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( z.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImageAdd: second "
           << endl << "output vector is not in the domain" << endl;
      exit(1);
   }
   if( num==1 )
   {
      ops[0]->NormalImageAdd( x,z,y,a,b );
      return;
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=0;i<num-1;i++ )
   {
      L = ops[i];
      out = L->Range().Member();
      L->Image( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[num-1];
   out = L->Domain().Member();
   L->NormalImage( *in,*out );
   if( in != &x ) HCL_delete( in );
   in = out;
   for( i=num-2;i>0;i-- )
   {
      L = ops[i];
      out = L->Domain().Member();
      L->AdjImage( *in,*out );
      HCL_delete( in );
      in = out;
   }
   L = ops[0];
   L->AdjImageAdd( *in,z,y,a,b );
   HCL_delete( in );
}

void HCL_CompLinearOp_d::InvImageAdd( const HCL_Vector_d & x,
                                 const HCL_Vector_d & z,
                                 HCL_Vector_d & y,double a,double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImageAdd: input vector is "
              "not in the range" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImageAdd: output vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( z.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImageAdd: second input "
              "vector is not in the domain" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=num-1;i>0;i-- )
   {
      L = ops[i];
      out = L->Domain().Member();
      L->InvImage( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[0];
   L->InvImageAdd( *in,z,y,a,b );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::InvAdjImageAdd( const HCL_Vector_d & x,
                               const HCL_Vector_d & z,HCL_Vector_d & y,
                               double a,double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImageAdd: input vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImageAdd: output vector is "
              "not in the range" << endl;
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImageAdd: second "
           << endl << "output vector is not in the range" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=0;i<num-1;i++ )
   {
      L = ops[i];
      out = L->Range().Member();
      L->InvAdjImage( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[num-1];
   L->InvAdjImageAdd( *in,z,y,a,b );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::ImageAdd( const HCL_Vector_d & x,
                                   HCL_Vector_d & y,
                                   double a,double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::ImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::ImageAdd: input vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::ImageAdd: output vector is "
              "not in the range" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=0;i<num-1;i++ )
   {
      L = ops[i];
      out = L->Range().Member();
      L->Image( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[num-1];
   L->ImageAdd( *in,y,a,b );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::AdjImageAdd( const HCL_Vector_d & x,
                                      HCL_Vector_d & y,
                                      double a,double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImageAdd: input vector is "
              "not in the range" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::AdjImageAdd: output vector is "
              "not in the domain" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=num-1;i>0;i-- )
   {
      L = ops[i];
      out = L->Domain().Member();
      L->AdjImage( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[0];
   L->AdjImageAdd( *in,y,a,b );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::NormalImageAdd( const HCL_Vector_d & x,
                                         HCL_Vector_d & y,double a,
                                         double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImageAdd: input vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::NormalImageAdd: output vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( num==1 )
   {
      ops[0]->NormalImageAdd( x,y,a,b );
      return;
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=0;i<num-1;i++ )
   {
      L = ops[i];
      out = L->Range().Member();
      L->Image( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[num-1];
   out = L->Domain().Member();
   L->NormalImage( *in,*out );
   if( in != &x ) HCL_delete( in );
   in = out;
   for( i=num-2;i>0;i-- )
   {
      L = ops[i];
      out = L->Domain().Member();
      L->AdjImage( *in,*out );
      HCL_delete( in );
      in = out;
   }
   L = ops[0];
   L->AdjImageAdd( *in,y,a,b );
   HCL_delete( in );
}

void HCL_CompLinearOp_d::InvImageAdd( const HCL_Vector_d & x,
                                 HCL_Vector_d & y,double a,double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImageAdd: input vector is "
              "not in the range" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvImageAdd: output vector is "
              "not in the domain" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=num-1;i>0;i-- )
   {
      L = ops[i];
      out = L->Domain().Member();
      L->InvImage( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[0];
   L->InvImageAdd( *in,y,a,b );
   if( in != &x ) HCL_delete( in );
}

void HCL_CompLinearOp_d::InvAdjImageAdd( const HCL_Vector_d & x,
                               HCL_Vector_d & y,
                               double a,double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImageAdd: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImageAdd: input vector is "
              "not in the domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_CompLinearOp_d::InvAdjImageAdd: output vector is "
              "not in the range" << endl;
      exit(1);
   }
   HCL_Vector_d * in = (HCL_Vector_d*)&x;
   HCL_Vector_d * out;
   HCL_LinearOp_d * L;
   int i;
   for( i=0;i<num-1;i++ )
   {
      L = ops[i];
      out = L->Range().Member();
      L->InvAdjImage( *in,*out );
      if( in != &x ) HCL_delete( in );
      in = out;
   }
   L = ops[num-1];
   L->InvAdjImageAdd( *in,y,a,b );
   if( in != &x ) HCL_delete( in );
}

ostream & HCL_CompLinearOp_d::Write( ostream & str ) const
{
   str << "Composition operator with factors"
       << endl;
   int i;
   for( i=0;i<numinit;i++ )
      ops[i]->Write( str );
   return str;
}
