/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_DiagFunctional_d.h"

HCL_EvalDiagFcnl_d::HCL_EvalDiagFcnl_d()
{
   cerr << "Error: HCL_EvalDiagFcnl_d default constructor called"
        << endl;
   exit(1);
}

HCL_EvalDiagFcnl_d::HCL_EvalDiagFcnl_d( const HCL_EvalDiagFcnl_d & )
{
   cerr << "Error: HCL_EvalDiagFcnl_d copy constructor called"
        << endl;
   exit(1);
}

HCL_EvalDiagFcnl_d::HCL_EvalDiagFcnl_d( HCL_ProductVector_d & x,
                                        HCL_Functional_d ** flist )
{
   num = x.Number();
   dom = (HCL_ProductSpace_d*)&(x.Space());dom->IncCount();
   evallist = new HCL_EvaluateFunctional_d*[num];
   int i;
   for( i=1;i<=num;i++ )
      evallist[i-1] = flist[i-1]->Evaluate( x(i) );
   Zlist = new HCL_ZeroLinearOp_d*[num*num];
   for( i=0;i<num*num;i++ ) Zlist[i] = NULL;
}

HCL_EvalDiagFcnl_d::~HCL_EvalDiagFcnl_d()
{
   int i;
   for( i=0;i<num*num;i++ )
      HCL_delete( Zlist[i] );
   for( i=0;i<num;i++ )
      HCL_delete( evallist[i] );
   delete[] evallist;
   HCL_delete( dom );
}

double HCL_EvalDiagFcnl_d::Value() const
{
   double v = 0.0;
   for( int i=0;i<num;i++ )
      v += evallist[i]->Value();
   return v;
}

void HCL_EvalDiagFcnl_d::PartialGradient( int i,HCL_Vector_d & g ) const
{
   if( i < 1 || i > ProductDomain().Number() )
   {
      cerr << "Error in HCL_EvalDiagFcnl_d::PartialGradient: invalid index"
           << endl;
      exit(1);
   }
   if( g.Space() != ProductDomain()(i) )
   {
      cerr << "Error in HCL_EvalDiagFcnl_d::PartialGradient: vector is not "
              "in domain" << endl;
      exit(1);
   }
   evallist[i-1]->Gradient( g );
}

HCL_LinearOp_d * HCL_EvalDiagFcnl_d::Hessian() const
{
   HCL_DiagBlockLinearOp_d * hess = new HCL_DiagBlockLinearOp_d( dom,dom );
   int i;
   for( i=0;i<num;i++ )
   {
      HCL_LinearOp_d * tmp = evallist[i]->Hessian();
      hess->SetNext( tmp );
      HCL_delete( tmp );
   }
   return hess;
}

const HCL_LinearOp_d & HCL_EvalDiagFcnl_d::HessianRef() const
{
   // Casting away const
   HCL_EvalDiagFcnl_d * t = (HCL_EvalDiagFcnl_d*)this;
   if( !SavedHessian )
      t->SavedHessian = new HCL_DiagBlockLinearOp_d( &(ProductDomain()),
                                                  &(ProductDomain()) );
   int i;
   HCL_DiagBlockLinearOp_d & H = (HCL_DiagBlockLinearOp_d&)(*SavedHessian);
   for( i=1;i<=ProductDomain().Number();i++ )
      if( !H.IsSet( i ) )
      {
         HCL_LinearOp_d * L = HessianBlock( i,i );
         H.Set( i,L );
         HCL_delete( L );
      }
   return *SavedHessian;      
}

HCL_LinearOp_d * HCL_EvalDiagFcnl_d::HessianBlock(int i,int j ) const
{
   if( i < 1 || j < 1 || i > num || j > num )
   {
      cerr << "Error in HCL_EvalDiagFcnl_d::HessianBlock: invalid "
              "indices" << endl;
      exit(1);
   }
   if( i == j )
   {
      HCL_LinearOp_d * L = evallist[i-1]->Hessian();
      return L;
   }
   else
   {
      HCL_ZeroLinearOp_d * L = new HCL_ZeroLinearOp_d( &(ProductDomain()(j)),
                                                       &(ProductDomain()(i)) );
      return L;
   }
}

const HCL_LinearOp_d &
          HCL_EvalDiagFcnl_d::HessianBlockRef(int i,int j ) const
{
   if( i < 1 || j < 1 || i > num || j > num )
   {
      cerr << "Error in HCL_EvalDiagFcnl_d::HessianBlockRef: invalid "
              "indices" << endl;
      exit(1);
   }
   if( i == j )
   {
      // Casting away const
      HCL_EvalDiagFcnl_d * t = (HCL_EvalDiagFcnl_d*)this;
      if( !SavedHessian )
         t->SavedHessian = new HCL_DiagBlockLinearOp_d( &(ProductDomain()),
                                                     &(ProductDomain()) );
      HCL_DiagBlockLinearOp_d & H=(HCL_DiagBlockLinearOp_d&)(*SavedHessian);
      if( !H.IsSet( i ) )
      {
         HCL_LinearOp_d * L = HessianBlock( i,i );
         H.Set( i,L );
         HCL_delete( L );
      }
      return H(i);
   }
   else
   {
      if( !Zlist[num*(i-1)+j-1] )
         Zlist[num*(i-1)+j-1] = new HCL_ZeroLinearOp_d( &(ProductDomain()(j)),
                                                      &(ProductDomain()(i)) );
      return *Zlist[num*(i-1)+j-1];
   }
}

ostream & HCL_EvalDiagFcnl_d::Write( ostream & str ) const
{
   str << "HCL_EvalDiagFcnl_d with the following terms:" << endl;
   for( int i=0;i<num;i++ )
   {
      str << "Term " << i+1 << ":" << endl;
      evallist[i]->Write( str );
   }
   return str;
}

HCL_DiagFunctional_d::HCL_DiagFunctional_d()
{
   cerr << "Error: HCL_DiagFunctional_d default constructor called"
        << endl;
   exit(1);
}

HCL_DiagFunctional_d::HCL_DiagFunctional_d( const HCL_DiagFunctional_d & F )
{
   cerr << "Error: HCL_DiagFunctional_d copy constructor called"
        << endl;
   exit(1);
}

HCL_DiagFunctional_d::HCL_DiagFunctional_d( HCL_ProductSpace_d * d )
{
   num = d->Number();
   dom = d;d->IncCount();
   flist = new HCL_Functional_d*[num];
   int i;
   for( i=0;i<num;i++ ) flist[i] = NULL;
}

HCL_DiagFunctional_d::~HCL_DiagFunctional_d()
{
   int i;
   for( i=0;i<num;i++ )
      HCL_delete( flist[i] );
   delete[] flist;
   HCL_delete( dom );
}

int HCL_DiagFunctional_d::Initialized() const
{
   int i;
   for( i=0;i<num;i++ )
      if( !flist[i] )
         return 0;
   return 1;
}

double HCL_DiagFunctional_d::MaxStep( const HCL_Vector_d & xx,
                                      const HCL_Vector_d & ddir) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagFunctional_d::MaxStep: first input vector "
              "is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( ddir.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagFunctional_d::MaxStep: first input vector "
              "is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & dir = (HCL_ProductVector_d&)x;

   if( !Initialized() )
   {
      cerr << "Error in HCL_DiagFunctional_d::MaxStep: not all terms have "
              "been set" << endl;
      exit(1);
   }

   double s = FLT_MAX;
   int i;
   for( i=0;i<num;i++ )
      s = min( s,flist[i]->MaxStep( x(i),dir(i) ) );
   return s;
}

HCL_Functional_d & HCL_DiagFunctional_d::operator()( int i ) const
{
   if( i < 1 || i > num )
   {
      cerr << "Error in HCL_DiagFunctional_d::operator(): invalid index"
           << endl;
      exit(1);
   }
   if( !flist[i-1] )
   {
      cerr << "Error in HCL_DiagFunctional_d::operator(): requested term "
              "has not been set" << endl;
      exit(1);
   }
   return *flist[i-1];
}

HCL_EvaluateFunctional_d * HCL_DiagFunctional_d::Evaluate( 
                               const HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagFunctional_d::Evaluate: vector is "
              "not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   return new HCL_EvalDiagFcnl_d( x,flist );
}

void HCL_DiagFunctional_d::SetNext( HCL_Functional_d * f )
{
   if( Initialized() )
   {
      cerr << "Error in HCL_DiagFunctional_d::SetNext: too many "
              "functionals" << endl;
      exit(1);
   }
   int i=0,j;
   for( j=0;j<num;j++ )
      if( flist[j] ) i++;
   if( f->Domain() != (*dom)(i+1) )
   {
      cerr << "Error in HCL_DiagFunctional_d::SetNext: domains do "
              "not agree" << endl;
      exit(1);
   }
   flist[i] = f;
   f->IncCount();
}

ostream & HCL_DiagFunctional_d::Write( ostream & str ) const
{
   str << "HCL_DiagFunctional_d defined on" << endl;
   Domain().Write( str );
   str << "with components:" << endl;
   for( int i=1;i<=Number();i++ )
      if( flist[i-1] )
      {
         str << "Term " << i << ":" << endl;
         flist[i-1]->Write( str );
      }
   return str;
}

