/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#include "HCL_FunctionalProductDomain_d.h"

void HCL_EvalFunctionalProductDomain_d::HessianBlockImage( int i,int j,
                                        const HCL_Vector_d & dxj,
                                              HCL_Vector_d & dxi ) const
{
   cerr << "Error: HCL_EvalFunctionalProductDomain_d::HessianBlockImage "
           "is not implemented" << endl;
   exit(1);
}

HCL_EvalFunctionalProductDomain_d::HCL_EvalFunctionalProductDomain_d()
{
   PartialGradFlags = NULL;
}

HCL_EvalFunctionalProductDomain_d::HCL_EvalFunctionalProductDomain_d(
                  const HCL_EvalFunctionalProductDomain_d & f )
{
   if( f.PartialGradFlags )
   {
      PartialGradFlags = new int[f.ProductDomain().Number()];
      int i;
      for( i=0;i<f.ProductDomain().Number();i++ )
         PartialGradFlags[i] = f.PartialGradFlags[i];
   }
   else
      PartialGradFlags = NULL;
}

HCL_EvalFunctionalProductDomain_d::~HCL_EvalFunctionalProductDomain_d()
{
   if( PartialGradFlags )
      delete[] PartialGradFlags;
}

HCL_ProductSpace_d & HCL_EvalFunctionalProductDomain_d::ProductDomain() const
{
   return (HCL_ProductSpace_d&)Domain();
}

void HCL_EvalFunctionalProductDomain_d::Gradient( HCL_Vector_d & gg ) const
{
   if( gg.Space() != Domain() )
   {
      cerr << "Error in HCL_EvalFunctionalProductDomain_d::Gradient: "
              "vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & g = (HCL_ProductVector_d&)gg;
   int i;
   for( i=1;i<=ProductDomain().Number();i++ )
      PartialGradient( i,g(i) );
}

const HCL_Vector_d & HCL_EvalFunctionalProductDomain_d::GradientRef() const
{
   // Casting away const
   HCL_EvalFunctionalProductDomain_d * t =
                    (HCL_EvalFunctionalProductDomain_d*)this;
   int i;
   if( !SavedGradient )
   {
      t->PartialGradFlags = new int[ProductDomain().Number()];
      for( i=0;i<ProductDomain().Number();i++ )
         t->PartialGradFlags[i] = 0;
      t->SavedGradient = ProductDomain().Member();
   }
   HCL_ProductVector_d & g = (HCL_ProductVector_d&)(*SavedGradient);
   for( i=1;i<=ProductDomain().Number();i++ )
      if( !PartialGradFlags[i-1] )
      {
         PartialGradient( i,g(i) );
         t->PartialGradFlags[i-1] = 1;
      }

   return *SavedGradient;
}

const HCL_Vector_d & HCL_EvalFunctionalProductDomain_d::PartialGradientRef(
                                                      int i ) const
{
   // Casting away const
   HCL_EvalFunctionalProductDomain_d * t =
                    (HCL_EvalFunctionalProductDomain_d*)this;
   if( !SavedGradient )
   {
      t->PartialGradFlags = new int[ProductDomain().Number()];
      int j;
      for( j=0;j<ProductDomain().Number();j++ )
         t->PartialGradFlags[j] = 0;
      t->SavedGradient = ProductDomain().Member();
   }
   HCL_ProductVector_d & g = (HCL_ProductVector_d&)(*SavedGradient);
   if( !PartialGradFlags[i-1] )
      PartialGradient( i,g(i) );

   return g(i);
}

HCL_LinearOp_d * HCL_EvalFunctionalProductDomain_d::Hessian() const
{
   HCL_BlockLinearOp_d * HH = new HCL_BlockLinearOp_d( &(ProductDomain()),
                                                       &(ProductDomain()) );
   HCL_BlockLinearOp_d & H = (HCL_BlockLinearOp_d&)(*HH);
   int i,j;
   for( i=1;i<=ProductDomain().Number();i++ )
      for( j=1;j<=ProductDomain().Number();j++ )
      {
         HCL_LinearOp_d * L = HessianBlock( i,j );
         H.Set( i,j,L );
         HCL_delete( L );
      }
   return HH;
}

const HCL_LinearOp_d & HCL_EvalFunctionalProductDomain_d::HessianRef() const
{
   // Casting away const
   HCL_EvalFunctionalProductDomain_d * t =
                    (HCL_EvalFunctionalProductDomain_d*)this;
   if( !SavedHessian )
      t->SavedHessian = new HCL_BlockLinearOp_d( &(ProductDomain()),
                                                 &(ProductDomain()) );
   int i,j;
   HCL_BlockLinearOp_d & H = (HCL_BlockLinearOp_d&)(*SavedHessian);
   for( i=1;i<=ProductDomain().Number();i++ )
      for( j=1;j<=ProductDomain().Number();j++ )
         if( !H.IsSet( i,j ) )
         {
            HCL_LinearOp_d * L = HessianBlock( i,j );
            H.Set( i,j,L );
            HCL_delete( L );
         }
   return *SavedHessian;      
}

HCL_LinearOp_d * HCL_EvalFunctionalProductDomain_d::HessianBlock(
                                   int i,int j ) const
{
   return new HCL_FunctionalDefaultHessianBlock_d( i,j,this );
}

const HCL_LinearOp_d &
      HCL_EvalFunctionalProductDomain_d::HessianBlockRef(int i,int j ) const
{
   // Casting away const
   HCL_EvalFunctionalProductDomain_d * t =
                    (HCL_EvalFunctionalProductDomain_d*)this;
   if( !SavedHessian )
      t->SavedHessian = new HCL_BlockLinearOp_d( &(ProductDomain()),
                                                 &(ProductDomain()) );
   HCL_BlockLinearOp_d & H = (HCL_BlockLinearOp_d&)(*SavedHessian);
   if( !H.IsSet( i,j ) )
   {
      HCL_LinearOp_d * L = HessianBlock( i,j );
      H.Set( i,j,L );
      HCL_delete( L );
   }
   return H(i,j);      
}

double HCL_FunctionalProductDomain_d::Value1( const HCL_Vector_d & x ) const
{
   cerr << "Error: HCL_FunctionalProductDomain_d::Value1 is not implemented"
        << endl;
   exit(1);
   return 0;
}

void HCL_FunctionalProductDomain_d::PartialGradient1( int i,
                                    const HCL_Vector_d & x,
                                          HCL_Vector_d & g ) const
{
   cerr << "Error: HCL_FunctionalProductDomain_d::PartialGradient1 is "
           "not implemented" << endl;
   exit(1);
}

void HCL_FunctionalProductDomain_d::HessianBlockImage( int i,int j,
                                    const HCL_Vector_d & x,
                                    const HCL_Vector_d & dxj,
                                          HCL_Vector_d & dxi ) const
{
   cerr << "Error: HCL_FunctionalProductDomain_d::HessianBlockImage "
           "is not implemented" << endl;
   exit(1);
}

HCL_LinearOp_d * HCL_FunctionalProductDomain_d::HessianBlock1(int i,int j,
                     const HCL_Vector_d & x ) const
{
   return (HCL_LinearOp_d*)NULL;
}

HCL_FunctionalProductDomain_d::HCL_FunctionalProductDomain_d()
{
}

HCL_FunctionalProductDomain_d::HCL_FunctionalProductDomain_d(
                  const HCL_FunctionalProductDomain_d & )
{
}

HCL_FunctionalProductDomain_d::~HCL_FunctionalProductDomain_d()
{
}

HCL_ProductSpace_d & HCL_FunctionalProductDomain_d::ProductDomain() const
{
   return (HCL_ProductSpace_d&)Domain();
}

HCL_EvaluateFunctional_d * HCL_FunctionalProductDomain_d::Evaluate(
                           const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::Evaluate: input vector "
              "is not in domain" << endl;
      cerr << "Domain:" << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector:" << endl;
      x.Space().Write( cerr );
      exit(1);
   }
   return new HCL_FunctionalProductDomainDefaultEval_d( this,x );
}

void HCL_FunctionalProductDomain_d::PartialGradient( int i,
          const HCL_Vector_d & x,HCL_Vector_d & g ) const
{
   HCL_EvalFunctionalProductDomain_d * eval = 
           (HCL_EvalFunctionalProductDomain_d*)Evaluate( x );
   eval->PartialGradient( i,g );
   HCL_delete( eval );
}

HCL_LinearOp_d * HCL_FunctionalProductDomain_d::HessianBlock( int i,int j,
                               const HCL_Vector_d & x ) const
{
   HCL_EvalFunctionalProductDomain_d * eval = 
           (HCL_EvalFunctionalProductDomain_d*)Evaluate( x );
   HCL_LinearOp_d * h = eval->HessianBlock( i,j );
   HCL_delete( eval );
   return h;
}

// This method checks the analytic second derivatives against the
// second order finite-difference formula
//
//     \[D_{ij}f(x)(\delta x_j,\delta x_i) \doteq 
//          f(x+\tilde{\delta x_i}+\tilde{\delta x_j})+
//          f(x-\tilde{\delta x_i}-\tilde{\delta x_j})-
//          f(x+\tilde{\delta x_i})-
//          f(x-\tilde{\delta x_i})-
//          f(x+\tilde{\delta x_j})-
//          f(x-\tilde{\delta x_j})+2f(x)\].
//
//  Here $\tilde{\delta x_i}$ is the vector in the domain of $f$ with every
//  component equal to zero except the $i$th, which equals $\delta x_i$.

int HCL_FunctionalProductDomain_d::CheckHessianBlock( const HCL_Vector_d & y,
                             const HCL_Vector_d & pi,
                             const HCL_Vector_d & pj, int i,int j,
                             ostream & str,int n,double hmin,
                             double hmax ) const
{
   if( i==j )
      return CheckHessianBlock( y,pi,i,str,n,hmin,hmax );
      
   if( i < 1 || i > ProductDomain().Number() || j < 1 ||
       j > ProductDomain().Number() )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::"
              "CheckHessianBlock:" << endl
           << "   invalid index" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::"
              "CheckHessianBlock:" << endl
           << "   vector y is not in Domain" << endl;
      exit(1);
   }
   if( pi.Space() != ProductDomain()(i) )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::"
              "CheckHessianBlock:" << endl
           << "   vector pi is not in Domain" << endl;
      exit(1);
   }
   if( pj.Space() != ProductDomain()(j) )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::"
              "CheckHessianBlock:" << endl
           << "   vector pj is not in Domain" << endl;
      exit(1);
   }

   if( hmax <= hmin )
   {
      double temp = hmax;
      hmax = hmin;
      hmin = temp;
   }
   if( hmin <= 0.0 )
   {
      hmin = 0.1;
      hmax = 1.0;
   }
   if( n <= 0 )
      n = 10;

   // There are now six directions to check: pi+pj,-pi-pj,pi,-pi,pj,-pj
   HCL_ProductVector_d * yp = (HCL_ProductVector_d*)ProductDomain().Member();
   yp->Zero();
   (*yp)(i).Copy( pi );
   (*yp)(j).Copy( pj );
   double hlimit = MaxStep( y,*yp );
   yp->Neg();
   hlimit = min( MaxStep( y,*yp ), hlimit );
   yp->Zero();
   (*yp)(i).Copy( pi );
   hlimit = min( MaxStep( y,*yp ), hlimit );
   yp->Neg();
   hlimit = min( MaxStep( y,*yp ), hlimit );
   yp->Zero();
   (*yp)(j).Copy( pj );
   hlimit = min( MaxStep( y,*yp ), hlimit );
   yp->Neg();
   hlimit = min( MaxStep( y,*yp ), hlimit );
   HCL_delete( yp );

   if( hlimit <= 0.0 )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::"
              "CheckHessianBlock:" << endl
           << "   direction is not feasible" << endl;
      exit(1);
   }
   if( hmax >= hlimit )
   {
      hmax = 0.99*hlimit;
      if( hmin >= hmax )
         hmin = hmax/n;
   }

   HCL_ProductVector_d * x1 = (HCL_ProductVector_d*)(Domain().Member());
   HCL_Vector_d * pi1 = ProductDomain()(i).Member();
   double v0,dv;

   v0 = Value( y );
   HCL_LinearOp_d * D2fy = HessianBlock( i,j,y );
   D2fy->Image( pj,*pi1 );
   HCL_delete( D2fy );
   dv = pi.Inner( *pi1 );
   HCL_delete( pi1 );

#if SINGLE_PRECISION
   int nd = 8;
#else
   int nd = 16;
#endif
   int oldprecision = str.precision( nd );

   double dvmag = abs(dv);
   int rflag = 1;
   HCL_FPArch_d arch;
   if( dvmag < arch.sfmin )
   {
      rflag = 0;
      str << "HCL_FunctionalProductDomain_d::CheckHessianBlock: norm of "
             "second variation is too "
          << endl << "small; displaying absolute error" << endl;
   }

   if( rflag )
      str << setw(5) << "h" << setw(nd+6) << "2nd variation" << setw(nd+6)
           << "cent. diff." << setw(nd+4) << "Rel. Err." << endl;
   else
      str << setw(5) << "h" << setw(nd+6) << "2nd variation" << setw(nd+6)
           << "cent. diff." << setw(nd+4) << "Error" << endl;
   int ii;
   double hstep = (hmax-hmin)/(n-1);
   for( ii=n-1;ii>=0;ii-- )
   {
      double v = 2.0*v0;
      double h = hmin+ii*hstep;
      x1->Copy( y );
      (*x1)(i).AddScale( h,pi );
      (*x1)(j).AddScale( h,pj );
      v += Value( *x1 );

      x1->Copy( y );
      (*x1)(i).AddScale( -h,pi );
      (*x1)(j).AddScale( -h,pj );
      v += Value( *x1 );

      x1->Copy( y );
      (*x1)(i).AddScale( h,pi );
      v -= Value( *x1 );

      x1->Copy( y );
      (*x1)(i).AddScale( -h,pi );
      v -= Value( *x1 );

      x1->Copy( y );
      (*x1)(j).AddScale( h,pj );
      v -= Value( *x1 );

      x1->Copy( y );
      (*x1)(j).AddScale( -h,pj );
      v -= Value( *x1 );

      v /= (2.0*h*h);
      double n1 = abs(v-dv);

      if( rflag )
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << dv << " " << setw(nd+6)
             << v << " " << setw(nd+6) << n1/dvmag << endl;
      else
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << dv << " " << setw(nd+6)
             << v << " " << setw(nd+6) << n1 << endl;
   }
   str.precision( oldprecision );
   return 0;
}

int HCL_FunctionalProductDomain_d::CheckHessianBlock( const HCL_Vector_d & y,
                             const HCL_Vector_d & pi,
                             int i,ostream & str,int n,double hmin,
                             double hmax ) const
{
   if( i < 1 || i > ProductDomain().Number() )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::CheckHessianBlock: "
               "invalid index" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::CheckHessianBlock: "
               "vector y is not in Domain" << endl;
      exit(1);
   }
   if( pi.Space() != ProductDomain()(i) )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::CheckHessianBlock: "
              "vector pi is not in Domain" << endl;
      exit(1);
   }

   if( hmax <= hmin )
   {
      double temp = hmax;
      hmax = hmin;
      hmin = temp;
   }
   if( hmin <= 0.0 )
   {
      hmin = 0.1;
      hmax = 1.0;
   }
   if( n <= 0 )
      n = 10;

   HCL_ProductVector_d * yp = (HCL_ProductVector_d*)ProductDomain().Member();
   yp->Zero();
   (*yp)(i).Copy( pi );
   double hlimit = MaxStep( y,*yp );
   yp->Neg();
   hlimit = min( MaxStep( y,*yp ), hlimit );
   HCL_delete( yp );

   if( hlimit <= 0.0 )
   {
      cerr << "Error in HCL_FunctionalProductDomain_d::CheckHessianBlock: "
              "direction is not feasible" << endl;
      exit(1);
   }
   if( hmax >= hlimit )
   {
      hmax = 0.99*hlimit;
      if( hmin >= hmax )
         hmin = hmax/n;
   }

   HCL_ProductVector_d * x1 = (HCL_ProductVector_d*)(Domain().Member());
   HCL_Vector_d * pi1 = ProductDomain()(i).Member();
   double v0,dv;

   v0 = Value( y );
   HCL_LinearOp_d * D2fy = HessianBlock( i,i,y );
   D2fy->Image( pi,*pi1 );
   dv = pi.Inner( *pi1 );
   HCL_delete( pi1 );
   HCL_delete( D2fy );

#if SINGLE_PRECISION
   int nd = 8;
#else
   int nd = 16;
#endif
   int oldprecision = str.precision( nd );

   double dvmag = abs(dv);
   int rflag = 1;
   HCL_FPArch_d arch;
   if( dvmag < arch.sfmin )
   {
      rflag = 0;
      str << "HCL_FunctionalProductDomain_d::CheckHessianBlock: norm of "
             "second variation is too "
          << endl << "small; displaying absolute error" << endl;
   }

   if( rflag )
      str << setw(5) << "h" << setw(nd+6) << "2nd variation" << setw(nd+6)
           << "cent. diff." << setw(nd+4) << "Rel. Err." << endl;
   else
      str << setw(5) << "h" << setw(nd+6) << "2nd variation" << setw(nd+6)
           << "cent. diff." << setw(nd+4) << "Error" << endl;
   double hstep = (hmax-hmin)/(n-1);
   int ii;
   for( ii=n-1;ii>=0;ii-- )
   {
      double v = -2.0*v0;
      double h = hmin+ii*hstep;
      x1->Copy( y );
      (*x1)(i).AddScale( h,pi );
      v += Value( *x1 );

      x1->Copy( y );
      (*x1)(i).AddScale( -h,pi );
      v += Value( *x1 );

      v /= (h*h);
      double n1 = abs(v-dv);

      if( rflag )
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << dv << " " << setw(nd+6)
             << v << " " << setw(nd+6) << n1/dvmag << endl;
      else
         str << setprecision(6) << setw(8) << h << " " << setprecision(nd)
             << setw(nd+6) << dv << " " << setw(nd+6)
             << v << " " << setw(nd+6) << n1 << endl;
   }
   str.precision( oldprecision );
   return 0;
}

int HCL_FunctionalProductDomain_d::CheckHessianBlockAdj(
                             const HCL_Vector_d & y,
                             const HCL_Vector_d & pi,
                             const HCL_Vector_d & pj, int i,int j,
                             ostream & str,int n,double hmin,
                             double hmax ) const
{
   CheckHessianBlock( y,pi,pj,i,j,str,n,hmin,hmax );
   HCL_LinearOp_d * L = HessianBlock( i,j,y );
   return L->CheckAdj( 2 );
}

HCL_FunctionalProductDomainDefaultEval_d::HCL_FunctionalProductDomainDefaultEval_d()
{
   cerr << "Error: HCL_FunctionalProductDomainDefaultEval_d default constructor "
           "not implemented" << endl;
   exit(1);
}

HCL_FunctionalProductDomainDefaultEval_d::HCL_FunctionalProductDomainDefaultEval_d(
                           const HCL_FunctionalProductDomainDefaultEval_d & )
{
   cerr << "Error: HCL_FunctionalProductDomainDefaultEval_d copy constructor "
           "not implemented" << endl;
   exit(1);
}

HCL_FunctionalProductDomainDefaultEval_d::HCL_FunctionalProductDomainDefaultEval_d(
                           const HCL_FunctionalProductDomain_d * ff,
                           const HCL_Vector_d & in )
{
   f = ff;f->IncCount();
   x = in.Space().Member();
   x->Copy( in );
}

HCL_FunctionalProductDomainDefaultEval_d::~HCL_FunctionalProductDomainDefaultEval_d()
{
   HCL_delete( x );
   HCL_delete( (HCL_FunctionalProductDomain_d*)f );
}

double HCL_FunctionalProductDomainDefaultEval_d::Value() const
{
   return f->Value1( *x );
}

void HCL_FunctionalProductDomainDefaultEval_d::PartialGradient( int i,
                   HCL_Vector_d & y ) const
{
   f->PartialGradient1( i,*x,y );
}

HCL_LinearOp_d *
HCL_FunctionalProductDomainDefaultEval_d::HessianBlock( int i,int j ) const
{
   HCL_LinearOp_d * H = f->HessianBlock1( i,j,*x );
   if( H )
      return H;
   else
      return new HCL_FunctionalDefaultHessianBlock_d( i,j,this );
}

ostream &
HCL_FunctionalProductDomainDefaultEval_d::Write( ostream & str ) const
{
   str << "HCL_FunctionalProductDomainDefaultEval_d based on the following "
          "functional" << endl;
   return f->Write( str );
}

HCL_FunctionalDefaultHessianBlock_d::HCL_FunctionalDefaultHessianBlock_d()
{
   cerr << "Error: HCL_FunctionalDefaultHessianBlock_d default "
           "constructor not implemented"
        << endl;
   exit(1);
}

HCL_FunctionalDefaultHessianBlock_d::HCL_FunctionalDefaultHessianBlock_d(
              const HCL_FunctionalDefaultHessianBlock_d & L )
{
   cerr << "Error: HCL_FunctionalDefaultHessianBlock_d copy "
           "constructor not implemented"
        << endl;
   exit(1);
}

HCL_FunctionalDefaultHessianBlock_d::HCL_FunctionalDefaultHessianBlock_d(
              int ii,int jj,const HCL_EvalFunctionalProductDomain_d * ev )
{
   int m = ev->ProductDomain().Number();
   if( ii<1 || ii>m || jj<1 || jj>m )
   {
      cerr << "Error: invalid block indices in " << endl
           << "HCL_FunctionalDefaultHessianBlock_d constructor" << endl;
      cerr << "(" << i << "," << j << ") was passed; both indices " << endl
           << "must be beteen 1 and " << m << endl;
      exit(1);
   }
   i = ii;
   j = jj;
   eval = ev;ev->IncCount();
}

HCL_FunctionalDefaultHessianBlock_d::~HCL_FunctionalDefaultHessianBlock_d()
{
   HCL_delete( (HCL_EvalFunctionalProductDomain_d*)eval );
}

HCL_VectorSpace_d & HCL_FunctionalDefaultHessianBlock_d::Domain() const
{
   return eval->ProductDomain()(j);
}

HCL_VectorSpace_d & HCL_FunctionalDefaultHessianBlock_d::Range() const
{
   return eval->ProductDomain()(i);
}

void HCL_FunctionalDefaultHessianBlock_d::Image( const HCL_Vector_d & x,
                                                  HCL_Vector_d & y ) const
{
   eval->HessianBlockImage( i,j,x,y );
}

void HCL_FunctionalDefaultHessianBlock_d::AdjImage( const HCL_Vector_d & y,
                                                    HCL_Vector_d & x ) const
{
   eval->HessianBlockImage( j,i,y,x );
}

void HCL_FunctionalDefaultHessianBlock_d::InvImage(const HCL_Vector_d & y,
                                                    HCL_Vector_d & x) const
{
   cerr << "Error: HCL_FunctionalDefaultHessianBlock_d::InvImage "
           "is not implemented" << endl;
   exit(1);
}

void HCL_FunctionalDefaultHessianBlock_d::InvAdjImage(const HCL_Vector_d & x,
                                                 HCL_Vector_d & y) const
{
   cerr << "Error: HCL_FunctionalDefaultHessianBlock_d::InvAdjImage "
           "is not implemented" << endl;
   exit(1);
}

ostream & HCL_FunctionalDefaultHessianBlock_d::Write( ostream & str ) const
{
   str << "HCL_FunctionalDefaultHessianBlock_d ( " << i << "," << j
       << ")" << endl << "based on the following evaluation object"
       << endl;
   return eval->Write( str );
}

HCL_EvalRestrictedFunctional_d::HCL_EvalRestrictedFunctional_d()
{
   cerr << "Error: HCL_EvalRestrictedFunctional_d default constructor "
           "called" << endl;
   exit(1);
}

HCL_EvalRestrictedFunctional_d::HCL_EvalRestrictedFunctional_d(
                                  const HCL_EvalRestrictedFunctional_d & )
{
   cerr << "Error: HCL_EvalRestrictedFunctional_d copy constructor "
           "called" << endl;
   exit(1);
}

HCL_EvalRestrictedFunctional_d::HCL_EvalRestrictedFunctional_d( int ii,
                        HCL_EvalFunctionalProductDomain_d * ev )
{
   int n = (ev->ProductDomain()).Number();
   if( ii < 1 || ii > n )
   {
      cerr << "Error in HCL_EvalRestrictedFunctional_d constructor: "
              "index must be between 1 and " << n << endl;
      exit(1);
   }
   i = ii;
   eval = ev;ev->IncCount();
}

HCL_EvalRestrictedFunctional_d::~HCL_EvalRestrictedFunctional_d()
{
   HCL_delete( eval );
}

HCL_VectorSpace_d & HCL_EvalRestrictedFunctional_d::Domain() const
{
   return (eval->ProductDomain())(i);
}

HCL_EvalFunctionalProductDomain_d &
            HCL_EvalRestrictedFunctional_d::UnrestrictedEval()
{
      return *eval;
}

double HCL_EvalRestrictedFunctional_d::Value() const
{
   return eval->ValueRef();
}

void HCL_EvalRestrictedFunctional_d::Gradient( HCL_Vector_d & g ) const
{
   if( g.Space() != Domain() )
   {
      cerr << "Error in HCL_EvalRestrictedFunctional_d::Gradient: "
              "vector is not in domain" << endl;
      exit(1);
   }
   g.Copy( eval->PartialGradientRef( i ) );
}

HCL_LinearOp_d * HCL_EvalRestrictedFunctional_d::Hessian() const
{
   HCL_LinearOp_d * t = (HCL_LinearOp_d*)(&(eval->HessianBlockRef(i,i)));
   t->IncCount();
   return t;
}

ostream & HCL_EvalRestrictedFunctional_d::Write( ostream & str ) const
{
   str << "HCL_EvalRestrictedFunctional_d object (i = " << i << ")"
       << endl;
   str << "Underlying HCL_EvalFunctionalProductDomain_d object: " << endl;
   return eval->Write( str );
}

HCL_RestrictedFunctional_d::HCL_RestrictedFunctional_d()
{
   cerr << "Error: HCL_RestrictedFunctional_d default constructor "
           "called" << endl;
   exit(1);
}

HCL_RestrictedFunctional_d::HCL_RestrictedFunctional_d(
                                  const HCL_RestrictedFunctional_d & )
{
   cerr << "Error: HCL_RestrictedFunctional_d copy constructor "
           "called" << endl;
   exit(1);
}

HCL_RestrictedFunctional_d::HCL_RestrictedFunctional_d( 
                        HCL_FunctionalProductDomain_d * ff,
                        const HCL_ProductVector_d & x,int ii )
{
   int n = ff->ProductDomain().Number();
   if( ii < 1 || ii > n )
   {
      cerr << "Error in HCL_RestrictedFunctional_d constructor: "
              "index must be between 1 and " << n << endl;
      exit(1);
   }
   i = ii;
   f = ff;f->IncCount();
   if( x.Space() != f->Domain() )
   {
      cerr << "Error in HCL_RestrictedFunctional_d constructor: "
              "vector is not in domain " << n << endl;
      exit(1);
   }
   y = (HCL_ProductVector_d*)f->Domain().Member();
   y->Copy( x );
}
HCL_RestrictedFunctional_d::HCL_RestrictedFunctional_d( 
                        HCL_FunctionalProductDomain_d * ff,
                        HCL_ProductVector_d * x,int ii )
{
   int n = ff->ProductDomain().Number();
   if( ii < 1 || ii > n )
   {
      cerr << "Error in HCL_RestrictedFunctional_d constructor: "
              "index must be between 1 and " << n << endl;
      exit(1);
   }
   i = ii;
   f = ff;f->IncCount();
   if( x->Space() != f->Domain() )
   {
      cerr << "Error in HCL_RestrictedFunctional_d constructor: "
              "vector is not in domain " << n << endl;
      exit(1);
   }
   y = x;x->IncCount();
}

HCL_RestrictedFunctional_d::~HCL_RestrictedFunctional_d()
{
   HCL_delete( y );
   HCL_delete( f );
}

HCL_VectorSpace_d & HCL_RestrictedFunctional_d::Domain() const
{
   return (f->ProductDomain())(i);
}

HCL_EvaluateFunctional_d * HCL_RestrictedFunctional_d::Evaluate(
                               const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_RestrictedFunctional_d::Evaluate: "
              "vector is not in domain" << endl;
      exit(1);
   }
   (*y)(i).Copy( x );
   HCL_EvalFunctionalProductDomain_d * ev =
        (HCL_EvalFunctionalProductDomain_d*)f->Evaluate( *y );
   HCL_EvaluateFunctional_d * eval = 
               new HCL_EvalRestrictedFunctional_d( i,ev );
   HCL_delete( ev );
   return eval;
}
   
ostream & HCL_RestrictedFunctional_d::Write( ostream & str ) const
{
   str << "HCL_RestrictedFunctional_d object (i = " << i << ")"
       << endl;
   str << "Underlying HCL_FunctionalProductDomain_d: " << endl;
   return f->Write( str );
}
