/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_GenericProductVector_d.h"

HCL_GenericProductSpace_d::HCL_GenericProductSpace_d()
{
   cerr << "Error: HCL_GenericProductSpace_d default constructor called"
        << endl;
   exit(1);
}

HCL_GenericProductSpace_d::HCL_GenericProductSpace_d(
                               const HCL_GenericProductSpace_d & x)
{
   len = x.Number();
   int i;
   SpaceList = new HCL_VectorSpace_d*[len];
   for( i=0;i<len;i++ )
      SpaceList[i] = x(i+1).Replicate();
}

HCL_GenericProductSpace_d::HCL_GenericProductSpace_d( int n )
{
   len = n;
   SpaceList = new HCL_VectorSpace_d*[n];
   int i;
   for( i=0;i<len;i++ )
      SpaceList[i] = NULL;
}

HCL_GenericProductSpace_d::HCL_GenericProductSpace_d( 
          const HCL_VectorSpace_d & S1,const HCL_VectorSpace_d & S2 )
{
   len = 2;
   SpaceList = new HCL_VectorSpace_d*[2];
   SpaceList[0] = S1.Replicate();
   SpaceList[1] = S2.Replicate();
}

HCL_GenericProductSpace_d::HCL_GenericProductSpace_d( HCL_VectorSpace_d * S1,
                                                    HCL_VectorSpace_d * S2 )
{
   len = 2;
   SpaceList = new HCL_VectorSpace_d*[2];
   SpaceList[0] = S1;S1->IncCount();
   SpaceList[1] = S2;S2->IncCount();
}

HCL_GenericProductSpace_d::HCL_GenericProductSpace_d(
                              const HCL_VectorSpace_d & S,int n )
{
   len = n;
   SpaceList = new HCL_VectorSpace_d*[n];
   SpaceList[0] = S.Replicate();
   int i;
   for( i=1;i<n;i++ )
   {
      SpaceList[i] = SpaceList[0];
      SpaceList[0]->IncCount();
   }
}

HCL_GenericProductSpace_d::HCL_GenericProductSpace_d(
                              HCL_VectorSpace_d * S,int n )
{
   len = n;
   SpaceList = new HCL_VectorSpace_d*[n];
   int i;
   for( i=0;i<n;i++ )
   {
      SpaceList[i] = S;
      S->IncCount();
   }
}

HCL_GenericProductSpace_d::~HCL_GenericProductSpace_d()
{
   for( int i=0;i<len;i++ )
      HCL_delete( SpaceList[i] );
   delete[] SpaceList;
}

int HCL_GenericProductSpace_d::Initialized() const
{
   int i;
   for( i=0;i<len;i++ )
      if( !SpaceList[i] )
         return 0;
   return 1;
}

int HCL_GenericProductSpace_d::IsSet( int i ) const
{
   if( i < 1 || i > len )
   {
      cerr << "Error in HCL_GenericProductSpace_d::IsSet: invalid index"
           << endl;
      exit(1);
   }
   if( SpaceList[i-1] )
      return 1;
   else
      return 0;
}

HCL_VectorSpace_d & HCL_GenericProductSpace_d::operator()( int i ) const
{
   if( i < 1 || i > len )
   {
      cerr << "Error in HCL_GenericProductSpace_d::IsSet: invalid index"
           << endl;
      exit(1);
   }
   if( !SpaceList[i-1] )
   {
      cerr << "Error in HCL_GenericProductSpace_d::operator(): attempt to "
              "access an uninitialized factor" << endl;
      exit(1);
   }
   return *SpaceList[i-1];
}

HCL_Vector_d * HCL_GenericProductSpace_d::Member() const
{
   if( !Initialized() )
   {
      cerr << "Error in HCL_GenericProductSpace_d::Member: attempt to use "
           << "uninitialized object" << endl;
      exit(1);
   }
   return new HCL_GenericProductVector_d( (HCL_GenericProductSpace_d*)this );
}

HCL_VectorSpace_d * HCL_GenericProductSpace_d::Replicate() const
{
   HCL_GenericProductSpace_d * t = (HCL_GenericProductSpace_d*)this;
   t->IncCount();
   return t;
}

void HCL_GenericProductSpace_d::SetNext( HCL_VectorSpace_d * S )
{
   if( Initialized() )
   {
      cerr << "Error in HCL_GenericProductSpace_d::SetNext: too many factors"
           << endl;
      exit(1);
   }
   int i;
   for( i=0;i<len;i++ )
      if( !SpaceList[i] )
      {
         SpaceList[i] = S;
         S->IncCount();
         return;
      }
}

void HCL_GenericProductSpace_d::Set( int i,HCL_VectorSpace_d * S )
{
   if( SpaceList[i-1] )
      HCL_delete( SpaceList[i-1] );
   SpaceList[i-1] = S;
   S->IncCount();
}

ostream & HCL_GenericProductSpace_d::Write( ostream & str ) const
{
   str << "   HCL_GenericProductSpace_d with factors" << endl;
   for(int i=1; i <= Number(); i++)
      if( SpaceList[i-1] )
         (*this)(i).Write(str);
      else
         str << "(This factor not set)" << endl;
   return str;
}

HCL_GenericProductVector_d::HCL_GenericProductVector_d()
{
}

HCL_GenericProductVector_d::HCL_GenericProductVector_d(const HCL_GenericProductVector_d & p)
{
   cerr << "Error: HCL_GenericProductVector_d copy constructor called"
        << endl;
   exit(1);
}

HCL_GenericProductVector_d::HCL_GenericProductVector_d(const HCL_GenericProductSpace_d & p)
{
   sp = new HCL_GenericProductSpace_d( p );
   int i;
   len = p.Number();
   vlist = new HCL_Vector_d*[len];
   for( i=0;i<len;i++ )
      vlist[i] = (*sp)(i+1).Member();
}

HCL_GenericProductVector_d::HCL_GenericProductVector_d( HCL_GenericProductSpace_d * p )
{
   sp = p;sp->IncCount();
   len = sp->Number();
   int i;
   vlist = new HCL_Vector_d*[len];
   for( i=0;i<len;i++ )
      vlist[i] = (*sp)(i+1).Member();
}

HCL_GenericProductVector_d::HCL_GenericProductVector_d( HCL_Vector_d * v1,HCL_Vector_d * v2 )
{
   len = 2;
   sp = new HCL_GenericProductSpace_d( &(v1->Space()),&(v2->Space()) );
   vlist = new HCL_Vector_d*[2];
   vlist[0] = v1;v1->IncCount();
   vlist[1] = v2;v2->IncCount();
}

HCL_GenericProductVector_d::HCL_GenericProductVector_d( int n )
{
   len = n;
   sp = new HCL_GenericProductSpace_d( n );
   vlist = new HCL_Vector_d*[n];
   int i;
   for( i=0;i<len;i++ ) vlist[i] = NULL;
}

HCL_GenericProductVector_d::~HCL_GenericProductVector_d()
{
   for( int i=0;i<len;i++ )
      HCL_delete( vlist[i] );
   delete[] vlist;
   HCL_delete( sp );
}

int HCL_GenericProductVector_d::Initialized() const
{
   int i;
   for( i=0;i<len;i++ )
      if( !vlist[i] )
         return 0;
   return 1;
}

int HCL_GenericProductVector_d::IsSet( int i ) const
{
   if( i < 1 || i > len )
   {
      cerr << "Error in HCL_GenericProductVector_d::IsSet: invalid index"
           << endl;
      exit(1);
   }
   if( vlist[i-1] )
      return 1;
   else
      return 0;
}

void HCL_GenericProductVector_d::SetNext( HCL_Vector_d * x )
{
   if( Initialized() )
   {
      cerr << "Error in HCL_GenericProductVector_d::SetNext: too many factors"
           << endl;
      exit(1);
   }
   int i;
   for( i=0;i<len;i++ )
      if( !vlist[i] )
      {
         vlist[i] = x;
         x->IncCount();
         sp->Set( i+1,&(x->Space()) );
         return;
      }
}

void HCL_GenericProductVector_d::Set( int i,HCL_Vector_d * S )
{
   if( vlist[i-1] )
      HCL_delete( vlist[i-1] );
   vlist[i-1] = S;
   S->IncCount();
   sp->Set( i,&(S->Space()) );
}
