/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_LinCombFcnl_d.h"

HCL_EvalLinCombFcnl_d::HCL_EvalLinCombFcnl_d()
{
   cerr << "Error: HCL_EvalLinCombFcnl_d default constructor called"
        << endl;
   exit(1);
}

HCL_EvalLinCombFcnl_d::HCL_EvalLinCombFcnl_d( 
                                 const HCL_EvalLinCombFcnl_d & )
{
   cerr << "Error: HCL_EvalLinCombFcnl_d copy constructor called"
        << endl;
   exit(1);
}

HCL_EvalLinCombFcnl_d::HCL_EvalLinCombFcnl_d( HCL_VectorSpace_d * d,int n )
{
   num = n;
   numinit = 0;
   dom = d;d->IncCount();
   evallist = new HCL_EvaluateFunctional_d*[num];
   weights = new double[num];
}

HCL_EvalLinCombFcnl_d::~HCL_EvalLinCombFcnl_d()
{
   delete[] weights;
   for( int i=0;i<numinit;i++ )
      HCL_delete( evallist[i] );
   delete[] evallist;
}

double HCL_EvalLinCombFcnl_d::Value() const
{
   if( numinit < num )
   {
      cerr << "Error in HCL_EvalLinCombFcnl_d::Value: not initialized"
           << endl;
      exit(1);
   }
   double s=0.0;
   for( int i=0;i<num;i++ )
      s += weights[i]*(evallist[i]->Value());
   return s;
}

void HCL_EvalLinCombFcnl_d::Gradient( HCL_Vector_d & g ) const
{
   if( numinit < num )
   {
      cerr << "Error in HCL_EvalLinCombFcnl_d::Gradient: not initialized"
           << endl;
      exit(1);
   }
   evallist[0]->Gradient( g );
   g.Mul( weights[0] );
   HCL_Vector_d * temp = Domain().Member();
   for( int i=1;i<numinit;i++ )
   {
      evallist[i]->Gradient( *temp );
      g.AddScale( weights[i],*temp );
   }
   HCL_delete( temp );
}

HCL_LinearOp_d * HCL_EvalLinCombFcnl_d::Hessian() const
{
   if( numinit < num )
   {
      cerr << "Error in HCL_EvalLinCombFcnl_d::Hessian: not initialized"
           << endl;
      exit(1);
   }
   HCL_LinCombLinearOp_d * temp = new HCL_LinCombLinearOp_d( dom,dom,num );
   for( int i=0;i<numinit;i++ )
   {
      HCL_LinearOp_d * thess = evallist[i]->Hessian();
      temp->SetNext( thess,weights[i] );
      HCL_delete( thess );
   }
   return temp;
}

void HCL_EvalLinCombFcnl_d::SetNext( HCL_EvaluateFunctional_d * eval,
                                         double w )
{
   if( numinit == num )
   {
      cerr << "Error in HCL_EvalLinCombFcnl_d::SetNext: "
              "already initialized" << endl;
      exit(1);
   }
   evallist[numinit] = eval;eval->IncCount();
   weights[numinit] = w;
   numinit++;
}

const HCL_EvaluateFunctional_d & HCL_EvalLinCombFcnl_d::
                                     GetTerm( int n ) const
{
   if( n < 1 || n > numinit )
   {
      cerr << "Error in HCL_EvalLinCombFcnl_d::GetTerm: invalid index"
           << endl;
      exit(1);
   }
   return *evallist[n-1];
}

ostream & HCL_EvalLinCombFcnl_d::Write( ostream & str ) const
{
   str << "HCL_EvalLinCombFcnl_d object; " << numinit << " terms out of "
       << num << " have been set" << endl;
   str << "The weights are " << endl << "   " << endl;
   int i;
   for( i=0;i<numinit;i++ ) str << weights[i] << " ";
   str << endl << endl;
   str << "The evaluation objects are " << endl;
   for( i=0;i<numinit;i++ ) evallist[i]->Write( str );
   return str;
}

HCL_LinCombFcnl_d::HCL_LinCombFcnl_d()
{
   cerr << "Error: HCL_LinCombFcnl_d default constructor called"
        << endl;
   exit(1);
}

HCL_LinCombFcnl_d::HCL_LinCombFcnl_d( const HCL_LinCombFcnl_d & )
{
   cerr << "Error: HCL_LinCombFcnl_d copy constructor called"
        << endl;
   exit(1);
}

HCL_LinCombFcnl_d::HCL_LinCombFcnl_d( HCL_VectorSpace_d * d,int n )
{
   dom = d;d->IncCount();
   num = n;
   numinit = 0;
   flist = new HCL_Functional_d*[num];
   ParamTable = new Table( NULL,NULL,num );
}

HCL_LinCombFcnl_d::HCL_LinCombFcnl_d( HCL_Functional_d * f1,
                 double w1,HCL_Functional_d * f2,double w2 )
{
   if( f1->Domain() != f2->Domain() )
   {
      cerr << "Error in HCL_LinCombFcnl_d constructor: domains must"
              " agree" << endl;
      exit(1);
   }

   dom = &(f1->Domain());dom->IncCount();
   num = 2;
   numinit = 2;
   flist = new HCL_Functional_d*[2];
   flist[0] = f1;f1->IncCount();
   flist[1] = f2;f2->IncCount();
   ParamTable = new Table;
   ParamTable->PutArrayValue( "Weights",1,w1 );
   ParamTable->PutArrayValue( "Weights",2,w2 );
}

HCL_LinCombFcnl_d::~HCL_LinCombFcnl_d()
{
   HCL_delete( ParamTable );
   for( int i=0;i<numinit;i++ )
      HCL_delete( flist[i] );
   delete[] flist;
   HCL_delete( dom );
}

double HCL_LinCombFcnl_d::MaxStep( const HCL_Vector_d & x, 
                                       const HCL_Vector_d &dir) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombFcnl_d::MaxStep: vector is "
              "not in domain" << endl;
      exit(1);
   }
   if( dir.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombFcnl_d::MaxStep: vector is "
              "not in domain" << endl;
      exit(1);
   }
   if( numinit < num )
   {
      cerr << "Error in HCL_LinCombFcnl_d::MaxStep: not initialized "
              "yet" << endl;
      exit(1);
   }

   double s = flist[0]->MaxStep( x,dir );
   int i;
   for( i=1;i<numinit;++i )
      s = min( s,flist[i]->MaxStep( x,dir ) );
   return s;
}

HCL_EvaluateFunctional_d * HCL_LinCombFcnl_d::Evaluate(
                               const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombFcnl_d::Evaluate: vector is "
              "not in domain" << endl;
      exit(1);
   }
   if( numinit < num )
   {
      cerr << "Error in HCL_LinCombFcnl_d::Evaluate: not initialized "
              "yet" << endl;
      exit(1);
   }

   HCL_EvalLinCombFcnl_d * eval = new HCL_EvalLinCombFcnl_d( dom,num );
   for( int i=0;i<num;i++ )
   {
      double w;
      ParamTable->GetArrayValue( "Weights",i+1,w );
      HCL_EvaluateFunctional_d * teval = flist[i]->Evaluate( x );
      eval->SetNext( teval,w );
      HCL_delete( teval );
   }
   return eval;
}

void HCL_LinCombFcnl_d::SetNext( HCL_Functional_d * f,double w )
{
   if( numinit == num )
   {
      cerr << "Error in HCL_LinCombFcnl_d::SetNext: already initialized"
           << endl;
      exit(1);
   }
   if( f->Domain() != Domain() )
   {
      cerr << "Error in HCL_LinCombFcnl_d::SetNext: domains do not agree"
           << endl;
      exit(1);
   }

   flist[numinit] = f;f->IncCount();
   numinit++;
   ParamTable->PutArrayValue( "Weights",numinit,w );
}

ostream & HCL_LinCombFcnl_d::Write( ostream & str ) const
{
   str << "HCL_LinCombFcnl_d object; " << numinit << " terms out of "
       << num << " have been set" << endl;
   str << "The weights are " << endl << "   " << endl;
   int i;
   for( i=0;i<numinit;i++ )
   {
      double w;
      ParamTable->GetArrayValue( "Weights",i+1,w );
      str << w << " ";
   }
   str << endl << endl;
   str << "The functionals are " << endl;
   for( i=0;i<numinit;i++ )
   {
      str << "Term " << i+1 << endl;
      flist[i]->Write( str );
   }
   return str;
}
