/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_LinCombLinearOp_d.h"

HCL_LinCombLinearOp_d::HCL_LinCombLinearOp_d()
{
   cerr << "Error: HCL_LinCombLinearOp_d default constructor called"
        << endl;
   exit(1);
}

HCL_LinCombLinearOp_d::HCL_LinCombLinearOp_d(
                          const HCL_LinCombLinearOp_d & L )
{
   cerr << "Error: HCL_LinCombLinearOp_d copy constructor called"
        << endl;
   exit(1);
}

HCL_LinCombLinearOp_d::HCL_LinCombLinearOp_d( HCL_VectorSpace_d * d,
                                              HCL_VectorSpace_d * r,int n )
{
   num = n;
   dom = d;d->IncCount();
   range = r;r->IncCount();
   ops = new HCL_LinearOp_d *[num];
   ParamTable = new Table( NULL,NULL,num );
   numinit = 0;
}

HCL_LinCombLinearOp_d::HCL_LinCombLinearOp_d(
                                       HCL_LinearOp_d * op1,double a1,
                                       HCL_LinearOp_d * op2,double a2 )
{
   if( op1->Domain() != op2->Domain() || op1->Range() != op2->Range() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d constructor: spaces do not "
              "agree" << endl;
      exit(1);
   }
   num = 2;
   numinit = 2;
   ops = new HCL_LinearOp_d*[2];
   ops[0] = op1;op1->IncCount();
   ops[1] = op2;op2->IncCount();
   ParamTable = new Table( NULL,NULL,2 );
   ParamTable->PutArrayValue( "Weights",1,a1 );
   ParamTable->PutArrayValue( "Weights",2,a2 );
   dom = &(op1->Domain());dom->IncCount();
   range = &(op1->Range());range->IncCount();
}

HCL_LinCombLinearOp_d::~HCL_LinCombLinearOp_d()
{
   HCL_delete( range );
   HCL_delete( dom );
   HCL_delete( ParamTable );
   for( int i=0;i<numinit;i++ )
      HCL_delete( ops[i] );
   delete[] ops;
}

void HCL_LinCombLinearOp_d::SetNext( HCL_LinearOp_d * L,double a )
{
   if( numinit == num )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::SetNext: "
              "too many factors" << endl;
      exit(1);
   }
   if( L->Domain() != Domain() || L->Range() != Range() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::SetNext: "
              "Domain or Range do not agree" << endl;
      exit(1);
   }
   ops[numinit] = L;L->IncCount();
   numinit++;
   ParamTable->PutArrayValue( "Weights",numinit,a );
}

void HCL_LinCombLinearOp_d::Image( const HCL_Vector_d & x,
                                      HCL_Vector_d & y ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_LinCombLinearOp_d:Image: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d:Image: vector is "
              "not in the Domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d:Image: vector is "
              "not in the Range" << endl;
      exit(1);
   }
   double wt;
   y.Zero();
   for( int i=0;i<num;i++ )
   {
      ParamTable->GetArrayValue( "Weights",i+1,wt );
      ops[i]->ImageAdd( x,y,wt );
   }
}

void HCL_LinCombLinearOp_d::AdjImage( const HCL_Vector_d & x,
                                            HCL_Vector_d & y ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_LinCombLinearOp_d:AdjImage: operator has not "
              "been initialized" << endl;
      exit(1);
   }
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d:AdjImage: input vector is "
              "not in the range" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d:AdjImage: output vector is "
              "not in the domain" << endl;
      exit(1);
   }
   double wt;
   y.Zero();
   for( int i=0;i<num;i++ )
   {
      ParamTable->GetArrayValue( "Weights",i+1,wt );
      ops[i]->AdjImageAdd( x,y,wt );
   }
}

void HCL_LinCombLinearOp_d::InvImage( const HCL_Vector_d & x,
                                            HCL_Vector_d & y ) const
{
   cerr << "Error: HCL_LinCombLinearOp_d::InvImage is not implemented"
        << endl;
   exit(1);
}

void HCL_LinCombLinearOp_d::InvAdjImage( const HCL_Vector_d & x,
                                               HCL_Vector_d & y ) const
{
   cerr << "Error: HCL_LinCombLinearOp_d::InvAdjImage is not implemented"
        << endl;
   exit(1);
}

void HCL_LinCombLinearOp_d::ImageAdd( const HCL_Vector_d & x,
                                      const HCL_Vector_d & z,
                                      HCL_Vector_d & y,double a,
                                      double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::ImageAdd: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::ImageAdd: input vector "
              "is not in domain"
           << endl;
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::ImageAdd: second input vector "
              "is not in range"
           << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::ImageAdd: output vector "
              "is not in range"
           << endl;
      exit(1);
   }

   double wt;
   ParamTable->GetArrayValue( "Weights",1,wt );
   ops[0]->ImageAdd( x,z,y,a*wt,b );
   for( int i=1;i<num;i++ )
   {
      ParamTable->GetArrayValue( "Weights",i+1,wt );
      ops[i]->ImageAdd( x,y,a*wt );
   }
}

void HCL_LinCombLinearOp_d::AdjImageAdd( const HCL_Vector_d & x,
                                       const HCL_Vector_d & z,
                                       HCL_Vector_d & y, double a,
                                       double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::AdjImageAdd: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::AdjImageAdd: input vector is "
              "not in range" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::AdjImageAdd: output vector is "
              "not in domain" << endl;
      exit(1);
   }
   if( z.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::AdjImageAdd: second input "
           << endl << "vector is not in domain" << endl;
      exit(1);
   }

   double wt;
   ParamTable->GetArrayValue( "Weights",1,wt );
   ops[0]->AdjImageAdd( x,z,y,a*wt,b );
   for( int i=1;i<num;i++ )
   {
      ParamTable->GetArrayValue( "Weights",i+1,wt );
      ops[i]->AdjImageAdd( x,y,a*wt );
   }
}

void HCL_LinCombLinearOp_d::ImageAdd( const HCL_Vector_d & x,
                                    HCL_Vector_d & y,double a,
                                    double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::ImageAdd: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::ImageAdd: input vector "
              "is not in domain"
           << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::ImageAdd: output vector "
              "is not in range"
           << endl;
      exit(1);
   }

   double wt;
   ParamTable->GetArrayValue( "Weights",1,wt );
   ops[0]->ImageAdd( x,y,a*wt,b );
   for( int i=1;i<num;i++ )
   {
      ParamTable->GetArrayValue( "Weights",i+1,wt );
      ops[i]->ImageAdd( x,y,a*wt );
   }
}

void HCL_LinCombLinearOp_d::AdjImageAdd( const HCL_Vector_d & x,
                                       HCL_Vector_d & y, double a,
                                       double b ) const
{
   if( numinit != num )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::AdjImageAdd: operator not "
              "initialized" << endl;
      exit(1);
   }
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::AdjImageAdd: input vector is "
              "not in range" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_LinCombLinearOp_d::AdjImageAdd: output vector is "
              "not in domain" << endl;
      exit(1);
   }

   double wt;
   ParamTable->GetArrayValue( "Weights",1,wt );
   ops[0]->AdjImageAdd( x,y,a*wt,b );
   for( int i=1;i<num;i++ )
   {
      ParamTable->GetArrayValue( "Weights",i+1,wt );
      ops[i]->AdjImageAdd( x,y,a*wt );
   }
}

ostream & HCL_LinCombLinearOp_d::Write( ostream & str ) const
{
   str << "Linear Combination operator with terms"
       << endl;
   int i;
   for( i=0;i<numinit;i++ )
      ops[i]->Write( str );
   str << "and weights" << endl;
   for( i=0;i<numinit;i++ )
   {
      double wt;
      ParamTable->GetArrayValue( "Weights",i+1,wt );
      str << "   " << wt << endl;
   }
   return str;
}
