/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#include "HCL_LNLOp_d.h"

HCL_ZeroBiLinearOp_d::HCL_ZeroBiLinearOp_d()
{
   cerr << "Error: HCL_ZeroBiLinearOp_d default constructor called" << endl;
   exit(1);
}

HCL_ZeroBiLinearOp_d::HCL_ZeroBiLinearOp_d( const HCL_ZeroBiLinearOp_d & )
{
   cerr << "Error: HCL_ZeroBiLinearOp_d copy constructor called" << endl;
   exit(1);
}

HCL_ZeroBiLinearOp_d::HCL_ZeroBiLinearOp_d( HCL_VectorSpace_d * d1,
                                            HCL_VectorSpace_d * d2,
                                            HCL_VectorSpace_d * r )
{
   dom1 = d1;dom1->IncCount();
   dom2 = d2;dom2->IncCount();
   range = r;range->IncCount();
}

HCL_ZeroBiLinearOp_d::~HCL_ZeroBiLinearOp_d()
{
   HCL_delete( range );
   HCL_delete( dom1 );
   HCL_delete( dom2 );
}

HCL_VectorSpace_d & HCL_ZeroBiLinearOp_d::Domain1() const
{
   return *dom1;
}

HCL_VectorSpace_d & HCL_ZeroBiLinearOp_d::Domain2() const
{
   return *dom2;
}

HCL_VectorSpace_d & HCL_ZeroBiLinearOp_d::Range() const
{
   return *range;
}

void HCL_ZeroBiLinearOp_d::Image( const HCL_Vector_d & x,
                                  const HCL_Vector_d & y,
                                        HCL_Vector_d & z ) const
{
   if( x.Space() != Domain1() )
   {
      cerr << "Error in HCL_ZeroBiLinearOp_d::Image: vector is not in domain"
           << endl;
      exit(1);
   }
   if( y.Space() != Domain2() )
   {
      cerr << "Error in HCL_ZeroBiLinearOp_d::Image: vector is not in domain"
           << endl;
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_ZeroBiLinearOp_d::Image: vector is not in range"
           << endl;
      exit(1);
   }
   z.Zero();
}

HCL_LinearOp_d * HCL_ZeroBiLinearOp_d::Op1( const HCL_Vector_d & x ) const
{
   return new HCL_ZeroLinearOp_d( dom2,range );
}

HCL_LinearOp_d * HCL_ZeroBiLinearOp_d::Op2( const HCL_Vector_d & x ) const
{
   return new HCL_ZeroLinearOp_d( dom1,range );
}

HCL_LinearOp_d * HCL_ZeroBiLinearOp_d::Op3( const HCL_Vector_d & x ) const
{
   return new HCL_ZeroLinearOp_d( dom1,dom2 );
}

ostream & HCL_ZeroBiLinearOp_d::Write( ostream & str ) const
{
   str << "HCL_ZeroBiLinearOp_d with domain 1" << endl;
   dom1->Write( str );
   str << "and domain 2" << endl;
   dom2->Write( str );
   str << "and range" << endl;
   return range->Write( str );
}

HCL_EvalLNLOp_d::HCL_EvalLNLOp_d()
{
   cerr << "Error: HCL_EvalLNLOp_d default constructor called" << endl;
   exit(1);
}

HCL_EvalLNLOp_d::HCL_EvalLNLOp_d( const HCL_EvalLNLOp_d & )
{
   cerr << "Error: HCL_EvalLNLOp_d copy constructor called" << endl;
   exit(1);
}

HCL_EvalLNLOp_d::HCL_EvalLNLOp_d( HCL_LinearOp_d * LL,
                                    const HCL_Vector_d & xx )
{
   L = LL;L->IncCount();
   x = LL->Domain().Member();
   x->Copy( xx );
}

HCL_EvalLNLOp_d::~HCL_EvalLNLOp_d()
{
   HCL_delete( L );
   HCL_delete( x );
}

HCL_VectorSpace_d & HCL_EvalLNLOp_d::Domain() const
{
   return L->Domain();
}

HCL_VectorSpace_d & HCL_EvalLNLOp_d::Range() const
{
   return L->Range();
}

void HCL_EvalLNLOp_d::Image( HCL_Vector_d & y ) const
{
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_EvalLNLOp_d::Image: vector is not in range"
           << endl;
      exit(1);
   }
   L->Image( *x,y );
}

void HCL_EvalLNLOp_d::ImageAdd( HCL_Vector_d & y,
                                double a,double b ) const
{
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_EvalLNLOp_d::ImageAdd: output vector "
           << endl << "is not in range" << endl;
      exit(1);
   }
   L->ImageAdd( *x,y,a,b );
}

void HCL_EvalLNLOp_d::ImageAdd( const HCL_Vector_d & z,HCL_Vector_d & y,
                                double a,double b ) const
{
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_EvalLNLOp_d::ImageAdd: output vector "
           << endl << "is not in range" << endl;
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_EvalLNLOp_d::ImageAdd: input vector "
           << endl << "is not in range" << endl;
      exit(1);
   }
   L->ImageAdd( *x,z,y,a,b );
}

HCL_LinearOp_d * HCL_EvalLNLOp_d::Deriv() const
{
   L->IncCount();
   return L;
}

HCL_BiLinearOp_d * HCL_EvalLNLOp_d::SecondDeriv() const
{
   return new HCL_ZeroBiLinearOp_d( &(Domain()),&(Domain()),&(Range()) );
}

ostream & HCL_EvalLNLOp_d::Write( ostream & str ) const
{
   str << "HCL_EvalLNLOp_d defined by the following linear operator:" << endl;
   return L->Write( str );
}

HCL_LNLOp_d::HCL_LNLOp_d()
{
   cerr << "Error: HCL_LNLOp_d default constructor called" << endl;
   exit(1);
}

HCL_LNLOp_d::HCL_LNLOp_d( const HCL_LNLOp_d & )
{
   cerr << "Error: HCL_LNLOp_d copy constructor called" << endl;
   exit(1);
}

HCL_LNLOp_d::HCL_LNLOp_d( const HCL_LinearOp_d * LL )
{
   L = (HCL_LinearOp_d*)LL;L->IncCount();
}

HCL_LNLOp_d::~HCL_LNLOp_d()
{
   HCL_delete( L );
}

HCL_VectorSpace_d & HCL_LNLOp_d::Domain() const
{
   return L->Domain();
}

HCL_VectorSpace_d & HCL_LNLOp_d::Range() const
{
   return L->Range();
}

HCL_EvaluateOp_d * HCL_LNLOp_d::Evaluate( const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_LNLOp_d::Evaluate: vector is not "
              "in domain" << endl;
      exit(1);
   }
   return new HCL_EvalLNLOp_d( L,x );
}

ostream & HCL_LNLOp_d::Write( ostream & str ) const
{
   str << "HCL_LNLOp_d defined by the following linear operator:" << endl;
   return L->Write( str );
}
