#include "HCL_LS_d.h"

HCL_EvalLSFcnlGN_d::HCL_EvalLSFcnlGN_d()
{
   cerr << "Error: HCL_EvalLSFcnlGN_d default constructor not implemented"
        << endl;
   exit(1);
}

HCL_EvalLSFcnlGN_d::HCL_EvalLSFcnlGN_d( HCL_EvalLSFcnlGN_d & F)
{
   cerr << "Error: HCL_EvalLSFcnlGN_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_EvalLSFcnlGN_d::HCL_EvalLSFcnlGN_d(const HCL_Vector_d & Input,
				       HCL_LinearOp_d * Operator,
				       HCL_Vector_d * Data,
				       double PenaltyParameter,
				       HCL_LinearOp_d * RegularizingOperator,
				       HCL_Vector_d * ModelPrior)
{
   opflag=0;
   L=Operator;L->IncCount();
   N = NULL;Neval = NULL;
   d =Data;d->IncCount();
   x = Input.Space().Member();
   x->Copy( Input );
   eps=PenaltyParameter;
   R=NULL;
   xprior=NULL;
   p=NULL;
   if (RegularizingOperator) {
     R=RegularizingOperator; R->IncCount();
   }
   if (ModelPrior) {
     xprior=ModelPrior; xprior->IncCount();
   }
   if (R) {
     p=R->Range().Member();
   }
   else {
     p=L->Domain().Member();
   }

   if (x->Space() != L->Domain())
   {
      cerr<<"Error: HCL_EvalLSFcnlGN_d"<<endl;
      cerr<<"input vector not in domain"<<endl;
      exit(1);
   }

   if (d->Space() != L->Range())
   {
      cerr<<"Error: HCL_EvalLSFcnlGN_d"<<endl;
      cerr<<"data vector not in range"<<endl;
      exit(1);
   }


   if (R) {
     if (L->Domain() != R->Domain()) {
       cerr<<"Error: HCL_EvalLSFcnlGN_d"<<endl;
       cerr<<"Operator, regularizing operator have different domains"<<endl;
       exit(1);
     }
   }
   if (xprior) {
     if (xprior->Space() != L->Domain()) {
       cerr<<"Error: HCL_EvalLSFcnlGN_d"<<endl;
       cerr<<"model prior vector not in domain"<<endl;
       exit(1);
     }
   }
   
   r = L->Range().Member();
   L->Image(*x,*r);
   r->Sub(*d);

   // compute penalty term

   HCL_FPArch_d fp;
   
   if (R) {
     HCL_Vector_d * dx=NULL;
     if (xprior) {
       dx=R->Domain().Member();
       dx->Sub(*x,*xprior);
       R->Image(*dx,*p);
       HCL_delete(dx);
     }
     else {
       R->Image(*x,*p);
     }
   }
   else if (eps*eps > fp.eps) {
     if (xprior) {
       p->Sub(*x,*xprior);
     }
     else {
       p->Copy(*x);
     }
   }
}

HCL_EvalLSFcnlGN_d::HCL_EvalLSFcnlGN_d( const HCL_Vector_d & Input,
					HCL_Op_d * Operator, 
					HCL_Vector_d * Data,
					double PenaltyParameter,
					HCL_LinearOp_d * RegularizingOperator,
					HCL_Vector_d * ModelPrior )
{
   opflag=1;
   N=Operator;N->IncCount();
   L = NULL;
   d =Data;d->IncCount();
   x = Input.Space().Member();
   x->Copy( Input );
   eps=PenaltyParameter;
   R=NULL;
   xprior=NULL;
   p=NULL;
   if (RegularizingOperator) {
     R=RegularizingOperator; R->IncCount();
   }
   if (ModelPrior) {
     xprior=ModelPrior; xprior->IncCount();
   }
   if (R) {
     p=R->Range().Member();
   }
   else {
     p=N->Domain().Member();
   }

   if (x->Space() != N->Domain())
   {
      cerr<<"Error: HCL_EvalLSFcnlGN_d"<<endl;
      cerr<<"input vector not in domain"<<endl;
      exit(1);
   }

   if (d->Space() != N->Range())
   {
      cerr<<"Error: HCL_EvalLSFcnlGN_d"<<endl;
      cerr<<"data vector not in range"<<endl;
      exit(1);
   }

   if (R) {
     if (N->Domain() != R->Domain()) {
       cerr<<"Error: HCL_EvalLSFcnlGN_d"<<endl;
       cerr<<"Operator, regularizing operator have different domains"<<endl;
       exit(1);
     }
     if (xprior) {
       if (xprior->Space() != N->Domain()) {
	 cerr<<"Error: HCL_EvalLSFcnlGN_d"<<endl;
	 cerr<<"model prior vector not in domain"<<endl;
	 exit(1);
       }
     }
   }

   Neval = N->Evaluate(*x);

   // compute residual

   r=N->Range().Member();
   r->Sub( Neval->ImageRef(),*d );

   // compute penalty term

   HCL_FPArch_d fp;
   
   if (R) {
     HCL_Vector_d * dx=NULL;
     if (xprior) {
       dx=R->Domain().Member();
       dx->Sub(*x,*xprior);
       R->Image(*dx,*p);
       HCL_delete(dx);
     }
     else {
       R->Image(*x,*p);
     }
   }
   else if (eps*eps > fp.eps) {
     if (xprior) {
       p->Sub(*x,*xprior);
     }
     else {
       p->Copy(*x);
     }
   }
}

HCL_EvalLSFcnlGN_d::~HCL_EvalLSFcnlGN_d()
{
   HCL_delete(L);
   HCL_delete(r);
   HCL_delete(Neval);
   HCL_delete(x);
   HCL_delete(d);
   HCL_delete(N);
   if (R) HCL_delete(R);
   if (xprior) HCL_delete(xprior);
   if (p) HCL_delete(p);
}

HCL_VectorSpace_d & HCL_EvalLSFcnlGN_d::Domain() const
{
   if (opflag)
      return N->Domain();
    else
      return L->Domain();
}

double HCL_EvalLSFcnlGN_d::Value() const
{
  HCL_FPArch_d fp;

  if (R || eps*eps > fp.eps) {
    return 0.5*(r->Norm2() + eps*eps*p->Norm2());
  }
  else {
    return 0.5*(r->Norm2());
  }
}

void HCL_EvalLSFcnlGN_d::Gradient( HCL_Vector_d & g ) const
{
   if (g.Space() != Domain())
   {
      cerr<<"Error: HCL_EvalLSFcnlGN_d::Gradient"<<endl;
      cerr<<"gradient vector not in domain"<<endl;
      exit(1);
   }

   if (opflag)
      (Neval->DerivRef()).AdjImage(*r,g);
   else
      L->AdjImage(*r,g);

   // regularize
   HCL_FPArch_d fp;
   if (eps*eps > fp.eps) {
     if (R && p) {
       HCL_Vector_d * tmp=(R->Domain()).Member();
       R->AdjImage(*p,*tmp);
       g.ScaleAdd(eps*eps,*tmp);
       HCL_delete(tmp);
     }
     else {
       g.ScaleAdd(eps*eps,*p);
     }
   }
       
}

HCL_LinearOp_d * HCL_EvalLSFcnlGN_d::Hessian() const {
  HCL_FPArch_d fp;
  if( !opflag ) {
    if (eps*eps > fp.eps) {
      HCL_LinearOp_d * n1 = new HCL_Normal_d( L );
      HCL_LinearOp_d * n2;
      if (R) {
	n2 = new HCL_Normal_d( R );
      }
      else {
	n2 = new HCL_ScaleLinearOp_d(&(L->Domain()),1.0);
      }
      HCL_LinearOp_d * t = new HCL_LinCombLinearOp_d(n1, 1.0, n2, eps*eps);
      HCL_delete(n1);
      HCL_delete(n2);
      return t;
    }
    else {
      return new HCL_Normal_d(L);
    }
  }
  else {
    if (eps*eps>fp.eps) {
      HCL_LinearOp_d * HH = Neval->Deriv();
      HCL_LinearOp_d * n1 = new HCL_Normal_d( HH );
      HCL_LinearOp_d * n2;
      if (R) {
	n2 = new HCL_Normal_d( R );
      }
      else {
	n2 = new HCL_ScaleLinearOp_d(&(L->Domain()),1.0);
      }
      HCL_LinearOp_d * t = new HCL_LinCombLinearOp_d(n1, 1.0, n2, eps*eps);
      HCL_delete(n1);
      HCL_delete(n2);
      HCL_delete( HH );
      return t;
    }
    else {
      HCL_LinearOp_d * HH = Neval->Deriv();
      HCL_LinearOp_d * t = new HCL_Normal_d( HH );
      HCL_delete( HH );
      return t;
    }
  }
}

ostream & HCL_EvalLSFcnlGN_d::Write( ostream & str ) const
{
   str<<endl<<"*** LeastSquaresFcnl Evaluation Object ***"<<endl;
   str<<"Input Vector:"<<endl;
   x->Write(str);
   if (opflag)
   {
      str<<"*** [Nonlinear] Operator:"<<endl;
      N->Write(str);
   }
   else
   {
      str<<"*** [Linear] Operator:"<<endl;
      L->Write(str);
   }
   str<<"*** Data:"<<endl;
   d->Write(str);
   return str;
}

HCL_LeastSquaresFcnlGN_d::HCL_LeastSquaresFcnlGN_d()
{
   cerr << "Error: HCL_LeastSquaresFcnlGN_d default constructor not implemented"
        << endl;
   exit(1);
}

HCL_LeastSquaresFcnlGN_d::HCL_LeastSquaresFcnlGN_d(
                              const HCL_LeastSquaresFcnlGN_d & )
{
   cerr << "Error: HCL_LeastSquaresFcnlGN_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_LeastSquaresFcnlGN_d::HCL_LeastSquaresFcnlGN_d( HCL_LinearOp_d * Operator,
						    HCL_Vector_d * Data,
						    double PenaltyParameter,
						    HCL_LinearOp_d * RegularizingOperator,
						    HCL_Vector_d * ModelPrior)
{
   opflag=0;
   L=Operator;L->IncCount();
   N = NULL;
   d = Data;d->IncCount();
   eps=PenaltyParameter;
   R=NULL;
   xprior=NULL;
   if (RegularizingOperator) {
     R=RegularizingOperator; R->IncCount();
   }
   if (ModelPrior) {
     xprior=ModelPrior; xprior->IncCount();
   }
}

HCL_LeastSquaresFcnlGN_d::HCL_LeastSquaresFcnlGN_d( HCL_Op_d * Operator,
						    HCL_Vector_d * Data,
						    double PenaltyParameter,
						    HCL_LinearOp_d * RegularizingOperator,
						    HCL_Vector_d * ModelPrior)
{
   opflag=1;
   N=Operator;N->IncCount();
   L = NULL;
   d = Data;d->IncCount();
   eps=PenaltyParameter;
   R=NULL;
   xprior=NULL;
   if (RegularizingOperator) {
     R=RegularizingOperator; R->IncCount();
   }
   if (ModelPrior) {
     xprior=ModelPrior; xprior->IncCount();
   }
}

HCL_LeastSquaresFcnlGN_d::~HCL_LeastSquaresFcnlGN_d()
{
   HCL_delete(d);
   HCL_delete(N);
   HCL_delete(L);
   if (R) HCL_delete(R);
   if (xprior) HCL_delete(xprior);
}

HCL_VectorSpace_d & HCL_LeastSquaresFcnlGN_d::Domain() const
{
   if (opflag)
     return N->Domain();
   else
     return L->Domain();
}

  // Evaluation

HCL_EvaluateFunctional_d * HCL_LeastSquaresFcnlGN_d::Evaluate(
                               const HCL_Vector_d & x ) const
{
   if (opflag)
     return new HCL_EvalLSFcnlGN_d( x, N, d, eps, R, xprior);
   else
     return new HCL_EvalLSFcnlGN_d( x, L, d, eps, R, xprior);
}

ostream & HCL_LeastSquaresFcnlGN_d::Write( ostream & str ) const
{
   str<<endl<<"*** LeastSquaresFcnlGN ***"<<endl;
   if (opflag)
   {
      str<<"*** [Nonlinear] Operator:"<<endl;
      N->Write(str);
   }
   else
   {
      str<<"*** [Linear] Operator:"<<endl;
      L->Write(str);
   }
   str<<"*** Data:"<<endl;
   d->Write(str);
   return str;
}

HCL_EvalLSFcnlGNProductDomain_d::HCL_EvalLSFcnlGNProductDomain_d()
{
   cerr << "Error: HCL_EvalLSFcnlGNProductDomain_d default "
           "constructor not implemented" << endl;
   exit(1);
}

HCL_EvalLSFcnlGNProductDomain_d::HCL_EvalLSFcnlGNProductDomain_d(
                           const HCL_EvalLSFcnlGNProductDomain_d & )
{
   cerr << "Error: HCL_EvalLSFcnlGNProductDomain_d copy constructor "
           "not implemented" << endl;
   exit(1);
}

HCL_EvalLSFcnlGNProductDomain_d::HCL_EvalLSFcnlGNProductDomain_d(
                               const HCL_ProductVector_d & Input,
                                     HCL_OpProductDomain_d * Operator,
                                     HCL_Vector_d * Data)
{
   N = Operator;N->IncCount();
   d = Data;d->IncCount();
   Neval = (HCL_EvalOpProductDomain_d*)N->Evaluate( Input );
   r = d->Space().Member();
   Neval->Image( *r );
   r->Sub( *d );
}

HCL_EvalLSFcnlGNProductDomain_d::~HCL_EvalLSFcnlGNProductDomain_d()
{
   HCL_delete( r );
   HCL_delete( Neval );
   HCL_delete( d );
   HCL_delete( N );
}

HCL_VectorSpace_d & HCL_EvalLSFcnlGNProductDomain_d::Domain() const
{
   return N->Domain();
}

double HCL_EvalLSFcnlGNProductDomain_d::Value() const
{
   return 0.5*r->Norm2();
}

void HCL_EvalLSFcnlGNProductDomain_d::PartialGradient( int i,
                                        HCL_Vector_d & g ) const
{
   if( i < 1 || i > ProductDomain().Number() )
   {
      cerr << "Error in HCL_EvalLSFcnlGNProductDomain_d::PartialGradient: "
              "invalid index" << endl;
      exit(1);
   }
   if( g.Space() != ProductDomain()(i) )
   {
      cerr << "Error in HCL_EvalLSFcnlGNProductDomain_d::PartialGradient: "
              "vector is not in proper space" << endl;
      exit(1);
   }

   Neval->PartialDerivRef( i ).AdjImage( *r,g );
}

HCL_LinearOp_d * HCL_EvalLSFcnlGNProductDomain_d::Hessian() const
{
   HCL_LinearOp_d * D = Neval->Deriv();
   HCL_LinearOp_d * t = new HCL_Normal_d( D );
   HCL_delete( D );
   return t;
}

HCL_LinearOp_d * HCL_EvalLSFcnlGNProductDomain_d::HessianBlock(
                    int i,int j ) const
{
   if( i < 1 || i > ProductDomain().Number() || j < 1 ||
                j > ProductDomain().Number() )
   {
      cerr << "Error in HCL_EvalLSFcnlGNProductDomain_d::HessianBlock: "
              "invalid index" << endl;
      exit(1);
   }
   HCL_LinearOp_d  * L1 = Neval->PartialDeriv( j );
   HCL_LinearOp_d  * L2 = Neval->PartialDeriv( i );
   HCL_Adjoint_d * L2t = new HCL_Adjoint_d( L2 );
   HCL_delete( L2 );
   HCL_CompLinearOp_d * H = new HCL_CompLinearOp_d( L1,L2t );
   HCL_delete( L1 );
   HCL_delete( L2t );
   return H;
}

ostream & HCL_EvalLSFcnlGNProductDomain_d::Write( ostream & str ) const
{
   str << "HCL_EvalLSFcnlGNProductDomain_d defined by " << endl;
   str << "Operator: " << endl;
   N->Write( str );
   str << "Data: " << endl;
   return d->Write( str );
}

HCL_LeastSquaresFcnlGNProductDomain_d::HCL_LeastSquaresFcnlGNProductDomain_d()
{
   cerr << "Error: HCL_LeastSquaresFcnlGNProductDomain_d default "
           "constructor not implemented" << endl;
   exit(1);
}

HCL_LeastSquaresFcnlGNProductDomain_d::HCL_LeastSquaresFcnlGNProductDomain_d(
                             const HCL_LeastSquaresFcnlGNProductDomain_d & )
{
   cerr << "Error: HCL_LeastSquaresFcnlGNProductDomain_d copy "
           "constructor not implemented" << endl;
   exit(1);
}

HCL_LeastSquaresFcnlGNProductDomain_d::HCL_LeastSquaresFcnlGNProductDomain_d(
   HCL_OpProductDomain_d * Operator,HCL_Vector_d * Data )
{
   if( Data->Space() != Operator->Range() )
   {
      cerr << "Error in HCL_LeastSquaresFcnlGNProductDomain_d: data is "
              "not in domain of operator" << endl;
      exit(1);
   }
   N = Operator;N->IncCount();
   d = Data;d->IncCount();
}

HCL_LeastSquaresFcnlGNProductDomain_d::~HCL_LeastSquaresFcnlGNProductDomain_d()
{
   HCL_delete( d );
   HCL_delete( N );
}

HCL_VectorSpace_d & HCL_LeastSquaresFcnlGNProductDomain_d::Domain() const
{
   return N->Domain();
}

HCL_EvaluateFunctional_d *  HCL_LeastSquaresFcnlGNProductDomain_d::
        Evaluate( const HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_LeastSquaresFcnlGNProductDomain_d::Evaluate:"
              " vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   return new HCL_EvalLSFcnlGNProductDomain_d( x,N,d );
}

ostream & HCL_LeastSquaresFcnlGNProductDomain_d::Write( ostream & str ) const
{
   str << "HCL_LeastSquaresFcnlGNProductDomain_d defined by " << endl;
   str << "Operator: " << endl;
   N->Write( str );
   str << "Data: " << endl;
   return d->Write( str );
}

HCL_EvalLSFcnl_d::HCL_EvalLSFcnl_d()
{
   cerr << "Error: HCL_EvalLSFcnl_d default constructor not implemented"
        << endl;
   exit(1);
}

HCL_EvalLSFcnl_d::HCL_EvalLSFcnl_d( HCL_EvalLSFcnl_d & F)
{
   cerr << "Error: HCL_EvalLSFcnl_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_EvalLSFcnl_d::HCL_EvalLSFcnl_d( const HCL_Vector_d & Input,
 					HCL_Op_d * Operator, 
					HCL_Vector_d * Data,
					double PenaltyParameter,
					HCL_LinearOp_d * RegularizingOperator,
					HCL_Vector_d * ModelPrior )
{
   N=Operator;N->IncCount();
   d =Data;d->IncCount();
   x = Input.Space().Member();
   x->Copy( Input );
   eps=PenaltyParameter;
   R=NULL;
   xprior=NULL;
   p=NULL;
   if (RegularizingOperator) {
     R=RegularizingOperator; R->IncCount();
   }
   if (ModelPrior) {
     xprior=ModelPrior; xprior->IncCount();
   }
   if (R) {
     p=R->Range().Member();
   }
   else {
     p=N->Domain().Member();
   }

   if (x->Space() != N->Domain())
   {
      cerr<<"Error: HCL_EvalLSFcnl_d"<<endl;
      cerr<<"input vector not in domain"<<endl;
      exit(1);
   }

   if (d->Space() != N->Range())
   {
      cerr<<"Error: HCL_EvalLSFcnl_d"<<endl;
      cerr<<"data vector not in range"<<endl;
      exit(1);
   }

   if (R) {
     if (N->Domain() != R->Domain()) {
       cerr<<"Error: HCL_EvalLSFcnl_d"<<endl;
       cerr<<"Operator, regularizing operator have different domains"<<endl;
       exit(1);
     }
     if (xprior) {
       if (xprior->Space() != N->Domain()) {
	 cerr<<"Error: HCL_EvalLSFcnl_d"<<endl;
	 cerr<<"model prior vector not in domain"<<endl;
	 exit(1);
       }
     }
   }

   Neval = N->Evaluate(*x);

   // compute residual

   r=N->Range().Member();
   r->Sub( Neval->ImageRef(),*d );

   // compute penalty term

   HCL_FPArch_d fp;
   
   if (R) {
     HCL_Vector_d * dx=NULL;
     if (xprior) {
       dx=R->Domain().Member();
       dx->Sub(*x,*xprior);
       R->Image(*dx,*p);
       HCL_delete(dx);
     }
     else {
       R->Image(*x,*p);
     }
   }
   else if (eps*eps > fp.eps) {
     if (xprior) {
       p->Sub(*x,*xprior);
     }
     else {
       p->Copy(*x);
     }
   }
}

HCL_EvalLSFcnl_d::~HCL_EvalLSFcnl_d()
{
   HCL_delete(r);
   HCL_delete(Neval);
   HCL_delete(x);
   HCL_delete(d);
   HCL_delete(N);
   if (R) HCL_delete(R);
   if (xprior) HCL_delete(xprior);
   if (p) HCL_delete(p);
}

HCL_VectorSpace_d & HCL_EvalLSFcnl_d::Domain() const
{
   return N->Domain();
}

double HCL_EvalLSFcnl_d::Value() const
{
  HCL_FPArch_d fp;

  if (R || eps*eps > fp.eps) {
    return 0.5*(r->Norm2() + eps*eps*p->Norm2());
  }
  else {
    return 0.5*(r->Norm2());
  }
}

void HCL_EvalLSFcnl_d::Gradient( HCL_Vector_d & g ) const
{
   if (g.Space() != Domain())
   {
      cerr<<"Error: HCL_EvalLSFcnl_d::Gradient"<<endl;
      cerr<<"gradient vector not in domain"<<endl;
      exit(1);
   }

   (Neval->DerivRef()).AdjImage(*r,g);

   // regularize
   HCL_FPArch_d fp;
   if (eps*eps > fp.eps) {
     if (R && p) {
       HCL_Vector_d * tmp=(R->Domain()).Member();
       R->AdjImage(*p,*tmp);
       g.ScaleAdd(eps*eps,*tmp);
       HCL_delete(tmp);
     }
     else {
       g.ScaleAdd(eps*eps,*p);
     }
   }
       
}

HCL_LinearOp_d * HCL_EvalLSFcnl_d::Hessian() const {
  HCL_FPArch_d fp;
  if (eps*eps>fp.eps) {
    HCL_LinearOp_d * HH = Neval->Deriv();
    HCL_LinearOp_d * n1 = new HCL_Normal_d( HH );
    HCL_delete( HH );
    HCL_LinearOp_d * n2;
    if (R) {
	n2 = new HCL_Normal_d( R );
    }
    else {
	n2 = new HCL_ScaleLinearOp_d(&(Domain()),1.0);
    }
    HCL_LinCombLinearOp_d * t = new HCL_LinCombLinearOp_d(
                                     &(Domain()),&(Domain()),3 );
    t->SetNext( n1,1.0 );
    HCL_delete( n1 );
    t->SetNext( n2,eps*eps );
    HCL_delete( n2 );
    HCL_BiLinearOp_d * D2Nx = Neval->SecondDeriv();
    HCL_LinearOp_d * L3 = D2Nx->Op3( *r );
    t->SetNext( L3,1.0 );
    HCL_delete( L3 );
    return t;
  }
  else {
    HCL_LinearOp_d * HH = Neval->Deriv();
    HCL_LinearOp_d * n1 = new HCL_Normal_d( HH );
    HCL_delete( HH );
    HCL_LinCombLinearOp_d * t = new HCL_LinCombLinearOp_d(
                                    &(Domain()),&(Domain()),2 );
    t->SetNext( n1,1.0 );
    HCL_BiLinearOp_d * D2Nx = Neval->SecondDeriv();
    HCL_LinearOp_d * L3 = D2Nx->Op3( *r );
    t->SetNext( L3,1.0 );
    HCL_delete( L3 );
    return t;
  }
}

ostream & HCL_EvalLSFcnl_d::Write( ostream & str ) const
{
   str<<endl<<"*** LeastSquaresFcnl Evaluation Object ***"<<endl;
   str<<"Input Vector:"<<endl;
   x->Write(str);
   str<<"*** Operator:"<<endl;
   N->Write(str);
   str<<"*** Data:"<<endl;
   d->Write(str);
   return str;
}

HCL_LeastSquaresFcnl_d::HCL_LeastSquaresFcnl_d()
{
   cerr << "Error: HCL_LeastSquaresFcnl_d default constructor not implemented"
        << endl;
   exit(1);
}

HCL_LeastSquaresFcnl_d::HCL_LeastSquaresFcnl_d(
                              const HCL_LeastSquaresFcnl_d & )
{
   cerr << "Error: HCL_LeastSquaresFcnl_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_LeastSquaresFcnl_d::HCL_LeastSquaresFcnl_d( HCL_Op_d * Operator,
				    HCL_Vector_d * Data,
				    double PenaltyParameter,
				    HCL_LinearOp_d * RegularizingOperator,
				    HCL_Vector_d * ModelPrior)
{
   N=Operator;N->IncCount();
   d = Data;d->IncCount();
   eps=PenaltyParameter;
   R=NULL;
   xprior=NULL;
   if (RegularizingOperator) {
     R=RegularizingOperator; R->IncCount();
   }
   if (ModelPrior) {
     xprior=ModelPrior; xprior->IncCount();
   }
}

HCL_LeastSquaresFcnl_d::~HCL_LeastSquaresFcnl_d()
{
   HCL_delete(d);
   HCL_delete(N);
   if (R) HCL_delete(R);
   if (xprior) HCL_delete(xprior);
}

HCL_VectorSpace_d & HCL_LeastSquaresFcnl_d::Domain() const
{
   return N->Domain();
}

  // Evaluation

HCL_EvaluateFunctional_d * HCL_LeastSquaresFcnl_d::Evaluate(
                               const HCL_Vector_d & x ) const
{
     return new HCL_EvalLSFcnl_d( x, N, d, eps, R, xprior);
}

ostream & HCL_LeastSquaresFcnl_d::Write( ostream & str ) const
{
   str<<endl<<"*** LeastSquaresFcnl ***"<<endl;
   str<<"*** Operator:"<<endl;
   N->Write(str);
   str<<"*** Data:"<<endl;
   d->Write(str);
   return str;
}

HCL_EvalLSFcnlProductDomain_d::HCL_EvalLSFcnlProductDomain_d()
{
   cerr << "Error: HCL_EvalLSFcnlProductDomain_d default "
           "constructor not implemented" << endl;
   exit(1);
}

HCL_EvalLSFcnlProductDomain_d::HCL_EvalLSFcnlProductDomain_d(
                           const HCL_EvalLSFcnlProductDomain_d & )
{
   cerr << "Error: HCL_EvalLSFcnlProductDomain_d copy constructor "
           "not implemented" << endl;
   exit(1);
}

HCL_EvalLSFcnlProductDomain_d::HCL_EvalLSFcnlProductDomain_d(
                               const HCL_ProductVector_d & Input,
                                     HCL_OpProductDomain_d * Operator,
                                     HCL_Vector_d * Data)
{
   N = Operator;N->IncCount();
   d = Data;d->IncCount();
   Neval = (HCL_EvalOpProductDomain_d*)N->Evaluate( Input );
   r = d->Space().Member();
   Neval->Image( *r );
   r->Sub( *d );
}

HCL_EvalLSFcnlProductDomain_d::~HCL_EvalLSFcnlProductDomain_d()
{
   HCL_delete( r );
   HCL_delete( Neval );
   HCL_delete( d );
   HCL_delete( N );
}

HCL_VectorSpace_d & HCL_EvalLSFcnlProductDomain_d::Domain() const
{
   return N->Domain();
}

double HCL_EvalLSFcnlProductDomain_d::Value() const
{
   return 0.5*r->Norm2();
}

void HCL_EvalLSFcnlProductDomain_d::PartialGradient( int i,
                                        HCL_Vector_d & g ) const
{
   if( i < 1 || i > ProductDomain().Number() )
   {
      cerr << "Error in HCL_EvalLSFcnlProductDomain_d::PartialGradient: "
              "invalid index" << endl;
      exit(1);
   }
   if( g.Space() != ProductDomain()(i) )
   {
      cerr << "Error in HCL_EvalLSFcnlProductDomain_d::PartialGradient: "
              "vector is not in proper space" << endl;
      exit(1);
   }

   Neval->PartialDerivRef( i ).AdjImage( *r,g );
}

HCL_LinearOp_d * HCL_EvalLSFcnlProductDomain_d::HessianBlock(
                    int i,int j ) const
{
   if( i < 1 || i > ProductDomain().Number() || j < 1 ||
                j > ProductDomain().Number() )
   {
      cerr << "Error in HCL_EvalLSFcnlProductDomain_d::HessianBlock: "
              "invalid index" << endl;
      exit(1);
   }
   HCL_LinearOp_d  * L1 = Neval->PartialDeriv( j );
   HCL_LinearOp_d  * L2 = Neval->PartialDeriv( i );
   HCL_Adjoint_d * L2t = new HCL_Adjoint_d( L2 );
   HCL_delete( L2 );
   HCL_CompLinearOp_d * H1 = new HCL_CompLinearOp_d( L1,L2t );
   HCL_delete( L1 );
   HCL_delete( L2t );
   HCL_BiLinearOp_d * B = Neval->SecondPartialDeriv( i,j );
   HCL_LinearOp_d * H2 = B->Op3( *r );
   HCL_LinCombLinearOp_d * H = new HCL_LinCombLinearOp_d( H1,1.0,H2,1.0 );
   HCL_delete( H2 );
   HCL_delete( H1 );
   return H;
}

ostream & HCL_EvalLSFcnlProductDomain_d::Write( ostream & str ) const
{
   str << "HCL_EvalLSFcnlProductDomain_d defined by " << endl;
   str << "Operator: " << endl;
   N->Write( str );
   str << "Data: " << endl;
   return d->Write( str );
}

HCL_LeastSquaresFcnlProductDomain_d::HCL_LeastSquaresFcnlProductDomain_d()
{
   cerr << "Error: HCL_LeastSquaresFcnlProductDomain_d default "
           "constructor not implemented" << endl;
   exit(1);
}

HCL_LeastSquaresFcnlProductDomain_d::HCL_LeastSquaresFcnlProductDomain_d(
                             const HCL_LeastSquaresFcnlProductDomain_d & )
{
   cerr << "Error: HCL_LeastSquaresFcnlProductDomain_d copy "
           "constructor not implemented" << endl;
   exit(1);
}

HCL_LeastSquaresFcnlProductDomain_d::HCL_LeastSquaresFcnlProductDomain_d(
   HCL_OpProductDomain_d * Operator,HCL_Vector_d * Data )
{
   if( Data->Space() != Operator->Range() )
   {
      cerr << "Error in HCL_LeastSquaresFcnlProductDomain_d: data is "
              "not in domain of operator" << endl;
      exit(1);
   }
   N = Operator;N->IncCount();
   d = Data;d->IncCount();
}

HCL_LeastSquaresFcnlProductDomain_d::~HCL_LeastSquaresFcnlProductDomain_d()
{
   HCL_delete( d );
   HCL_delete( N );
}

HCL_VectorSpace_d & HCL_LeastSquaresFcnlProductDomain_d::Domain() const
{
   return N->Domain();
}

HCL_EvaluateFunctional_d *  HCL_LeastSquaresFcnlProductDomain_d::
        Evaluate( const HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_LeastSquaresFcnlProductDomain_d::Evaluate:"
              " vector is not in domain" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   return new HCL_EvalLSFcnlProductDomain_d( x,N,d );
}

ostream & HCL_LeastSquaresFcnlProductDomain_d::Write( ostream & str ) const
{
   str << "HCL_LeastSquaresFcnlProductDomain_d defined by " << endl;
   str << "Operator: " << endl;
   N->Write( str );
   str << "Data: " << endl;
   return d->Write( str );
}
