#include "HCL_MatrixOps_d.h"

HCL_SymMatOp_d::HCL_SymMatOp_d()
{
   cerr << "Error: HCL_SymMatOp_d default constructor not implemented" << endl;
   exit(1);
}

HCL_SymMatOp_d::HCL_SymMatOp_d( const HCL_SymMatOp_d & )
{
   cerr << "Error: HCL_SymMatOp_d copy constructor not implemented" << endl;
   exit(1);
}

HCL_SymMatOp_d::HCL_SymMatOp_d( GenSymMatrix_d * AA,int flag,
                                HCL_VectorSpace_d * d )
{
   if( d->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_SymMatOp_d constructor: dimension of domain ("
           << d->Dim() << ") does not agree with size of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   Matrix = AA;
   DeleteMat = flag;
   dom = d;d->IncCount();
   range = d;d->IncCount();
}

HCL_SymMatOp_d::HCL_SymMatOp_d( GenSymMatrix_d * AA,int flag,
                                HCL_VectorSpace_d * d,
                                HCL_VectorSpace_d * r )
{
   if( d->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_SymMatOp_d constructor: dimension of domain ("
           << d->Dim() << ") does not agree with size of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   if( r->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_SymMatOp_d constructor: dimension of range ("
           << r->Dim() << ") does not agree with size of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   Matrix = AA;
   DeleteMat = flag;
   dom = d;d->IncCount();
   range = r;r->IncCount();
}

HCL_SymMatOp_d::~HCL_SymMatOp_d()
{
   HCL_delete( range );
   HCL_delete( dom );
   if( DeleteMat ) delete Matrix;
}

void HCL_SymMatOp_d::Image( const HCL_Vector_d & xx,
                                  HCL_Vector_d & yy ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SymMatOp_d::Image: input vector is not in domain"
           << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SymMatOp_d::Image: output vector is not in range"
           << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SymMatOp_d::Image: input vector has NULL "
              "data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SymMatOp_d::Image: output vector has NULL "
              "data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Y.mult( *((GenSymMatrix_d *)Matrix),(GeneralMatrix_d)X );
}

void HCL_SymMatOp_d::AdjImage( const HCL_Vector_d & yy,
                                     HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SymMatOp_d::AdjImage: output vector is not in "
              "domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SymMatOp_d::AdjImage: input vector is not in range"
           << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SymMatOp_d::AdjImage: output vector has NULL "
              "data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SymMatOp_d::AdjImage: input vector has NULL "
              "data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   X.mult( *((GenSymMatrix_d *)Matrix),(GeneralMatrix_d)Y );
}

void HCL_SymMatOp_d::InvImage(const HCL_Vector_d & yy,
                                    HCL_Vector_d & xx) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SymMatOp_d::InvImage: output vector is not in "
              "domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SymMatOp_d::InvImage: input vector is not in range"
           << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SymMatOp_d::InvImage: output vector has NULL "
              "data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SymMatOp_d::InvImage: input vector has NULL "
              "data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->solve( X,Y );
}

void HCL_SymMatOp_d::InvAdjImage(const HCL_Vector_d & xx,
                                       HCL_Vector_d & yy) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SymMatOp_d::InvAdjImage: input vector is not "
              "in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SymMatOp_d::InvAdjImage: output vector is not "
              "in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SymMatOp_d::InvAdjImage: input vector has NULL "
              "data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SymMatOp_d::InvAdjImage: output vector has NULL "
              "data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->solve( Y,X );
}

HCL_UpperTriangularMatOp_d * HCL_SymMatOp_d::Cholesky() const
{
   if( Domain() != Range() )
   {
      cerr << "Error in HCL_SymMatOp_d::Cholesky: Cannot form the Cholesky"
           << endl << "factorization, because the domain and range are not "
              "equal" << endl;
      exit(1);
   }
   UpperTriangle_d * R = new UpperTriangle_d( Matrix->numRows() );
   int flag = Matrix->cholesky( *R );
   if( flag )
   {
      delete R;
      return (HCL_UpperTriangularMatOp_d*)NULL;
   }
   else
   {
      HCL_UpperTriangularMatOp_d * U =
               new HCL_UpperTriangularMatOp_d( R,1,&(Domain()),&(Domain()) );
      return U;
   }
}

ostream & HCL_SymMatOp_d::Write( ostream & str ) const
{
   str << "HCL_SymMatOp_d with domain" << endl;
   Domain().Write( str );
   str << "and range" << endl;
   Range().Write( str );
   str << "The matrix is" << endl << *Matrix << endl;
   return str;
}

HCL_GeneralMatOp_d::HCL_GeneralMatOp_d()
{
   cerr << "Error: HCL_GeneralMatOp_d default constructor not implemented"
        << endl;
   exit(1);
}

HCL_GeneralMatOp_d::HCL_GeneralMatOp_d( const HCL_GeneralMatOp_d & )
{
   cerr << "Error: HCL_GeneralMatOp_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_GeneralMatOp_d::HCL_GeneralMatOp_d( GeneralMatrix_d * AA,int flag,
                                        HCL_VectorSpace_d * d,
                                        HCL_VectorSpace_d * r )
{
   if( d->Dim() != AA->numCols() )
   {
      cerr << "Error in HCL_GeneralMatOp_d constructor: dimension of domain ("
           << d->Dim() << ") does not agree with # of columns of matrix ("
           << AA->numCols() << ")" << endl;
      exit(1);
   }
   if( r->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_GeneralMatOp_d constructor: dimension of range ("
           << r->Dim() << ") does not agree with # of rows of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   Matrix = AA;
   DeleteMat = flag;
   dom = d;d->IncCount();
   range = r;r->IncCount();
}

HCL_GeneralMatOp_d::~HCL_GeneralMatOp_d()
{
   HCL_delete( range );
   HCL_delete( dom );
   if( DeleteMat ) delete Matrix;
}

void HCL_GeneralMatOp_d::Image( const HCL_Vector_d & xx,
                                      HCL_Vector_d & yy ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_GeneralMatOp_d::Image: input vector is "
              "not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_GeneralMatOp_d::Image: output vector is "
              "not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_GeneralMatOp_d::Image: input vector has NULL "
              "data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_GeneralMatOp_d::Image: output vector has NULL "
              "data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Y.mult( *Matrix,X );
}

void HCL_GeneralMatOp_d::AdjImage( const HCL_Vector_d & yy,
                                         HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_GeneralMatOp_d::AdjImage: output vector is "
              "not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_GeneralMatOp_d::AdjImage: input vector is "
              "not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_GeneralMatOp_d::AdjImage: output vector has NULL "
              "data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_GeneralMatOp_d::AdjImage: input vector has NULL "
              "data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   X.multTrans( *Matrix,Y );
}

void HCL_GeneralMatOp_d::InvImage(const HCL_Vector_d & yy,
                                    HCL_Vector_d & xx) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_GeneralMatOp_d::InvImage: output vector is not in "
              "domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_GeneralMatOp_d::InvImage: input vector is  "
              "not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_GeneralMatOp_d::InvImage: output vector has NULL "
              "data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_GeneralMatOp_d::InvImage: input vector has NULL "
              "data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->solve( X,Y );
}

void HCL_GeneralMatOp_d::InvAdjImage(const HCL_Vector_d & xx,
                                       HCL_Vector_d & yy) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_GeneralMatOp_d::InvAdjImage: input vector is not "
              "in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_GeneralMatOp_d::InvAdjImage: output vector is not "
              "in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_GeneralMatOp_d::InvAdjImage: input vector "
              "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_GeneralMatOp_d::InvAdjImage: output vector "
              "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->transposeSolve( Y,X );
}

ostream & HCL_GeneralMatOp_d::Write( ostream & str ) const
{
   str << "HCL_GeneralMatOp_d with domain" << endl;
   Domain().Write( str );
   str << "and range" << endl;
   Range().Write( str );
   str << "The matrix is" << endl << *Matrix << endl;
   return str;
}

HCL_SymSparseRCSMatOp_d::HCL_SymSparseRCSMatOp_d()
{
   cerr << "Error: HCL_SymSparseRCSMatOp_d default constructor not "
           "implemented" << endl;
   exit(1);
}

HCL_SymSparseRCSMatOp_d::HCL_SymSparseRCSMatOp_d(
                         const HCL_SymSparseRCSMatOp_d & )
{
   cerr << "Error: HCL_SymSparseRCSMatOp_d copy constructor not "
           "implemented" << endl;
   exit(1);
}

HCL_SymSparseRCSMatOp_d::HCL_SymSparseRCSMatOp_d( SymSparseRCSMatrix_d * AA,
                         int flag,HCL_VectorSpace_d * d )
{
   if( d->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d constructor: dimension of "
              "domain (" << d->Dim()
           << ") does not agree with size of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   Matrix = AA;
   DeleteMat = flag;
   dom = d;d->IncCount();
   range = d;d->IncCount();
}

HCL_SymSparseRCSMatOp_d::HCL_SymSparseRCSMatOp_d( SymSparseRCSMatrix_d * AA,
                                int flag,
                                HCL_VectorSpace_d * d,
                                HCL_VectorSpace_d * r )
{
   if( d->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d constructor: dimension "
              "of domain (" << d->Dim()
           << ") does not agree with size of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   if( r->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d constructor: dimension "
              "of range ("
           << r->Dim() << ") does not agree with size of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   Matrix = AA;
   DeleteMat = flag;
   dom = d;d->IncCount();
   range = r;r->IncCount();
}

HCL_SymSparseRCSMatOp_d::~HCL_SymSparseRCSMatOp_d()
{
   HCL_delete( range );
   HCL_delete( dom );
   if( DeleteMat ) delete Matrix;
}

void HCL_SymSparseRCSMatOp_d::Image( const HCL_Vector_d & xx,
                                           HCL_Vector_d & yy ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::Image: input vector is "
              "not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::Image: output vector is "
              "not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::Image: input vector has NULL "
              "data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::Image: output vector has "
              "NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Y.mult( *((SymSparseRCSMatrix_d *)Matrix),(GeneralMatrix_d)X );
}

void HCL_SymSparseRCSMatOp_d::AdjImage( const HCL_Vector_d & yy,
                                              HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::AdjImage: output vector "
              "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::AdjImage: input vector "
              "is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::AdjImage: output vector "
              "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::AdjImage: input vector "
              "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   X.mult( *((SymSparseRCSMatrix_d *)Matrix),(GeneralMatrix_d)Y );
}

void HCL_SymSparseRCSMatOp_d::InvImage(const HCL_Vector_d & yy,
                                             HCL_Vector_d & xx) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::InvImage: output vector "
              "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::InvImage: input vector "
              "is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::InvImage: output vector "
              "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::InvImage: input vector "
              "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->solve( X,Y );
}

void HCL_SymSparseRCSMatOp_d::InvAdjImage(const HCL_Vector_d & xx,
                                                HCL_Vector_d & yy) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::InvAdjImage: input "
              "vector is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::InvAdjImage: output "
              "vector is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::InvAdjImage: input "
              "vector has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SymSparseRCSMatOp_d::InvAdjImage: output "
              "vector has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->solve( Y,X );
}

ostream & HCL_SymSparseRCSMatOp_d::Write( ostream & str ) const
{
   str << "HCL_SymSparseRCSMatOp_d with domain" << endl;
   Domain().Write( str );
   str << "and range" << endl;
   Range().Write( str );
   str << "The matrix is" << endl << *Matrix << endl;
   return str;
}

HCL_SparseRCSMatOp_d::HCL_SparseRCSMatOp_d()
{
   cerr << "Error: HCL_SparseRCSMatOp_d default constructor not "
           "implemented" << endl;
   exit(1);
}

HCL_SparseRCSMatOp_d::HCL_SparseRCSMatOp_d(
                         const HCL_SparseRCSMatOp_d & )
{
   cerr << "Error: HCL_SparseRCSMatOp_d copy constructor not "
           "implemented" << endl;
   exit(1);
}

HCL_SparseRCSMatOp_d::HCL_SparseRCSMatOp_d( SparseRCSMatrix_d * AA,
                                int flag,
                                HCL_VectorSpace_d * d,
                                HCL_VectorSpace_d * r )
{
   if( d->Dim() != AA->numCols() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d constructor: dimension "
              "of domain (" << d->Dim()
           << ") does not agree with size of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   if( r->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d constructor: dimension "
              "of range ("
           << r->Dim() << ") does not agree with size of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   Matrix = AA;
   DeleteMat = flag;
   dom = d;d->IncCount();
   range = r;r->IncCount();
}

HCL_SparseRCSMatOp_d::~HCL_SparseRCSMatOp_d()
{
   HCL_delete( range );
   HCL_delete( dom );
   if( DeleteMat ) delete Matrix;
}

void HCL_SparseRCSMatOp_d::Image( const HCL_Vector_d & xx,
                                           HCL_Vector_d & yy ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::Image: input vector is "
              "not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::Image: output vector is "
              "not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::Image: input vector has NULL "
              "data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::Image: output vector has "
              "NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Y.mult( *((SparseRCSMatrix_d *)Matrix),(GeneralMatrix_d)X );
}

void HCL_SparseRCSMatOp_d::AdjImage( const HCL_Vector_d & yy,
                                              HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::AdjImage: output vector "
              "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::AdjImage: input vector "
              "is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::AdjImage: output vector "
              "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::AdjImage: input vector "
              "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   X.multTrans( *Matrix,Y );
}

void HCL_SparseRCSMatOp_d::InvImage(const HCL_Vector_d & yy,
                                             HCL_Vector_d & xx) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::InvImage: output vector "
              "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::InvImage: input vector "
              "is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::InvImage: output vector "
              "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::InvImage: input vector "
              "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->solve( X,Y );
}

void HCL_SparseRCSMatOp_d::InvAdjImage(const HCL_Vector_d & xx,
                                                HCL_Vector_d & yy) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::InvAdjImage: input "
              "vector is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::InvAdjImage: output "
              "vector is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::InvAdjImage: input "
              "vector has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_SparseRCSMatOp_d::InvAdjImage: output "
              "vector has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   // Need a routine to (a) form the tranpose of the matrix or
   // (b) solve the system with the transposed matrix.
   cerr << "Error: HCL_SparseRCSMatOp_d::InvAdjImage is not implemented"
        << endl;
}

ostream & HCL_SparseRCSMatOp_d::Write( ostream & str ) const
{
   str << "HCL_SparseRCSMatOp_d with domain" << endl;
   Domain().Write( str );
   str << "and range" << endl;
   Range().Write( str );
   str << "The matrix is" << endl << *Matrix << endl;
   return str;
}

HCL_UpperTriangularMatOp_d::HCL_UpperTriangularMatOp_d()
{
   cerr << "Error: HCL_UpperTriangularMatOp_d default constructor "
        << endl << "not implemented" << endl;
   exit(1);
}

HCL_UpperTriangularMatOp_d::HCL_UpperTriangularMatOp_d(
                            const HCL_UpperTriangularMatOp_d & )
{
   cerr << "Error: HCL_UpperTriangularMatOp_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_UpperTriangularMatOp_d::HCL_UpperTriangularMatOp_d(
                            UpperTriangle_d * AA,int flag,
                            HCL_VectorSpace_d * d,HCL_VectorSpace_d * r )
{
   if( d->Dim() != AA->numCols() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d constructor: dimension "
           << endl << "of domain ("
           << d->Dim() << ") does not agree with # of columns of matrix ("
           << AA->numCols() << ")" << endl;
      exit(1);
   }
   if( r->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d constructor: dimension "
           << endl << "of range ("
           << r->Dim() << ") does not agree with # of rows of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   Matrix = AA;
   DeleteMat = flag;
   dom = d;d->IncCount();
   range = r;r->IncCount();
}

HCL_UpperTriangularMatOp_d::~HCL_UpperTriangularMatOp_d()
{
   HCL_delete( range );
   HCL_delete( dom );
   if( DeleteMat ) delete Matrix;
}

void HCL_UpperTriangularMatOp_d::Image( const HCL_Vector_d & xx,
                                      HCL_Vector_d & yy ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::Image: input vector is "
              "not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::Image: output vector is "
              "not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::Image: input vector has "
           << endl << "NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::Image: output vector has "
           << endl << "NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Y.mult( *((UpperTriangle_d *)Matrix),(GeneralMatrix_d)X );
}

void HCL_UpperTriangularMatOp_d::AdjImage( const HCL_Vector_d & yy,
                                         HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::AdjImage: output vector"
           << endl << "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::AdjImage: input vector is "
              "not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::AdjImage: output vector "
           << endl << "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::AdjImage: input vector "
           << endl << "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   X.multTrans( *Matrix,Y );
}

void HCL_UpperTriangularMatOp_d::InvImage(const HCL_Vector_d & yy,
                                    HCL_Vector_d & xx) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::InvImage: output vector "
           << "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::InvImage: input vector "
           << endl << "is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::InvImage: output vector"
           << "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::InvImage: input vector"
           << "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->solve( X,Y );
}

void HCL_UpperTriangularMatOp_d::InvAdjImage(const HCL_Vector_d & xx,
                                       HCL_Vector_d & yy) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::InvAdjImage: input vector"
           << "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::InvAdjImage: output"
           << "vector is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::InvAdjImage: input vector "
              "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_UpperTriangularMatOp_d::InvAdjImage: output vector"
           << endl << "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->transposeSolve( Y,X );
}

ostream & HCL_UpperTriangularMatOp_d::Write( ostream & str ) const
{
   str << "HCL_UpperTriangularMatOp_d with domain" << endl;
   Domain().Write( str );
   str << "and range" << endl;
   Range().Write( str );
   str << "The matrix is" << endl << *Matrix << endl;
   return str;
}

HCL_LowerTriangularMatOp_d::HCL_LowerTriangularMatOp_d()
{
   cerr << "Error: HCL_LowerTriangularMatOp_d default constructor "
        << endl << "not implemented" << endl;
   exit(1);
}

HCL_LowerTriangularMatOp_d::HCL_LowerTriangularMatOp_d(
                            const HCL_LowerTriangularMatOp_d & )
{
   cerr << "Error: HCL_LowerTriangularMatOp_d copy constructor not implemented"
        << endl;
   exit(1);
}

HCL_LowerTriangularMatOp_d::HCL_LowerTriangularMatOp_d(
                            LowerTriangle_d * AA,int flag,
                            HCL_VectorSpace_d * d,HCL_VectorSpace_d * r )
{
   if( d->Dim() != AA->numCols() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d constructor: dimension "
           << endl << "of domain ("
           << d->Dim() << ") does not agree with # of columns of matrix ("
           << AA->numCols() << ")" << endl;
      exit(1);
   }
   if( r->Dim() != AA->numRows() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d constructor: dimension "
           << endl << "of range ("
           << r->Dim() << ") does not agree with # of rows of matrix ("
           << AA->numRows() << ")" << endl;
      exit(1);
   }
   Matrix = AA;
   DeleteMat = flag;
   dom = d;d->IncCount();
   range = r;r->IncCount();
}

HCL_LowerTriangularMatOp_d::~HCL_LowerTriangularMatOp_d()
{
   HCL_delete( range );
   HCL_delete( dom );
   if( DeleteMat ) delete Matrix;
}

void HCL_LowerTriangularMatOp_d::Image( const HCL_Vector_d & xx,
                                      HCL_Vector_d & yy ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::Image: input vector is "
              "not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::Image: output vector is "
              "not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::Image: input vector has "
           << endl << "NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::Image: output vector has "
           << endl << "NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Y.mult( *((LowerTriangle_d *)Matrix),(GeneralMatrix_d)X );
}

void HCL_LowerTriangularMatOp_d::AdjImage( const HCL_Vector_d & yy,
                                         HCL_Vector_d & xx ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::AdjImage: output vector"
           << endl << "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::AdjImage: input vector is "
              "not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::AdjImage: output vector "
           << endl << "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::AdjImage: input vector "
           << endl << "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   X.multTrans( *Matrix,Y );
}

void HCL_LowerTriangularMatOp_d::InvImage(const HCL_Vector_d & yy,
                                    HCL_Vector_d & xx) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::InvImage: output vector "
           << "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::InvImage: input vector "
           << endl << "is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::InvImage: output vector"
           << "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::InvImage: input vector"
           << "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->solve( X,Y );
}

void HCL_LowerTriangularMatOp_d::InvAdjImage(const HCL_Vector_d & xx,
                                       HCL_Vector_d & yy) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::InvAdjImage: input vector"
           << "is not in domain" << endl;
      cerr << "Domain is: " << endl;
      Domain().Write( cerr );
      cerr << "Space of input vector is: " << endl;
      xx.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & x = (HCL_Vector_d&)xx; // casting away const
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::InvAdjImage: output"
           << "vector is not in range" << endl;
      cerr << "Range is: " << endl;
      Range().Write( cerr );
      cerr << "Space of output vector is: " << endl;
      yy.Space().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & y = (HCL_Vector_d&)yy; // casting away const
   double * xdat = x.Data();
   if( !xdat )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::InvAdjImage: input vector "
              "has NULL data pointer" << endl;
      cerr << "Input vector is" << endl;
      x.Write( cerr );
      exit(1);
   }
   double * ydat = y.Data();
   if( !ydat )
   {
      cerr << "Error in HCL_LowerTriangularMatOp_d::InvAdjImage: output vector"
           << endl << "has NULL data pointer" << endl;
      cerr << "Output vector is" << endl;
      y.Write( cerr );
      exit(1);
   }
   GeneralMatrix_d X( x.Dim(),xdat ),Y( y.Dim(),ydat );
   Matrix->transposeSolve( Y,X );
}

ostream & HCL_LowerTriangularMatOp_d::Write( ostream & str ) const
{
   str << "HCL_LowerTriangularMatOp_d with domain" << endl;
   Domain().Write( str );
   str << "and range" << endl;
   Range().Write( str );
   str << "The matrix is" << endl << *Matrix << endl;
   return str;
}
