/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_ProductVector_d.h"

int HCL_ProductSpace_d::operator==(const HCL_VectorSpace_d & xx) const
{
   if( this == &xx )
      return 1;
   if( xx.GetType() != GetType() )
      return 0;
   HCL_ProductSpace_d & S = (HCL_ProductSpace_d&)xx;

   if( S.Number() != Number() )
      return 0;
   for( int i=1;i<=Number();i++ )
      if( S(i) != (*this)(i) )
         return 0;

   return 1;
}

int HCL_ProductSpace_d::Dim() const
{
   int d=0,i;
   for( i=1;i<=Number();i++ ) d += (*this)(i).Dim();
   return d;
}

ostream & HCL_ProductSpace_d::Write( ostream & str ) const
{
   str << "HCL_ProductSpace_d" << endl;
   for(int i=1; i <= Number(); i++) {
     str<<"   Factor "<<i<<":"<<endl;
     (*this)(i).Write(str);
   }
   return str;
}

double HCL_ProductVector_d::Inner( const HCL_Vector_d & xx ) const
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::Inner: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   double s = 0.0;
   for( int i=1;i<=Number();i++ )
      s += (*this)(i).Inner( x(i) );
   return s;
}

double HCL_ProductVector_d::Norm2() const
{
   double s = 0.0;
   for( int i=1;i<=Number();i++ )
      s += (*this)(i).Norm2();
   return s;
}

double HCL_ProductVector_d::Norm() const
{
   return sqrt(Norm2());
}

void HCL_ProductVector_d::Neg()
{
   for( int i=1;i<=Number();i++ )
      (*this)(i).Neg();
}

void HCL_ProductVector_d::Mul( const double & a )
{
   for( int i=1;i<=Number();i++ )
      (*this)(i).Mul( a );
}

void HCL_ProductVector_d::Mul(const double & a,const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::Mul: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   for( int i=1;i<=Number();i++ )
      (*this)(i).Mul( a,x(i) );
}

void HCL_ProductVector_d::Add( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::Add: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   for( int i=1;i<=Number();i++ )
      (*this)(i).Add( x(i) );
}

void HCL_ProductVector_d::Add( const HCL_Vector_d & xx,
                               const HCL_Vector_d & yy )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::Add: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::Add: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   for( int i=1;i<=Number();i++ )
      (*this)(i).Add( x(i),y(i) );
}

void HCL_ProductVector_d::Sub( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::Sub: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   for( int i=1;i<=Number();i++ )
      (*this)(i).Sub( x(i) );
}

void HCL_ProductVector_d::Sub( const HCL_Vector_d & xx,
                               const HCL_Vector_d & yy )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::Sub: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::Sub: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   for( int i=1;i<=Number();i++ )
      (*this)(i).Sub( x(i),y(i) );
}

void HCL_ProductVector_d::ScaleAdd( const double & a,const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::ScaleAdd: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   for( int i=1;i<=Number();i++ )
      (*this)(i).ScaleAdd( a,x(i) );
}

void HCL_ProductVector_d::AddScale( const double & a,const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::AddScale: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   for( int i=1;i<=Number();i++ )
      (*this)(i).AddScale( a,x(i) );
}

void HCL_ProductVector_d::AddScale( const double & a,const HCL_Vector_d & xx,
                                    const HCL_Vector_d & yy )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::AddScale: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::AddScale: vector does not belong"
              " to the correct space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   for( int i=1;i<=Number();i++ )
      (*this)(i).AddScale( a,x(i),y(i) );
}

void HCL_ProductVector_d::Copy( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector_d::Copy: vector does not belong"
              " to the correct space" << endl;
      cerr << "This space is" << endl;
      Space().Write(cout);
      cerr << "Space of vector to be copied is" << endl;
      xx.Space().Write(cout);
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   for( int i=1;i<=Number();i++ )
      (*this)(i).Copy( x(i) );
}

void HCL_ProductVector_d::Zero()
{
   for( int i=1;i<=Number();i++ )
      (*this)(i).Zero();
}

void HCL_ProductVector_d::Random()
{
   for( int i=1;i<=Number();i++ )
      (*this)(i).Random();
}

ostream & HCL_ProductVector_d::Write( ostream & str ) const
{
   str<<"HCL_ProductVector_d"<<endl;
   for( int i=1;i<=Number();i++ ) 
   {
      str<<"   Factor "<<i<<":"<<endl;
      (*this)(i).Write( str );
   }
   return str;
}

void HCL_ProductVector_d::Fill( double a )
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Fill( a );
}

void HCL_ProductVector_d::Add( double a )
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Add( a );
}

void HCL_ProductVector_d::Add( const HCL_Vector_d & xx, double a )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Add: vector is not in same space"
           << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Add( x(i),a );
}

void HCL_ProductVector_d::DiagScale( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::DiagScale: vector is not "
              "in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).DiagScale( x(i) );
}

void HCL_ProductVector_d::DiagScale( const HCL_Vector_d & xx,
                                     const HCL_Vector_d & yy )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::DiagScale: first vector is not "
              "in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::DiagScale: second vector is not "
              "in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).DiagScale( x(i),y(i) );
}

void HCL_ProductVector_d::DiagRecipScale( const HCL_Vector_d & xx,
                                          double tol)
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::DiagRecipScale: vector is not "
              "in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).DiagRecipScale( x(i),tol );
}

void HCL_ProductVector_d::DiagRecipScale( const HCL_Vector_d & xx,
                                          const HCL_Vector_d & yy,
                                          double tol)
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::DiagRecipScale: first vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::DiagRecipScale: second vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).DiagRecipScale( x(i),y(i),tol );
}

double HCL_ProductVector_d::Max() const
{
   double m=(*this)(1).Max();
   int i;
   for( i=2;i<=Number();i++ )
      m = max( m,(*this)(i).Max() );
   return m;
}

void HCL_ProductVector_d::Max( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Max: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Max( x(i) );
}

void HCL_ProductVector_d::Max( const HCL_Vector_d & xx,
                               const HCL_Vector_d & yy )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Max: first vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Max: second vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Max( x(i),y(i) );
}

double HCL_ProductVector_d::Min() const
{
   double m=(*this)(1).Min();
   int i;
   for( i=2;i<=Number();i++ )
      m = min( m,(*this)(i).Min() );
   return m;
}

void HCL_ProductVector_d::Min( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Min: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Min( x(i) );
}

void HCL_ProductVector_d::Min( const HCL_Vector_d & xx,
                               const HCL_Vector_d & yy )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Min: first vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Min: second vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Min( x(i),y(i) );
}

double HCL_ProductVector_d::Sum() const
{
   double s = 0.0;
   int i;
   for( i=1;i<=Number();i++ )
      s += (*this)(i).Sum();
   return s;
}

void HCL_ProductVector_d::Abs()
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Abs();
}

void HCL_ProductVector_d::Abs( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Abs: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Abs( x(i) );
}

void HCL_ProductVector_d::Sign()
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Sign();
}

void HCL_ProductVector_d::Sign( double t )
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Sign( t );
}

void HCL_ProductVector_d::Sign( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Sign: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Sign( x(i) );
}

      /// \TEX{$c_i = {\rm sign}\,(x_i-y_i)$}
void HCL_ProductVector_d::Sign( const HCL_Vector_d & xx,
                                const HCL_Vector_d & yy )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Sign: first vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Sign: second vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Sign( x(i),y(i) );
}

void HCL_ProductVector_d::Sign( const HCL_Vector_d & xx,double t )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Sign: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Sign( x(i),t );
}

void HCL_ProductVector_d::Sign( double t, const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Sign: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Sign( t,x(i) );
}

void HCL_ProductVector_d::Greater( double t )
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Greater( t );
}

void HCL_ProductVector_d::Greater( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Greater: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Greater( x(i) );
}

      /// Mask function: \TEX{$c_i = 1$} if \TEX{$x_i$} > \TEX{$y_i$}, else 0
void HCL_ProductVector_d::Greater( const HCL_Vector_d & xx,
                                   const HCL_Vector_d & yy )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Greater: first vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   if( yy.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Greater: second vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & y = (HCL_ProductVector_d&)yy;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Greater( x(i),y(i) );
}

void HCL_ProductVector_d::Greater( const HCL_Vector_d & xx, double t )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Greater: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Greater( x(i),t );
}

void HCL_ProductVector_d::Greater( double t, const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Greater: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Greater( t,x(i) );
}

void HCL_ProductVector_d::Power( double p )
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Power( p );
}

void HCL_ProductVector_d::Power( double p, const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Power: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Power( p,x(i) );
}

void HCL_ProductVector_d::Sqrt()
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Sqrt();
}

void HCL_ProductVector_d::Sqrt( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Sqrt: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Sqrt( x(i) );
}

void HCL_ProductVector_d::Exp()
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Exp();
}

void HCL_ProductVector_d::Exp( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Exp: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Exp( x(i) );
}

void HCL_ProductVector_d::Log()
{
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Log();
}

void HCL_ProductVector_d::Log( const HCL_Vector_d & xx )
{
   if( xx.Space() != Space() )
   {
      cerr << "Error in HCL_ProductVector::Log: vector "
              "is not in same space" << endl;
      exit(1);
   }
   HCL_ProductVector_d & x = (HCL_ProductVector_d&)xx;
   int i;
   for( i=1;i<=Number();i++ )
      (*this)(i).Log( x(i) );
}
