/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_QuadFcnl_d.h"

HCL_EvalQuadFunctional_d::HCL_EvalQuadFunctional_d()
{
   cerr << "Error: HCL_EvalQuadFunctional_d default "
           "constructor called" << endl;
   exit(1);
}

HCL_EvalQuadFunctional_d::HCL_EvalQuadFunctional_d(
                          const HCL_EvalQuadFunctional_d & )
{
   cerr << "Error: HCL_EvalQuadFunctional_d copy constructor called"
        << endl;
   exit(1);
}

HCL_EvalQuadFunctional_d::HCL_EvalQuadFunctional_d(
                             const HCL_Vector_d & x,HCL_LinearOp_d * a,
                             HCL_Vector_d * bb,double aalpha )
{
   if( a->Domain() != a->Range() )
   {
      cerr << "Error in HCL_EvalQuadFunctional_d constructor: domain and "
              "range of operator must agree" << endl;
      exit(1);
   }
   A = a;a->IncCount();
   b = bb;if( b != NULL ) b->IncCount();
   alpha = aalpha;
   input = A->Domain().Member();
   input->Copy( x );
   Ax = NULL;
}

HCL_EvalQuadFunctional_d::~HCL_EvalQuadFunctional_d()
{
   HCL_delete( Ax );
   HCL_delete( input );
   HCL_delete( b );
   HCL_delete( A );
}

double HCL_EvalQuadFunctional_d::Value() const
{
   // Casting away const
   HCL_EvalQuadFunctional_d * t = (HCL_EvalQuadFunctional_d*)this;
   if( Ax == NULL )
   {
      t->Ax = A->Domain().Member();
      A->Image( *input,*(t->Ax) );
   }
   double v;
   v = 0.5*(input->Inner( *Ax ));
   if( b != NULL ) v += input->Inner( *b );
   v += alpha;
   return v;
}

void HCL_EvalQuadFunctional_d::Gradient( HCL_Vector_d & g ) const
{
   // Casting away const
   HCL_EvalQuadFunctional_d * t = (HCL_EvalQuadFunctional_d*)this;
   if( Ax == NULL )
   {
      t->Ax = A->Domain().Member();
      A->Image( *input,*(t->Ax) );
   }
   if( b != NULL )
      g.Add( *Ax,*b );
   else
      g.Copy( *Ax );
}

HCL_LinearOp_d * HCL_EvalQuadFunctional_d::Hessian() const
{
   A->IncCount();
   return A;
}

ostream & HCL_EvalQuadFunctional_d::Write( ostream & str ) const
{
   str << "HCL_EvalQuadFunctional_d defined by:" << endl;
   str << "A = " << endl;
   A->Write( str );
   if( b != NULL )
   {
      str << "b = " << endl;
      b->Write( str );
   }
   str << "alpha = " << alpha << endl;
   return str;
}

HCL_QuadFunctional_d::HCL_QuadFunctional_d()
{
   cerr << "Error: HCL_QuadFunctional_d default constructor called"
        << endl;
   exit(1);
}

HCL_QuadFunctional_d::HCL_QuadFunctional_d( const HCL_QuadFunctional_d & )
{
   cerr << "Error: HCL_QuadFunctional_d copy constructor called"
        << endl;
   exit(1);
}

HCL_QuadFunctional_d::HCL_QuadFunctional_d( HCL_LinearOp_d * a,
                                            HCL_Vector_d * bb,double aalpha )
{
   if( a->Domain() != a->Range() )
   {
      cerr << "Error in HCL_QuadFunctional_d constructor: domain and "
              "range of operator must agree" << endl;
      exit(1);
   }
   A = a;a->IncCount();
   if( bb != NULL )
   {
      if( bb->Space() != A->Domain() )
      {
         cerr << "Error in HCL_QuadFunctional_d constructor: gradient "
                 "must lie in domain of operator" << endl;
         exit(1);
      }
      b = bb;b->IncCount();
   }
   else
      b = NULL;
   alpha = aalpha;
}

HCL_QuadFunctional_d::~HCL_QuadFunctional_d()
{
   HCL_delete( b );
   HCL_delete( A );
}

HCL_EvaluateFunctional_d * HCL_QuadFunctional_d::Evaluate(
                               const HCL_Vector_d & x ) const
{
   return new HCL_EvalQuadFunctional_d( x,A,b,alpha );
}
ostream & HCL_QuadFunctional_d::Write( ostream & str ) const
{
   str << "HCL_QuadFunctional_d defined by:" << endl;
   str << "A = " << endl;
   A->Write( str );
   if( b != NULL )
   {
      str << "b = " << endl;
      b->Write( str );
   }
   str << "alpha = " << alpha << endl;
   return str;
}
