#include "HCL_SubProductProjector_d.h"

HCL_SubProductProjector_d::HCL_SubProductProjector_d()
{
   cerr << "Error: HCL_SubProductProjector_d default constructor is not "
           "implemented" << endl;
   exit(1);
}

HCL_SubProductProjector_d::HCL_SubProductProjector_d(
                           const HCL_SubProductProjector_d & )
{
   cerr << "Error: HCL_SubProductProjector_d copy constructor is not "
           "implemented" << endl;
   exit(1);
}

HCL_SubProductProjector_d::HCL_SubProductProjector_d(
                           HCL_ProductSpace_d * d,iArray * indices )
{
   if( indices->Number() > d->Number() )
   {
      cerr << "Error in HCL_SubProductProjector_d constructor: iArray object"
           << endl << "has too many elements" << endl;
      cerr << "Number of factors in domain: " << d->Number() << endl;
      cerr << "Number of factors in iArray: " << indices->Number() << endl;
      exit(1);
   }
   rng1 = NULL;
   ind = indices;ind->IncCount();
   dom = d;d->IncCount();
   rng = new HCL_GenericProductSpace_d( ind->Number() );
   int i;
   for( i=1;i<=ind->Number();i++ )
      rng->SetNext( &((*dom)( ind->Index(i) )) );
}

HCL_SubProductProjector_d::HCL_SubProductProjector_d( HCL_ProductSpace_d * d,
                                                      int jj )
{
   if( jj<1 || jj>d->Number() )
   {
      cerr << "Error in HCL_SubProductProjector_d constructor: invalid index"
           << endl;
      cerr << "Number of factors in domain: " << d->Number() << endl;
      cerr << "Index requested for range: " << jj << endl;
      exit(1);
   }
   dom = d;d->IncCount();
   j = jj;
   rng = NULL;
   rng1 = &((*dom)(j));
   ind = NULL;
}

HCL_SubProductProjector_d::~HCL_SubProductProjector_d()
{
   HCL_delete( ind );
   HCL_delete( rng1 );
   HCL_delete( rng );
   HCL_delete( dom );
}

void HCL_SubProductProjector_d::Image( const HCL_Vector_d & x,
                                                     HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_SubProductProjector_d::Image: input vector is "
              "not in domain" << endl;
      cerr << "Space of input vector: " << endl;
      x.Write( cerr );
      cerr << "Domain:" << endl;
      Domain().Write( cerr );
      exit(1);
   }
   HCL_ProductVector_d & in = (HCL_ProductVector_d&)x;
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_SubProductProjector_d::Image: output vector is "
              "not in range" << endl;
      cerr << "Space of output vector: " << endl;
      y.Write( cerr );
      cerr << "Range:" << endl;
      Range().Write( cerr );
      exit(1);
   }

   if( rng1 )
      y.Copy( in(j) );
   else
   {
      HCL_GenericProductVector_d & out = (HCL_GenericProductVector_d&)y;
      int i;
      for( i=1;i<=ind->Number();i++ )
         out(i).Copy( in( ind->Index(i)) );
   }
}

void HCL_SubProductProjector_d::AdjImage( const HCL_Vector_d & y,
                                                  HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_SubProductProjector_d::Image: output vector is "
              "not in domain" << endl;
      cerr << "Space of output vector: " << endl;
      x.Write( cerr );
      cerr << "Domain:" << endl;
      Domain().Write( cerr );
      exit(1);
   }
   HCL_ProductVector_d & out = (HCL_ProductVector_d&)x;
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_SubProductProjector_d::Image: input vector is "
              "not in range" << endl;
      cerr << "Space of input vector: " << endl;
      y.Write( cerr );
      cerr << "Range:" << endl;
      Range().Write( cerr );
      exit(1);
   }
   out.Zero();
   if( rng1 )
      out(j).Copy( y );
   else
   {
      HCL_GenericProductVector_d & in = (HCL_GenericProductVector_d&)y;
      int i;
      for( i=1;i<=ind->Number();i++ )
         out(ind->Index(i)).Copy( in(i) );
   }
}
   
ostream & HCL_SubProductProjector_d::Write( ostream & str ) const
{
   str << "HCL_SubProductProjector_d with domain" << endl;
   Domain().Write( str );
   str << "and range" << endl;
   Range().Write( str );
   return str;
}

