/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_ScaleLinearOp_d.h"

HCL_ScaleLinearOp_d::HCL_ScaleLinearOp_d()
{
   cerr << "Error: HCL_ScaleLinearOp_d default constructor called"
        << endl;
   exit(1);
}

HCL_ScaleLinearOp_d::HCL_ScaleLinearOp_d( const HCL_ScaleLinearOp_d & L )
{
   cerr << "Error: HCL_ScaleLinearOp_d copy constructor called"
        << endl;
   exit(1);
}

HCL_ScaleLinearOp_d::HCL_ScaleLinearOp_d( HCL_VectorSpace_d * SP,double mu )
{
   sp = SP;sp->IncCount();
   ParamTable = new Table;
   ParamTable->PutValue( "Scale",mu );
}

HCL_ScaleLinearOp_d::~HCL_ScaleLinearOp_d()
{
   HCL_delete( ParamTable );
   HCL_delete( sp );
}

HCL_VectorSpace_d & HCL_ScaleLinearOp_d::Domain() const
{
   return *sp;
}

HCL_VectorSpace_d & HCL_ScaleLinearOp_d::Range() const
{
   return *sp;
}

Table & HCL_ScaleLinearOp_d::Parameters()
{
  return *ParamTable;
}

void HCL_ScaleLinearOp_d::Image( const HCL_Vector_d & x,
                                       HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:Image: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:Image: vector is not in "
              "Range" << endl;
      exit(1);
   }
   double mu;
   ParamTable->GetValue( "Scale",mu );
   HCL_FPArch_d arch;
   if( fabs(mu-1.0)<=arch.eps ) y.Copy(x);
   else if( fabs(mu) <=arch.sfmin ) y.Zero();
   else y.Mul( mu,x );
}

void HCL_ScaleLinearOp_d::AdjImage( const HCL_Vector_d & x,
                                          HCL_Vector_d & y ) const
{
   Image( x,y );
}

void HCL_ScaleLinearOp_d::NormalImage( const HCL_Vector_d & x,
                                             HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:NormalImage: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:NormalImage: vector is not in "
              "Range" << endl;
      exit(1);
   }
   double mu;
   ParamTable->GetValue( "Scale",mu );
   HCL_FPArch_d arch;
   if( fabs(mu-1.0)<=arch.eps ) y.Copy(x);
   else if( fabs(mu)<=arch.sfmin ) y.Zero();
   else y.Mul( mu*mu,x );
}

void HCL_ScaleLinearOp_d::InvImage( const HCL_Vector_d & x,
                                          HCL_Vector_d & y ) const
{
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:InvImage: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:InvImage: vector is not in "
              "Range" << endl;
      exit(1);
   }
   double mu;
   ParamTable->GetValue( "Scale",mu );
   HCL_FPArch_d arch;
   if( fabs(mu-1.0)<=arch.eps ) y.Copy(x);
   else if( fabs(mu) < arch.sfmin )
   {
      cerr << "Error in HCL_ScaleLinearOp_d::InvImage: the operator is not"
           << endl << "safely invertible!" << endl;
      exit(1);
   }
   else y.Mul( 1.0/mu,x );
}

void HCL_ScaleLinearOp_d::InvAdjImage( const HCL_Vector_d & x,
                                             HCL_Vector_d & y ) const
{
   InvImage( x,y );
}

void HCL_ScaleLinearOp_d::ImageAdd(const HCL_Vector_d & x,
                                   const HCL_Vector_d & z,
                                   HCL_Vector_d & y,
                                   double a,double b) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:ImageAdd: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:ImageAdd: vector is not in "
              "Range" << endl;
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:ImageAdd: second input vector "
           << endl << "is not in Range" << endl;
      exit(1);
   }
   double mu;
   ParamTable->GetValue( "Scale",mu );
   a *= mu;

   HCL_FPArch_d arch;
   if( fabs(a-1.0)<=arch.eps )
   {
      if( fabs(b-1.0)<=arch.eps )
         y.Add( x,z );
      else
         y.AddScale( b,x,z );
   }
   else if( fabs(a)<=arch.small )
   {
      if( fabs(b-1.0)<=arch.eps )
         y.Copy( z );
      else
         y.Mul( b,z );
   }
   else
   {
      if( fabs(b-1.0)<=arch.eps )
         y.AddScale( a,z,x );
      else
      {
         y.Mul( a,x );
         y.AddScale( b,z );
      }
   }
}

void HCL_ScaleLinearOp_d::AdjImageAdd(const HCL_Vector_d & x,
                                      const HCL_Vector_d & z,
                                      HCL_Vector_d & y,
                                      double a,double b) const
{
   ImageAdd( x,z,y,a,b );
}

void HCL_ScaleLinearOp_d::NormalImageAdd(const HCL_Vector_d & x,
                                         const HCL_Vector_d & z,
                                         HCL_Vector_d & y,
                                         double a,double b) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:NormalImageAdd: input vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:NormalImageAdd: output vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   if( z.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:NormalImageAdd: second input"
           << endl << "vector is not in domain" << endl;
      exit(1);
   }
   double mu;
   ParamTable->GetValue( "Scale",mu );
   a *= (mu*mu);

   HCL_FPArch_d arch;
   if( fabs(a-1.0)<=arch.eps )
   {
      if( fabs(b-1.0)<=arch.eps )
         y.Add( x,z );
      else
         y.AddScale( b,x,z );
   }
   else if( fabs(a)<=arch.small )
   {
      if( fabs(b-1.0)<=arch.eps )
         y.Copy( z );
      else
         y.Mul( b,z );
   }
   else
   {
      if( fabs(b-1.0)<=arch.eps )
         y.AddScale( a,z,x );
      else
      {
         y.Mul( a,x );
         y.AddScale( b,z );
      }
   }
}

void HCL_ScaleLinearOp_d::InvImageAdd(const HCL_Vector_d & x,
                                      const HCL_Vector_d & z,
                                      HCL_Vector_d & y,
                                      double a,double b) const
{
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:InvImageAdd: vector is not in "
              "domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:InvImageAdd: vector is not in "
              "range" << endl;
      exit(1);
   }
   if( z.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:InvImageAdd: second input vector "
           << endl << "is not in range" << endl;
      exit(1);
   }
   double mu;
   ParamTable->GetValue( "Scale",mu );
   HCL_FPArch_d arch;
   if( fabs(mu)<(arch.sfmin)*fabs(a) )
   {
      cerr << "Error in HCL_ScaleLinearOp_d::InvImageAdd: the operator is not"
           << endl << "safely invertible!" << endl;
      exit(1);
   }
   a /= mu;

   if( fabs(a-1.0)<=arch.eps )
   {
      if( fabs(b-1.0)<=arch.eps )
         y.Add( x,z );
      else
         y.AddScale( b,x,z );
   }
   else if( fabs(a)<=arch.small )
   {
      if( fabs(b-1.0)<=arch.eps )
         y.Copy( z );
      else
         y.Mul( b,z );
   }
   else
   {
      if( fabs(b-1.0)<=arch.eps )
         y.AddScale( a,z,x );
      else
      {
         y.Mul( a,x );
         y.AddScale( b,z );
      }
   }
}

void HCL_ScaleLinearOp_d::InvAdjImageAdd(const HCL_Vector_d & x,
                                         const HCL_Vector_d & z,
                                         HCL_Vector_d & y,
                                         double a,double b) const
{
   InvImageAdd( x,z,y,a,b );
}

void HCL_ScaleLinearOp_d::ImageAdd(const HCL_Vector_d & x,HCL_Vector_d & y,
                                   double a,double b) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:ImageAdd: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:ImageAdd: vector is not in "
              "Range" << endl;
      exit(1);
   }
   double mu;
   ParamTable->GetValue( "Scale",mu );
   a *= mu;

   HCL_FPArch_d arch;
   if( fabs(a-1.0)<=arch.eps )
   {
      if( fabs(b-1.0)<=arch.eps )
         y.Add( x );
      else
         y.ScaleAdd( b,x );
   }
   else if( fabs(a)<=arch.small )
   {
      if( fabs(b-1.0)>arch.eps )
         y.Mul( b );
   }
   else
   {
      if( fabs(b-1.0)<=arch.eps )
         y.AddScale( a,x );
      else
      {
         y.Mul( b );
         y.AddScale( a,x );
      }
   }
}

void HCL_ScaleLinearOp_d::AdjImageAdd(const HCL_Vector_d & x,
                                      HCL_Vector_d & y,
                                      double a,double b) const
{
   ImageAdd( x,y,a,b );
}

void HCL_ScaleLinearOp_d::NormalImageAdd(const HCL_Vector_d & x,
                                         HCL_Vector_d & y,
                                         double a,double b) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:NormalImageAdd: input vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:NormalImageAdd: output vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   double mu;
   ParamTable->GetValue( "Scale",mu );
   a *= (mu*mu);

   HCL_FPArch_d arch;
   if( fabs(a-1.0)<=arch.eps )
   {
      if( fabs(b-1.0)<=arch.eps )
         y.Add( x );
      else
         y.ScaleAdd( b,x );
   }
   else if( fabs(a)<=arch.small )
   {
      if( fabs(b-1.0)>arch.eps )
         y.Mul( b );
   }
   else
   {
      if( fabs(b-1.0)<=arch.eps )
         y.AddScale( a,x );
      else
      {
         y.Mul( b );
         y.AddScale( a,x );
      }
   }
}

void HCL_ScaleLinearOp_d::InvImageAdd(const HCL_Vector_d & x,
                                      HCL_Vector_d & y,
                                      double a,double b) const
{
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:InvImageAdd: vector is not in "
              "domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ScaleLinearOp_d:InvImageAdd: vector is not in "
              "range" << endl;
      exit(1);
   }
   double mu;
   ParamTable->GetValue( "Scale",mu );
   HCL_FPArch_d arch;
   if( fabs(mu)<(arch.sfmin)*fabs(a) )
   {
      cerr << "Error in HCL_ScaleLinearOp_d::InvImageAdd: the operator is not"
           << endl << "safely invertible!" << endl;
      exit(1);
   }
   a /= mu;

   if( fabs(a-1.0)<=arch.eps )
   {
      if( fabs(b-1.0)<=arch.eps )
         y.Add( x );
      else
         y.ScaleAdd( b,x );
   }
   else if( fabs(a)<=arch.small )
   {
      if( fabs(b-1.0)>arch.eps )
         y.Mul( b );
   }
   else
   {
      if( fabs(b-1.0)<=arch.eps )
         y.AddScale( a,x );
      else
      {
         y.Mul( b );
         y.AddScale( a,x );
      }
   }
}

void HCL_ScaleLinearOp_d::InvAdjImageAdd(const HCL_Vector_d & x,
                                         HCL_Vector_d & y,
                                         double a,double b) const
{
   InvImageAdd( x,y,a,b );
}

ostream & HCL_ScaleLinearOp_d::Write( ostream & str ) const
{
   double mu;
   ParamTable->GetValue( "Scale",mu );
   HCL_FPArch_d arch;
   if( fabs(mu-1.0)<=arch.eps )
      str << "Identity operator defined on" << endl;
   else if( fabs(mu)<=arch.sfmin )
      str << "Zero operator defined on" << endl;
   else
      str << "Scaling operator with scale " << mu << " defined on" 
          << endl;
   sp->Write( str );
   return str;
}

HCL_DiagScaleLinearOp_d::HCL_DiagScaleLinearOp_d()
{
   cerr << "Error: HCL_DiagScaleLinearOp_d default constructor called"
        << endl;
   exit(1);
}

HCL_DiagScaleLinearOp_d::HCL_DiagScaleLinearOp_d(
                         const HCL_DiagScaleLinearOp_d & L )
{
   cerr << "Error: HCL_DiagScaleLinearOp_d copy constructor called"
        << endl;
   exit(1);
}

HCL_DiagScaleLinearOp_d::HCL_DiagScaleLinearOp_d( 
                                const HCL_Vector_d & d )
{
   diag = (HCL_Vector_d*)(d.Space().Member());
   diag->Copy( d );
}

HCL_DiagScaleLinearOp_d::HCL_DiagScaleLinearOp_d( HCL_Vector_d * d )
{
   diag = d;d->IncCount();
}

HCL_DiagScaleLinearOp_d::~HCL_DiagScaleLinearOp_d()
{
   HCL_delete( diag );
}

HCL_VectorSpace_d & HCL_DiagScaleLinearOp_d::Domain() const
{
   return diag->Space();
}

HCL_VectorSpace_d & HCL_DiagScaleLinearOp_d::Range() const
{
   return diag->Space();
}

void HCL_DiagScaleLinearOp_d::Image( const HCL_Vector_d & x,
                                           HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagScaleLinearOp_d:Image: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_DiagScaleLinearOp_d:Image: vector is not in "
              "Range" << endl;
      exit(1);
   }
   y.DiagScale( x,*diag );
}

void HCL_DiagScaleLinearOp_d::AdjImage( const HCL_Vector_d & x,
                                              HCL_Vector_d & y ) const
{
   Image( x,y );
}

void HCL_DiagScaleLinearOp_d::NormalImage( const HCL_Vector_d & x,
                                                 HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagScaleLinearOp_d:NormalImage: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagScaleLinearOp_d:NormalImage: vector is not in "
              "Range" << endl;
      exit(1);
   }
   y.DiagScale( x,*diag );
   y.DiagScale( *diag );
}

void HCL_DiagScaleLinearOp_d::InvImage( const HCL_Vector_d & x,
                                           HCL_Vector_d & y ) const
{
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_DiagScaleLinearOp_d:InvImage: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_DiagScaleLinearOp_d:InvImage: vector is not in "
              "Range" << endl;
      exit(1);
   }
   y.DiagRecipScale( *diag,x );
}

void HCL_DiagScaleLinearOp_d::InvAdjImage( const HCL_Vector_d & x,
                                              HCL_Vector_d & y ) const
{
   InvImage( x,y );
}

ostream & HCL_DiagScaleLinearOp_d::Write( ostream & str ) const
{
   str << "Diagonal Scaling operator with diagonal " << endl;
   diag->Write( str );
   return str;
}

HCL_ZeroLinearOp_d::HCL_ZeroLinearOp_d()
{
   cerr << "Error: HCL_ZeroLinearOp_d default constructor called"
        << endl;
   exit(1);
}

HCL_ZeroLinearOp_d::HCL_ZeroLinearOp_d( const HCL_ZeroLinearOp_d & L )
{
   cerr << "Error: HCL_ZeroLinearOp_d copy constructor called"
        << endl;
   exit(1);
}

HCL_ZeroLinearOp_d::HCL_ZeroLinearOp_d( HCL_VectorSpace_d * d,
                                        HCL_VectorSpace_d * r )
{
   dom = d;dom->IncCount();
   range = r;range->IncCount();
}

HCL_ZeroLinearOp_d::~HCL_ZeroLinearOp_d()
{
   HCL_delete( range );
   HCL_delete( dom );
}

HCL_VectorSpace_d & HCL_ZeroLinearOp_d::Domain() const
{
   return *dom;
}

HCL_VectorSpace_d & HCL_ZeroLinearOp_d::Range() const
{
   return *range;
}

void HCL_ZeroLinearOp_d::Image( const HCL_Vector_d & x,
                                       HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:Image: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:Image: vector is not in "
              "Range" << endl;
      exit(1);
   }
   y.Zero();
}

void HCL_ZeroLinearOp_d::AdjImage( const HCL_Vector_d & y,
                                         HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:AdjImage: vector is not in "
              "range" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:AdjImage: vector is not in "
              "domain" << endl;
      exit(1);
   }
   x.Zero();
}

void HCL_ZeroLinearOp_d::NormalImage( const HCL_Vector_d & x,
                                            HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:NormalImage: vector is not in "
              "domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:NormalImage: vector is not in "
              "domain" << endl;
      exit(1);
   }
   y.Zero();
}

void HCL_ZeroLinearOp_d::InvImage( const HCL_Vector_d & x,
                                       HCL_Vector_d & y ) const
{
   cerr << "Error in HCL_ZeroLinearOp_d::InvImage: the zero operator "
           "is not invertible!" << endl;
   exit(1);
}

void HCL_ZeroLinearOp_d::InvAdjImage( const HCL_Vector_d & y,
                                         HCL_Vector_d & x ) const
{
   cerr << "Error in HCL_ZeroLinearOp_d::InvAdjImage: the zero operator "
           "is not invertible!" << endl;
   exit(1);
}

void HCL_ZeroLinearOp_d::ImageAdd( const HCL_Vector_d & x,
                                   const HCL_Vector_d & z,
                                   HCL_Vector_d & y,
                                   double a,double b) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: vector is not in "
              "Range" << endl;
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: second input vector "
           << endl << "is not in Range" << endl;
      exit(1);
   }

   HCL_FPArch_d arch;
   if( fabs(b-1.0)<=arch.eps )
      y.Copy( z );
   else
      y.Mul( b,z );
}

void HCL_ZeroLinearOp_d::AdjImageAdd( const HCL_Vector_d & x,
                                      const HCL_Vector_d & z,
                                      HCL_Vector_d & y,
                                      double a,double b) const
{
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: vector is not in "
              "range" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: vector is not in "
              "domain" << endl;
      exit(1);
   }
   if( z.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: second input vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }

   HCL_FPArch_d arch;
   if( fabs(b-1.0)<=arch.eps )
      y.Copy( z );
   else
      y.Mul( b,z );
}

void HCL_ZeroLinearOp_d::NormalImageAdd( const HCL_Vector_d & x,
                                         const HCL_Vector_d & z,
                                         HCL_Vector_d & y,
                                         double a,double b) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:NormalImageAdd: input vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:NormalImageAdd: output vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   if( z.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:NormalImageAdd: second input"
           << endl << "vector is not in domain" << endl;
      exit(1);
   }

   HCL_FPArch_d arch;
   if( fabs(b-1.0)<=arch.eps )
      y.Copy( z );
   else
      y.Mul( b,z );
}

void HCL_ZeroLinearOp_d::InvImageAdd(const HCL_Vector_d & x,
                                      const HCL_Vector_d & z,
                                      HCL_Vector_d & y,
                                      double a,double b) const
{
   cerr << "Error in HCL_ZeroLinearOp_d::InvImageAdd: the zero "
        << endl << "operator is not invertible!" << endl;
   exit(1);
}

void HCL_ZeroLinearOp_d::InvAdjImageAdd(const HCL_Vector_d & x,
                                         const HCL_Vector_d & z,
                                         HCL_Vector_d & y,
                                         double a,double b) const
{
   cerr << "Error in HCL_ZeroLinearOp_d::InvAdjImageAdd: the zero "
        << endl << "operator is not invertible!" << endl;
   exit(1);
}

void HCL_ZeroLinearOp_d::ImageAdd(const HCL_Vector_d & x,HCL_Vector_d & y,
                                   double a,double b) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: vector is not in "
              "Domain" << endl;
      exit(1);
   }
   if( y.Space() != Range() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: vector is not in "
              "Range" << endl;
      exit(1);
   }

   HCL_FPArch_d arch;
   if( fabs(b-1.0)>arch.eps )
      y.Mul( b );
}

void HCL_ZeroLinearOp_d::AdjImageAdd(const HCL_Vector_d & x,
                                      HCL_Vector_d & y,
                                      double a,double b) const
{
   if( x.Space() != Range() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: vector is not in "
              "range" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:ImageAdd: vector is not in "
              "domain" << endl;
      exit(1);
   }

   HCL_FPArch_d arch;
   if( fabs(b-1.0)>arch.eps )
      y.Mul( b );
}

void HCL_ZeroLinearOp_d::NormalImageAdd(const HCL_Vector_d & x,
                                         HCL_Vector_d & y,
                                         double a,double b) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:NormalImageAdd: input vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }
   if( y.Space() != Domain() )
   {
      cerr << "Error in HCL_ZeroLinearOp_d:NormalImageAdd: output vector "
           << endl << "is not in domain" << endl;
      exit(1);
   }

   HCL_FPArch_d arch;
   if( fabs(b-1.0)>arch.eps )
      y.Mul( b );
}

void HCL_ZeroLinearOp_d::InvImageAdd(const HCL_Vector_d & x,
                                      HCL_Vector_d & y,
                                      double a,double b) const
{
   cerr << "Error in HCL_ZeroLinearOp_d::InvImageAdd: the zero "
        << endl << "operator is not invertible!" << endl;
   exit(1);
}

void HCL_ZeroLinearOp_d::InvAdjImageAdd(const HCL_Vector_d & x,
                                         HCL_Vector_d & y,
                                         double a,double b) const
{
   cerr << "Error in HCL_ZeroLinearOp_d::InvAdjImageAdd: the zero "
        << endl << "operator is not invertible!" << endl;
   exit(1);
}

ostream & HCL_ZeroLinearOp_d::Write( ostream & str ) const
{
   str << "Zero operator defined with domain" << endl;
   dom->Write( str );
   str << "and range" << endl;
   return range->Write( str );
}

