/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_SmoothL1Fcnl_d.h"

HCL_SmoothL1Fcnl_d::HCL_SmoothL1Fcnl_d()
{
   cerr << "Error: HCL_SmoothL1Fcnl_d default constructor "
           "called" << endl;
   exit(1);
}

HCL_SmoothL1Fcnl_d::HCL_SmoothL1Fcnl_d(
                           const HCL_SmoothL1Fcnl_d & x )
{
   cerr << "Error: HCL_SmoothL1Fcnl_d copy constructor called"
        << endl;
   exit(1);
}

HCL_SmoothL1Fcnl_d::HCL_SmoothL1Fcnl_d( HCL_LinearOp_d * a,
                                                      HCL_Vector_d * d,
                                                      double e )
{
   A = a; A->IncCount();
   if( d->Space() != A->Range() )
   {
      cerr<<"Error: HCL_SmoothL1Fcnl constructor"<<endl;
      cerr<<"space of data not equal to range of operator"<<endl;
      cerr<<"data:"<<endl;
      d->Write(cerr);
      cerr<<"operator"<<endl;
      A->Write(cerr);
      exit(1);
    }
    data = d; data->IncCount();
    eps = e;
}

HCL_SmoothL1Fcnl_d::~HCL_SmoothL1Fcnl_d()
{
   HCL_delete(A);
   HCL_delete(data);
}

HCL_EvaluateFunctional_d * 
HCL_SmoothL1Fcnl_d::Evaluate( const HCL_Vector_d & x ) const
{
  if (A->Domain() !=x.Space())
  {
      cerr<<"Error: HCL_SmoothL1Fcnl_d::Evaluate"<<endl;
      cerr<<"input vector not in domain of operator"<<endl;
      cerr<<"input vector:"<<endl;
      x.Write(cerr);
      cerr<<"operator:"<<endl;
      A->Write(cerr);
      exit(1);
   }
   return new HCL_EvalSmoothL1Fcnl_d( x,A,data,eps );
}

ostream & HCL_SmoothL1Fcnl_d::Write( ostream & str ) const
{
   str << "HCL_SmoothL1Fcnl_d with linear operator "
          "(epsilon = " << eps << ")" << endl;
   return A->Write( str );
}

HCL_EvalSmoothL1Fcnl_d::HCL_EvalSmoothL1Fcnl_d()
{
   cerr << "Error: HCL_EvalSmoothL1Fcnl_d default constructor "
           "called" << endl;
   exit(1);
}

HCL_EvalSmoothL1Fcnl_d::HCL_EvalSmoothL1Fcnl_d(
                           const HCL_EvalSmoothL1Fcnl_d & x )
{
   cerr << "Error: HCL_EvalSmoothL1Fcnl_d copy constructor called"
        << endl;
   exit(1);
}

HCL_EvalSmoothL1Fcnl_d::~HCL_EvalSmoothL1Fcnl_d()
{
   HCL_delete(d2fr);
   HCL_delete(dfr);
   HCL_delete(t1);
   HCL_delete(r);
   HCL_delete(fr);
   HCL_delete(dom);
   HCL_delete(A);
}

ostream & HCL_EvalSmoothL1Fcnl_d::Write( ostream & str ) const
{
   str << "HCL_EvalSmoothL1Fcnl_d"<< endl;
   return str;
}

HCL_EvalSmoothL1Fcnl_d::HCL_EvalSmoothL1Fcnl_d(
                               const HCL_Vector_d & x,HCL_LinearOp_d * a,
                               HCL_Vector_d * data,double e )
{

  // this construction presumes that data membership in operator
  // range has been checked elsewhere

  A = a;a->IncCount();
  eps = e;
  // record domain

  dom = &(A->Domain()); dom->IncCount();

  // allocate workspace

  fr = (A->Range().Member());
  dfr = NULL;
  d2fr = NULL;

  r = (A->Range().Member());
  t1 = (A->Range().Member());

  // compute value

  A->Image( x,*r );
  r->Sub( *data );     // r holds the data residual

  t1->DiagScale( *r,*r );
  t1->Add( eps*eps );   // Now t1 holds r^2+eps^2

  fr->Sqrt( *t1 );      // fr holds f(residual)
}

double HCL_EvalSmoothL1Fcnl_d::Value() const
{
   return fr->Sum();
}

void HCL_EvalSmoothL1Fcnl_d::Gradient(HCL_Vector_d & g ) const
{
   if( g.Space() != Domain() )
   {
      cerr << "Error in HCL_EvalSmoothL1Fcnl_d::Gradient: vector is "
              "not in domain" << endl;
      exit(1);
   }
   if( !dfr )
   {
      // Casting away const
      HCL_EvalSmoothL1Fcnl_d * t =
               (HCL_EvalSmoothL1Fcnl_d*)this;
      t->dfr = (A->Range().Member());
      (t->dfr)->DiagRecipScale( *fr,*r );  // dfr holds f'(residual)
   }

   A->AdjImage( *dfr,g );
}

HCL_LinearOp_d * HCL_EvalSmoothL1Fcnl_d::Hessian() const
{
   if( !d2fr )
   {
      // Casting away const
      HCL_EvalSmoothL1Fcnl_d * t =
               (HCL_EvalSmoothL1Fcnl_d*)this;
      t->d2fr = (A->Range().Member());
      (t->d2fr)->Fill( eps*eps );
      (t->d2fr)->DiagRecipScale( *t1 );
      (t->d2fr)->DiagRecipScale( *fr );   // d2fr holds f''(residual)
   }

   HCL_DiagScaleLinearOp_d * D = new HCL_DiagScaleLinearOp_d( *d2fr );
   HCL_CompLinearOp_d * hess = 
          new HCL_CompLinearOp_d( &(A->Domain()),&(A->Domain()),3 );
   HCL_Adjoint_d * At = new HCL_Adjoint_d( A );
   hess->SetNext( A );
   hess->SetNext( D );
   HCL_delete( D );
   hess->SetNext( At );
   HCL_delete( At );

   return hess;
}
