/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_TensorBiLinearOp_d.h"

HCL_TensorBiLinearOp_d::HCL_TensorBiLinearOp_d()
{
   cerr << "Error: HCL_TensorBiLinearOp_d default constructor is not "
           "implemented" << endl;
   exit(1);
}

HCL_TensorBiLinearOp_d::HCL_TensorBiLinearOp_d(
                        const HCL_TensorBiLinearOp_d & L )
{
   cerr << "Error: HCL_TensorBiLinearOp_d copy constructor is not "
           "implemented" << endl;
   exit(1);
}

HCL_TensorBiLinearOp_d::HCL_TensorBiLinearOp_d( HCL_VectorSpace_d * D1,
                            HCL_VectorSpace_d * D2,
                            HCL_VectorSpace_d * R,General3Tensor_d * t,
                            int deleteflag )
{
   if( D1->Dim() != t->Size( 3 ) )
   {
      cerr << "Error in HCL_TensorBiLinearOp_d constructor: dimension of "
           << endl << "Domain1 (" << D1->Dim()
           << ") must match the size of the tensor" << endl
           << "in the third dimension (" << t->Size( 3 ) << ")" << endl;
      exit(1);
   }
   if( D2->Dim() != t->Size( 2 ) )
   {
      cerr << "Error in HCL_TensorBiLinearOp_d constructor: dimension of "
           << endl << "Domain2 (" << D2->Dim()
           << ") must match the size of the tensor" << endl
           << "in the second dimension (" << t->Size( 2 ) << ")" << endl;
      exit(1);
   }
   if( R->Dim() != t->Size( 1 ) )
   {
      cerr << "Error in HCL_TensorBiLinearOp_d constructor: dimension of "
           << endl << "Range (" << R->Dim()
           << ") must match the size of the tensor" << endl
           << "in the first dimension (" << t->Size( 1 ) << ")" << endl;
      exit(1);
   }

   dom1 = D1;D1->IncCount();
   dom2 = D2;D2->IncCount();
   rng = R;R->IncCount();
   T = t;
   d1 = t->Size( 1 );
   d2 = t->Size( 2 );
   d3 = t->Size( 3 );
   DeleteTensor = deleteflag;
}

HCL_TensorBiLinearOp_d::~HCL_TensorBiLinearOp_d()
{
   if( DeleteTensor )
      delete T;
   HCL_delete( rng );
   HCL_delete( dom2 );
   HCL_delete( dom1 );
}

void HCL_TensorBiLinearOp_d::Image( const HCL_Vector_d & x,
                                    const HCL_Vector_d & y,
                                          HCL_Vector_d & z ) const
{
   if( x.Space() != Domain1() )
   {
      cerr << "Error in HCL_TensorBiLinearOp_d::Image: first input is not "
              "in Domain1" << endl;
      cerr << "Space of first input:" << endl;
      x.Space().Write( cerr );
      cerr << "Domain1:" << endl;
      Domain1().Write( cerr );
      exit(1);
   }
   if( y.Space() != Domain2() )
   {
      cerr << "Error in HCL_TensorBiLinearOp_d::Image: second input is not "
              "in Domain2" << endl;
      cerr << "Space of second input:" << endl;
      y.Space().Write( cerr );
      cerr << "Domain2:" << endl;
      Domain2().Write( cerr );
      exit(1);
   }
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_TensorBiLinearOp_d::Image: output vector is not "
              "in range" << endl;
      cerr << "Space of output vector:" << endl;
      z.Space().Write( cerr );
      cerr << "Range:" << endl;
      Range().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & xx = (HCL_Vector_d&)x; // casting away const
   HCL_Vector_d & yy = (HCL_Vector_d&)y; // casting away const
   General1Tensor_d X( xx.Dim(),xx.Data() ),Y( yy.Dim(),yy.Data() ),
                    Z( z.Dim(),z.Data() );
   T->Multiply( 3,2,X,Y,Z );
}

HCL_LinearOp_d * HCL_TensorBiLinearOp_d::Op1( const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain1() )
   {
      cerr << "Error in HCL_TensorBiLinearOp_d::Op1: input is not "
              "in Domain1" << endl;
      cerr << "Space of input:" << endl;
      x.Space().Write( cerr );
      cerr << "Domain1:" << endl;
      Domain1().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & xx = (HCL_Vector_d&)x; // casting away const
   General1Tensor_d X( xx.Dim(),xx.Data() );
   GeneralMatrix_d * A = new GeneralMatrix_d( d1,d2 );
   General2Tensor_d a( d1,d2,A->data );
   T->Multiply( 3,X,a );
   HCL_GeneralMatOp_d * L1 = new HCL_GeneralMatOp_d( A,1,dom2,rng );
   return L1;
}

HCL_LinearOp_d * HCL_TensorBiLinearOp_d::Op2( const HCL_Vector_d & y ) const
{
   if( y.Space() != Domain2() )
   {
      cerr << "Error in HCL_TensorBiLinearOp_d::Op2: input is not "
              "in Domain2" << endl;
      cerr << "Space of input:" << endl;
      y.Space().Write( cerr );
      cerr << "Domain2:" << endl;
      Domain2().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & yy = (HCL_Vector_d&)y; // casting away const
   General1Tensor_d Y( yy.Dim(),yy.Data() );
   GeneralMatrix_d * A = new GeneralMatrix_d( d1,d3 );
   General2Tensor_d a( d1,d3,A->data );
   T->Multiply( 2,Y,a );
   HCL_GeneralMatOp_d * L2 = new HCL_GeneralMatOp_d( A,1,dom1,rng );
   return L2;
}

HCL_LinearOp_d * HCL_TensorBiLinearOp_d::Op3( const HCL_Vector_d & z ) const
{
   if( z.Space() != Range() )
   {
      cerr << "Error in HCL_TensorBiLinearOp_d::Op3: output is not "
              "in Range" << endl;
      cerr << "Space of output:" << endl;
      z.Space().Write( cerr );
      cerr << "Range:" << endl;
      Range().Write( cerr );
      exit(1);
   }
   HCL_Vector_d & zz = (HCL_Vector_d&)z; // casting away const
   General1Tensor_d Z( zz.Dim(),zz.Data() );
   GeneralMatrix_d * A = new GeneralMatrix_d( d2,d3 );
   General2Tensor_d a( d2,d3,A->data );
   T->Multiply( 1,Z,a );
   HCL_GeneralMatOp_d * L3 = new HCL_GeneralMatOp_d( A,1,dom1,dom2 );
   return L3;
}

ostream & HCL_TensorBiLinearOp_d::Write( ostream & str ) const
{
   str << "HCL_TensorBiLinearOp_d with" << endl;
   str << "Domain1:" << endl;
   Domain1().Write( str );
   str << "Domain2:" << endl;
   Domain2().Write( str );
   str << "Range:" << endl;
   Range().Write( str );
   return str;
}
