/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include <stdlib.h>
#include "HCL_lmbfgsOp1_d.h"

HCL_lmbfgsOp1_d::HCL_lmbfgsOp1_d()
{
   cerr << "Error: HCL_lmbfgsOp1_d default constructor called" << endl;
   exit(1);
}

HCL_lmbfgsOp1_d::HCL_lmbfgsOp1_d( const HCL_lmbfgsOp1_d & H )
{
   cerr << "Error: HCL_lmbfgsOp1_d copy constructor called" << endl;
   exit(1);
}

HCL_lmbfgsOp1_d::HCL_lmbfgsOp1_d( HCL_ScaleLinearOp_d * h0,int maxnum,
      HCL_LinearOp_d * Pin,HCL_LinearSolver_d * ls )
{
   sp = &(h0->Domain());sp->IncCount();
   P = Pin;if( Pin != NULL ) Pin->IncCount();
   lsolver = ls;if( ls != NULL ) ls->IncCount();
   if( P != NULL )
   {
      temp_vec = P->Domain().Member();
      temp_vec1 = P->Range().Member();
   }
   else
   {
      temp_vec = NULL;
      temp_vec1 = NULL;
      lsolver = NULL;
   }   
   H0 = h0;h0->IncCount();
   MaxNum = maxnum;
   CurNum = 0;
   CurAllocated = 0;
   Y = new HCL_Vector_d*[MaxNum];
   S = new HCL_Vector_d*[MaxNum];
   rho = new double[MaxNum];
}

HCL_lmbfgsOp1_d::~HCL_lmbfgsOp1_d()
{
   delete[] rho;
   int i;
   for( i=0;i<CurAllocated;i++ ) HCL_delete( S[i] );
   delete[] S;
   for( i=0;i<CurAllocated;i++ ) HCL_delete( Y[i] );
   delete[] Y;
   HCL_delete( H0 );
   HCL_delete( temp_vec1 );
   HCL_delete( temp_vec );
   HCL_delete( lsolver );
   HCL_delete( P );
   HCL_delete( sp );
}

HCL_VectorSpace_d & HCL_lmbfgsOp1_d::Domain() const { return *sp; }

HCL_VectorSpace_d & HCL_lmbfgsOp1_d::Range() const { return *sp; }

void HCL_lmbfgsOp1_d::Reset() { CurNum = 0; }

void HCL_lmbfgsOp1_d::InvImage(const HCL_Vector_d & y,
                                     HCL_Vector_d & x) const
{
   cerr << "Error: HCL_lmbfgsOp1_d::InvImage is not implemented" << endl;
   exit(1);
}

void HCL_lmbfgsOp1_d::InvAdjImage(const HCL_Vector_d & x,
                                        HCL_Vector_d & y) const
{
   cerr << "Error: HCL_lmbfgsOp1_d::InvAdjImage is not implemented" << endl;
   exit(1);
}

void HCL_lmbfgsOp1_d::AdjImage( const HCL_Vector_d & y,HCL_Vector_d & x ) const
{
   Image( y,x );
}

void HCL_lmbfgsOp1_d::Image( const HCL_Vector_d & xx,
                                   HCL_Vector_d & yy ) const
{
   if( xx.Space() != Domain() )
   {
      cerr << "Error in HCL_lmbfgsOp1_d::Image: vector is not in Domain" 
           << endl;
      exit(1);
   }
   if( yy.Space() != Range() )
   {
      cerr << "Error in HCL_lmbfgsOp1_d::Image: vector is not in Range"
           << endl;
      exit(1);
   }

   double scale;
   H0->Parameters().GetValue( "Scale",scale );


   if( P == NULL )
   {
      // Handle the special case of no updates

      if( CurNum == 0 )
      {
         yy.Mul( scale,xx );
         return;
      }

      // This is the general case---the initial approximation has been updated
       
      double * alpha = new double[CurNum];
      int i;
      yy.Copy( xx );
      for( i=CurNum-1;i>=0;--i )
      {
         alpha[i] = rho[i]*(S[i]->Inner( yy ));
         yy.AddScale( -alpha[i],*(Y[i]) );
      }

      yy.Mul( scale );

      for( i=0;i<CurNum;++i )
      {
         double beta = rho[i]*(Y[i]->Inner( yy ));
         yy.AddScale( alpha[i]-beta,*(S[i]) );
      }
      delete[] alpha;
      return;
   }
   else
   {
      // Handle the special case of no updates

      if( CurNum == 0 )
      {
         if( lsolver )
            lsolver->Solve( *P,xx,yy );
         else
            P->InvImage( xx,yy );
         yy.Mul( scale );
         return;
      }

      // This is the general case---the initial approximation has been updated
       
      double * alpha = new double[CurNum];
      int i;
      temp_vec1->Copy( xx );
      for( i=CurNum-1;i>=0;--i )
      {
         alpha[i] = rho[i]*(S[i]->Inner( *temp_vec1 ));
         temp_vec1->AddScale( -alpha[i],*(Y[i]) );
      }

      temp_vec1->Mul( scale );
      if( lsolver )
         lsolver->Solve( *P,*temp_vec1,yy );
      else
         P->InvImage( *temp_vec1,yy );

      for( i=0;i<CurNum;++i )
      {
         double beta = rho[i]*(Y[i]->Inner( yy ));
         yy.AddScale( alpha[i]-beta,*(S[i]) );
      }
      delete[] alpha;
      return;
   }
}

void HCL_lmbfgsOp1_d::Update( const HCL_Vector_d & x,
                              const HCL_Vector_d & xnext,
                              const HCL_Vector_d & g,
                              const HCL_Vector_d & gnext )
{
   if( MaxNum == 0 )
      return;
   if( x.Space() != *sp || xnext.Space() != *sp ||
       g.Space() != *sp || gnext.Space() != *sp )
   {
      cerr << "Error in HCL_lmbfgsOp1_d::Update: attempt to update "
              "operator with vector from wrong space" << endl;
      exit(1);
   }
   int i;
   if( CurNum < MaxNum )
   {
      if( CurNum == CurAllocated )
      {
         CurAllocated++;
         S[CurNum] = x.Space().Member();
         Y[CurNum] = g.Space().Member();
      }
      S[CurNum]->Sub( xnext,x );
      Y[CurNum]->Sub( gnext,g );
      rho[CurNum] = 1.0/(S[CurNum]->Inner( *(Y[CurNum]) ));
      CurNum++;
   }
   else
   {
      for( i=0;i<MaxNum-1;i++ )
         rho[i] = rho[i+1];
      HCL_Vector_d * temp;
      temp = Y[0];
      for( i=0;i<MaxNum-1;i++ )
         Y[i] = Y[i+1];
      Y[MaxNum-1] = temp;
      Y[MaxNum-1]->Sub( gnext,g );

      temp = S[0];
      for( i=0;i<MaxNum-1;i++ )
         S[i] = S[i+1];
      S[MaxNum-1] = temp;
      S[MaxNum-1]->Sub( xnext,x );
      rho[MaxNum-1] = 1.0/(S[MaxNum-1]->Inner( *(Y[MaxNum-1]) ));
   }
   if( P == NULL )
      H0->Parameters().PutValue( "Scale",
               double(1.0/(rho[CurNum-1]*(Y[CurNum-1]->Norm2()))) );
   else
   {
      if( lsolver )
         lsolver->Solve( *P,*(Y[CurNum-1]),*temp_vec );
      else
         P->InvImage( *(Y[CurNum-1]),*temp_vec );
      H0->Parameters().PutValue( "Scale",
            double(1.0/(rho[CurNum-1]*(Y[CurNum-1]->Inner( *temp_vec )))) );
   }
}

double HCL_lmbfgsOp1_d::GetScale() const
{
   double scale;
   H0->Parameters().GetValue( "Scale",scale );
   return scale;
}

ostream & HCL_lmbfgsOp1_d::Write( ostream & str ) const
{
   str << "HCL_lmbfgsOp1_d with Domain " << endl;
   Domain().Write( str );
   str << "and Range " << endl;
   Range().Write( str );
   str << "CurNum: " << CurNum << " MaxNum: " << MaxNum << endl;
   double scale;H0->Parameters().GetValue( "Scale",scale );
   str << "Scale: " << scale << endl;
   str << "S: " << endl;
   int i;
   for( i=0;i<CurNum;i++ )
      S[i]->Write( str );
   str << "Y: " << endl;
   for( i=0;i<CurNum;i++ )
      Y[i]->Write( str );
   str << "rho:" << endl;
   for( i=0;i<CurNum;i++ )
      str << " " << rho[i] << endl;
   return str;
}
