/*
                 THE HILBERT CLASS LIBRARY                        
                Release 0.8, September 1997                       
                                                                  
       copyright (C) 1995, 1996, 1997 by Mark S. Gockenbach,      
              Matthew Petro, and William W. Symes                 
                                                                  
The source code contained in this file FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "Std_includes.h"
#include "Table.h"



Table::str_table::str_table(int size)
{
    assert(size > 0);
    table = new pair[size];
    table_size = 0;
    allocated_size = size;
}

Table::str_table::~str_table()
{
    for (int i = 0; i < table_size; ++i) {
       if (table[i].value)
	  delete[] table[i].value;
    }

    delete[] table;

}

int 
Table::str_table::put(const char *key, char *value)
{
   if (strlen(key) > KEY_LENGTH)
      cout << "Table::str_table::put: Warning: Parameter name too long: \""
           << key << '"' << endl;
   int ind;
   if (find(key, ind)) {
      if (table[ind].value)
	 delete[] table[ind].value;
      table[ind].value = value;
      return 0;
   }
   else {
      return insert(key, value, ind);
   }
}

char *
Table::str_table::get(const char *key) const
{
   if (strlen(key) > KEY_LENGTH)
      cout << "Table::str_table::get: Warning: Parameter name too long: \""
           << key << '"' << endl;
   int ind;
   if (find(key, ind)) 
      return table[ind].value;
   else
      return NULL;
}

char * Table::str_table::key(int & ind) {
   assert(ind <= table_size && ind >= 0);
   return table[ind].key;
}

int 
Table::str_table::insert(const char *key, char *value, int ind)
{
   assert(ind <= table_size && ind >= 0);
   if (table_size == allocated_size) {
//      cout << "resizing table " << endl;
      allocated_size *= 2;
      pair *table1 = new pair[allocated_size];
      for (int i = 0; i < table_size; ++i)
         table1[i] = table[i];
      delete[] table;
      table = table1;
   }
   
   for (int i = table_size-1; i >= ind; --i)
      table[i+1] = table[i];
   table[ind].value = value;
   strncpy(table[ind].key, key, KEY_LENGTH);
   table_size++;
   return 0;
}

int 
Table::str_table::find(const char *key, int &ind) const
{
   int a = 0, b = table_size-1;
   
   if (b < a) {
      ind = 0;
      return 0;
   }

   int res = strncmp(key, table[b].key, KEY_LENGTH);
   if (res > 0) {  // key > last elt.
      ind = b+1;
      return 0;
   }
   else if (res == 0) {  // key == last elt.
      ind = b;
      return 1;
   }
      
   
   while(1) {
      int m = (a+b)/2;
      int res = strncmp(key, table[m].key, KEY_LENGTH);
      if (res == 0) {     // key == table[m].key
         ind = m;
         return 1;
      }
      else if (res < 0) {   // key < table[m].key
	 if (b == m) {
	    ind = b;
	    return 0;
	 }
         b = m;
      }
      else {                // key > table[m].key
	 if (a == m) {
	    ind = a+1;
	    return 0;
	 }
         a = m;
      }
   }
}


Table::Table(char *fname, char *prefix_str, int size): prefixes(3),table(size)
{
   char w[KEY_LENGTH+1];
   char *pp = prefix_str;
   int numread;
   if (prefix_str) {
      while (sscanf(pp, "%s%n", w, &numread) > 0) {
         pp += numread;
         prefixes.put(w, NULL);
      }
   }
   Merge(fname);
}

Table::Table( const Table & SS ): prefixes(1),table(1) {
  Table & S = (Table&)SS;
  char * keytmp;
  char * valuetmp;
  int keylen;
  int valuelen;
  int j;
  for (int i=0;i<S.Size();i++) {
    keylen=strlen(S.Key(i))+1;
    keytmp=new char[keylen];
    for (j=0;j<keylen;j++) {
      keytmp[j]='\0';
    }
    strcpy(keytmp,S.Key(i));
    valuelen=strlen(S.strValue(keytmp))+1;
    valuetmp=new char[valuelen];
    for (j=0;j<valuelen;j++) {
      valuetmp[j]='\0';
    }
    strcpy(valuetmp,S.strValue(keytmp));
    //    table.put(keytmp,valuetmp);
    PutValue(keytmp,valuetmp);
    delete [] valuetmp;
    delete [] keytmp;
  }
}

Table::~Table(){}

int
Table::GetValue(const char *key, int &value) const
{
   assert(key);
   char *value_str = table.get(key);
   if (value_str == NULL)
      return 1;
   
   return !sscanf(value_str, "%d", &value);
}

int
Table::GetValue(const char *key, double &value) const
{
   assert(key);
   char *value_str = table.get(key);
   if (value_str == NULL)
      return 1;
   
   return !sscanf(value_str, "%lf", &value);
}

int
Table::GetValue(const char *key, float &value) const
{
   assert(key);
   char *value_str = table.get(key);
   if (value_str == NULL)
      return 1;
   
   return !sscanf(value_str, "%f", &value);
}

int
Table::GetValue(const char *key, char *value, int length) const
{
   assert(key);
   char *value_str = table.get(key);
   if (value_str == NULL)
      return 1;
   
   if (length > 0) {
      int len = strlen(value_str);
      if (length < len) {
         cout << "Table::GetValue: Warning: Truncating string for key \""<<key<<'\"'<<endl;
         strncpy(value, value_str, length-1);
         value[length] = '\0';
      }
      else {
         strcpy(value, value_str);
      }
   }
   else 
      strcpy(value, value_str);
   
   return 0;
}

int
Table::PutValue(const char *key, int value)
{
   assert(key);
   char * buf = new char[12];
   sprintf(buf, "%d", value);
   return table.put(key, buf);
}

int
Table::PutValue(const char *key, double value)
{
   assert(key);
   char * buf = new char[25];
   sprintf(buf, "%.15g", value);
   return table.put(key, buf);
}

int
Table::PutValue(const char *key, float value)
{
   assert(key);
   char * buf = new char[25];
   sprintf(buf, "%.15g", value);
   return table.put(key, buf);
}

int
Table::PutValue(const char *key, const char *value)
{
   assert(key);
   if (value) {
     char * buf = new char[strlen(value)+1];
     for (int i=0;i<strlen(value)+1; i++) {
       buf[i]='\0';
     }
     strcpy(buf,value);
     return table.put(key,buf);
   }
   return 1;
}


int
Table::GetArrayValue(const char *key, int ind, int &value) const
{
   assert(key);
   char newkey[KEY_LENGTH+12];
   sprintf(newkey, "%s%d", key, ind);
   return GetValue(newkey, value);
}

int
Table::GetArrayValue(const char *key, int ind, float &value) const
{
   assert(key);
   char newkey[KEY_LENGTH+12];
   sprintf(newkey, "%s%d", key, ind);
   return GetValue(newkey, value);
}

int
Table::GetArrayValue(const char *key, int ind, double &value) const
{
   assert(key);
   char newkey[KEY_LENGTH+12];
   sprintf(newkey, "%s%d", key, ind);
   return GetValue(newkey, value);
}

int
Table::GetArrayValue(const char *key, int ind, char *value, int length) const
{
   assert(value);
   assert(key);
   char newkey[KEY_LENGTH+12];
   sprintf(newkey, "%s%d", key, ind);
   return GetValue(newkey, value, length);
}


int
Table::PutArrayValue(const char *key, int ind, int value)
{
   assert(key);
   char newkey[KEY_LENGTH+12];
   sprintf(newkey, "%s%d", key, ind);
   return PutValue(newkey, value);
}

int
Table::PutArrayValue(const char *key, int ind, float value)
{
   assert(key);
   char newkey[KEY_LENGTH+12];
   sprintf(newkey, "%s%d", key, ind);
   return PutValue(newkey, value);
}

int
Table::PutArrayValue(const char *key, int ind, double value)
{
   assert(key);
   char newkey[KEY_LENGTH+12];
   sprintf(newkey, "%s%d", key, ind);
   return PutValue(newkey, value);
}

int
Table::PutArrayValue(const char *key, int ind, const char *value)
{
   assert(key);
   char newkey[KEY_LENGTH+12];
   sprintf(newkey, "%s%d", key, ind);
   return PutValue(newkey, value);
}

      

int
Table::Merge(char *fname)
{
   char buff[MAX_LINE_LENGTH];
   char name[KEY_LENGTH], prefix[KEY_LENGTH];

   if (fname == NULL)
      return 0;
   
   FILE *f = fopen(fname, "r");
   if (f == NULL) {
//      cout << "Table::Merge: Unable to open parameter file \"" << fname
//           << "\"." << endl;
      return 1;
   }

readloop:

   while (fgets(buff, MAX_LINE_LENGTH, f)) {
       if (strlen(buff) >= MAX_LINE_LENGTH-1) {
          cout << "Table::Merge: Input line too long; Aborting." << endl;
          break;
       }
       
       char *p = buff;
       while (isspace(*p))
           p++;
       
       char *p1 = strpbrk(p, ":= \t");
       if (p1 == NULL || p == p1)
          goto readloop;   // bad line
       
       if (*p1 == ':') {  // parameter name of form    prefix::name
         if (*(p1+1) != ':')
            goto readloop;  // bad line
         
         char t = *p1;
         *p1 = '\0';
         if (p1-p > KEY_LENGTH) {
            cout << "Table::Merge: Warning: prefix too long: " << p << endl;
         }
         strncpy(prefix, p, KEY_LENGTH);
         *p1 = t;

         
         int ind;
         if (!prefixes.find(prefix, ind)) {
            goto readloop;  // prefix doesnt match
	 }
         
         p1 += 2;
         
         char *p2 = strpbrk(p1, "= \t");
         if (p2 == NULL || p2 == p1)
            goto readloop;   // bad line
         
	 t = *p2;
         *p2 = '\0';
         if (p2 - p1 > KEY_LENGTH) {
            cout << "Table::Merge: Warning: name too long: " << p << endl;
         }
         strncpy(name, p1, KEY_LENGTH);
         *p2 = t;

	 p1 = p2;
       }
       else {  // parameter name of form     name
	 char t = *p1;
	 *p1 = '\0';
         if (p1-p > KEY_LENGTH) {
            cout << "Table::Merge: Warning: name too long: " << p << endl;
         }
         strncpy(name, p, KEY_LENGTH);
         *p1 = t;
       }
       
       while (isspace(*p1))
          p1++;
       if (*p1 != '=') {
          goto readloop;  // bad line
       }
       p1++;
       while (isspace(*p1))
          p1++;
       
       int end = strlen(p1)-1;
       while (end > 0 && isspace(*(p1 + end)))
          --end;
       p1[end+1] = '\0';
       
       
       if (*p1 == '\"' && *(p1 + end) == '\"') {  // remove any enclosing quotes
          *(p1 + end) = '\0';
           p1++;
       }
                 
       char *value = strdup(p1);

       
       table.put(name, value);
   }
   fclose(f);
   return 0;
}

ostream &
Table::Write(ostream&s) const
{
   
   s << "Table (";
   for (int i = 0; i < prefixes.table_size; ++i) {
      s << prefixes.table[i].key;
      if (i != prefixes.table_size-1) s << " ";
   }
   s << ")" << endl << "{" << endl;
   for (int j = 0; j < table.table_size; ++j) {
      s << "  " << table.table[j].key << " = " << table.table[j].value << endl;
   }
   s << "}" << endl;
    
   return s;
}

ostream& operator<<(ostream&s, Table&t) { return t.Write(s); }


  
