/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_BiLinearOp_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef _HCL_BILINEAROP_S_H
#define _HCL_BILINEAROP_S_H

#include "HCL_LinearOp_s.h"

/** {\bf HCL\_BiLinearOp\_s} is the base class for bilinear operators.  A
bilinear operator is of the form 
\TEX{
\[
B:X\times Y\rightarrow Z,
\]}
where X, Y, and Z are vector spaces and B(.,y) and B(x,.) are linear
operators.  Bilinear operators belong to the core of HCL because the
second derivative of a nonlinear operator (at a point) is a bilinear operator.

The primary class methods are:
\begin{itemize}
\item {\bf Domain1} Returns a reference to the domain of the first argument
of the operator.
\item {\bf Domain2} Returns a reference to the domain of the second argument
of the operator.
\item {\bf Range} Returns a reference to the range of the operator.
\item {\bf Image} Computes the action of the operator on a pair of vectors:
\begin{verbatim}
   B.Image( x,y,z );  // z <-- B(x,y)
\end{verbatim}
\item {\bf Op1} Creates the operator B(x,.).  Note that this operator
maps Y into Z.
\item {\bf Op2} Creates the operator B(.,y).  Note that this operator
maps X into Z.
\item {\bf Op3} Creates the operator x |--> B(x,.)'z as a linear operator
with adjoint.  Note that this operator maps X into Y.
\end{itemize}
*/
class HCL_BiLinearOp_s: public HCL_Base
{
   friend class HCL_RestrictedBiLinearOp_s;
   protected:

   /** The method {\bf PartialAdjImage} implements the mappings
       \TEX{$(x,z)\mapsto B(x,\cdot)^*z$} (i=1) and
       \TEX{$(y,z)\mapsto B(\cdot,y)^*z$} ($i=2$).  These methods are
       implemented as error messages in this base class.  If they are
       over-ridden in a derived class, then the default implementations
       of {\bf Op1}, {\bf Op2}, and {\bf Op3} may be used.  Otherwise,
       the implementor of the derived class is expected to over-ride
       {\bf Op1}, {\bf Op2}, and {\bf Op3}. */
      virtual void PartialAdjImage( int i,const HCL_Vector_s & xy,
                                          const HCL_Vector_s & z,
                                                HCL_Vector_s & yx ) const;
   public:  

   // Default constructor.
      HCL_BiLinearOp_s(){}
   // Copy constructor.
      HCL_BiLinearOp_s( const HCL_BiLinearOp_s & L ) {}
   // Destructor.
      virtual ~HCL_BiLinearOp_s(){}

   /** Access to the domain of the first argument.  Returns a reference
       to an {\bf HCL\_VectorSpace\_s}. Pure virtual: must be implemented 
       in any derived instantiable class. */
      virtual HCL_VectorSpace_s & Domain1() const = 0;
   /** Access to the domain of the second argument.  Returns a reference
       to an {\bf HCL\_VectorSpace\_s}. Pure virtual: must be implemented in any derived instantiable class. */
      virtual HCL_VectorSpace_s & Domain2() const = 0;
   /** Access to the range.  Returns a reference to a
       {\bf HCL\_VectorSpace\_s}. Pure virtual: must be implemented in any derived instantiable class.*/
      virtual HCL_VectorSpace_s & Range() const = 0;

   /** {\bf Image} computes the action of the operator on the pair (x,y),
       giving z. Pure virtual: must be implemented in any derived instantiable class. */
      virtual void Image( const HCL_Vector_s & x,const HCL_Vector_s & y,
                                HCL_Vector_s & z ) const = 0;
   /// {\bf Op1} creates the linear operator B(x,.). Implemented in this class.
      virtual HCL_LinearOp_s * Op1( const HCL_Vector_s & x ) const;
   /// {\bf Op2} creates the linear operator B(.,y). Implemented in this class.
      virtual HCL_LinearOp_s * Op2( const HCL_Vector_s & y ) const;
   /** Access to the "partial adjoint" of the bilinear operator.
       Assume B:XxY --> Z is the bilinear operator.
       Then {\bf Op3} returns the operator x |---> (B(x,.))'z.
       Its adjoint is the operator y |---> (B(.,y))'z.
       Implemented in this class. */
      virtual HCL_LinearOp_s * Op3( const HCL_Vector_s & z ) const;

   /** {\bf CheckAdj} invokes {\bf Op1}, {\bf Op2}, and {\bf Op3}
       at random points, and then tests them for consistency with
       their adjoints using the {\bf HCL\_LinearOp\_s} method
       {\bf CheckAdj}.  The flag {\bf Display} indicates how the results
       are to be displayed: 0 means no display to the screen, 1 means
       a message is sent to the screen only if there is an error, while
       2 means that the results are sent to the screen whether an error is
       detected or not.   The second input sets the tolerance for
       the floating-point tests; an error is reported if two results
       which ought to be equal differ by a relative amount greater
       than tol times machine epsilon. Implemented in this class. */
      int CheckAdj( int Display = 1,float tol = 1000 );

};

/** {\bf HCL\_RestrictedBiLinearOp\_s} creates one of the three
linear operators related to a bilinear operator.  Suppose
\TEX{\[
B:X\times Y\rightarrow Z
\]}
is a bilinear operator represented by an instance of
{\bf HCL\_BiLinearOp\_s}.  The three related operators are
\TEX{\[
L_1:Y\rightarrow Z,\ L_1:y\mapsto B(x,y),
\]}
\TEX{\[
L_2:X\rightarrow Z,\ L_2:x\mapsto B(x,y),
\]}
\TEX{\[
L_3:X\rightarrow Y,\ L_3:x\mapsto B(x,\cdot)^*z.
\]}
Note that to define \TEX{$L_1$}, a fixed x in X is required; similarly
y in Y defines \TEX{$L_2$}, and z in Z defines \TEX{$L_3$}.
Note also that the fourth natural operator,
\TEX{\[
L_4:Y\rightarrow X,\ L_4:y\mapsto B(\cdot,y)^*z
\]}
turns out to be the adjoint of \TEX{$L_3$}. 

This class is concrete: all of its methods are implemented.

*/

class HCL_RestrictedBiLinearOp_s: public HCL_LinearOp_s
{
   private:
      const HCL_BiLinearOp_s * Bop;
      HCL_Vector_s * fixed;   // "fixed" vector
      int OpFlag;             // OpFlag is 1, 2, or 3
   public:

   // Default constructor---disabled.
      HCL_RestrictedBiLinearOp_s();
   // Copy constructor---disabled.
      HCL_RestrictedBiLinearOp_s( const HCL_RestrictedBiLinearOp_s & );
   /** Usual constructor; needs a pointer to the underlying bilinear operator,
       a flag indicating which restricted operator is to be represented
       (1, 2, or 3), and the fixed vector.  Note that an independent copy
       of the fixed vector is made. */
      HCL_RestrictedBiLinearOp_s( const HCL_BiLinearOp_s * t,
                                  int of,const HCL_Vector_s & v );
   // Destructor.
      virtual ~HCL_RestrictedBiLinearOp_s();

  /** Domain space access. Returns a reference to the domain of
      the operator, which is an {\bf HCL\_VectorSpace\_s}. */
      virtual HCL_VectorSpace_s & Domain() const;
  /** Range space access. Returns a reference to the range of
      the operator, which is an {\bf HCL\_VectorSpace\_s}. */
      virtual HCL_VectorSpace_s & Range() const;
  /// Image compute the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_s & x,
                                HCL_Vector_s & y ) const;
  /** AdjImage computes the action of the adjoint on y, giving x. */
      virtual void AdjImage( const HCL_Vector_s & y,
                                   HCL_Vector_s & x ) const;

   /** {\bf InvImage} computes the action of the inverse on y, giving x.
      This method is implemented to return an error message. */
      virtual void InvImage(const HCL_Vector_s & y,
                                  HCL_Vector_s & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y.  This method is implemented to return an error message. */
      virtual void InvAdjImage(const HCL_Vector_s & x,
                                     HCL_Vector_s & y) const;
   /** {\bf Write} invokes the {\bf Write} method of the underlying
       bilinear operator. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
