/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_CMinAL_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_AUGLAGGRAD_S_
#define HCL_AUGLAGGRAD_S_

#include <iostream.h>
#include <stdlib.h>
#include "HCL_UMin_s.h"
#include "HCL_ALFcnl_s.h"
#include "HCL_Arch_s.h"

/** {\bf HCL\_CMinAL\_s} implements the augmented Lagrangian algorithm for an
equality constrained nonlinear programs.  The augmented Lagrangian
algorithm is a Sequential Unconstrained Minimization Technique, and is based on
a (user-provided) {\bf HCL\_UMin\_s} object, which performs the unconstrained
minimization.

The problem addressed by this algorithm is
\TEX{\begin{eqnarray*}
\min&f(x)&\\
s.t.& H(x)=0,&
\end{eqnarray*}}
where \TEX{$f$} is a real-valued function (an instance of
{\bf HCL\_Functional\_s}) and \TEX{$H$} is an operator (an instance of
{\bf HCL\_Op\_s}).

The augmented Lagrangian algorithm takes an initial estimate of the solution
and an initial estimate of the Lagrange multiplier \TEX{$\lambda$}, and
minimizes the augmented Lagrangian functional
\TEX{\[
L(x)=f(x)+(G(x),\lambda)+\frac{\rho}{2}(G(x),G(x))
\]}
(subject to no constraints).  The estimate of the Lagrange multiplier is
then updated and the process is repeated.  The iteration halts when the
gradient of the Lagrangian (the augmented Lagrangian without the penalty
(third) term) is approximately zero.  */
class HCL_CMinAL_s: public HCL_Base
{
   private:
      Table * ParamTable;

   /**@name Input Parameters
      In a parameter file, the names may be prepended by "CMinAL::"
      or simply "CMin::". */
   //@{
      /// (10) Maximum number of iterations.
      int MaxItn;
      /** (1e-4) Gradient tolerance.  The algorithm attempts to locate
          a point where the relative norm of the gradient of the Lagrangian
          is less than this value. */
      float GradTol;
      /** (1e-4) Constraint tolerance.  The algorithm attempts to locate
          a point where the norm of the constraint violation is less than
          this value. */
      float ConstrTol;
      /** (0) Display level.  This determines how much information should be
          displayed during the execution of the algorithm.  Possible values
          are: 0 - No output; 1 - Function value and gradient norm after
          final iteration; 2 - Function value and gradient norm after every
          iteration. */
      int DispFlag;
      /** (0) Dump level.  This determines how much information should be
          sent to the dump file during the execution of the algorithm.
          Possible values are: 0 - No output; 1 - Function value and
          gradient norm after final iteration; 2 - Function value and
          gradient norm after every iteration. */
      int DumpFlag;
      /** (HCL\_UMin\_lbfgs.DumpFile) Dump file name. */
      char DumpFile[81];
      /** (6) Display precision---the number of digits sent to the screen. */
      int DispPrecision;
      /** (6) Dump precision---the number of digits sent to the file. */
      int DumpPrecision;
      /** (1) Initial penalty parameter. */
      float InitPenParam;
      /** (4) Factor for increasing penalty parameter. */
      float PenParamIncFac;
      /** (1/4) Norm of the constraint violation must decrease by this
          factor at each step; otherwise, the penalty parameter is
          increased by the factor PenParamIncFac. */
      float IncPenaltyThreshold;
      /** (1e8) The penalty parameter will not be increased above
          this value. */
      float MaxPenParam;
      /** (9/10) Norm of the gradient of the Lagrangian must decrease by
          this factor at each step; otherwise, the gradient tolerance for
          the unconstrained minimizer is decreased by the factor
          UMinGradTolDecFac. */
      float GradDecThreshold;
      /** (1/2) Factor by which to decrease the gradient tolerance for
          the unconstrained minimizer if the gradient of the Lagrangian
          is not decreased. */
      float UMinGradTolDecFac;
      /** (1e-8) The unconstrained minimizer's GradTol will not be decreased
          below this value. */
      float UMinMinGradTol;
   //@}

      float Typf;
      float MinStep;
      int TraceSteps,SpecFlag,OldDispPrec;

      int Itn,TermCode;
      float fx,Gxnorm,Hxnorm,gradlnorm,PenParam,ConstrViolNorm;
      HCL_Vector_s * Hx;
      HCL_Vector_s * lambda;
      HCL_Vector_s * gradl;
      HCL_EvalALFcnl_s * Feval;

      HCL_UMin_s * umin;

      void Initialize( const HCL_Op_s & H );
      void Clean();
   /** Check stopping criterion.  Return value is nonzero if a stopping test
       has been met. */
      virtual int StoppingTest( HCL_EvalALFcnl_s & Feval );

   public:

   // Default constructor---disabled.
      HCL_CMinAL_s();
   /// Copy constructor---disabled.
      HCL_CMinAL_s( const HCL_CMinAL_s & x );
   /** The constructor takes a (pointer to an) {\bf HCL\_UMin\_s} object.
       Optionally, a file containing algorithmic parameters can be provided.
       If given, this file contains algorithmic parameters (such as stopping
       tolerances), a full list of which is given elsewhere in the
       documentation of this class.  For example, to set the maximum number of
       iterations to 100, the file should contain a line of the
       form CMinAL::MaxItn = 100 or, alternately, CMin::MaxItn = 100. */
      HCL_CMinAL_s( HCL_UMin_s * um,char * fname = NULL );
   // Destructor.
      virtual ~HCL_CMinAL_s();

   /** Access to parameter table.  Algorithmic parameters may be accessed
       with Parameters().GetValue( "NAME",val ) or changed with
       Parameters().PutValue( "NAME",val ). */
      virtual Table & Parameters() const { return *ParamTable; }

   /** Print current status.  The level of output is controlled by the
       parameter DispFlag---the larger the value of DispFlag, the more
       information is displayed. */
      virtual void Display();

   /** Return a reference to the last evaluation object, thus giving
       access to all of the information at the computed solution. */
      virtual HCL_EvalALFcnl_s & LastEval();
   /** {\bf Minimize} attempts to solve the problem of minimizing f(x) subject
       to the constraint H(x) = 0.  A starting guess must be provided.
       Currently the Lagrange multiplier is initialized to zero. */
      virtual int Minimize( HCL_Functional_s & f,HCL_Op_s & H,
                            HCL_Vector_s & x );

   /** Prints description of the object. */
      virtual ostream & Write(ostream &) const;
};

#endif
