/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_Functional_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_FUNCTIONAL_S_H_
#define HCL_FUNCTIONAL_S_H_

#include "HCL_LinearOp_s.h"
#include "HCL_Arch_s.h"
#include "maxmin_s.h"

// "Look-ahead declaration
class HCL_FunctionalDefaultHessian_s;

/** This class represents an "evaluation" object that will be created
by a instance of {\bf HCL\_Functional} object to implement the value, gradient,
and Hessian of the functional at a point.  That is, if the
{\bf HCL\_Functional} represents $f:X\rightarrow {\bf{\rm R}}$, the
the evaluation object represents the triple
         $(f(x),\nabla f(x),\nabla^2 f(x))$.
The purpose of having an evaluation object is to deal with the common
situation in which the functional value and derivatives share some
intermediate calculations. A typical implementation of such a functional
involves defining an evaluation object whose constructor performs these
intermediate calculations.  The {\bf Value} method then completes the
computation of the functional value, while the {\bf Gradient} and
{\bf Hessian} methods complete the calculation of the derivatives.

Note that if one is defining a functional for which there are no
(significant) common computations shared by $f(x)$, $\nabla f(x)$,
and $\nabla^2 f(x)$, then need not define an evaluation object explicitly.
Instead, the concrete class derived from {\bf HCL\_Functional} can create
an instance of {\bf HCL\_FunctionalDefaultEval}.  For more details, see
the documentation for {\bf HCL\_Functional}, and also the Technical
Report referenced at the end of this documentation.

There is also a special mechanism for implementing the Hessian operator.
The method {\bf Hessian} must create an instance of {\bf HCL\_LinearOp}
to represent $\nabla^2 f(x)$.  In many cases, this operator can be
implemented using existing HCL classes (such as the matrix operator
classes).  However, when this is not the case, one can use the default
class {\bf HCL\_FunctionalDefaultHessian} as follows.  In the class derived
from {\bf HCL\_EvaluateFunctional}, the virtual protected methods
{\bf HessianImage} and {\bf HessianInvImage} (if desired) are overridden
to implement the action of the Hessian operator and its inverse.
The method {\bf Hessian} is {\it not} overriddent in the derived class,
for then the default implementation will create an instance of
{\bf HCL\_FunctionalDefaultHessian} (which will contain a pointer to
the evaluation object and can then invoke {\bf HessianImage} and
{\bf HessianInvImage} as needed).

When an optimization (or other) algorithm is implemented, the functional
value, gradient, and Hessian should be accessed through
the {\bf ValueRef}, {\bf GradientRef}, and {\bf HessianRef} methods.
These are methods implemented in the base class (i.e. they do not need to
be redefined in a derived class).  The {\bf ValueRef} method invokes
{\bf Value} to get the function value, stores it, and sets a flag to indicate
that the value has already been computed.  On subsequent calls, {\bf ValueRef}
merely returns the stored value.  {\bf GradientRef} and {\bf HessianRef}
work in a similar fashion; note that these method manage the storage of
the gradient and Hessian, so the calling program does not need to allocate
storage.

Here is a summary of the primary methods of the class
{\bf HCL\_EvaluateFunctional}:
\begin{itemize}
\item {\bf Domain} Returns a reference to the domain of the underlying
functional.
\item {\bf Value} Computes the value of the functional at the point
at which this object was created.  Must be implemented by the creator
of a concrete class.
\item {\bf ValueRef} Calls {\bf Value}, if necessary, to compute
the functional value, saves it, and returns it.  Implemented in the
base class.
\item {\bf Gradient} Computes the gradient of the functional at the point
at which this object was created.  Must be implemented by the creator
of a concrete class.
\item {\bf GradientRef} Calls {\bf Gradient}, if necessary, to compute
the functional gradient, saves it, and returns a reference to it.
Implemented in the base class.
\item {\bf Hessian} Computes the Hessian of the functional at the point
at which this object was created.  Must be implemented by the creator
of a concrete class.
\item {\bf HessianRef} Calls {\bf Hessian}, if necessary, to compute
the functional Hessian, saves it, and returns a reference to it.
Implemented in the base class.
\end{itemize}

For more information about functional and evaluation classes, see 
    "Implementing functionals in HCL", Technical Report 99-24,
    Department of Computational and Applied Mathematics, Rice University,
    PO Box 1892, Houston, TX 77251-1892.
This report is available on the HCL home page.
*/

class HCL_EvaluateFunctional_s: public HCL_Base
{
   friend class HCL_FunctionalDefaultHessian_s;
   protected:
      int valFlag;
      float SavedValue;
      HCL_Vector_s * SavedGradient;
      HCL_LinearOp_s * SavedHessian;

      /** {\bf HessianImage} computes the action of the Hessian
          operator on the input {\bf dx}, yielding {\bf dy}.  The method
          {\bf HessianImage} (and, optionally,  {\bf HessianInvImage}) must be
          implemented in a derived class if the implementor wishes to
          implement the Hessian operator using {\bf
          HCL\_FunctionalDefaultHessian}. This method is implemented in
          this class as an error, and need not be overridden if
          {\bf HCL\_FunctionalDefaultHessian} is not to be used.  */
      virtual void HessianImage( const HCL_Vector_s & dx,
                                       HCL_Vector_s & dy ) const;
      /** {\bf HessianInvImage} computes the action of the inverse
          Hessian operator on the input {\bf dy}, yielding the output
          {\bf dx}.  This method can be overridden in a derived class
          if {\bf HCL\_FunctionalDefaultHessian} is used to implement
          the Hessian operator, and the inverse image of the operator
          is needed. This method is implemented in this class as an
          error, and need not be overridden if {\bf
          HCL\_FunctionalDefaultHessian} is not to be used.  */
      virtual void HessianInvImage( const HCL_Vector_s & dy,
                                          HCL_Vector_s & dx ) const;

   public:

   // Default constructor
      HCL_EvaluateFunctional_s();
   // Copy constructor
      HCL_EvaluateFunctional_s( const HCL_EvaluateFunctional_s & );
   // Destructor
      virtual ~HCL_EvaluateFunctional_s();

   /** Domain returns a reference to the domain of the underlying
      functional. Pure virtual: must be implemented in any derived
      instantiable class. */

      virtual HCL_VectorSpace_s & Domain() const = 0;

   /** {\bf Value} computes the value of the function.  This function
      must be implemented in each concrete derived class.  It is
      recommended that any calculations that are required for both the
      value and derivatives be done in the constructor; {\bf Value}
      should then complete the computation of the functional
      value. Pure virtual: must be implemented in any derived
      instantiable class.*/

      virtual float Value() const = 0;
   /** {\bf ValueRef} calls {\bf Value} to compute the value of the
       functional at the given point, saves it, and returns it.
       Successive calls to {\bf ValueRef} will retrieve the stored
       value.  This method is implemented in this (the base) class,
       and need not be redefined in a concrete derived class. This
       method is implemented in this class. */

      virtual float ValueRef() const;

   /** {\bf Gradient} computes the gradient of the function.  It is
       recommended that any calculations that are required for both
       the value and derivatives be done in the constructor; {\bf
       Gradient} should then complete the computation of the
       gradient. Pure virtual: must be implemented in any derived
       instantiable class. */

      virtual void Gradient( HCL_Vector_s & g ) const = 0;
   /** {\bf GradientRef} computes the gradient of the function and returns
       a reference to it.  This causes the Evaluation object to manage
       the storage. This method is implemented in this class. */

      virtual const HCL_Vector_s & GradientRef() const;

   /** {\bf Hessian} computes the Hessian of the function.  It is
       recommended that any calculations that are required for both
       the value and derivatives be done in the constructor; {\bf
       Hessian} should then complete the computation of the
       Hessian. Note that {\bf Hessian} is implemented in the base
       class to create an instance of {\bf HCL\_FunctionalDefaultEval}
       and need not be overridden in a derived class unless there is a
       better way to implement the Hessian.  Note, however, that if
       {\bf Hessian} is not overridden, then the Hessian operator
       cannot be used unless the protected methods {\bf HessianImage}
       and {\bf HessianInvImage} (if desired) are overridden in the
       derived class.  */

      virtual HCL_LinearOp_s * Hessian() const;

   /** HessianRef computes the Hessian of the function and returns a
       reference to it.  This causes the Evaluation object to manage
       the storage. This method is implemented in this class. */

      virtual const HCL_LinearOp_s & HessianRef() const;
};

// "Look-ahead" declaration.
class HCL_FunctionalDefaultEval_s;

/** {\bf HCL\_Functional} is the base class for all real-valued
functions.
The purpose of this class is to represent objective functions
in optimization problems.

The primary methods of this class are
\begin{itemize}
\item {\bf Domain} Returns a reference to the domain of the functional,
      which is a vector space.  This is used primarily for error-checking;
      for example:
\begin{verbatim}
   if( x.Space() != f.Domain() )
      // Error condition
\end{verbatim}
\item {\bf Value} Returns the value of the functional at a point.
      Example:
\begin{verbatim}
   float fx = f.Value( x );
\end{verbatim}
\item {\bf Gradient} Computes the gradient of the functional at a point.
      Example:
\begin{verbatim}
   HCL_Vector_s * g = f.Domain().Member(); // Get a vector to hold the gradient
   f.Gradient( x,*g ); // Fill *g with the gradient of f at x
\end{verbatim}
\item {\bf Hessian} Computes the Hessian of the functional at a point.
      Example:
\begin{verbatim}
   HCL_LinearOp_s * L = f.Hessian( x ); // Get the Hessian operator at x.
\end{verbatim}
\item {\bf Evaluate} Returns an "evaluation object", that is, an object
      of class {\bf HCL\_EvaluateFunctional_s}, which knows the value,
      gradient, and Hessian of the functional at given point.  Typically, this
      method should be used to compute the value, gradient, and/or Hessian
      of the functional, rather than {\bf Value}, {\bf Gradient}, and/or
      {\bf Hessian}.  For example, implementors of optimization algorithms
      should use {\bf Evaluate} unless it is known that only one of the
      value, gradient, Hessian will be used at a given point.  Otherwise,
      the optimization algorithm is failing to take advantage of possible
      efficiencies that may be built into a given functional class.
\end{itemize}

There are several ways to implements functionals; the various options are
described in detail, with examples, in the Technical Report referenced at
the end of this documentation.  In brief, the basic issue is that a functional
class must have a corresponding evaluation class (derived from
{\bf HCL\_EvaluateFunctional}).  The two basic options are:
\begin{itemize}
\item Define a single class, derived from {\bf HCL\_Functional}, and use
the default evaluation class {\bf HCL\_FunctionalDefaultEval}.  This is
done automatically if the virtual function {\bf Evaluate} is not overridden
in the derived functional class.  However, the implementor of the derived
class must then override the protected methods {\bf Value1}, {\bf Gradient1},
and {\bf HessianImage} (and {\bf HessianInvImage}, if desired) or
{\bf Hessian1}.  This is the method of choice if there is no efficiency
to be gained by sharing intermediate computations among the value, gradient,
and Hessian.
\item Define a class derived from {\bf HCL\_Functional} and another
class derived from {\bf HCL\_EvaluateFunctional}.  Override the virtual
function {\bf Evaluate} in {\bf HCL\_Evaluate} to create an instance of
the evaluation class.  This is the method of choice if there is significant
gain in sharing intermediate computations among the value, gradient, and
Hessian.  The user-defined evaluation class provides a place to store
these intermediate calculations.
\end{itemize}

For more information about functional and evaluation classes, see 
    "Implementing functionals in HCL", Technical Report 99-24,
    Department of Computational and Applied Mathematics, Rice University,
    PO Box 1892, Houston, TX 77251-1892.
This report is available on the HCL home page.
*/

class HCL_Functional_s: public HCL_Base
{
   friend class HCL_FunctionalDefaultEval_s;

   protected:

   /** {\bf Value1} computes the value of the functional at {\bf
     x}. Default implementation returns an error message and exits.
     The protected methods {\bf Value1}, {\bf Gradient1}, and {\bf
     Hessian1} or {\bf HessianImage}, {\bf HessianInvImage} must be
     implemented in a derived class if the implementor wishes to use
     {\bf HCL\_FunctionalDefaultEval} for the evaluation object.  */

   virtual float Value1( const HCL_Vector_s & x ) const; 
   
   /** {\bf Gradient1} computes the gradient of the functional at {\bf
     x} and stores it in {\bf y}. Default implementation returns an
     error message and exits. The protected methods {\bf Value1}, {\bf
     Gradient1}, and {\bf Hessian1} or {\bf HessianImage}, {\bf
     HessianInvImage} must be implemented in a derived class if the
     implementor wishes to use {\bf HCL\_FunctionalDefaultEval} for
     the evaluation object.  */

   virtual void Gradient1( const HCL_Vector_s & x,HCL_Vector_s & y ) const; 

   /** {\bf HessianImage} computes the action of the Hessian of the
     functional (at {\bf x}) on the input {\bf dx}, giving {\bf dy}.
     Default implementation returns an error message and exits. To
     obtain the functionality of this method in a derived class, only
     one of {\bf Hessian1} and the pair {\bf HessianImage}, {\bf
     HessianInvImage} should be implemented. The protected methods
     {\bf Value1}, {\bf Gradient1}, and {\bf Hessian1} or {\bf
     HessianImage}, {\bf HessianInvImage} must be implemented in a
     derived class if the implementor wishes to use {\bf
     HCL\_FunctionalDefaultEval} for the evaluation object.  */

   virtual void HessianImage( const HCL_Vector_s & x, 
			      const HCL_Vector_s & dx, 
			      HCL_Vector_s & dy ) const;

   /** {\bf HessianInvImage} computes the action of the inverse
     Hessian of the functional (at {\bf x}) on the input {\bf dy},
     giving {\bf dx}. Default implementation returns an error message
     and exits. Note that if the inverse Hessian is never to be used,
     this method can be implemented as an error.  To obtain the
     functionality of this method in a derived class, you must
     override this default implementation. Only one of {\bf Hessian1}
     and the pair {\bf HessianImage}, {\bf HessianInvImage} should be
     implemented. The protected methods {\bf Value1}, {\bf Gradient1},
     and {\bf Hessian1} or {\bf HessianImage}, {\bf HessianInvImage}
     must be implemented in a derived class if the implementor wishes
     to use {\bf HCL\_FunctionalDefaultEval} for the evaluation
     object.  */

   virtual void HessianInvImage( const HCL_Vector_s & x, 
				 const HCL_Vector_s & dy,
				 HCL_Vector_s & dx ) const; 

   /** {\bf Hessian1} creates an instance of {\bf HCL\_LinearOp} to
     represent the Hessian of the functional at {\bf x}, and returns a
     pointer to it. Note that the calling program must delete it using
     {\bf HCL\_delete} when it is no longer needed.  Only one of {\bf
     Hessian1} and the pair {\bf HessianImage}, {\bf HessianInvImage}
     should be implemented. The protected methods {\bf Value1}, {\bf
     Gradient1}, and {\bf Hessian1} or {\bf HessianImage}, {\bf
     HessianInvImage} must be implemented in a derived class if the
     implementor wishes to use {\bf HCL\_FunctionalDefaultEval} for
     the evaluation object.  */

   virtual HCL_LinearOp_s * Hessian1( const HCL_Vector_s & x ) const; 

public:  

   // Default constructor. 
      HCL_Functional_s(){}
   // Copy constructor.
      HCL_Functional_s( const HCL_Functional_s & ){}
   // destructor
      virtual ~HCL_Functional_s(){}

   /** Domain space access.  Returns a reference to the {\bf
     HCL\_VectorSpace\_s} which forms the domain of the functional.
     Note that, in the current design, the domain of a functional must
     be an entire vector space, although many nonlinear functionals
     are only defined on an open subset of a vector space.  This is a
     flaw in the design of HCL which may be corrected in some future
     version of the software.  A partial work-around to this problem
     is provided by the {\bf MaxStep} method.  Pure virtual: must be
     implemented in any derived instantiable class.  */

   virtual HCL_VectorSpace_s & Domain() const = 0;

    /** {\bf MaxStep} computes the largest value of {\bf a} such that
      {\bf x+a*dir} lies in the domain of definition of the
      functional.  By default (i.e. unless this virtual
      function is implemented in a derived class), the domain is
      assumed to be the whole space, and MaxStep return the largest
      floating point number.  This method is used in some of the
      optimization algorithms and provides a partial solution to the
      problem that many functionals are not defined on an entire
      vector space, but rather only on a subset. */

   virtual float MaxStep( const HCL_Vector_s & x, 
			   const HCL_Vector_s & dir) const;

   /** {\bf Evaluate} creates an "evaluation" object, which knows how
     to compute the function value, gradient,and Hessian at the
     given x.  The purpose of using {\bf Evaluate} (rather than {\bf
     Value}, {\bf Gradient}, and {\bf Hessian}) is for efficiency;
     many functionals are such that the evaluation of the functional
     and its derivatives require common intermediate calculations.
     The evaluation object mechanism gives the functional class
     implementor a place to store these intermediate results so that
     they do not have to be re-computed if the the gradient is
     requested at some point after the value, for example.  The
     implementor of an optimization algorithm should respect this by
     using {\bf Evaluate} to get a value, gradient, and/or Hessian
     rather than using {\bf Value}, {\bf Gradient}, and/or {\bf
     Hessian}.  Note as explained in the class documentation, if
     there is no advantage in having an evaluation object, then the
     implementor can simply not override {\bf Evaluate} in the
     derived class, in which case the default implementation will
     create an instance of {\bf HCL\_FunctionalDefaultEval}.  In
     this case, the implementor must override the virtual protected
     methods {\bf Value1}, {\bf Gradient1}, and either {\bf
     Hessian1} or {\bf HessianImage}, {\bf HessianInvImage}.  */

   virtual HCL_EvaluateFunctional_s * 
                  Evaluate( const HCL_Vector_s & x ) const;

   /** {\bf Value} evaluates the function at x.  It does so by
     creating an evaluation object with the {\bf Evaluate} method,
     getting the value from it, deleting the evaluation object, and
     returning the value.  {\bf Value} is defined in this (the base)
     class, so it need not be redefined in a derived class. */

   virtual float Value( const HCL_Vector_s & x ) const;

   /** {\bf Gradient} computes the gradient at x.  It does so by creating
     an evaluation object with the {\bf Evaluate} method, getting the
     gradient from it, and deleting the evaluation object. {\bf Gradient}
     is defined in this (the base) class, so it need not be redefined in
     a derived class. */

   virtual void Gradient( const HCL_Vector_s & x,
			  HCL_Vector_s & g ) const;

   /** {\bf Hessian} computes the gradient at x.  It does so by
     creating an evaluation object with the {\bf Evaluate} method,
     getting the Hessian from it, and deleting the evaluation
     object. {\bf Hessian} is defined in this (the base) class, so
     it need not be redefined in a derived class. */

   virtual HCL_LinearOp_s * Hessian( const HCL_Vector_s & x ) const;

   /** Scan computes the value of the function at N points.  The
     points are equally spaced and lie on the line segment from
     x+hmin*dx to x+hmax*dx.  The results are written to the file
     fname (if provided) and to the screen. This method is implemented
     in this class. */

   void Scan( const HCL_Vector_s & x,
		 const HCL_Vector_s & dx, 
		 int N,
		 float hmin,
		 float hmax,
		 char * fname = NULL );

   /** Check analytic gradient.  This method computes the relative difference
     between the directional derivative as computed from the analytic
     gradient and the same estimated by a central difference formula.
     This is done at n equally-spaced points on the line segment
     [x+hmin*y,x+hmax*y].  The parameters hmin, hmax, and n are optional
     and default to 0.1, 1.0, and 10, respectively.
     If the analytic gradient is correctly implemented and the function
     is three times continuously differentiable, then these errors
     will lie on a parabola when plotted against the step size. The
     results are sent to the output stream str. This method is implemented
     in this class. */

   int CheckGrad( const HCL_Vector_s & x,
		  const HCL_Vector_s & y,
		  ostream & str,
		  int n=10,
		  float hmin=0.1,
		  float hmax=1.0 );

   /** Check analytic Hessian.  This method computes the relative difference
     between the second directional derivative as computed from the analytic
     Hessian and the same estimated by a central difference formula.
     This is done at n equally-spaced points on the line segment
     [x+hmin*y,x+hmax*y].  The parameters hmin, hmax, and n are optional
     and default to 0.1, 1.0, and 10, respectively.
     If the analytic Hessian is correctly implemented and the function
     is four times continuously differentiable, then these errors
     will lie on a parabola when plotted against the step size. The
     results are sent to the output stream str. This method is implemented
     in this class. */

   int CheckHess( const HCL_Vector_s &,const HCL_Vector_s &,
                     ostream & str,int n=10,float hmin=0.1,float hmax=1.0 );

};

/** The purpose of the class {\bf HCL\_FunctionalDefaultEval} is to
  relieve the implementor of a derived functional class from the
  necessity of also defining a derived evaluation class if there is no
  need to do so.  The purpose of the evaluation class is to represent
  the triple $(f(x),\nabla f(x),\nabla^2 f(x))$.  If the implementor of
  the derived functional class has overridden the protected methods {\bf
  Value1}, {\bf Gradient1}, and either {\bf Hessian1} or {\bf
  HessianImage}, {\bf HessianInvImage}, then this default evaluation
  class can call those methods to compute the value, gradient, or
  Hessian of the functional.  To make this possible, the constructor of
  {\bf HCL\_FunctionalDefaultEval} saves a copy of {\bf x} and a pointer
  to the functional class.

  See the documentation for {\bf HCL\_EvaluateFunctional} for more information
  about evaluation objects for functionals. 

  This class is concrete: all of its methods are implemented.

*/

class HCL_FunctionalDefaultEval_s: public HCL_EvaluateFunctional_s
{
   private:
      HCL_Vector_s * x;
      const HCL_Functional_s * f;
   protected:

   /** {\bf HessianImage} computes the image of the Hessian on dx,
       yielding dy.  This method calls the corresponding method from
       the underlying functional class. */
      virtual void HessianImage( const HCL_Vector_s & dx,
                                       HCL_Vector_s & dy ) const
      {
         f->HessianImage( *x,dx,dy );
      }
   /** {\bf HessianInvImage} computes the image of the inverse Hessian
       on {\bf dy}, yielding {\bf dx}.  This method calls the
       corresponding method from the underlying functional class. */

      virtual void HessianInvImage( const HCL_Vector_s & dy,
                                          HCL_Vector_s & dx ) const
      {
         f->HessianInvImage( *x,dy,dx );
      }

   public:

   // Default constructor---disabled.
      HCL_FunctionalDefaultEval_s();
   // Copy constructor---disabled.
      HCL_FunctionalDefaultEval_s( const HCL_FunctionalDefaultEval_s & );
   /** Usual constructor; requires a pointer to a functional and a reference
       to the input vector. */
      HCL_FunctionalDefaultEval_s( const HCL_Functional_s * ff,
                                   const HCL_Vector_s & in );
   // Destructor
      virtual ~HCL_FunctionalDefaultEval_s();

   /** {\bf Domain} returns a reference to the domain of the underlying
       functional. */
      virtual HCL_VectorSpace_s & Domain() const { return f->Domain(); }

   /// {\bf Value} computes the value of the functional at {\bf x}.
      virtual float Value() const;
   /** {\bf Gradient} computes the gradient of the functional at {\bf x}
       and stores it in {\bf y}. */

       virtual void Gradient(HCL_Vector_s & y ) const;

   /** {\bf Hessian} computes the Hessian of the functional.  Note
       that this method constructs a {\bf HCL\_LinearOp\_s}, which the
       calling program is responsible for deleting with {\bf
       HCL\_delete}. */

      virtual HCL_LinearOp_s * Hessian() const;
 
   /** Debugging information. */
      virtual ostream & Write( ostream & str ) const;
};

/** The purpose of the class {\bf HCL\_FunctionalDefaultHessian} is to relieve
the implementor of a derived functional class from the necessity of also
defining a class to represent the Hessian if there is no existing HCL
class that can be used for this purpose.  The implementor need only,
in either the derived functional class or the corresponding derived
evaluation class, implement the methods {\bf HessianImage} and (if desired)
{\bf HessianInvImage}.  The constructor of {\bf HCL\_FunctionalDefaultHessian}
saves a pointer to the evaluation class, and refers calls to {\bf Image} and
{\bf InvImage} back to {\bf HessianImage} and {\bf HessianInvImage},
respectively. (Note that a Hessian is symmetric, so calls to the adjoint
methods are handled similarly.) */

class HCL_FunctionalDefaultHessian_s: public HCL_LinearOp_s
{
   private:
       const HCL_EvaluateFunctional_s * eval;
   public:  

  // Default constructor---disabled.
      HCL_FunctionalDefaultHessian_s();
  // Copy constructor---disabled.
      HCL_FunctionalDefaultHessian_s(
                    const HCL_FunctionalDefaultHessian_s & L );
  /** Usual constructor; just needs a functional's evaluation object. */
      HCL_FunctionalDefaultHessian_s( const HCL_EvaluateFunctional_s * ev );
  // destructor
      virtual ~HCL_FunctionalDefaultHessian_s();

  /** Domain space access. Returns a reference to the domain of
      the functional, which is a {\bf HCL\_VectorSpace\_s}. */
      virtual HCL_VectorSpace_s & Domain() const;
  /** Range space access. Returns a reference to the range of
      the functional, which is a {\bf HCL\_VectorSpace\_s}. */
      virtual HCL_VectorSpace_s & Range() const;
  /// Image compute the action of the functional on x, giving y.
      virtual void Image( const HCL_Vector_s & x,
                                HCL_Vector_s & y ) const;
   /// AdjImage computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_s & y,
                                   HCL_Vector_s & x ) const;

   /** {\bf InvImage} computes the action of the inverse on y, giving
       x.  Note that this method is enabled only if the underlying
       functional class has a working {\bf HessianInvImage} method. */

      virtual void InvImage(const HCL_Vector_s & y,
                                  HCL_Vector_s & x) const;

   /** {\bf InvAdjImage} computes the action of the inverse adjoint on
       x, giving y. Note that this method is enabled only if the
       underlying functional class has a working {\bf
       HessianInvAdjImage} method. */

      virtual void InvAdjImage(const HCL_Vector_s & x,
                                     HCL_Vector_s & y) const;
   /** Debugging information.  */
      virtual ostream & Write( ostream & str ) const;
};

#endif











