/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_GenericProductVector_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_GenericProductVector_d_H
#define HCL_GenericProductVector_d_H

#include <stdlib.h>
#include "HCL_ProductVector_s.h"

class HCL_GenericProductVector_s;

/** {\bf HCL\_GenericProductSpace\_s} is a concrete product vector space
class; it represents a vector space which is defined as a product of
several existing vector spaces.

There are three ways to create such a class.  There is a special constructor
for the case of two factors; pointers to the two vector spaces are
passed to this constructor.
\begin{verbatim}
   MySpace1_s X; // MySpace1_s and MySpace2_s are concrete classes
   MySpace2_s Y; // derived from HCL_VectorSpace_s
   HCL_GenericProductSpace_s Z( &X,&Y );

   // Z is now ready to use
   .
   .
   .
\end{verbatim}

There is also a special constructor for the case of a product space with
n identical factors:
\begin{verbatim}
   MySpace1_s X; // MySpace1_s is a concrete class derived
                 // from HCL_VectorSpace_s
   HCL_GenericProductSpace_s Z( &X,10 );

   // Z is now ready to use
   .
   .
   .
\end{verbatim}

The general constructor just takes the number of factors; the individual
factors must then be specified using the {\bf SetNext} method.
\begin{verbatim}
   MySpace1_s X; // MySpace1_s, MySpace2_s, MySpace3_s are concrete classes
   MySpace2_s Y; // derived from HCL_VectorSpace_s
   MySpace3_s Z;
   HCL_GenericProductSpace_s W( 3 );
   W.SetNext( &X );
   W.SetNext( &Y );
   W.SetNext( &Z );

   // W is now ready to use
   .
   .
   .
\end{verbatim}

See {\bf HCL\_GenericProductVector\_s} and the base class
{\bf HCL\_ProductSpace\_s} for more details.
 */
class HCL_GenericProductSpace_s : public HCL_ProductSpace_s
{
   private:
      static char Type;
   protected:
      int len;
      HCL_VectorSpace_s ** SpaceList;
   public:  

   // Default constructor---disabled.
      HCL_GenericProductSpace_s();
   // Copy constructor.
      HCL_GenericProductSpace_s(const HCL_GenericProductSpace_s & x);
   /** This constructor creates a product space with n factors.  The
       factors must be specified with the {\bf SetNext} or {\bf Set}
       methods. */
      HCL_GenericProductSpace_s( int n );
   /** Provided for convenience; creates a product space with 2 factors. */
      HCL_GenericProductSpace_s( const HCL_VectorSpace_s & S1,
                                 const HCL_VectorSpace_s & S2 );
   /** Provided for convenience; creates a product space with 2 factors
       (copies pointers). */
      HCL_GenericProductSpace_s( HCL_VectorSpace_s * S1,
                                 HCL_VectorSpace_s * S2 );
   /** This constructor creates a product space with n identical factors. */
      HCL_GenericProductSpace_s( const HCL_VectorSpace_s & S,int n );
   /** This constructor creates a product space with n identical factors
       (copies pointers). */
      HCL_GenericProductSpace_s( HCL_VectorSpace_s * S,int n );
   // Destructor.
      virtual ~HCL_GenericProductSpace_s();

   /** {\bf Initialized} is a boolean method indicating whether the factors
       in the space have been initialized. */
      int Initialized() const;
   /** {\bf IsSet} is a boolean method indicating whether the indicated
       factor has been set. */
      int IsSet( int i ) const;
   /** Get type of vector space for comparison.
       GetType makes it possible to check that two vector spaces
       are of the same derived class before comparing them. */
      virtual void * GetType() const { return (void*)&Type; }

   /** Virtual constructor.  {\bf Member} creates a vector from the
       corresponding vector class.  Note that this is a dynamically allocated
       object that must be deleted, using {\bf HCL\_delete}, by the calling
       program. */
      HCL_Vector_s * Member() const;
   /** Virtual copy constructor. {\bf Replicate} creates a copy of this
       vector space.  Note that this is a dynamically allocated
       object that must be deleted, using {\bf HCL\_delete}, by the calling
       program. */
      virtual HCL_VectorSpace_s * Replicate() const;

   /// {\bf Number} returns the number of factors.
      virtual int Number() const { return len; }
   /// The () operator returns a reference to the ith factor.
      HCL_VectorSpace_s & operator()( int i ) const;
   /** SetNext allows the user to specify the next factor in the product. */
      void SetNext( HCL_VectorSpace_s * S );
   /** {\bf Set} allows the user to specify an arbitrary factor in the
       product. */
      void Set( int i,HCL_VectorSpace_s * S );

   /// Write prints out the factors in the product.
      virtual ostream & Write(ostream & str) const;
};

/** {\bf HCL\_GenericProductVector\_s} is a concrete product vector
class; it represents a vector from a vector space which is defined as a
product of several existing vector spaces.

There are three ways to create such a vector.

First of all, if x and y are two existing vectors (i.e. instances of
classes derived from {\bf HCL\_Vector\_s}), then the vector z = (x,y)
can be created as follows:
\begin{verbatim}
   HCL_GenericProductVector_s z( &x,&y );
\end{verbatim}
Note that using this construction, z points to x and y, and any change
in z is also a change in x and/or y, and vice versa.

Secondly, if x1, x2, ... xn are existing vectors, then the vector
z = (x1,x2,...,xn) can be created as follows:
\begin{verbatim}
   HCL_GenericProductVector_s z( n );  // n is the number of factors
   z.SetNext( &x1 );
   z.SetNext( &x2 );
   .
   .
   .
   z.SetNext( &xn );
\end{verbatim}
Again, z points to the vector x1, x2, ..., xn, and any changes in one
is a change in the other.

Finally, the general constructor takes an instance of
{\bf HCL\_GenericProductSpace\_s} (either a reference or a pointer), and
creates a vector from that space. */

class HCL_GenericProductVector_s: public HCL_ProductVector_s
{
   protected:
      HCL_GenericProductSpace_s * sp;
      HCL_Vector_s ** vlist;
      int len;
   public:

   /** Default constructor---for use by derived classes that wish to
       assign the space and length. */
      HCL_GenericProductVector_s();
   // Copy constructor---disabled.
      HCL_GenericProductVector_s(const HCL_GenericProductVector_s & p);
   /** Constructor to create a vector from the space. Creates an independent
       copy of the space. */
      HCL_GenericProductVector_s(const HCL_GenericProductSpace_s & p);
   /// Constructor to create a vector from the space (copies pointer).
      HCL_GenericProductVector_s( HCL_GenericProductSpace_s * p );
   /** Creates a vector from a product space with two factors.  This
       object copies pointers to the two given vectors. */
      HCL_GenericProductVector_s( HCL_Vector_s * v1,HCL_Vector_s * v2 );
   /** This constructor creates a product vector with n factors.  The
       factors must be specified with the {\bf SetNext} method. */
      HCL_GenericProductVector_s( int n );
   /// Destructor.
      virtual ~HCL_GenericProductVector_s();

   /** {\bf Initialized} is a boolean method indicating whether the factors
       in the vector have been initialized. */
      int Initialized() const;
   /** {\bf IsSet} is a boolean method indicating whether the indicated
       factor has been set. */
      int IsSet( int i ) const;

   /// The () operator returns a reference to the ith factor.
      virtual HCL_Vector_s & operator()(int i) const { return *(vlist[i-1]); }
   /** {\bf Space} returns a reference to the space to which this vector
       belongs. */
      virtual HCL_VectorSpace_s & Space() const { return *sp;}
   /// {\bf Number} returns the number of factors in the product.
      virtual int Number() const { return sp->Number(); }

   /** {\bf SetNext} specifies the next vector in the product.  This is
       used with the constructor that takes only the number of vectors
       in the product. */
      void SetNext( HCL_Vector_s * x );
   /** {\bf Set} specifies an arbitrary vector in the product.  This is
       used with the constructor that takes only the number of vectors
       in the product.  */
      void Set( int i,HCL_Vector_s * x );
};
  
#endif




