/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_LinCombFcnl_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_LINCOMBFCNL_S_H_
#define HCL_LINCOMBFCNL_S_H_

#include "Table.h"
#include "HCL_Functional_s.h"
#include "HCL_LinCombLinearOp_s.h"

/** This class represents the "evaluation" object for
{\bf HCL\_LinCombFcnl\_s}, which implements a linear combination of
real-valued functionals:
      $F(x)=\alpha_1 f_1(x) + \alpha_2 f_2(x) +\ldots+ \alpha_n f_n(x)$.
The evaluation object is an object knowing
how to compute the triple
      $(F(x),\nabla F(x),\nabla^2 F(x))$,
i.e. the value, gradient, and Hessian of the functional at a single
point.  For more details, see {\bf HCL\_LinCombFcnl\_s} and
{\bf HCL\_EvaluateFunctional\_s}. */

class HCL_EvalLinCombFcnl_s: public HCL_EvaluateFunctional_s
{
   private:
      int num;
      int numinit;
      HCL_VectorSpace_s * dom;
      HCL_EvaluateFunctional_s ** evallist;
      float * weights;
   public:

   // Default constructor---disabled.
      HCL_EvalLinCombFcnl_s();
   // Copy constructor---disabled.
      HCL_EvalLinCombFcnl_s( const HCL_EvalLinCombFcnl_s & );
   /** Usual constructor; takes the domain and the number of terms.  Each of
       the terms in the linear combination, consisting of a scalar and
       a {\bf HCL\_EvaluateFunctional\_s}, must be added with the
       {\bf SetNext} method. */
      HCL_EvalLinCombFcnl_s( HCL_VectorSpace_s * d,int n );
   // Destructor
      virtual ~HCL_EvalLinCombFcnl_s();

   /** Domain space access.  Returns a reference to the domain of the
       functional, which is required to be a vector space. */
      virtual HCL_VectorSpace_s & Domain() const { return *dom; }

   /// Value computes the value of the function.
      virtual float Value() const;
   /// Gradient computes the gradient of the function.
      virtual void Gradient( HCL_Vector_s & g ) const;
   /// Hessian computes the Hessian of the function.
      virtual HCL_LinearOp_s * Hessian() const;

   /** SetNext allows the next term in the linear combination to be added.
       This method takes the evaluation object and the weight. */
      void SetNext( HCL_EvaluateFunctional_s * eval,float w );
   /** {\bf GetTerm} extracts one of the evaluation objects from the
       linear combination.  In this way, one can efficiently access
       the individual values, gradients, and Hessian of the terms in the
       linear combination. */
      const HCL_EvaluateFunctional_s & GetTerm( int ) const;

   /** Debugging information.  Prints out the weights and the individual
       evaluation objects. */
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_LinCombFcnl\_s} represents a linear combination of
real-valued functionals:
      $F(x)=\alpha_1 f_1(x) + \alpha_2 f_2(x) +\ldots+ \alpha_n f_n(x)$.
This object is useful for representing penalty functionals and similar
linear combinations.

For the primary methods of this class, see {\bf HCL\_Functional\_s}.

There are two ways to construct a linear combination of functionals.
There is a special constructor for th common case of two terms:
\[
J(x) = \alpha_1f_1(x) + \alpha_2f_2(x).
\]
For the more general case of $n$ terms, there is a constructor which
takes a pointer to the domain of the functional and the integer $n$.
The terms in the linear combination must then be added one-by-one
using the {\bf SetNext} method.

Example:
\begin{verbatim}

   // Assume HCL_Functional_s's f1, f2, g1, g2, and g3 have been defined.

   // Create the linear combination F(x) = f1(x) + 100*f2(x)
   HCL_LinCombFcnl_s F( &f1,1.0,&f2,100.0 ); 

   // Create the linear combination G(x) = g1(x) + 100*g2(x) + 10*g3(x)
   HCL_LinCombFcnl_s G( &(g1.Domain()),3 );
   G.SetNext( &g1,1.0 );
   G.SetNext( &g2,100.0 );
   G.SetNext( &g3,10.0 );

\end{verbatim}
*/

class HCL_LinCombFcnl_s: public HCL_Functional_s
{
   private:
      HCL_VectorSpace_s * dom;
      int numinit;
      int num;
      Table * ParamTable;
      HCL_Functional_s ** flist;

   public:  

   // Default constructor---disabled.
      HCL_LinCombFcnl_s();
   // Copy constructor---disabled.
      HCL_LinCombFcnl_s( const HCL_LinCombFcnl_s & );
   /** Usual constructor; takes the domain and the number of functions
       in the sum.  The terms in the linear combination must be added
       one-by-one using the {\bf SetNext} method. */
      HCL_LinCombFcnl_s( HCL_VectorSpace_s * dom,int n );
   /** Special constructor.  This constructor makes it easy to create
       a linear combination with two terms. It just takes pointers to
       the two functionals and the two weights. */
      HCL_LinCombFcnl_s( HCL_Functional_s * f1,float w1,
                               HCL_Functional_s * f2,float w2 );
   // Destructor.
      virtual ~HCL_LinCombFcnl_s();

   /** Domain space access.  Returns a reference to the domain of the
       functional, which is required to be a vector space. */
      virtual HCL_VectorSpace_s & Domain() const { return *dom; }
    /** MaxStep computes the longest vector from x in the direction dir
	that will lie in the domain of the functional; that is, it computes
        the greatest a such that x + a*dir lies in the domain.
        It does this by calling {\bf MaxStep} for each functional in
        the linear combination, and taking the minimum of the results. */
      virtual float MaxStep( const HCL_Vector_s & x, 
			      const HCL_Vector_s &dir) const;

   /** Access to parameter table.  The weights for the linear combination
       can be accessed with Parameters().GetArrayValue( "Weights",i,val )
       or changed with Parameters().PutArrayValue( "Weights",i,val ). */
      Table & Parameters() { return *ParamTable; }

   /** {\bf Evaluate} creates an "evaluation" object, which knows how to
       compute the function value and gradient at the given x. */
      virtual HCL_EvaluateFunctional_s * 
                  Evaluate( const HCL_Vector_s & x ) const;

   /** SetNext allows the next term in the linear combination to be added.
       This method takes the functional and the weight. */
      void SetNext( HCL_Functional_s * eval,float w );

   /** Debugging information.  Prints out the weights and the individual
       functionals. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
