/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_LinCombLinearOp_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_LINCOMBLINEAROP_S_H_
#define HCL_LINCOMBLINEAROP_S_H_

#include "HCL_LinearOp_s.h"
#include "Table.h"

/** {\bf HCL\_LinCombLinearOp\_s} is a concrete class implementing a linear
combination of one or more linear operators.  This class is useful
for combining existing linear operators to create a new operator.

The primary methods are exactly the same as those of {\bf HCL\_LinearOp\_s}.
In addition, the {\bf Parameters} method allows the weights in the
linear combination to be accessed or changed.

There are two constructors.  The usual constructor takes (pointers to) the
domain and range spaces, as well as the number of operators in the sum.
The operators and their weights are then specified one-by-one using the
{\bf SetNext} method.  Here is an example:
\begin{verbatim}
// X and Y are instances of HCL_VectorSpace_s, and L1,L2,L3 are instances
// of HCL_LinearOp_s, each with domain X and range Y.
// The following constructs the operator M = 2L1-3.7L2+L3.

HCL_LinCombLinearOp_s M( &X,&Y,3 );
M.SetNext( &L1,2.0 );
M.SetNext( &L2,-3.7 );
M.SetNext( &L3,1.0 );

// M is now ready to use.
\end{verbatim}

The second constructor is provided for convenience in the special case of
a linear combination of two operators.
\begin{verbatim}
// The next line constructs the operator M = 2L1-3.7L2.

HCL_LinCombLinearOp_s M( &L1,2.0,&L2,-3.7 );

// M is now ready to use.
\end{verbatim}
*/
class HCL_LinCombLinearOp_s: public HCL_LinearOp_s
{
   private:
      int num;
      int numinit;
      HCL_LinearOp_s ** ops;
      Table * ParamTable;
      HCL_VectorSpace_s * dom;
      HCL_VectorSpace_s * range;
   public:

   // Default constructor---disabled
      HCL_LinCombLinearOp_s();
   // Copy constructor---disabled
      HCL_LinCombLinearOp_s( const HCL_LinCombLinearOp_s & L );
   /** Usual constructor.  Requires that the domain and range space be
       given (for error checking), along with the number of terms in
       the sum.  The operators and weights are added one-by-one using
       the SetNext method. */
      HCL_LinCombLinearOp_s( HCL_VectorSpace_s * d,
                             HCL_VectorSpace_s * r,int n );
   /** Provided for convenience.  This constructor makes it easy to
       create a linear combination with two terms.  Pointers to the
       two operators (Op1 and Op2) and their weights (a1 and a2)
       must be given. */
      HCL_LinCombLinearOp_s( HCL_LinearOp_s * op1,float a1,
                             HCL_LinearOp_s * op2,float a2 );
   // Destructor
      virtual ~HCL_LinCombLinearOp_s();

   /** Domain space access. Returns a reference to the domain of
       the operator, which is an {\bf HCL\_VectorSpace\_s}. */
      virtual HCL_VectorSpace_s & Domain() const { return *dom; }
   /** Range space access. Returns a reference to the range of
       the operator, which is an {\bf HCL\_VectorSpace\_s}. */
      virtual HCL_VectorSpace_s & Range() const { return *range; }

   /** Access to parameter table.  The weights for the linear combination
       can be accessed with Parameters().GetArrayValue( "Weights",i,val )
       or changed with Parameters().PutArrayValue( "Weights",i,val ). */
      Table & Parameters() { return *ParamTable; }

   /// Image computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_s & x,
                                HCL_Vector_s & y ) const;
   /// AdjImage computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_s & y,
                                   HCL_Vector_s & x ) const;
   /// {\bf InvImage} returns an error message (unimplemented).
      virtual void InvImage(const HCL_Vector_s & y,
                                  HCL_Vector_s & x) const;
   /** {\bf InvAdjImage} returns an error message (unimplemented).*/
      virtual void InvAdjImage(const HCL_Vector_s & x,
                                     HCL_Vector_s & y) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b z$}. */
       virtual void ImageAdd(const HCL_Vector_s & x,
                             const HCL_Vector_s & z,
			     HCL_Vector_s & y,
                             float a=1.0,
                             float b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b z$}. */
       virtual void AdjImageAdd(const HCL_Vector_s & x,
                             const HCL_Vector_s & z,
			     HCL_Vector_s & y,
                             float a=1.0,
                             float b=1.0) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b y$}. */
       virtual void ImageAdd(const HCL_Vector_s & x,
			     HCL_Vector_s & y,
                             float a=1.0,
                             float b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b y$}. */
       virtual void AdjImageAdd(const HCL_Vector_s & x,
			     HCL_Vector_s & y,
                             float a=1.0,
                             float b=1.0) const;

   /** SetNext is used for initializing the terms in the linear combination.
       A pointer to the next operator (L) and its weight (a) must be
       given. */
      void SetNext( HCL_LinearOp_s * L,float a );

   /** Debugging information.  Prints out the weights in the linear
       combination, and invokes the {\bf Write} method for
       each linear operator. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
