/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_LinearOp_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef _HCL_LINEAROP_S_H
#define _HCL_LINEAROP_S_H

#include "HCL_Vector_s.h"

/** {\bf HCL\_LinearOp\_s} is the base class for all linear operators. 
This class represents a linear operator mapping one vector space to
another.  The primary methods are:
\begin{itemize}
\item {\bf Domain} Returns a reference to the domain of the operator.
This is used primarily for error-checking.
\item {\bf Range} Returns a reference to the range of the operator.
This is used primarily for error-checking.
\item {\bf Image} Computes the action of the operator on a vector.
Here is an example illustrating the use of {\bf Domain}, {\bf Range},
and {\bf Image}:
\begin{verbatim}
   void fcn( HCL_LinearOp_s & L,HCL_Vector_s & x,HCL_Vector_s & y )
   {
      if( x.Space() != L.Domain() || y.Space() != L.Range() )
         // Error condition
      .
      .
      .
      L.Image( x,y ); // y <-- Lx
      .
      .
      .
   }
\end{verbatim}
\item {\bf ImageAdd} Computes \TEX{$y\leftarrow aLx+bz$} or
\TEX{$y\leftarrow aLx+by$}, where a and b are scalars.
\item {\bf AdjImage} Computes the action of the adjoint of the operator on
a vector.  This method is used as follows:
\begin{verbatim}
      L.AdjImage( y,x ); // x <-- L'y
\end{verbatim}
The method is implemented as an error in this base class.  In a derived
class, this method must be implemented to behave as advertised if the operator
will be used by algorithms which require the adjoint.  If the operator is not
intended to be used by such algorithms, then this method need not be
over-ridden.
\item {\bf AdjImageAdd} Analogous to {\bf ImageAdd}, but for the adjoint.
\item {\bf InvImage} Computes the action of the inverse of the operator on
a vector.  This method is used as follows:
\begin{verbatim}
      L.InvImage( y,x ); // x <-- L^{-1}y
\end{verbatim}
The method is implemented as an error in this base class.  In a derived
class, this method must be implemented to behave as advertised if the operator
will be used by algorithms which require the inverse.  If the operator is not
intended to be used by such algorithms, then this method need not be
over-ridden.
\item {\bf InvImageAdd} Analogous to {\bf ImageAdd}, but for the inverse.
\item {\bf InvAdjImage} Computes the action of the inverse of the adjoint on
a vector. The method is implemented as an error in this base class.  In a
derived class, this method must be implemented to behave as advertised if
the operator will be used by algorithms which require the inverse of the
adjoint.  If the operator is not intended to be used by such algorithms, then
this method need not be over-ridden.
\item {\bf InvAdjImageAdd} Analogous to {\bf ImageAdd}, but for the inverse of
the adjoint.
\item {\bf CheckAdj} tests that the operator and its adjoint satisfy
the "adjointness" relationship.
\item {\bf NormalImage} Computes the action of the normal operator on
a vector.  This method is used as follows:
\begin{verbatim}
      L.NormalImage( x,z ); // z <-- L'Ly
\end{verbatim}
The method is implemented in this base class to call the {\bf Image} method
followed by the {\bf AdjImage} method.  In a derived class, this method can be
reimplemented if there is a more efficient way to compute the normal operator.
\item {\bf NormalImageAdd} Analogous to {\bf ImageAdd}, but for the normal
operator.
\item {\bf CheckNormal} tests that the {\bf Image}, {\bf AdjImage}, and
{\bf NormalImage} methods are consistent.
\end{itemize}
Note that this class makes no assumption about the representation of
the operator.  Examples include operators defined by dense or sparse
matrices and solution operators of (linear) PDEs defined by simulations.
*/
class HCL_LinearOp_s: public HCL_Base
{
public:  

  // Default constructor
       HCL_LinearOp_s(){}
  // Copy constructor
       HCL_LinearOp_s( const HCL_LinearOp_s & L ) {}
  // destructor
       virtual ~HCL_LinearOp_s(){}

  /** Domain space access. Returns a reference to the domain of the
    operator, which is an {\bf HCL\_VectorSpace\_s}.  Pure virtual:
    must be implemented in any derived instantiable class.  */

  virtual HCL_VectorSpace_s & Domain() const = 0;

  /** Range space access. Returns a reference to the range of the
    operator, which is an {\bf HCL\_VectorSpace\_s}.  Pure virtual:
    must be implemented in any derived instantiable class. */

  virtual HCL_VectorSpace_s & Range() const = 0;

  /** {\bf Image} computes the action of the operator on x, giving y.
    Pure virtual: must be implemented in any derived instantiable
    class. */

  virtual void Image( const HCL_Vector_s & x,
		      HCL_Vector_s & y ) const = 0;

  /** {\bf AdjImage} computes the action of the adjoint on y, giving
    x.  This method is implemented in the base class to return an
    error message. To obtain a working {\bf AdjImage} in any
    derived instantiable class, you must override this default. */

  virtual void AdjImage( const HCL_Vector_s & y,
			 HCL_Vector_s & x ) const;

  /** {\bf CheckAdj} tests that the operator and its adjoint satisfy
    the "adjointness" relationship. The return value is 1 if an
    error is detected, zero otherwise.  The first input is a display flag;
    zero means no display, 1 means display a message only if there is
    an error, and 2 means displays the results of the test, whether there
    is an error or not.  The second input sets the tolerance for
    the floating-point tests; an error is reported if two results
    which ought to be equal differ by a relative amount greater
    than tol times machine epsilon. This method is implemented in
    this class. */
    
  int CheckAdj(int Display = 1,float tol = 1000 );
  
  /** {\bf NormalImage} computes the action of the normal on x, giving
    z. This method is given a default implementation in this
    class. Custom implementations may be more efficient than the
    default. */

  virtual void NormalImage( const HCL_Vector_s & x,
			    HCL_Vector_s & z ) const;

  /** {\bf CheckNormal} tests that the normal operator N is really defined
     by N = A'A.  The return value is 1 if an error is detected, zero
     otherwise.  The first input is a display flag; zero means no
     display, 1 means display a message only if there is an error,
     and 2 means displays the results of the test, whether there is
     an error or not.  The second input sets the tolerance for
     the floating-point tests; an error is reported if two results
     which ought to be equal differ by a relative amount greater
     than tol times machine epsilon. This method is implemented 
     in this class. */

  int CheckNormal( int display = 1,float tol = 100 );

  /** {\bf InvImage} computes the action of the inverse on y, giving x.
    This method is implemented in the base class to return an 
    error message. To obtain a working {\bf InvImage} in any
    derived instantiable class, you must override this default. */

  virtual void InvImage(const HCL_Vector_s & y,
			HCL_Vector_s & x) const;

  /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
     giving y.  This method is implemented in the base class to return an 
     error message. To obtain a working {\bf InvAdjImage} in any
    derived instantiable class, you must override this default. */

  virtual void InvAdjImage(const HCL_Vector_s & x,
			   HCL_Vector_s & y) const;

  /** {\bf ImageAdd} implements \TEX{$y = a A x + b z$}.  This provides
    for efficiency in creating linear combinations of the operator output
    and another vector.
    An efficient implementation will avoid temporary storage of intermediate
    results, as well as multiple passes through the data. The
    scalars default to 1.0; if left at their default values, no
    multiplies are actually performed.  For convenience we provide
    a default (inefficient) implementation: efficient implementations
    for concrete operators will override this. */

  virtual void ImageAdd(const HCL_Vector_s & x,
			const HCL_Vector_s & z,
			HCL_Vector_s & y,
			float a=1.0,
			float b=1.0) const;

  /** {\bf AdjImageAdd} implements \TEX{$y = a A^* x + b z$}.  This
       provides for efficiency in creating linear combinations of the
       operator output and another vector.  An efficient
       implementation will avoid temporary storage of intermediate
       results, as well as multiple passes through the data. The
       scalars default to 1.0; if left at their default values, no
       multiplies are actually performed.  For convenience we provide
       a default (inefficient) implementation: efficient
       implementations for concrete operators will override this. */

  virtual void AdjImageAdd(const HCL_Vector_s & x,
			   const HCL_Vector_s & z,
			   HCL_Vector_s & y,
			   float a=1.0,
			   float b=1.0) const;

  /** {\bf NormalImageAdd} implements \TEX{$y = a A^*A x + b z$}.  This
       provides for efficiency in creating linear combinations of the
       operator output and another vector, as frequently occurs in iterative
       algorithms.  An efficient implementation will avoid temporary storage
       of intermediate results, as well as multiple passes through the data.
       The scalars default to 1.0; if left at their default values, no
       multiplies are actually performed.  For convenience we provide
       a default (inefficient) implementation: efficient implementations
       for concrete operators will override this. */

  virtual void NormalImageAdd(const HCL_Vector_s & x,
			      const HCL_Vector_s & z,
			      HCL_Vector_s & y,
			      float a=1.0,
			      float b=1.0) const;
  
  /** {\bf InvImageAdd} implements \TEX{$y = a A^{-1} x + b z$}.  This
       provides for efficiency in creating linear combinations of the
       operator output and another vector, as frequently occurs in
       iterative algorithms.  An efficient implementation will avoid
       temporary storage of intermediate results, as well as multiple
       passes through the data.  The scalars default to 1.0; if left
       at their default values, no multiplies are actually performed.
       For convenience we provide a default (inefficient)
       implementation: efficient implementations for concrete
       operators will override this. */

  virtual void InvImageAdd(const HCL_Vector_s & x,
			   const HCL_Vector_s & z,
			   HCL_Vector_s & y,
			   float a=1.0,
			   float b=1.0) const;

  /** {\bf InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b z$}.  This
    provides for efficiency in creating linear combinations of the
    operator output and another vector, as frequently occurs in iterative
    algorithms.  An efficient implementation will avoid temporary storage
       of intermediate results, as well as multiple passes through the data.
       The scalars default to 1.0; if left at their default values, no
       multiplies are actually performed.  For convenience we provide
       a default (inefficient) implementation: efficient implementations
       for concrete operators will override this. */
    
  virtual void InvAdjImageAdd(const HCL_Vector_s & x,
			      const HCL_Vector_s & z,
			      HCL_Vector_s & y,
			      float a=1.0,
			      float b=1.0) const;

  /** {\bf ImageAdd} implements \TEX{$y = a A x + b y$}.  This provides
       for efficiency in creating linear combinations of the operator output
       and another vector.
       An efficient implementation will avoid temporary storage of intermediate
       results, as well as multiple passes through the data. The
       scalars default to 1.0; if left at their default values, no
       multiplies are actually performed.  For convenience we provide
       a default (inefficient) implementation: efficient implementations
       for concrete operators will override this. */

  virtual void ImageAdd(const HCL_Vector_s & x,
			HCL_Vector_s & y,
			float a=1.0,
			float b=1.0) const;

  /** {\bf AdjImageAdd} implements \TEX{$y = a A^* x + b y$}.  This provides
       for efficiency in creating linear combinations of the operator output
       and another vector.
       An efficient implementation will avoid temporary storage of intermediate
       results, as well as multiple passes through the data. The
       scalars default to 1.0; if left at their default values, no
       multiplies are actually performed.  For convenience we provide
       a default (inefficient) implementation: efficient implementations
       for concrete operators will override this. */

  virtual void AdjImageAdd(const HCL_Vector_s & x,
			   HCL_Vector_s & y,
			   float a=1.0,
			   float b=1.0) const;

  /** {\bf NormalImageAdd} implements \TEX{$y = a A^*A x + b y$}.  This
       provides for efficiency in creating linear combinations of the
       operator output and another vector, as frequently occurs in iterative
       algorithms.  An efficient implementation will avoid temporary storage
       of intermediate results, as well as multiple passes through the data.
       The scalars default to 1.0; if left at their default values, no
       multiplies are actually performed.  For convenience we provide
       a default (inefficient) implementation: efficient implementations
       for concrete operators will override this. */

  virtual void NormalImageAdd(const HCL_Vector_s & x,
			      HCL_Vector_s & y,
			      float a=1.0,
			      float b=1.0) const;

  /** {\bf InvImageAdd} implements \TEX{$y = a A^{-1} x + b y$}.  This
       provides for efficiency in creating linear combinations of the
       operator output and another vector, as frequently occurs in iterative
       algorithms.  An efficient implementation will avoid temporary storage
       of intermediate results, as well as multiple passes through the data.
       The scalars default to 1.0; if left at their default values, no
       multiplies are actually performed.  For convenience we provide
       a default (inefficient) implementation: efficient implementations
       for concrete operators will override this. */

  virtual void InvImageAdd(const HCL_Vector_s & x,
			   HCL_Vector_s & y,
			   float a=1.0,
			   float b=1.0) const;

  /** {\bf InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b y$}.  This
       provides for efficiency in creating linear combinations of the
       operator output and another vector, as frequently occurs in iterative
       algorithms.  An efficient implementation will avoid temporary storage
       of intermediate results, as well as multiple passes through the data.
       The scalars default to 1.0; if left at their default values, no
       multiplies are actually performed.  For convenience we provide
       a default (inefficient) implementation: efficient implementations
       for concrete operators will override this. */

  virtual void InvAdjImageAdd(const HCL_Vector_s & x,
			      HCL_Vector_s & y,
			      float a=1.0,
			      float b=1.0) const;

};

#endif
