/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_MatTR_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#ifndef HCL_MATTR_S_H_
#define HCL_MATTR_S_H_

#include "HCL_MatrixOps_s.h"
#include "HCL_Rn_s.h"
#include "Table.h"

#define BOUNDARY_SOLUTION 101
#define INTERIOR_SOLUTION 102
#define FAILED 103

/** {\bf HCL\_MatTR\_s} implements an algorithm for solving the
trust region subproblem in the case in which the operator is
represented by a symmetric matrix.  The algorithm is based on
the following paper:

    J. J. More' and D. C. Sorensen, "Computing a trust region step",
    SIAM J. Sci. Stat. Comput. Vol. 4, No. 3, 553-572, 1983.
*/
class HCL_MatTR_s: public HCL_Base
{
   private:

      Table * ParamTable;
      int n;

   // Input parameters (default value)
   /**@name Input parameters */
   //@{
   /// Initial trust region radius (1.0).
      float Delta;
   /** Tolerance for the step length.  The iteration is halted when
       the step is within this relative tolerance of the boundary (1.0e-2). */
      float Tol;
   /** Secondary stopping tolerance (0.0). */
      float Tol2;
   /// Maximum number of iterations (20).
      int MaxItn;
   /** Display Flag (0). This flag controls how much detail about the
       progress of the iteration is sent to the screen.  Possible values
       are 0 (no display), 1 (summary statement at the end),
       2 (summary statement every {\bf DispFreq} iterations),
       3 (output of algorithmic parameters at every iterations---used for
       debugging purposes). */
      int DispFlag;
   /// Display precision; controls how many digits are sent to the screen (6).
      int DispPrecision;
   /** Display frequency; controls how often details of iteration are
       displayed on screen (MaxItn/5). */
      int DispFreq;
   /// minimum step length, that is, minimum Delta (1.0e-6).
      float MinStep;
   /// maximum step length, that is, maximum Delta (1.0e6).
      float MaxStep;
   /** Boolean flag; start current run with multiplier from
       previous run (0). */
      int UseLastLam;
   //@}


   // The following is for documentation purposes only.
#ifdef NOT_DEFINED
   /**@name Required output parameters */
   //@{
   /** Termination code.  Possible values are
          BOUNDARY_SOLUTION 101
          INTERIOR_SOLUTION 102
          FAILED            103 */
      int TermCode;
   /// Lagrange multiplier.
      float Multiplier;
   /// Reduction in the quadratic model.
      float Reduction;
   /// Length of step taken.
      float StepLength;
   //@}
#endif

   // Some temporary vectors

   // Scaling operators

      HCL_SymMatOp_s * C;
      HCL_SymMatOp_s * sqrtC;
      HCL_SymMatOp_s * B;
      HCL_RnVector_s * g;
      HCL_RnVector_s * p;
      HCL_RnVector_s * q;

   // Other local variables

      float lam,LastLam;
      float lam_u,lam_l,lam_s;    // Safeguards for lam
      float gnorm;
      float pnorm,pnorm2;
      float Bnorm;
      int OldDispPrec;
      int TermCode;
      int Itn;
      float tau,tau2;
      float sing,sing2;

   // Initialize (allocate temporary vectors, etc.)

      void Initialize();

   // De-allocate temporaries, etc.

      void Clean();

   public:

   // Copy constructor---disabled.
      HCL_MatTR_s( const HCL_MatTR_s & x );
   /** Usual constructor; optionally takes the name of a file from
       which to read algorithmic parameters.  If such a file is
       provided, it must be a plain text file with entries of the
       form

                 ParameterName = value

       or

                 MatTR::ParameterName = value

       The second form allows the same file to contain parameters for
       more than one algorithm to be given in the same file.  For example,
       one can entries such as

                 UMin::DispFlag = 2
                 MatTR::DispFlag = 1

       which set the display flags for the minimization algorithm and
       the trust region solver to different values. */
      HCL_MatTR_s( char * fname = NULL );

   // Destructor.
      ~HCL_MatTR_s();

   /** The method {\bf Parameters} returns the parameter table of
       the algorithm.  Algorithmic parameters can be accessed
       via {\bf Parameters().GetValue( "name",val )} and changed
       via {\bf Parameters().PutValue( "name",val )}. */
      virtual Table & Parameters() const;

   /** {\bf SetScaling} defines a new inner product in terms of
       a symmetric, positive definite operator S: <x,y> = (x,Sy). 
       This method is included for future use; at present, scaling is
       not implemented. */
      void SetScaling( HCL_SymMatOp_s * S );
   /// {\bf UnSetScaling} returns the inner product to the default.
      void UnSetScaling();

   /** The method {\bf Solve} attempts to compute an approximate
       solution of the trust region subproblem defined by
       the inputs of the method.  Note that the trust region
       radius {\bf Delta} must defined through the {\bf Parameters}
       method. */
      int Solve( HCL_SymMatOp_s & B,const HCL_RnVector_s & g,
                                HCL_RnVector_s & p );
    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
