/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_PCG_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_PCG_S_H_
#define HCL_PCG_S_H_

#include "Std_includes.h"
#include "lapack.h"
#include "HCL_Arch_s.h"
#include "timer_s.h"
#include "ProtectedDivision_s.h"
#include "Table.h"
#include "HCL_LinearSolver_s.h"
#include "HCL_Adjoint_s.h"
#include "HCL_Recorder_s.H"
#include "maxmin_s.h"

/** {\bf HCL\_PCG\_s} implements the preconditioned conjugate gradients
algorithm for solving a linear operator equation with a symmetric positive
definite operator. By passing an (optional pointer to an) 
{\bf HCL\_Recorder\_s} object, the user can record all of the intermediate
estimates of the solution generated during the iteration. */
class HCL_PCG_s: public HCL_LinearSolver_s
{
   private:
      Table * ParamTable;
      HCL_LinearOp_s * MInv;
      HCL_Recorder_s * Rec;

   public:

   // Default constructor---disabled.
      HCL_PCG_s();
  // Copy constructor---disabled.
      HCL_PCG_s( const HCL_PCG_s & S );
  /** Constructor.  Takes (a pointer to) the preconditioner and
    (a pointer to) the linear solver (used for inverting the preconditioner),
    also (a pointer to) an optional Recorder object.
      @param fname Parameter file name, or NULL
      @param m Preconditioner
      @param ls Linear solver to invert preconditioner
      @param rec is an optional Recorder object
    */
      HCL_PCG_s( char *fname, HCL_LinearOp_s * m,
		 HCL_LinearSolver_s * ls,
		 HCL_Recorder_s * rec = NULL);

  /** Alternate constructor.  Takes (a pointer to) the linear operator 
      that computes the inverse of the preconditioner, also (a pointer
      to) an optional Recorder object.
      @param fname Parameter file name, or NULL
      @param minv Inverse of the preconditioner or NULL
      @param rec Optional Recorder object
    */
      HCL_PCG_s( char *fname, 
		 HCL_LinearOp_s * minv, 
		 HCL_Recorder_s * rec = NULL );

   // Destructor.
      virtual ~HCL_PCG_s();

   /** Access to parameter table.  Algorithmic parameters can be
       accessed with Parameters().GetValue("NAME",val) or changed with
       Parameters().PutValue("NAME",val).
    */
	Table & Parameters() const { return *ParamTable; }

   /** Solve (approximately) the equation Ax=b for x.
       @param A Linear operator
       @param b RHS vector
       @param x Initially, this must be a starting vector for the
                iteration.  On exit, this contains the approximate solution
       @return The value of TermCode (see the output parameters section of this
               document.) 
    */
      virtual int Solve( const HCL_LinearOp_s & A,const HCL_Vector_s & b,
                                HCL_Vector_s & x ) const;

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;

      enum {
	 Success = 0,
	 TooManyIterations = 1
      } TermCodes;

#ifdef UNDEFINED  // for documentation purposes only

   /**@name         Input Parameters
      In a parameter file, the names may be prepended by "LinearSolver::",
      "LinearSolverPCG::" or "PCG::" */
   //@{

     // dummy
     int dummy;
     /// Maximum number of iterations.\\Default = 100
     int MaxItn;

     /** Stopping criteria.  The algorithm will terminate when the 
        relative residual is less than Tol.  A value of 0.0 will imply
        a tolerance of 500*(machine epsilon).\\
        Default = 1.0e-3
      */
     float Tol;

      /** Display level.  This determines how much information should be
          displayed during the execution of the algorithm.  Common values
          are: \\
          0 - No output \\
          1 - One-line summary output \\
          2 - Detailed header and footer with timing information \\
          3 - Per-iteration information \\
 
          Default = 0
       */
     int DispFlag;
   //@}

   /**@name       Output parameters */
   //@{
     /** Norm of the residual error. */
     float ResidualNorm;

      /** Termination code.  Value is one of: \\
 
          0 - Success \\
          1 - Too many iterations taken \\
       */
     int TermCode;
   //@}

#endif
};

/** {\bf HCL\_CG\_s} implements the (unpreconditioned) conjugate gradients
algorithm for solving a linear operator equation with a symmetric positive
definite operator. By passing an (optional pointer to an) 
{\bf HCL\_Recorder\_s} object, the user can record all of the intermediate
estimates of the solution generated during the iteration. */
class HCL_CG_s: public HCL_LinearSolver_s
{
   private:
      Table * ParamTable;
      HCL_Recorder_s * Rec;

   public:

  // Copy constructor---disabled.
      HCL_CG_s( const HCL_CG_s & S );

  /** Constructor.  
      @param fname Parameter file name, or NULL
      @param rec Optional (pointer to) Recorder object
    */
      HCL_CG_s( char *fname = NULL, HCL_Recorder_s * rec = NULL);
  // Destructor.
      virtual ~HCL_CG_s();

   /** Access to parameter table.  Algorithmic parameters can be
       accessed with Parameters().GetValue("NAME",val) or changed with
       Parameters().PutValue("NAME",val).
    */
	Table & Parameters() const { return *ParamTable; }

   /** Solve (approximately) the equation Ax=b for x.
       @param A Linear operator
       @param b RHS vector
       @param x Initially, this must be a starting vector for the
                iteration.  On exit, this contains the approximate solution
       @return The value of TermCode (see the Parameters section of this
               document.) 
    */
      virtual int Solve( const HCL_LinearOp_s & A,const HCL_Vector_s & b,
                                HCL_Vector_s & x ) const;

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;

      enum {
	 Success = 0,
	 TooManyIterations = 1
      } TermCodes;

#ifdef UNDEFINED  // for documentation purposes only

   /**@name         Input Parameters
      In a parameter file, the names may be prepended by "LinearSolver::",
      "LinearSolverCG::" or "CG::" */
   //@{

     // dummy
     int dummy;
     /** Stopping criteria.  The algorithm will terminate when the 
        relative error is less than Tol.  A value of 0.0 will imply
	a tolerance of 500*(machine epsilon). 
	\\
	Default = 1.0e-3
      */
     float Tol;

     /** Maximum number of iterations.
	 \\
	 Default = 100
      */
     int MaxItn;

      /** Display level.  This determines how much information should be
          displayed during the execution of the algorithm.  Common values
          are: \\
          0 - No output \\
          1 - One-line summary output \\
          2 - Detailed header and footer with timing information \\
          3 - Per-iteration information \\
 
          Default = 0
       */
     int DispFlag;
   //@}

   /**@name       Output parameters */
   //@{
     /** Norm of the residual error. */
     float ResidualNorm;

      /** Termination code.  Value is one of: \\
 
          0 - Success \\
          1 - Too many iterations taken \\
       */
     int TermCode;
   //@}

#endif

};

#endif




