/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_TrustRegionSolvers_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_TRUSTREGIONSOLVERS_H_
#define HCL_TRUSTREGIONSOLVERS_H_

#include "Table.h"
#include "HCL_LinearOp_s.h"
#include "HCL_LinearSolver_s.h"
#include "HCL_Arch_s.h"

#define BOUNDARY_SOLUTION 101
#define INTERIOR_SOLUTION 102
#define FAILED 103

/** The class {\bf HCL\_TrustRegionSolver\_s} is an abstract base clase for
algorithms for solving the trust region subproblem: Minimize
\TEX{\[
            q(s) = (g,s) + \frac{1}{2}(s,Bs)
\]}
subject to the constraint
\TEX{\[
            (s,Cs)\le Delta^2
\]}
where $()$ denotes the inner product, $g$ is a given vector, $B$ is a symmetric linear operator, and
$C$ is a symmetric and positive definite linear operator. */

class HCL_TrustRegionSolver_s: public HCL_Base
{
   private:

// The following are for documentation purposes only
#ifdef NOT_DEFINED

   // Parameters found in ParamTable for derived solvers.

   // Input parameters (default value)
   /**@name Input parameters */
   //@{
   /// Initial trust region radius (1.0).
      float Delta;
   /// minimum step length, that is, minimum Delta (1.0e-6).
      float MinStep;
   /// maximum step length, that is, maximum Delta (1.0e6).
      float MaxStep;
   //@}


   /**@name Required output parameters */
   //@{
   /** Termination code.  Possible values are
          BOUNDARY_SOLUTION 101
          INTERIOR_SOLUTION 102
          FAILED            103
       Other values may be added by derived classes.*/
      int TermCode;
   /// Lagrange multiplier.
      float Multiplier;
   /// Reduction in the quadratic model.
      float Reduction;
   /// Length of step taken.
      float StepLength;
   //@}
#endif

   public:

      /** The method {\bf Parameters} returns the parameter table of
          the algorithm.  Algorithmic parameters can be accessed
          via {\bf Parameters().GetValue( "name",val )} and changed
          via {\bf Parameters().PutValue( "name",val )}. */
      virtual Table & Parameters() const = 0;

      /** {\bf SetScaling} defines a new inner product in terms of
          a symmetric, positive definite operator S: <x,y> = (x,Sy).
          An optional linear solver can be used to compute
          \TEX{$S^{-1}y$} when needed. */
      virtual void SetScaling( HCL_LinearOp_s * S,
                               HCL_LinearSolver_s * lsolver=NULL ) = 0;
      /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling() = 0;

      /** The method {\bf Solve} attempts to compute an approximate
          solution of the trust region subproblem defined by
          the inputs of the method.  Note that the trust region
          radius {\bf Delta} must defined through the {\bf Parameter}
          method. */
      virtual void Solve( const HCL_LinearOp_s & B,const HCL_Vector_s & g,
                                HCL_Vector_s & p ) = 0;
};

/** The class HCL\_TRCG\_s is an implentation of the Steihaug-Toint algorithm
for approximately solving the trust region subproblem.  This algorithm
is based on using the conjugate gradient (CG) method to compute the Newton
step, with modifications for handling the case in which the
approximate step, as computed by CG, leaves the trust region, and the
case in which negative curvature is encountered.  Note, in particular,
that in spite of the use of CG, the linear operator defining the
quadratic model is not assumed to be positive definite.

For more information about the underlying algorithm, see

   "The conjugate gradient method and trust regions in large scale
    optimization"  by Trond Steihaug,  SIAM J. Numer. Anal. Vol. 20,
    No. 3, 626--637
*/

class HCL_TRCG_s: public HCL_TrustRegionSolver_s
{
   private:

      Table * ParamTable;
      int init;

   // Input parameters (default value)
   /**@name Input parameters */
   //@{
   /// Initial trust region radius (1.0).
      float Delta;
   /** Tolerance for computing the Newton step.  The CG iteration
       is halted when the relative residual falls below this value (1.0e-2). */
      float Tol;
   /// Maximum number of CG iterations (20).
      int MaxItn;
   /** Display Flag (0). This flag controls how much detail about the
       progress of the CG iteration is sent to the screen.  Possible values
       are 0 (no display), 1 (summary statement at the end),
       2 (summary statement every {\bf DispFreq} iterations),
       3 (output of algorithmic parameters at every iterations---used for
       debugging purposes). */
      int DispFlag;
   /// Display precision; controls how many digits are sent to the screen (6).
      int DispPrecision;
   /** Display frequency; controls how often details of CG iteration are
       displayed on screen (MaxItn/5). */
      int DispFreq;
   /// minimum step length, that is, minimum Delta (1.0e-6).
      float MinStep;
   /// maximum step length, that is, maximum Delta (1.0e6).
      float MaxStep;
   //@}


   // The following is for documentation purposes only.
#ifdef NOT_DEFINED
   /**@name Required output parameters */
   //@{
   /** Termination code.  Possible values are
          BOUNDARY_SOLUTION 101
          INTERIOR_SOLUTION 102
          FAILED            103 */
      int TermCode;
   /// Lagrange multiplier.
      float Multiplier;
   /// Reduction in the quadratic model.
      float Reduction;
   /// Length of step taken.
      float StepLength;
   //@}
#endif

   // Some temporary vectors

      HCL_Vector_s * r;
      HCL_Vector_s * Cinvr;
      HCL_Vector_s * dir;
      HCL_Vector_s * Bdir;
      HCL_Vector_s * temp1;
      HCL_Vector_s * temp2;

      HCL_LinearOp_s * C;
      HCL_LinearSolver_s * lsolver;

   // Other local variables
      int OldDispPrec;

   // Initialize (allocate temporary vectors, etc.)
      void Initialize( HCL_VectorSpace_s & dom );

   // De-allocate temporaries, etc.
      void Clean();

   public:

   // Copy constructor---disabled.
      HCL_TRCG_s( const HCL_TRCG_s & x );
   /** Usual constructor; optionally takes the name of a file from
       which to read algorithmic parameters.  If such a file is
       provided, it must be a plain text file with entries of the
       form

                 ParameterName = value

       or

                 TRSolver::ParameterName = value

       The second form allows the same file to contain parameters for
       more than one algorithm to be given in the same file.  For example,
       one can entries such as

                 UMin::DispFlag = 2
                 TRSolver::DispFlag = 1

       which set the display flags for the minimization algorithm and
       the trust region solver to different values. */
      HCL_TRCG_s( char * fname = NULL );

   // Destructor.
      virtual ~HCL_TRCG_s();

   /** The method {\bf Parameters} returns the parameter table of
       the algorithm.  Algorithmic parameters can be accessed
       via {\bf Parameters().GetValue( "name",val )} and changed
       via {\bf Parameters().PutValue( "name",val )}. */
      virtual Table & Parameters() const;

   /** {\bf SetScaling} defines a new inner product in terms of
       a symmetric, positive definite operator S: <x,y> = (x,Sy).
       An optional linear solver can be used to compute
       \TEX{$S^{-1}y$} when needed. */
      virtual void SetScaling( HCL_LinearOp_s * S,
                               HCL_LinearSolver_s * lsolver=NULL );
   /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling();

   /** The method {\bf Solve} attempts to compute an approximate
       solution of the trust region subproblem defined by
       the inputs of the method.  Note that the trust region
       radius {\bf Delta} must defined through the {\bf Parameter}
       method. */
      virtual void Solve( const HCL_LinearOp_s & B,const HCL_Vector_s & g,
                                HCL_Vector_s & p );

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
