/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_TrustRegion_SS_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#ifndef HCL_TRUSTREGION_SS_H_
#define HCL_TRUSTREGION_SS_H_

#include "HCL_TrustRegionSolvers_s.h"
#include "HCL_ItEigSolvers_s.h"
#include "HCL_GenericProductVector_s.h"
#include "HCL_LinearSolver_s.h"

/** The class {\bf HCL\_TRSSOp\_s} defines the bordered operator used
by {\bf HCL\_TrustRegion\_SS\_s}.  This is a linear operator, and hence
an instance of {\bf HCL\_LinearOp\_s}. */

class HCL_TRSSOp_s: public HCL_LinearOp_s
{
   private:
      HCL_GenericProductSpace_s * dom;
      HCL_Vector_s * g;
      HCL_LinearOp_s * A;
      HCL_LinearOp_s * C;
      HCL_LinearSolver_s * lsolver;
      Table * ParamTable;
   public:

   // Default constructor---disabled.
      HCL_TRSSOp_s();
   // Copy constructor---disabled.
      HCL_TRSSOp_s( const HCL_TRSSOp_s & );
   /** Constructor; takes the operator A, the vector g, the scalar alpha,
       and, optionally, a preconditioner and corresponding linear solver. */
      HCL_TRSSOp_s( const HCL_LinearOp_s * AA, const HCL_Vector_s * gg,
                    float a,HCL_LinearOp_s * CC=NULL,
                    HCL_LinearSolver_s * ls=NULL );
   // Destructor.
      ~HCL_TRSSOp_s();

   /// Access to parameters, specifically, the scalar alpha.
      Table & Parameters() const;

   /** Domain space access. Returns a reference to the domain of
       the operator, which is a {\bf HCL\_VectorSpace\_s}. */
      virtual HCL_VectorSpace_s & Domain() const;
   /** Range space access. Returns a reference to the range of
       the operator, which is a {\bf HCL\_VectorSpace\_s}. */
      virtual HCL_VectorSpace_s & Range() const;
   /// Image compute the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_s & x,
                                HCL_Vector_s & y ) const;
   /// AdjImage computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_s & y,
                                   HCL_Vector_s & x ) const;

   /// {\bf InvImage} returns an error message and stops the program.
      virtual void InvImage(const HCL_Vector_s & y,
                                  HCL_Vector_s & x) const;
   /** {\bf InvAdjImage} returns an error message and stops the program. */
      virtual void InvAdjImage(const HCL_Vector_s & x,
                                     HCL_Vector_s & y) const;

   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};


/* Developer's notes:

   Algorithm does not set return value "Multiplier" (this is important!).
   Use of scaled inner product must be checked; it probably doesn't work.

*/

/** {\bf NOTE: This class is experimental and has not been carefully
debugged.  It is known to fail on occasions when it should not.}

The class {\bf HCL\_TrustRegion\_SS\_s} is an implementation of the
Santosa-Sorensen algorithm for solving the trust region subproblem.
This algorithm involves transforming the subproblem into a family of
parameterized eigenvalues problems, which are solving using the k-step
Arnoldi method.  The parameter in the eigenvalue problem is adjusted to
produce a solution to the trust region subproblem, along with its Lagrange
multiplier, simultaneously. For more information about the algorithm, see

    "A new matrix-free algorithm for the large-scale trust-region
     subproblem" by Sandra A. Santos and Danny C. Sorensen
     Technical report TR95-20, Department of Computational and Applied
     Mathematics, Rice University, Houston, TX 77251-1892.

In the initial implementation, the trust-region constraint is defined
by the unscaled Euclidean norm.
*/

class HCL_TrustRegion_SS_s: public HCL_TrustRegionSolver_s
{
   private:

      Table * ParamTable;

   // Input parameters (default value)
   /**@name Input parameters */
   //@{
   /// Initial trust region radius (1.0).
      float Delta;
   /** Tolerance for the step length.  The iteration is halted when
       the step is within this relative tolerance of the boundary (1.0e-2). */
      float Tol;
   /// Maximum number of iterations (20).
      int MaxItn;
   /** Display Flag (0). This flag controls how much detail about the
       progress of the iteration is sent to the screen.  Possible values
       are 0 (no display), 1 (summary statement at the end),
       2 (summary statement every {\bf DispFreq} iterations),
       3 (output of algorithmic parameters at every iterations---used for
       debugging purposes). */
      int DispFlag;
   /// Display precision; controls how many digits are sent to the screen (6).
      int DispPrecision;
   /** Display frequency; controls how often details of iteration are
       displayed on screen (MaxItn/5). */
      int DispFreq;
   /// minimum step length, that is, minimum Delta (1.0e-6).
      float MinStep;
   /// maximum step length, that is, maximum Delta (1.0e6).
      float MaxStep;
   /// Tolerance for the hard case (1.0e-2). */
      float HardCaseTol;
   //@}

   // The following is for documentation purposes only.
#ifdef NOT_DEFINED
   /**@name Required output parameters */
   //@{
   /** Termination code.  Possible values are
          BOUNDARY_SOLUTION 101
          INTERIOR_SOLUTION 102
          FAILED            103 */
      int TermCode;
   /// Lagrange multiplier.
      float Multiplier;
   /// Reduction in the quadratic model.
      float Reduction;
   /// Length of step taken.
      float StepLength;
   //@}
#endif

   // Some temporary objects

      HCL_IRArnoldi_s * eigsolver;
      HCL_TRSSOp_s * Balpha;
      HCL_RnVector_s * gammaVec;
      HCL_Vector_s * xtemp;
      HCL_GenericProductVector_s * y;

      HCL_LinearOp_s * C;
      HCL_LinearSolver_s * lsolver;

   // Some other local variables

      int OldDispPrec;
      float gnorm;           // Norm of g (used in detecting hard case)
      float alpha;           // The parameter in the bordered matrix
      float gamma;           // The first component of the eigenvector
      float lam1_prev;       // The smallest eigenvalue of Balpha from the
                              // previous iteration
      float lam1,lam2;       // The two smallest eigenvalues of Balpha
      float phi,phi_prev;    // The current and previous values of phi
      float dphi,dphi_prev;  // The current and previous derivatives of phi
      float alpha_l,alpha_u; // Lower and upper bounds on alpha (safeguard)
      float lam_l,lam_u;     // Lower and upper bounds on lambda (safeguard)
      float xnorm;           // Norm of approximate solution
      float xnorm_prev;      // Norm of approximate solution (previous itn)
      float ZeroTol;         // Tolerance for deciding when lam1 >= 0.0
                              // (use lam1 >= -ZeroTol)

   // Compute (scaled) vector norm

      float GetxNorm( const HCL_Vector_s & ) const;

   // Apply CG in case an interior solution is detected.

      int GetInteriorSolution( const HCL_LinearOp_s & A,
                               const HCL_Vector_s & g,HCL_Vector_s & p );

   // Initialize (allocate temporary vectors, etc.)

      void Initialize( HCL_VectorSpace_s & dom );

   // De-allocate temporaries, etc.

      void Clean();
      void DumpProblem( const HCL_LinearOp_s & A,const HCL_Vector_s & g,
                                                       HCL_Vector_s & x );
   public:

   // Copy constructor---disabled.
      HCL_TrustRegion_SS_s( const HCL_TrustRegion_SS_s & x );
   /** Usual constructor; optionally takes the name of a file from
       which to read algorithmic parameters.  If such a file is
       provided, it must be a plain text file with entries of the
       form

                 ParameterName = value

       or

                 TRSolver::ParameterName = value

       The second form allows the same file to contain parameters for
       more than one algorithm to be given in the same file.  For example,
       one can entries such as

                 UMin::DispFlag = 2
                 TRSolver::DispFlag = 1

       which set the display flags for the minimization algorithm and
       the trust region solver to different values. */
      HCL_TrustRegion_SS_s( char * fname = NULL );

   // Destructor.
      virtual ~HCL_TrustRegion_SS_s();

      /** The method {\bf Parameters} returns the parameter table of
          the algorithm.  Algorithmic parameters can be accessed
          via {\bf Parameters().GetValue( "name",val )} and changed
          via {\bf Parameters().PutValue( "name",val )}. */
      virtual Table & Parameters() const;

      /** {\bf SetScaling} defines a new inner product in terms of
          a symmetric, positive definite operator S: <x,y> = (x,Sy). */
      virtual void SetScaling( HCL_LinearOp_s * S,
                               HCL_LinearSolver_s * lsolver=NULL );
      /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling();

      /** The method {\bf Solve} attempts to compute an approximate
          solution of the trust region subproblem defined by
          the inputs of the method.  Note that the trust region
          radius {\bf Delta} must defined through the {\bf Parameter}
          method. */
      virtual void Solve( const HCL_LinearOp_s & A,const HCL_Vector_s & g,
                                HCL_Vector_s & p );

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
