/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_UMinMethod_d.H.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                     Release 1.0 - July 2000                       
                                                                  
                          copyright (C) 1995-2000      
            Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef _UMINMETH_
#define _UMINMETH_

#include "HCL_UMin_s.h"
#include "HCL_UMin_lbfgs_s.h"
#include "HCL_UMinTR_s.h"
#include "HCL_UMinNLCG_s.h"
#include "Table.h"

/** HCL\_UMinMethod provides easy access to the UMin methods.
  The available choices of direction update and line search, by name, 
  form the parameter list of the constructor, which invokes the appropriate
  constructor for the method combination chosen by the user.

  IMPORTANT NOTE: this version (9/97) does not enable specification
  of various sorts of preconditioning options offered by the UMin
  constructors. For the moment users wishing to use these features 
  will need to use the UMin constructors directly. */

class HCL_UMinMethod_s: public HCL_UMin_s {

private:

  HCL_LineSearch_s * lnsrch;
  HCL_UMin_s * meth;
  HCL_TRCG_s * trs;

public:

  /**@name Constructors and destructor */
  //@{
  /// Default constructor.
  HCL_UMinMethod_s(){}
  /// Copy constructor - disabled.
  HCL_UMinMethod_s( const HCL_UMinMethod_s & um ){
    cerr<<"Error: HCL_UMinMethod"<<endl;
    cerr<<"no copy constructor defined"<<endl;
    exit(1);
  }
  /** Usual constructor. 
    Legal choices of UpdateMethod:
    \begin{itemize}
    \item BFGS (limited memory BFGS, following Nocedal)
    \item NLCG (nonlinear conjugate gradient algorithm following Fletcher)
    \end{itemize}
    Legal choices of LineSearchMethod:
    \begin{itemize}
    \item DennisSchnabel
    \item Fletcher
    \item MoreThuente
    \end{itemize}
    See the documentation for the classes implementing these algorithms for
    more information (HCL\_UMin_s, HCL\_LineSearch_s).
    */

  HCL_UMinMethod_s( char * UpdateMethod, 
			char * LineSearchMethod,
			char * ParameterFile) {

    // select line search method

    trs = NULL;
    lnsrch = NULL;

    if (LineSearchMethod) {

      if (!strcmp(LineSearchMethod, "DennisSchnabel")) {
	lnsrch = new HCL_LineSearch_DS_s( ParameterFile );
      }
      else if (!strcmp(LineSearchMethod, "MoreThuente")) {
	lnsrch = new HCL_LineSearch_DS_s( ParameterFile );
      }
      else if (!strcmp(LineSearchMethod, "Fletcher")) {
	lnsrch = new HCL_LineSearch_Fl_s( ParameterFile );
      }
      else {
	cerr<<"Error: HCL_UMinMethod constructor"<<endl;
	cerr<<"parameter LineSearchMethod must be one of:"<<endl;
	cerr<<"   * DennisSchabel"<<endl;
	cerr<<"   * MoreThuente"<<endl;
	cerr<<"   * Fletcher"<<endl;
	cerr<<"Your choice was: "<<LineSearchMethod<<endl;
	exit(1);
      }
    }

    // select update method

    if (!strcmp(UpdateMethod,"BFGS") && lnsrch ) {
      meth = new HCL_UMin_lbfgs_s(lnsrch, ParameterFile);
    }
    else if (!strcmp(UpdateMethod,"NLCG") && lnsrch ) {
      meth = new HCL_UMinNLCG_s(lnsrch, ParameterFile);
    }
    else if (!strcmp(UpdateMethod,"TRCG")) {
      trs = new HCL_TRCG_s(ParameterFile);
      meth = new HCL_UMinTR_s(trs, ParameterFile);
    }
    else {
      cerr<<"Error: HCL_UMinMethod constructor"<<endl;
      cerr<<"parameter UpdateMethod must be one of:"<<endl;
      cerr<<"   * BFGS"<<endl;
      cerr<<"   * NLCG"<<endl;
      cerr<<"   * TRCG"<<endl;
      cerr<<"Your choice was: "<<UpdateMethod<<endl;
      exit(1);
    }
  }

  /// Destructor
  virtual ~HCL_UMinMethod_s(){
    HCL_delete(meth);
    if (trs) HCL_delete(trs);
    if (lnsrch) HCL_delete(lnsrch);
  }
  //@}

  virtual Table & Parameters() const {
    return meth->Parameters();
  }
  
  virtual HCL_EvaluateFunctional_s & LastEval() const {
    return meth->LastEval();
  }
  
  virtual void SetScaling( HCL_LinearOp_s * S,
			   HCL_LinearSolver_s * lsolver ){};
  virtual void SetScaling( HCL_LinearOp_s * S ){};
  virtual void UnSetScaling(){};

  virtual int Minimize( HCL_Functional_s & f,HCL_Vector_s & x ) {
    return meth->Minimize(f,x);
  }

  virtual ostream & Write(ostream & str) const {
    str<<"HCL_UMinMethod: optimization method selector"<<endl;
    str<<"Update method:"<<endl;
    meth->Write(str);
    if (lnsrch) {
      str<<"Line Search method:"<<endl;
      lnsrch->Write(str);
    }
    if (trs) {
      str<<"Trust Region Search method:"<<endl;
      trs->Write(str);
    }
    return str;
  }

};


#endif


