/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_UMinSubspace_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_UMINSUBSPACE_S_H_
#define HCL_UMINSUBSPACE_S_H_

#include "HCL_UMinProductDomain_s.h"
#include "HCL_MatTR_s.h"
#include "HCL_Basis_s.h"

/** {\bf HCL_UMinSubspace_s} is a variation on the idea of 
Kennett for minimizing a function with two or more distinct classes
of problem variables.  In the HCL setting, this means that the objective
function is defined on a product space and hence is an instance of
{\bf HCL\_FunctionalProductDomain\_s}.

The Kennett idea is to define a low-dimensional affine set in the
space of the problem variables spanned by the components of the gradient
and possibly the image of the gradient under the Hessian (and its image
under the Hessian, and so on).  The local quadratic model of the
objective function is then minimized over this low dimensional subspace.
In this version of the Kennett algorithm, a trust-region is imposed
on the step, and the More'-Sorensen algorithm ({\bf HCL\_MatTR\_s} for dense
trust region problems is used. */

class HCL_UMinSubspace_s: public HCL_UMinProductDomain_s
{
   private:
      Table * ParamTable;

   /**@name Input Parameters
      In a parameter file, the names may be prepended by "UMinSubspace::"
      or "UMinProductDomain::". */
   //@{
      /// (100) Maximum number of iterations.
      int MaxItn;
      /** (1) Typical value of the functional near the solution.  Setting
          this makes the gradient stopping tolerance more meaningful. */
      float Typf;
      /** (1) Typical norm of the unknown vector x near the solution.
          Setting this makes the gradient stopping tolerance more
          meaningful. */
      float TypxNorm;
      /** (1e-2) Gradient tolerance.  The algorithm attempts to locate
          a point where the relative gradient norm (i.e. the norm of the
          gradient scaled by the size of the vector x and by f(x))
          is less than this value. */
      float GradTol;
      /** (1e-20) Minimum allowable step.  If the algorithm takes a (relative)
          step less than this value in norm, it halts and reports
          "Possible convergence". */
      float MinStep;
      /** (1e+20) Maximum allowable step.  If the algorithm takes too many
          consecutive steps of length MaxStep, it assumes that the iterates
          are diverging and reports "Possible divergence". */
      float MaxStep;
      /** (5) Maximum number of steps of length MaxStep allowed before
          the algorithm decides that the iterates are diverging. */
      int CscMaxLimit;
      /** (0) Display level.  This determines how much information should be
          displayed during the execution of the algorithm.  Possible values
          are: 0 - No output; 1 - Function value and gradient norm after
          final iteration; 2 - Function value and gradient norm after every
          iteration. */
      int DispFlag;
      /** (0) Dump level.  This determines how much information should be
          sent to the dump file during the execution of the algorithm.
          Possible values are: 0 - No output; 1 - Function value and
          gradient norm after final iteration; 2 - Function value and
          gradient norm after every iteration. */
      int DumpFlag;
      /** (HCL\_UMinSubspace.DumpFile) Dump file name. */
      char DumpFile[81];
      /** (6) Display precision---the number of digits sent to the screen. */
      int DispPrecision;
      /** (6) Dump precision---the number of digits sent to the file. */
      int DumpPrecision;
      /** (0) If nonzero, the iterates are sent to a file using the
          {\bf Write} method from the vector class. */
      int TraceSteps;
      /** (HCL\_UMinSubspace.StepFile) File name for recording iterates. */
      char StepFile[81];
      /** (2) This number determines the dimension of the subspace generated
          from the components of the gradient by applying the Hessian. */
      int KrylovSubspaceLevel;
      /** (1e-1) Tolerance for reducing the trust region radius.  If the ratio
          of the actual to the predicted reduction falls below LowerTol,
          the radius is cut in half. */
      float LowerTol;
      /** (9e-1) Tolerance for increasing the trust region radius.  If the
          ratio of the actual to the predicted reduction is greater than
          UpperTol, the radius is floatd. */
      float UpperTol;
      /** (0) Flag indicating whether re-orthogonalization is to be applied
              to the basis for subspace. */
      int Reorthogonalize;
      /** (1.0) Initial trust region radius. */
      float Delta;
   //@}

      // Local variables

      int Itn,CscMax;
      float fx;
      float gnorm;
      float xnorm;
      float stepnorm;
      float ared,pred,ratio;
      float StepLength;
      int trcode;
      int OldDispPrec;
      int TermCode;
      ofstream * dump;
      ofstream * Steps;
      HCL_EvalFunctionalProductDomain_s * eval;
      HCL_LinearOp_s * S;
      HCL_LinearSolver_s * lsolver;
      HCL_ProductVector_s * tmpvec;
      HCL_ProductVector_s * step;
      HCL_ProductVector_s * xnext;
      HCL_MatTR_s * trsolver;

      void Initialize( HCL_VectorSpace_s & dom );
      int StoppingTest( HCL_ProductVector_s & x,
                        HCL_ProductVector_s * x1=NULL );
      void Display0( const HCL_Functional_s & f,const HCL_Vector_s & x );
      void Display( const HCL_Functional_s & f,const HCL_Vector_s & x );
      void Clean();

   /** Compute the basis for the subspace. */
      HCL_BasisProductSpace_s * Basis();
   /** Compute the projected operator and gradient. */
      void ProjectedProblem( HCL_BasisProductSpace_s & Bas,HCL_SymMatOp_s* & B,
                             HCL_RnVector_s* & g ) const;

   public:

   // Default constructor---disabled.
      HCL_UMinSubspace_s();
   // Copy constructor---disabled.
      HCL_UMinSubspace_s( const HCL_UMinSubspace_s & );
   /** Usual constructor; the only argument is the name of an optional
       parameter file.   If such a file is provided, it must be a plain
       text file with entries of the form

                 ParameterName = value

       or

                 UMinProductDomain::ParameterName = value

       or

                 UMinSubspace::ParameterName = value

       The second and third forms allow the same file to contain parameters for
       more than one algorithm to be given in the same file.  For example,
       one can have entries such as

                 UMinSubspace::DispFlag = 2
                 TRSolver::DispFlag = 1

       which set the display flags for the minimization algorithm and
       the trust region solver to different values. */
      HCL_UMinSubspace_s( char * fname=NULL );
   // Destructor.
      ~HCL_UMinSubspace_s();

   /** Access to parameter table. Algorithmic parameters can be accessed
       with Parameters().GetValue( "NAME",val ) or changed with
       Parameters().PutValue( "NAME",val ). */
      virtual Table & Parameters() const;

   /** LastEval returns a reference to the functional's evaluation object
       at the most recent point.  Thus, after the minimization has completed,
       the calling routine has access to the last function value and
       gradient. */
      virtual HCL_EvalFunctionalProductDomain_s & LastEval() const;

   /** {\bf SetScaling} defines a new inner product in terms of
       a symmetric, positive definite operator S: <x,y> = (x,Sy). This
       method is included for future use; it is not currently implemented. */
      virtual void SetScaling( HCL_LinearOp_s * S,
                               HCL_LinearSolver_s * lsolver=NULL );
   /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling();
   /** Minimize attempts to find a local minimizer of the functional f,
       using x as a starting guess. The return value is the termination
       code, as given above. */
      virtual int Minimize( HCL_FunctionalProductDomain_s & f,
                            HCL_ProductVector_s & x );
    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
