/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_UMin_lbfgs_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_UMIN_LBFGS_S_H
#define HCL_UMIN_LBFGS_S_H

#include "HCL_ScaleLinearOp_s.h"
#include "HCL_lmbfgsOp1_s.h"
#include "HCL_UMin_s.h"
#include "HCL_LineSearch_s.h"

/** {\bf HCL\_UMin\_lbfgs\_s} implements the limited memory BFGS algorithm
for unconstrained minimization. This algorithm is due to Nocedal
(see, for example, D. C. Liu and J. Nocedal, ``On the limited memory BFGS
method for large scale optimization methods'' Mathematical
Programming 45 (1989), pp. 503-528).

The minimization algorithm is based on the limited memory BFGS
inverse Hessian approximation (represented by a class, but hidden
from the user), and on a line search algorithm.

The line search is represented by an abstract base class,
{\bf HCL\_LineSearch\_s}; this allows the algorithm to be executed
with different choices of the line search algorithm.

The purpose of the algorithm is to take a functional with gradient
and a starting guess in the domain of the functional, and to produce
a local minimizer via a descent algorithm.  If desired, the calling
routine may provide a (symmetric positive definite) linear operator
(and a corresponding linear solver, if needed) to be used to change the
inner product on the solution space.  If these are provided, it is assumed
that the functional computes the gradient with respect to the default inner
product (i.e. the one provided by the vector class).

The primary methods of this class are:
\begin{itemize}
\item {\bf Minimize}, which performs the minimization;
\item {\bf Parameters},which provides a means of accessing or changing
      algorithmic parameters which control the iteration.  These
      parameters are typically set initially by passing a file name to
      the constructor; this file contains parameter values.
      Any parameters not specified for the constructor are assigned
      default values.  See the documentation of the parameters for more
      information.
\item {\bf LastEval}, which provides access to the "evaluation" object
      at the final point.
\end{itemize}

Use of this minimizer typically involves the following steps:
\begin{itemize}
\item Construct functional with gradient, i.e. an object of a class
      derived from {\bf HCL\_Functional\_s};
\item Construct a starting vector in the domain of the functional;
\item Construct a line search object, i.e. an object of a class
      derived from {\bf HCL\_LineSearch\_s};
\item Construct the minimizer by passing the constructor the line
      search object and a parameter file;
\item Call the {\bf Minimize} method;
\item Manipulate the value and/or gradient at the approximate minimizer
      through the {\bf LastEval} method.
\end{itemize}

Here's an example:
\begin{verbatim}
   MyFcnl f;  // MyFcnl must be derived from HCL\_Functional_s
   MyVector x( "x0" ); // Assume MyVector has a constructor which
                       //reads from a file
   HCL_LineSearch_DS_s lsearch( "lsearch.dat" );  // Dennis-Schnabel
                                                  // line search
   HCL_UMin_lbfgs_s umin( "umin.dat",&lsearch );  // Parameters read
                                                  // from "umin.dat"
   umin.Minimize( f,x );  // Search for local minimum
   cout << "Final value of f: " << umin.LastEval().ValueRef() << endl;
   cout << "Final gradient norm: "
        << umin.LastEval().GradientRef().Norm() << endl;
\end{verbatim}
(If the display flag is turned on, the final function value and gradient
norm will be displayed anyway, so the last two lines are not necessary.
But they illustrate that the last function value and gradient are stored
 in case they are needed.) */

class HCL_UMin_lbfgs_s: public HCL_UMin_s
{
   private:

      Table * ParamTable;

   /**@name Input Parameters
      In a parameter file, the names may be prepended by "UMinlbfgs::"
      or simply "UMin::". */
   //@{
      /// (100) Maximum number of iterations.
      int MaxItn;
      /** (1.0) Typical value of the functional near the solution.  Setting
          this makes the gradient stopping tolerance more meaningful. */
      float Typf;
      /** (1.0) Typical norm of the unknown vector x near the solution.
          Setting this makes the gradient stopping tolerance more
          meaningful. */
      float TypxNorm;
      /** (1.0e-2) Gradient tolerance.  The algorithm attempts to locate
          a point where the relative gradient norm (i.e. the norm of the
          gradient scaled by the size of the vector x and by f(x))
          is less than this value. */
      float GradTol;
      /** (1e-20) Minimum allowable step.  If the algorithm takes a (relative)
          step less than this value in norm, it halts and reports
          "Possible convergence". */
      float MinStep;
      /** (1e+20) Maximum allowable step.  If the algorithm takes too many
          consecutive steps of length MaxStep, it assumes that the iterates
          are diverging and reports "Possible divergence". */
      float MaxStep;
      /** (5) Maximum number of steps of length MaxStep allowed before
          the algorithm decides that the iterates are diverging. */
      int CscMaxLimit;
      /** (0) Display level.  This determines how much information should be
          displayed during the execution of the algorithm.  Possible values
          are: 0 - No output; 1 - Function value and gradient norm after
          final iteration; 2 - Function value and gradient norm after every
          iteration. */
      int DispFlag;
      /** (0) Dump level.  This determines how much information should be
          sent to the dump file during the execution of the algorithm.
          Possible values are: 0 - No output; 1 - Function value and
          gradient norm after final iteration; 2 - Function value and
          gradient norm after every iteration. */
      int DumpFlag;
      /** (HCL\_UMin\_lbfgs.DumpFile) Dump file name. */
      char DumpFile[81];
      /** (6) Display precision---the number of digits sent to the screen. */
      int DispPrecision;
      /** (6) Dump precision---the number of digits sent to the file. */
      int DumpPrecision;
      /** (0) If nonzero, the iterates are sent to a file using the
          {\bf Write} method from the vector class. */
      int TraceSteps;
      /** (HCL\_UMin\_lbfgs.StepFile) File name for recording iterates. */
      char StepFile[81];
      /** (1) Scale for the initial inverse Hessian approximation.  The
          initial inverse Hessian approximation is a multiple of the identity.
          This scale is adjusted after each iteration using information
          gained from the step taken; however, its initial value must be
          provided. */
     float InvHessianScale;
     /** (4) Maximum number of BFGS updates. See paper by Liu and Nocedal
         for details. */
      int MaxUpdates;
   //@}


      // local variables

      int OldDispPrec;
      int Itn,CscMax;
      int TermCode;
      int StDescDir;
      float f_x,f_xnext;
      float gnorm,snorm,xnorm;
      ofstream * dump;
      ofstream * Steps;

      // temporary vectors, etc.

      HCL_EvaluateFunctional_s * feval;
      int fevalFlag;
      HCL_ScaleLinearOp_s * H0;
      HCL_lmbfgsOp1_s * H;
      HCL_Vector_s * xnext;
      HCL_Vector_s * dir;
      HCL_Vector_s * tmpvec;

      HCL_LinearOp_s * S;
      HCL_LinearSolver_s * lsolver;

      HCL_LineSearch_s *lsearch;

   // Allocate temporaries, etc.

      void Initialize( HCL_VectorSpace_s & dom );

   // De-allocate temporaries, etc.

      void Clean();

   // print current status

      void Display( const HCL_Functional_s & f,
                    const HCL_Vector_s & x );

   // check stopping criterion

      virtual int StoppingTest( const HCL_Vector_s & x,
                                HCL_EvaluateFunctional_s & eval,
                                int ls_result, int MaxTkn);

   public:

   // Copy constructor---disabled.
      HCL_UMin_lbfgs_s( const HCL_UMin_lbfgs_s & um );
   /** Usual constructor.
       @param linesearch Linesearch class (optional). If no line search
              is specified, then the Dennis-Schnabel line search
              ({\bf HCL\_LineSearch\_DS\_s}) will be used.
       @param fname Parameter file name or NULL.  If given, this file
              contains algorithmic parameters (such as stopping tolerances),
              a full list of which is given elsewhere in the documentation
              of this class.  For example, to set the maximum number of
              iterations to 100, the file should contain a line of the
              form UMinlbfgs::MaxItn = 100 or, alternately, UMin::MaxItn = 100.
              If the line seach is not provided, then this file will also
              be passed to the line search constructor, and so can contain
              entries such as LineSearch::DispFlag = 1.
    */
      HCL_UMin_lbfgs_s( HCL_LineSearch_s * linesearch = NULL,
                        char * fname = NULL );

   // Destructor.
      virtual ~HCL_UMin_lbfgs_s();

   // Reset for another run
      void Reset() { Itn = 0;H->Reset(); }

   /** Access to parameter table. Algorithmic parameters can be accessed
       with Parameters().GetValue( "NAME",val ) or changed with
       Parameters().PutValue( "NAME",val ). */
      Table & Parameters() const { return *ParamTable; }

   /// Returns a reference to the last evaluation object.
      virtual HCL_EvaluateFunctional_s & LastEval() const;

   /** {\bf SetScaling} defines a new inner product in terms of
       a symmetric, positive definite operator S: <x,y> = (x,Sy). */
      virtual void SetScaling( HCL_LinearOp_s * S,
                               HCL_LinearSolver_s * lsolver );
   /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling();

   /** Minimize attempts to find a local minimizer of the functional f,
       using x as a starting guess. The return value is the termination
       code, as defined in the base class. */
      int Minimize( HCL_Functional_s & f,HCL_Vector_s & x );

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};
    
#endif
