/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_Vector_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef _HCL_VECTOR_S_H_
#define _HCL_VECTOR_S_H_

#include "HCL_Base.h"

class HCL_Vector_s;

/** {\bf HCL\_VectorSpace\_s} is the base class for all vector spaces in HCL.
(In HCL, "vector space" is synonomous with "Hilbert space".)
The purpose of a vector space in HCL is two-fold.
\begin{itemize}
\item Instead of being a set of vectors (as it should be mathematically), a
{\bf HCL\_VectorSpace\_s} is really a description of a certain type of
vector.  Therefore, two vectors are compatible (e.g. can be added) if
they belong to the same vector space, i.e. if their descriptions are
the same (note that a vector knows its space; see {\bf HCL\_Vector\_s}).

Example:
\begin{verbatim}
if( x.Space() != y.Space() ) // Check whether x and y are compatible
   // Error condition
\end{verbatim}
\item A {\bf HCL\_VectorSpace\_s} knows how to construct a vector belonging to
it, using the {\bf Member} method.  Thus, if in an algorithm one needs a
temporary vector (to store an intermediate calculation, say), it can be
constructed from the vector space, {\it without the algorithm knowing the
exact type of vector involved}.

Example:
\begin{verbatim}
HCL_Vector_s * xtemp = x.Space().Member(); // Get a vector like x
\end{verbatim}
\end{itemize}

A {\bf HCL\_VectorSpace\_s} returns its dimension via the {\bf Dim} method.

{\bf For implementors:} This class also provides a {\bf Test} method,
which performs a series of tests to check that the algebraic
operations are implemented consistently in the corresponding {\bf
HCL\_Vector\_s} class. For example, {\bf Test} verifies that addition
is commutative, that the two versions of {\bf Add} produce consistent
results, and so on. This method is intended to be useful for
implementors of new vector classes. The concrete vector classes
provided with HCL have already passed.

All of the methods in this class are pure virtual, with the sole exception 
of {\bf operator !=}.
*/

class HCL_VectorSpace_s: public HCL_Base
{
   public:  

   // default constructor
      HCL_VectorSpace_s(){}
   /** Constructs identical copy.  {\bf For implementors:} A copy constructor 
     provides a simple means to implement {\bf Replicate} in a concrete class 
     derived from {\bf HCL\_VectorSpace\_s}. */
      HCL_VectorSpace_s(const HCL_VectorSpace_s & x){}
   // destructor
      virtual ~HCL_VectorSpace_s(){}

   /** Returns nonzero if spaces are identical. {\bf For
     implementors:} This comparitor must be implemented for each
     concrete derived class. It should begin by comparing the
     addresses of the two spaces. If these are the same, no further
     work is required, and the operator returns a nonzero. Otherwise,
     it is necessary to compare the data defining the spaces. To
     access this data, it is necessary to cast the comparison space
     (i.e. the argument of {\bf operator ==}) to the type of {\bf
     this}. To do this safely, each vector space class should declare
     a static char (named {\bf Type}, for example) in private data,
     and implement {\bf void * GetType()}. {\bf GetType} should return
     the address of the static char, cast to {\bf void *}:
     
     \begin{verbatim} 
     void * MyVectorSpace::GetType() const {
       return (void *)&Type;
     }
     \end{verbatim}

     Since {\bf GetType} is declared public and pure virtual in the
     base class, its value is available to other vector spaces before
     the cast. The return values of {\bf GetType} should be compared
     for the two spaces; if they are the same, then the second space
     is an object of the same class as the first.  The cast can then
     be done safely, and then the characteristic information about the
     spaces can be compared. See the comparitor for {\bf
     HCL\_RnSpace\_s} for a simple example. [This primitive runtime
     type check, the only runtime type check built into HCL, will
     become obsolete in a future version incorporating the ISO
     standard dynamic cast.]  */

      virtual int operator == (const HCL_VectorSpace_s &) const = 0;

   /** Returns nonzero if spaces differ. {\bf For implementors:} This
       comparitor is implemented in this base class, and need not be
       overridden. */

      virtual int operator != (const HCL_VectorSpace_s & x) const {return !(operator ==(x)); }

   /* Get type of vector space for comparison.  {\bf GetType} makes it
       possible to check that two vector spaces are of the same
       derived class before comparing them.  Each derived class has a
       static char field called {\bf Type}; {\bf GetType} just returns
       the address of {\bf Type}, cast to void*.  This is a primitive
       form of Run Time Type Information. */

      virtual void * GetType() const = 0;

   /** Returns address of dynamically allocated (and uninitialized)
       vector in this space.  This method constructs a vector
       belonging to this vector space and returns a pointer to it.
       Note that the calling routine is responsible for deleting the
       vector when it is no longer needed.  {\bf Member} is used to
       create temporary vectors "abstractly", i.e. without knowing the
       exact type of the vector desired.  See the documentation for
       this class (above). */

      virtual HCL_Vector_s * Member() const = 0;

   /** Returns address of independent, identical copy of this vector
       space.  This method constructs a copy of this vector space and
       returns a pointer to it. {\bf For implementors:} For concrete
       child classes of {\bf HCL\_VectorSpace\_s} having a copy
       constructor, {\bf Replicate} can just be implemented to call
       the copy constructor and return the the result:
       \begin{verbatim} 
       MyVectorSpace::Replicate() { 
         return new MyVectorSpace( *this ); 
       } 
       \end{verbatim} */

      virtual HCL_VectorSpace_s * Replicate() const = 0;

   /** Returns the dimension of the vector space. */
      virtual int Dim() const = 0;

   /** Test routine, useful for implementors of new vector
       classes to check for coding errors.  This routine checks that the
       axioms of a vector space hold (up to round-off errors) for
       random vectors.  The return value is the number of errors
       detected.  The first input is a display flag; if it is zero, no
       results are displayed; if it is 1, the number of errors
       detected is displayed; if it is greater than 1, the result of
       each individual test is displayed.  The second input sets the
       tolerance for the floating-point tests; an error is reported if
       two results which ought to be equal differ by a relative amount
       greater than tol times machine epsilon. */

      int Test( int display = 0,float tol = 100 ) const;

   protected:

    // Test routines.
   int TestAdd(HCL_Vector_s&, HCL_Vector_s&, float, int) const;
   int TestSub(HCL_Vector_s&, HCL_Vector_s&, float, int) const;
   int TestMulDist(HCL_Vector_s&, HCL_Vector_s&, float,int) const;
   int TestAddAssoc(HCL_Vector_s&, HCL_Vector_s&, HCL_Vector_s&,
                    float,int) const;
   int TestAddComm(HCL_Vector_s&, HCL_Vector_s&, float,int) const;
   int TestSubAddNeg(HCL_Vector_s&, HCL_Vector_s&, float,int) const;
   int TestScaleAdd(HCL_Vector_s&, HCL_Vector_s&, float,int) const;
   int TestNegMul(HCL_Vector_s&, float,int) const;
   int TestSubDist(HCL_Vector_s&, HCL_Vector_s&, HCL_Vector_s&,
                   float,int) const;
   int TestIdentity(HCL_Vector_s&, float,int) const;
   int TestZeroMul0(HCL_Vector_s&, float,int) const;
   int TestAddZero(HCL_Vector_s&, float,int) const;
   int TestNegZero(float,int) const;
   int TestAddNeg(HCL_Vector_s&, float,int) const;
   int TestMul(HCL_Vector_s& x, float,int) const;
   int TestSAddAddS(HCL_Vector_s& x, HCL_Vector_s& y, float,int) const;
   int TestNormNorm2(HCL_Vector_s& x, float, int) const;
   int TestNormInner(HCL_Vector_s& x, float,int) const;
   int Compare(HCL_Vector_s&, HCL_Vector_s&, float, float *, float *)
      const;
   void Message( const char * str1,const char * str2,float ndiff,
                 float n1 ) const;
   void Message1( const char * str1,const char * str2,float ndiff,
                 float n1 ) const;
};

/** {\bf HCL\_Vector\_s} is the base class for all vectors in HCL.  Vectors
are implemented as concrete classes derived from {\bf HCL\_Vector\_s}
so that the optimization code need make {\em no} assumptions about
the data structures used to implement the vectors.  This is essential
when nonstandard data structures (such as disk files for very large
problems) are used.

Note that each class derived from {\bf HCL\_Vector\_s} must correspond
to a class derived from {\bf HCL\_VectorSpace\_s}.  That is, if one wishes
to create a class {\bf MyVector}, derived from {\bf HCL\_Vector\_s},
one is obligated to first define {\bf MyVectorSpace}, derived from
{\bf HCL\_VectorSpace\_s}. */
class HCL_Vector_s: public HCL_Base
{
 public:

   // Default constructor.
      HCL_Vector_s(){}
   // Copy constructor
      HCL_Vector_s( const HCL_Vector_s & x){}
   // destructor
      virtual ~HCL_Vector_s(){}


   /**@name Vector operations (z is the object invoking the method). */
   //@{
   /// z <-- x
      virtual void Copy( const HCL_Vector_s & x ) = 0;
   /// z <-- (-z)
      virtual void Neg() = 0;
   /// z <-- a*z
      virtual void Mul( const float & a )  = 0;
   /// z <-- z + x
      virtual void Add( const HCL_Vector_s & x ) = 0;
   /// z <-- a*x
      virtual void Mul( const float & a,const HCL_Vector_s & x ) = 0;
   /// z <-- x + y
      virtual void Add( const HCL_Vector_s & x,
                        const HCL_Vector_s & y ) = 0;
   /// z <-- z - x
      virtual void Sub( const HCL_Vector_s & x ) = 0;
   /// z <-- x - y
      virtual void Sub(const HCL_Vector_s & x,
                       const HCL_Vector_s & y ) = 0;
   /// z <-- a*z + x
      virtual void ScaleAdd(const float & a,
                            const HCL_Vector_s & x ) = 0;
   /// z <-- z + a*x
      virtual void AddScale(const float & a,
                            const HCL_Vector_s & x ) = 0;
   /// z <-- x + a*y
      virtual void AddScale(const float & a, const HCL_Vector_s & x,
                            const HCL_Vector_s & y ) = 0;

   /// Inner product of z with x.
      virtual float Inner( const HCL_Vector_s & x ) const = 0;
   /// Norm of z.
      virtual float Norm() const = 0;
   /// Norm squared of z.
      virtual float Norm2() const = 0;
   /// Returns a reference to the space of which z is a member
      virtual HCL_VectorSpace_s & Space() const = 0;
   /// z <-- 0
      virtual void Zero() = 0;
   /** Returns a "random" vector.  The meaning of random is left up to
       the implementor of each derived class. */
     virtual void Random() = 0;
   //@}

   /**@name Access to components.

      {\bf Dim} and {\bf Data} return the dimension of the data and a
      pointer to the internal data array, respectively. Efficient
      access to the data of the vector object is usually most easily
      implemented using these two functions. */

      //@{
     /// {\bf Dim} returns dimension of space.
      virtual int Dim() const { return Space().Dim(); }
      /** {\bf Data} returns a pointer to the array of components, which
          is assumed to exist or to be created for this purpose.  The
          default implementation returns NULL so that the method need not
          be re-implemented for, e.g., out-of-core classes. */
      virtual float * Data();
      //@}

      /**@name Array operations - ``Matlab'' methods. 
	In describing these,
       \TEX{$c_i$} denotes the ith component of the current array. */
      //@{
	/// \TEX{$c_i = a,\,\,i=1,n$}
     virtual void Fill( float a ) = 0;
     /** \TEX{$c_i = c_i + a$} - overload of Add with scalar argument
         {\em a la} Matlab. */
     virtual void Add( float a ) = 0;
     /** \TEX{$c_i = x_i + a$} - overload of Add with scalar argument
         {\em a la} Matlab. */
     virtual void Add( const HCL_Vector_s & x, float a ) = 0;
     /// \TEX{$c_i = c_i*x_i$} - matrix multiply by diag(x)
     virtual void DiagScale( const HCL_Vector_s & x ) = 0;
     /// \TEX{$c_i = y_i*x_i$}
      virtual void DiagScale( const HCL_Vector_s & x,
                              const HCL_Vector_s & y ) = 0;
   /** \TEX{$c_i = c_i/{\rm maxmod}(x_i,tol)$} - matrix division by diag(x). 
     Safeguarded against overflow - if tol is less
     that the smallest safe divisor sfmin as defined by LAPACK::[SD]LAMCH, then
     tol is replaced by sfmin - thus the default is tol=0.0, which effectively
     means tol=sfmin. */
      virtual void DiagRecipScale( const HCL_Vector_s & x,
				   float tol=0.0) = 0;
   /** \TEX{$c_i = y_i/{\rm maxmod}(x_i,tol)$} - matrix division by diag(x).
     Safeguarded against overflow - if tol is less
     that the smallest safe divisor sfmin as defined by LAPACK::[SD]LAMCH, then
     tol is replaced by sfmin - thus the default is tol=0.0, which effectively
     means tol=sfmin. */
      virtual void DiagRecipScale( const HCL_Vector_s & x,
				   const HCL_Vector_s & y,
				   float tol=0.0) = 0;

   /// "Matlab" name for DiagScale, first version
      virtual void DotStar( const HCL_Vector_s & x ) {
        this->DiagScale(x);
      }
   /// "Matlab" name for DiagScale, second version
      virtual void DotStar( const HCL_Vector_s & x,
                            const HCL_Vector_s & y ) {
        this->DiagScale(x,y);
      }
   /// "Matlab" name for DiagRecipScale, first version.
      virtual void DotSlash( const HCL_Vector_s & x,
			     float tol=0.0) {
	this->DiagRecipScale(x,tol);
      }
   /// "Matlab" name for DiagRecipScale, second version.
      virtual void DotSlash( const HCL_Vector_s & x,
			     const HCL_Vector_s & y,
			     float tol=0.0) {
        this->DiagRecipScale(x,y,tol);
      }
      /// returns largest of components or -FLT\_MAX 
      virtual float Max() const = 0;
      /// \TEX{$c_i = {\rm max}(c_i,x_i) $}
      virtual void Max( const HCL_Vector_s & x ) = 0;
      /// \TEX{$c_i = {\rm max}(y_i,x_i) $}
      virtual void Max( const HCL_Vector_s & x,
                        const HCL_Vector_s & y ) = 0;
      /// returns smallest of components or FLT\_MAX
      virtual float Min() const = 0;
      /// \TEX{$c_i = {\rm min}(c_i,x_i) $}
      virtual void Min( const HCL_Vector_s & x ) = 0;
      /// \TEX{$c_i = {\rm min}(y_i,x_i) $}
      virtual void Min( const HCL_Vector_s & x,
                        const HCL_Vector_s & y ) = 0;
      /// returns \TEX{$\sum_{i=1}^n\,c_i$}
      virtual float Sum() const = 0;
      /// \TEX{$c_i = |c_i|$}
      virtual void Abs() = 0;
      /// \TEX{$c_i = |x_i|$}
      virtual void Abs( const HCL_Vector_s & x ) = 0;
      /** \TEX{$c_i = {\rm sign}\,(c_i)$}. Sign returns 1 for positive
           arguments, -1 for negative arguments, and 0 for zero arguments. */
      virtual void Sign() = 0;
      /// \TEX{$c_i = {\rm sign}\,(c_i - t)$}
      virtual void Sign( float t ) = 0;
      /// \TEX{$c_i = {\rm sign}\,(c_i-x_i)$}
      virtual void Sign( const HCL_Vector_s & x ) = 0;
      /// \TEX{$c_i = {\rm sign}\,(x_i-y_i)$}
      virtual void Sign( const HCL_Vector_s & x,
                         const HCL_Vector_s & y ) = 0;
      /// \TEX{$c_i = {\rm sign}\,(x_i - t)$}
      virtual void Sign( const HCL_Vector_s & x,float t ) = 0;
      /// \TEX{$c_i = {\rm sign}\,(t - x_i)$}
      virtual void Sign( float t, const HCL_Vector_s & y ) = 0;
      /// Mask function: \TEX{$c_i = 1$} if \TEX{$c_i$} > \TEX{$t$}, else 0
      virtual void Greater( float t ) = 0;
      /// Mask function: \TEX{$c_i = 1$} if \TEX{$c_i$} > \TEX{$x_i$}, else 0
      virtual void Greater( const HCL_Vector_s & x ) = 0;
      /// Mask function: \TEX{$c_i = 1$} if \TEX{$x_i$} > \TEX{$y_i$}, else 0
      virtual void Greater( const HCL_Vector_s & x,
                         const HCL_Vector_s & y ) = 0;
      /// Mask function: \TEX{$c_i = 1 $} if \TEX{$x_i$} > \TEX{$t$}, else 0
      virtual void Greater( const HCL_Vector_s & x, float t ) = 0;
      /// Mask function: \TEX{$c_i = 1 $} if \TEX{$t$} > \TEX{$x_i$}, else 0
      virtual void Greater( float t, const HCL_Vector_s & x ) = 0;
      /** \TEX{$c_i = c_i^{p}$} - arbitrary real power.
	Should only be invoked for nonintegral \TEX{$p$} when
	\TEX{$c_i \ge 0$}, but have not found a good way to enforce 
	this constraint, so rely on runtime */
      virtual void Power( float p ) = 0;
      /** \TEX{$c_i = x_i^{p}$} - arbitrary real power.
	should only be invoked for nonintegral \TEX{$p$} when
	\TEX{$x_i \ge 0$}, but have not found a good way to enforce 
	this constraint, so rely on runtime */
      virtual void Power( float p, const HCL_Vector_s & x ) = 0;
      /** \TEX{$c_i=\sqrt{c_i}$} - componentwise square root.
      exits with error condition if any component is negative. */
      virtual void Sqrt() = 0;
      /** \TEX{$c_i=\sqrt{x_i}$} - componentwise square root.
      exits with error condition if any component is negative. */
      virtual void Sqrt( const HCL_Vector_s & x ) = 0;
      /// \TEX{$c_i=\exp{c_i}$} 
      virtual void Exp() = 0;
      /// \TEX{$c_i=\exp{x_i}$} 
      virtual void Exp( const HCL_Vector_s & x ) = 0;
      /** \TEX{$c_i=\log{c_i}$} - componentwise natural log.
	exits with error condition if any component is nonpositive. */
      virtual void Log() = 0;
      /** \TEX{$c_i=\log{x_i}$} - componentwise natural log.
	exits with error condition if any component is nonpositive. */
      virtual void Log( const HCL_Vector_s & x ) = 0;
      //@}

};

#endif
