/*********************************************************************
 This file was automatically generated by d2s from
 the file SGFlite_d.H.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef _SGF_D
#define _SGF_D

#include "HCL_Arch_s.h"
#include "HCL_Vector_s.h"
#include "maxmin_s.h"
#include "absTable.H"
#include "Table.h"
#include "HCL_Recorder_s.H"

// tolerances, max dimension 
#define SPACE_CMP_TOL 1.0e-5
#define STEP_TOL 1.0e-10
// NOTE: for naming conventions in constructor to work,
// must have dimensions in single digits, so MAX_DIM <= 9
#define MAX_DIM 9
// normalization for random arrays
#ifndef RAND_MAX
#define RAND_MAX 32768
#endif
// max length of Index list for default constructor
#define BIG_INT 32767
#define DATAFILENAME_LEN 256

class SGFVector_s;

/** 

{\tt SGFSpace} objects define the geometry and attributes of uniform
rectilinear grids. The class accomodates up to 9-dimensional data, and
records its organization and any auxiliary scalar information required
to properly interpret the data. Vectors in an {\tt SGFSpace} are 
\Ref{SGFVector_s}s.

Motivated by and designed after the Stanford Exploration Project
``classic'' seismic disk data format. After seismic usage, grid
geometry and auxiliary information are ``headers'' or ``header words''
in this documentation.

Double and single precision versions defined. The description below
refers to the float precision class {\tt SGFSpace\_s}. In single
precision, all floats change to floats, otherwise everything is the
same. {\tt SGFSpace} stores header words (grid parameters etc.) internally
in a parameter table, which is an instance of an \Ref{AbstractTable}
object. The table is read-only data: access is provided through a {\tt
const} reference (method {\tt ParametersRef()}) and through returning
a pointer to an independent copy (method {\tt Parameters()}).  Neither
of these access methods permit the user to alter the parameters of the
{\tt SGFSpace} object. Creating a new {\tt SGFSpace} with a new
parameter table having only a few differences with that of the current
{\tt SGFSpace} - a fairly frequently occurring task - might be
accomplished with code resembling this fragment:

\begin{verbatim} 
AbstractTable * tnew = oldsp.Parameters();
tnew->PutValue("WineType","Beaujolais"); 
SGFSpace * newsp = new
SGFSpace(tnew); 
HCL_delete(tnew); 
\end{verbatim} 
Note that {\tt
Parameters} allocates storage, which the calling locus of control must
manage.

{\bf Required} header words: 
\begin{itemize} 
\item number of samples in each direction: int {\tt n1,n2,..., ndim} 
\item step size in each direction: float {\tt d1,d2,..., ddim} 
\item coordinate of first sample in each direction:
float {\tt o1, o2,..., odim} 
\end{itemize}

{\bf Optional} header words: 
\begin{itemize}
\item data dimension: int {\tt dim} (if not present, inferred from
dimensional parameters)
\item data type identifier: int {\tt esize} = 8 for float precision,
4 for single precision [provided for backward compatibility with other
software reading SEP disk data format] 
\item name of data file ({\tt NULL} or arbitrary for definition of
Space only): char * {\tt in}
\item any user-specified integer, real (float
or float) or string data (``attributes'') needed for proper
interpretation of the data in the particular {\tt SGFSpace}.
\end{itemize}

{\bf Notes:} 
\begin{itemize} 
\item The {\tt in} string should be the
filename of the binary data, when the {\tt SGFSpace} object is the
{\tt SGFSpace} of an \Ref{SGFVector_s}. To define the space by itself
(i.e. the grid plus its additional attributes, if any), initialize
this string arbitrarily (eg. to "NULL"), or simply leave this 
parameter out of the list.
\end{itemize}

{\bf Example:} The following file, input to the {\tt SGFSpace} usual
constructor, constructs an {\tt SGFSpace} implementing a 2-dimensional
grid having additional attributes of WaterDepth and WineType.

Since no binary data filename appears ({\tt in=...}), this
file is suitable only to define an {\tt SGFSpace}. A similar file
specifying a vector in this space would include instead an actual
filename ({\tt in=<filename>}) as a ``pointer - to - data''.

\begin{verbatim}
  n1=101
  n2=201
  d1=0.01
  d2=27.9
  o1=0.0
  o2=34723.39487
  WaterDepth=154.5
  WineType=VinRougeOrdinaire
\end{verbatim}

*/
 
class SGFSpace_s: public HCL_VectorSpace_s {

private:

  AbstractTable * t; // table of parameters
  static char Type;   // type ptr
  int buflen;         // dimension, i.e. length of vector (not number of axes)
  float vol;         // volume of elementary cell in grid

public:

  /**@name Constructors */
  //@{
  /// Default constructor
  SGFSpace_s(){
    cout<<"Error: SGFSpace_s default constructor called"<<endl;
    exit(1);
  }
  /// Copy constructor 
  SGFSpace_s( const SGFSpace_s & S );
  /** Usual constructor - reads grid parameter table from file. Also permits user
      to override default dimension computation. */
  SGFSpace_s( const char * fname, int dim=0 );
  /** Constructor accepting externally constructed table. Also permits user
      to override default dimension computation. */
  SGFSpace_s( const AbstractTable & tbl, int dim=0 );
  /** Constructor accepting pointer to externally constructed table. Also
      permits user to override default dimension computation. */
  SGFSpace_s( AbstractTable * tbl, int dim=0 );
  /// Destructor
  ~SGFSpace_s() {
    HCL_delete(t);
  }
  //@}

  /* access to Type */
  virtual void * GetType() const { 
    return (void*)&Type; 
  }

  /**@name Access to header table */
  //@{
  /// returns header word as int
  virtual int GetValue(const char *name, int &value) const {
    return t->GetValue(name,value);
  }
  /// returns header word as float
  virtual int GetValue(const char *name, float &value) const {
    return t->GetValue(name,value);
  }
  /// returns header word as character array
  virtual int GetValue(const char *name, char *value, int length = 0) const {
    return t->GetValue(name,value,length);
  }
  ///
  virtual int GetArrayValue(const char *name, int ind, int &value) const {
    return t->GetArrayValue(name,ind,value);
  }
  ///
  virtual int GetArrayValue(const char *name, int ind, float &value) const {
    return t->GetArrayValue(name,ind,value);
  }
  ///
  virtual int GetArrayValue(const char *name, int ind, char *value, int length = 0) const {
    return t->GetArrayValue(name,ind,value,length);
  }

  /// returns const reference to table
  virtual const AbstractTable & ParametersRef() const { return *t; }
  /** returns independent copy of table - uses {\tt new}, so calling
    unit must manage storage */
  virtual AbstractTable * Parameters() const { return t->Replicate(); }
  //@}

  /// returns dimension = total number of samples
  int Dim() const {return buflen;}
  /// returns volume of basic grid cell
  float Vol() const {return vol;}

  // comparison 

  virtual int operator == ( const HCL_VectorSpace_s & x ) const;

  virtual int operator != ( const HCL_VectorSpace_s & x ) const {
    return !operator==( x );
  }

  /* virtual constructor */

  virtual HCL_Vector_s * Member() const;

  /* replicator */

  virtual HCL_VectorSpace_s * Replicate() const {
    return new SGFSpace_s(*this);
  }

  /** intersection of this SGFSpace with another. Nonvoid return if
    grids are compatible (same increments dx, origins ox differ by
    integer multiples of increments). */

  SGFSpace_s * Intersect( const SGFSpace_s & B ) const;

  /* I/O */

  virtual ostream & Write( ostream & str ) const;

};

/** 

{\tt SGFVector} implements sampled functions on a uniform rectilinear
grid as {\tt HCL\_Vector}s. ``SGF'' stands for Sampled Grid
Function. It designates a data structure for arbitrary (< 10)
dimensional regularly sampled real single and float precision
hypercubical data, plus auxiliary scalar data (integer, single or
float precision floating point, or string) describing the grid
geometry and other information necessary to properly interpret the
data.  Many natural data types in science and engineering fit this
description.  Other types of data do not conform to the SGF design,
for example because the data sampling is geometrically irregular. Such
data and associated processing tasks are beyond the scope of SGF.

From the user point of view, SGF is in fact identical to the Stanford
Exploration Project's ``classic'' (pre-1996) disk data storage format,
with a few restrictions and extensions noted below.

The \Ref{SGFSpace_s}
class implements the vector spaces in which {\tt
SGFVector}s live.  Grid and scalar attribute specification is proper
to \Ref{SGFSpace_s}, and the user should consult the documentation on
that class for details. Every {\tt SGFVector} knows what space it
belongs to (as does every {\tt HCL\_Vector}), and in particular has access
to all of the dimensional anpd auxiliary information proper to the
space. The {\tt SGFVector} class provides convenient access functions,
described below, which return values of these parameters by keyword
(the keywords are described in the \Ref{SGFSpace_s} documentation).

Sample data of an {\tt SGFVector} may be stored in a (native) binary data
file, designated by the {\tt in} keyword in the {\tt SGFSpace}
parameter table.  The storage order is as indicated by the ordering of
directions in the {\tt SGFSpace}: the first direction has the fastest
index, followed by the second, etc. For example if the first direction
represents time, the second a space coordinate, then the fastest index
is time.

If no {\tt in=} keyword-value pair appears in the parameter table, then
no file is associated with the {\tt SGFVector}, and its data will be lost
when the object is deleted. 

The Data() method returns a float or float pointer to the {\bf data
buffer} holding the current in-core record.

{\tt SGFVector} has two constructors, one for ``old'' vectors based on
preexisting data files, the other for ``new'' vectors for which no
data file need exist.  The ``old'' vector constructor reads the data
from disk, whereas the ``new'' vector constructor presumes that the
data is either nonexistent or out-of-date, so does not populate the
data buffer on construction.

{\bf Notes:}
\begin{itemize}
\item The vector constructors take a filename argument - optional for
``new'' vectors, mandatory for ``old'' ones. Passing this argument
causes standard SEP-format files to be created: an ascii header
file and a binary data file, the latter having the same name as
the former except for an appended {\tt @}. Data in the buffer
is flushed to the binary file on destruction, if the file exists.
\item A ``new'' vector created without a filename argument is not
connected to a file, and its data is lost on destruction. Such vectors
are suitable as non-checkpointed workspace in applications programs.
Note that this is the mode of creation implicit in the action of
the {\tt Member} method of the \Ref{SGFSpace_s} class.
\end{itemize} */

class SGFVector_s: public HCL_Vector_s {

private:

  SGFSpace_s * sp;
  int save_file;
  float * databuf;
  int buflen;
  FILE * file_ptr;
  char * datafile;
  char * hdrfile;

  float zero;

public:

  /**@name Constructors and destructor */
  //@{
  /// Default constructor
  SGFVector_s(){
    cerr << "Error: SGFVector_s default constructor called" << endl;
    exit(1);
  }
  /// Copy constructor - disabled
  SGFVector_s( const SGFVector_s & S ) {
    cerr << "Error: SGFVector_s copy constructor called" << endl;
    exit(1);
  }
  /** Usual constructor for old data. String fname must be name of existing 
    SGFSpace header file, with in keyword pointing to binary data file. */
  SGFVector_s( char * fname);
  /** Usual constructor for new data. Constructs vector in space given
    by first arguement. If string fname given, opens data file on that
    name, and sets flag to write data to this file on object destruction.
    Otherwise data is lost when object is destroyed. */
  SGFVector_s( const SGFSpace_s * s, char * fname = NULL);

  /** Destructor */
  ~SGFVector_s();
  //@}

  //*** dimension access
  virtual int Dim() const { return sp->Dim(); }
  virtual float Vol() const { return sp->Vol(); }

  /**@name Header and data access methods */
  //@{
  /** returns length in words of data buffer */
  int BufLen() { return buflen; }
  /// returns header word as int
  virtual int GetValue(const char *name, int &value) const {
    return sp->GetValue(name,value);
  }
  /// returns header word as float
  virtual int GetValue(const char *name, float &value) const {
    return sp->GetValue(name,value);
  }
  /// returns header word as character array
  virtual int GetValue(const char *name, char *value, int length = 0) const {
    return sp->GetValue(name,value,length);
  }
  ///
  virtual int GetArrayValue(const char *name, int ind, int &value) const {
    return sp->GetArrayValue(name,ind,value);
  }
  ///
  virtual int GetArrayValue(const char *name, int ind, float &value) const {
    return sp->GetArrayValue(name,ind,value);
  }
  ///
  virtual int GetArrayValue(const char *name, int ind, char *value, int length = 0) const {
    return sp->GetArrayValue(name,ind,value,length);
  }
  /// returns const reference to table
  virtual const AbstractTable & ParametersRef() const { return sp->ParametersRef(); }
  /** returns independent copy of table - uses {\tt new}, so calling
    unit must manage storage */
  virtual AbstractTable * Parameters() const { return sp->Parameters(); }
  /** Return pointer to data buffer */
  float * Data() { return databuf; }
  //@}

  /* space access */

  virtual HCL_VectorSpace_s & Space() const { return *(SGFSpace_s *)sp;}

  /* copying */

  virtual void Copy( const HCL_Vector_s & xx );

  /* required operations */

  virtual void Neg();
  virtual void Mul( const float & a );
  virtual void Mul( const float & a,const HCL_Vector_s & xx );
  virtual void Add( const HCL_Vector_s & xx );
  virtual void Add( const HCL_Vector_s & xx,
		   const HCL_Vector_s & yy );

  virtual float Inner( const HCL_Vector_s & xx ) const;

  /* derived operations */

  virtual void Sub( const HCL_Vector_s & xx );
  virtual void Sub( const HCL_Vector_s & xx,
		   const HCL_Vector_s & yy );
  virtual void ScaleAdd( const float  & a,
			const HCL_Vector_s & xx );
  virtual void AddScale( const float  & a,
			const HCL_Vector_s & xx );
  virtual void AddScale( const float  & a,
			const HCL_Vector_s & xx,
			const HCL_Vector_s & yy );
  float Norm() const { return sqrt( Norm2() ); }
  float Norm2() const;
  
  /* Fill */

  virtual void Fill( float a );
  virtual void Zero();
  virtual void Random();

  /* View - experimental interface to external graphics */

  int ChangeViewMethod(char * meth);

  int ChangeViewParameters(char * par);

  int View() const;

  // Subarray extraction
  
  virtual HCL_Vector_s * SubArray(int begin, 
					   int end) const {
    cerr<<"Error: SGFVector"<<endl;
    cerr<<"subarray operators not implemented yet"<<endl;
    return NULL;
  }

  virtual HCL_Vector_s * SubArray(int begin, 
					   int stride,
					   int end) const {
    cerr<<"Error: SGFVector"<<endl;
    cerr<<"subarray operators not implemented yet"<<endl;
    return NULL;
  }

  // scalar argument overloads of Add

  virtual void Add( float a );
  virtual void Add( const HCL_Vector_s & x, float a );

  // Diagonal scaling
  
  virtual void DiagScale( const HCL_Vector_s & x );
  virtual void DiagScale( const HCL_Vector_s & x,
			  const HCL_Vector_s & y );
  virtual void DiagRecipScale( const HCL_Vector_s & x,
			       float tol=0.0);
  virtual void DiagRecipScale( const HCL_Vector_s & x,
			       const HCL_Vector_s & y,
			       float tol=0.0);

  // "MATLAB" methods

  virtual float Max() const;
  virtual float Min() const;
  virtual float Sum() const;
  virtual void Max( const HCL_Vector_s & x );
  virtual void Max( const HCL_Vector_s & x,
		    const HCL_Vector_s & y );
  virtual void Min( const HCL_Vector_s & x );
  virtual void Min( const HCL_Vector_s & x,
		    const HCL_Vector_s & y );
  virtual void Abs();
  virtual void Abs( const HCL_Vector_s & x );
  virtual void Sign();
  virtual void Sign( float t );
  virtual void Sign( const HCL_Vector_s & x );
  virtual void Sign( const HCL_Vector_s & x,
		     const HCL_Vector_s & y );
  virtual void Sign( const HCL_Vector_s & y,float t );
  virtual void Sign( float,const HCL_Vector_s & y );
  virtual void Greater( float t );
  virtual void Greater( const HCL_Vector_s & x );
  virtual void Greater( const HCL_Vector_s & x,
			const HCL_Vector_s & y );
  virtual void Greater( const HCL_Vector_s & y,float t );
  virtual void Greater( float,const HCL_Vector_s & y );
  virtual void Power( float );
  virtual void Power( float p, const HCL_Vector_s & x );
  virtual void Sqrt();
  virtual void Sqrt( const HCL_Vector_s & x );
  virtual void Exp();
  virtual void Exp( const HCL_Vector_s & x );
  virtual void Log();
  virtual void Log( const HCL_Vector_s & x );

  /**@name Write method */
  //@{
  /// Prints SGFSpace header table and name of binary data file 
  virtual ostream & Write( ostream & str ) const {
    str << "*** SGFVector_s Object: Out-of-Core HCL_Vector *** " << endl;
    str << "Space (contents of header file "<<hdrfile<<"):" <<endl;
    sp->Write( str );
    str << "data file name "<<datafile<<endl;
    return str;
  }
  //@}
};

/*
class SGFRecorder_s: public HCL_Recorder_s {

private:

  int n;
  int m;
  SGFSpace_s * rsp;
  SGFSpace_s * tsp;
  SGFVector_s * tape;

public:

  // Constructors
  SGFRecorder_s() {}

  SGFRecorder_s( SGFRecorder_s & R ) {
    cerr<<"Error: no copy constructor for class SGFRecorder"<<endl;
    exit(1);
  }

  SGFRecorder_s( SGFSpace_s * RecordSpace, 
		 int MaxRecords,
		 char * fname = NULL);

  ~SGFRecorder_s();

  /// Record - store input vector

  virtual void Record( HCL_Vector_s & Input );

  /// Rewind - go to first vector recorded

  virtual void Rewind();

  /// Playback - copy stored data onto vector parameter

  virtual void PlayBack( HCL_Vector_s & Output );

  /// Next - advance to next recorded vector - returns 0 when at end

  virtual int Next();

  /// Number - how many vectors currently stored

  virtual int Number();

  /// Erase - deletes currently stored vectors;

  virtual void Erase();

};
*/
#endif







