/*********************************************************************
 This file was automatically generated by d2s from
 the file timer_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef TIMER_S_H
#define TIMER_S_H

#include "Std_includes.h"
#include <sys/times.h>

#ifndef CLK_TCK
#define   CLK_TCK sysconf(3)     /* 3 is _SC_CLK_TCK */
#endif

/** {\bf timer_s} is a simple class used for timing the execution of
code segments.  To use this class, define one or more real variables
to hold the total times, and initialize them to zero.  Then, to time
a segment of code, enclose it in braces, instantiate a timer_s object
at the beginning of the new scope, passing the appropriate variable to
the constructor.  At the end of the code segment delimited by the
braces, the timer_s object will be destroyed; its destructor increments
the variable with the elapsed time.
\begin{verbatim}

   // SolveTime will measure the total time used by the CG solver

   float SolveTime = 0.0;
   HCL_CG_s cg;

   // First solve

   {
      timer_s tme( &SolveTime );
      cg.Solve( A,b,x );
   }

   // ...

   // Another solve

   {
      timer_s tme( &SolveTime );
      cg.Solve( A,b,x );
   }

   // ... (Maybe some more solves)

   // Print out time spent in the CG solver

   cout << "Total time in CG: " << SolveTime << endl;
\end{verbatim}
*/

class timer_s {

 private:

   float start;
   float finish;
   float *user_ptr;

   float time(); // return current time in seconds. 

 public:

   /** The constructor takes a pointer to the (timing) variable to be
       incremented by the elapsed time. */   
   timer_s(float *dp) { 
      user_ptr = dp;
      start = time();
   }

   // Destructor.
   ~timer_s() {
      finish = time();

      *user_ptr += (finish-start);
   }

   /** {\bf mark} returns the time elapsed since construction of the timer
       object or since the last call to {\bf mark}. */
   float mark();/* {
      finish = time();
      float inc = finish-start;
      *user_ptr += inc;
      start = finish;
      
      return inc;
      }*/
   /** {\bf elapsed} returns the time elapsed since construction of the timer.
       */
   float elapsed() {
      finish = time();
      *user_ptr += (finish-start);
      start = finish;
      
      return *user_ptr;
   }

};

#endif
