/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_MatTR_d.C.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#include "HCL_MatTR_s.h"

HCL_MatTR_s::HCL_MatTR_s( const HCL_MatTR_s & x )
{
   cerr << "Error: HCL_MatTR_s copy constructor called" << endl;
   exit(1);
}

HCL_MatTR_s::HCL_MatTR_s( char * fname )
{
   if( fname )
      ParamTable = new Table( fname,"MatTR TRSolver" );
   else
      ParamTable = new Table;
}

HCL_MatTR_s::~HCL_MatTR_s()
{
   HCL_delete( ParamTable );
}

Table & HCL_MatTR_s::Parameters() const
{
   return *ParamTable;
}

void HCL_MatTR_s::Initialize()
{
   if( ParamTable->GetValue( "Delta",Delta ) )
      Delta = 1.0;
   if( ParamTable->GetValue( "Tol",Tol ) )
      Tol = 1.0e-2;
   if( ParamTable->GetValue( "Tol2",Tol2 ) )
      Tol2 = 0.0;
   if( ParamTable->GetValue( "UseLastLam",UseLastLam ) )
      UseLastLam = 0;
   if( ParamTable->GetValue( "DispFlag",DispFlag ) )
      DispFlag = 0;
   if( ParamTable->GetValue( "DispPrecision",DispPrecision ) )
      DispPrecision = 6;
   OldDispPrec = cout.precision( DispPrecision );
   Itn = 0;
   g = new HCL_RnVector_s( n );
   q = new HCL_RnVector_s( n );
   C = NULL;
   sqrtC = NULL;
}

void HCL_MatTR_s::Clean()
{
   cout.precision( OldDispPrec );
   if( C )
   {
      HCL_delete( p );
      HCL_delete( B );
   }
   HCL_delete( sqrtC );
   HCL_delete( C );
   HCL_delete( g );
   HCL_delete( q );
}

void HCL_MatTR_s::SetScaling( HCL_SymMatOp_s * S )
{
   C = S;S->IncCount();
//   sqrtC = C->SquareRoot();
}

void HCL_MatTR_s::UnSetScaling()
{
   HCL_delete( sqrtC );
   HCL_delete( C );
   C = NULL;
   sqrtC = NULL;
}

int HCL_MatTR_s::Solve( HCL_SymMatOp_s & Q,const HCL_RnVector_s & b,
                        HCL_RnVector_s & w )
{
   // Error checking

   if( Q.Domain() != w.Space() )
   {
      cerr << "Error in HCL_MatTR_s::Solve: vector and matrix are not "
              "consistent" << endl;
      exit(1);
   }
   if( Q.Range() != b.Space() )
   {
      cerr << "Error in HCL_MatTR_s::Solve: vector and matrix are not "
              "consistent" << endl;
      exit(1);
   }

   int i;
   n = w.Dim();
   Initialize();
   float Delta2 = Delta*Delta;

   // Modify the problem if scaling is being used.

//   if( !C )
//   {
      B = &Q;
      g->Copy( b );
      p = &w;
//   }
//   else  // This needs to be fixed.  I have only computed C^{-1/2}*Q.
//   {
//      p = new HCL_RnVector_s( n );
//      B = new HCL_SymMatOp_s( n );
//      sqrtC->SolvePositiveDefinite( b,*g );
//      B->Zero();
//      HCL_RnVector_s t1( n ),t2( n );
//      for( j=1;j<=n;j++ )
//      {
//         for( i=1;i<=n;i++ ) t1(i) = Q(i,j);
//         sqrtC->SolvePositiveDefinite( t1,t2 );
//         for( i=1;i<=j;i++ ) (*B)(i,j) = t2(i);
//      }
//   }

   // Negate g

   g->Neg();

   // Choose initial safeguards

//   lam_s = -(*B)(1,1);
//   for( i=2;i<=n;i++ ) lam_s = max( lam_s,-(*B)(i,i) );
// For now, compute and use -lam_1 for lam_s
   float * evals = new float[n];
   B->Matrix->eig( evals );
   lam_s = -evals[0];
   HCL_FPArch_s arch;
   if( abs(lam_s)/max(abs(evals[0]),abs(evals[n-1])) < 10*arch.eps )
      lam_s = -arch.eps;
   delete[] evals;
//   cout.precision(16);
//   for( i=1;i<=n;i++ )
//      cout << B->Eigenvalue(i) << endl;

   gnorm = g->Norm();
   Bnorm = B->Matrix->normOnes();
   lam_l = max( 0.0,lam_s );
   lam_l = max( lam_l,gnorm/Delta-Bnorm );

   lam_u = gnorm/Delta+Bnorm;

   if( UseLastLam )
      lam = LastLam;
   else
      lam = 0.0;   // Should refine this choice of lam0

   // Main loop

   TermCode = 0;
   while( !TermCode )
   {
      // Safeguard lam

      lam = max( lam,lam_l );
      lam = min( lam,lam_u );
      if( lam <= lam_s ) lam = max( float(0.001*lam_u),sqrt(lam_u*lam_l) );

      // Form B + lam*I

      GenSymMatrix_s Q1( *(B->Matrix) );
      for( i=1;i<=n;i++ ) Q1(i,i) += lam;
      HCL_SymMatOp_s B1( &Q1,0,&(B->Domain()),&(B->Range()) );

      // Check to see if B1 is positive definite

      HCL_UpperTriangularMatOp_s * R = B1.Cholesky();
      int pdflag = int(R);
      float phi;
      HCL_RnVector_s * zhat = NULL;
      if( pdflag > 0 )
      {
         // Solve B1p = -g (recall that g now holds -g)

         HCL_RnVector_s temp( n );
         R->InvAdjImage( *g,temp );
         R->InvImage( temp,*p );

         pnorm = p->Norm();
         pnorm2 = pnorm*pnorm;
         phi = 1.0/Delta - 1.0/pnorm;

         // if ||p|| < Delta, compute tau and zhat

         if( phi < 0.0 )
         {
            zhat = new HCL_RnVector_s( n );
            GeneralMatrix_s tmp1( n,zhat->Data() );
            R->Matrix->LINPACKSmallestSingularVectorEst( tmp1 );
            float pdotz = p->Inner( *zhat );
            float s = (pdotz>=0?1.0:-1.0);
            float t = Delta2-pnorm2;
            tau = t/(pdotz+s*sqrt(pdotz*pdotz+t));
            tau2 = tau*tau;
            HCL_RnVector_s tmp( n );
            R->Image( *zhat,tmp );
            sing = tmp.Norm();
            sing2 = sing*sing;
         }
      }
      else
         lam_s = max( lam_s,lam );

      // Update safeguards

      if( pdflag > 0 && phi < 0.0 )
         lam_u = min( lam_u,lam );
      else
         lam_l = max(lam_l,lam );
      // Eventually must implement update to lam_s from More'-Sorensen
//      if( pdflag > 0 && phi < 0.0 )
//         lam_s =max( lam_s,lam-sing2 );
//      else
//         ;
      lam_l = max( lam_l,lam_s );

      // Check convergence criteria

      Itn++;
      float err = abs( Delta-pnorm )/Delta;
      if( lam >= 0.0 && pdflag > 0 )
      {
         if( phi < 0.0 )
         {
            HCL_RnVector_s Rp( n );
            R->Image( *p,Rp );
            float Rpnorm = Rp.Norm();
            float Rp2 = Rpnorm*Rpnorm;
            if( tau2*sing2 <= Tol*(2.0-Tol)*
                max( Tol2,Rp2+lam*Delta2 ) && (err > Tol ||
                tau2*sing2 > lam*(Delta2-pnorm2)) )
            {
               p->AddScale( tau,*zhat );
               pnorm = p->Norm();
               err = abs( Delta-pnorm )/Delta;
               TermCode = BOUNDARY_SOLUTION;
            }
         }
         if( err <= Tol )
            TermCode = BOUNDARY_SOLUTION;
         else if( phi <= 0.0 && lam == 0.0 )
            TermCode = INTERIOR_SOLUTION;
      }
      HCL_delete( zhat );

      if( DispFlag > 1 || (DispFlag && TermCode) )
         cout << Itn << " lambda: " << lam << " ||p||: "
              << pnorm << " Relative error: " << err << endl;

      // Update lam

      if( pdflag > 0 )
      {
         R->InvAdjImage( *p,*q );
         float qnorm = q->Norm();
         float t = pnorm/qnorm;
         lam += t*t*(pnorm-Delta)/Delta;
      }
      else
         lam = lam_s;
      HCL_delete( R );
   }
   HCL_RnVector_s tmp( n );
   B->Image( *p,tmp );
   float red = -(0.5*p->Inner( tmp ) - g->Inner( *p )); // g holds -g
   ParamTable->PutValue( "Reduction",red );
   ParamTable->PutValue( "StepLength",pnorm );
   ParamTable->PutValue( "TermCode",TermCode );
   ParamTable->PutValue( "Multiplier",lam );

   // Change variables here, if necessary!!!!

   Clean();
   LastLam = lam;
   return TermCode;
}

ostream & HCL_MatTR_s::Write( ostream & str ) const
{
   str << "HCL_MatTR_s matrix-based trust region solver"
       << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}
