/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_PCG_d.C.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_PCG_s.h"

HCL_PCG_s::HCL_PCG_s()
{
   cerr << "Error: HCL_PCG_s default constructor called" << endl;
   exit(1);
}

HCL_PCG_s::HCL_PCG_s( const HCL_PCG_s & S )
{
   cerr << "Error: HCL_PCG_s copy constructor called" << endl;
   exit(1);
}

HCL_PCG_s::HCL_PCG_s( char *fname, 
		      HCL_LinearOp_s * m,
		      HCL_LinearSolver_s * ls,
		      HCL_Recorder_s * rec)
{
   MInv = new HCL_InvLinearOp_s(m, ls);
   ParamTable = new Table( fname, "LinearSolver LinearSolverPCG PCG", 8 );
   if (rec) {
     Rec = rec; Rec->IncCount();
   }
   else {
     Rec = (HCL_Recorder_s *)NULL;
   }
}

HCL_PCG_s::HCL_PCG_s( char *fname, 
		      HCL_LinearOp_s * minv, 
		      HCL_Recorder_s * rec)
{
   MInv = minv;  MInv->IncCount();
   ParamTable = new Table( fname, "LinearSolver LinearSolverPCG PCG", 8 );
   if (rec) {
     Rec = rec; Rec->IncCount();
   }
   else {
     Rec = (HCL_Recorder_s *)NULL;
   }
}

HCL_PCG_s::~HCL_PCG_s()
{
   HCL_delete( ParamTable );
   HCL_delete( MInv );
   if (Rec) {
     HCL_delete(Rec);
   }
}

int
HCL_PCG_s::Solve( const HCL_LinearOp_s & A,const HCL_Vector_s & b,
                             HCL_Vector_s & x ) const
{
   float total_t = 0;
   float user_t = 0, precond_t = 0;
   int user_count = 0, precond_count = 0;
   timer_s tme(&total_t);

   if( A.Domain() != MInv->Range() )
   {
      cerr << "Error in HCL_PCG_s::Solve: domains of A and M "
              "do not agree" << endl;
      exit(1);
   }
   if( A.Range() != MInv->Domain() )
   {
      cerr << "Error in HCL_PCG_s::Solve: ranges of A and M "
              "do not agree" << endl;
      exit(1);
   }
   if( A.Domain() != x.Space() )
   {
      cerr << "Error in HCL_PCG_s::Solve: x not in Domain of A" << endl;
      exit(1);
   }
   if( A.Range() != b.Space() )
   {
      cerr << "Error in HCL_PCG_s::Solve: b not in Range of A" << endl;
      exit(1);
   }

   // Get algorithmic parameters

   HCL_FPArch_s fp;

   float Tol;
   if( ParamTable->GetValue( "Tol",Tol ) )
      Tol = 1.0e-3;
   Tol = max(Tol, 500*fp.eps);

   int ItnMax;
   if( ParamTable->GetValue( "MaxItn",ItnMax ) )
      ItnMax = 100;
   int DispFlag;
   if( ParamTable->GetValue( "DispFlag",DispFlag ) )
      DispFlag = 0;
   
   if (DispFlag>1) {
      cout << endl << "================  HCL_PCG::Solve  ================" 
	   << endl;
   }

   // Allocate needed objects

   HCL_Vector_s *r = MInv->Domain().Member();
   HCL_Vector_s *u = MInv->Range().Member();
   HCL_Vector_s *p = MInv->Range().Member();
   HCL_Vector_s *v = MInv->Domain().Member();

   // compute initial residual

   {
      timer_s tme(&user_t);
      A.ImageAdd( x,b,*r,-1.0,1.0 );
   }
   ++user_count;

   // compute u = Minv*r

   {
      timer_s tme(&precond_t);
      MInv->Image(*r, *u);
   }
   ++precond_count;
   p->Copy( *u );

   // Main iteration

   float res0 = b.Norm();
   float res = r->Norm();
   //   float ratio = res/res0;
   float ratio;
   if (ProtectedDivision(res,res0,ratio)) {
     cerr<<"Error: HCL_PCG::Solve"<<endl;
     cerr<<"unsafe division of normal residual "<<res<<endl;
     cerr<<"by residual "<<res0<<endl;
     exit(1);
   }
   int itn = 0;
   if( DispFlag > 2 )
      cout << "HCL_PCG::Solve: iteration = "<< itn << " rel. error = " 
	   << ratio << endl;

   float rtu = r->Inner( *u );
   while( ratio > Tol && itn < ItnMax )
   {
      // compute v = Ap

      {
         timer_s tme(&user_t);
         A.Image( *p,*v );
      }
      ++user_count;

      // compute step length alpha

      float ptv = p->Inner( *v );
      //      float alpha = rtu/ptv;
      float alpha;
      if (ProtectedDivision(rtu,ptv,alpha)) {
	cerr<<"Error: HCL_PCG::Solve"<<endl;
	cerr<<"division defining CG parameter alpha is unsafe"<<endl;
	cerr<<"numerator   = "<<rtu<<endl;
	cerr<<"denominator = "<<ptv<<endl;
	exit(1);
      }
	  
      // update x and r

      x.AddScale( alpha,*p );
      r->AddScale( -alpha,*v );

      // record result

      if (Rec) {
	Rec->Record(x);
      }

      // compute u = Minv*r

      {
	 timer_s tme(&precond_t);
         MInv->Image(*r, *u);
      }
      ++precond_count;

      // compute the new relative residual

      res = r->Norm();
      //      ratio = res/res0;
      if (ProtectedDivision(res,res0,ratio)) {
	cerr<<"Error: HCL_PCG::Solve"<<endl;
	cerr<<"unsafe division of normal residual "<<res<<endl;
	cerr<<"by residual "<<res0<<endl;
	exit(1);
      }
      itn++;

      if( DispFlag > 2 )
	 cout << "HCL_PCG::Solve: iteration = "<< itn << " rel. error = " 
	      << ratio << endl;

      // compute beta and the new search direction p

      float rtu1 = r->Inner( *u );
      float beta = rtu1/rtu;
      p->ScaleAdd( beta,*u );
      rtu = rtu1;
   }
   
   HCL_delete( v );
   HCL_delete( p );
   HCL_delete( u );
   HCL_delete( r );

   int TermCode;

   if (ratio > Tol) {
      TermCode = TooManyIterations;
      if (DispFlag>0) {
	 cout << "HCL_PCG_s::Solve: Solve FAILED - Too many iterations - "
	         "rnorm = " << res << endl;
      }
   }
   else {
      TermCode = Success;
      if (DispFlag>0) {
	 cout << "HCL_PCG_s::Solve: Solve SUCCEEDED - rnorm = " << res 
	      << endl;
      }
   }

   if (DispFlag>1) {
      
      tme.mark();
      cout << "=============  HCL_PCG::Solve Summary ============" << endl;
      cout << "Residual norm:                       " << res << endl;
      cout << "Total number of iterations:          " << itn << endl;
      cout << "Total number of user A*x calls:      " << user_count << endl;
      cout << "Total number of user Inv[M]*x calls: " << precond_count 
           << endl;
      cout << "Total time in PCG iteration:         " << total_t << endl;
      cout << "Total time in user A*x routine:      " << user_t << endl;
      cout << "Total time in user Inv[M]*x routine: " << precond_t << endl;
      cout << "==================================================" << endl;
   }
   
   ParamTable->PutValue( "TermCode",TermCode );   
   ParamTable->PutValue( "ResidualNorm", res );
   
   return TermCode;
}

HCL_CG_s::HCL_CG_s( const HCL_CG_s & S )
{
   cerr << "Error: HCL_CG_s copy constructor called" << endl;
   exit(1);
}

HCL_CG_s::HCL_CG_s( char *fname, HCL_Recorder_s * rec )
{
   ParamTable = new Table( fname, "LinearSolver LinearSolverCG CG", 8 );
   if (rec) {
     Rec = rec; Rec->IncCount();
   }
   else {
     Rec = (HCL_Recorder_s *)NULL;
   }

}

HCL_CG_s::~HCL_CG_s()
{
   HCL_delete( ParamTable );
   if (Rec) {
     HCL_delete(Rec);
   }
}

int
HCL_CG_s::Solve( const HCL_LinearOp_s & A,const HCL_Vector_s & b,
                            HCL_Vector_s & x ) const
{
   float total_t = 0;
   float user_t = 0;
   int user_count = 0;
   timer_s tme(&total_t);

   if( A.Domain() != x.Space() )
   {
      cerr << "Error in HCL_CG_s::Solve: x not in Domain of A" << endl;
      exit(1);
   }
   if( A.Range() != b.Space() )
   {
      cerr << "Error in HCL_CG_s::Solve: b not in Range of A" << endl;
      exit(1);
   }
   

   // allocate the needed storage

   HCL_Vector_s *r = A.Range().Member();
   HCL_Vector_s *p = A.Domain().Member();
   HCL_Vector_s *v = A.Range().Member();

   // Get algorithmic parameters

   HCL_FPArch_s fp;

   float Tol;
   if( ParamTable->GetValue( "Tol",Tol ) )
      Tol = 1.0e-3;
   Tol = max(Tol, 500*fp.eps);

   int ItnMax;
   if( ParamTable->GetValue( "MaxItn",ItnMax ) )
      ItnMax = 100;
   int DispFlag;
   if( ParamTable->GetValue( "DispFlag",DispFlag ) )
      DispFlag = 0;
   
   if (DispFlag>1) {
      cout << endl << "=================  HCL_CG::Solve  ================" 
	   << endl;
   }

   // Compute initial residual

   {
      timer_s tme(&user_t);
      A.ImageAdd( x,b,*r,-1.0,1.0 );
   }
   ++user_count;

   p->Copy( *r );

   // Main iteration

   float res0 = b.Norm();
   float res = r->Norm();
   //   float ratio = res/res0;
   float ratio;
   if (ProtectedDivision(res,res0,ratio)) {
     cerr<<"Error: HCL_CG::Solve"<<endl;
     cerr<<"unsafe division of normal residual "<<res<<endl;
     cerr<<"by residual "<<res0<<endl;
     exit(1);
   }
   int itn = 0;
   
   if( DispFlag > 2 )
      cout << "HCL_CG::Solve: iteration = "<< itn << " rel. error = " 
	   << ratio << endl;

   float rtr = res*res;
   while( ratio > Tol && itn < ItnMax )
   {
      // compute v = Ap

      {
	 timer_s tme(&user_t);
	 A.Image( *p,*v );
      }
      ++user_count;

      // compute step length alpha

      float ptv = p->Inner( *v );
      //      float alpha = rtr/ptv;
      float alpha;
      if (ProtectedDivision(rtr,ptv,alpha)) {
	cerr<<"Error: HCL_CG::Solve"<<endl;
	cerr<<"division defining CG parameter alpha is unsafe"<<endl;
	cerr<<"numerator   = "<<rtr<<endl;
	cerr<<"denominator = "<<ptv<<endl;
	exit(1);
      }

      // update x and r

      x.AddScale( alpha,*p );
      r->AddScale( -alpha,*v );

      // record result

      if (Rec) {
	Rec->Record(x);
      }

      // compute the new relative residual

      res = r->Norm();
      //      ratio = res/res0;
      if (ProtectedDivision(res,res0,ratio)) {
	cerr<<"Error: HCL_CG::Solve"<<endl;
	cerr<<"unsafe division of normal residual "<<res<<endl;
	cerr<<"by residual "<<res0<<endl;
	exit(1);
      }
      itn++;
      if( DispFlag > 2 )
	 cout << "HCL_CG::Solve: iteration = "<< itn << " rel. error = " 
	      << ratio << endl;

      // compute beta and the new search direction p

      float rtr1 = res*res;
      float beta = rtr1/rtr;
      p->ScaleAdd( beta,*r );
      rtr = rtr1;
   }
   
   HCL_delete( v );
   HCL_delete( p );
   HCL_delete( r );

   int TermCode;
   if (ratio > Tol) {
      TermCode = TooManyIterations;
      if (DispFlag>0) {
	 cout << "HCL_CG_s::Solve: Solve FAILED - Too many iterations - "
	         "rnorm = " << res << endl;
      }
   }
   else {
      TermCode = Success;
      if (DispFlag>0) {
	 cout << "HCL_CG_s::Solve: Solve SUCCEEDED - rnorm = " << res 
	      << endl;
      }
   }

   if (DispFlag>1) {
      
      tme.mark();
      cout << "=============  HCL_CG::Solve Summary =============" << endl;
      cout << "Residual norm:                       " << res << endl;
      cout << "Total number of iterations:          " << itn << endl;
      cout << "Total number of user A*x calls:      " << user_count << endl;
      cout << "Total time in CG iteration:          " << total_t << endl;
      cout << "Total time in user A*x routine:      " << user_t << endl;
      cout << "==================================================" << endl;
   }
   
   ParamTable->PutValue( "TermCode",TermCode );   
   ParamTable->PutValue( "ResidualNorm", res );
   
   return TermCode;
}

ostream & HCL_PCG_s::Write( ostream & str ) const
{
   str << "HCL_PCG_s preconditioned conjugate gradient algorithm"
       << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}

ostream & HCL_CG_s::Write( ostream & str ) const
{
   str << "HCL_CG_s conjugate gradient algorithm"
       << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}


