/*********************************************************************
 This file was automatically generated by d2s from
 the file HCL_TRSlv_d.C.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_TrustRegionSolvers_s.h"

HCL_TRCG_s::HCL_TRCG_s( const HCL_TRCG_s & x )
{
   cerr << "Error: HCL_TRCG_s copy constructor called" << endl;
   exit(1);
}

Table & HCL_TRCG_s::Parameters() const
{
   return *ParamTable;
}

HCL_TRCG_s::HCL_TRCG_s( char * fname )
{
   if( fname != NULL )
      ParamTable = new Table( fname,"TRSolver" );
   else
      ParamTable = new Table;
   C = NULL;
   lsolver = NULL;   
}

HCL_TRCG_s::~HCL_TRCG_s()
{
   HCL_delete( ParamTable );
}

void HCL_TRCG_s::Initialize( HCL_VectorSpace_s & dom )
{
   if( Parameters().GetValue( "Delta",Delta ) )
   {
      Delta = 1.0;
      ParamTable->PutValue( "Delta",Delta );
   }
   if( Parameters().GetValue( "Tol",Tol ) )
   {
      Tol = 1.0e-2;
      ParamTable->PutValue( "Tol",Tol );
   }
   if( Parameters().GetValue( "MaxItn",MaxItn ) )
   {
      MaxItn = 20;
      ParamTable->PutValue( "MaxItn",MaxItn );
   }
   if( Parameters().GetValue( "DispFreq",DispFreq ) )
      DispFreq =  MaxItn/5;
   
   if( Parameters().GetValue( "DispFlag",DispFlag ) )
      DispFlag = 0;
   if( Parameters().GetValue( "DispPrecision",DispPrecision ) )
      DispPrecision = 6;
   OldDispPrec = cout.precision( DispPrecision );

   if( C != NULL )
   {
      if( C->Domain() != dom )
      {
         cerr << "Error in HCL_TRCG_s::Initialize: domain does not match"
                 " preconditioner" << endl;
         exit(1);
      }
      Cinvr = dom.Member();
   }
   else
      Cinvr = NULL;

   r = dom.Member();
   dir = dom.Member();
   Bdir = dom.Member();
   temp1 = dom.Member();
   temp2 = dom.Member();
}

void HCL_TRCG_s::Clean()
{
   cout.precision( OldDispPrec );
   HCL_delete( temp2 );
   HCL_delete( temp1 );
   HCL_delete( Bdir );
   HCL_delete( dir );
   HCL_delete( Cinvr );
   HCL_delete( r );
}

void HCL_TRCG_s::SetScaling( HCL_LinearOp_s * S,
                             HCL_LinearSolver_s * ls )
{
   HCL_delete( C );
   HCL_delete( lsolver );
   C = S;S->IncCount();
   lsolver = ls;if( ls ) ls->IncCount();
}
   
void HCL_TRCG_s::UnSetScaling()
{
   HCL_delete( C );
   HCL_delete( lsolver );
   C = NULL;
   lsolver = NULL;
}

void HCL_TRCG_s::Solve( const HCL_LinearOp_s & B,const HCL_Vector_s & g,
                              HCL_Vector_s & p )
{
   if( p.Space() != B.Domain() )
   {
      cerr << "Error in HCL_TRCG_s::Solve: vector does not belong to "
              "domain" << endl;
      exit(1);
   }
   if( g.Space() != B.Domain() )
   {
      cerr << "Error in HCL_TRCG_s::Solve: gradient does not belong "
              "to domain" << endl;
      exit(1);
   }

   // Initialize

   Initialize( B.Domain() );
   float gnorm;
   if( lsolver )
   {
      lsolver->Solve( *C,g,*temp2 );
      gnorm = sqrt(g.Inner( *temp2 ));
   }
   else if( C )
   {
      C->InvImage( g,*temp2 );
      gnorm = sqrt(g.Inner( *temp2 ));
   }
   else
      gnorm = g.Norm();
#if 0
   HCL_FPArch_s arch;

   // If the gradient (at zero) is zero or tiny, then generate
   // a random initial point; otherwise, the initial point is zero.

   if( gnorm < pow(arch.eps,1.5)*Delta )
   {
      if( DispFlag )
         cout << "HCL_TRCG_s: initial gradient too small; perturbing "
                 "initial point" << endl;
      p.Random();
      p.Mul( (0.1*Delta)/p.Norm() );
      B.Image( p,*r );
      r->Neg();
      gnorm = r->Norm();
   }
   else
   {
      p.Zero();
      r->Mul( -1.0,g );
   }
#endif
   p.Zero();
   r->Mul( -1.0,g );

   float eta;
   if( lsolver )
   {
      lsolver->Solve( *C,*r,*Cinvr );
      dir->Copy( *Cinvr );
      eta = r->Inner( *Cinvr );
   }
   else if( C )
   {
      C->InvImage( *r,*Cinvr );
      dir->Copy( *Cinvr );
      eta = r->Inner( *Cinvr );
   }
   else
   {
      dir->Copy( *r );
      eta = r->Norm2();
   }
   if( DispFlag > 2 ) cout << "   eta = " << eta << endl;
   int itn = 0;
   int done = 0;

   // Main loop

   if( DispFlag > 1 && itn%DispFreq == 0 )
      cout << "HCL_TRCG_s: iteration: " << itn << " relative residual: "
           << 1.0 << endl;
   float pnorm;
   while( !done )
   {

      // Check for direction of nonpositive curvature

      B.Image( *dir,*Bdir );
      float gamma = dir->Inner( *Bdir );
      if( DispFlag > 2 ) cout << "   gamma = " << gamma << endl;

      itn++;
      if( gamma <= 0.0 )
      {

         // Direction of nonpositive curvature found; ride it to the
         // boundary

         if( DispFlag > 2 )
         {
             cout << "Negative curvature: " << endl;
             cout << "p = " << endl;
             p.Write( cout );
             cout << "dir = " << endl;
             dir->Write( cout );
         }

         // Solve the quadratic equation giving the step length to the
         // boundary

         float a,b,c,tau;
         if( C != NULL )
         {
            C->Image( *dir,*temp1 );
            a = dir->Inner( *temp1 );
         }
         else
            a = dir->Norm2();
         if( C != NULL )
         {
            C->Image( p,*temp1 );
            b = 2.0*(dir->Inner( *temp1 ));
            c = p.Inner( *temp1 ) - Delta*Delta;
         }
         else
         {
            b = 2.0*(dir->Inner( p ));
            c = p.Norm2() - Delta*Delta;
         }
         tau = (-b+sqrt(b*b-4.0*a*c))/(2.0*a);

         // Step to the boundary and return

         p.AddScale( tau,*dir );
         B.Image( p,*temp1 );
         float red = -(0.5*(p.Inner( *temp1 )) + g.Inner( p ));
         temp1->Add( g );
         float lam = -p.Inner( *temp1 )/(Delta*Delta);
         ParamTable->PutValue( "Multiplier",lam );
         ParamTable->PutValue( "Reduction",red );
         ParamTable->PutValue( "TermCode",BOUNDARY_SOLUTION );
         ParamTable->PutValue( "StepLength",Delta );

         if( DispFlag )
            cout << "HCL_TRCG_s: " << itn
                 << " Encountered direction of negative curvature; "
	            "Reduction = " << red << endl;
         Clean();
         return;
      }

      // Compute next step

      float alpha = eta/gamma;
      if( DispFlag > 2 ) cout << "   alpha = " << alpha << endl;
      temp2->AddScale( alpha,p,*dir );

      // Check to see if the trust region has been exceeded

      if( C != NULL )
      {
         C->Image( *temp2,*temp1 );
         pnorm = sqrt(temp2->Inner( *temp1 ));
      }
      else
         pnorm = temp2->Norm();

      if( DispFlag > 2 ) cout << "   pnorm = " << pnorm << endl;

      if( pnorm >= Delta )
      {

         // Trust region has been exceeded; backtrack to the boundary

         // Solve the quadratic equation giving the step length to the
         // boundary

         float a,b,c,tau;
         if( C != NULL )
         {
            C->Image( *dir,*temp1 );
            a = dir->Inner( *temp1 );
         }
         else
            a = dir->Norm2();
         if( C != NULL )
         {
            C->Image( p,*temp1 );
            b = 2.0*(dir->Inner( *temp1 ));
            c = p.Inner( *temp1 ) - Delta*Delta;
         }
         else
         {
            b = 2.0*(dir->Inner( p ));
            c = p.Norm2() - Delta*Delta;
         }
         tau = (-b+sqrt(b*b-4.0*a*c))/(2.0*a);

         // Step to the boundary and return

         p.AddScale( tau,*dir );
         B.Image( p,*temp1 );
         float red = -(0.5*(p.Inner( *temp1 )) + g.Inner( p ));
         temp1->Add( g );
         float lam = -p.Inner( *temp1 )/(Delta*Delta);
         ParamTable->PutValue( "Multiplier",lam );
         ParamTable->PutValue( "Reduction",red );
         ParamTable->PutValue( "TermCode",BOUNDARY_SOLUTION );
         ParamTable->PutValue( "StepLength",Delta );

         if( DispFlag )
            cout << "HCL_TRCG_s: " << itn << " Step exceeded trust region; "
                    "Reduction = " << red << endl;
         Clean();
         return;

      }

      // Compute new residual and its scaled norm

      p.Copy( *temp2 );
      r->AddScale( -alpha,*Bdir );
      float res;
      if( lsolver )
      {
         lsolver->Solve( *C,*r,*Cinvr );
         float t = r->Inner( *Cinvr );
         res = sqrt(t);
      }
      if( C )
      {
         C->InvImage( *r,*Cinvr );
         float t = r->Inner( *Cinvr );
         res = sqrt(t);
      }
      else
         res = r->Norm();
      if( DispFlag > 1 && itn%DispFreq == 0 )
         cout << "HCL_TRCG_s: iteration: " << itn << " relative residual: "
              << res/gnorm << endl;
      if( res/gnorm < Tol )
      {
         B.Image( p,*temp1 );
         float red = -(0.5*(p.Inner( *temp1 )) + g.Inner( p ));
         ParamTable->PutValue( "Multiplier",0.0 );
         ParamTable->PutValue( "Reduction",red );
         ParamTable->PutValue( "TermCode",INTERIOR_SOLUTION );
         ParamTable->PutValue( "StepLength",pnorm );

         if( DispFlag )
            cout << "HCL_TRCG_s: " << itn
                 << " Approximate Newton step computed; "
	            "Reduction = " << red << endl;
         Clean();
         return;
      }
      else
      {
         float beta = res*res/eta;
         if( C != NULL )
            dir->ScaleAdd( beta,*Cinvr );
         else
            dir->ScaleAdd( beta,*r );
         eta = res*res;
      }

      if( itn >= MaxItn )
         done = 1;
   }
   B.Image( p,*temp1 );
   float red = -(0.5*(p.Inner( *temp1 )) + g.Inner( p ));
   ParamTable->PutValue( "Multiplier",0.0 );
   ParamTable->PutValue( "Reduction",red );
   ParamTable->PutValue( "TermCode",INTERIOR_SOLUTION );
   ParamTable->PutValue( "StepLength",pnorm );
   if( DispFlag )
      cout << "HCL_TRCG_s: " << itn << " Iteration limit reached; Reduction = "
           << red << endl;
   Clean();
   return;
}

ostream & HCL_TRCG_s::Write( ostream & str ) const
{
   str << "HCL_TRCG_s Steihaug-Toint trust region solver"
       << endl;
   str << "Parameter table:" << endl;
   return Parameters().Write( str );
}
