// Shannon Scott   5/27/1999
// Initial draft of a lower-triangular matrix class
// Data is stored column wise, starting with the left-most.
// Matrix must be square.
// Matrix indexing is one based.

#ifndef LOWER_H
#define LOWER_H

class LowerTriangle_d : public Matrix_d
{
  LowerTriangle_d(void);
public:
  LowerTriangle_d(int size);
  LowerTriangle_d(int i, int j);
  LowerTriangle_d(const LowerTriangle_d& copy);

  ~LowerTriangle_d(void);

  int zero(void)      { return setAll(0.0); };
  int ones(void)      { return setAll(1.0); };
  int diag(double val);
  int setAll(double val);
  int setDiag(double val);
  int random(void);

  double  operator()(int i, int j) const;
  double& operator()(int i, int j);

  double get(int i, int j) const;
  void   set(int i, int j, double val);
  
  int numRows(void) const { return size; };
  int numCols(void) const { return size; };

  friend ostream& operator<< (ostream&, const LowerTriangle_d&);

  int add(const LowerTriangle_d& B);
  int add(const LowerTriangle_d& A, const LowerTriangle_d& B);

  int solve(GeneralMatrix_d& X, const GeneralMatrix_d& B);
  int transposeSolve(GeneralMatrix_d& X, const GeneralMatrix_d& B);
  
  int extractLower(const Matrix_d& A);

  friend class GeneralMatrix_d;
protected:
  double* data;
  int size;
  int nAlloc;  // Number of doubles allocated to data
  double dummy;
};

//****************************************************************************
inline double LowerTriangle_d::operator()(int i, int j) const
{
  REQUIRE( i <= size, "Row index beyond matrix bounds." );
  REQUIRE( j <= size, "Column index beyond matrix bounds." );
  
  if( i < j )
    return 0.0;
  else
    return data[ i + ((2*size-j)*(j-1))/2 - 1 ];
};

//****************************************************************************
inline double& LowerTriangle_d::operator()(int i, int j)
{
  REQUIRE( i <= size, "Row index beyond matrix bounds." );
  REQUIRE( j <= size, "Column index beyond matrix bounds." );
  REQUIRE( i >= j, "These values must be zero." );
  
  return data[ i + ((2*size-j)*(j-1))/2 - 1 ];
};

//****************************************************************************
inline double LowerTriangle_d::get(int i, int j) const
{
  REQUIRE( i <= size, "Row index beyond matrix bounds." );
  REQUIRE( j <= size, "Column index beyond matrix bounds." );
  
  if( i < j )
    return 0.0;
  else
    return data[ i + ((2*size-j)*(j-1))/2 - 1 ];
};

//****************************************************************************
inline void LowerTriangle_d::set(int i, int j, double val)
{
  REQUIRE( i <= size, "Row index beyond matrix bounds." );
  REQUIRE( j <= size, "Column index beyond matrix bounds." );
  REQUIRE( i >= j, "These values must be zero." );
  
  data[ i + ((2*size-j)*(j-1))/2 - 1 ] = val;
};

#endif
