// Shannon Scott  6/18/1999
// Row-compressed sparse matrix class.
// sparse_rcs_d.H

#ifndef SPARSE_RCS_MATRIX_D_H
#define SPARSE_RCS_MATRIX_D_H

// Remember to add resize function

class SparseRCSMatrix_d : public Matrix_d
{
public:
  SparseRCSMatrix_d(void);
  SparseRCSMatrix_d(int i, int j);
  SparseRCSMatrix_d(int i, int j, int _nNonZero);
  SparseRCSMatrix_d(const SparseRCSMatrix_d& copy);

  SparseRCSMatrix_d(const Matrix_d& copy);
  SparseRCSMatrix_d(const Matrix_d& copy, Index rows, Index cols);

  SparseRCSMatrix_d(const GeneralMatrix_d&      A);
  SparseRCSMatrix_d(const UpperTriangle_d&      U);
  SparseRCSMatrix_d(const LowerTriangle_d&      L);
  SparseRCSMatrix_d(const Permutation&          P);
  SparseRCSMatrix_d(const BandedMatrix_d&       A);
  SparseRCSMatrix_d(const SymSparseRCSMatrix_d& A);
  
  ~SparseRCSMatrix_d(void);

  // Data access.  Does not change the dimensions of this.
  // Methods returning pointers allocate memory which must
  // be deleted by the user.
  double get(int i, int j) const;
  void   set(int i, int j, double val);

  double& element(int i, int j) { return operator()(i,j); };
  
  double  operator()(int i, int j) const;
  double& operator()(int i, int j);

  inline int numRows(void) const { return nRows; };
  inline int numCols(void) const { return nCols; };

  int set(const SparseRCSMatrix_d& B);

  int set(const GeneralMatrix_d& B);
  int set(const GeneralMatrix_d& B, Index rows, Index cols);
  int set(Index rowsA, Index colsA, const GeneralMatrix_d& B,
	  Index rowsB = Index(), Index colsB = Index());
  
  int set(const Matrix_d&);
  int set(const Matrix_d& B, Index rows, Index cols);
  int set(Index rowsA, Index colsA, const Matrix_d& B,
	  Index rowsB = Index(), Index colsB = Index());
    
  // Methods specific to sparse matrices
  int numNonZero(void) const { return nNonZero; };
  int numAlloc  (void) const { return nAlloc;   };
  void trim(void);                      // Decrease allocated space to
                                        // minimum possible
  void remove(int i, int j);            // Remove an entry (results
                                        // in the entry equalling zero
  void removeZeros(void);               // Removes any zeros
					// inadvertantly entered in matrix
  
  int ones(void)         { return -1; };
  int zero(void);
  int diag(double val)   { return -1; };
  int setAll(double val) { return -1; };

  // Add a single matrix
  int add(const GeneralMatrix_d&      A);
  int add(const SparseRCSMatrix_d&    A);
  int add(const SymSparseRCSMatrix_d& A);
  int add(const BandedMatrix_d&       A);

  // Add a transposed matrix
  int addTrans(const GeneralMatrix_d&      A);
  int addTrans(const SparseRCSMatrix_d&    A);
  int addTrans(const BandedMatrix_d&       A);

  // Add a scaled matrix to this matrix, C += a * A
  int add(const GeneralMatrix_d&      A, double a);
  int add(const SparseRCSMatrix_d&    A, double a);
  int add(const SymSparseRCSMatrix_d& A, double a);
  int add(const BandedMatrix_d&       A, double a);

  // Add a scaled transposed matrix to this: C += a * A'
  int addTrans(const GeneralMatrix_d&      A, double a);
  int addTrans(const SparseRCSMatrix_d&    A, double a);
  int addTrans(const BandedMatrix_d&       A, double a);

  // Multiply matrix by a scalar
  int mult(double a);
  
  // Add product of matrices to this, C += A * B
  int mult(const SparseRCSMatrix_d& A, const SparseRCSMatrix_d& B);

  // Multiply matrices, C = a*A*B + b*C
  int mult(const SparseRCSMatrix_d& A, const SparseRCSMatrix_d& B,
	   double a, double b);
  
  // Produce the product Y = C*X
  int image(const GeneralMatrix_d&   X, GeneralMatrix_d& Y) const;
  int image(const SparseRCSMatrix_d& X, GeneralMatrix_d& Y) const;

  //int cg(GeneralMatrix_d& X, const GeneralMatrix_d& B,
  // unsigned maxiter, double tol);
  
  int solve(GeneralMatrix_d& X, const GeneralMatrix_d& B);
  
  double normFrobenius(void);
  double normInf(void);
  double normOnes(void);

  void dump(ostream& os);

  double*   data;
  int* column;
  int* rowstart;
  int nAlloc;
  int nNonZero;
  
  friend ostream& operator<< (ostream& os, const SparseRCSMatrix_d& A);
  friend class GeneralMatrix_d;
protected:

  void resize(void);    // Resize the storage space
  
  int nRows;
  int nCols;
};

#endif
