// Shannon Scott
// 7/21/1999
// Symmetric sparse matrix class.  This uses the format supported by
// the Yale Sparse MAtrix Package.  It is a row-compressed format, in
// which only the upper triangle is stored.  I have basically copied
// the non-symmetric row-compressed matrix class with appropriate
// modifications
// Note: There is a slight difference in the way Yale Sparse Matrix
// Package stores its row_start array.  (I made the entries
// zero-based, fortran won't like that).  So, this will have to be
// changed to use YSMP.
// 8/10/1999, I have changed the rowstart array to be compatible with YSMP.

#ifndef SYM_SPARSE_RCS_MATRIX_D_H
#define SYM_SPARSE_RCS_MATRIX_D_H

class SymSparseRCSMatrix_d : public Matrix_d
{
  SymSparseRCSMatrix_d(void){};
public:
  SymSparseRCSMatrix_d(int i, int j);
  SymSparseRCSMatrix_d(int i, int j, int _nNonZero);
  SymSparseRCSMatrix_d(const SymSparseRCSMatrix_d& copy);

  SymSparseRCSMatrix_d(const Matrix_d& copy);
  SymSparseRCSMatrix_d(const Matrix_d& copy, Index rows, Index cols);

  SymSparseRCSMatrix_d(const GeneralMatrix_d& A);
  SymSparseRCSMatrix_d(const UpperTriangle_d& U);
  SymSparseRCSMatrix_d(const LowerTriangle_d& L);
  SymSparseRCSMatrix_d(const Permutation&     P);
  
  ~SymSparseRCSMatrix_d(void);

  double  operator()(int i, int j) const;
  double& operator()(int i, int j);

  int numRows(void) const { return nRows; };
  int numCols(void) const { return nCols; };

  double get(int i, int j) const;
  void   set(int i, int j, double val);
  
  int set(const SymSparseRCSMatrix_d& B);
  
  int set(const GeneralMatrix_d& B);
  int set(const GeneralMatrix_d& B, Index rows, Index cols);
  int set(Index rowsA, Index colsA, const GeneralMatrix_d& B,
	  Index rowsB = Index(), Index colsB = Index());
  
  int set(const Matrix_d&);
  int set(const Matrix_d& B, Index rows, Index cols);
  int set(Index rowsA, Index colsA, const Matrix_d& B,
	  Index rowsB = Index(), Index colsB = Index());
  
  // Methods specific to sparse matrices
  int numNonZero(void)       { return nNonZero; };
  int numAlloc  (void)       { return nAlloc;   };
  
  void trim(void);                      // Decrease allocated space to
                                        // minimum possible
  void remove(int i, int j);            // Remove an entry (results
                                        // in the entry equalling zero
  void removeZeros(void);               // Removes any zeros
					// inadvertantly entered in matrix
  
  int ones(void)         { return -1; };
  int zero(void);
  int diag(double val)   { return -1; };
  int setAll(double val) { return -1; };

  int add(const SymSparseRCSMatrix_d& A);

  //int mult(const GeneralMatrix_d& X, GeneralMatrix_d& B);
  //int mult(const SymSparseRCSMatrix_d& X, GeneralMatrix_d& B);

  int solve( GeneralMatrix_d& X, const GeneralMatrix_d& B) const;
  
  //double normFrobenius(void);
  //double normInf(void);
  //double normOnes(void);

  void dump(ostream& os);
  
  double*   data;
  int* column;
  int* rowstart;
  int nAlloc;
  int nNonZero;

  friend ostream& operator<< (ostream& os, const SymSparseRCSMatrix_d& A);
  
  friend class GeneralMatrix_d;
  friend class SparseRCSMatrix_d;
protected:

  void resize(void);    // Resize the storage space
  
  int nRows;
  int nCols;
};

#endif
