// Shannon Scott   6/3/1999
// First draft of an upper-triangular matrix class
// Data is stored column-wise
// Matrix must be square
// Matrix indexing is ones-based

#ifndef UPPER_D_H
#define UPPER_D_H

//#include "matrix.H"
//#include "m_error.H"

class UpperTriangle_d : public Matrix_d
{
  UpperTriangle_d(void);
public:
  UpperTriangle_d(int size);
  UpperTriangle_d(int i, int j);
  UpperTriangle_d(const UpperTriangle_d& copy);

  ~UpperTriangle_d(void);

  int zero(void)          { return setAll(0.0); };
  int ones(void)          { return setAll(1.0); };
  int setAll(double val);
  int diag(double val);
  int random(void);

  double  operator()(int i, int j) const;
  double& operator()(int i, int j);
  
  double get(int i, int j) const;  
  void   set(int i, int j, double val);
  
  int numRows(void) const { return size; };
  int numCols(void) const { return size; };

  int add(const UpperTriangle_d& B);
  int add(const UpperTriangle_d& A, const UpperTriangle_d& B);

  int solve(GeneralMatrix_d& X, const GeneralMatrix_d& B);
  int transposeSolve(GeneralMatrix_d& X, const GeneralMatrix_d& B);
  
  int extractUpper(const Matrix_d& A);

  void LINPACKSmallestSingularVectorEst( GeneralMatrix_d & x );

  friend ostream& operator<< (ostream& os, const UpperTriangle_d& A);
  
  friend class GeneralMatrix_d;
  friend class GenSymMatrix_d;
protected:
  double* data;
  int size;
  int nAlloc;
  double dummy;
};



inline double UpperTriangle_d::operator()(int i, int j) const
{
  REQUIRE( i <= size, "Row index beyond matrix bounds." );
  REQUIRE( j <= size, "Column index beyond matrix bounds." );
  if( i > j )
    return 0.0;
  else
    return data[ i + (j * (j-1)) / 2 - 1 ];
};


inline double& UpperTriangle_d::operator()(int i, int j)
{
  REQUIRE( i <= size, "Row index beyond matrix bounds." );
  REQUIRE( j <= size, "Column index beyond matrix bounds." );
  REQUIRE( i <= j, "Data in the lower triangle is not stored." );
  return data[ i + (j * (j-1)) / 2 - 1];
};


inline double UpperTriangle_d::get(int i, int j) const
{
  REQUIRE( i <= size, "Row index beyond matrix bounds." );
  REQUIRE( j <= size, "Column index beyond matrix bounds." );
  if( i > j )
    return 0.0;
  else
    return data[ i + (j * (j-1)) / 2 - 1 ];
};


inline void UpperTriangle_d::set(int i, int j, double val)
{
  REQUIRE( i <= size, "Row index beyond matrix bounds." );
  REQUIRE( j <= size, "Column index beyond matrix bounds." );
  REQUIRE( i <= j, "Data in the lower triangle is not stored." );
  data[ i + (j * (j-1)) / 2 - 1] = val;
};

#endif
