// Shannon Scott
// 7/15/1999

#include "mat_inc_d.H"
#include <fstream.h>

//****************************************************************************
GenSymMatrix_d::GenSymMatrix_d(void)
{
  ERROR( "This has not been implemented." );
}

//****************************************************************************
GenSymMatrix_d::GenSymMatrix_d(int _size)
{
  size = _size;
  nAlloc = (size * (size + 1)) / 2;
  data = new double[ nAlloc ];
}

//****************************************************************************
GenSymMatrix_d::GenSymMatrix_d(int i, int j)
{
  REQUIRE( i == j, "Matrix must be square." );
  size = i;
  nAlloc = (size * (size + 1)) / 2;
  data = new double[ nAlloc ];
}

//****************************************************************************
GenSymMatrix_d::GenSymMatrix_d(const GenSymMatrix_d& copy)
{
  size = copy.numRows();
  nAlloc = copy.nAlloc;
  data = new double[ nAlloc ];
  memcpy( data, copy.data, nAlloc * sizeof( double ) );
}

//****************************************************************************
GenSymMatrix_d::GenSymMatrix_d(char * fname)
{
  ifstream in( fname );
  if( !in )
  {
     cerr << "Error in GenSymMatrix_d constructor: cannot open file "
          << fname << endl;
     exit(1);
  }
  in >> size;
  nAlloc = size*(size+1)/2;
  data = new double[nAlloc];
  int i;
  for( i=0;i<nAlloc;i++ ) in >> data[i];
}

//****************************************************************************
GenSymMatrix_d::~GenSymMatrix_d(void)
{
  delete[] data;
}

double GenSymMatrix_d::normFrobenius(void) const
{
  double norm = 0.0;
  int i,j;
  for( i=1; i<= size;i++ )
  {
    for( j=1;j<i;j++ )
      norm += 2*(*this)(i,j);
    norm += (*this)(i,i);
  }
  return sqrt( norm );
}

//********************************************************************
double GenSymMatrix_d::normOnes(void) const
{
  double norm = 0.0;
  int i,j;
  for( j=1; j<=size; ++j)
    {
      double subnorm = 0.0;
      for( i=1; i<=size; ++i )
	{
	  subnorm += fabs( (*this)(i,j) );
	}
      if( subnorm > norm )
	norm = subnorm;
    }
  return norm;
}

//********************************************************************
double GenSymMatrix_d::normInf(void) const
{
  return normOnes();
}

//****************************************************************************
int GenSymMatrix_d::setAll(double val)
{
  for(int i=0; i < nAlloc; ++i)
    {
      data[i] = val;
    }
  return 0;
}

//****************************************************************************
int GenSymMatrix_d::random()
{
  int DIST = 2;  // uniform distribution in (-1,1)
#ifdef SINGLE_PRECISION
  F77NAME(slarnv)(&DIST, matrix_rnd_seed, &nAlloc, data);
#else
  F77NAME(dlarnv)(&DIST, matrix_rnd_seed, &nAlloc, data);
#endif

  return 0;
}


//****************************************************************************
int GenSymMatrix_d::eig( double * val ) const
{
   char jobz = 'N';
   char uplo = 'U';
   GenSymMatrix_d * t = (GenSymMatrix_d*)this;
   int n = t->size;
   double * ap = new double[n*(n+1)/2];
   int i;
   for( i=0;i<n*(n+1)/2;i++ ) ap[i] = data[i];
   double * work = new double[3*n];
   int info;

#ifdef SINGLE_PRECISION
   F77NAME(sspev)( &jobz,&uplo,&n,ap,val,(double*)NULL,&n,work,&info );
#else
   F77NAME(dspev)( &jobz,&uplo,&n,ap,val,(double*)NULL,&n,work,&info );
#endif
   if( info < 0 )
   {
      cerr << "Error in LAPACK routine dspev: info = " << info << endl
           << "(" << info << "-th argument had an illegal value)" << endl;
      exit(1);
   }
   else if( info > 0 )
   {
      cerr << "Error in LAPACK routine dspev: info = " << info << endl
           << "(algorithm failed to converge)" << endl;
      exit(1);
   }
   delete[] work;
   delete[] ap;
   return info;
}

//****************************************************************************
int GenSymMatrix_d::eig( double * val,GeneralMatrix_d & V ) const
{
   char jobz = 'V';
   char uplo = 'U';
   GenSymMatrix_d * t = (GenSymMatrix_d*)this;
   int n = t->size;
   double * ap = new double[n*(n+1)/2];
   int i;
   for( i=0;i<n*(n+1)/2;i++ ) ap[i] = data[i];
   double * work = new double[3*n];
   int info;

#ifdef SINGLE_PRECISION
   F77NAME(sspev)( &jobz,&uplo,&n,ap,val,V.data,&n,work,&info );
#else
   F77NAME(dspev)( &jobz,&uplo,&n,ap,val,V.data,&n,work,&info );
#endif
   if( info < 0 )
   {
      cerr << "Error in LAPACK routine dspev: info = " << info << endl
           << "(" << info << "-th argument had an illegal value)" << endl;
      exit(1);
   }
   else if( info > 0 )
   {
      cerr << "Error in LAPACK routine dspev: info = " << info << endl
           << "(algorithm failed to converge)" << endl;
      exit(1);
   }
   delete[] work;
   delete[] ap;
   return info;
}
   
//****************************************************************************
int GenSymMatrix_d::cholesky(UpperTriangle_d& U)
{
  REQUIRE( U.size == size, "Incorrect size for UpperTriangle_d");
  
  // Copy data from this matrix to U
  for(int i=0; i < nAlloc; ++i)
    U.data[i] = data[i];

  int info;
  int n = size;
  #ifdef SINGLE_PRECISION
  F77NAME(spptrf)( "U", &n, U.data, &info );
  #else
  F77NAME(dpptrf)( "U", &n, U.data, &info );
  #endif
  return info;
}

//****************************************************************************
int GenSymMatrix_d::solvePositiveDefinite(GeneralMatrix_d& X,
                                          const GeneralMatrix_d& B)
{
  REQUIRE( X.numRows() == size,
	   "Row dimension of X does not agree. ");
  REQUIRE( B.numRows() == size,
	   "Row dimension of B does not agree." );
  REQUIRE( B.numCols() == X.numCols(),
	   "Number of columns in X and B do not agree." );
  
  UpperTriangle_d U(size);
  int info;
  int n = size;
  int m = X.numCols();

  // Copy B to X (since B would bo overwritten if sent to LAPACK)
  X.set( B );
  
  info = cholesky(U);
  REQUIRE( info == 0, "Cholesky factorization failed." );

  #ifdef SINGLE_PRECISION
  F77NAME(spptrs)("U", &n, &m, U.data, X.data, &n, &info);
  #else
  F77NAME(dpptrs)("U", &n, &m, U.data, X.data, &n, &info);
  #endif
  
  return info;
}

int GenSymMatrix_d::solve(GeneralMatrix_d& X, const GeneralMatrix_d& B)
{
//  return solvePositiveDefinite( X,B );

  REQUIRE( X.numRows() == size,
	   "Row dimension of X does not agree. ");
  REQUIRE( B.numRows() == size,
	   "Row dimension of B does not agree." );
  REQUIRE( B.numCols() == X.numCols(),
	   "Number of columns in X and B do not agree." );
  
  UpperTriangle_d U(size);
  int info;
  int n = size;
  int m = X.numCols();

  // Copy B to X (since B would be overwritten if sent to LAPACK)
  X.set( B );
  
  int * ipiv = new int[n];
  double * adat = new double[nAlloc];
  for(int i=0; i < nAlloc; ++i)
    adat[i] = data[i];

  #ifdef SINGLE_PRECISION
  F77NAME(ssptrf)( "U", &n, adat, ipiv, &info );
  #else
  F77NAME(dsptrf)( "U", &n, adat, ipiv, &info );
  #endif
  REQUIRE( info == 0, "Factorization failed." );

  #ifdef SINGLE_PRECISION
  F77NAME(ssptrs)("U", &n, &m, adat, ipiv, X.data, &n, &info);
  #else
  F77NAME(dsptrs)("U", &n, &m, adat, ipiv, X.data, &n, &info);
  #endif

  delete[] adat;
  delete[] ipiv;

  return info;
}

//****************************************************************************
ostream& operator<< (ostream& os, const GenSymMatrix_d& A)
{
  int i, j;
  
  for(i=1; i <= A.size; ++i)     // iterate over rows
    {
      for(j=1; j <= A.size; ++j)
	{
	  os << A(i, j) << '\t';
	}
      os << endl;
    }

  return os;
}





