// Shannon Scott 5/25/1999
// Initial rough-draft of an index class, used to simulate
// colon notation in matrix algebra.
// Indexing is ones-based.  0 is used to denote either a reset state, or an
// unbounded range.

#include "index.H"

//************************************************************************
// The "everything" index.  This will simple step through each index from 
// 1 to the end of...
//************************************************************************
Index::Index(void)
{
  Start = 1;
  Stop = 0;
  Step = 1;
  Current = 0;
}

//***********************************************************************
Index::Index(int _start, int _step, int _stop)
{
  assert( _start <= _stop );
  assert( _step <= (_stop - _start) ); 

  Start = _start;
  Stop = _stop;
  Step = _step;
  Current = 0;
}

//*********************************************************************
int Index::get(int i)
{
  int retval = Start + (i-1) * Step;
  REQUIRE( retval <= Stop, "i out of bounds." );
  return retval;
}

//**********************************************************************
int Index::size(void)
{
  return (Stop - Start)/Step + 1;
}

//*********************************************************************
// Initialize the index object.  This does mostly bounds-checking.
//*********************************************************************
int Index::init(int lbound, int ubound)
{
  Current = 0;

  if( Start == 0 )
    Start = lbound;

  if( Stop == 0 )
    Stop = ubound;

  if( Start < lbound || Stop > ubound )
    return 1;
  else
    return 0;
}

//*********************************************************************
// This function is really not necessary.  Repeated calls to getNext()
// will result in the index reseting itself at appropriate times, but
// I thought this might make people feel safer.
//***********************************************************************
int Index::getFirst(void)
{
  Current = Start;
  return Current;
}

//**********************************************************************
// Take a step.  If the step goes past the end, set current to zero. If
// current is zero, loop-back to the start.
//*********************************************************************
int Index::getNext(void)
{
  if( Current == 0 )
    Current = Start;
  else
    {
      Current += Step;         // Take a step

      if( Current > Stop )     // Is it past the end?
	Current = 0;
    }

  return Current; 
}


