// Shannon Scott
// 5/27/1999, 7/16/99

#include "mat_inc_d.H"

//*******************************************************************
LowerTriangle_d::LowerTriangle_d(void)
{
  // I don't know what to do with the default constructor, yet.
  ERROR("This has not been implemented");
}

//*******************************************************************
LowerTriangle_d::LowerTriangle_d(int _size)
{
  size = _size;
  nAlloc = (size * (size + 1)) / 2; 
  data = new double[ nAlloc ];
}

//******************************************************************
LowerTriangle_d::LowerTriangle_d(int i, int j)
{
  REQUIRE( i == j, "Lower triangular matrix must be square.");
  size = i;
  nAlloc = (size * (size + 1)) / 2; 
  data = new double[ nAlloc ];
}

//******************************************************************
LowerTriangle_d::LowerTriangle_d(const LowerTriangle_d& copy)
{
  size = copy.size;
  nAlloc = (size * (size + 1)) / 2;
  data = new double[ nAlloc ];

  int i;
  for(i=0; i < nAlloc; ++i)
    data[i] = copy.data[i];
}

//*****************************************************************
LowerTriangle_d::~LowerTriangle_d(void)
{
  delete[] data;
}

//*****************************************************************
int LowerTriangle_d::diag(double val)
{
  int i, j, k=0;

  for(i=0; i < size; ++i)
    {
      data[k] = val;
      ++k;
      for(j=0; j < size-i-1; ++k, ++j)
	data[k] = 0.0;
    }
  return 0;
}

//*****************************************************************
int LowerTriangle_d::setAll(double val) 
{
  int i;
  for(i=0; i < nAlloc; ++i)
    data[i] = val;
  return 0;
}

//*****************************************************************
int LowerTriangle_d::setDiag(double val)
{
  int i, n;

  for(i=0, n=size;   i < (size*(size+1))/2;   i += n, n-- )
    {
      data[i] = val;
    }
  
  return 0;
}

//****************************************************************************
int LowerTriangle_d::random()
{
  int DIST = 2;  // uniform distribution in (-1,1)
#ifdef SINGLE_PRECISION
  F77NAME(slarnv)(&DIST, matrix_rnd_seed, &nAlloc, data);
#else
  F77NAME(dlarnv)(&DIST, matrix_rnd_seed, &nAlloc, data);
#endif

  return 0;
}

//*******************************************************************
ostream& operator<< (ostream& os, const LowerTriangle_d& A)
{
  int i, j;

  for(i=1; i <= A.size; ++i)     // Iterate over rows
    {
      for(j=1; j <= A.size; ++j) // Iterate over columns
	{
	  os << A(i,j) << '\t';
	}
      cout << endl;
    }

  return os;
}

//********************************************************************
int LowerTriangle_d::add(const LowerTriangle_d& A)
{
  REQUIRE( size == A.size, "Dimensions do not agree" );

  nAlloc = (size*(size+1))/2;
  int i;

  for(i=0; i < nAlloc; ++i)
    data[i] += A.data[i];
  return 0;
}

//********************************************************************
int LowerTriangle_d::add(const LowerTriangle_d& A, const LowerTriangle_d& B)
{
  REQUIRE( size == A.size, "Dimensions do not agree" );
  REQUIRE( size == B.size, "Dimensions do not agree" );

  int i;
  for(i=0; i < nAlloc; ++i)
    data[i] = A.data[i] + B.data[i];

  return 0;
}

//**********************************************************************
int LowerTriangle_d::solve(GeneralMatrix_d& X, const GeneralMatrix_d& B)
{
  REQUIRE( X.nRows == size, "Range of X does not agree.");
  REQUIRE( B.nRows == size, "Range of B does not agree.");
  REQUIRE( X.nCols == B.nCols, "Domain of X and B do not agree.");

  int info = 0;
  int n = size;
  
  // Copy B to X (may encapsulate this in a function later)
  if( &X != &B )
    {
      for(int i=0; i < B.nRows*B.nCols; ++i)
	X.data[i] = B.data[i];
    }

  // Solve the system
  #ifdef SINGLE_PRECISION
  F77NAME(stptrs)("L", "N", "N", &n, (int*)&(X.nCols),
		  data, X.data, &n, &info);
  #else
  F77NAME(dtptrs)("L", "N", "N", &n, (int*)&(X.nCols),
		  data, X.data, &n, &info);
  #endif

  REQUIRE( info == 0, "Solution of system failed");
  return info;
}

//**********************************************************************
int LowerTriangle_d::transposeSolve(GeneralMatrix_d& X,
                                   const GeneralMatrix_d& B)
{
  REQUIRE( X.nRows == size, "Range of X does not agree.");
  REQUIRE( B.nRows == size, "Range of B does not agree.");
  REQUIRE( X.nCols == B.nCols, "Domain of X and B do not agree.");

  int info = 0;
  int n = size;
  
  // Copy B to X (may encapsulate this in a function later)
  if( &X != &B )
    {
      for(int i=0; i < B.nRows*B.nCols; ++i)
	X.data[i] = B.data[i];
    }

  // Solve the system
  #ifdef SINGLE_PRECISION
  F77NAME(stptrs)("L", "T", "N", &n, (int*)&(X.nCols),
		  data, X.data, &n, &info);
  #else
  F77NAME(dtptrs)("L", "T", "N", &n, (int*)&(X.nCols),
		  data, X.data, &n, &info);
  #endif

  REQUIRE( info == 0, "Solution of system failed");
  return info;
}

//**************************************************************************
int LowerTriangle_d::extractLower(const Matrix_d& A)
{
  REQUIRE( A.numRows() == A.numCols(), "A is not square");
  REQUIRE( A.numRows() == size, "Dimensions do not agree");

  int i, j;

  for(j=1; j <= size; ++j)           // Iterate over columns
    {
      for(i=j; i <= size; ++i)
	{
	  data[size*(j-1)-((j-1)*(j-2))/2+(i-j)] = A.get(i,j); 
	}
    }
  
  return 0;
}







