#include "tensor_d.h"

General1Tensor_d::General1Tensor_d()
{
   cerr << "Error: General1Tensor_d default constructor not implemented"
        << endl;
   exit(1);
}

General1Tensor_d::General1Tensor_d( const General1Tensor_d & )
{
   cerr << "Error: General1Tensor_d copy constructor not implemented"
        << endl;
   exit(1);
}

General1Tensor_d::General1Tensor_d( int d1,double * d,int flag )
{
   if( d1 < 1 )
   {
      cerr << "Error in General1Tensor_d constructor: dimension "
              "must be positive" << endl;
      cerr << "   Input dimension was " << d1 << endl;
      exit(1);
   }
   size = d1;
   if( d )
   {
      data = d;
      DeleteDataFlag = flag;
   }
   else
   {
      data = new double[size];
      int i;
      for( i=0;i<size;i++ ) data[i] = 0.0;
      DeleteDataFlag = 1;
   }
}
   
General1Tensor_d::~General1Tensor_d()
{
   if( DeleteDataFlag )
      delete[] data;
}

double General1Tensor_d::Multiply( const General1Tensor_d & x ) const
{
   if( x.Size() != Size() )
   {
      cerr << "Error in General1Tensor_d::Multiply: input tensor is not "
              "compatible." << endl;
      cerr << "   Input tensor has size " << x.Size() << endl;
      cerr << "   This tensor has size " << Size() << endl;
      exit(1);
   }
   double s=0.0;
   int i;
   for( i=1;i<=size;i++ )
      s += x(i)*data[i-1];
   return s;
}

General2Tensor_d::General2Tensor_d()
{
   cerr << "Error: General2Tensor_d default constructor not implemented"
        << endl;
   exit(1);
}

General2Tensor_d::General2Tensor_d( const General2Tensor_d & )
{
   cerr << "Error: General2Tensor_d copy constructor not implemented"
        << endl;
   exit(1);
}

General2Tensor_d::General2Tensor_d( int d1,int d2,double * d,int flag )
{
   if( d1 < 1 || d2 < 1 )
   {
      cerr << "Error in General1Tensor_d constructor: dimensions "
              "must be positive" << endl;
      cerr << "   Input dimensions were " << d1 << ", " << d2 << endl;
      exit(1);
   }
   rows = d1;
   cols = d2;
   size = d1*d2;
   if( d )
   {
      data = d;
      DeleteDataFlag = flag;
   }
   else
   {
      data = new double[size];
      int i;
      for( i=0;i<size;i++ ) data[i] = 0.0;
      DeleteDataFlag = 1;
   }
}
   
General2Tensor_d::~General2Tensor_d()
{
   if( DeleteDataFlag )
      delete[] data;
}

int General2Tensor_d::Size( int i ) const
{
   if( i==1 ) return rows;
   else if( i==2 ) return cols;
   else
   {
      cerr << "Error in General2Tensor_d::Size: size of dimension "
           << i << " requested" << endl;
      exit(1);
   }
}

void General2Tensor_d::Multiply( int i1,const General1Tensor_d & x,
                                              General1Tensor_d & y ) const
{
   if( i1 < 1 || i1 > 2 )
   {
      cerr << "Error in General2Tensor_d::Multiply: product along dimension "
           << i1 << " requested" << endl;
      exit(1);
   }
   if( x.Size() != Size( i1 ) )
   {
      cerr << "Error in General2Tensor_d::Multiply: input tensor is not "
              "compatible." << endl;
      cerr << "   Input tensor has size " << x.Size() << endl;
      cerr << "   This tensor has size " << rows << "x" << cols << endl;
      exit(1);
   }
   if( y.Size() != Size( 3-i1 ) )
   {
      cerr << "Error in General2Tensor_d::Multiply: output tensor is not "
              "compatible." << endl;
      cerr << "   Output tensor has size " << y.Size() << endl;
      cerr << "   This tensor has size " << rows << "x" << cols << endl;
      exit(1);
   }
   if( i1==1 )
   {
      int i,j;
      for( j=1;j<=cols;j++ )
      {
         y(j) = 0.0;
         for( i=1;i<=rows;i++ )
            y(j) += (*this)(i,j)*x(i);
      }
   }
   else
   {
      int i,j;
      for( i=1;i<=rows;i++ )
      {
         y(i) = 0.0;
         for( j=1;j<=cols;j++ )
            y(i) += (*this)(i,j)*x(j);
      }
   }
}

General3Tensor_d::General3Tensor_d()
{
   cerr << "Error: General3Tensor_d default constructor not implemented"
        << endl;
   exit(1);
}

General3Tensor_d::General3Tensor_d( const General3Tensor_d & )
{
   cerr << "Error: General3Tensor_d copy constructor not implemented"
        << endl;
   exit(1);
}

General3Tensor_d::General3Tensor_d( int d1,int d2,int d3,double * d,int flag )
{
   if( d1 < 1 || d2 < 1 || d3 < 1 )
   {
      cerr << "Error in General1Tensor_d constructor: dimensions "
              "must be positive" << endl;
      cerr << "   Input dimensions were " << d1 << ", " << d2
           << ", " << d3 << endl;
      exit(1);
   }
   mats = d1;
   rows = d2;
   cols = d3;
   size = d1*d2*d3;
   matsize = d2*d3;
   if( d )
   {
      data = d;
      DeleteDataFlag = flag;
   }
   else
   {
      data = new double[size];
      int i;
      for( i=0;i<size;i++ ) data[i] = 0.0;
      DeleteDataFlag = 1;
   }
}
   
General3Tensor_d::~General3Tensor_d()
{
   if( DeleteDataFlag )
      delete[] data;
}

int General3Tensor_d::Size( int i ) const
{
   if( i==1 ) return mats;
   else if( i==2 ) return rows;
   else if( i==3 ) return cols;
   else
   {
      cerr << "Error in General3Tensor_d::Size: size of dimension "
           << i << " requested" << endl;
      exit(1);
   }
}

void General3Tensor_d::Multiply( int i1,int i2,const General2Tensor_d & A,
                                 General1Tensor_d & z ) const
{
   if( i1 < 1 || i1 > 3  || i2 < 1 || i2 > 3 || i1==i2 )
   {
      cerr << "Error in General3Tensor_d::Multiply: invalid product "
              "requested" << endl;
      cerr << "   Product along dimensions " << i1 << ", " << i2 
           << " requested" << endl;
      exit(1);
   }
   if( A.Size( 1 ) != Size( i1 ) )
   {
      cerr << "Error in General3Tensor_d::Multiply: input tensor is not "
              "compatible." << endl;
      cerr << "   Input tensor has size " << A.Size(1) << "x"
           << A.Size(2) << endl;
      cerr << "   This tensor has size " << mats << "x" << rows << "x"
           << cols << endl;
      exit(1);
   }
   if( A.Size( 2 ) != Size( i2 ) )
   {
      cerr << "Error in General3Tensor_d::Multiply: input tensor is not "
              "compatible." << endl;
      cerr << "   Input tensor has size " << A.Size(1) << "x"
           << A.Size(2) << endl;
      cerr << "   This tensor has size " << mats << "x" << rows << "x"
           << cols << endl;
      exit(1);
   }
   if( z.Size() != Size( 6-i1-i2 ) )
   {
      cerr << "Error in General3Tensor_d::Multiply: output tensor is not "
              "compatible." << endl;
      cerr << "   Output tensor has size " << z.Size() << endl;
      cerr << "   This tensor has size " << mats << "x" << rows << "x"
           << cols << endl;
      exit(1);
   }
   if( i1==1 && i2==2 )
   {
      int i,j,k;
      for( k=1;k<=cols;k++ )
      {
         z(k) = 0.0;
         for( i=1;i<=mats;i++ )
            for( j=1;j<=rows;j++ )
               z(k) += (*this)(i,j,k)*A(i,j);
      }
   }
   else if( i1==2 && i2==1 )
   {
      int i,j,k;
      for( k=1;k<=cols;k++ )
      {
         z(k) = 0.0;
         for( i=1;i<=mats;i++ )
            for( j=1;j<=rows;j++ )
               z(k) += (*this)(i,j,k)*A(j,i);
      }
   }
   else if( i1==1 && i2==3 )
   {
      int i,j,k;
      for( j=1;j<=rows;j++ )
      {
         z(j) = 0.0;
         for( i=1;i<=mats;i++ )
            for( k=1;k<=cols;k++ )
               z(j) += (*this)(i,j,k)*A(i,k);
      }
   }
   else if( i1==3 && i2==1 )
   {
      int i,j,k;
      for( j=1;j<=rows;j++ )
      {
         z(j) = 0.0;
         for( i=1;i<=mats;i++ )
            for( k=1;k<=cols;k++ )
               z(j) += (*this)(i,j,k)*A(k,i);
      }
   }
   else if( i1==2 && i2==3 )
   {
      int i,j,k;
      for( i=1;i<=mats;i++ )
      {
         z(i) = 0.0;
         for( j=1;j<=rows;j++ )
            for( k=1;k<=cols;k++ )
               z(i) += (*this)(i,j,k)*A(j,k);
      }
   }
   else if( i1==3 && i2==2 )
   {
      int i,j,k;
      for( i=1;i<=mats;i++ )
      {
         z(i) = 0.0;
         for( j=1;j<=rows;j++ )
            for( k=1;k<=cols;k++ )
               z(i) += (*this)(i,j,k)*A(k,j);
      }
   }
}

void General3Tensor_d::Multiply( int i1,const General1Tensor_d & z,
                                              General2Tensor_d & A ) const
{
   if( i1 < 1 || i1 > 3  )
   {
      cerr << "Error in General3Tensor_d::Multiply: product along dimension "
           << i1 << " requested" << endl;
      exit(1);
   }
   if( A.Size( 1 ) != Size( (i1*i1-5*i1+8)/2 ) )
   {
      cerr << "Error in General3Tensor_d::Multiply: output tensor is not "
              "compatible." << endl;
      cerr << "   Output tensor has size " << A.Size(1) << "x"
           << A.Size(2) << endl;
      cerr << "   This tensor has size " << mats << "x" << rows << "x"
           << cols << endl;
      exit(1);
   }
   if( A.Size( 2 ) != Size( (-i1*i1+3*i1+4)/2 ) )
   {
      cerr << "Error in General3Tensor_d::Multiply: output tensor is not "
              "compatible." << endl;
      cerr << "   Output tensor has size " << A.Size(1) << "x"
           << A.Size(2) << endl;
      cerr << "   This tensor has size " << mats << "x" << rows << "x"
           << cols << endl;
      exit(1);
   }
   if( z.Size() != Size( i1 ) )
   {
      cerr << "Error in General3Tensor_d::Multiply: output tensor is not "
              "compatible." << endl;
      cerr << "   Output tensor has size " << z.Size() << endl;
      cerr << "   This tensor has size " << mats << "x" << rows << "x"
           << cols << endl;
      exit(1);
   }
   if( i1==1 )
   {
      int i,j,k;
      for( j=1;j<=rows;j++ )
         for( k=1;k<=cols;k++ )
         {
            A(j,k) = 0.0;
            for( i=1;i<=mats;i++ )
               A(j,k) += (*this)(i,j,k)*z(i);
         }
   }
   else if( i1==2 )
   {
      int i,j,k;
      for( i=1;i<=mats;i++ )
         for( k=1;k<=cols;k++ )
         {
            A(i,k) = 0.0;
            for( j=1;j<=rows;j++ )
               A(i,k) += (*this)(i,j,k)*z(j);
         }
   }
   else if( i1==3 )
   {
      int i,j,k;
      for( i=1;i<=mats;i++ )
         for( j=1;j<=rows;j++ )
         {
            A(i,j) = 0.0;
            for( k=1;k<=cols;k++ )
               A(i,j) += (*this)(i,j,k)*z(k);
         }
   }
}

void General3Tensor_d::Multiply( int i1,int i2,const General1Tensor_d & x,
                       const General1Tensor_d & y,General1Tensor_d & z ) const
{
   if( i1 < 1 || i1 > 3  || i2 < 1 || i2 > 3 || i1==i2 )
   {
      cerr << "Error in General3Tensor_d::Multiply: invalid product "
              "requested" << endl;
      cerr << "   Product along dimensions " << i1 << ", " << i2 
           << " requested" << endl;
      exit(1);
   }
   if( x.Size() != Size( i1 ) )
   {
      cerr << "Error in General3Tensor_d::Multiply: input tensor is not "
              "compatible." << endl;
      cerr << "   Input tensor has size " << x.Size() << endl;
      cerr << "   This tensor has size " << mats << "x" << rows << "x"
           << cols << endl;
      exit(1);
   }
   if( y.Size() != Size( i2 ) )
   {
      cerr << "Error in General3Tensor_d::Multiply: input tensor is not "
              "compatible." << endl;
      cerr << "   Input tensor has size " << y.Size() << endl;
      cerr << "   This tensor has size " << mats << "x" << rows << "x"
           << cols << endl;
      exit(1);
   }
   if( z.Size() != Size( 6-i1-i2 ) )
   {
      cerr << "Error in General3Tensor_d::Multiply: output tensor is not "
              "compatible." << endl;
      cerr << "   Output tensor has size " << z.Size() << endl;
      cerr << "   This tensor has size " << mats << "x" << rows << "x"
           << cols << endl;
      exit(1);
   }
   if( i1==1 && i2==2 )
   {
      int i,j,k;
      for( k=1;k<=cols;k++ )
      {
         z(k) = 0.0;
         for( i=1;i<=mats;i++ )
            for( j=1;j<=rows;j++ )
               z(k) += (*this)(i,j,k)*x(i)*y(j);
      }
   }
   else if( i1==2 && i2==1 )
   {
      int i,j,k;
      for( k=1;k<=cols;k++ )
      {
         z(k) = 0.0;
         for( i=1;i<=mats;i++ )
            for( j=1;j<=rows;j++ )
               z(k) += (*this)(i,j,k)*x(j)*y(i);
      }
   }
   else if( i1==1 && i2==3 )
   {
      int i,j,k;
      for( j=1;j<=rows;j++ )
      {
         z(j) = 0.0;
         for( i=1;i<=mats;i++ )
            for( k=1;k<=cols;k++ )
               z(j) += (*this)(i,j,k)*x(i)*y(k);
      }
   }
   else if( i1==3 && i2==1 )
   {
      int i,j,k;
      for( j=1;j<=rows;j++ )
      {
         z(j) = 0.0;
         for( i=1;i<=mats;i++ )
            for( k=1;k<=cols;k++ )
               z(j) += (*this)(i,j,k)*x(k)*y(i);
      }
   }
   else if( i1==2 && i2==3 )
   {
      int i,j,k;
      for( i=1;i<=mats;i++ )
      {
         z(i) = 0.0;
         for( j=1;j<=rows;j++ )
            for( k=1;k<=cols;k++ )
               z(i) += (*this)(i,j,k)*x(j)*y(k);
      }
   }
   else if( i1==3 && i2==2 )
   {
      int i,j,k;
      for( i=1;i<=mats;i++ )
      {
         z(i) = 0.0;
         for( j=1;j<=rows;j++ )
            for( k=1;k<=cols;k++ )
               z(i) += (*this)(i,j,k)*x(k)*y(j);
      }
   }
}
