// Shannon Scott   6/3/1999
// First draft of an upper-triangular matrix class
// Data is stored column-wise
// Matrix must be square
// Matrix indexing is ones-based

#include "mat_inc_d.H"

//********************************************************************
UpperTriangle_d::UpperTriangle_d(void)
{
  ERROR("This has not been implemented");
}

//********************************************************************
UpperTriangle_d::UpperTriangle_d(int _size)
{
  size = _size;
  nAlloc = (size * (size + 1)) / 2;
  data = new double[ nAlloc ];
}

//********************************************************************
UpperTriangle_d::UpperTriangle_d(int i, int j)
{
  REQUIRE( i == j, "Upper Triangular matrix must be square.");
  size = i;
  nAlloc = (size * (size + 1)) / 2;
  data = new double[ nAlloc ];
}

//**********************************************************************
UpperTriangle_d::UpperTriangle_d(const UpperTriangle_d& copy)
{
  size = copy.size;
  nAlloc = copy.nAlloc;
  data = new double[ nAlloc ];
  memcpy(data, copy.data, nAlloc * sizeof(double));
}

//***********************************************************************
UpperTriangle_d::~UpperTriangle_d(void)
{
  delete[] data;
}

//***********************************************************************
int UpperTriangle_d::setAll(double val)
{
  int i;
  for(i=0; i < nAlloc; ++i)
    data[i] = val;

  return 0;
}

//**********************************************************************
int UpperTriangle_d::diag(double val)
{
  ERROR("This has not been implemented");
  return 1;
}

//****************************************************************************
int UpperTriangle_d::random()
{
  int DIST = 2;  // uniform distribution in (-1,1)
#ifdef SINGLE_PRECISION
  F77NAME(slarnv)(&DIST, matrix_rnd_seed, &nAlloc, data);
#else
  F77NAME(dlarnv)(&DIST, matrix_rnd_seed, &nAlloc, data);
#endif

  return 0;
}

//**********************************************************************
ostream& operator<< (ostream& os, const UpperTriangle_d& A)
{
  int i, j;
  
  for(i=1; i <= A.size; ++i)
    {
      for(j=1; j < i; ++j)
	os << "0\t";
      
      for(; j <= A.size; ++j)
	os << A.data[ i + (j * (j-1)) / 2 - 1 ] << '\t';

      os << endl;
    }

  //for(int i=0; i < (A.size*(A.size+1))/2; ++i)
  //  cout << A.data[i] << '\t';
  //cout << endl;

  return os;
}

//**********************************************************************
int UpperTriangle_d::add(const UpperTriangle_d& B)
{
  REQUIRE( B.size == size, "Dimensions do not agree" );
  
  for(int i=0; i < size; ++i)
    data[i] += B.data[i];

  return 0;
}

//**********************************************************************
int UpperTriangle_d::add(const UpperTriangle_d& A, const UpperTriangle_d& B)
{
  REQUIRE( A.size == B.size, "Dimensions of A and B do not agree." );
  REQUIRE( A.size == size, "Dimensions do not agree." );

  for(int i=0; i < nAlloc ; ++i)
    {
      data[i] = A.data[i] + B.data[i];
    }
  return 0;
}

//***********************************************************************
int UpperTriangle_d::solve(GeneralMatrix_d& X, const GeneralMatrix_d& B)
{
  REQUIRE( X.nRows == size, "Range of X does not agree.");
  REQUIRE( B.nRows == size, "Range of B does not agree.");
  REQUIRE( X.nCols == B.nCols, "Domain of X and B do not agree.");

  int info = 0;
  int n = size;
  
  // Copy B to X (may encapsulate this in a function later)
  if( &X != &B )
    {
      for(int i=0; i < B.nRows*B.nCols; ++i)
	X.data[i] = B.data[i];
    }

  // Solve the system
  #ifdef SINGLE_PRECISION
  F77NAME(stptrs)("U", "N", "N", &n, (int*)&(X.nCols),
		  data, X.data, &n, &info);
  #else
  F77NAME(dtptrs)("U", "N", "N", &n, (int*)&(X.nCols),
		  data, X.data, &n, &info);
  #endif

  REQUIRE( info == 0, "Solution of system failed");
  return info;
}

//***********************************************************************
int UpperTriangle_d::transposeSolve(GeneralMatrix_d& X,
                                    const GeneralMatrix_d& B)
{
  REQUIRE( X.nRows == size, "Range of X does not agree.");
  REQUIRE( B.nRows == size, "Range of B does not agree.");
  REQUIRE( X.nCols == B.nCols, "Domain of X and B do not agree.");

  int info = 0;
  int n = size;
  
  // Copy B to X (may encapsulate this in a function later)
  if( &X != &B )
    {
      for(int i=0; i < B.nRows*B.nCols; ++i)
	X.data[i] = B.data[i];
    }

  // Solve the system
  #ifdef SINGLE_PRECISION
  F77NAME(stptrs)("U", "T", "N", &n, (int*)&(X.nCols),
		  data, X.data, &n, &info);
  #else
  F77NAME(dtptrs)("U", "T", "N", &n, (int*)&(X.nCols),
		  data, X.data, &n, &info);
  #endif

  REQUIRE( info == 0, "Solution of system failed");
  return info;
}

//**************************************************************************
int UpperTriangle_d::extractUpper(const Matrix_d& A)
{
  REQUIRE( A.numRows() == A.numCols(), "A is not square");
  REQUIRE( A.numRows() == size, "Dimensions do not agree");

  int i, j;

  for(j=1; j <= size; ++j)       // Iterate over columns
    {
      for(i=1; i <= j; ++i)      // Iterate over rows
	{
	  data[ i + (j * (j-1)) / 2 - 1 ] = A.get(i,j);
	}
    }
  
  return 0;
}

void UpperTriangle_d::LINPACKSmallestSingularVectorEst( GeneralMatrix_d & x )
{
  REQUIRE( x.numCols() == 1, "Input matrix must be a column vector");
  REQUIRE( x.numRows() == size, "Input matrix has wrong number of rows");

   int i,j,n = size;
   double * w = new double[size];

   w[0] = 1.0/get(1,1);
   for( j=2;j<=n;j++ )
   {
      double s = 0.0;
      for( i=1;i<=j-1;i++ ) s += w[i-1]*get(i,j);
      double e = (s>=0?-1.0:1.0);
      w[j-1] = (e-s)/get(j,j);
   }

   GeneralMatrix_d y( size,w );
   solve( x,y );
   x.mult( 1.0/x.normFrobenius() );

   delete[] w;
}
