/*********************************************************************
 This file was automatically generated by d2s from
 the file banded_d.H.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

// Shannon Scott
// 7/16/1999
// Banded matrix using LAPACK compressed banded storage.
// An m x n matrix with ku super-diagonals and kl sub-diagonals is
// stored column-wise in a (ku+kl+1) x n matrix.  (Note that ku+kl+1
// is sufficient room to store the non-zero elements of each column).
// Element (i,j) is stored by ones-based indexing at (ku + 1 + i - j, j),
// or (ku + i - j, j - 1) for zero-based indexing, which translates to
// (ku + i - j + (ku + kl + 1) * (j - 1)) in a one-dimensional array
// (zero-based indexing).

#ifndef BANDED_S_H
#define BANDED_S_H

class BandedMatrix_s : public Matrix_s
{
  BandedMatrix_s(void){};
public:
  // Banded matrix m-by-n with ku super-diagonals and kl sub-diagonals
  BandedMatrix_s(int m, int n, int ku, int kl);
  BandedMatrix_s(const BandedMatrix_s& copy);

  ~BandedMatrix_s(void);

  float  operator()(int i, int j) const;
  float& operator()(int i, int j);

  float get(int i, int j) const;
  void   set(int i, int j, float val);

  int numRows(void) const { return nrows; };
  int numCols(void) const { return ncols; };

  int numSubDiag(void) const   { return kl; };
  int numSuperDiag(void) const { return ku; };

  int zero(void) { return setAll( 0.0 ); };
  int ones(void) { return setAll( 1.0 ); };
  int setAll(float val);
  
  friend ostream& operator<< (ostream&, const BandedMatrix_s&);

private:
  float* data;
  float dummy;
  int nrows, ncols;
  int ku, kl;
  int nAlloc;
};

//****************************************************************************
inline float BandedMatrix_s::operator()(int i, int j) const
{
  REQUIRE( i <= nrows, "Row index out of bounds." );
  REQUIRE( j <= ncols, "Column index out of bounds." );
  //return data[ ku + 1 + i - j + nrows * (j - 1) ];
  if( i >= macroMax( 1, j-ku ) && i <= macroMin( nrows, j+kl ) )
    return data[ ku + i - j + (ku + kl + 1) * (j - 1) ];
  else
    return 0;
}

//****************************************************************************
inline float& BandedMatrix_s::operator()(int i, int j)
{
  REQUIRE( i <= nrows, "Row index out of bounds." );
  REQUIRE( j <= ncols, "Column index out of bounds." );
  //return data[ ku + 1 + i - j + nrows * (j - 1) ];
  if( i >= macroMax( 1, j-ku ) && i <= macroMin( nrows, j+kl ) )
    return data[ ku + i - j + (ku + kl + 1) * (j - 1) ];
  else
    return dummy;
}

//****************************************************************************
inline float BandedMatrix_s::get(int i, int j) const
{
  REQUIRE( i <= nrows, "Row index out of bounds." );
  REQUIRE( j <= ncols, "Column index out of bounds." );
  //return data[ ku + 1 + i - j + nrows * (j - 1) ];
  if( i >= macroMax( 1, j-ku ) && i <= macroMin( nrows, j+kl ) )
    return data[ ku + i - j + (ku + kl + 1) * (j - 1) ];
  else
    return 0;
}

//****************************************************************************
inline void BandedMatrix_s::set(int i, int j, float val)
{
  REQUIRE( i <= nrows, "Row index out of bounds." );
  REQUIRE( j <= ncols, "Column index out of bounds." );
  //data[ ku + 1 + i - j + nrows * (j - 1) ] = val;
  if( i >= macroMax( 1, j-ku ) && i <= macroMin( nrows, j+kl ) )
    data[ ku + i - j + (ku + kl + 1) * (j - 1) ] = val;
}

#endif



