/*********************************************************************
 This file was automatically generated by d2s from
 the file tensor_d.h.  Do not edit. 
*********************************************************************/

#define SINGLE_PRECISION 1

#ifndef TENSOR_S_H_
#define TENSOR_S_H_

#include <stdlib.h>
#include <math.h>
#include <iostream.h>

/** {\bf General1Tensor_s} implements a simple (dense) 1-tensor, that is,
a vector $x_i$.  The basic operations are access to the entries and
computation of the dot product $x_iy_i$. */
class General1Tensor_s
{
   private:
      int size;
      int DeleteDataFlag;

   public:

      float * data;

   // Default constructor---disabled.
      General1Tensor_s();
   // Copy constructor---disabled.
      General1Tensor_s( const General1Tensor_s & );
   /** Constructor: requires the dimension and initializes the
       tensor to zero. Optional argments are a data array used to fill
       the entries of the tensor and a flag indicating whether the data
       array is to be deleted upon destruction. */
      General1Tensor_s( int d1,float * d=NULL,int flag=0 );
   // Destructor.
      ~General1Tensor_s();

   /** {\bf Size} provides access to the dimension of the tensor. */
      int Size() const { return size; }
   /** Access to the entries.  T(i) returns the ith entry of the
       tensor. (Indexing starts at 1.)  Value version for using entries. */
      float  operator()( int i ) const
      {
         return data[i-1];
      }
   /** Access to the entries.  T(i) returns the ith entry of the
       tensor. (Indexing starts at 1.)  Reference version for changing
       entries. */
      float & operator()( int i )
      {
         return data[i-1];
      }

   /** {\bf Multiply} computes the product of the 1-tensor with another
       1-tensor. */
      float Multiply( const General1Tensor_s & x ) const;
};

/** {\bf General2Tensor_s} implements a simple (dense) 2-tensor, that is,
a matrix $N_{ij}$.  The basic operations are access to the entries and
computation of the two single sums $N_{ij}x_i$, $N_{ij}y_j$. */
class General2Tensor_s
{
   private:
      int rows,cols;
      int size;
      int DeleteDataFlag;

   public:

      float * data;

   // Default constructor---disabled.
      General2Tensor_s();
   // Copy constructor---disabled.
      General2Tensor_s( const General2Tensor_s & );
   /** Constructor: requires the two dimensions and initializes the
       tensor to zero. Optional argments are a data array used to fill
       the entries of the tensor and a flag indicating whether the data
       array is to be deleted upon destruction. */
      General2Tensor_s( int d1,int d2,float * d=NULL,int flag=0 );
   // Destructor.
      ~General2Tensor_s();

   /** {\bf Size} provides access to the two dimensions of the tensor.
      T.Size(i) returns the number of entries in the ith dimension. */
      int Size( int i ) const;
   /** Access to the entries.  T(i,j) returns the (i,j) entry of the
       tensor. (Indexing starts at 1.)  Value version for using entries. */
      float  operator()( int i, int j ) const
      {
         return data[(j-1)*rows+i-1];
      }
   /** Access to the entries.  T(i,j) returns the (i,j) entry of the
       tensor. (Indexing starts at 1.)  Reference version for changing
       entries. */
      float & operator()( int i, int j )
      {
         return data[(j-1)*rows+i-1];
      }

   /** {\bf Multiply} computes the product of the 2-tensor with a 1-tensor.
       The flag i1 indicates which indices are to be summed.  Index i1
       (i1=1, 2) will be matched with the index of the 1-tensor. */
      void Multiply( int i1, const General1Tensor_s & x,
                                   General1Tensor_s & z ) const;
};

/** {\bf General3Tensor_s} implements a simple (dense) 3-tensor, that is,
a 3-dimensional array $N_{ijk}$.  The basic operations are access to the
entries and computation of the three float sums $N_{ijk}x_ky_j$,
$N_{ijk}x_kz_i$, and $N_{ijk}y_jz_i$ and the three single sums
$N_{ijk}x_i$, $N_{ijk}y_j$, and $N_{ijk}z_i$.  */
class General3Tensor_s
{
   private:
      int rows,cols,mats;
      int matsize,size;
      int DeleteDataFlag;

   public:

      float * data;

   // Default constructor---disabled.
      General3Tensor_s();
   // Copy constructor---disabled.
      General3Tensor_s( const General3Tensor_s & );
   /** Constructor: requires the three dimensions and initializes the
       tensor to zero. Optional argments are a data array used to fill
       the entries of the tensor and a flag indicating whether the data
       array is to be deleted upon destruction. */
      General3Tensor_s( int d1,int d2,int d3,float * d=NULL,int flag=0 );
   // Destructor.
      ~General3Tensor_s();

   /** {\bf Size} provides access to the three dimensions of the tensor.
      T.Size(i) returns the number of entries in the ith dimension. */
      int Size( int i) const;
   /** Access to the entries.  T(i,j,k) returns the (i,j,k) entry of the
       tensor. (Indexing starts at 1.)  Value version for using entries. */
      float  operator()( int i, int j,int k ) const
      {
         return data[(i-1)*matsize+(k-1)*rows+j-1];
      }
   /** Access to the entries.  T(i,j,k) returns the (i,j,k) entry of the
       tensor. (Indexing starts at 1.)  Reference version for changing
       entries. */
      float & operator()( int i, int j,int k )
      {
         return data[(i-1)*matsize+(k-1)*rows+j-1];
      }

   /** {\bf Multiply} computes the product of the 3-tensor with a 2-tensor.
       The flags i1,i2 indicate which indices are to be summed.  Index i1
       (i1=1, 2, or 3) will be matched with the first index of the 2-tensor,
       while index i2 (i2=1, 2, or 3, i2\neq i1) will be matched with the
       second index of the 2-tensor. */
      void Multiply( int i1,int i2,const General2Tensor_s & A,
                                         General1Tensor_s & z ) const;
   /** {\bf Multiply} computes the product of the 3-tensor with a 1-tensor.
       The flag i1 indicates which index is to be summed.  Index i1
       (i1=1, 2, or 3) will be matched with the index of the 1-tensor. */
      void Multiply( int i1,const General1Tensor_s & A,
                                  General2Tensor_s & z ) const;
   /** {\bf Multiply} computes the product of the 3-tensor with two 1-tensors.
       The flags i1,i2 indicate which indices are to be summed.  Index i1
       (i1=1, 2, or 3) will be matched with the index of the first 1-tensor,
       while index i2 (i2=1, 2, or 3, i1\neq i2) will be matched with the
       second 1-tensor. */
      void Multiply( int i1,int i2,const General1Tensor_s & x,
                     const General1Tensor_s & y,General1Tensor_s & z ) const;
};

#endif



