/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0, July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "lj.h"

double LennardJonesEnergy_d::Value1( const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in LennardJonesEnergy::Value1: input "
              "vector is not in domain" << endl;
      exit(1);
   }
   HCL_RnVector_d & c = (HCL_RnVector_d&)x;
   int n = dom->Dim();
   double val;
   F77NAME(fnlj)( n,c.Data(),val );
   return val;
}

void LennardJonesEnergy_d::Gradient1( const HCL_Vector_d & x,
                                            HCL_Vector_d & y ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in LennardJonesEnergy::Gradient1: input "
              "vector is not in domain" << endl;
      exit(1);
   }
   HCL_RnVector_d & c = (HCL_RnVector_d&)x;
   if( y.Space() != Domain() )
   {
      cerr << "Error in LennardJonesEnergy::Gradient1: output "
              "vector is not in domain" << endl;
      exit(1);
   }
   HCL_RnVector_d & g = (HCL_RnVector_d&)y;
   int n = dom->Dim();
   F77NAME(gradlj)( n,c.Data(),g.Data() );
}

HCL_LinearOp_d * LennardJonesEnergy_d::Hessian1(
                 const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in LennardJonesEnergy::Hessian1: input "
              "vector is not in domain" << endl;
      exit(1);
   }
   HCL_RnVector_d & c = (HCL_RnVector_d&)x;
   int n = dom->Dim();
   GeneralMatrix_d * h = new GeneralMatrix_d( n,n );
   F77NAME(hesslj)( n,c.Data(),h->data );
   HCL_GeneralMatOp_d * H = new HCL_GeneralMatOp_d( h,1,&(c.Space()),
                                                    &(c.Space()) );
   return H;
}

LennardJonesEnergy_d::LennardJonesEnergy_d()
{
   cerr << "Error: LennardJonesEnergy_d default constructor is not "
           "implemented" << endl;
   exit(1);
}

LennardJonesEnergy_d::LennardJonesEnergy_d( const LennardJonesEnergy_d & )
{
   cerr << "Error: LennardJonesEnergy_d copy constructor is not "
           "implemented" << endl;
   exit(1);
}

LennardJonesEnergy_d::LennardJonesEnergy_d( int n )
{
   dom = new HCL_RnSpace_d( 3*n-6 );
}

LennardJonesEnergy_d::~LennardJonesEnergy_d()
{
   HCL_delete( dom );
}

HCL_VectorSpace_d & LennardJonesEnergy_d::Domain() const
{
   return *dom;
}

ostream & LennardJonesEnergy_d::Write( ostream & str ) const
{
   str << "Lennard-Jones energy functional on " << (dom->Dim()+6)/3
       << " particles"
       << endl;
   str << "Domain is" << endl;
   return Domain().Write( str );
}
