/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0, July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "EllipticSolOp3_d.h"

EvalEllipticSolOp3_d::EvalEllipticSolOp3_d()
{
   cerr << "Error: EvalEllipticSolOp3_d default constructor not implemented"
        << endl;
   exit(1);
}

EvalEllipticSolOp3_d::EvalEllipticSolOp3_d( const EvalEllipticSolOp3_d & )
{
   cerr << "Error: EvalEllipticSolOp3_d copy constructor not implemented"
        << endl;
   exit(1);
}

EvalEllipticSolOp3_d::EvalEllipticSolOp3_d( HCL_RnSpace_d * d,
                                  HCL_RnSpace_d * r,
                                  int MM,int NN,int LL,double * NL,
                                  int * NP,int * EL,const HCL_RnVector_d & aa )
{
   dom = d;d->IncCount();
   rng = r;r->IncCount();
   M = MM;
   N = NN;
   L = LL;
   NodeList = new double[4*M];
   int i;
   for( i=0;i<4*M;i++ ) NodeList[i] = NL[i];
   NodePtrs = new int[N];
   for( i=0;i<N;i++ ) NodePtrs[i] = NP[i];
   ElList = new int[3*L];
   for( i=0;i<3*L;i++ ) ElList[i] = EL[i];

   HCL_RnVector_d & a = (HCL_RnVector_d&)aa;
   SymSparseRCSMatrix_d * K = new SymSparseRCSMatrix_d( N,N,5*N );
   HCL_RnVector_d b( N );
   F77NAME(assemble)( (int&)M,(int&)N,(int&)L,NodeList,NodePtrs,ElList,
                      a.Data(),K->rowstart,K->column,K->data,K->nNonZero,
                      K->nAlloc,b.Data() );
   Ka = new HCL_SymSparseRCSMatOp_d( K,1,rng );
   u = new HCL_RnVector_d( N );
   Ka->InvImage( b,*u );
   Bu = NULL;
}

EvalEllipticSolOp3_d::~EvalEllipticSolOp3_d()
{
   HCL_delete( Bu );
   HCL_delete( u );
   HCL_delete( Ka );
   delete[] ElList;
   delete[] NodePtrs;
   delete[] NodeList;
   HCL_delete( rng );
   HCL_delete( dom );
}

void EvalEllipticSolOp3_d::GetBu() const
{
   EvalEllipticSolOp3_d * t = (EvalEllipticSolOp3_d*)this;
   SparseRCSMatrix_d * B = new SparseRCSMatrix_d( N,M,5*M );
   F77NAME(assemble2)( (int&)(t->M),(int&)(t->N),(int&)(t->L),t->NodeList,
                      t->NodePtrs,t->ElList,
                      (t->u)->Data(),B->rowstart,B->column,B->data,B->nNonZero,
                      B->nAlloc );
   t->Bu = new HCL_SparseRCSMatOp_d( B,1,dom,rng );
}

void EvalEllipticSolOp3_d::SecondDerivImage( const HCL_Vector_d & dx1,
                                             const HCL_Vector_d & dx2,
                                                   HCL_Vector_d & dy ) const
{
   if( dx1.Space() != Domain() )
   {
      cerr << "Error in EvalEllipticSolOp3_d::SecondDerivImage: first "
              "input perturbation is not in Domain" << endl;
      cerr << "Space of first input perturbation" << endl;
      dx1.Space().Write( cerr );
      cerr << "Domain" << endl;
      Domain().Write( cerr );
      exit(1);
   }
   if( dx2.Space() != Domain() )
   {
      cerr << "Error in EvalEllipticSolOp3_d::SecondDerivImage: second "
              "input perturbation is not in Domain" << endl;
      cerr << "Space of second input perturbation" << endl;
      dx2.Space().Write( cerr );
      cerr << "Domain" << endl;
      Domain().Write( cerr );
      exit(1);
   }
   if( dy.Space() != Range() )
   {
      cerr << "Error in EvalEllipticSolOp3_d::SecondDerivImage: output "
              "perturbation is not in range" << endl;
      cerr << "Space of output perturbation" << endl;
      dy.Space().Write( cerr );
      cerr << "Range" << endl;
      Range().Write( cerr );
      exit(1);
   }
   HCL_RnVector_d & da = (HCL_RnVector_d&)dx1;
   HCL_RnVector_d & db = (HCL_RnVector_d&)dx2;
   HCL_RnVector_d & du = (HCL_RnVector_d&)dy;

   SymSparseRCSMatrix_d * K1 = new SymSparseRCSMatrix_d( N,N,5*N );
   HCL_RnVector_d b1( N );
   F77NAME(assemble)( (int&)M,(int&)N,(int&)L,NodeList,NodePtrs,ElList,
                      da.Data(),K1->rowstart,K1->column,K1->data,K1->nNonZero,
                      K1->nAlloc,b1.Data() );
   HCL_SymSparseRCSMatOp_d * Kda = new HCL_SymSparseRCSMatOp_d( K1,1,rng );

   HCL_RnVector_d du1( N ),du2( N ),tmp1( N ),tmp2( N );
   Kda->Image( *u,tmp1 );
   Ka->InvImage( tmp1,du1 );

   K1 = new SymSparseRCSMatrix_d( N,N,5*N );
   F77NAME(assemble)( (int&)M,(int&)N,(int&)L,NodeList,NodePtrs,ElList,
                      db.Data(),K1->rowstart,K1->column,K1->data,K1->nNonZero,
                      K1->nAlloc,b1.Data() );
   HCL_SymSparseRCSMatOp_d * Kdb = new HCL_SymSparseRCSMatOp_d( K1,1,rng );

   Kdb->Image( *u,tmp2 );
   Ka->InvImage( tmp2,du2 );

   Kda->Image( du2,tmp1 );
   Kdb->Image( du1,tmp2 );
   HCL_delete( Kdb );
   HCL_delete( Kda );

   tmp1.Add( tmp2 );

   Ka->InvImage( tmp1,du );
}

void EvalEllipticSolOp3_d::SecondDerivPartialAdjImage(const HCL_Vector_d & dx1,
                                             const HCL_Vector_d & dy,
                                                   HCL_Vector_d & dx2 ) const
{
   if( dx1.Space() != Domain() )
   {
      cerr << "Error in EllipticSolOpSecondDeriv1_d::"
              "SecondDerivPartialAdjImage: first " << endl
           << "input perturbation is not in Domain" << endl;
      cerr << "Space of first input perturbation" << endl;
      dx1.Space().Write( cerr );
      cerr << "Domain" << endl;
      Domain().Write( cerr );
      exit(1);
   }
   if( dx2.Space() != Domain() )
   {
      cerr << "Error in EllipticSolOpSecondDeriv1_d::"
              "SecondDerivPartialAdjImage: "
              "output perturbation is not in Domain" << endl;
      cerr << "Space of output perturbation" << endl;
      dx2.Space().Write( cerr );
      cerr << "Domain" << endl;
      Domain().Write( cerr );
      exit(1);
   }
   if( dy.Space() != Range() )
   {
      cerr << "Error in EllipticSolOpSecondDeriv1_d::"
              "SecondDerivPartialAdjImage: "
              "second input perturbation is not in range" << endl;
      cerr << "Space of second input perturbation" << endl;
      dy.Space().Write( cerr );
      cerr << "Range" << endl;
      Range().Write( cerr );
      exit(1);
   }
   HCL_RnVector_d & da = (HCL_RnVector_d&)dx1;
   HCL_RnVector_d & db = (HCL_RnVector_d&)dx2;
   HCL_RnVector_d & dv = (HCL_RnVector_d&)dy;

   if( !Bu ) GetBu();

   HCL_RnVector_d du( N ),tmp1( N ),tmp2( N ),tmp3( M ),tmp4( N );

   Ka->InvImage( dv,tmp1 );

   SymSparseRCSMatrix_d * K1 = new SymSparseRCSMatrix_d( N,N,5*N );
   HCL_RnVector_d b( N );
   F77NAME(assemble)( (int&)M,(int&)N,(int&)L,NodeList,NodePtrs,ElList,
                      da.Data(),K1->rowstart,K1->column,K1->data,K1->nNonZero,
                      K1->nAlloc,b.Data() );
   HCL_SymSparseRCSMatOp_d * Kda = new HCL_SymSparseRCSMatOp_d( K1,1,rng );
   Kda->Image( tmp1,tmp2 );
   Kda->Image( *u,tmp4 );
   Ka->InvImage( tmp4,du );
   HCL_delete( Kda );

   Ka->InvImage( tmp2,tmp4 );
   Bu->AdjImage( tmp4,db );

   SparseRCSMatrix_d B( N,M,5*M );
   F77NAME(assemble2)( (int&)M,(int&)N,(int&)L,NodeList,NodePtrs,ElList,
                      du.Data(),B.rowstart,B.column,B.data,B.nNonZero,
                      B.nAlloc );
   HCL_SparseRCSMatOp_d Bdu( &B,0,dom,rng );
   Bdu.AdjImage( tmp1,tmp3 );
   db.Add( tmp3 );
}

void EvalEllipticSolOp3_d::Image( HCL_Vector_d & y ) const
{
   if( y.Space() != Range() )
   {
      cerr << "Error in EllipticSolOp3_d::Image: output vector is not "
              "in range" << endl;
      cerr << "Space of output vector" << endl;
      y.Space().Write( cerr );
      cerr << "Range:"<< endl;
      Range().Write( cerr );
      exit(1);
   }
   y.Copy( *u );
}

HCL_LinearOp_d * EvalEllipticSolOp3_d::Deriv() const
{
   if( !Bu ) GetBu();

   HCL_InvLinearOp_d * Kinv = new HCL_InvLinearOp_d( Ka );
   HCL_ScaleLinearOp_d * Neg =
                       new HCL_ScaleLinearOp_d( &(Kinv->Range()),-1.0 );

   HCL_CompLinearOp_d * D = new HCL_CompLinearOp_d( &(Bu->Domain()),
                                                    &(Kinv->Range()),3 );
   D->SetNext( Bu );
   D->SetNext( Kinv );
   HCL_delete( Kinv );
   D->SetNext( Neg );
   HCL_delete( Neg );

   return D;
}

ostream & EvalEllipticSolOp3_d::Write( ostream & str ) const
{
   str << "EvalEllipticSolOp3_d with domain" << endl;
   Domain().Write( str );
   str << "and range" << endl;
   Range().Write( str );
   return str;
}

EllipticSolOp3_d::EllipticSolOp3_d()
{
   cerr << "Error: EllipticSolOp3_d default constructor not implemented"
        << endl;
   exit(1);
}

EllipticSolOp3_d::EllipticSolOp3_d( const EllipticSolOp3_d & )
{
   cerr << "Error: EllipticSolOp3_d copy constructor not implemented"
        << endl;
   exit(1);
}

EllipticSolOp3_d::EllipticSolOp3_d( char * fname )
{
   ifstream meshfile( fname );
   meshfile >> amin >> M >> N >> L;

   NodeList = new double[4*M];
   NodePtrs = new int[N];
   ElList = new int[3*L];

   int i;
   for( i=0;i<M;i++ )
      for( int j=0;j<4;j++ )
         meshfile >> NodeList[j*M+i];
   for( i=0;i<N;i++ )
      meshfile >> NodePtrs[i];
   for( i=0;i<L;i++ )
      for( int j=0;j<3;j++ )
         meshfile >> ElList[j*L+i];

   dom = new HCL_RnSpace_d( M );
   rng = new HCL_RnSpace_d( N );
}

EllipticSolOp3_d::~EllipticSolOp3_d()
{
   HCL_delete( rng );
   HCL_delete( dom );
   delete[] ElList;
   delete[] NodePtrs;
   delete[] NodeList;
}

double EllipticSolOp3_d::MaxStep( const HCL_Vector_d & x, 
                                 const HCL_Vector_d &dir) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in EllipticSolOp3_d::MaxStep: input point is not "
              "in domain" << endl;
      cerr << "Space of input point" << endl;
      x.Space().Write( cerr );
      cerr << "Domain" << endl;
      Domain().Write( cerr );
      exit(1);
   }
   if( dir.Space() != Domain() )
   {
      cerr << "Error in EllipticSolOp3_d::MaxStep: input direction is not "
              "in domain" << endl;
      cerr << "Space of input direction" << endl;
      dir.Space().Write( cerr );
      cerr << "Domain" << endl;
      Domain().Write( cerr );
      exit(1);
   }
   HCL_RnVector_d & a = (HCL_RnVector_d&)x;
   HCL_RnVector_d & da = (HCL_RnVector_d&)dir;

   HCL_FPArch_d arch;
   double alpha = arch.large;
   for( int i=1;i<=M;i++ )
      if( a(i) < amin )
         return -1.0;
      else if( da(i) < 0.0 )
         alpha = min( alpha,(amin-a(i))/da(i) );
   return alpha;
}

HCL_EvaluateOp_d * EllipticSolOp3_d::Evaluate( const HCL_Vector_d & x ) const
{
   if( x.Space() != Domain() )
   {
      cerr << "Error in EllipticSolOp3_d::Evaluate: input point is not "
              "in domain" << endl;
      cerr << "Space of input point" << endl;
      x.Space().Write( cerr );
      cerr << "Domain" << endl;
      Domain().Write( cerr );
      exit(1);
   }
   if( MaxStep( x,x ) < 0.0 )
   {
      cerr << "Error in EllipticSolOp3_d::Evaluate: input point is not "
              "in domain (though it belongs to the correct vector space)"
           << endl;
      exit(1);
   }
   HCL_RnVector_d & a = (HCL_RnVector_d&)x;
   return new EvalEllipticSolOp3_d( dom,rng,M,N,L,NodeList,NodePtrs,ElList,a );
}

ostream & EllipticSolOp3_d::Write( ostream & str ) const
{
   str << "EllipticSolOp3_d with domain" << endl;
   Domain().Write( str );
   str << "and range" << endl;
   return Range().Write( str );
}
