/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0, July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef ELLIPTICSOLOP3_D_H_
#define ELLIPTICSOLOP3_D_H_

#include "HCL_Op_d.h"
#include "HCL_Rn_d.h"
#include "HCL_MatrixOps_d.h"
#include "ffem.h"
#include "HCL_Adjoint_d.h"
#include "HCL_ScaleLinearOp_d.h"
#include "HCL_CompLinearOp_d.h"

class EvalEllipticSolOp3_d: public HCL_EvaluateOp_d
{
   private:
      HCL_RnSpace_d * dom;
      HCL_RnSpace_d * rng;
      int M,N,L;
      double * NodeList;
      int * NodePtrs;
      int * ElList;
      double amin;
      HCL_SymSparseRCSMatOp_d * Ka;
      HCL_SparseRCSMatOp_d * Bu;
      HCL_RnVector_d * u;

      void GetBu() const;

   protected:
      virtual void SecondDerivImage( const HCL_Vector_d & dx1,
                                     const HCL_Vector_d & dx2,
                                           HCL_Vector_d & dy ) const;
      virtual void SecondDerivPartialAdjImage(const HCL_Vector_d & dx1,
                                              const HCL_Vector_d & dy,
                                                    HCL_Vector_d & dx2 ) const;
   public:

   // Default constructor---disabled.
      EvalEllipticSolOp3_d();
   // Copy constructor---disabled.
      EvalEllipticSolOp3_d( const EvalEllipticSolOp3_d & );
   // Usual constructor.
      EvalEllipticSolOp3_d( HCL_RnSpace_d * d,HCL_RnSpace_d * r,
                                  int MM,int NN,int LL,double * NL,
                                  int * NP,int * EL,const HCL_RnVector_d & a );
   // Destructor
      virtual ~EvalEllipticSolOp3_d();

   /** {\bf Domain} returns a reference to the domain of the underlying
       operator. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom;; }
   /// {\bf Range} returns a reference to the range of the underlying operator.
      virtual HCL_VectorSpace_d & Range() const { return *rng; }

   /// {\bf Image} computes the image of the operator.
      virtual void Image( HCL_Vector_d & y ) const;
   /** {\bf Deriv} computes the derivative of the operator.  Note that
       this method constructs a {\bf HCL\_LinearOp\_d}, which the calling
       program is responsible for deleting with {\bf HCL\_delete}. */
      virtual HCL_LinearOp_d * Deriv() const;

   /** Debugging information. */
      virtual ostream & Write( ostream & str ) const;
};

class EllipticSolOp3_d: public HCL_Op_d
{
   private:
      HCL_RnSpace_d * dom;
      HCL_RnSpace_d * rng;
      int M,N,L;
      double * NodeList;
      int * NodePtrs;
      int * ElList;
      double amin;

   public:

   // Default constructor---disabled.
      EllipticSolOp3_d();
   // Copy constructor---disabled.
      EllipticSolOp3_d( const EllipticSolOp3_d & );
   // Usual constructor; need the name of a file containing the mesh data.
      EllipticSolOp3_d( char * fname );
   // Destructor.
      ~EllipticSolOp3_d();

  // Domain space access.
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
  // Range space access.
      virtual HCL_VectorSpace_d & Range() const { return *rng; }

  /* {\bf MaxStep} computes the longest vector from x in the direction dir
     that will lie in the domain of definition of the operator, i.e. it
     computes the greatest a such that x + a*dir lies in the domain. */
      virtual double MaxStep( const HCL_Vector_d & x, 
			      const HCL_Vector_d &dir) const;

   /** {\bf Evaluate} creates an evaluation object which knows how
       to compute the image, first derivative, and second derivative, along
       with all relevant adjoints, at a point. */ 
      virtual HCL_EvaluateOp_d * Evaluate( const HCL_Vector_d & x ) const;

  // Debugging information.
      virtual ostream & Write( ostream & str ) const;
};

#endif
